#ifndef MPC_CORRIDORZONE_HPP
#define MPC_CORRIDORZONE_HPP

#include <list>

namespace mpc
{

  typedef struct CORRIDOR_LAT_ITV
  {
    float lowB;
    float upB;
  } CorridorLatItv;

  typedef struct CORRIDOR_LONG_ITV
  {
    int minB;
    int maxB;
  } CorridorLongItv;

                                    /**************************************************/
                                    /*************** Class CorridorZone ***************/
                                    /**************************************************/

  /** \class CorridorZone
   *  \brief Topologic free zone defined on a navigation corridor
   */
  class CorridorZone
  {
  protected : 
    std::list <CorridorLatItv> listLatItv_;            /**< list of intervals */
    BoolObsMap* obsmap_;                               /**< obstacle map attached to the current corridor */

    PositionXYTh* inputPos_;                           /**< input position of the corridor */
    PositionXYTh* outputPos_;                          /**< output position of the corridor */
    
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    CorridorZone();

  public : 
    /**
     *  \brief init constructor
     */
    CorridorZone(BoolObsMap* obsmap);

    /**
     *  \brief destructor
     */
    virtual ~CorridorZone();


    /**************** Zone computation ****************/

  protected : 
    /*
     *  \brief compute the input position of the zone
     */
    void computeInputPos();

    /*
     *  \brief compute the output position of the zone
     */
    void computeOutputPos();

  public  :
    /**
     *  \brief get back the input position of the zone
     *  \return the input position of the zone
     */
    PositionXYTh getInputPos() const;

    /**
     *  \brief get back the input position of the zone
     *  \return the input position of the zone
     */
    PositionXYTh getOutputPos() const;
  };


                                    /**************************************************/
                                    /***************** Class ZoneGraph ****************/
                                    /**************************************************/

  /** \class ZoneGraph
   *  \brief Graph representing the topologic free zones defined on a navigation corridor
   */
  class ZoneGraph
  {
  };

};

#endif // MPC_CORRIDORZONE_HPP
