#include "BoundingShape.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /************** Class BoundingShape ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
BoundingShape::BoundingShape()
{
  this->type_ = UNDEFINED_BSHAPE;
}

/*
 *  \brief copy constructor
 *  \param bb : the bounding box to copy
 */
BoundingShape::BoundingShape(const BoundingShape& bb)
{
  this->type_ = bb.type_;
}

/*
 *  \brief destructor
 */
BoundingShape::~BoundingShape()
{
}


                                    /**************************************************/
                                    /**************** Class PointShape ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
PointShape::PointShape() : BoundingShape()
{
  this->type_ = POINT_BSHAPE;
}

/*
 *  \brief copy constructor
 *  \param bs : the bounding shape to copy
 */
PointShape::PointShape(const PointShape& ps) : BoundingShape(ps)
{
}

/*
 *  \brief destructor
 */
PointShape::~PointShape()
{
}


/*************** Display functions ****************/

/*
 *  \brief display the point shape's features
 */
void PointShape::display() const
{
  cout << "Point shape" << endl;
}

/*
 *  \brief draw the bounding shape in an openGL viewer
 */
void PointShape::toGL() const
{
}


                                    /**************************************************/
                                    /***************** Class BDisk ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
BDisk::BDisk() : PointShape()
{
  this->type_ = SPHERE_BSHAPE;
}

/*
 *  \brief copy constructor
 *  \param bs : the spherical bounding shape to copy
 */
BDisk::BDisk(const BDisk& bs) : PointShape(bs)
{
  this->radius_ = bs.radius_;
}

/*
 *  \brief init constructor
 *  \param r : radius of the spherical bounding shape
 */
BDisk::BDisk(double r)
{
  this->type_ = SPHERE_BSHAPE;
  this->radius_ = r;  
}

/*
 *  \brief destructor
 */
BDisk::~BDisk()
{
}


/*************** Display functions ****************/

/*
 *  \brief display the spherical bouding shape's features
 */
void BDisk::display() const
{
  cout << "Spherical bounding shape" << endl;
  cout << "Radius : " << this->radius_ << endl;
}

/*
 *  \brief draw the spherical bounding shape in an openGL viewer
 */
void BDisk::toGL() const
{
}


                                    /**************************************************/
                                    /***************** Class BPolygon *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
BPolygon::BPolygon() : BoundingShape()
{
  this->type_ = POLYGON_BSHAPE;
}

/*
 *  \brief copy constructor
 *  \param bs : the spherical bounding shape to copy
 */
BPolygon::BPolygon(const BPolygon& bp) : BoundingShape(bp)
{
}

/*
 *  \brief destructor
 */
BPolygon::~BPolygon()
{
}


/******************* Accessors ********************/

/*
 *  \brief getter : return the i-th position defining the bounds 
 *  \param i : index of the bound to return
 *  \return a reference to the i-th position defining the bounds 
 */
Position& BPolygon::getBound(int i) const
{
  cout << "BPolygon::getBound : Not implemented yet!" << endl;
  exit(1);
}

/*
 *  \brief getter : return the number of bounds of the polygon
 *  \return the number of bounds of the polygon 
 */
int BPolygon::getNbBounds() const
{
  cout << "BPolygon::getNbBounds : Not implemented yet!" << endl;
  exit(1);
}

/*
 *  \brief add a bound to the polygon
 *  \param the new position defining a bound of the polygon
 */
void BPolygon::addBound(const Position& p)
{
  cout << "BPolygon::addBound : Not implemented yet!" << endl;
  exit(1);
}

/*
 *  \brief delete all bounds of the polygon
 */
void BPolygon::clearBounds()
{
  cout << "BPolygon::clearBounds : Not implemented yet!" << endl;
  exit(1);
}


/*************** Display functions ****************/

/*
 *  \brief display the bounding polygon's features
 */
void BPolygon::display() const
{
  cout << "BPolygon::display() : Not implemented yet!" << endl;
  exit(1);
}

/*
 *  \brief draw the bounding polygon in an openGL viewer
 */
void BPolygon::toGL() const
{
  cout << "BPolygon::toGL : Not implemented yet!" << endl;
  exit(1);
}


                                    /**************************************************/
                                    /******************* Class BBox *******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
BBox::BBox()
{
  this->type_ = BOX_BSHAPE;
  this->lg() = 0;
  this->wd() = 0;
}

/*
 *  \brief copy constructor
 *  \param bs : the spherical bounding shape to copy
 */
BBox::BBox(const BBox& bb) : PointShape(bb)
{
  this->type_ = BOX_BSHAPE;
  this->lg() = bb.lg();
  this->wd() = bb.wd();
}

/*
 *  \brief init constructor
 *  \param l : length of the bounding box
 *  \param w : width of the bounding box
 */
BBox::BBox(const double& l,const double& w)
{
  this->type_ = BOX_BSHAPE;
  this->lg() = l;
  this->wd() = w;
}

/*
 *  \brief destructor
 */
BBox::~BBox()
{
}


//   /************* Obstruction functions **************/

//   virtual bool nearPosFrom(ConfigR* c, ConfigR& near, double& dist);       
//                                                       // return the nearest position of the obstacle from config c
//                                                       // return true if c is in the rectangle, return false else

/*************** Growing functions ****************/

/*
 *  \brief grows the dimension of the bouding box
 *  \param v : dimension to grow the box
 */  
void BBox::grows(const double v)
{
  this->lg() += v;
  this->wd() += v;
}


/*************** Display functions ****************/

/*
 *  \brief display the bounding sphere's features
 */
void BBox::display() const
{
  cout << "Bounding box" << endl;
  cout << "Length : " << this->lg() << endl;
  cout << "Width : " << this->wd() << endl;
}

/*
 *  \brief draw the bounding sphere in an openGL viewer
 */
void BBox::toGL() const
{
}


                                    /**************************************************/
                                    /*********** Static collision functions ***********/
                                    /**************************************************/

/*
 *  \brief collision test between two arbitrary bounding shapes
 *  \param p1 : absolute position of the first bounding shape
 *  \param bs1 : the first bounding shape
 *  \param p2 : absolute position of the second bounding shape
 *  \param bs2 : the second bounding shape
 *  \param nearP1 : nearest position from bs1 to bs2
 *  \param nearP2 : nearest position from bs2 to bs1
 *  \param dist : distance between nearP1 and nearP2
 *  \return true if both bounding shapes are in collision
 */
bool mpc::areInCollision(const PositionXY& p1,const BoundingShape& bs1,const PositionXY& p2,const BoundingShape& bs2,
		    PositionXY& nearP1, PositionXY& nearP2, double& dist)
{
  switch(bs1.type())
    {
    case UNDEFINED_BSHAPE : 
      return false;
      break;

    case POINT_BSHAPE : 
      switch(bs2.type())
	{
	case UNDEFINED_BSHAPE : 
	  return false;
	  break;
	case POINT_BSHAPE : 
	  return areInCollision(*((PositionXY*)(&p1)),*((PointShape*)(&bs1)),*((PositionXY*)(&p2)),*((PointShape*)(&bs2)),
				nearP1,nearP2,dist);
	  break;
	case SPHERE_BSHAPE : 
	  return areInCollision(*((PositionXY*)(&p2)),*((BDisk*)(&bs2)),*((PositionXY*)(&p2)),*((BDisk*)(&bs2)),
				nearP1,nearP2,dist);
	  break;
	case BOX_BSHAPE : 
	  return areInCollision(*((PositionXYTh*)(&p2)),*((BBox*)(&bs2)),*((PositionXYTh*)(&p1)),*((BDisk*)(&bs1)),
				nearP1,nearP2,dist);
	  break;
	default : 
	  cout << "mpc::areInCollision : bounding shape type not handled" << endl;
	  exit(1);
	  break;
	}
      break;

    case SPHERE_BSHAPE : 
      switch(bs2.type())
	{
	case UNDEFINED_BSHAPE : 
	  return false;
	  break;
	case POINT_BSHAPE : 
	  return areInCollision(*((PositionXY*)(&p1)),*((BDisk*)(&bs1)),*((PositionXY*)(&p2)),*((PointShape*)(&bs2)),
				nearP1,nearP2,dist);
	  break;
	case SPHERE_BSHAPE : 
	  return areInCollision(*((PositionXY*)(&p1)),*((BDisk*)(&bs1)),*((PositionXY*)(&p2)),*((BDisk*)(&bs2)),
				nearP1,nearP2,dist);
	  break;
	case BOX_BSHAPE : 
	  return areInCollision(*((PositionXYTh*)(&p2)),*((BBox*)(&bs2)),*((PositionXYTh*)(&p1)),*((BDisk*)(&bs1)),
				nearP1,nearP2,dist);
	  break;
	default : 
	  cout << "mpc::areInCollision : bounding shape type not handled" << endl;
	  exit(1);
	  break;
	}
      break;

    case BOX_BSHAPE : 
      switch(bs2.type())
	{
	case UNDEFINED_BSHAPE : 
	  return false;
	  break;
	case POINT_BSHAPE : 
	  return areInCollision(*((PositionXYTh*)(&p1)),*((BBox*)(&bs1)),*((PositionXYTh*)(&p2)),*((PointShape*)(&bs2)),
				nearP1,nearP2,dist);
	  break;
	case SPHERE_BSHAPE : 
	  return areInCollision(*((PositionXYTh*)(&p1)),*((BBox*)(&bs1)),*((PositionXYTh*)(&p2)),*((BDisk*)(&bs2)),
				nearP1,nearP2,dist);
	  break;
	case BOX_BSHAPE : 
	  return areInCollision(*((PositionXYTh*)(&p1)),*((BBox*)(&bs1)),*((PositionXYTh*)(&p2)),*((BBox*)(&bs2)),
				nearP1,nearP2,dist);
	  break;
	default : 
	  cout << "mpc::areInCollision : bounding shape type not handled" << endl;
	  exit(1);
	  break;
	}
      break;
    default : 
      cout << "mpc::areInCollision : bounding shape type not handled" << endl;
      exit(1);
      break;
    }

  return false;
}

/*
 *  \brief collision test between two point shapes
 *  \param p1 : absolute position of the first point shape
 *  \param bs1 : the first point shape
 *  \param p2 : absolute position of the second point shape
 *  \param bs2 : the second point shape
 *  \param nearP1 : nearest position from ps1 to ps2
 *  \param nearP2 : nearest position from ps2 to ps1
 *  \param dist : distance between nearP1 and nearP2
 *  \return true if both point shapes are in collision
 */
bool mpc::areInCollision(const PositionXY& p1,const PointShape& ps1,const PositionXY& p2,const PointShape& ps2,
		    PositionXY& nearP1, PositionXY& nearP2, double& dist)
{
  // compute the square distance between both bounding boxes
  double sqDist = p1.sqDist(p2);

  nearP1 = p1;
  nearP2 = p2;

  dist = sqrt(sqDist);

  if(dist == 0) return true;
  else return false;
}

/*
 *  \brief collision test between a point shape and a bouding sphere
 *  \param p1 : absolute position of the first point shape
 *  \param bs : the bounding sphere
 *  \param p2 : absolute position of the second point shape
 *  \param ps : the point shape
 *  \param nearP1 : nearest position from bs to ps
 *  \param nearP2 : nearest position from ps to bs
 *  \param dist : distance between nearP1 and nearP2
 *  \return true if the point shape and the bouding sphere are in collision
 */
bool mpc::areInCollision(const PositionXY& p1,const BDisk& bs,const PositionXY& p2,const PointShape& ps,
		    PositionXY& nearP1, PositionXY& nearP2, double& dist)
{
  // compute the square distance between both bounding boxes
  double distCenter = p1.dist(p2);
  nearP2 = p2;

  if(distCenter == 0)
    {
      nearP1 = p2;
      dist = -bs.radius();
      return true;
    }
  else
    {
      // compute the vector \overrightarrow{p1 p2}
      PositionXY p1p2 = p2 - p1;
      
      // compute the closest position from the border
      // of the sphere to the point ps
      PositionXY dirP1P2 = p1p2.normalized();
      nearP1 = p1 + (dirP1P2*bs.radius());

      dist = distCenter - bs.radius();

      if(dist < 0) return true;
      else return false;
    }
}

/*
 *  \brief collision test between two bounding spheres
 *  \param p1 : absolute position of the first bounding sphere
 *  \param bs1 : the first bounding sphere
 *  \param p2 : absolute position of the second bounding sphere
 *  \param bs2 : the second bounding sphere
 *  \param nearP1 : nearest position from bs1 to bs2
 *  \param nearP2 : nearest position from bs2 to bs1
 *  \param dist : distance between nearP1 and nearP2
 *  \return true if both bounding spheres are in collision
 */
bool mpc::areInCollision(const PositionXY& p1,const BDisk& bs1,const PositionXY& p2,const BDisk& bs2,
		    PositionXY& nearP1, PositionXY& nearP2, double& dist)
{
  // compute the square distance between both bounding boxes
  double distCenter = p1.dist(p2);

  if(distCenter == 0)
    {
      nearP1 = p1;
      nearP2 = p2;
      dist = max(bs1.radius(),bs2.radius());
      return true;
    }
  else
    {
      // compute the vector \overrightarrow{p1 p2}
      PositionXY p1p2 = p2 - p1;
      
      // compute the closest position from 
      // both borders of the spheres
      PositionXY dirP1P2 = p1p2.normalized();
      nearP1 = p1 + (dirP1P2*bs1.radius());
      nearP2 = p2 + (dirP1P2*(-bs2.radius()));

      dist = distCenter - bs1.radius() - bs2.radius();
      
      if(dist < 0)
	return true;
      else 
	return false;
    }
}

/*
 *  \brief collision test between a point shape and a bouding box
 *  \param p1 : absolute position of the bounding box
 *  \param bb : the bounding box
 *  \param p2 : absolute position of the point shape
 *  \param ps : the point shape
 *  \param nearP1 : nearest position from bb to ps
 *  \param nearP2 : nearest position from ps to bb
 *  \param dist : distance between nearP1 and nearP2
 *  \return true if the point shape and the bouding box are in collision
 */
bool mpc::areInCollision(const PositionXYTh& p1,const BBox& bb,const PositionXYTh& p2,const PointShape& ps,
		    PositionXY& nearP1, PositionXY& nearP2, double& dist)
{
  PositionXYTh posCinF = coordInFrame(p2, p1);

  double X = posCinF.x();
  double Y = posCinF.y();

  bool inside = false;

  double lg = bb.lg();
  double wd = bb.wd();

  if(X <= -0.5*lg)
    {
      if(Y <= -0.5*wd)
	{
	  nearP1 = bb.getLLCorner(p1);
	}
      else if(Y > -0.5*wd && Y < 0.5*wd)
	{
	  nearP1 = (PositionXY) coordFromFrame(PositionXYTh(-0.5*lg,Y,0.0),p1);
	}
      else // (Y >= 0.5*wd)
	{
	  nearP1 = bb.getULCorner(p1);
	}
    }
  else if(X > -0.5*lg && X < 0.5*lg)
    {
      if(Y <= -0.5*wd)
	{
	  nearP1 = coordFromFrame(PositionXYTh(X,-0.5*wd,0.0),p1);
	}
      else if(Y > -0.5*wd && Y < 0.5*wd)
	{
	  if(X < 0)
	    {
	      if(Y < 0)
		{
		  if(0.5*lg + X < 0.5*wd + Y) nearP1 = coordFromFrame(PositionXYTh(-0.5*lg,Y,0.0),p1);
		  else nearP1 = coordFromFrame(PositionXYTh(X,-0.5*wd,0.0),p1);
		}
	      else // Y >= 0
		{
		  if(0.5*lg + X < 0.5*wd - Y) nearP1 = coordFromFrame(PositionXYTh(-0.5*lg,Y,0.0),p1);
		  else nearP1 = coordFromFrame(PositionXYTh(X,0.5*wd,0.0),p1);
		}
	    }
	  else // X >= 0
	    {
	      if(Y < 0)
		{
		  if(0.5*lg - X < 0.5*wd + Y) nearP1 = coordFromFrame(PositionXYTh(0.5*lg,Y,0.0),p1);
		  else nearP1 = coordFromFrame(PositionXYTh(X,-0.5*wd,0.0),p1);
		}
	      else // Y >= 0
		{
		  if(0.5*lg - X < 0.5*wd - Y) nearP1 = coordFromFrame(PositionXYTh(0.5*lg,Y,0.0),p1);
		  else nearP1 = coordFromFrame(PositionXYTh(X,0.5*wd,0.0),p1);
		}
	    }
	  inside = true;
	}
      else // (Y >= 0.5*wd_)
	{
	  nearP1 = coordFromFrame(PositionXYTh(X,0.5*wd,0.0),p1);
	}
    }
  else // (X >= 0.5*lg_)
    {
      if(Y <= -0.5*wd)
	{
	  nearP1 = bb.getLRCorner(p1);
	}
      else if(Y > -0.5*wd && Y < 0.5*wd)
	{
	  nearP1 = coordFromFrame(PositionXYTh(0.5*lg,Y,0.0),p1);
	}
      else // (Y >= 0.5*wd_)
	{
	  nearP1 = bb.getURCorner(p1);
	}
    }

  // compute the distance between c and near
  double dx = p2.x() - nearP1.at(0);
  double dy = p2.y() - nearP1.at(1);
  dist = sqrt(dx*dx + dy*dy);

//   if(!inside) cout << "posCinF : "; posCinF.display(); cout << "  box : (" << lg << "," << wd << ")  dist : " << dist << endl;

  return inside;
}

/*
 *  \brief collision test between a bounding box and a bounding sphere
 *  \param p1 : absolute position of the bounding box
 *  \param bb : the bounding box
 *  \param p1 : absolute position of the bounding sphere
 *  \param bs : the bounding sphere
 *  \param nearP1 : nearest position from bb to bs
 *  \param nearP2 : nearest position from bs to bb
 *  \param dist : distance between nearP1 and nearP2
 *  \return true if the bounding box and the bouding sphere are in collision
 */
bool mpc::areInCollision(const PositionXYTh& p1,const BBox& bb,const PositionXYTh& p2,const BDisk& bs,
		    PositionXY& nearP1, PositionXY& nearP2, double& dist)
{
  // consider the sphere as a point
  areInCollision(p1,bb,p2,(const PointShape&)bs,nearP1,nearP2,dist);

  // compute the nearest point from the sphere
  if(dist == 0)
    {
      dist -= bs.radius();
      return true;
    }
  else
    {
      // compute the vector p2p1
      PositionXY vecP2P1 = p1 - p2;
      vecP2P1.normalize();
      
      // compute the nearest point from the sphere
      nearP2 = ((PositionXY)p2) + (vecP2P1*bs.radius());
      
      dist -= bs.radius();
      if(dist < 0) return true;
      else return false;      
    }
}

/*
 *  \brief collision test between two bounding boxes
 *  \param p1 : absolute position of the first bounding box
 *  \param bb1 : the first bounding box
 *  \param p2 : absolute position of the second bounding box
 *  \param bb2 : the second bounding box
 *  \param nearP1 : nearest position from bb1 to bb2
 *  \param nearP2 : nearest position from bb2 to bb1
 *  \param dist : distance between nearP1 and nearP2
 *  \return true if both bounding boxes are in collision
 */
bool mpc::areInCollision(const PositionXYTh& p1,const BBox& bb1,const PositionXYTh& p2,const BBox& bb2,
		    PositionXY& nearP1, PositionXY& nearP2, double& dist)
{
  cerr << "mpc::areInCollision(box,box) : Not implemented yet!" << endl;
  exit(1);
  return false;
}

/*
 *  \brief collision test between a bounding disk and a segment
 *  \param p1 : absolute position of the bounding disk
 *  \param bd : the bounding disk
 *  \param p1 : first end position of the segment
 *  \param p2 : second end position of the segment
 *  \param itsc1 : first intersection between the disk and the segment
 *  \param itsc2 : second intersection between the disk and the segment
 *  \param dist : closest distance from the segment to the disk (negative distance if they are in collision)
 *  \return the number of intersections (0, 1 or 2)
 */
int mpc::areInCollision(const PositionXYTh& pos,const BDisk& bd,const PositionXY& p1,const PositionXY& p2,
			PositionXY& itsc1, PositionXY& itsc2, double& dist)
{
  cerr << "mpc::areInCollision(disk,segment) : Not implemented yet!" << endl;
  exit(1);
}

/*
 *  \brief collision test between a bounding box and a segment
 *  \param p1 : absolute position of the bounding box
 *  \param bd : the bounding box
 *  \param p1 : first end position of the segment
 *  \param p2 : second end position of the segment
 *  \param itsc1 : first intersection between the disk and the segment
 *  \param itsc2 : second intersection between the disk and the segment
 *  \param dist : closest distance from the segment to the box (negative distance if they are in collision)
 *  \return the number of intersections (0, 1 or 2)
 */
int mpc::areInCollision(const PositionXYTh& pos,const BBox& bb,const PositionXY& p1,const PositionXY& p2,
		   PositionXY& itsc1, PositionXY& itsc2)
{
  // get the four points defining the box
  PositionXY ulc = bb.getULCorner(pos);
  PositionXY urc = bb.getURCorner(pos);
  PositionXY llc = bb.getLLCorner(pos);
  PositionXY lrc = bb.getLRCorner(pos);

  // check for each boundary of the box, if it intersect the segment
  int nbItsc = 0;

  PositionXY itscULUR;
  if(mpc::intersecSgmts(ulc,urc,p1,p2,itscULUR))
    {
      itsc1 = itscULUR;
      nbItsc++;
    }
  PositionXY itscULLL;
  if(mpc::intersecSgmts(ulc,llc,p1,p2,itscULLL))
    {
      if(nbItsc == 0) itsc1 = itscULLL;
      else itsc2 = itscULLL;
      nbItsc++;
    }
  PositionXY itscLLLR;
  if(mpc::intersecSgmts(llc,lrc,p1,p2,itscLLLR))
    {
      if(nbItsc == 0) itsc1 = itscLLLR;
      else if(nbItsc == 1) itsc2 = itscLLLR;
      else cerr << "mpc::areInCollision(box,segment) : Too many intersections!" << endl;
      nbItsc++;
    }
  PositionXY itscURLR;
  if(mpc::intersecSgmts(urc,lrc,p1,p2,itscURLR))
    {
      if(nbItsc == 0) itsc1 = itscURLR;
      else if(nbItsc == 1) itsc2 = itscURLR;
      else cerr << "mpc::areInCollision(box,segment) : Too many intersections!" << endl;
      nbItsc++;
    }
  return nbItsc;
}


/*
 *  \brief collision test between a bounding box and a segment
 *  \param p1 : absolute position of the bounding box
 *  \param bd : the bounding box
 *  \param p1 : first end position of the segment
 *  \param p2 : second end position of the segment
 *  \param itsc1 : first intersection between the disk and the segment
 *  \param itsc2 : second intersection between the disk and the segment
 *  \param dist : closest distance from the segment to the box (negative distance if they are in collision)
 *  \return the number of intersections (0, 1 or 2)
 *
 *  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Les calculs de distances doivent tre tests !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 */
int mpc::areInCollision(const PositionXYTh& pos,const BBox& bb,const PositionXY& p1,const PositionXY& p2,
			PositionXY& itsc1, PositionXY& itsc2, double& dist)
{
  int res = mpc::areInCollision(pos,bb,p1,p2,itsc1,itsc2);

  // compute the orthogonal projection of each vertex of the 
  // bounding box on the segment to see which is the closest one
  PositionXY ulc = bb.getULCorner(pos);
  PositionXY urc = bb.getURCorner(pos);
  PositionXY llc = bb.getLLCorner(pos);
  PositionXY lrc = bb.getLRCorner(pos);      
  
  double dUL = closeDistSgmt(ulc,p1,p2);
  double dUR = closeDistSgmt(urc,p1,p2);
  double dLL = closeDistSgmt(llc,p1,p2);
  double dLR = closeDistSgmt(lrc,p1,p2);

  if(res == 0)
    {
      // case where there is no intersection :       
      dist = min(dUL,min(dUR,min(dLL,dLR)));
    }
  else
    {
      // compute the closest distance from 
      // the center of the box to the segment
      cerr << "mpc::areInCollision(box,segment) : This minimal distance is not the good one!" << endl;
      dist = -min(dUL,min(dUR,min(dLL,dLR)));
    }

  return res;
}
