#include "ControlTime.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /**************** Class ControlTime ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
ControlTime::ControlTime() : Control()
{
}

/**
 *  \brief copy constructor
 *  \param c : the control-time to copy
 */
ControlTime::ControlTime(const ControlTime& c) : Control(c)
{
}

/**
 *  \brief init construtor (init a control of n parameters)
 *  \param n : the number of parameters of the control
 */
ControlTime::ControlTime(int n) : Control(n)
{
}

/**
 *  \brief destructor
 */
ControlTime::~ControlTime()
{
}


                                    /**************************************************/
                                    /*************** Class PM3DControlT ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PM3DControlT::PM3DControlT() : ControlTime(4)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
PM3DControlT::PM3DControlT(const PM3DControlT& c) : ControlTime(4)
{
  this->ax() = c.ax();
  this->ay() = c.ay();
  this->az() = c.az();
  this->t()  = c.t();
}

/**
 *  \brief init constructor
 *  \param acc : the acceleration control
 *  \param ss : the steering speed control
 *  \param t : the time of application of the control
 */
PM3DControlT::PM3DControlT(double ax,double ay,double az,double t) : ControlTime(4)
{
  this->ax() = ax;
  this->ay() = ay;
  this->az() = az;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
PM3DControlT::~PM3DControlT()
{
}


/*************** Display functions ****************/

/**
 *  \brief display the vector's coordinates
 */
void PM3DControlT::display() const
{
  cout << "(" << this->ax() << "," << this->ay() << "," << this->az() << "," << this->t() << ")";
}


                                    /**************************************************/
                                    /**************** Class CLControlT ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CLControlT::CLControlT() : ControlTime(3)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
CLControlT::CLControlT(const CLControlT& c) : ControlTime(3)
{
  this->acc() = c.acc();
  this->ss()  = c.ss();
  this->t()   = c.t();
}

/**
 *  \brief init constructor
 *  \param acc : the acceleration control
 *  \param ss : the steering speed control
 *  \param t : the time of application of the control
 */
CLControlT::CLControlT(double acc,double ss,double t) : ControlTime(3)
{
  this->acc() = acc;
  this->ss()  = ss;
  this->t()   = t;
}

/**
 *  \brief destructor
 */
CLControlT::~CLControlT()
{
}


/*************** Display functions ****************/

/**
 *  \brief display the vector's coordinates
 */
void CLControlT::display() const
{
  cout << "(" << this->acc() << "," << this->ss() << "," << this->t() << ")";
}


                                    /**************************************************/
                                    /**************** Class SSControlT ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SSControlT::SSControlT() : ControlTime(3)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
SSControlT::SSControlT(const SSControlT& c) : ControlTime(3)
{
  this->acc() = c.acc();
  this->aacc() = c.aacc();
  this->t()    = c.t();
}

/**
 *  \brief init constructor
 *  \param accx : the acceleration control on X
 *  \param accy : the acceleration control on Y
 *  \param aacc : the angular acceleration control
 *  \param t : the time of application of the control
 */
SSControlT::SSControlT(double acc,double aacc,double t) : ControlTime(3)
{
  this->acc() = acc;
  this->aacc() = aacc;
  this->t()    = t;
}

/**
 *  \brief destructor
 */
SSControlT::~SSControlT()
{
}


/*************** Display functions ****************/
  
/**
 *  \brief display the vector's coordinates
 */
void SSControlT::display() const
{
  cout << "(" << this->acc() << "," << this->aacc() << "," << this->t() << ")";
}

                                    /**************************************************/
                                    /**************** Class DDControlT ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DDControlT::DDControlT() : ControlTime(3)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
DDControlT::DDControlT(const DDControlT& c) : ControlTime(3)
{
  this->wracc() = c.wracc();
  this->wlacc() = c.wlacc();
  this->t()     = c.t();
}

/**
 *  \brief init constructor
 *  \param wracc : the right wheel acceleration control
 *  \param wlacc : the left wheel acceleration control
 *  \param t : the time of application of the control
 */
DDControlT::DDControlT(double wracc,double wlacc,double t) : ControlTime(3)
{
  this->wracc() = wracc;
  this->wlacc() = wlacc;
  this->t()     = t;
}

/**
 *  \brief destructor
 */
DDControlT::~DDControlT()
{
}


/*************** Display functions ****************/

/**
 *  \brief display the vector's coordinates
 */
void DDControlT::display() const
{
  cout << "(" << this->wracc() << "," << this->wlacc() << "," << this->t() << ")";
}

                                    /**************************************************/
                                    /**************** Class DD2ControlT ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DD2ControlT::DD2ControlT() : ControlTime(3)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
DD2ControlT::DD2ControlT(const DD2ControlT& c) : ControlTime(3)
{
  this->v()  = c.v();
  this->av() = c.av();
  this->t()  = c.t();
}

/**
 *  \brief init constructor
 *  \param acc : the linear acceleration control
 *  \param aacc : the angular acceleration control
 *  \param t : the time of application of the control
 */
DD2ControlT::DD2ControlT(double v,double av,double t) : ControlTime(3)
{
  this->v()  = v;
  this->av() = av;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
DD2ControlT::~DD2ControlT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input state
 *  \param s2 : the state to copy
 */
void DD2ControlT::operator=(const DD2ControlT& c2)
{
  this->v() = c2.v();
  this->w() = c2.w();
  this->t() = c2.t();
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
DD2ControlT DD2ControlT::operator+(const DD2ControlT& c2) const
{
  return DD2ControlT(this->v() + c2.v(),
		      this->w() + c2.w(),
		      this->t() + c2.t());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
DD2ControlT DD2ControlT::operator-(const DD2ControlT& c2) const
{
  return DD2ControlT(this->v() - c2.v(),
		      this->w() - c2.w(),
		      this->t() - c2.t());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
DD2ControlT DD2ControlT::operator*(double d) const
{
  return DD2ControlT(this->v() * d,
		      this->w() * d,
		      this->t() * d);
}


/*************** Display functions ****************/

/**
 *  \brief display the vector's coordinates
 */
void DD2ControlT::display() const
{
  cout << "(" << this->v() << "," << this->av() << "," << this->t() << ")";
}


//                                     /**************************************************/
//                                     /*************** Class ManaControlT ***************/
//                                     /**************************************************/

// /********** Constructors and destructors **********/

// /**
//  *  \brief default constructor
//  */
// ManaControlT::ManaControlT() : ControlTime(3)
// {
// }

// /**
//  *  \brief copy constructor
//  *  \param c : the control to copy
//  */
// ManaControlT::ManaControlT(const ManaControlT& c) : ControlTime(3)
// {
//   this->v() = c.v();
//   this->w() = c.w();
//   this->t() = c.t();
// }

// /**
//  *  \brief init constructor
//  *  \param v : the linear velocity control
//  *  \param w : the angular velocity control
//  *  \param t : the time of application of the control
//  */
// ManaControlT::ManaControlT(double v,double w,double t) : ControlTime(3)
// {
//   this->v() = v;
//   this->w() = w;
//   this->t() = t;
// }

// /**
//  *  \brief destructor
//  */
// ManaControlT::~ManaControlT()
// {
// }


// /******************* Operators ********************/

// /**
//  *  \brief operator = : copy the input state
//  *  \param s2 : the state to copy
//  */
// void ManaControlT::operator=(const ManaControlT& c2)
// {
//   this->v() = c2.v();
//   this->w() = c2.w();
//   this->t() = c2.t();
// }

// /**
//  *  \brief operator + : sumation
//  *  \param s2 : the second vector to add
//  *  \return the sum of both vectors
//  */
// ManaControlT ManaControlT::operator+(const ManaControlT& c2) const
// {
//   return ManaControlT(this->v() + c2.v(),
// 		      this->w() + c2.w(),
// 		      this->t() + c2.t());
// }

// /**
//  *  \brief operator + : substraction
//  *  \param s2 : the second vector to substract
//  *  \return the substraction of both vectors
//  */
// ManaControlT ManaControlT::operator-(const ManaControlT& c2) const
// {
//   return ManaControlT(this->v() - c2.v(),
// 		      this->w() - c2.w(),
// 		      this->t() - c2.t());
// }

// /**
//  *  \brief operator * : scalar multiplication
//  *  \param d : the scalar to multiply with
//  *  \return the product of the current vector and the scalar
//  */
// ManaControlT ManaControlT::operator*(double d) const
// {
//   return ManaControlT(this->v() * d,
// 		      this->w() * d,
// 		      this->t() * d);
// }


// /*************** Display functions ****************/
  
// /**
//  *  \brief display the vector's coordinates
//  */
// void ManaControlT::display() const
// {
//   cout << "(" << this->v() << "," << this->w() << "," << this->t() << ")";
// }


                                    /**************************************************/
                                    /************* Class DD2AccControlT **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DD2AccControlT::DD2AccControlT() : ControlTime(3)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
DD2AccControlT::DD2AccControlT(const DD2AccControlT& c) : ControlTime(3)
{
  this->acc() = c.acc();
  this->aacc() = c.aacc();
  this->t() = c.t();
}

/**
 *  \brief init constructor
 *  \param acc : the linear acceleration control
 *  \param aacc : the angular acceleration control
 *  \param t : the time of application of the control
 */
DD2AccControlT::DD2AccControlT(double acc,double aacc,double t) : ControlTime(3)
{
  this->acc() = acc;
  this->aacc() = aacc;
  this->t() = t;
}

/**
 *  \brief destructor
 */
DD2AccControlT::~DD2AccControlT()
{
}


/******************* Operators ********************/

/**
 *  \brief operator = : copy the input state
 *  \param s2 : the state to copy
 */
void DD2AccControlT::operator=(const DD2AccControlT& c2)
{
  this->acc() = c2.acc();
  this->aacc() = c2.aacc();
  this->t() = c2.t();
}

/**
 *  \brief operator + : sumation
 *  \param s2 : the second vector to add
 *  \return the sum of both vectors
 */
DD2AccControlT DD2AccControlT::operator+(const DD2AccControlT& c2) const
{
  return DD2AccControlT(this->acc() + c2.acc(),
			 this->aacc() + c2.aacc(),
			 this->t() + c2.t());
}

/**
 *  \brief operator + : substraction
 *  \param s2 : the second vector to substract
 *  \return the substraction of both vectors
 */
DD2AccControlT DD2AccControlT::operator-(const DD2AccControlT& c2) const
{
  return DD2AccControlT(this->acc() - c2.acc(),
			 this->aacc() - c2.aacc(),
			 this->t() - c2.t());
}

/**
 *  \brief operator * : scalar multiplication
 *  \param d : the scalar to multiply with
 *  \return the product of the current vector and the scalar
 */
DD2AccControlT DD2AccControlT::operator*(double d) const
{
  return DD2AccControlT(this->acc() * d,
			 this->aacc() * d,
			 this->t() * d);
}


/*************** Display functions ****************/
  
/**
 *  \brief display the vector's coordinates
 */
void DD2AccControlT::display() const
{
  cout << "(" << this->acc() << "," << this->aacc() << "," << this->t() << ")";
}


                                    /**************************************************/
                                    /**************** Class AASControlT ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
AASControlT::AASControlT() : ControlTime(3)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
AASControlT::AASControlT(const AASControlT& c) : ControlTime(3)
{
  this->acc() = c.acc();
  this->as()  = c.as();
  this->t()   = c.t();
}

/**
 *  \brief init constructor
 *  \param acc : the linear acceleration control
 *  \param as : the angular speed control
 *  \param t : the time of application of the control
 */
AASControlT::AASControlT(double acc,double as,double t) : ControlTime(3)
{
  this->acc() = acc;
  this->as()  = as;
  this->t()   = t;
}

/**
 *  \brief destructor
 */
AASControlT::~AASControlT()
{
}


/*************** Display functions ****************/
  
/**
 *  \brief display the vector's coordinates
 */
void AASControlT::display() const
{
  cout << "(" << this->acc() << "," << this->as() << "," << this->t() << ")";
}


                                    /**************************************************/
                                    /**************** Class CCControlT ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CCControlT::CCControlT() : ControlTime(3)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
CCControlT::CCControlT(const CCControlT& c) : ControlTime(3)
{
  this->v()  = c.v();
  this->ph() = c.ph();
  this->t()  = c.t();
}

/**
 *  \brief init constructor
 *  \param v : the velocity control
 *  \param ss : the steering angle control
 *  \param t : the time of application of the control
 */
CCControlT::CCControlT(double v,double ph,double t) : ControlTime(3)
{
  this->v()  = v;
  this->ph() = ph;
  this->t()  = t;
}

/**
 *  \brief destructor
 */
CCControlT::~CCControlT()
{
}


/*************** Display functions ****************/
  
/**
 *  \brief display the vector's coordinates
 */
void CCControlT::display() const
{
  cout << "(" << this->v() << "," << this->ph() << "," << this->t() << ")";
}
