#include "../inc/Utils.hpp"
using namespace mpc;

#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>

#include <iostream>
using namespace std;


                 /**************************************************************************************/
                 /********************************* characters treatment *******************************/
                 /**************************************************************************************/

// Function which return true if the input charachter is alphanumeric
bool mpc::isAlphaNum(char c)
{
	if( ((int)c >= (int)'0') && ((int)c <= (int)'9') ) return true;
	else if( ((int)c >= (int)'a') && ((int)c <= (int)'z') ) return true;
	else if( ((int)c >= (int)'A') && ((int)c <= (int)'Z') ) return true;
	else return false;
}

//Function which return true if the input charachter is numeric
bool mpc::isNum(char c)
{
	if( ((int)c >= (int)'0') && ((int)c <= (int)'9') ) return true;
	else return false;
}

// add a suffixe before ".pgm" extension to a filename
string mpc::addSuffixe(string s,string suf)
{
	string sres = string(s);
	sres.insert(sres.size() - 4,suf);
	return sres;
}

// return the extension type of the filename
/********************************************
 *** return null if no extension is found ***
 ********************************************/
char* mpc::getExtFilename(const char* filename)
{
  // define the maximal length of the extension
  int tailleExtMax = 10;

  // check if the filename is not null
  if(filename == NULL)
    {
      return NULL;
    }
  else
    {
      bool trouvePt = false;
      bool trouveFin = false;
      int i=0;
      int j=0;
      int plLastDot = 0;
      char* ext = NULL;

      // search for a dot in the filename
      while(!trouveFin)
	{
	  if(filename[i] == '.')
	    {
	      // a dot has been found in the filename
	      // all characters after the dot will be 
	      // defined as the extension of the file
	      trouvePt = true;
	      plLastDot = i;
	      i++;
	    }
	  else if(filename[i] == '\0')
	    {
	      // not dot has been found before the end of the filename
	      trouveFin = true;
	    }
	  else
	    {
	      // not still on the dot or at the end of the filename
	      i++;
	    }
	}
      
      // check if a dot has been found
      if(!trouvePt)
	{
	  // no dot has been found
	  return NULL;
	}
      else
	{
	  // a dot has been found : allocate the extension
	  // and replace the pointer after the last dot
	  i = plLastDot + 1;
	  ext = (char*) malloc (tailleExtMax * sizeof(char));
	}

      // get back the extension (all characters after the dot)
      while(filename[i] != '\0')
	{
	  ext[j] = filename[i];
	  i++;
	  j++;
	}

      ext[j] = '\0';

      return ext;
    }
}

// return true if filename has the extension type ext
bool mpc::isExtFilename(const char* filename, const char* ext)
{
  // get back the extension of the filename
  char* extFilename = getExtFilename(filename);

  // check if the extension of the filename is not null
  if(extFilename == NULL)
    {
      return false;
    }

  // compare the extension of the filename with the input extension
  if(strcmp(extFilename,ext) == 0)
    {
      free (extFilename);
      return true;
    }
  else 
    {
      free (extFilename);
      return false;
    }
}


                 /**************************************************************************************/
                 /********************************** numbers treatment *********************************/
                 /**************************************************************************************/

// return the minimum of both values
double mpc::min(double a,double b)
{
  if(a <= b) return a;
  else return b;
}

// return the maximum of both values
double mpc::max(double a,double b)
{
  if(a >= b) return a;
  else return b;
}

// return the absolute value of v
double mpc::valAbs(double v)
{
  if(v >= 0) return v;
  else return -v;
}

/**
 *  \brief Function which convert an integer into a string
 *  \param nb : the number to convert
 *  \param chres : the resulting string
 */
void mpc::int2String(int nb,char* chres)
{
	int i=0;
	while(nb>=10)
	{
		chres[i] = (char)(((int)'0') + nb%10);
		nb = nb/10;
		i++;
	}
	chres[i] = (char)(((int)'0') + nb);
	inverseTab(chres,i+1);
	chres[i+1] = '\0';
}

/**
 *  \brief Function which convert an double into a string
 *  \param nb : the number to convert
 *  \param chres : the resulting string
 *  \param fl : the number of character after the point
 */
void mpc::double2String(double nb,char* chres,int fl)
{
  // save the part after the dot
  double val = (nb - ((int)nb));
  int valI = (int)nb;
  
  // compute the part before the dot
  int i=0;
  while(valI>=10)
    {
      chres[i] = (char)(((int)'0') + valI%10);
      valI = valI/10;
      i++;
    }
  chres[i] = (char)(((int)'0') + valI);
  inverseTab(chres,i+1);  
  chres[i+1] = '.';
  
  // compute the part after the dot
  for(int j=i+2; j<i+2+fl; j++)
    {
      val *= 10;
      int valInt = ((int) val)%10;
      chres[j] = (char)(((int)'0') + valInt);
    }
  chres[i+2+fl] = '\0';
}

// /*
//  *  \brief compute the round integer from a float value
//  */
// int round(float nb);


                 /**************************************************************************************/
                 /***************************** tables and vectors treatment ***************************/
                 /**************************************************************************************/

/**
 *  \brief copy the content of a table
 *  \param from : the table to copy
 *  \param to : the resulting table
 *  \param n : the size of both tables
 *
 *  both from and to must be allocated first and have the same size
 */
void mpc::copyTab(double* from,double* to,int n)
{
  for(int i=0; i<n; i++) to[i] = from[i];
}

/**
 *  \brief display a table of doubles
 *  \param tab : the table to display
 *  \param n : the size of the table
 */
void mpc::displayTab(const double* tab,int n)
{
  cout << "(";
  for(int i=0; i<n-1; i++) cout << tab[i] << ",";
  if(n != 0) cout << tab[n-1];
  cout << ")";
  return;
}

/**
 *  \brief display a table of integers
 *  \param tab : the table to display
 *  \param n : the size of the table
 */
void mpc::displayTab(int* tab,int n)
{
  cout << "(";
  for(int i=0; i<n-1; i++) cout << tab[i] << ",";
  if(n != 0) cout << tab[n-1];
  cout << ")";
  return;
}

//Function which inverse a table
void mpc::inverseTab(char* tab,int n)
{
	int i;
	char tmp;
	for(i=0;i<n/2;i++)
	{
		tmp = tab[i];
		tab[i] = tab[n-1-i];
		tab[n-1-i] = tmp;
	}
}

// display vector v
void mpc::displayVectDble(vector<double>* v)
{
	for(unsigned int i=0;i<v->size();i++)
	{
		cout << (*v)[i] << " ";
	}
	cout << endl;
}

// sort the vector of double values
void mpc::sortVectDble(vector<double>& v)
{
  for(unsigned int i=0;i<v.size()-1;i++)
    {
      // search for the minimal value of the vector
      double min = 0;
      int ind = -1;
      for(unsigned int j=i; j<v.size(); j++)
	{
	  if(ind == -1 || v[j] < min)
	    {
	      min = v[j];
	      ind = j;
	    }
	}
      
      // invert the current value and the minimal value
      double tmp = v[ind];
      v[ind] = v[i];
      v[i] = tmp;
    }
}

// delete the double values from the input vector
void mpc::delDoubleValues(vector<double>& v)
{
  for(unsigned int i=0; i<v.size()-1; i++)
    {
      vector<double>::iterator it = v.begin();
      for(int j=0; j<=i; j++) it++;
      double val = v[i];
      
      while(it != v.end())
	{
	  if(*it == val) it = v.erase(it);
	  else it++;
	}
    }
}

// sort and merge two two vectors and return the result in v1
void mpc::mergeAndSortVectDble(vector<double>& v1,vector<double>& v2)
{
  // add v2 at the end of v1
  for(unsigned int i=0; i<v2.size(); i++)
    v1.push_back(v2[i]);

  // delete the double values
  delDoubleValues(v1);

  // sort the vector
  sortVectDble(v1);
}

// return a vector vres containing the values of v1 include in the 
// interval [binf;bsup]
// the vector v1 is supposed to be sorted
void mpc::getValuesBetween(double binf,double bsup,
		      const vector<double>& v1,
		      vector<double>& vres)
{
  // check if binf and bsup are consistent
  if(binf > bsup)
    {
      return;
    }

  bool trouveDeb = false;
  bool trouveFin = false;
  unsigned int i=0;
  while(i<v1.size() && trouveFin == false)
    {
      // check if the inferior bound has been found
      if(!trouveDeb)
	{
	  if(v1[i] >= binf)
	    {
	      trouveDeb = true;

	      // check if the superior bound has been already found
	      if(v1[i] > bsup)
		{
		  // in that case the result vector is empty
		  trouveFin = true;
		}
	      else
		{
		  // if the superior bound has not been reached, we add the 
		  // current value to the result vector
		  vres.push_back(v1[i]);
		  i++;
		}
	    }
	  else
	    {
	      // the inferior bound has still not been reached
	      i++;
	    }
	}
      else
	{
	  // check if the superior bound has been found
	  if(v1[i] > bsup)
	    {
	      trouveFin = true;
	    }
	  else
	    {
	      // if the superior bound has not been reached, we add the 
	      // current value to the result vector
	      vres.push_back(v1[i]);
	      i++;
	    }
	}
    }
}

/**
 *  \brief compute the index of a 1D tab from a MultiD tab
 *  \param dim : the dimension of the multidim tab
 *  \param sampMD : the sampling of the multidim tab
 *  \param indMD : the input indices in the multidim tab
 *  \param ind1D : the resulting index in the 1D tab
 */
void mpc::indMD_2_ind1D(int dim,int* sampMD,int* indMD,int& ind1D)
{
  ind1D = 0;
  int shift = 1;
  
  for(int i=dim-1; i>=0; i--)
    {
      // compute the shift corresponding to the actual dimension
      if(i != dim-1) shift *= sampMD[i+1];
      
      ind1D += indMD[i] * shift;
    }
}

/**
 *  \brief compute the index of a MultiD tab from a 1D tab
 *  \param dim : the dimension of the multidim tab
 *  \param sampMD : the sampling of the multidim tab
 *  \param ind1D : the input index in the 1D tab
 *  \param indMD : the resulting indices in the multidim tab
 */
void mpc::ind1D_2_indMD(int dim,int* sampMD,int ind1D,int* indMD)
{
  int reste1D = ind1D;

  for(int i=0; i<dim-1; i++)
    {
      int shift = 1;
      for(int j=i+1; j<dim; j++) shift *= sampMD[j];

      indMD[i] = reste1D / shift;
      reste1D = reste1D % shift;

      if(i == dim-2) indMD[dim-1] = reste1D;
    }
}


                 /**************************************************************************************/
                 /************************************ curve smoothing *********************************/
                 /**************************************************************************************/

/**
 *  \brief smooth a vector of double values
 *  \param v : the vector to smooth
 *  \param n : (2n+1) value used to smooth each parameter
 */
void mpc::smoothVec(vector<double>& v,int n)
{
  int N = v.size();
  vector<double> vtmp(N);

  if(N < 2*n+1) return;

  for(int i=0; i<N; i++)
    {
      double res = 0;
      double nbval;

      if(i < n)
	{
	  nbval = n+i+1;
	  for(int j=0; j<=i+n; j++) res += v.at(j);
	}
      else if(i > N-n-1)
	{
	  nbval = n+N-i;
	  for(int j=i-n; j<N; j++) res += v.at(j);
	}
      else
	{
	  nbval = 2*n+1;
	  for(int j=i-n; j<=i+n; j++) res += v.at(j);
	}
      vtmp[i] = res / nbval;
    }

  for(int i=0; i<N; i++) v[i] = vtmp[i];
  return;
}

// return a vector of Pascal triangle numbers
void mpc::PascTabV2(vector<double>& v,unsigned int n)
{
	if(n == 1)
	{
		v.push_back(1);
		return;
	}
	else if(n == 2)
	{
		v.push_back(1);
		v.push_back(1);
		return;
	}
	else
	{
		vector<double> vtmp = vector<double>();
		PascTabV2(vtmp,n-1);
		v.push_back(1);
		for(unsigned int i=1;i<n-1;i++)
		{
			v.push_back(vtmp[i-1] + vtmp[i]);
		}
		v.push_back(1);
		return;
	}
}

// return a vector of Pascal triangle numbers for 1st degree derivative
void mpc::PascTab1stV2(vector<double>& v,unsigned int n)
{
	if(n == 1)
	{
		v.push_back(1);
		return;
	}
	else if(n == 2)
	{
		v.push_back(-1);
		v.push_back(1);
		return;
	}
	else
	{
		vector<double> vtmp = vector<double>();
		PascTab1stV2(vtmp,n-1);
		v.push_back(-1);
		for(unsigned int i=1;i<n-1;i++)
		{
			v.push_back(vtmp[i-1] + vtmp[i]);
		}
		v.push_back(1);
		return;
	}
}


                 /**************************************************************************************/
                 /*********************************** random functions *********************************/
                 /**************************************************************************************/

// initialize the random function
int mpc::initRandom()
{
  int rand = getTimeOfDayUSec();
  srandom(rand);
  return rand;
}

// initialize the random function from a defined sequence
int mpc::initRandom(int seq)
{
  srandom(seq);
  return seq;
}

// return a real random value between 0 and 1
double mpc::getRandom()
{
  double res = (double) (1.0 * (random() / (RAND_MAX + 1.0)));
  return res;
}

// return a real random value between -1 and 1
double mpc::getRandomM11()
{
  double res = -1.0 + (double) (2.0 * (random() / (RAND_MAX + 1.0)));
  return res;
}

// return an integer random value between bsup and binf
int mpc::getIntRandom(int binf,int bsup)
{
  double realRes = ((double)binf) + ((bsup - binf + 0.99) * (random() / (RAND_MAX + 1.0)));
  return ((int)realRes);
}

// return a double random value between bsup and binf
double mpc::getDoubleRandom(double binf,double bsup)
{
  return ((double)binf) + ((bsup - binf) * (((double)random()) / ((double)(RAND_MAX + 1.0))));
}


                 /**************************************************************************************/
                 /************************************ time functions **********************************/
                 /**************************************************************************************/

// return a real gettimeofday function
double mpc::getTimeOfDay()
{
  struct timeval t1;
  gettimeofday(&t1,NULL);

  double time = ((double)t1.tv_sec) + 0.000001 * ((double)t1.tv_usec);
  return time;
}

// return only the usec value of the gettimeofday function
int mpc::getTimeOfDayUSec()
{
  struct timeval t1;
  gettimeofday(&t1,NULL);

  int time = ((int)t1.tv_usec);
  return time;
}


                 /**************************************************************************************/
                 /******************************* intersection functions *******************************/
                 /**************************************************************************************/

// compute the coefficients (a,b,c) of the equation given by a point p and a vector v
void mpc::coefEqu(Vec2D p,Vec2D v,double* a,double* b,double *c)
{
  *a = v.getVal(1);
  *b = -v.getVal(0);
  *c = v.getVal(0) * p.getVal(1) - v.getVal(1) * p.getVal(0);
}

// compute the intersection between the line defined by (p1,v1) and the line defined by (p2,v2)
// return false if both vectors are colinears
bool mpc::intersecLine2D(Vec2D p1, Vec2D v1, Vec2D p2, Vec2D v2, Vec2D& pres)
{
  if(v1.getVal(0)*v2.getVal(1) - v1.getVal(1)*v2.getVal(0) != 0)
	{
		double a1,b1,c1;
		double a2,b2,c2;
		coefEqu(p1,v1,&a1,&b1,&c1);
	        coefEqu(p2,v2,&a2,&b2,&c2);
		
		// defining the result
		pres.setVal(0,-(c1*b2 - c2*b1)/(a1*b2 - a2*b1));
		pres.setVal(1,-(c1*a2 - c2*a1)/(b1*a2 - b2*a1));
		//pres.y = (-c1 + a1*(c1*b2 - c2*b1)/(a1*b2 - a2*b1))/(b1);
		return true;
	}
	else
	{
		return false;
	}
} 

// compute the intersection between the segments [p1;p2] and [p3;p4]
// return false if there is no intersection
bool mpc::intersecSeg2D(Vec2D p1, Vec2D p2, Vec2D p3, Vec2D p4, Vec2D& pres)
{
  // compute the vector p1p2
  Vec2D vecP1P2(p2.getVal(0) - p1.getVal(0),
		p2.getVal(1) - p1.getVal(1));

  // compute the vector p3p4
  Vec2D vecP3P4(p4.getVal(0) - p3.getVal(0),
		p4.getVal(1) - p3.getVal(1));

  bool noncolinear = intersecLine2D(p1, vecP1P2, p3, vecP3P4, pres);

  if(!noncolinear) return false;
  else
    {
      // compute the distance P1P2 (square)
      double dxP1P2 = p2.getVal(0) - p1.getVal(0);
      double dyP1P2 = p2.getVal(1) - p1.getVal(1);
      double sdDistP1P2 = (dxP1P2*dxP1P2 + dyP1P2*dyP1P2);

      // compute the distance P1pres (square)
      double dxP1itsc = pres.getVal(0) - p1.getVal(0);
      double dyP1itsc = pres.getVal(1) - p1.getVal(1);
      double sdDistP1pres = (dxP1itsc*dxP1itsc + dyP1itsc*dyP1itsc);

      // compute the distance P2pres (square)
      double dxP2itsc = pres.getVal(0) - p2.getVal(0);
      double dyP2itsc = pres.getVal(1) - p2.getVal(1);
      double sdDistP2pres = (dxP2itsc*dxP2itsc + dyP2itsc*dyP2itsc);
      
      // compute the distance P3P4 (square)
      double dxP3P4 = p4.getVal(0) - p3.getVal(0);
      double dyP3P4 = p4.getVal(1) - p3.getVal(1);
      double sdDistP3P4 = (dxP3P4*dxP3P4 + dyP3P4*dyP3P4);

      // compute the distance P3pres (square)
      double dxP3itsc = pres.getVal(0) - p3.getVal(0);
      double dyP3itsc = pres.getVal(1) - p3.getVal(1);
      double sdDistP3pres = (dxP3itsc*dxP3itsc + dyP3itsc*dyP3itsc);

      // compute the distance P4pres (square)
      double dxP4itsc = pres.getVal(0) - p4.getVal(0);
      double dyP4itsc = pres.getVal(1) - p4.getVal(1);
      double sdDistP4pres = (dxP4itsc*dxP4itsc + dyP4itsc*dyP4itsc);


      // check if intersec in on the segment P1P2
      if(sdDistP1pres <= sdDistP1P2 && sdDistP2pres <= sdDistP1P2 && sdDistP3pres <= sdDistP3P4 && sdDistP4pres <= sdDistP3P4)
	{
	  return true;
	}
      else return false;
    }
}

// // compute the two intersections itsc1 and itsc2 between the line defined by (p1,v1) and the circle of center p2 and radius rad2, return false if there is no intersection
// bool intersecLineCircle2D(Vec p1, Vec v1, Vec p2, double rad2, Vec& itsc1, Vec& itsc2)
// {
  
// }

// compute the nearest intersection itsc1 between the line defined by (p1,v1) and the circle of center p2 and radius rad2 from the point p1, return false if there is no intersection
bool mpc::intersecLineCircle2D(Vec p1, Vec v1, Vec p2, double rad2, Vec& itsc1)
{
  Vec v1n = v1.normalized();
  Vec p1p2 = p2 - p1;
  Vec M = p1 + (v1n * (v1n*p1p2));
  Vec p2M = M - p2;
  double normP2M = p2M.norm();

  // check if there is an intersection
  if(normP2M < rad2)
    {
      // define the nearest intersection
      itsc1 = M - v1 * (sqrt(rad2*rad2 - normP2M*normP2M));
      return true;
    }
  else if(normP2M == rad2)
    {
      // define the only intersection
      itsc1 = M;
      return true;
    }
  else
    {
      // there is no intersection
      return false;
    }
}


                 /**************************************************************************************/
                 /*********************************** Angle functions **********************************/
                 /**************************************************************************************/

// return the value in ]-PI;PI] of the angle a
double mpc::normalizedAngle(double a)
{
  if(-M_PI < a && a <= M_PI)
    {
      return a;
    }
  else if(a > 1000 || a < -1000)
    {
      //cout << "normalizedAngle : Warning. Strange angle : " << a << endl;
      return 0;
    }
  else if(-M_PI > a)
    {
      while(a < -M_PI) a += 2*M_PI;
      return a;
    }
  else // a >= M_PI
    {
      while(a >= M_PI) a -= 2*M_PI;
      return a;
    }
}

// // return the value in ]-PI;PI] of the angle a
// double mpc::normalizedAngle(double a)
// {
//   if(-M_PI < a && a <= M_PI)
//     {
//       return a;
//     }
//   else if(a > 1000 || a < -1000)
//     {
//       cout << "normalizedAngle : Warning. Strange angle : " << a << endl;
//       return 0;
//     }
//   else if(-M_PI > a)
//     {
//       a -= M_PI;
//       double res1 = a / (-2.0*M_PI);
//       int res2 = (int) res1;
//       double res = a + M_PI + (2.0*M_PI * res2);
// //       cout << "a : " << a << "  na : " << res << "  res2 : " << res2 << endl; 
//       return res;
//     }
//   else // a >= M_PI
//     {
//       a += M_PI;
//       double res1 = a / (2.0*M_PI);
//       int res2 = (int) res1;
//       double res = a - M_PI - (2.0*M_PI * res2);
// //       cout << "a : " << a << "  na : " << res << endl;
//       return res;
//     }
// }

// compute the angle from pos1 between pos2 and pos3
double mpc::computeAngleFromBetween(Vec pos1,Vec pos2,Vec pos3, leftright_t lr)
{
  // check if the three position have the same dimension
  if(pos1.size() != pos2.size() || pos2.size() != pos3.size())
    {
      return 0;
    }
  // check if the first position is not the same than one of the other positions
  else if(pos1 == pos2 || pos1 == pos3)
    {
      return 0;
    }
  else
    {
      // compute the direction of the vector po1pos2
      double dirPos1Pos2 = atan2(pos2.getVal(1) - pos1.getVal(1), pos2.getVal(0) - pos1.getVal(0));

      // compute the direction of the vector po1pos3
      double dirPos1Pos3 = atan2(pos3.getVal(1) - pos1.getVal(1), pos3.getVal(0) - pos1.getVal(0));
      
      if(lr == LEFT)
	{
	  if(dirPos1Pos2 < dirPos1Pos3)
	    {
	      return dirPos1Pos3 - dirPos1Pos2;
	    }
	  else
	    {
	      return dirPos1Pos3 - dirPos1Pos2 + 2.0*M_PI;
	    }
	}
      else
	{
	  if(dirPos1Pos2 < dirPos1Pos3)
	    {
	      return dirPos1Pos3 - dirPos1Pos2 - 2.0*M_PI;
	    }
	  else
	    {
	      return dirPos1Pos3 - dirPos1Pos2;
	    }
	}
    }
}

// compute the minimal absolute difference between angles a1 and a2 (a1 - a2)
double mpc::computeMinDiffAngle(double a1,double a2)
{
  double res;
  if((a1 >= 0 && a2 >= 0) || (a1 <= 0 && a2 <= 0)) res = a1 - a2;
  else
    {
      if(fabs(a1 - a2) < M_PI) res = a1 - a2;
      else if(a1 > a2) res = -2*M_PI + a1 - a2;
      else res = 2*M_PI + a1 - a2;
    }

  if(fabs(res) > M_PI) 
    {
//       cout << "c'est ici a1 : " << a1 << "  a2 : " << a2 << "  res : " << res << endl;
    }
  return res;
}

// compute an approximation of the curvature between two configs
double mpc::computeCurvature(Vec2D p1,double dir1,Vec2D p2,double dir2)
{
  if(dir1 == dir2)
    {
      return 0;
    }
  else
    {
      // search if the turn is left or right
      leftright_t lr;
      double dir11;
      if(dir1 > dir2) dir11 = dir1;
      else dir11 = dir1 + 2*M_PI;
      if(dir11 - dir2 <= M_PI) lr = RIGHT;
      else lr = LEFT;
      
      // compute the resulting shifted vector
      Vec2D vDir1,vDir2;
      if(lr == LEFT)
	{
	  vDir1 = Vec2D(-sin(dir1),cos(dir1));
	  vDir2 = Vec2D(-sin(dir2),cos(dir2));
	}
      else
	{
	  vDir1 = Vec2D(sin(dir1),-cos(dir1));
	  vDir2 = Vec2D(sin(dir2),-cos(dir2));
	}
      
      // compute the intersection of both lines
      Vec2D pres;
      bool resIntersec = intersecLine2D(p1, vDir1, p2, vDir2, pres);
      if(!resIntersec) return 0;
      else
	{
	  // compute the distance between p1 and pres
	  double dx1 = pres.getVal(0) - p1.getVal(0);
	  double dy1 = pres.getVal(1) - p1.getVal(1);
	  double dist1 = sqrt(dx1*dx1 + dy1*dy1);

	  // compute the distance between p2 and pres
	  double dx2 = pres.getVal(0) - p2.getVal(0);
	  double dy2 = pres.getVal(1) - p2.getVal(1);
	  double dist2 = sqrt(dx2*dx2 + dy2*dy2);

	  // return the medium invert curvature circle radius
	  if(lr == LEFT) return 2.0/(dist1 + dist2);
	  else return -2.0/(dist1 + dist2);
	}
    }
}

// return true if Tgt is in the field of view of From with a FOV angle FOV
bool mpc::isInTheFieldOfView(Vec2D posFrom,double dirFrom,Vec2D posTgt,double dirTgt,double angleFOV)
{
//   // compute extremal angles
//   double dirMax = normalizedAngle(dirFrom + fieldOfView);
//   double dirMin = normalizedAngle(dirFrom - fieldOfView);
//   Vec2D posDirMax(posFrom.getVal(0) + cos(dirMax),posFrom.getVal(1) + sin(dirMax));
//   Vec2D posDirMin(posFrom.getVal(0) + cos(dirMin),posFrom.getVal(1) + sin(dirMin));

  //double anglePosTgtFromDirMin = computeAngleFromBetween(Vec posFrom,Vec posDirMin,Vec posTgt,LEFT);
  Vec2D posFromPosTgt(posTgt.getVal(0) - posFrom.getVal(0),posTgt.getVal(1) - posFrom.getVal(1));
  double dirFromTgt = atan2(posFromPosTgt.getVal(1),posFromPosTgt.getVal(0));
  double diffDirPos = computeMinDiffAngle(dirFrom,dirFromTgt);

  double diffDir = computeMinDiffAngle(dirFrom,dirTgt);
  
  if(fabs(diffDirPos) < angleFOV && fabs(diffDir) < angleFOV)
    return true;
  else return false;
}

// convert degrees to radians
double mpc::degToRad(const double d)
{
  return normalizedAngle(0.017453293*d);
}

// convert radians to degrees
double mpc::radToDeg(const double r)
{
  return (57.295779513*normalizedAngle(r));
}


                 /**************************************************************************************/
                 /******************************** Coordinate functions ********************************/
                 /**************************************************************************************/

// return the coordinate of the point P in the frame F
// from frame (0,0)
Vec2D mpc::coordInFrame(Vec2D P,Vec3D F)
{
  // compute the vector OP where O is the origin of the frame F
  Vec2D vecOP(P.getVal(0) - F.getVal(0),
	      P.getVal(1) - F.getVal(1));

  if(vecOP.getVal(0) == 0 && vecOP.getVal(0) == 0)
    {
      return Vec2D(0.0,0.0);
    }

  // compute the direction of vector OP in the current frame
  double dirOP = atan2(vecOP.getVal(1),vecOP.getVal(0));

  // compute the direction of vector OP in the frame F
  double dirOPinF = normalizedAngle(dirOP - F.getVal(2));

  // compute the norm of vector OP
  double dx = vecOP.getVal(0);
  double dy = vecOP.getVal(1);
  double normOP = sqrt(dx*dx + dy*dy);

  // compute the coordinate of the point P in the frame F
  return Vec2D(normOP * cos(dirOPinF),
	       normOP * sin(dirOPinF));
}

// return the coordinate of the point P in the frame (0.0)
// from frame F
Vec2D mpc::coordFromFrame(Vec2D P,Vec3D F)
{
  // compute the distance between (0,0) and the origin O of F
  double dx = F.getVal(0);
  double dy = F.getVal(1);
  double distO0 = sqrt(dx*dx+dy*dy);
  
  if(distO0 == 0)
    {
      return Vec2D(0.0,0.0);
    }

  // compute the direction of the vector between (0,0) and O
  double dirO0 = atan2(-dy,-dx);

  // compute the direction of (0,0) from F
  double dir0fromF = normalizedAngle(dirO0 - F.getVal(2));

  // compute the coord of the frame (0,0) from frame F
  Vec3D coord00fromF(distO0 * cos(dir0fromF),
		     distO0 * sin(dir0fromF),
		     -F.getVal(2));

  // return the coordinate of the point P in the frame (0.0)
  return coordInFrame(P, coord00fromF);
}


                 /**************************************************************************************/
                 /******************************* Integration functions ********************************/
                 /**************************************************************************************/

// compute a Runge Kutta integration of the function fct over [binf;bsup] knowing the initial value init
double mpc::RungeKutta(equation_t* fct, void* param, double init, double binf, double bsup, int order)
{
  int N = order;

  vector<double> pasc;
  PascTabV2(pasc, N);

  // compute the sum of pascal coefficients
  double sumPasc = 0;
  for(int i=0; i<N; i++)
    sumPasc += pasc[i];

  // initialize the result
  double res = init;

  for(int i=0; i<N; i++)
    {
      // compute the current time
      double coef = ((double)i)/((double)(N-1));
      double curVal = binf + coef * (bsup - binf);

      double pascCoef;
      if(i == 0 || i == N-1) pascCoef = pasc[i] / sumPasc;
      else pascCoef = 2.0 * pasc[i] / sumPasc;

      double curKRes = pascCoef * fct(param, curVal) * (bsup - binf);
      
      res += curKRes;
    }
  return res;
}

void mpc::gauleg(const double x1, const double x2, double x[], double w[], int n)
{
  const double EPS=1.0e-14;
  int m,j,i;
  double z1,z,xm,xl,pp,p3,p2,p1;
  
  m=(n+1)/2;
  xm=0.5*(x2+x1);
  xl=0.5*(x2-x1);
  for (i=0;i<m;i++) {
    z=cos(3.141592654*(i+0.75)/(n+0.5));
    do {
      p1=1.0;
      p2=0.0;
      for (j=0;j<n;j++) {
	p3=p2;
	p2=p1;
	p1=((2.0*j+1.0)*z*p2-j*p3)/(j+1);
      }
      pp=n*(z*p1-p2)/(z*z-1.0);
      z1=z;
      z=z1-p1/pp;
    } while (fabs(z-z1) > EPS);
    x[i]=xm-xl*z;
    x[n-1-i]=xm+xl*z;
    w[i]=2.0*xl/((1.0-z*z)*pp*pp);
    w[n-1-i]=w[i];
  }
}

double mpc::qgaus(function_t* fct, void* param, const double a, const double b)
{
  static const int n = 20;
  double x[n];
  double w[n];
  int j;
  double s;
  
  gauleg(a, b, x, w, n);      
  s=0;
  for (j=0;j<n;j++) {        
    s += w[j]*((fct)(x[j],param));
  }
  return s;
}


                 /**************************************************************************************/
                 /*************************** Equations solvation functions ****************************/
                 /**************************************************************************************/

// compute the root of a first order polynom ax + b = 0
// return the number of roots found
int mpc::rootPoly1st(double a,double b,double& r1)
{
  if(a == 0) return 0;
  else // a != 0
    {
      r1 = -b / a;
      return 1;
    }
}

// compute the discriminant of a second order polynom ax^2 + bx + c = 0
double mpc::discriminant(double a,double b,double c)
{
  return b*b - 4*a*c;
}

// compute the real roots of a second order polynom ax^2 + bx + c = 0
// return the number of real roots found
int mpc::realRootPoly2nd(double a,double b,double c,double& r1,double& r2)
{
  if(a == 0)
    {
      // first order polynomial
      return rootPoly1st(b,c,r1);
    }

  double disc = discriminant(a,b,c);

  if(disc < 0)
    {
      return 0;
    }
  else if(disc == 0)
    {
      r1 = -b / (2.0*a);
      return 1;
    }
  else // disc > 0
    {
      r1 = (-b - sqrt(disc)) / (2.0*a);
      r2 = (-b + sqrt(disc)) / (2.0*a);
      return 2;
    }
}

// compute the complex roots of a second order polynom ax^2 + bx + c = 0
// return the number of complex roots found
int mpc::complexRootPoly2nd(double a,double b,double c,complex<double>& r1,complex<double>& r2)
{
  if(a == 0)
    {
      // first order polynomial
      double r1real;
      int res = rootPoly1st(b,c,r1real);
      r1 = complex<double>(r1real);
      return res;
    }

  double disc = discriminant(a,b,c);

  if(disc < 0)
    {
      r1 = complex<double>(-b / (2.0*a), -sqrt(-disc) / (2.0*a));
      r2 = complex<double>(-b / (2.0*a), +sqrt(-disc) / (2.0*a));
      return 2;
    }
  else if(disc == 0)
    {
      r1 = complex<double>(-b / (2.0*a));
      return 1;
    }
  else // disc > 0
    {
      r1 = complex<double>((-b - sqrt(disc)) / (2.0*a));
      r2 = complex<double>((-b + sqrt(disc)) / (2.0*a));
      return 2;
    }
}


// compute the real roots of a second order polynom ax^3 + bx^2 + cx + d = 0
// return the number of real roots found
int mpc::realRootPoly3rd(double a,double b,double c,double d,double& r1,double& r2,double& r3)
{
  double eps = 0.00000001;

  if(a == 0)
    {
      // second order polynomial
      return realRootPoly2nd(b,c,d,r1,r2);
    }

  complex<double> r1cplx,r2cplx,r3cplx;
  complexRootPoly3rd(a,b,c,d,r1cplx,r2cplx,r3cplx);

//   cout << "real(r1cplx) : " << real(r1cplx) << endl;
//   cout << "imag(r1cplx) : " << imag(r1cplx) << endl;
//   cout << "real(r2cplx) : " << real(r2cplx) << endl;
//   cout << "imag(r2cplx) : " << imag(r2cplx) << endl;  
//   cout << "real(r3cplx) : " << real(r3cplx) << endl;
//   cout << "imag(r3cplx) : " << imag(r3cplx) << endl;  

  int res=0;
  if(fabs(imag(r1cplx)) < eps) { r1 = real(r1cplx); res++; }
  if(fabs(imag(r2cplx)) < eps) { 
    if(res == 0) r1 = real(r2cplx);
    else r2 = real(r2cplx);
    res++; 
  }
  if(fabs(imag(r3cplx)) < eps) { 
    if(res == 0) r1 = real(r3cplx);
    else if(res == 1) r2 = real(r3cplx);
    else r3 = real(r3cplx);
    res++; 
  }

  return res;
}

// compute the real roots of a second order polynom ax^3 + bx^2 + cx + d = 0
// return the number of real roots found
int mpc::complexRootPoly3rd(double a,double b,double c,double d,complex<double>& r1,complex<double>& r2,complex<double>& r3)
{
  if(a == 0)
    {
      // second order polynomial
      return complexRootPoly2nd(b,c,d,r1,r2);
    }

  complex<double> j1(-0.5, 0.5*sqrt(3));
  complex<double> j2(-0.5, -0.5*sqrt(3));

  // change variables (x = z - b/3a) to have a polynom of the form z^3 + p*z + q = 0
  double p = (1.0/a) * ((-b*b)/(3.0*a) + c);
  double q = (1.0/(27.0*a*a*a)) * (2*b*b*b - 9*a*b*c + 27.0*a*a*d);

  if(p == 0)
    {
      // it exists just one real root : cubroot(-q)
      // the two other roots are complex
      double r = pow(-q,1.0/3.0) - b/(3.0*a);
      r1 = complex<double>(r);
      r2 = r*j1;
      r3 = r*j2;
      return 3;
    }
  else // p != 0
    {
      double rr = (q*q) + (4.0*p*p*p / 27.0);
      complex<double> Yp3_1;
      complex<double> Yp3_2;
      
      if(rr < 0)
	{
	  Yp3_1 = complex<double>(-0.5*q, -0.5*sqrt(-rr));
	  Yp3_2 = complex<double>(-0.5*q, 0.5*sqrt(-rr));
	}
      else if(rr == 0)
	{
	  Yp3_1 = complex<double>(-0.5*q);
	}
      else // rr > 0
	{
	  Yp3_1 = complex<double>(-0.5*q - 0.5*sqrt(rr));
	  Yp3_2 = complex<double>(-0.5*q + 0.5*sqrt(rr));
	}

      complex<double> rYp3_1 = pow(Yp3_1, 1.0/3);
      complex<double> rYp3_2 = pow(Yp3_2, 1.0/3);
      
      complex<double> rYp3_11,rYp3_12,rYp3_13,rYp3_21,rYp3_22,rYp3_23;
      rYp3_11 = rYp3_1;
      rYp3_12 = rYp3_1*j1;
      rYp3_13 = rYp3_1*j2;
      
      if(rr != 0)
	{
	  rYp3_21 = rYp3_2;
	  rYp3_22 = rYp3_2*j1;
	  rYp3_23 = rYp3_2*j2;
	}

      complex<double> w1,w2,w3;
      w1 = rYp3_11 - (p / (3.0*rYp3_11));
      w2 = rYp3_12 - (p / (3.0*rYp3_12));
      w3 = rYp3_13 - (p / (3.0*rYp3_13));

      r1 = w1 - (b / (3.0*a));
      r2 = w2 - (b / (3.0*a));
      r3 = w3 - (b / (3.0*a));
      return 3;
    }
}


                 /**************************************************************************************/
                 /******************************** Intervals functions *********************************/
                 /**************************************************************************************/

/**
 *  \brief search for the interval of a regular sampling to which belongs a value v
 *  \param v : the searched value
 *  \param minB : the minimal bound of the sampling
 *  \param maxB : the maximal bound of the sampling
 *  \param nbSp : the number of values of the sampling (number of intervals + 1)
 *  \param minI : the resulting index of the lowest bound of the searched interval
 *  \param maxI : the resulting index of the highest bound of the searched interval
 *
 *  If v is outside of the interval [minB;maxB], minI and maxI are bot equals
 *  to the nearest extremal bound of [minB;maxB]
 *
 *  If v is equals to a bound of the sampling intervals, the resulting indices are [ind(v),ind(v)+1]
 */
void mpc::searchForItvSampling(double v,double minB,double maxB,int nbSp,int& minI,int& maxI)
{
  // check if v is outside of the bounds [minB;maxB]
  if(v < minB)
    {
      minI = 0;
      maxI = 0;
    }
  else if(v >= maxB)
    {
      minI = nbSp-1;
      maxI = nbSp-1;
    }
  else
    {
      double dv = v - minB;
      double ratio = dv / (maxB - minB);

      minI = ((int)(ratio * (nbSp-1)));
      maxI = minI + 1;
    }
}


                 /**************************************************************************************/
                 /******************************* Input/output functions *******************************/
                 /**************************************************************************************/

// wait for the user to press a key of the keyboard
void mpc::wait_for_key()
{
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__TOS_WIN__)  // every keypress registered, also arrow keys
    cout << endl << "Press any key to continue..." << endl;

    FlushConsoleInputBuffer(GetStdHandle(STD_INPUT_HANDLE));
    _getch();
#elif defined(unix) || defined(__unix) || defined(__unix__)
    cout << endl << "Press ENTER to continue..." << endl;

    std::cin.clear();
    std::cin.ignore(std::cin.rdbuf()->in_avail());
    std::cin.get();
#endif
    return;
}


                 /**************************************************************************************/
                 /********************************* Distance functions *********************************/
                 /**************************************************************************************/

/**
 *  \brief compute the minimal time needed to travel a distance given constraints
 *  \param dist : the distance to travel
 *  \param v : the initial velocity
 *  \param vmax : the maximal velocity
 *  \param amax : the maximal acceleration
 *  \return the minimal time needed to travel dist given constraints
 */
double mpc::minTravelTime(double dist,double v,double vmax,double amax)
{
  // check the straightforward case of the null distance
  if(dist == 0) return 0;

  // compute the minimal time needed to reach the maximal speed
  double t1 = (vmax - v)/amax;

  // compute the minimal distance needed to reach the maximal speed
  double d1 = v*t1 + 0.5*amax*t1*t1;

  if(d1 <= dist)
    {
      // case where the maximal velocity can be reached
      double d2 = dist - d1;
      double t2 = d2 / vmax;
      return (t1 + t2);
    }
  else
    {
      // case where the maximal velocity cannot be reached
      double delta = v*v + 2*amax*dist;
      double t3 = (-v + sqrt(delta))/amax;
      return t3;
    }
}

/**
 *  \brief compute the maximal time needed to travel a distance given constraints
 *  \param dist : the distance to travel
 *  \param v : the initial velocity
 *  \param vmax : the maximal velocity
 *  \param amax : the maximal acceleration
 *  \return the maximal time needed to travel dist given constraints
 *  \return INFINITY if the robot has the time to stop before to reach that position
 */
double mpc::maxTravelTime(double dist,double v,double vmax,double amax)
{
  // compute the maximal time for the robot to stop
  double maxTStop = fabs(v) / amax;

  // compute the minimal distance travelled during that time
  double minDStop;

  if(v == 0)
    return INFINITY;
  if(v > 0)
    {
      minDStop = v*maxTStop - 0.5*amax*maxTStop*maxTStop;
    }
  else // if(v < 0)
    {
      minDStop = fabs(v)*maxTStop - 0.5*amax*maxTStop*maxTStop;
    }

  // check if the robot has the time to stop
  if(dist > minDStop)
    {
      // case where the robot has the time to stop
      return INFINITY;
    }
  else
    {
//       cout << "minDStop : " << minDStop << "   ";

      // case where the robot doesn't have the time to stop
      double delta;
      if(v > 0) delta = v*v - 2*amax*dist;
      else delta = fabs(v*v - 2*amax*dist);
      if(delta < 0)
	{
	  cout << "maxTravelTime : delta < 0 !!!!!!" << endl;
	  exit(1);
	}
      if(v > 0) return ((sqrt(delta) - v) / (-amax));
      else return ((sqrt(delta) + v) / (-amax));
    }
}

/**
 *  \brief compute the minimal time and distance to reach the speed vgoal from vinit given constraints
 *  \param vinit : the initial velocity
 *  \param vgoal : the goal velocity
 *  \param amax : the maximal acceleration
 *  \param tres : the resulting time
 *  \param dres : the resulting distance
 */
void mpc::minTDtoReachSpd(double vinit,double vgoal,double amax,double& tres,double& dres)
{
  double diffv = fabs(vgoal - vinit);
  tres = diffv / amax;
  if(vgoal >= vinit)
    {
      dres = vinit*tres + 0.5*amax*tres*tres;
    }
  else
    {
      dres = vinit*tres - 0.5*amax*tres*tres;
    }
}

/**
 *  \brief compute the minimal distance travelled during a given time and constraints
 *  \param t : the duration of the motion
 *  \param v : the initial velocity
 *  \param vmax : the maximal velocity
 *  \param amax : the maximal acceleration
 *  \return the minimal distance travelled during a given time and constraints
 */
double mpc::minTravelDist(double t,double v,double vmax,double amax)
{
  // compute the time needed for the robot to stop
  double tstop = fabs(v) / amax;
  
  // check if the robot has the time to stop
  if(tstop < t)
    {
      // compute the distance travelled until the robot stops
      double dstop = v*tstop - 0.5*amax*tstop*tstop;
      return dstop;
    }
  else
    {
      // compute the total distance travelled
      double d = v*t - 0.5*amax*t*t;
      return d;
    }
}

/**
 *  \brief compute the maximal distance travelled during a given time and constraints
 *  \param t : the duration of the motion
 *  \param v : the initial velocity
 *  \param vmax : the maximal velocity
 *  \param amax : the maximal acceleration
 *  \return the maximal distance travelled during a given time and constraints
 */
double mpc::maxTravelDist(double t,double v,double vmax,double amax)
{
  // compute the time needed for the robot to reach the maximal speed
  double tMaxSpd = (vmax - v) / amax;

  // check if the robot has the time to reach the maximal speed
  if(tMaxSpd < t)
    {
      // compute the total distance travelled
      double d = v*tMaxSpd + 0.5*amax*tMaxSpd*tMaxSpd + vmax*(t-tMaxSpd);
      return d;
    }
  else
    {
      // compute the total distance travelled
      double d = v*t + 0.5*amax*t*t;
      return d;
    }
}
