#include "DDTrajDeform.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

                                    /**************************************************/
                                    /*************** Class DDTrajDeform ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
DDTrajDeform::DDTrajDeform()
{
}

/**
 *  \brief init constructor
 *  \param traj : the initial trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \param sys : the features of the robot system used
 *  \param typeS : pointer to the table of state types (for simulation display)
 *  \param tns : pointer to the table of nearest states (for simulation display)
 *  \param tfi : pointer to the table of internal forces (for simulation display)
 *  \param tfe : pointer to the table of external forces (for simulation display)
 */
DDTrajDeform::DDTrajDeform(DDSTraj* traj,Environment* env,DiffDriveSys* sys)
{
}

/**
 *  \brief init constructor
 *  \param traj : the initial trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \param sys : the features of the robot system used
 *  \param typeS : pointer to the table of state types (for simulation display)
 *  \param tns : pointer to the table of nearest states (for simulation display)
 *  \param tfi : pointer to the table of internal forces (for simulation display)
 *  \param tfe : pointer to the table of external forces (for simulation display)
 */
DDTrajDeform::DDTrajDeform(DDSTraj* traj,Environment* env,DiffDriveSys* sys,int** typeS,StateTime*** tns,StateTime** tfi,StateTime** tfe)
{
}

/**
 *  \brief destructor
 */
DDTrajDeform::~DDTrajDeform()
{
}


/*************** Forces computation ***************/

/**
 *  \brief computing of the internal forces
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param next : the next node of the trajectory
 */
DDStateT DDTrajDeform::computeFInt(const DDStateT& prev, const DDStateT& cur, const DDStateT& next, bool last, Vec_DP** param, int *fea)
{
}

/**
 *  \brief compute the total external forces
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 */
PositionXYT DDTrajDeform::computeFExt(const StateTime& cur, Environment* env, double time, bool last)
{
}

/**
 *  \brief compute the total forces
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param next : the next node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 */
CLStateT DDTrajDeform::computeForces(const DDStateT& prev, const DDStateT& cur, const DDStateT& next, Environment* env, double time, bool last)
{
}

/**
 *  \brief apply the internal forces on the current node
 *  \param i : index of the current node on the trajectory
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param next : the next node of the trajectory
 *  \param time : the current time
 */
void DDTrajDeform::applyFInt(int i,const DDStateT& prev, DDStateT& cur, const DDStateT& next, double time)
{
}

/**
 *  \brief compute the total external forces
 *  \param i : index of the current node on the trajectory
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 */
void DDTrajDeform::applyFExt(int i,DDStateT& cur, Environment* env, double time)
{
}


/******* Connectivity and Collision checking ******/

/**
 *  \brief check the connectivity of the trajectory
 *  \param curTime : the current elapsed time
 *  \return the time to the first disconnection found (-1 if no disconnection has been found)
 */
double DDTrajDeform::checkConnectivity(double curTime)
{
}

/**
 *  \brief check the connectivity between two nodes of the trajectory
 *  \param i : 
 *  \param oldCur : 
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param force : the resulting force to apply to restore the connectivity
 *
 *  SHOULD NOT BE USED!!!
 */
void DDTrajDeform::checkConnectivity(int i, const DDStateT& oldCur, const DDStateT& prev, const DDStateT& cur, DDStateT& force)
{
}
  
/**
 *  \brief adjust the time of the nodes of the trajectory to maintain its consistency
 *  \param curStep : index of the node already reached
 */
void DDTrajDeform::adjustTime(int curStep)
{
}


/******************* Resampling *******************/
  
/**
 *  \brief resample the trajectory
 *  \param curStep : index of the node already reached
 */
void DDTrajDeform::resample(int curStep)
{
}


/*************** Deformation process **************/

/**
 *  \brief init the deformation and check the temporal space between two consecutive states
 *  \param curStep : index of the node of the trajectory that has been already reached
 */
void DDTrajDeform::initDeform(int curStep)
{
}
  
/**
 *  \brief finish the deformation
 *  \param time : the current time
 *  \param curStep : index of the node of the trajectory that has been already reached
 */
int DDTrajDeform::finishDeform(double time,int curStep)
{
}

/**
 *  \brief deform the current trajectory
 *  \param time : the current time
 *  \param curStep : index of the node of the trajectory that has been already reached
 */
int DDTrajDeform::deform(double time, int curStep)
{
}


/*************** Feedback control methods ****************/

/**
 *  \brief init the list of feedback controls
 *  \param listCtrl : the input list of feedback controls
 */
void DDTrajDeform::initListCtrl(std::list <Vec_DP*>* listCtrl)
{
}


/***************** Trajectory computation ****************/

/**
 *  \brief initialize the trajectory generator and compute a trajectory from prev to next
 *  \param prev : the starting node of the trajectory generation
 *  \param next : the goal node of the trajectory generation
 *  \param carModel : the trajectory generator used
 *  \param param : the vector of parameters used for the trajectory generation
 *  \param dt : resulting duration of the generated trajectory
 *  \return the trajectory generator return status
 */
int DDTrajDeform::genTraj(const DDStateT& prev,const DDStateT& next,DDTrajGen& carModel,Vec_DP** param,double& dt) const
{
}

/**
 *  \brief compute the intermediate node to add between prev and next (strict case)
 *  \param prev : the starting node of the trajectory generation
 *  \param next : the goal node of the trajectory generation
 *  \param res : the resulting intermediate node
 *  \param param : the vector of parameters used for the trajectory generation
 *  \return 0 -> convergence has been ensured
 *  \return 1 -> convergence has not been ensured
 *
 *  strict case : compute the node only if the convergence has been ensured
 */
int DDTrajDeform::computeItdStateT_Strict(const DDStateT& prev,const DDStateT& next,DDStateT& res,Vec_DP** param)
{
}

/**
 *  \brief compute the intermediate node to add between prev and next (average case)
 *  \param prev : the starting node of the trajectory generation
 *  \param next : the goal node of the trajectory generation
 *  \param res : the resulting intermediate node
 *  \param param : the vector of parameters used for the trajectory generation
 *  \return 0 -> near convergence has been ensured
 *  \return 1 -> far convergence has been ensured
 *  \return 2 -> convergence hasn't been ensured
 *
 *  average case : compute a intermediate node even if the convergence hasn't been ensured
 */
int DDTrajDeform::computeItdStateT_Avg(const DDStateT& prev,const DDStateT& next,DDStateT& res,Vec_DP** param)
{
}


/**************** Trajectory modifications ***************/

/**
 *  \brief return the state of the deformed trajectory at time t
 *  \param t : the input time
 *  \return the state of the deformed trajectory at time t
 */
DDStateT DDTrajDeform::getStateTraj(double t) const
{
}
  
/**
 *  \brief locate the loops along the trajectory
 *  \param curStep : index of the node already reached by the system
 *  \param vecLoop : the resulting vector of loops
 *  \return the number of loops found (put in the vector vecLoop)
 */
int DDTrajDeform::locateLoop(int curStep,std::vector<int*> &vecLoop) const
{
}

/**
 *  \brief remove the loops from the trajectory
 *  \param curStep : index of the node already reached by the system
 */
void DDTrajDeform::removeLoop(int curStep)
{
}

/**
 *  \brief insert a state in the deformed trajectory
 *  \param s : the state to insert
 *  \param pos : its position in the trajectory's nodes vector
 *  \param param : the vector of parameters for the trajectory generation used to reach the new state
 *  \param fea : the feasibility status of the new inserted node
 */
void DDTrajDeform::insertStateInTraj(const DDStateT& s,int pos,Vec_DP** param,int fea)
{
}

/**
 *  \brief delete one or several states from the trajectory including start but not including end
 *  \param start : the first node to delete
 *  \param end : the node after the last node to delete
 */
void DDTrajDeform::deleteStateFromTraj(int start,int end)
{
}

/******************* Utility functions *******************/
  
/**
 *   \brief initialize the vector of constraints
 */
void DDTrajDeform::initConstraints()
{
}

/**
 *  \brief display the refined trajectory in an openGL viewer
 */
void DDTrajDeform::displayTraj() const
{
}
