#include "PMTrajDeform.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

                                    /**************************************************/
                                    /*************** Class PMTrajDeform ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
PMTrajDeform::PMTrajDeform() : TrajDeform()
{
  tableFInt_ = NULL;
  tableFExt_ = NULL;
}

/**
 *  \brief init constructor
 *  \param traj : the initial trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \param sys : the features of the robot system used
 *  \param typeS : pointer to the table of state types (for simulation display)
 *  \param tns : pointer to the table of nearest states (for simulation display)
 *  \param tfi : pointer to the table of internal forces (for simulation display)
 *  \param tfe : pointer to the table of external forces (for simulation display)
 */
PMTrajDeform::PMTrajDeform(PMSTraj* traj,Environment* env,PointMassSys* sys) : TrajDeform(traj,env,sys)
{
  tableFInt_ = NULL;
  tableFExt_ = NULL;
}

/**
 *  \brief init constructor
 *  \param traj : the initial trajectory to deform
 *  \param env : the environment in which the system evolve
 *  \param sys : the features of the robot system used
 *  \param dbg : the debugging level used
 */
PMTrajDeform::PMTrajDeform(PMSTraj* traj,Environment* env,PointMassSys* sys,int dbg) : TrajDeform(traj,env,sys,dbg)
{
  tableFInt_ = NULL;
  tableFExt_ = NULL;
}

// /**
//  *  \brief init constructor
//  *  \param traj : the initial trajectory to deform
//  *  \param env : the environment in which the system evolve
//  *  \param sys : the features of the robot system used
//  *  \param typeS : pointer to the table of state types (for simulation display)
//  *  \param tns : pointer to the table of nearest states (for simulation display)
//  *  \param tfi : pointer to the table of internal forces (for simulation display)
//  *  \param tfe : pointer to the table of external forces (for simulation display)
//  */
// PMTrajDeform::PMTrajDeform(PMSTraj* traj,Environment* env,PointMassSys* sys,int** typeS,PositionXYT*** tns,StateTime** tfi,StateTime** tfe)
//   : TrajDeform(traj,env,sys,typeS,tns,tfi,tfe)
// {
// }

/**
 *  \brief destructor
 */
PMTrajDeform::~PMTrajDeform()
{
//   if(tableFInt_ != NULL) delete tableFInt_;
//   if(tableFExt_ != NULL) delete tableFExt_;
}


/*************** Getters and Setters **************/

// /**
//  *  \brief set the debugging level
//  *  \param dbg : the new debugging level
//  */
// void PMTrajDeform::setDebug(int dbg)
// {
//   this->debug_ = dbg;

// //   typeState = new int*;
// //   tableNearState = new PositionXYT**;
// //   tableFInt = new PMStateT*;
// //   tableFExt = new PMStateT*;

// //   *typeState = NULL;
// //   *tableNearState = NULL;
// //   *tableFInt = NULL;
// //   *tableFExt = NULL;
// }

/**
 *  \brief set the debugging level
 *  \param dbg : the new debugging level
 */
void PMTrajDeform::setDebug(int dbg) 
{ 
  this->debug_ = dbg;

  tableFInt_ = NULL;
  tableFExt_ = NULL;
}


/*************** Forces computation ***************/

/**
 *  \brief computing of the internal forces
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param next : the next node of the trajectory
 */
PMStateT PMTrajDeform::computeFInt(const PMStateT& prev, const PMStateT& cur, const PMStateT& next)
{
  PMStateT sBary((prev.x() + next.x())*0.5, (prev.y() + next.y())*0.5,(prev.sx() + next.sx())*0.5, (prev.sy() + next.sy())*0.5,(prev.t() + next.t())*0.5);

  PMStateT sRes = ((sBary - cur) * (K_INT));
  return sRes;
}

// /**
//  *  \brief compute the total forces
//  *  \param prev : the previous node of the trajectory
//  *  \param cur : the current node of the trajectory
//  *  \param next : the next node of the trajectory
//  *  \param env : the environment of the system
//  *  \param time : the current time
//  *  \param last : true if the current node is the last node of the trajectory
//  */
// PMStateT PMTrajDeform::computeForces(const PMStateT& prev, const PMStateT& cur, const PMStateT& next, Environment* env, double time, bool last)
// {
// }

/**
 *  \brief apply the internal forces on the current node
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param next : the next node of the trajectory
 */
void PMTrajDeform::applyFInt(int i,const PMStateT& prev, PMStateT& cur, const PMStateT& next)
{
  // !!! NOT USED YET

  // compute the internal force
  PMStateT force = this->computeFInt(prev,cur,next);

  // aply it on the current node
  cur = cur + force;

  if(debug_ >= 1)
    {
      // save the force applied for display
      tableFInt_[i] = force;
    }
}

/**
 *  \brief compute the total external forces
 *  \param cur : the current node of the trajectory
 *  \param env : the environment of the system
 *  \param time : the current time
 *  \param last : true if the current node is the last node of the trajectory
 */
void PMTrajDeform::applyFExt(int i,PMStateT& cur, Environment* env, double time, bool last)
{
  // !!! NOT USED YET

//   PMStateT curStateWithShiftedT = cur;
//   curStateWithShiftedT.t() -= time;

//   cout << "cur:"; cur.display(); cout << "  curStateWithShiftedT:"; curStateWithShiftedT.display(); cout << endl;

  // compute the external force
  PositionXYT pforce = this->computeFExt(cur,env,time,last);
//   PositionXYT pforce = this->computeFExt(curStateWithShiftedT,env,time,last);
  PMStateT force(force.x(),force.y(),0,0,force.t());

  // aply it on the current node
  cur = cur + force;

  if(debug_ >= 1)
    {
      // save the force applied for display
      tableFExt_[i] = force;

//       // shift back the time of the nearest obstacle
//       if(tableNearState_[I] != NULL) tableNearState_[I]->t() += time;
    }
}


/******* Connectivity and Collision checking ******/

/**
 *  \brief check the connectivity between two nodes of the trajectory
 *  \param i : 
 *  \param oldCur : 
 *  \param prev : the previous node of the trajectory
 *  \param cur : the current node of the trajectory
 *  \param force : the resulting force to apply to restore the connectivity
 */
void PMTrajDeform::checkConnectivity(int i, const PMStateT& oldCur, const PMStateT& prev, const PMStateT& cur, PMStateT& force)
{
  double maxXSpd = this->sys()->getVXMax();
  double maxXAcc = this->sys()->getAXMax();
  double maxYSpd = this->sys()->getVYMax();
  double maxYAcc = this->sys()->getAYMax();
  double maxSpd = sqrt(maxXSpd*maxXSpd + maxYSpd*maxYSpd);
  double maxAcc = sqrt(maxXAcc*maxXAcc + maxYAcc*maxYAcc);

  // check the temporal consistency of the curve
  double oldDX = oldCur.x() - prev.x();
  double oldDY = oldCur.y() - prev.y();
  bool noMoreConnec1 = false;
  bool noMoreConnec2 = false;
  bool noMoreConnec11 = false;
  bool noMoreConnec22 = false;
  
  if(oldCur.t() + force.t() < prev.t() + oldDX / maxSpd)
    {
      typeState_[i] = -2;
      noMoreConnec1 = true;
      if(useTemporalConsistency_) force.t() = (prev.t() + fabs(oldDX) / maxSpd - oldCur.t());
    }
  if(oldCur.t() + force.t() < prev.t() + oldDY / maxSpd)
    {
      typeState_[i] = -2;
      noMoreConnec2 = true;
      if(useTemporalConsistency_) force.t() = (prev.t() + fabs(oldDY) / maxSpd - oldCur.t());
    }
  if(!noMoreConnec1 && !noMoreConnec2)
    {
      typeState_[i] = 0;
    }
  
  if(fabs(force.t()) > 2.0)
    {
      cout << "prev.t() : " << prev.t() << endl;
      cout << "oldCur.t() : " << oldCur.t() << endl;
      cout << "oldDX : " << oldDX << endl;
      cout << "oldDY : " << oldDY << endl;
      cout << "maxSpd : " << maxSpd << endl;
    }

  // check the connectivity of the curve
  double newDX = oldCur.x() + force.x() - prev.x();
  double newDY = oldCur.y() + force.y() - prev.y();
  double DMAX = maxSpd * (oldCur.t() + force.t() - prev.t());
  bool noMoreConnec3 = false;
  bool noMoreConnec4 = false;
  bool noMoreConnec5 = false;
  bool noMoreConnec6 = false;
  
  if(DMAX < 0) DMAX = -DMAX;
  
  if( newDX > DMAX )
    {
      typeState_[i] = -2;
      noMoreConnec3 = true;
    }
  else if ( newDX < -DMAX )
    {
      typeState_[i] = -2;
      noMoreConnec4 = true;
    }
  if( newDY > DMAX ) 
    {
      typeState_[i] = -2;
      noMoreConnec5 = true;
    }
  else if ( newDY < -DMAX ) 
    {
      typeState_[i] = -2;
      noMoreConnec6 = true;
    }
  if(!noMoreConnec1 && !noMoreConnec2 && !noMoreConnec3 && !noMoreConnec4 && !noMoreConnec5 && !noMoreConnec6)
    {
      typeState_[i] = 0;
    }
}


/******************* Resampling *******************/
  
/**
 *  \brief resample the trajectory
 *  \param curStep : index of the node already reached
 */
void PMTrajDeform::resample(int curStep)
{
  double maxTime = 2*this->timeStep_;
  double maxDist = 2*this->posStep_;

  // check if the space between two consecutive states of the trajectory are too large
  int i=curStep;
  while(i < this->traj_->getSize() - 1)
    {
      // get back the current and the next states
      PMStateT curS  = this->traj()->at(i);
      PMStateT nextS = this->traj()->at(i+1);

      double dx = nextS.x() - curS.x();
      double dy = nextS.y() - curS.y();
      double sqDist = dx*dx + dy*dy;
      
      // check if the temporal space is too large
      if(nextS.t() - curS.t() > maxTime || sqDist > maxDist)
	{
	  // add a new state between both current and next
	  PMStateT midS = (curS + nextS) * 0.5;

	  // insert it in the trajectory
	  this->traj()->insert(i+1, midS);
	}

      i++;
    }
}


/*************** Deformation process **************/

/**
 *  \brief init the deformation and check the temporal space between two consecutive states
 *  \param time : the current time
 *  \param curStep : index of the node of the trajectory that has been already reached
 */
void PMTrajDeform::initDeform(double time,int curStep)
{
  // clear the tables of forces, states types and nearest states
  if(debug_ >= 1)
    {
      if(typeState_ != NULL) delete [] typeState_;
      if(tableNearState_ != NULL)delete [] tableNearState_;
      if(tableFInt_ != NULL)delete [] tableFInt_;
      if(tableFExt_ != NULL)delete [] tableFExt_;
    }

  if(this->useResampling_) this->resample(curStep);

  // reinit the tables of states types and nearest states
  if(debug_ >= 1)
    {
      typeState_ = new int [this->traj()->getSize()];
      for(int i=0; i<this->traj()->getSize(); i++)
	{
	  typeState_[i] = -1;
	}
      tableNearState_ = (new PositionXYT* [this->traj()->getSize()]);
      for(int i=0; i<this->traj()->getSize(); i++)
	{
      tableNearState_[i] = NULL;
	}
      tableFInt_ = new PMStateT [this->traj()->getSize()];  
      tableFExt_ = new PMStateT [this->traj()->getSize()];
    }
}

/**
 *  \brief finish the deformation
 *  \param time : the current time
 *  \param curStep : index of the node of the trajectory that has been already reached
 */
int PMTrajDeform::finishDeform(double time,int curStep)
{
  this->oldTime_ = time;
  
  return 0;
}

/**
 *  \brief deform the current trajectory
 *  \param time : the current time
 *  \param curStep : index of the node of the trajectory that has been already reached
 */
int PMTrajDeform::deform(double time, int curStep)
{
  cout << "////////////////////////////////////////////////////////////////////////////////////" << endl;
  cout << "////////////////////////////////////////////////////////////////////////////////////" << endl;
  cout << "////////////////////////////////////////////////////////////////////////////////////" << endl;

  double maxXSpd = this->sys()->getVXMax();
  double maxYSpd = this->sys()->getVYMax();
  double maxSpd = sqrt(maxXSpd*maxXSpd + maxYSpd*maxYSpd);
  double elapsedTime = time - this->oldTime_;
  
  // init the deformation
  this->initDeform(time,curStep);

  if(this->traj_ == NULL)
    {
      // if the trajectory hasn't been computed, the deformation must be stoped (this cas should'nt appear)
      return -1;
    }
  else
    {
      // save the old trajectory
      PMSTraj oldTraj(*((PMSTraj*)(this->traj_)));

      // apply the repulsive forces from the obstacles
      for(int i=1; i<oldTraj.getSize(); i++)
	{
	  I = i;

	  if(i > curStep)
	    {
	      bool last;
	      if(i == oldTraj.getSize()-1) last = true;
	      else last = false;
	      
	      PMStateT fExtDyn;
	      PMStateT fExtDyn2;
	      
	      if((oldTraj)[i].t() < time + (reliabilityTime))
		{
		  // compute the external forces
 		  PositionXYT fextdyn = this->computeFExt(oldTraj.at(i), this->env_, time, last, time); // * elapsedTime;
		  fExtDyn2 = PMStateT(fextdyn.x(),fextdyn.y(),0,0,fextdyn.t());

		  if(debug_ >= 1)
		    {
		      // save the force applied for display
		      tableFExt_[i] = fExtDyn2;
		      
// 		      // shift back the time of the nearest obstacle
// 		      if(tableNearState_[i] != NULL) tableNearState_[i]->t() += time;
		    }

		  fExtDyn = fExtDyn2;

		  // check the connectivity of the curve
		  this->checkConnectivity(i, oldTraj.at(i), this->traj()->at(i-1), this->traj()->at(i), fExtDyn);
		}

	      PositionXYT fextst = this->computeStFExt2((oldTraj)[i], this->env_, last);
	      PMStateT fExtSt(fextst.x(),fextst.y(),0,0,fextst.t());

	      PMStateT forces = (fExtDyn) + (fExtSt);
// 	      PMStateT forces = (fExtDyn);

	      if(fabs(forces.t()) > 2.0) 
		{
// 		  cout << "i : " << i << "  fExtDyn : "; fExtDyn.display(); cout << "  fExtDyn2 : "; fExtDyn2.display(); cout << "  forces : "; forces.display(); cout << "  forces.t() : " << forces.t() << endl;
		}

	      // apply the forces
	      (*(this->traj()))[i] += forces;
	    }
	}
      
      // adjust the speed of each state
      for(int i=1; i<this->traj_->getSize()-1; i++)
	{
	  // get back the i-1 th i th and i+1 th states
	  PMStateT prev = this->traj()->at(i-1);
	  PMStateT next = this->traj()->at(i+1);

	  // compute the distance between the i-1 th and i+1 th configurations
	  double dx = next.x() - prev.x();
	  double dy = next.y() - prev.y();

	  // compute the time between the i-1 th and i+1 th times
	  double dt = next.t() - prev.t();

	  // compute the new speeds of the current state
	  double sx = dx/dt;
	  double sy = dy/dt;

	  // set the speeds of the current state
	  (*(this->traj()))[i].sx() = sx;
	  (*(this->traj()))[i].sy() = sy;
	}

      // apply the internal forces
      for(int i=1; i<this->traj_->getSize()-1; i++)
	{
	  if(i > curStep)
	    {
	      PMStateT forces = this->computeFInt(this->traj()->at(i-1), this->traj()->at(i), this->traj()->at(i+1));
	      tableFInt_[i] = forces;

	      (*(this->traj()))[i] += forces;
	    }
	}

      // test if any states are in collision
      double tSecur = this->checkCollision(this->traj(),this->env_,time);
    }

  // finish the deformation
  return this->finishDeform(time,curStep);
}


/**************** Display functions ***************/

/**
 *  \brief display the external forces (from dynamic obstacles) in an openGL viewer
 */
void PMTrajDeform::dispFExt()
{
  const GLfloat orangeMaterial[]={1.,1.,0.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, orangeMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, orangeMaterial);

  int N = traj()->size();

  for(int i=0; i<N; i++)
    {
      glBegin(GL_LINES);
      glVertex3f(traj()->at(i).x(), 
		 traj()->at(i).y(), 
		 traj()->at(i).t());
      glVertex3f(traj()->at(i).x() + 30 * tableFExt_[i].x(),
		 traj()->at(i).y() + 30 * tableFExt_[i].y(),
		 traj()->at(i).t() + 30 * tableFExt_[i].t());
      glEnd();
    }
}

/**
 *  \brief display the external forces (from static obstacles) in an openGL viewer
 */
void PMTrajDeform::dispStFExt()
{
}

/**
 *  \brief display the internal forces in an openGL viewer
 */
void PMTrajDeform::dispFInt()
{
  const GLfloat purpleMaterial[]={1.,0.,1.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, purpleMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, purpleMaterial);

  int N = traj()->size();

  for(int i=0; i<N; i++)
    {
      glBegin(GL_LINES);
      glVertex3f(traj()->at(i).x(), 
		 traj()->at(i).y(), 
		 traj()->at(i).t());
      glVertex3f(traj()->at(i).x() + 30 * tableFInt_[i].x(),
		 traj()->at(i).y() + 30 * tableFInt_[i].y(),
		 traj()->at(i).t() + 30 * tableFInt_[i].t());
      glEnd();
    }
}
