#include "environment/inc/DistObsMap.hpp"
using namespace mpc;

#include "common/inc/Utils.hpp"

#include <stdlib.h>
#include <iostream>
#include <stdio.h>
#include <cmath>
#include <cstring>
using namespace std;

                                    /**************************************************/
                                    /****************** Class dm3_cell ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DM3_Cell::DM3_Cell()
{
//   this->distX = 0;
//   this->distY = 0;
  this->dist = 999999999;
//   this->nearX = 0;
//   this->nearY = 0;

//     this->cellStatus_ = CELL_FREE;
//     this->it_tcd = NULL;
}

/**
 *  \brief copy constructor
 */
DM3_Cell::DM3_Cell(const DM3_Cell& cell)
{
  this->posX = cell.posX;
  this->posY = cell.posY;
  this->distX = cell.distX;
  this->distY = cell.distY;
  this->dist = cell.dist;
  this->nearX = cell.nearX;
  this->nearY = cell.nearY;
  this->cellStatus = cell.cellStatus;
  this->it_tcd = cell.it_tcd;
}

/**
 *  \brief init constructor
 */
DM3_Cell::DM3_Cell(int x, int y, obs2D_cell_t status, int distX, int distY, int nearX, int nearY, std::list<DM3_Cell*>::iterator* it)
{
  this->posX = x;
  this->posY = y;

  this->cellStatus = status;

  if(status == CELL_OBS)
    {
      this->distX = 0;
      this->distY = 0;
      this->dist = 0;
      this->nearX = x;
      this->nearY = y;
    }

  else if(distX < 0 || distY < 0)
    {
      this->distX = -1;
      this->distY = -1;
      this->dist = 999999999;
      this->nearX = -1;
      this->nearY = -1;
    }
  else
    {
      this->distX = distX;
      this->distY = distY;
      this->dist = distX*distX + distY*distY;
      this->nearX = nearX;
      this->nearY = nearY;
    }

  if(it != NULL) this->it_tcd = *it;
}

/**
 *  \brief destructor
 */
DM3_Cell::~DM3_Cell()
{
}


                                    /**************************************************/
                                    /***************** Class DistObsMap ***************/
                                    /**************************************************/

bool DistObsMap::unkwnIsObs_ = false;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DistObsMap::DistObsMap() : BoolObsMap()
{
  dist_ = NULL;
  near_ = NULL;

  tabBorder_ = NULL;
  nbBorder_ = 0;
  newTabBorder_ = NULL;
  newNbBorder_ = 0;
  status_ = NULL;

  nbFilledCells_ = 0;
  maxSqDistToObs_ = 0;
  tabCells_ = NULL;
  tabCptDist_ = NULL;
}
    
/**
 *  \brief copy constructor
 *  \param map : the map to copy
 */
DistObsMap::DistObsMap(const DistObsMap& map) : BoolObsMap(map)
{
  dist_ = NULL;
  near_ = NULL;

  tabBorder_ = NULL;
  nbBorder_ = 0;
  newTabBorder_ = NULL;
  newNbBorder_ = 0;
  status_ = NULL;

  this->initTableDist(map.nbCols_,map.nbLines_);

  // copy the distance and nearest obstacle tables
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	this->dist_[i][j] = map.dist_[i][j];
	this->near_[i][j] = map.near_[i][j];
      }

  nbFilledCells_ = 0;
  maxSqDistToObs_ = 0;
  tabCells_ = NULL;
  tabCptDist_ = NULL;
}

/**
 *  \brief init constructor
 *  \param nbCols : number of columns of the map table
 *  \param nbLines : number of lines of the map table
 *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
 *  \param xScale : X size of a cell in meters
 *  \param yScale : Z size of a cell in meters
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param thresVarSlope : the maximal slope derivative that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 */
DistObsMap::DistObsMap(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale, float thresSlope, float thresVarSlope, float safeDist)
  : BoolObsMap(nbCols, nbLines, xOrigin, yOrigin, xScale, yScale, thresSlope, thresVarSlope, safeDist)
{
  dist_ = NULL;
  near_ = NULL;

  tabBorder_ = NULL;
  nbBorder_ = 0;
  newTabBorder_ = NULL;
  newNbBorder_ = 0;
  status_ = NULL;

  nbFilledCells_ = 0;
  maxSqDistToObs_ = 0;
  tabCells_ = NULL;
  tabCptDist_ = NULL;

  this->initTableDist(nbCols,nbLines);
}

/**
 *  \brief init constructor (from a data elevation map)
 *  \param dem : the input dem
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 */
DistObsMap::DistObsMap(const SlopeMap* slopemap, float thresSlope, float safeDist)
{
  this->cellState_ = NULL;

  dist_ = NULL;
  near_ = NULL;

  tabBorder_ = NULL;
  nbBorder_ = 0;
  newTabBorder_ = NULL;
  newNbBorder_ = 0;
  status_ = NULL;

  nbFilledCells_ = 0;
  maxSqDistToObs_ = 0;
  tabCells_ = NULL;
  tabCptDist_ = NULL;

  this->buildFromSM(slopemap,thresSlope,safeDist);
}

/**
 *  \brief init constructor (from a data elevation map)
 *  \param dem : the input dem
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param thresVarSlope : the maximal slope derivative that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 */
DistObsMap::DistObsMap(const VarSlopeMap* slopemap, float thresSlope, float thresVarSlope, float safeDist)
{
  this->cellState_ = NULL;

  dist_ = NULL;
  near_ = NULL;

  tabBorder_ = NULL;
  nbBorder_ = 0;
  newTabBorder_ = NULL;
  newNbBorder_ = 0;
  status_ = NULL;

  nbFilledCells_ = 0;
  maxSqDistToObs_ = 0;
  tabCells_ = NULL;
  tabCptDist_ = NULL;

  this->buildFromVSM3(slopemap, thresSlope, thresVarSlope, safeDist);
//   this->buildFromVSM(slopemap, thresSlope, thresVarSlope, safeDist);
}

/**
 *  \brief init constructor (from an obstacle map)
 *  \param obsmap : the obstacle map
 */
DistObsMap::DistObsMap(const BoolObsMap* obsmap)
{
  this->cellState_ = NULL;

  dist_ = NULL;
  near_ = NULL;

  tabBorder_ = NULL;
  nbBorder_ = 0;
  newTabBorder_ = NULL;
  newNbBorder_ = 0;
  status_ = NULL;

  nbFilledCells_ = 0;
  maxSqDistToObs_ = 0;
  tabCells_ = NULL;
  tabCptDist_ = NULL;

  this->buildFromObsMap(obsmap);
}


/**
 *  \brief destructor
 */
DistObsMap::~DistObsMap()
{
  this->clearTables();
}

/**
 *  \brief return a copy of the current map
 *  \return a pointer to the copy of the current map
 *
 *  A new instance of DistObsMap is allocated
 */
DistObsMap* DistObsMap::copy() const
{
  return new DistObsMap(*this);
}


/***************** Initialization *****************/

/*
 *  \brief init the distance table from its dimesions
 *  \param number of columns of the table
 *  \param number of lines of the table
 */
void DistObsMap::initTableDist(int nbCols,int nbLines)
{
  dist_ = new pos_cell_xy* [nbCols];
  for(int i=0; i<nbCols; i++) dist_[i] = new pos_cell_xy [nbLines];

  for(int i=0; i<nbCols; i++) 
    for(int j=0; j<nbLines; j++) 
      {
	dist_[i][j].x = INFINITY;
	dist_[i][j].y = INFINITY;
      }

  near_ = new pos_cell_xy* [nbCols];
  for(int i=0; i<nbCols; i++) near_[i] = new pos_cell_xy [nbLines];
}

/*
 *  \brief init tables from their dimesions
 *  \param number of columns of the table
 *  \param number of lines of the table
 */
void DistObsMap::initTables(int nbCols,int nbLines)
{
  this->BoolObsMap::initTables(nbCols,nbLines);
  this->initTableDist(nbCols,nbLines);
}
    
/*
 *  \brief delete all the content of the slope map tables
 */
void DistObsMap::clearTables()
{
  this->BoolObsMap::clearTables();
  
//   if(distObs_ != NULL)
//     {
//       for(int i=0; i<nbCols_; i++) if(distObs_[i] != NULL) delete distObs_[i];
//       delete distObs_;
//     }

  if(dist_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(dist_[i] != NULL) delete [] dist_[i];
      delete [] dist_;
    }
  dist_ = NULL;

  if(near_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(near_[i] != NULL) delete [] near_[i];
      delete [] near_;
    }
  near_ = NULL;

  if(status_ != NULL)
    {
      for(int i=0; i<nbCols_; i++) if(status_[i] != NULL) delete [] status_[i];
      delete [] status_;
    }
  status_ = NULL;

  if(tabBorder_ != NULL) delete [] tabBorder_;
  tabBorder_ = NULL;
  if(newTabBorder_ != NULL) delete [] newTabBorder_;
  newTabBorder_ = NULL;

  if(tabCells_ != NULL) 
    {
      for(int i=0; i<nbCols_; i++) 
	{
	  if(tabCells_[i] != NULL) 
	    {
	      for(int j=0; j<nbLines_; j++) 
		{
		  if(tabCells_[i][j] != NULL) delete tabCells_[i][j];
		}
	      delete [] tabCells_[i];
	    }
	}
      delete [] tabCells_;
    }

  if(tabCptDist_ != NULL)
    {
//       for(int i=0; i<maxSqDistToObs_; i++) // for(int i=0; i<=maxSqDistToObs_; i++) ??? (i<= instead of i<)
      for(int i=0; i<=maxSqDistToObs_; i++)
	{
	  if(tabCptDist_[i] != NULL)
	    {
// 	      for(std::list<DM3_Cell*>::iterator it=tabCptDist_[i]->begin(); it != tabCptDist_[i]->end(); it++)
// 		{
// 		  if(*it != NULL) delete (*it);
// 		}

	      tabCptDist_[i]->clear();
 	      delete tabCptDist_[i];     //                       ------------------------------- uncommenting this line cause a Seg Fault!!!!
	      tabCptDist_[i] = NULL;
	    }
	}
      delete [] tabCptDist_;
      tabCptDist_ = NULL;
    }
}

/*
 *  \brief return true if the cell at index (i,j) is a border of an obstacle
 *  \return true if the cell at index (i,j) is a border of an obstacle
 */
bool DistObsMap::isBorderObs(int i,int j) const
{
  if(this->getCellType(i,j) != CELL_OBS) 
    { 
//       cout << "f ";
      return false;
    }

//   else if(i <= 0 || i >= nbCols_-1 || j <= 0 || j >= nbLines_-1) 
//     { 
// //       cout << "l ";
//       return false;
//     }

  else
    {
      bool found = false;
      int k = i-1;
      while(!found && k<=i+1)
	{
	  if(k < 0 || k >= nbCols_) k++;
	  else
	    {
	      int l=j-1;
	      while(!found && l<=j+1)
		{
		  if(l < 0 || l >= nbLines_) l++;
		  else
		    {
		      if(this->getCellType(k,l) != CELL_OBS)
			{
			  // 		  cout << "b ";
			  return true;
			}
		      else l++;
		    }
		}
	      k++;
	    }
	}
//       cout << "o ";
      return false;
    }
}

// /*
//  *  \brief return true if the cell at index (i,j) is a border of an obstacle
//  *  \return true if the cell at index (i,j) is a border of an obstacle
//  */
// bool DistObsMap::isBorderObs(int i,int j) const
// {
//   if((this->getCellType(i,j) == CELL_FREE) || (!DistObsMap::unkwnIsObs_ && this->getCellType(i,j) == CELL_UNKWN))
//     { 
//       return false;
//     }
//   else
//     {
//       bool found = false;
//       int k = i-1;
//       while(!found && k<=i+1)
// 	{
// 	  if(k < 0 || k >= nbCols_) k++;
// 	  else
// 	    {
// 	      int l=j-1;
// 	      while(!found && l<=j+1)
// 		{
// 		  if(l < 0 || l >= nbLines_) l++;
// 		  else
// 		    {
// 		      if((this->getCellType(k,l) == CELL_FREE) || (!DistObsMap::unkwnIsObs_ && this->getCellType(k,l) == CELL_UNKWN))
// 			{
// 			  return true;
// 			}
// 		      else l++;
// 		    }
// 		}
// 	      k++;
// 	    }
// 	}
//       return false;
//     }
// }

/*
 *  \brief init the table used for the distance computation from its dimesions
 *  \param number of columns of the table
 *  \param number of lines of the table
 */
void DistObsMap::initTableCptDist(int nbCols,int nbLines)
{
  tabBorder_ = new pos_cell_xy [nbCols*nbLines];
  nbBorder_ = 0;

  newTabBorder_ = new pos_cell_xy [nbCols*nbLines];
  newNbBorder_ = 0;

  status_ = NULL;
  status_ = new dm_cell_t* [nbCols];
  for(int i=0; i<nbCols; i++) status_[i] = new dm_cell_t [nbLines];

  for(int i=0; i<nbCols; i++) 
    {
      for(int j=0; j<nbLines; j++) 
	{
	  if(this->isBorderObs(i,j))
	    {
	      dist_[i][j].x = 0;
	      dist_[i][j].y = 0;
	      near_[i][j].x = i;
	      near_[i][j].y = j;
	      status_[i][j] = DM_CELL_DONE;
	      
	      tabBorder_[nbBorder_].x = i;
	      tabBorder_[nbBorder_].y = j;
	      nbBorder_++;
	    }
	  else
	    {
	      dist_[i][j].x = -1;
	      dist_[i][j].y = -1;
	      near_[i][j].x = -1;
	      near_[i][j].y = -1;
	      status_[i][j] = DM_CELL_ND;	      
	    }
	}
    }
}

/*
 *  \brief init the table used for the distance computation from its dimesions (for the 3rd version of the distmap computation)
 *  \param number of columns of the table
 *  \param number of lines of the table
 */
void DistObsMap::initTableCptDist3(int nbCols,int nbLines)
{
//   cout << "DistObsMap::initTableCptDist3 1" << endl;

  this->nbFilledCells_ = 0;
  this->maxSqDistToObs_ = nbCols*nbCols + nbLines*nbLines;

  // init status, dist and near tables depending on if cells are borders or not
  tabBorder_ = NULL;
  nbBorder_ = 0;

  newTabBorder_ = NULL;
  newNbBorder_ = 0;

  status_ = NULL;

//   cout << "DistObsMap::initTableCptDist3 2" << endl;
  // alloc the tabCells table used for the fast distmap computation
  this->tabCells_ = new DM3_Cell** [nbCols];
  for(int i=0; i<nbCols; i++) this->tabCells_[i] = new DM3_Cell*[nbLines];
  for(int i=0; i<nbCols; i++) 
    for(int j=0; j<nbLines; j++) 
      this->tabCells_[i][j] = NULL;
  
//   cout << "DistObsMap::initTableCptDist3 3" << endl;
  // alloc the tabCptDist table used for the fast distmap computation
  tabCptDist_ = new std::list<DM3_Cell*>* [maxSqDistToObs_+1];
  for(int i=0; i<=maxSqDistToObs_; i++) tabCptDist_[i] = NULL;

//   cout << "DistObsMap::initTableCptDist3 4" << endl;
  for(int i=0; i<nbCols; i++) 
    for(int j=0; j<nbLines; j++) 
      if(this->getCellType(i,j) == CELL_OBS)           // obstacle or unknown cells (but not borders)
	{
	  this->tabCells_[i][j] = new DM3_Cell(i, j, CELL_OBS);
	  nbFilledCells_++;
	}
  
//   cout << "DistObsMap::initTableCptDist3 5" << endl;
  // update each free neighbor of the border cells
  for(int i=0; i<nbCols; i++) 
    for(int j=0; j<nbLines; j++) 
      if(this->isBorderObs(i,j))
	{
	  for(int k=i-1; k<=i+1; k++)
	    for(int l=j-1; l<=j+1; l++)
	      {
// 		cout << "(i,j,k,l) : (" << i << "," << j << "," << k << "," << l << ")" << endl;
		if((k!=i || l!=j) && k>=0 && k<nbCols && l>=0 && l<nbLines && (this->getCellType(k,l) == CELL_FREE || (this->getCellType(k,l) == CELL_UNKWN && !DistObsMap::unkwnIsObs_)))
// 		  this->updateDist(k, l, i, j, tabCells_[i][j]);
		  this->updateDist(k, l, i, j);
	      }
	}
//   cout << "DistObsMap::initTableCptDist3 6" << endl;

//   cout << "####################### tab cpt dist apres init : " << endl;
//   this->displayTabCptDist();
//   cout << "#################################################" << endl;
}


/******************** Accessors *******************/

/*
 *  \brief set the cell (i,j) as undefined (no knowledge about the distance to obstacles)
 *  \param (i,j) : index of the cell in the grid
 */
void DistObsMap::setUndefCell(int i,int j)
{
  this->setCellType(i, j, CELL_UNKWN);
}

/*
 *  \brief set the parameters of cell (i,j)
 *  \param (i,j) : index of the cell in the grid     
 *  \param dist : the new distance the the closest obstacle
 *  \param near : the new index of the closest obstacle
 */
void DistObsMap::setDist(int i, int j, pos_cell_xy dist, pos_cell_xy near)
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    return;
  else if(dist.x <= 0)
    {
      this->setCellType(i, j, CELL_OBS);
      dist_[i][j].x = dist.x;
      dist_[i][j].y = dist.y;
//       near_[i][j].x = -1;
//       near_[i][j].y = -1;
      near_[i][j].x = i;
      near_[i][j].y = j;
    }
  else 
    {
      this->setCellType(i, j, CELL_FREE);
      dist_[i][j].x = dist.x;
      dist_[i][j].y = dist.y;
      near_[i][j].x = near.x;
      near_[i][j].y = near.y;
    }
}

/*
 *  \brief set the parameters of cell (i,j)
 *  \param (i,j) : index of the cell in the grid     
 *  \param state : status of the cell (CELL_UNKWN, CELL_OBS, CELL_FREE)
 *  \param dist : the new distance the the closest obstacle
 *  \param near : the new index of the closest obstacle
 */
void DistObsMap::setDistAndState(int i, int j, obs2D_cell_t state, pos_cell_xy dist, pos_cell_xy near)
{
  if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
    return;
  else
    {
      this->setCellType(i, j, state);
      dist_[i][j].x = dist.x;
      dist_[i][j].y = dist.y;
      near_[i][j].x = near.x;
      near_[i][j].y = near.y;
    }
}

// /*
//  *  \brief set the slope of cell (i,j)
//  *  \param (i,j) : index of the cell in the grid     
//  *  \param dist : the new distance the the closest obstacle
//  */
// void DistObsMap::setDist(int i,int j,float dist)
// {
//   if(i < 0 || i >= nbCols_ || j < 0 || j >= nbLines_)
//     return;
//   else if(dist <= 0)
//     {
//       this->setCellType(i, j, CELL_OBS);
//       distObs_[i][j] = dist;
//     }
//   else 
//     {
//       this->setCellType(i, j, CELL_FREE);
//       distObs_[i][j] = dist;
//     }
// }

/**
 *  \brief get the type of cell (i,j)
 *  \param (i,j) : index of the cell in the grid
 *  \param cellSt : the resulting cell type
 *  \param distX, distY : the resulting distance
 *  \param nearX, nearY : the nearest obstacle from the input position
 */
void DistObsMap::getCellInfo(int i, int j, obs2D_cell_t& cellSt, int& distX, int& distY, int& nearX, int& nearY) const
{
  cellSt = this->getCellType(i, j);

  if(cellSt == CELL_FREE || cellSt == CELL_UNKWN) 
    {
      distX = dist_[i][j].x;
      distY = dist_[i][j].y;

      nearX = near_[i][j].x;
      nearY = near_[i][j].y;
    }
  else
    {
      distX = 0;
      distY = 0;

      nearX = i;
      nearY = j;
    }
}

/**
 *  \brief get the type of cell (i,j)
 *  \param (i,j) : index of the cell in the grid
 *  \param cellSt : the resulting cell type
 *  \param dist : the resulting distance
 *  \param near : the nearest obstacle from the input position
 */
void DistObsMap::getCellInfo(int i, int j, obs2D_cell_t& cellSt, float& dist, PositionXY& near) const
{
  cellSt = this->getCellType(i, j);
//   if(cellSt == CELL_UNKWN) dist = 0;
//   else 
  if(cellSt == CELL_FREE || cellSt == CELL_UNKWN) 
    {
      int dxCell = dist_[i][j].x;
      int dyCell = dist_[i][j].y;
      dist = sqrt(dxCell*dxCell*xScale_*xScale_ + dyCell*dyCell*yScale_*yScale_);

      float nearX, nearY;
      getPosIndex(near_[i][j].x, near_[i][j].y, nearX, nearY);
      near.x() = nearX;
      near.y() = nearY;
    }
  else
    {
      int dxCell = dist_[i][j].x;
      int dyCell = dist_[i][j].y;
      dist = -sqrt(dxCell*dxCell*xScale_*xScale_ + dyCell*dyCell*yScale_*yScale_);
    }
}

/**
 *  \brief get the type of cell (i,j)
 *  \param (x,y) : input position on the map
 *  \param cellSt : the resulting cell type
 *  \param dist : the resulting distance
 *  \param near : the nearest obstacle from the input position
 */
void DistObsMap::getInfo(float x, float y, obs2D_cell_t& cellSt, float& dist, PositionXY& near) const
{
  // get back the index of the closest cell to the input position
  int i,j;
  int resIndex = getIndexPos(x, y, i, j);

  // get back informations about the current cell
  if(resIndex == 0) getCellInfo(i, j, cellSt, dist, near);
  else 
    {
      cellSt = CELL_UNKWN;
      dist = 0;
      near.x() = 0;
      near.y() = 0;
    }
}

/**
 *  \brief search for the cell maximizing its minimal distance to obstacles on a line (x1,y1) -> (x2,y2)
 *  \param (x1,y1) : coord of the first end point of the line
 *  \param (x2,y2) : coord of the second end point of the line
 *  \param pos : resulting cell position
 *  \param cellSt : resulting cell state
 *  \param dist : resulting cell minimal distance to obstacles
 *  \param near : resulting cell nearest obstacle position
 *  \return true if a free position has been found, false else
 */
bool DistObsMap::getCellMaxDistOnLine(float x1, float y1, float x2, float y2, PositionXY& pos, obs2D_cell_t& cellSt, float& dist, PositionXY& near) const
{
  // get back the list of cell indices belonging to the line defined between (x1,y1) and (x2,y2)
  std::vector<Map2D_index> resLineIndices;
  this->getIndicesLine(x1, y1, x2, y2, resLineIndices);

  // check if the list is not empty
  if(resLineIndices.size() == 0)
    {
      cerr << "Out of the field of view " << "x1 : " << x1 << "  y1 : " << y1 << "  x2 : " << x2 << "  y2 : " << y2 << endl;
      return false;
//       throw string("DistObsMap::getCellMaxDistOnLine : the vector resLineIndices is empty!");
    }

  int N = resLineIndices.size();
  double maxDist = 0;
  bool foundFree = false;

  for(int i=0; i<N; i++)
    {
      // get back informations about the current cell
      float curX, curY;
      obs2D_cell_t curCellSt;
      float curDist; 
      PositionXY curNear;
      
      // get back informations about the current cell
      getPosIndex(resLineIndices.at(i).col_x, resLineIndices.at(i).line_y, curX, curY);
      this->getCellInfo(resLineIndices.at(i).col_x, resLineIndices.at(i).line_y, curCellSt, curDist, curNear);

      if(curCellSt != CELL_OBS)
	{
	  if(!foundFree || curDist > maxDist)
	    {
	      foundFree = true;

	      pos = PositionXY(curX, curY);
	      cellSt = curCellSt;
	      dist = curDist;
	      maxDist = curDist;
	      near = curNear;
	    }
	}
    }
  return foundFree;
}

/**
 *  \brief search for the cell maximizing its minimal distance to obstacles on a line (x1,y1) -> (x2,y2)
 *  \param (x1,y1) : coord of the first end point of the line
 *  \param (x2,y2) : coord of the second end point of the line
 *  \param vecRes : vector containing the free segments found
 *  \return the number of free segments found
 */
int DistObsMap::getFreeSegmentsOnLine(float x1, float y1, float x2, float y2, vector<FreeSegment>& vecRes) const
{
  // get back the list of cell indices belonging to the line defined between (x1,y1) and (x2,y2)
  std::vector<Map2D_index> resLineIndices;
  this->getIndicesLine(x1, y1, x2, y2, resLineIndices);

  // check if the list is not empty
  if(resLineIndices.size() == 0)
    {
      cerr << "Out of the field of view " << "x1 : " << x1 << "  y1 : " << y1 << "  x2 : " << x2 << "  y2 : " << y2 << endl;
      return 0;
    }

  int N = resLineIndices.size();
//   cout << "DistObsMap::getFreeSegmentsOnLine -> resLineIndices.size() : " << resLineIndices.size() << endl;

  bool buildingFreeItv = false;

  PositionXY curMaxDistPos(0,0);
  float curMaxDist = 0.0;
  PositionXY curNearObs(0,0);

  FreeSegment curSeg;

  for(int i=0; i<N; i++)
    {
      // get back informations about the current cell
      float curX, curY;
      obs2D_cell_t curCellSt;
      float curDist; 
      PositionXY curNear;
      
      // get back informations about the current cell
      getPosIndex(resLineIndices.at(i).col_x, resLineIndices.at(i).line_y, curX, curY);
      this->getCellInfo(resLineIndices.at(i).col_x, resLineIndices.at(i).line_y, curCellSt, curDist, curNear);

//       cout << "DistObsMap::getFreeSegmentsOnLine -> i : " << i << "  curCellSt : " << curCellSt;

      if(curCellSt == CELL_OBS)
	{
	  if(buildingFreeItv)
	    {
// 	      cout << "   ...stops and saves free segment..." << endl;

	      // stops and saves the current built interval
	      buildingFreeItv = false;

	      // add the built interval to the resulting vector
	      vecRes.push_back(curSeg);
	    }
	  else
	    {
// 	      cout << "   ...another obstacle..." << endl;
	      // nothing to do?
	    }
	}
      else
	{
	  if(buildingFreeItv)
	    {
// 	      cout << "   ...continues the current free segment..." << endl;

	      // compute the distance from the first border point to the current position
	      double dx = curX - x1;
	      double dy = curY - y1;
	      double distToBorder = sqrt(dx*dx+dy*dy);

	      // set the maximal distance to border with the new value
	      curSeg.maxB = distToBorder;

	      curSeg.p2 = PositionXY(curX,curY);
	      float dx2 = curX - curSeg.p1.x();
	      float dy2 = curY - curSeg.p1.y();
	      curSeg.length = sqrt(dx2*dx2 + dy2*dy2);

	      // check if the new free cell is further from the obstacles than the previous ones
	      if(curDist > curSeg.dist)
		{
		  curSeg.maxDistPos = PositionXY(curX,curY);
		  curSeg.dist = curDist;
		  curSeg.nearObs = curNear;
		}
	    }
	  else
	    {
// 	      cout << "   ...starts a new free segment..." << endl;

	      // build a new interval
	      buildingFreeItv = true;

	      // compute the distance from the first border point to the current position
	      double dx = curX - x1;
	      double dy = curY - y1;
	      double distToBorder = sqrt(dx*dx+dy*dy);

	      // init the new interval parameters
	      curSeg.p1 = PositionXY(curX,curY);
	      curSeg.p2 = PositionXY(curX,curY);
	      curSeg.length = 0;
	      curSeg.minB = distToBorder;
	      curSeg.maxB = distToBorder;
	      curSeg.maxDistPos = PositionXY(curX,curY);
	      curSeg.dist = curDist;
	      curSeg.nearObs = curNear;
	    }
	}
    }

  // check if the last segment has been registered
  if(buildingFreeItv)
    {
//       cout << "   ...stops and saves free segment (final)..." << endl;

      // stops and saves the current built interval
      buildingFreeItv = false;
      
      // add the built interval to the resulting vector
      vecRes.push_back(curSeg);
    }

  // return the number of segments found
  return vecRes.size();
}

/**
 *  \brief return true if the segment defined between (x1,y1) and (x2,y2) is free
 *  \param (x1,y1) : coord of the first end point of the segment
 *  \param (x2,y2) : coord of the second end point of the sement
 *  \return true if the segment defined between (x1,y1) and (x2,y2) is free
 */
bool DistObsMap::isFreeSegmentsOnLine(float x1, float y1, float x2, float y2) const
{
  // get back the list of cell indices belonging to the line defined between (x1,y1) and (x2,y2)
  std::vector<Map2D_index> resLineIndices;
  this->getIndicesLine(x1, y1, x2, y2, resLineIndices);

  // check if the list is not empty
  if(resLineIndices.size() == 0)
    {
      cerr << "Out of the field of view " << "x1 : " << x1 << "  y1 : " << y1 << "  x2 : " << x2 << "  y2 : " << y2 << endl;
      return false;
    }

  int N = resLineIndices.size();
  bool col = false;
  int i=0; 

  while(i<N && !col)
    {
      // get back informations about the current cell
      obs2D_cell_t curCellSt;
      float curDist; 
      PositionXY curNear;
      
      // get back informations about the current cell
      this->getCellInfo(resLineIndices.at(i).col_x, resLineIndices.at(i).line_y, curCellSt, curDist, curNear);

      if(curCellSt == CELL_OBS) col = true;
      else i++;
    }

  return (!col);
}

// /*
//  *  \brief create a new cell in the grid, and return a pointer to it
//  *  \param k : index of the column in the grid of the new cell
//  *  \param l : index of the line in the grid of the new cell
//  */
// DM3_Cell* DistObsMap::init_cell(int k, in l)
// {
//   cerr << "DistObsMap::init_cell : Not defined yet!" << endl;
//   exit(1);
// }

/*
 *  \brief remove a cell from the sorted list
 *  \param cell : a pointer to the cell to remove
 */
void DistObsMap::removeCellFromSortedList(DM3_Cell* cell)
{
//   cout << "cell removed (" << cell->posX << "," << cell->posY << ")" << endl;

//   cout << "RM 1" << endl;
  // check if the input cell is not NULL
  if(cell == NULL) 
    {
      cerr << "DistObsMap::removeCellFromSortedList : The input cell is null." << endl;
      exit(1);
    }
//   cout << "RM 2... cell->dist : " << cell->dist << endl;

  // get back the list in which is currently saved the input cell
  list<DM3_Cell*>* curList = tabCptDist_[cell->dist];
//   cout << "RM 3" << endl;

  // check if the found list is not null
  if(curList == NULL)
    {
      cerr << "DistObsMap::removeCellFromSortedList : The found list is null." << endl;
      exit(1);
    }
//   cout << "RM 4" << endl;

  // remove the input cell from the found list
  curList->erase(cell->it_tcd);
//   cout << "RM 5" << endl;
}

/*
 *  \brief add a cell into the sorted list
 *  \param cell : the new cel to add into the list
 *  \param dist : the distance from the cell to add to its current closest neighbor
 *  \return an iterator to to the cell in the sorted list
 */
std::list<DM3_Cell*>::iterator DistObsMap::addCellInSortedList(DM3_Cell* cell, int dist)
{
  // check if a list has already been initialized at the input distance
  if(tabCptDist_[dist] == NULL)
    {
      tabCptDist_[dist] = new list<DM3_Cell*>;
    }

  // add the current cell to the list
  tabCptDist_[dist]->push_back(cell);

  // get back the iterator to the added cell
  std::list<DM3_Cell*>::iterator it = tabCptDist_[dist]->end();
  it--;
  return it;
}

/*
 *  \brief update the distance of a cell in the grid from one of its neighbors
 *  \param (k,l) : coordinates in the grid of the cell to update
 *  \param (i,j) : coordinates in the grid of the neighbor from which the cell (k,l) must be updated
 *  \param ngh : pointer to neighbor cell
 */
void DistObsMap::updateDist(int k, int l, int i, int j, int curDist)
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// a remplacer par DX = tabCells_[i][j]->nearX - k,  DY = tabCells_[i][j]->nearX - l
//   int DX = abs(k-i);
//   int DY = abs(l-j);
//   cout << "DX : " << DX << "  DY : " << DY << endl;
//   cout << "tabCells_[i][j]->distX : " << tabCells_[i][j]->distX << "  tabCells_[i][j]->distY : " << tabCells_[i][j]->distY << endl;

//   cout << "(i,j,k,l) : (" << i << "," << j << "," << k << "," << l << ") -> near(i,j) : (" << tabCells_[i][j]->nearX << "," << tabCells_[i][j]->nearY << ")" << endl;

//   int distX = tabCells_[i][j]->distX + DX;
//   int distY = tabCells_[i][j]->distY + DY;
  int distX = abs(tabCells_[i][j]->nearX - k);
  int distY = abs(tabCells_[i][j]->nearY - l);

  int dist = distX*distX + distY*distY;
//   cout << "(i,j):(k,l) -> (" << i << "," << j << "):(" << k << "," << l << ")" << endl;
//   cout << "distX : " << distX << "  distY : " << distY << "  dist : " << dist << endl;

  if(tabCells_[k][l] != NULL)
    {
//       cout << "tabCells_[k][l] not NULL" << endl;
      // check if the current distance is the better found until now
      if(dist < tabCells_[k][l]->dist)
	{
// 	  cout << "update cell (" << k << "," << l << ") from (" << i << "," << j << ") -> dist : " << dist << " (" << distX << "," << distY << ")" << "  old dist : " << tabCells_[k][l]->dist << "  ";

// 	  if(dist <= curDist)
// 	    cout << "DistObsMap::updateDist : dist = " << dist << " <= curDist = " << curDist << " _____ "<< "tabCells_[i][j]->nearX : " << tabCells_[i][j]->nearX << " k : " << k 
// 		 << "  tabCells_[i][j]->nearY : " << tabCells_[i][j]->nearY << " l : " << l << " ! " << endl;
// 	  else cout << endl;

	  // remove the cell from the sorted list
	  this->removeCellFromSortedList(tabCells_[k][l]);

	  // replace the best parameters found
	  tabCells_[k][l]->distX = distX;
	  tabCells_[k][l]->distY = distY;
	  tabCells_[k][l]->dist = dist;
// 	  tabCells_[k][l]->nearX = i;
// 	  tabCells_[k][l]->nearY = j;
	  tabCells_[k][l]->nearX = tabCells_[i][j]->nearX;
	  tabCells_[k][l]->nearY = tabCells_[i][j]->nearY;

	  // add the cell in the sorted list
	  tabCells_[k][l]->it_tcd = this->addCellInSortedList(tabCells_[k][l],dist);
	}
//       else cout << "nothing added" << endl;
    }
  else
    {
//       cout << "new cell (" << k << "," << l << ") -> dist : " << dist << endl;
      
//       cout << "tabCells_[k][l] NULL" << endl;
      // create a new cell
      tabCells_[k][l] = new DM3_Cell(k, l, CELL_FREE, distX, distY, tabCells_[i][j]->nearX, tabCells_[i][j]->nearY);

      // add the cell in the sorted list
      tabCells_[k][l]->it_tcd = this->addCellInSortedList(tabCells_[k][l],dist);

      nbFilledCells_++;
    }
}


/************** Obstacle map building *************/

/*
 *  \brief compute the distance to the closest obstacle for every cell
 */
void DistObsMap::computeDistObs()
{
  float minDist_prec = 1;
  int mdX_prec = 1;
  int mdY_prec = 0;
  
  while(nbBorder_ != 0)
    {
//       cout << "nbBorder_ : " << nbBorder_ << endl;

//       cout << "minDist_prec : " << minDist_prec << endl; 
//       cout << "mdX_prec : " << mdX_prec << endl; 
//       cout << "mdY_prec : " << mdY_prec << endl; 
//       cout << endl;

//       this->displayTabStatus();      
//       cout << endl;

//       this->display(true); 
//       cout << endl;
//       wait_for_key();

      newNbBorder_ = 0;

      float minDist = 999999999.9;
      int mdX = 0;
      int mdY = 0;

      for(int i=0; i<nbBorder_; i++)
	{
	  // get back the current border index
	  int xB = tabBorder_[i].x;
	  int yB = tabBorder_[i].y;

	  // update the map for each neighbour of the current border
	  for(int xV=xB-1; xV<=xB+1; xV++)
	    for(int yV=yB-1; yV<=yB+1; yV++)
	      {
		if(xV < 0 || xV >= nbCols_ || yV < 0 || yV >= nbLines_)
		  {
		    // nothing to do
		  }
		else // if((xV == xB-1 && yV == yB) || (xV == xB+1 && yV == yB) || (xV == xB && yV == yB-1) || (xV == xB && yV == yB+1))
		  {
		    // compute the current dx and dy from the border point
		    int curDX = dist_[xB][yB].x + abs(xV - xB);
		    int curDY = dist_[xB][yB].y + abs(yV - yB);
		    float curDist = curDX*curDX + curDY*curDY;

		    if(curDist <= minDist_prec)
		    {

			if(status_[xV][yV] == DM_CELL_ND)
			  {
			    status_[xV][yV] = DM_CELL_CUR;

			    near_[xV][yV] = near_[xB][yB];
			    dist_[xV][yV].x = curDX;
			    dist_[xV][yV].y = curDY;

			    newTabBorder_[newNbBorder_].x = xV;
			    newTabBorder_[newNbBorder_].y = yV;
			    newNbBorder_++;

			    if(curDX*curDX + curDY*curDY < minDist)
			      {
				minDist = curDX*curDX + curDY*curDY;
				mdX = curDX;
				mdY = curDY;
			      }
			  }
			else if(status_[xV][yV] == DM_CELL_CUR)
			  {
			    if(curDX*curDX + curDY*curDY < dist_[xV][yV].x*dist_[xV][yV].x + dist_[xV][yV].y*dist_[xV][yV].y)
			      {
				near_[xV][yV] = near_[xB][yB];
				dist_[xV][yV].x = curDX;
				dist_[xV][yV].y = curDY;			    
			      }

			    if(curDX*curDX + curDY*curDY < minDist)
			      {
				minDist = curDX*curDX + curDY*curDY;
				mdX = curDX;
				mdY = curDY;
			      }
			  }
			else
			  {
			    // nothing to do
			  }

		    }
		  }
	      }
	}

      // replace the new minimal distances
      mdX_prec = mdX;
      mdY_prec = mdY;
      if(mdX_prec <= mdY_prec) mdX_prec++;
      else mdY_prec++;
      minDist_prec = mdX_prec*mdX_prec + mdY_prec*mdY_prec;
      
//       this->displayTabStatus();
//       wait_for_key();

      // set the status of all cells modified at this step as done
      for(int i=0; i<nbCols_; i++)
	for(int j=0; j<nbLines_; j++)
	  {
	    if(status_[i][j] == DM_CELL_CUR) status_[i][j] = DM_CELL_DONE;
	  }

      // save the border table build during the past step
      nbBorder_ = newNbBorder_;
      for(int i=0; i<newNbBorder_; i++)
	{
	  tabBorder_[i].x = newTabBorder_[i].x;
	  tabBorder_[i].y = newTabBorder_[i].y;
	}
    }
}

/*
 *  \brief compute the distance to the closest obstacle for every cell (2nd version)
 */
void DistObsMap::computeDistObs2()
{
  float minDist_prec = 1;
  
  while(nbBorder_ != 0)
    {
//       cout << endl << endl;
//       cout << "nbBorder_ : " << nbBorder_ << endl;

//       cout << "minDist_prec : " << minDist_prec << endl; 
//       cout << endl;

//       this->displayTabStatus();      
//       cout << endl;

//       this->display(true); 
//       cout << endl;
//       wait_for_key();

      newNbBorder_ = 0;

      float minNotFilled = 999999999.9;
      int mdX = 0;
      int mdY = 0;

      for(int i=0; i<nbBorder_; i++)
	{
	  // get back the current border index
	  int xB = tabBorder_[i].x;
	  int yB = tabBorder_[i].y;

	  bool surroundByFar = false;

	  // update the map for each neighbour of the current border
	  for(int xV=xB-1; xV<=xB+1; xV++)
	    for(int yV=yB-1; yV<=yB+1; yV++)
	      {
		if(xV < 0 || xV >= nbCols_ || yV < 0 || yV >= nbLines_)
		  {
		    // nothing to do
		  }
		else // if((xV == xB-1 && yV == yB) || (xV == xB+1 && yV == yB) || (xV == xB && yV == yB-1) || (xV == xB && yV == yB+1))
		  {
		    // compute the current dx and dy from the border point
		    int curDX = dist_[xB][yB].x + abs(xV - xB);
		    int curDY = dist_[xB][yB].y + abs(yV - yB);
		    float curDist = curDX*curDX + curDY*curDY;

		    if(curDist <= minDist_prec)
		    {

			if(status_[xV][yV] == DM_CELL_ND)
			  {
			    status_[xV][yV] = DM_CELL_CUR;

			    near_[xV][yV] = near_[xB][yB];
			    dist_[xV][yV].x = curDX;
			    dist_[xV][yV].y = curDY;

			    newTabBorder_[newNbBorder_].x = xV;
			    newTabBorder_[newNbBorder_].y = yV;
			    newNbBorder_++;

// 			    if(curDX*curDX + curDY*curDY < minDist)
// 			      {
// 				minDist = curDX*curDX + curDY*curDY;
// 				mdX = curDX;
// 				mdY = curDY;
// 			      }
			  }
			else if(status_[xV][yV] == DM_CELL_CUR)
			  {
			    if(curDX*curDX + curDY*curDY < dist_[xV][yV].x*dist_[xV][yV].x + dist_[xV][yV].y*dist_[xV][yV].y)
			      {
				near_[xV][yV] = near_[xB][yB];
				dist_[xV][yV].x = curDX;
				dist_[xV][yV].y = curDY;			    
			      }

// 			    if(curDX*curDX + curDY*curDY < minDist)
// 			      {
// 				minDist = curDX*curDX + curDY*curDY;
// 				mdX = curDX;
// 				mdY = curDY;
// 			      }
			  }
			else
			  {
			    // nothing to do
			  }

		    }
		    else
		      {
// 			cout << curDist << " ";
			
			if(status_[xV][yV] == DM_CELL_ND)
			  {
			    if(curDist < minNotFilled) minNotFilled = curDist;
			    surroundByFar = true;
			  }
		      }
		  }
	      }

	  if(surroundByFar)
	    {
	      // add this border to the set of borders again
	      newTabBorder_[newNbBorder_].x = xB;
	      newTabBorder_[newNbBorder_].y = yB;
	      newNbBorder_++;	      
	    }
	}

      // replace the new minimal distances
      minDist_prec = minNotFilled;
//       mdX_prec = mdX;
//       mdY_prec = mdY;
//       if(mdX_prec <= mdY_prec) mdX_prec++;
//       else mdY_prec++;
//       minDist_prec = mdX_prec*mdX_prec + mdY_prec*mdY_prec;
      
//       this->displayTabStatus();
//       wait_for_key();

      // set the status of all cells modified at this step as done
      for(int i=0; i<nbCols_; i++)
	for(int j=0; j<nbLines_; j++)
	  {
	    if(status_[i][j] == DM_CELL_CUR) status_[i][j] = DM_CELL_DONE;
	  }

//       cout << "newNbBorder_ : " << newNbBorder_ << endl;

      // save the border table build during the past step
      nbBorder_ = newNbBorder_;
      for(int i=0; i<newNbBorder_; i++)
	{
	  tabBorder_[i].x = newTabBorder_[i].x;
	  tabBorder_[i].y = newTabBorder_[i].y;
	}
    }
}

/*
 *  \brief compute the distance to the closest obstacle for every cell (3rd version)
 */
void DistObsMap::computeDistObs3()
{
//   cout << "DistObsMap::computeDistObs3" << endl;

  int nbCels = nbCols_ * nbLines_;
  int curDist = 1;

  while(nbFilledCells_ != nbCels
	&& curDist < maxSqDistToObs_
	)
    {
//       cout << "------------------- curDist : " << curDist << "------------------ " << " ( nbFilledCells_ : " << nbFilledCells_ << " / nbCels : " << nbCels << " )" << endl;

      if(tabCptDist_[curDist] != NULL)
	{
// 	  cout << "tab cpt dist : " << endl;
// 	  this->displayTabCptDist();

// 	  cout << endl;
// 	  cout << "tab bool obs : " << endl;
// 	  this->displayTabBoolObs();
// // 	  cout << endl;

// 	  cout << endl;
// 	  cout << "tab cells : " << endl;
// 	  this->displayTabCells();
// 	  cout << endl;

	  int curSize = tabCptDist_[curDist]->size();
	  int cp = 0;

	  std::list<DM3_Cell*>::iterator it = tabCptDist_[curDist]->begin();
	  while(it != tabCptDist_[curDist]->end() && cp < curSize)
	    {
	      DM3_Cell* cell = *it;
	      int i=cell->posX;
	      int j=cell->posY;

// 	      cout << "current cell : (" << i << "," << j << ") --- " << cp << "/" << curSize << endl;

	      for(int k=i-1; k<=i+1; k++)
		for(int l=j-1; l<=j+1; l++)
		  {
		    if((k!=i || l!=j) && k>=0 && k<nbCols_ && l>=0 && l<nbLines_)
		      {
			this->updateDist(k, l, i, j, curDist);
		      }
		  }

	      it++;
	      cp++;
	    }

// 	  this->fill_tabDistNear_from_tabCell();
// 	  char str[50];
// 	  strcpy (str, "distmap-");
// 	  char strcp[10];
// 	  mpc::int2String(curDist, strcp);
// 	  strcat (str, strcp);
// 	  strcat (str, ".ppm");
// 	  this->saveIntoPPM(str);

//   	  tabCptDist_[curDist]->clear();
//  	  delete tabCptDist_[curDist];
//  	  tabCptDist_[curDist] = NULL;
	}
      curDist++;
    }
}

/*
 *  \brief fill the dist and near table from the cell table of the 3rd  version of the distmap computation
 */
void DistObsMap::fill_tabDistNear_from_tabCell()
{
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	if(tabCells_[i][j] != NULL)
	  {
	    pos_cell_xy dist;
	    dist.x = tabCells_[i][j]->distX;
	    dist.y = tabCells_[i][j]->distY;
	    
	    pos_cell_xy near;
	    near.x = tabCells_[i][j]->nearX;
	    near.y = tabCells_[i][j]->nearY;

	    obs2D_cell_t state = tabCells_[i][j]->cellStatus;

// 	    this->setDist(i, j, dist, near);
	    this->setDistAndState(i, j, state, dist, near);
	  }
	else this->setUndefCell(i, j);
      }
}

/*
 *  \brief set every unknown cell as defining an obstacle
 */
void DistObsMap::setUnknwnAsObs()
{
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	obs2D_cell_t cellSt = this->getCellType(i, j);
	if(cellSt == CELL_UNKWN) this->setCellType(i, j, CELL_OBS);
      }
}

/**
 *  \brief build the obstacle map from an input slope map
 *  \param sm : the input slope map
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 *  \return 0 if the process succeed, -1 else
 */
int DistObsMap::buildFromSM(const SlopeMap* sm, float thresSlope, float safeDist)
{
  if(sm == NULL) return -1;

  // initialize the slope map parameters
  this->nbCols_ = sm->nbCols();
  this->nbLines_ = sm->nbLines();
  
  this->xOrigin_ = sm->xOrigin();
  this->yOrigin_ = sm->yOrigin();
  
  this->xScale_ = sm->xScale();
  this->yScale_ = sm->yScale();
  
  // initialize the obstacle map tables
  this->clearTables();
  this->initTables(sm->nbCols(),sm->nbLines());

  // initialize the slope threshold and the safe distance
  this->thresSlope_ = thresSlope;
  this->safeDist_ = safeDist;

  // thresholding the input slope map to obtain an obstacle map
  this->thresSM(sm, thresSlope);

//   cout << "Table obs : " << endl;
//   for(int i=0; i<nbCols_; i++)
//     {
//       for(int j=0; j<nbLines_; j++) 
// 	{
// 	  if(cellState_[i][j] == CELL_UNKWN) cout << "U ";
// 	  else if(cellState_[i][j] == CELL_OBS) cout << "O ";
// 	  else cout << "F ";
// 	}
//       cout << endl;
//     }
//   cout << endl;

//   cout << "Bool Obs Map : " << endl;
//   this->displayTabBoolObs();
//   cout << endl;

  // set every unknown cell as defining an obstacle
  if(DistObsMap::unkwnIsObs_) this->setUnknwnAsObs();

  // increasing the obstacles with the input safe distance
  this->increaseObs(safeDist);

//   cout << "Table obs increased : " << endl;
//   for(int i=0; i<nbCols_; i++)
//     {
//       for(int j=0; j<nbLines_; j++) 
// 	{
// 	  if(cellState_[i][j] == CELL_UNKWN) cout << "U ";
// 	  else if(cellState_[i][j] == CELL_OBS) cout << "O ";
// 	  else cout << "F ";
// 	}
//       cout << endl;
//     }
//   cout << endl;

  // initialize the tables used for the distance computation
  this->initTableCptDist(sm->nbCols(),sm->nbLines());

  // compute the distance to the closest obstacle for every cell
//   this->computeDistObs();
  this->computeDistObs2();

  return 0;
}

/**
 *  \brief build the obstacle map from an input slope map
 *  \param sm : the input slope map
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 *  \return 0 if the process succeed, -1 else
 */
int DistObsMap::buildFromVSM(const VarSlopeMap* vsm, float thresSlope, float thresVarSlope, float safeDist)
{
  if(vsm == NULL) return -1;

  // initialize the slope map parameters
  this->nbCols_ = vsm->nbCols();
  this->nbLines_ = vsm->nbLines();
  
  this->xOrigin_ = vsm->xOrigin();
  this->yOrigin_ = vsm->yOrigin();
  
  this->xScale_ = vsm->xScale();
  this->yScale_ = vsm->yScale();
  
  // initialize the obstacle map tables
  this->clearTables();
  this->initTables(vsm->nbCols(),vsm->nbLines());

  // initialize the slope threshold and the safe distance
  this->thresSlope_ = thresSlope;
  this->safeDist_ = safeDist;

  // thresholding the input slope map to obtain an obstacle map
  this->thresVSM(vsm, thresSlope, thresVarSlope);

  // set every unknown cell as defining an obstacle
  if(DistObsMap::unkwnIsObs_) this->setUnknwnAsObs();

  // increasing the obstacles with the input safe distance
  this->increaseObs(safeDist);

  // initialize the tables used for the distance computation
  this->initTableCptDist(vsm->nbCols(),vsm->nbLines());

  // compute the distance to the closest obstacle for every cell
//   this->computeDistObs();
  this->computeDistObs2();

  return 0;
}

/**
 *  \brief build the obstacle map from an input slope map (third version)
 *  \param sm : the input slope map
 *  \param thresSlope : the maximal slope that the robot can climb
 *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
 *  \return 0 if the process succeed, -1 else
 */
int DistObsMap::buildFromVSM3(const VarSlopeMap* vsm, float thresSlope, float thresVarSlope, float safeDist)
{
  if(vsm == NULL) return -1;

  // initialize the slope map parameters
  this->nbCols_ = vsm->nbCols();
  this->nbLines_ = vsm->nbLines();
  
  this->xOrigin_ = vsm->xOrigin();
  this->yOrigin_ = vsm->yOrigin();
  
  this->xScale_ = vsm->xScale();
  this->yScale_ = vsm->yScale();
  
  // initialize the obstacle map tables
  this->clearTables();
  this->initTables(vsm->nbCols(),vsm->nbLines());

  // initialize the slope threshold and the safe distance
  this->thresSlope_ = thresSlope;
  this->safeDist_ = safeDist;

  // thresholding the input slope map to obtain an obstacle map
  this->thresVSM(vsm, thresSlope, thresVarSlope);

  // set every unknown cell as defining an obstacle
  if(DistObsMap::unkwnIsObs_) this->setUnknwnAsObs();

  // increasing the obstacles with the input safe distance
  this->increaseObs(safeDist);

  // initialize the tables used for the distance computation
  this->initTableCptDist3(vsm->nbCols(),vsm->nbLines());
 
  // compute the distance to the closest obstacle for every cell
//   this->computeDistObs();
  this->computeDistObs3();

  this->fill_tabDistNear_from_tabCell();

  return 0;
}

/**
 *  \brief build the distance map from an obstacle map
 *  \param obsmap : the input obstacle map
 *  \return 0 if the process succeed, -1 else
 */
int DistObsMap::buildFromObsMap(const BoolObsMap* obsmap)
{
  // initialize the obstacle map parameters
  this->nbCols_ = obsmap->nbCols();
  this->nbLines_ = obsmap->nbLines();
  
  this->xOrigin_ = obsmap->xOrigin();
  this->yOrigin_ = obsmap->yOrigin();
  
  this->xScale_ = obsmap->xScale();
  this->yScale_ = obsmap->yScale();

  // initialize the obstacle map tables
  this->clearTables();
  this->initTables(obsmap->nbCols(),obsmap->nbLines());

  // copy the obstacle map tables
  for(int i=0; i<nbCols_; i++)
    for(int j=0; j<nbLines_; j++)
      {
	this->setCellType(i,j,obsmap->getCellType(i,j));
      }

  // initialize the tables used for the distance computation
  this->initTableCptDist3(obsmap->nbCols(),obsmap->nbLines());
 
  // compute the distance to the closest obstacle for every cell
//   this->computeDistObs();
  this->computeDistObs3();

  this->fill_tabDistNear_from_tabCell();

  return 0;
}


/************ Input / Output functions ************/

/*
 *  \brief init the Z color parameters
 *  \param minColZ : output minimal Z value used to convert the DEM into an image
 *  \param maxColZ : output maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : output Z scale value used to convert the DEM into an image
 */
void DistObsMap::initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const
{
  minColZ = 1;

  // search for the maximal distance from an obstacle
  int maxSqDist = 1;

  for(int i=0; i<nbCols_; i++) 
    for(int j=0; j<nbLines_; j++) 
      {
	if(cellState_[i][j] != CELL_OBS)
	  {
	    float dx = dist_[i][j].x * xScale_;
	    float dy = dist_[i][j].y * yScale_;
	    int sqDist = dx*dx + dy*dy;

	    if(sqDist > maxSqDist) maxSqDist = sqDist;
	  }
      }

  minColZ = 0.0;
  maxColZ = sqrt(maxSqDist);
  scaleColZ = (maxColZ - minColZ) / colDepth_;

//   cout << "MAX DIST (maxColZ) : " << maxColZ << endl;
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \return the colour corresponding to a cell of the DEM
 */
int DistObsMap::greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const
{
}

/*
 *  \brief return the colour corresponding to a cell of the DEM
 *  \param minColZ : minimal Z value used to convert the DEM into an image
 *  \param maxColZ : maximal Z value used to convert the DEM into an image
 *  \param scaleColZ : Z scale value used to convert the DEM into an image
 *  \param R : the output red colour
 *  \param G : the output green colour
 *  \param B : the output blue colour
 */
void DistObsMap::rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const
{
  obs2D_cell_t cellSt;
  float dist;
  PositionXY near;
  getCellInfo(i, j, cellSt, dist, near);

//   if(cellSt == CELL_UNKWN)
//     {
//       R= (int) (0.5*colDepth_);
//       G= (int) (0.5*colDepth_);
//       B= (int) (0.5*colDepth_);
//     }
//   else 
    if(cellSt == CELL_OBS)
    {
      R=0;
      G=0;
      B=0;
    }
  else
    {
      R= (int) (colDepth_ - dist / scaleColZ);
      G= (int) (dist / scaleColZ);
      B= 0;

//       R= (int) (colDepth_ - 2.0 * dist / scaleColZ);
//       B= 0;
//       if(R < 0)
// 	{
// 	  B = fabs(R);
// 	  R = 0;
// 	}
//       G= (int) (dist / scaleColZ);
    }
}


/**************** Utility functions ***************/

/*
 *  \brief display the status table
 */
void DistObsMap::displayTabStatus() const
{
      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
	      if(status_[i][j] == DM_CELL_ND) cout << "-        ";
	      else if(status_[i][j] == DM_CELL_CUR) cout << "C        ";
	      else cout << dist_[i][j].x << ";" << dist_[i][j].y << "      ";
// 	      else cout << "X        ";
	    }
	  cout << endl;
	}
}

/*
 *  \brief display the boolean obstacle table
 */
void DistObsMap::displayTabBoolObs() const
{
  this->BoolObsMap::display(true);
}

/*
 *  \brief display the cell table sorted by distance
 */
void DistObsMap::displayTabCptDist() const
{
  cout << "====================================================================================================================" << endl;
  cout << "nbFilledCells : " << nbFilledCells_ << endl;
  cout << "maxSqDistToObs : " << maxSqDistToObs_ << endl;

  cout << endl << "tabCptDist : " << endl;
  for(int i=0; i<maxSqDistToObs_; i++)
    {
      if(tabCptDist_[i] != NULL)
	{
	  cout << "I : " << i << "  nbCells : " << tabCptDist_[i]->size() << " ----- ";

	  std::list<DM3_Cell*>::iterator it = tabCptDist_[i]->begin(); 
	  while(it != tabCptDist_[i]->end())
	    {
	      DM3_Cell* cell = *it;
	      cout << "(" << cell->posX << "," << cell->posY << ") ";
	      it++;
	    }
	  cout << endl;
	}
    }
  cout << "====================================================================================================================" << endl;
}

/*
 *  \brief display the table of cells (3rd distmap computation version)
 */
void DistObsMap::displayTabCells() const
{
      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
	      if(tabCells_[i][j] == NULL) cout << "N ";
	      else
		{
		  cout << tabCells_[i][j]->dist << " ";
		}
	    }
	  cout << endl;
	}
}

/**
 *  \brief display the main features of the DEM
 *  \param dispGrid : display the grid if true
 */
void DistObsMap::display(bool dispGrid) const
{
  // display the main features
  cout << "nbLines : " << nbLines_ <<  "  nbCols : " << nbCols_ << endl;
  cout << "xOrigin : " << xOrigin_ <<  "  yOrigin : " << yOrigin_ << endl;
  cout << "xScale : " << xScale_ <<  "  yScale : " << yScale_ << endl;
  cout << endl;

  // display the grid if needed
  if(dispGrid)
    {
      for(int i=0; i<nbCols_; i++)
	{
	  for(int j=0; j<nbLines_; j++)
	    {
	      obs2D_cell_t cellSt;
	      float dist;
	      PositionXY near;
	      this->getCellInfo(i,j,cellSt,dist,near);
	      
// 	      if(cellSt == CELL_UNKWN) cout << "? ";
// 	      else if(cellSt == CELL_FREE) cout << "F(" << dist << ") ";
// 	      else cout << "O(" << dist << ") ";

	      if(cellSt == CELL_UNKWN) printf("? ");
	      else if(cellSt == CELL_FREE) printf("F( %.2f) ", dist); // cout << "F(" << dist << ") ";
	      else printf("O(%.2f) ", dist); // cout << "O(" << dist << ") ";

	    }
	  cout << endl;
	}
      cout << endl;
    }
}
