#include "Environment.hpp"
using namespace mpc;

#include <cmath>
#include <iostream>
#include <fstream>

using namespace std;

//#ifdef GAZEBO_SCENE
//#include <QtXml/QDomDocument>
//#include <QtXml/QDomElement>
//#include <QtCore/QFile>
//#endif


/********** Constructors and destructors **********/

// default constructor
Environment::Environment()
{
//   map_         = NULL;
//   lObs_        = new vector<ObsDisk2D*>;
//   lStObs_      = new vector<Obstacle2D*>;
  lObs2D_      = new vector<Obs2D*>;
  lFixedObs2D_ = new vector<Obs2D*>;

//   this->Xdflt_ = 400.0;
//   this->Ydflt_ = 400.0;

  this->minX_ = 0.0;
  this->minY_ = 0.0;
  this->maxX_ = 400.0;
  this->maxY_ = 400.0;
}

// copy constructor
Environment::Environment(const Environment& e)
{
//   map_ = e.map_;
  
  // copy the frame of the environment
  this->frame_ = PositionXYTh(e.frame_);
//   this->frame_ = PositionXYThT(e.frame_);

  // copy the list of 2D obstacles
  lObs2D_ = new vector<Obs2D*>;
  if(e.lObs2D_ != NULL)
    {
      for(unsigned int i=0; i<e.lObs2D_->size(); i++)
	{
	  lObs2D_->push_back(e.lObs2D_->at(i)->copy());
	}
    }

  // copy the list of 2D obstacles
  lFixedObs2D_ = new vector<Obs2D*>;
  if(e.lFixedObs2D_ != NULL)
    {
      for(unsigned int i=0; i<e.lFixedObs2D_->size(); i++)
	{
	  lFixedObs2D_->push_back(e.lFixedObs2D_->at(i)->copy());
	}
    }

//   this->Xdflt_ = e.Xdflt_;
//   this->Ydflt_ = e.Ydflt_;

  this->minX_ = e.minX_;
  this->minY_ = e.minY_;
  this->maxX_ = e.maxX_;
  this->maxY_ = e.maxY_;
}

/*
 *  \brief copy constructor
 *  \param e : the environment to copy
 *  \param frame : the new frame in which the environment will be set
 */
Environment::Environment(const Environment& e,const PositionXYTh& frame)
// Environment::Environment(const Environment& e,const PositionXYThT& frame)
{
  map_ = e.map_;
  
  // copy the frame of the environment
  this->frame_ = PositionXYTh(frame);
//   this->frame_ = PositionXYThT(frame);

  // compute the transformation to apply from the previous frame to the new frame
  PositionXYTh fTransfo(frame.x() - e.frame().x(),
			frame.y() - e.frame().y(),
			computeMinDiffAngle(frame.th(),e.frame().th()));
//   PositionXYThT fTransfo(frame.x() - e.frame().x(),
// 			 frame.y() - e.frame().y(),
// 			 computeMinDiffAngle(frame.th(),e.frame().th()),
// 			 frame.t() - e.frame().t());

  // copy the list of 2D obstacles
  lObs2D_ = new vector<Obs2D*>;
  if(e.lObs2D_ != NULL)
    {
      for(unsigned int i=0; i<e.lObs2D_->size(); i++)
	{
	  lObs2D_->push_back(e.lObs2D_->at(i)->copyInFrame(fTransfo));
	}
    }

  // copy the list of a priori known 2D obstacles
  lFixedObs2D_ = new vector<Obs2D*>;
  if(e.lFixedObs2D_ != NULL)
    {
      for(unsigned int i=0; i<e.lFixedObs2D_->size(); i++)
	{
	  lObs2D_->push_back(e.lFixedObs2D_->at(i)->copyInFrame(fTransfo));
	}
    }

  //////////////////////////////////////////////// to be changed!!! ///////////////////////////////////////////
  this->minX_ = e.minX_;
  this->minY_ = e.minY_;
  this->maxX_ = e.maxX_;
  this->maxY_ = e.maxY_;
}

// // init constructor
// Environment::Environment(Map* m)
// {
//   map_    = m;
//   lObs_   = new vector<ObsDisk2D*>;
//   lStObs_ = new vector<Obstacle2D*>;
//   this->Xdflt_ = 400.0;
//   this->Ydflt_ = 400.0;
// }

// // init constructor
// Environment::Environment(Map* m, vector<ObsDisk2D*> *l, vector<Obstacle2D*> *ls)
// {
//   map_    = m;
//   lObs_   = l;
//   lStObs_ = ls;
//   this->Xdflt_ = 400.0;
//   this->Ydflt_ = 400.0;
// }

// destructor
Environment::~Environment()
{
//   if(this->lObs_ != NULL)
//     {
//       for(unsigned int i=0; i<this->lObs_->size(); i++)
// 	{
// 	  ObsDisk2D* obs;
// 	  if((obs = this->getObs(i)) != NULL)
// 	    {
// 	      delete obs;
// 	    }
// 	}
//     }
//   if(this->lStObs_ != NULL)
//     {
//       for(unsigned int i=0; i<this->lStObs_->size(); i++)
// 	{
// 	  Obstacle2D* obs;
// 	  if((obs = this->getStObs(i)) != NULL)
// 	    {
// 	      delete obs;
// 	    }
// 	}
//     }
  if(this->lObs2D_ != NULL)
    {
      for(unsigned int i=0; i<this->lObs2D_->size(); i++)
	{
	  Obs2D* obs;
	  if((obs = lObs2D_->at(i)) != NULL)
	    {
	      delete obs;
	    }
	}
    }
  if(this->lFixedObs2D_ != NULL)
    {
      for(unsigned int i=0; i<this->lFixedObs2D_->size(); i++)
	{
	  Obs2D* obs;
	  if((obs = lFixedObs2D_->at(i)) != NULL)
	    {
	      delete obs;
	    }
	}
    }
}


/******************* Accessors ********************/

// // return the number of obstacles
// int Environment::getNbObs() const
// {
//   if(this->lObs_ == NULL)
//     {
//       return -1;
//     }
//   else
//     {
//       return this->lObs_->size();
//     }
// }

// // return the number of static obstacles
// int Environment::getNbStObs() const
// {
//   if(this->lStObs_ == NULL)
//     {
//       return -1;
//     }
//   else
//     {
//       return this->lStObs_->size();
//     }
// }

// // append a new obstacle
// void Environment::appendObs(ObsDisk2D* obs)
// {
//   this->lObs_->push_back(obs);
// }

// // append a new static obstacle
// void Environment::appendStObs(Obstacle2D* obs)
// {
//   this->lStObs_->push_back(obs);
// }

// // remove the i-th obstacle
// void Environment::removeObs(int i)
// {
//   if(this->lObs_ != NULL && this->getNbObs() > i)
//     {
//       if((*(this->lObs_))[i] != NULL) delete (*(this->lObs_))[i];
//       //this->lObs_->removeAt(i);
//       vector<ObsDisk2D*>::iterator it = this->lObs_->begin();
//     }
// }

// // remove the i-th static obstacle
// void Environment::removeStObs(int i)
// {
//   if(this->lStObs_ != NULL && this->getNbStObs() > i)
//     {
//       if((*(this->lStObs_))[i] != NULL) delete (*(this->lStObs_))[i];
//       //this->lObs_->removeAt(i);
//       vector<Obstacle2D*>::iterator it = this->lStObs_->begin();
//     }
// }

// // remove all obstacles of the environment
// void Environment::clearObs()
// {
//   if(this->lObs_ != NULL)
//     {
//       for(int i=this->getNbObs()-1; i>=0; i--)
// 	{
// 	  if((*(this->lObs_))[i] != NULL) delete (*(this->lObs_))[i];
// 	  this->lObs_->pop_back();
// 	}
//     }
// }

// // remove all static obstacles of the environment
// void Environment::clearStObs()
// {
//   if(this->lStObs_ != NULL)
//     {
//       for(int i=this->getNbStObs()-1; i>=0; i--)
// 	{
// 	  if((*(this->lStObs_))[i] != NULL) delete (*(this->lStObs_))[i];
// 	  this->lStObs_->pop_back();
// 	}
//     }
// }

// // return the i-th obstacle
// ObsDisk2D* Environment::getObs(int i) const
// {
//   return (*(this->lObs_))[i];
// }

// // return the i-th static obstacle
// Obstacle2D* Environment::getStObs(int i) const
// {
//   return (*(this->lStObs_))[i];
// }

// // return the current static map
// Map* Environment::getMap() const
// {
//   return this->map_;
// }

// // set the current map
// void Environment::setMap(Map* map)
// {
//   this->map_ = map;
// }

// // return the center of the environment
// const PMConfig Environment::getCenter() const
// {
//   PMConfig res;

//   if(this->map_ != NULL)
//     {
//       res.x() = (this->map_->getWidth() / 2.0f);
//       res.y() = (this->map_->getHeight() / 2.0f);      
//     }

//   return res;
// }

// return the dimensions of the environment
double Environment::getDim() const
{
  if(this->map_ == NULL)
    {
      return 0;
    }
  else
    {
      double wt = this->map_->getWidth();
      double ht = this->map_->getHeight();
      return sqrt(wt * wt + ht * ht);
    }
}

// return the default X value
double Environment::getDfltX() const
{
//   return this->Xdflt_;
  return (maxX_ - minX_);
}

// return the default Y value
double Environment::getDfltY() const
{
//   return this->Ydflt_;
  return (maxY_ - minY_);
}

// set the default X value
void Environment::setDfltX(double x)
{
//   this->Xdflt_ = x;
  minX_ = 0.0;
  maxX_ = x;
}

// set the default Y value
void Environment::setDfltY(double y)
{
//   this->Ydflt_ = y;
  minY_ = 0.0;
  maxY_ = y;
}

// load a new map
int Environment::loadMap(char* nomfic)
{
  if(map_ != NULL) delete map_;
  map_ = new Map();

  int resLoad = map_->load(nomfic);
  return resLoad;
}

/**
 *  \brief return the center of the environment
 *  \return the center of the environment
 */
PositionXY Environment::getCenter() const
{
  PositionXY locCenter = this->getLocalCenter();
  PositionXYTh locCenterXYTh(locCenter.x(),locCenter.y(),0.0);
  
  PositionXYTh CenterInF = coordFromFrame(locCenterXYTh,frame_);

  return CenterInF;
}

/*
 *  \brief return the center of the environment (in the local frame)
 *  \return the center of the environment (in the local frame)
 */
PositionXY Environment::getLocalCenter() const
{
  if(this->getNbObs2D() == 0 && this->getNbFixedObs2D() == 0)
    {
      return(PositionXY(0.5*(maxX_-minX_),0.5*(maxY_-minY_)));
    }
  else
    {
      int nbObs = 0;
      PositionXY center(0,0);

      if(this->lObs2D_ != NULL)
	{
	  int N = this->getNbObs2D();  
	  for(int i=0; i<N; i++)
	    {
	      PositionXYTh localPos = this->getLocalObs2D(i).pos();
	      PositionXYTh globalPos = mpc::coordFromFrame(localPos,frame_);

	      center.x() += globalPos.x();
	      center.y() += globalPos.y();

	      nbObs++;
	    }
	}
      if(this->lFixedObs2D_ != NULL)
	{
	  int N = this->getNbFixedObs2D();  
	  for(int i=0; i<N; i++)
	    {
	      PositionXYTh globalPos = this->getFixedObs2D(i).pos();

	      center.x() += globalPos.x();
	      center.y() += globalPos.y();

	      nbObs++;
	    }
	}

      center.x() /= ((double)nbObs);
      center.y() /= ((double)nbObs);
      return center;
    }
}

/*
 *  \brief return the radius of the environment
 *  \return the radius of the environment
 */
double Environment::getRadius() const
{
  if(this->getNbObs2D() == 0 && this->getNbFixedObs2D() == 0)
    {
      double xrad = 0.5*(maxX_-minX_);
      double yrad = 0.5*(maxY_-minY_);
      return sqrt(xrad*xrad + yrad*yrad);
    }
  else
    {
      double sqMaxDist = -1;

      PositionXY center = this->getLocalCenter();
      
      if(this->lObs2D_ != NULL)
	{
	  int N = this->getNbObs2D();  
	  for(int i=0; i<N; i++)
	    {
	      PositionXYTh localPos = this->getLocalObs2D(i).pos();
	      PositionXYTh globalPos = mpc::coordFromFrame(localPos,frame_);

	      double dx = globalPos.x() - center.x();
	      double dy = globalPos.y() - center.y();
	      double sqDist = dx*dx + dy*dy;
	      if(sqDist > sqMaxDist) sqMaxDist = sqDist;
	    }
	}
      if(this->lFixedObs2D_ != NULL)
	{
	  int N = this->getNbFixedObs2D();  
	  for(int i=0; i<N; i++)
	    {
	      PositionXYTh globalPos = this->getFixedObs2D(i).pos();

	      double dx = globalPos.x() - center.x();
	      double dy = globalPos.y() - center.y();
	      double sqDist = dx*dx + dy*dy;
	      if(sqDist > sqMaxDist) sqMaxDist = sqDist;
	    }
	}

      return (1.5*sqrt(sqMaxDist));
    }
}

/*
 * \brief return the number of 2D obstacles
 * \return the number of 2D obstacles
 */
int Environment::getNbObs2D() const
{
  if(lObs2D_ != NULL) return lObs2D_->size();
  else return 0;
}

/*
 *  \brief append a new 2D obstacle
 *  \param obs : the new obstacle
 */
void Environment::appendObs2D(const Obs2D& obs)
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  if(lObs2D_ != NULL) lObs2D_->push_back(obs.copy());
  else 
    throw string("Environment::appendObs2D : Undefined list of 2D obstacles");
}

/*
 *  \brief remove all 2D obstacles of the environment
 */
void Environment::clearObs2D()
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  if(lObs2D_ != NULL) lObs2D_->clear();
  else 
    throw string("Environment::clearObs2D : Undefined list of 2D obstacles");  
}

/**
 *  \brief return the i-th 2D obstacle in the global frame
 *  \param i : index of the obstacle to retrieve
 *  \return the i-th 2D obstacle in the global frame
 */
Obs2D* Environment::getObs2D(int i) const
{
  return this->getLocalObs2D(i).copyFromFrame(frame_);
}

/*
 *  \brief return the i-th 2D obstacle in the local frame
 *  \param i : index of the obstacle to retrieve
 *  \return the i-th 2D obstacle in the local frame
 */
const Obs2D& Environment::getLocalObs2D(int i) const
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  if(lObs2D_ != NULL) return *(lObs2D_->at(i));
  else 
    throw string("Environment::getObs2D : Undefined list of 2D obstacles");    
}

/*
 * \brief return the number of 2D obstacles
 * \return the number of 2D obstacles
 */
int Environment::getNbFixedObs2D() const
{
  if(lFixedObs2D_ != NULL) return lFixedObs2D_->size();
  else return 0;
}

/*
 *  \brief append a new 2D obstacle
 *  \param obs : the new obstacle
 */
void Environment::appendFixedObs2D(const Obs2D& obs)
{
  if(lFixedObs2D_ != NULL) lFixedObs2D_->push_back(obs.copy());
  else 
    throw string("Environment::appendFixedObs2D : Undefined list of 2D obstacles");
}

/*
 *  \brief remove all 2D obstacles of the environment
 */
void Environment::clearFixedObs2D()
{
  if(lFixedObs2D_ != NULL) lFixedObs2D_->clear();
  else 
    throw string("Environment::clearFixedObs2D : Undefined list of 2D obstacles");  
}

/*
 *  \brief return the i-th 2D obstacle
 *  \param i : index of the obstacle to retrieve
 *  \return the i-th 2D obstacle
 */
const Obs2D& Environment::getFixedObs2D(int i) const
{
  return this->getLocalFixedObs2D(i);
}

/*
 *  \brief return the i-th 2D obstacle in the local frame
 *  \param i : index of the obstacle to retrieve
 *  \return the i-th 2D obstacle in the local frame
 */
const Obs2D& Environment::getLocalFixedObs2D(int i) const
{
  if(lFixedObs2D_ != NULL) return *(lFixedObs2D_->at(i));
  else 
    throw string("Environment::getFixedObs2D : Undefined list of 2D obstacles");    
}


/*************** Environment update ***************/

/*
 *  \brief simulate a random motion of all dynamic obstacles
 *  \param the time at which the environment is evaluated
 */
void Environment::simuRandMotion(double t)
{
  // verify if the configuration is free with the static obstacles
  if(this->lObs2D_ != NULL)
    {
      // move each obstacle of the environment
      int N = this->getNbObs2D();
      for(int i=0; i<N; i++)
	{
	  lObs2D_->at(i)->simuRandMotion(t);
	}

      // make the obstacles bounce on the walls 
      // of the environment if they move out of it
      double minX = minX_;
      double maxX = maxX_;
      double minY = minY_;
      double maxY = maxY_;
      for(int i=0; i<N; i++)
	{
	  Obs2D* curObs = lObs2D_->at(i);
	  if(curObs->lastPosT().x() < minX)
	    {
		  curObs->lastPosT().x() = minX;
		  curObs->lastSpd().x() *= -1;
	    }
	  else if(curObs->lastPosT().x() > maxX)
	    {
		  curObs->lastPosT().x() = maxX;
		  curObs->lastSpd().x() *= -1;
	    }
	  if(curObs->lastPosT().y() < minY)
	    {
		  curObs->lastPosT().y() = minY;
		  curObs->lastSpd().y() *= -1;
	    }
	  else if(curObs->lastPosT().y() > maxY)
	    {
		  curObs->lastPosT().y() = maxY;
		  curObs->lastSpd().y() *= -1;
	    }
	}
    }
}


/************* Obstruction functions **************/

// // return true if the obstacle obstruct the config c
// bool Environment::obstruct(const Config& c)
// {
//   // verify if the configuration is free with the static obstacles
//   if(this->lStObs_ != NULL)
//     {
//       int N = this->getNbStObs();
//       for(int i=0; i<N; i++)
// 	{
// 	  if(this->getStObs(i)->obstruct(c))
// 	    {
// 	      return true;
// 	    }
// 	}
//     }
//   return false;
// }

// // return true if the obstacles obstruct the line between c1 and c2
// bool Environment::obstruct(const Config& c1,const Config& c2)
// {
//   // verify if the configuration is free with the static obstacles
//   if(this->lStObs_ != NULL)
//     {
//       int N = this->getNbStObs();
//       for(int i=0; i<N; i++)
// 	{
// 	  if(this->getStObs(i)->obstruct(c1,c2))
// 	    {
// 	      return true;
// 	    }
// 	}
//     }
//   return false;
// }

// /**
//  *  \brief test if the environment obstruct the position pos
//  *  \param pos : the position to test
//  */
// bool Environment::obstruct(const PositionXY pos,double t)
// {
//   cerr << "Environment::obstruct : Not implemented anymore!" << endl;
//   exit(1);
// }

// /**
//  *  \brief test if the environment obstruct the position pos
//  *  \param pos : the position to test
//  *  \param near : the nearest position of the obstacle found
//  *  \param dist : the minimal distance between pos and the nearest obstacle
//  */
// bool Environment::obstruct(const PositionXY pos,PositionXY& near,double& dist)
// {
// //   boost::mutex::scoped_lock lock(mutexEnv);

//   // init the distance to an huge value
//   dist = 999999999;

//   // verify if the position is free with the static obstacles
//   bool trouve = false;

//   cerr << "Environment::obstruct : Not implemented anymore!" << endl;
//   exit(1);

// //   if(this->lStObs_ != NULL)
// //     {
// //       int N = this->getNbStObs();

// //       for(int i=0; i<N; i++)
// // 	{
// // 	  double curDist;
// // 	  PositionXY curNear;
// // 	  if(this->getStObs(i)->obstruct(pos,curNear,curDist))
// // 	    {
// // 	      trouve = true;
// // 	    }
// // 	  if(i == 0 || curDist < dist)
// // 	    {
// // 	      near = curNear;
// // 	      dist = curDist;
// // 	    }
// // 	}
// //     }
//   return trouve;
// }

// /**
//  *  \brief test if the environment obstruct the position pos
//  *  \param pos : the position to test
//  *  \param t : the input time at which to test collisions
//  *  \param near : the nearest position of the obstacle found
//  *  \param dist : the minimal distance between pos and the nearest obstacle
//  */
// bool Environment::obstruct(const PositionXY pos,double t,PositionXY& near,double& dist)
// {
// //   boost::mutex::scoped_lock lock(mutexEnv);

//   // init the distance to an huge value
//   dist = 999999999;
//   bool trouve = false;

// //   // verify if the position is free with the static obstacles
// //   if(this->lStObs_ != NULL)
// //     {
// //       int N = this->getNbStObs();

// //       for(int i=0; i<N; i++)
// // 	{
// // 	  double curDist;
// // 	  PositionXY curNear;
// // 	  if(this->getStObs(i)->obstruct(pos,curNear,curDist))
// // 	    {
// // 	      trouve = true;
// // 	    }
// // 	  if(i == 0 || curDist < dist)
// // 	    {
// // 	      near = curNear;
// // 	      dist = curDist;
// // 	    }
// // 	}
// //     }

//   // verify if the position is free with the 2D obstacles
//   if(this->lObs2D_ != NULL)
//     {
//       int N = this->getNbObs2D();

//       for(int i=0; i<N; i++)
// 	{
// 	  double curDist;
// 	  PositionXY curNear;
// 	  if(this->getObs2D(i).obstruct(pos,t,curNear,curDist))
// 	    {
// 	      trouve = true;
// 	    }
// 	  if(curDist < dist)
// 	    {
// 	      near = curNear;
// 	      dist = curDist;
// 	    }
// 	}
//     }
//   return trouve;
// }

/*
 *  \brief test if the environment obstruct the position pos
 *  \param pos : the position to test
 *  \param t : the input time at which to test collisions
 */
bool Environment::obstruct(const PositionXY& pos,double t) const
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  // verify if the position is free with the 2D moving obstacles
  if(this->lObs2D_ != NULL)
    {
      int N = this->getNbObs2D();
      bool found = false;

      for(int i=0; i<N; i++)
	{
	  double curDist;
	  PositionXY curNear;
	  
	  // get back the current obstacle in the global frame
	  Obs2D* curObs = this->getObs2D(i);

	  if(curObs != NULL)
	    {
	      if(curObs->obstruct(pos,t)) found = true;
	      delete curObs;
	      if(found) return true;
	    }
	}
    }

  // verify if the position is free with the 2D static obstacles
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  double curDist;
	  PositionXY curNear;
	  if(this->getFixedObs2D(i).obstruct(pos,t))
	    {
	      return true;
	    }
	}
    }

  return false;
}

/*
 *  \brief test if the environment obstruct the position pos
 *  \param pos : the position to test
 *  \param t : the input time at which to test collisions
 *  \param near : the nearest position of the obstacle found
 *  \param dist : the minimal distance between pos and the nearest obstacle
 */
bool Environment::obstruct(const PositionXY& pos,double t,PositionXY& near,double& dist) const
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  // init the distance to an huge value
  dist = 999999999;
  bool trouve = false;

  // verify if the position is free with the 2D moving obstacles
  if(this->lObs2D_ != NULL)
    {
      int N = this->getNbObs2D();

      for(int i=0; i<N; i++)
	{
	  double curDist;
	  PositionXY curNear;

	  // get back the current obstacle in the global frame
	  Obs2D* curObs = this->getObs2D(i);

	  if(curObs != NULL)
	    {
	      if(curObs->obstruct(pos,t,curNear,curDist)) trouve = true;
	      delete curObs;

	      if(curDist < dist)
		{
		  near = curNear;
		  dist = curDist;
		}
	    }
	}
    }

  // verify if the position is free with the 2D static obstacles
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  double curDist;
	  PositionXY curNear;
	  if(this->getFixedObs2D(i).obstruct(pos,t,curNear,curDist))
	    {
	      trouve = true;
	    }
	  if(curDist < dist)
	    {
	      near = curNear;
	      dist = curDist;
	    }
	}
    }

  return trouve;
}

/**
 *  \brief test if the environment obstruct the segment p1p2
 *  \param p1 : the first end position-time of the segment
 *  \param p2 : the second end position-time of the segment
 */
bool Environment::obstructSgmt(const PositionXYT& p1,const PositionXYT& p2) const
{
}

/**
 *  \brief test if the static environment obstruct the segment p1p2
 *  \param p1 : the first end position of the segment
 *  \param p2 : the second end position of the segment
 *  \param near : the nearest position of the obstacle found
 *  \param dist : the minimal distance between pos and the nearest obstacle
 */
bool Environment::obstructSgmt(const PositionXYT& p1,const PositionXYT& p2,PositionXYT& near,double& dist) const
{
}

/*
 *  \brief test if the environment obstruct the position pos
 *  \param pos : the position to test
 *  \param near : the nearest position of the obstacle found
 *  \param dist : the minimal distance between pos and the nearest obstacle
 */
bool Environment::staticObstruct(const PositionXY pos,PositionXY& near,double& dist)
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  // init the distance to an huge value
  dist = 999999999;
  bool trouve = false;

  // verify if the position is free with the 2D static obstacles
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  double curDist;
	  PositionXY curNear;
	  if(this->getFixedObs2D(i).obstruct(pos,0,curNear,curDist))
	    {
	      trouve = true;
	    }
	  if(curDist < dist)
	    {
	      near = curNear;
	      dist = curDist;
	    }
	}
    }

  return trouve;
}

/*
 *  \brief test if the static environment obstruct the position pos
 *  \param pos : the position to test
 */
bool Environment::staticObstruct(const PositionXY& pos) const
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  // verify if the position is free with the 2D static obstacles
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  double curDist;
	  PositionXY curNear;
	  if(this->getFixedObs2D(i).obstruct(pos))
	    {
	      return true;
	    }
	}
    }

  return false;
}

/*
 *  \brief test if the static environment obstruct the segment p1p2
 *  \param p1 : the first end position of the segment
 *  \param p2 : the second end position of the segment
 *  \param t : the input time at which to test collisions
 */
bool Environment::staticObstructSgmt(const PositionXY& p1,const PositionXY& p2) const
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  // verify if the position is free with the 2D static obstacles
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  double curDist;
	  PositionXY curNear;
	  if(this->getFixedObs2D(i).obstruct(p1,p2))
	    {
	      return true;
	    }
	}
    }

  return false;
}

/*
 *  \brief test if the static environment obstruct the segment p1p2
 *  \param p1 : the first end position of the segment
 *  \param p2 : the second end position of the segment
 *  \param t : the input time at which to test collisions
 *  \param near : the nearest position of the obstacle found
 *  \param dist : the minimal distance between pos and the nearest obstacle
 */
bool Environment::staticObstructSgmt(const PositionXY& p1,const PositionXY& p2,PositionXY& near,double& dist) const
{
//   boost::mutex::scoped_lock lock(mutexEnv);

  // init the distance to an huge value
  dist = 999999999;
  bool trouve = false;

  // verify if the position is free with the 2D static obstacles
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  double curDist;
	  PositionXY curNear;
	  if(this->getFixedObs2D(i).obstruct(p1,p2,0,curNear,curDist))
	    {
	      trouve = true;
	    }
	  if(curDist < dist)
	    {
	      near = curNear;
	      dist = curDist;
	    }
	}
    }

  return trouve;
}


/*************** Sampling functions ***************/

/*
 *  \brief sample the obstructed positions of the obstacle of the static environment
 *  \param sampleX,sampleY : sampling scale of the resulting map
 *  \return an occupancy grid of the environment
 *
 *  A new instance of occupancy grid is returned
 */
Map* Environment::sample(int sampleX,int sampleY) const
{
  // create the resulting map
  Map* map = new Map(sampleX,sampleY,255);
  map->setCompletelyFree();

  for(int i=0; i<this->getNbFixedObs2D(); i++)
    {
      // compute a sampling of the obstruction of the current obstacle
      this->getFixedObs2D(i).sample(map,minX_,maxX_,minY_,maxY_,sampleX,sampleY);
    }
  
  // set the borders of the map as obstructed
  map->setBordersFreedom(0.0);

  // return the resulting map
  return map;
}


/****************** IO functions ******************/

/*
 *  \brief load a segmented environment from an input file
 *  \param filename : the input filename of the environment to load
 *  \param dist : the secure distance considered to build boxes around the obstacles
 *  \return 0 if the process succeed, -1 else
 */
int Environment::loadSegmentedEnv(const char* filename,double dist)
{
  ifstream file(filename);

  // check if the file has been opened
  if(!file) return -1;

  // check the extension of the file
  int mode = -1;
  if(mpc::isExtFilename(filename,"map")) { mode = 1; cout << "extension : map" << endl; }
  if(mpc::isExtFilename(filename,"map2")) { mode = 2; cout << "extension : map2" << endl; }

  double minX=999999999;
  double minY=999999999;
  double maxX=-999999999;
  double maxY=-999999999;

  char buf[255];
  while(file.good())
    {
      file.getline(buf, 255);

      double x0, y0, x1, y1;
      double c1, c2, c3, c4, c5, c6;
      char endLine[255];
      int id;

      if(mode == 1)
	{
	  int ret = sscanf(buf,
			   "Map.InsertFeature segment %d / %lf %lf %lf %lf / %lf %lf %lf %lf %lf %lf / ~",
			   &id,
			   &x0, &y0, &x1, &y1,
			   &c1, &c2, &c3, &c4, &c5, &c6);
	  if(ret == 11)
	    {
	      // compute both ends positions of the segment in the input frame
	      // 	  PositionXYTh pos1 = coordInFrame(PositionXYTh(x0,y0,0),frame);
	      // 	  PositionXYTh pos2 = coordInFrame(PositionXYTh(x1,y1,0),frame);
	      PositionXY pos1(x0,y0);
	      PositionXY pos2(x1,y1);
	      
	      // create a new obstacle from the input segment
	      BBoxObs2D obs(pos1,pos2,dist);
	      
	      // append the new obstacle to the environment
	      this->appendFixedObs2D(obs);
	      
	      cout << "segment loaded : " << id << " " << x0 << " " << y0 << " " << x1 << " " << y1 << endl;  
	      
	      if(x0 < minX) minX = x0;
	      if(x1 < minX) minX = x1;
	      if(y0 < minY) minY = y0;
	      if(y1 < minY) minY = y1;
	      if(x0 > maxX) maxX = x0;
	      if(x1 > maxX) maxX = x1;
	      if(y0 > maxY) maxY = y0;
	      if(y1 > maxY) maxY = y1;
	    }
	  else
	    {
	      cout << "unable to parse line : " << buf << endl;
	    }
	}
      else if(mode == 2)
	{
	  int ret = sscanf(buf,
			   "    Segment p1=%lf %lf p2=%lf %lf ~",
			   &x0, &y0, &x1, &y1);
	  if(ret == 4)
	    {
	      // compute both ends positions of the segment in the input frame
	      // 	  PositionXYTh pos1 = coordInFrame(PositionXYTh(x0,y0,0),frame);
	      // 	  PositionXYTh pos2 = coordInFrame(PositionXYTh(x1,y1,0),frame);
	      PositionXY pos1(x0,y0);
	      PositionXY pos2(x1,y1);
	      
	      // create a new obstacle from the input segment
	      BBoxObs2D obs(pos1,pos2,dist);
	      
	      // append the new obstacle to the environment
	      this->appendFixedObs2D(obs);
	      
	      cout << "segment loaded : " << id << " " << x0 << " " << y0 << " " << x1 << " " << y1 << endl;  
	      
	      if(x0 < minX) minX = x0;
	      if(x1 < minX) minX = x1;
	      if(y0 < minY) minY = y0;
	      if(y1 < minY) minY = y1;
	      if(x0 > maxX) maxX = x0;
	      if(x1 > maxX) maxX = x1;
	      if(y0 > maxY) maxY = y0;
	      if(y1 > maxY) maxY = y1;
	    }
	  else
	    {
	      ret = sscanf(buf,
			   "    Segment id=%d p1=%lf %lf p2=%lf %lf cov1=%lf %lf %lf cov2=%lf %lf %lf ~",
			   &id,
			   &x0, &y0, &x1, &y1,
			   &c1, &c2, &c3, &c4, &c5, &c6);

	      if(ret == 11)
		{
		  PositionXY pos1(x0,y0);
		  PositionXY pos2(x1,y1);
		  
		  // create a new obstacle from the input segment
		  BBoxObs2D obs(pos1,pos2,dist);
		  
		  // append the new obstacle to the environment
		  this->appendFixedObs2D(obs);
		  
		  cout << "segment loaded : " << id << " " << x0 << " " << y0 << " " << x1 << " " << y1 << endl;  
		  
		  if(x0 < minX) minX = x0;
		  if(x1 < minX) minX = x1;
		  if(y0 < minY) minY = y0;
		  if(y1 < minY) minY = y1;
		  if(x0 > maxX) maxX = x0;
		  if(x1 > maxX) maxX = x1;
		  if(y0 > maxY) maxY = y0;
		  if(y1 > maxY) maxY = y1;
		}
	      else
		{
		  cout << "unable to parse line : " << buf << endl;
		}
	    }
	}
    }

  this->minX_ = minX;
  this->maxX_ = maxX;
  this->minY_ = minY;
  this->maxY_ = maxY;

  return 0;
}


/*************** Display functions ****************/

// /*
//  *  \brief draw the environment in an openGL viewer
//  *  \param dispFM : true if the forecast model of the dynamic obstacles must be displayed
//  *  \param time : the time at which the environment must be displayed
//  */
// void Environment::toGL(bool dispFM,double time) const
// {
//   // display each obstacle of the environment
//   if(this->lObs2D_ != NULL)
//     {
//       int N = this->getNbObs2D();

//       for(int i=0; i<N; i++)
// 	{
// 	  this->getLocalObs2D(i).toGL(frame_,dispFM);
// 	}
//     }

//   // display each a priori known obstacle of the environment
//   if(this->lFixedObs2D_ != NULL)
//     {
//       int N = this->getNbFixedObs2D();

//       for(int i=0; i<N; i++)
// 	{
// 	  this->getLocalFixedObs2D(i).toGL(PositionXYThT(0,0,0,time));
// 	}
//     }
// }

/*
 *  \brief draw the environment in an openGL viewer
 *  \param frame : the frame of the perception system from which is construct the environment model
 *  \param dispFM : true if the forecast model of the dynamic obstacles must be displayed
 */
void Environment::toGL(const PositionXYTh& frame,bool dispFM) const
// void Environment::toGL(const PositionXYThT& frame,bool dispFM) const
{
  // display each obstacle of the environment
  if(this->lObs2D_ != NULL)
    {
      int N = this->getNbObs2D();

      for(int i=0; i<N; i++)
	{
	  this->getLocalObs2D(i).toGL(frame,dispFM);
	}
    }

  // display each a priori known obstacle of the environment
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  //this->getFixedObs2D(i).display();
	  PositionXYThT frameT(frame.x(),frame.y(),frame.th(),0);
	  this->getLocalFixedObs2D(i).toGL(frameT,false);
	}
    }
}

/*
 *  \brief draw the environment in an openGL viewer
 *  \param dispFM : true if the forecast model of the dynamic obstacles must be displayed
 *  \param curTime : the current time at which must be displayed the static environment
 */
void Environment::toGL(bool dispFM,double curTime) const
{  
  // display each obstacle of the environment
  if(this->lObs2D_ != NULL)
    {
      int N = this->getNbObs2D();

      for(int i=0; i<N; i++)
	{
	  this->getLocalObs2D(i).toGL(frame_,dispFM);
	}
    }

  // display each a priori known obstacle of the environment
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  this->getLocalFixedObs2D(i).toGL(PositionXYThT(0,0,0,curTime));
	}
    }
}

/*
 *  \brief display the main features of the environment
 */
void Environment::display() const
{
  // display each obstacle of the environment
  cout << "Dynamic obstacles (in local frame) : " << endl;
  if(this->lObs2D_ != NULL)
    {
      int N = this->getNbObs2D();

      for(int i=0; i<N; i++)
	{
	  this->getLocalObs2D(i).display();
	}
    }

  cout << "Static obstacles (in local frame) : " << endl;
  if(this->lFixedObs2D_ != NULL)
    {
      int N = this->getNbFixedObs2D();

      for(int i=0; i<N; i++)
	{
	  this->getLocalFixedObs2D(i).display();
	}
    }
}

/**
 *  \brief display the number of obstacles of the environment
 */
void Environment::displayNbObs() const
{
  if(this->lObs2D_ != NULL) cout << "Dynamic obstacles : " << this->getNbObs2D() << " obstacles" << endl;
  else cout << "No dynamic obstacles" << endl;

  if(this->lFixedObs2D_ != NULL) cout << "Static obstacles : " << this->getNbFixedObs2D() << " obstacles" << endl;
  else cout << "No static obstacles" << endl;
}


// #ifdef GAZEBO_SCENE
// /*************** Sampling functions ***************/

// // load the environment from an xml file (gazebo world)
// // return status : 
// //  0 : the environment has been correctly loaded
// // -1 : the input file cannot be opened
// // -2 : the document cannot be loaded 
// // -3 : the input file is not a gazebo world or has an incorrect format

// int Environment::loadFromXMLFile(char* filename)
// {
//   QDomDocument doc( "gazeboSceneXML" );
//   QString strFilename(filename);
//   QFile file;
//   file.setFileName (strFilename);

//   // open the input file
//   if ( !file.open( QIODevice::ReadOnly ) )
//     return -1;
  
//   // fill the document
//   if ( !doc.setContent( &file ) ) {
//     file.close();
//     return -2;
//   }
//   file.close();

//   // get back the root of the document
//   QDomElement docElem = doc.documentElement();

//   // check if the input file is a gazebo world
//   if(docElem.tagName() != QString("gz:world"))
//     {
//       cout << "Environment::loadFromXMLFile : Warning. The input file is not a gazebo world. Cannot load the environment." << endl;
//       return -3;
//     }
  
//   // travel the document to create each element of the environment
  
// }
// #endif
