#ifndef MATHFUNC_HPP
#define MATHFUNC_HPP

#include <cmath>
#include "fresnel.hpp"
#include "Vec.hpp"

class MathFunc
{
public : 
  /*********** Geometric transformations ************/
  static Vec translate(Vec pos,Vec v);                // return the result of the translation of pos from vector v
  static Vec rotate(Vec pos,Vec c,double th);         // return the result of the rotation of pos around c from angle th

  /*************** Clothoid functions ***************/
  static double ClothX(double lg,double sigma);       // return x clothoid value from curve abscica lg and curve speed sig
  static double ClothY(double lg,double sigma);       // return x clothoid value from curve abscica lg and curve speed sig

  /************** Hermite interpolation *************/
  static double hermiteItp(double P1,double P2,double T1,double T2,double S);
  // compute the hermite interpolation from point and tangent (P1,T1) to point and tangent (P2,T2) at curve abscica S

  /*************** Distances computing **************/
  static double maxTravDist(double initSp,double maxSp,double maxAcc,double dt);
  // compute the maximal travelable distance from initial speed initSp during  
  // a time dt with speed and acceleration constraints maxSp and maxAcc
  static double minTravDist(double initSp,double minSp,double minAcc,double dt);
  // compute the minimal travelable distance from initial speed initSp during  
  // a time dt with speed and acceleration constraints minSp and minAcc
};

#endif
