#include "math/inc/PascalTriangle.hpp"
using namespace mpc;

#include <stdlib.h>
#include <iostream>
using namespace std;


                                    /**************************************************/
                                    /************** Class PascalTriangle **************/
                                    /**************************************************/

int ** PascalTriangle::tab_ = NULL;
int PascalTriangle::maxOrder_ = -1;

/**
 *  \brief Build the Pascal's triangle n-th order
 *  \param n : the order of the Pascal's triangle that must be computed
 */
void PascalTriangle::buildOrder(int n)
{
  if(n == 0) tab_[0][0] = 1;
  else
    {
      for(int i=0; i<=n; i++)
	{
	  if(i == 0 || i == n) tab_[n][i] = 1;
	  else tab_[n][i] = tab_[n-1][i-1] + tab_[n-1][i];
	}
    }
}

/**
 *  \brief Build the Pascal's triangle up to order n
 *  \param n : the order up to which the Pascal's triangle must be computed
 */
void PascalTriangle::buildUpToOrder(int n)
{
  // clean the tables if needed
  PascalTriangle::cleanTables();

  // allocate the table up to order n
  PascalTriangle::allocTablesUpToOrder(n);

  // build the Pascal's triangle table up to order n
  for(int i=0; i<=n; i++) PascalTriangle::buildOrder(i);
}

/**
 *  \brief return the value of the Pascal's triangle at order i, j-th value
 *  \param i : number of the Pascal's triangle order
 *  \param j : number of the value to compute
 *  \return the value of the Pascal's triangle at order i, j-th value
 *  \return -1 if the input index is not valid
 */
int PascalTriangle::getVal(int i, int j)
{
  if(PascalTriangle::indexIsValid(i, j))
    {
      return tab_[i][j];
    }
  else return -1;
}

/*
 *  \brief check if the input index is valid
 *  \return true if the input index is valid, false otherwise
 */
bool PascalTriangle::indexIsValid(int i, int j)
{
  if(i >= 0 && j >= 0 && i >= j) return true;
  else return false;
}

/*
 *  \brief check if the input order has already been computed
 *  \return true if the input order has already been computed, false otherwise
 */
bool PascalTriangle::orderAlreadyComputed(int i)
{
  if(i >= 0 && i <= PascalTriangle::maxOrder_) return true;
  else return false;
}

/*
 *  \brief clean the Pascal's triangle tables
 */
void PascalTriangle::cleanTables()
{
  if(PascalTriangle::tab_ != NULL) 
    {
      for(int i=0; i<PascalTriangle::maxOrder_; i++) delete [] tab_[i];
      delete [] PascalTriangle::tab_;

      PascalTriangle::tab_ = NULL;
      PascalTriangle::maxOrder_ = -1;
    }
}

/*
 *  \brief allocate the memory required to build the Pascal's triangle up to order n
 */
void PascalTriangle::allocTablesUpToOrder(int n)
{
  if(n >= 0)
    {
      // allocate the table
      PascalTriangle::tab_ = new int* [n+1];
      for(int i=0; i<=n; i++)
	{
	  PascalTriangle::tab_[i] = new int [i+1];
	}

      // set the current order
      PascalTriangle::maxOrder_ = n;
    }
}

/**
 *  \brief display the current Pascal's triangle table
 */
void PascalTriangle::display()
{
  if(PascalTriangle::maxOrder_ < 0)
    {
      cout << "The Pascal's triangle hasn't been comuted yet" << endl;
    }
  else
    {
      cout << "Pascal's triangle: " << endl;
      for(int i=0; i<=PascalTriangle::maxOrder_; i++)
	{
	  cout << i << ": ";
	  for(int j=0; j<=i; j++) cout << PascalTriangle::tab_[i][j] << " ";
	  cout << endl;
	}
    }
}
