#include "Polynomial.hpp"

#include "Utils.hpp"
#include "utilsLuis.h"
using namespace mpc;

#include <cmath>
#include <iostream>
using namespace std;

#include "math/inc/PascalTriangle.hpp"


/********** Constructors and destructors **********/

// default constructor
Polynomial::Polynomial() : Function()
{
  this->coef_ = vector<double>();
  this->coef_.push_back(0);

//   this->t0_ = 0;

  this->type_ = POLY_FUNC;
}

// copy constructor
Polynomial::Polynomial(const Polynomial& p) : Function()
{
  this->coef_ = vector<double>();
  for(int i=0; i<=p.order(); i++)
    {
      this->coef_.push_back(p.getCoef(i));
    }

//   this->t0_ = p.t0_;

  this->type_ = POLY_FUNC;
}

// init constructor
Polynomial::Polynomial(vector<double>& c) : Function()
{
  this->coef_ = c;
//   this->t0_ = 0;

  this->type_ = POLY_FUNC;
}

// init constructor (degree of the polynomial)
Polynomial::Polynomial(int deg)
{
  this->coef_ = vector<double>();
  for(int i=0; i<=deg; i++)
    {
      this->coef_.push_back(0);
    }

//   this->t0_ = 0;

  this->type_ = POLY_FUNC;
}

// init constructor
Polynomial::Polynomial(double cste) : Function()
{
  this->coef_ = vector<double>();
  this->coef_.push_back(cste);

//   this->t0_ = 0;

  this->type_ = POLY_FUNC;
}

// init constructor
Polynomial::Polynomial(double A,double B) : Function()
{
  this->coef_ = vector<double>();
  this->coef_.push_back(B);
  this->coef_.push_back(A);

//   this->t0_ = 0;

  this->type_ = POLY_FUNC;
}

// init constructor
Polynomial::Polynomial(double A,double B,double C) : Function()
{
  this->coef_ = vector<double>();
  this->coef_.push_back(C);
  this->coef_.push_back(B);
  this->coef_.push_back(A);

//   this->t0_ = 0;

  this->type_ = POLY_FUNC;
}

// init constructor
Polynomial::Polynomial(double A,double B,double C,double D) : Function()
{
  this->coef_ = vector<double>();
  this->coef_.push_back(D);
  this->coef_.push_back(C);
  this->coef_.push_back(B);
  this->coef_.push_back(A);

//   this->t0_ = 0;

  this->type_ = POLY_FUNC;
}

// destructor
Polynomial::~Polynomial()
{
  this->coef_.clear();
}

/**
 *  \brief return a copy of the current function
 *  \return a copy of the current function
 *
 *  A new instance of function is allocated and returned here
 */
Polynomial* Polynomial::copy() const
{
  Polynomial* res = new Polynomial();

  res->coef_ = vector<double>();
  for(int i=0; i<=this->order(); i++)
    {
      res->coef_.push_back(this->getCoef(i));
    }
  
//   res->t0_ = this->t0_;

  return res;
}


/**************** Setters & Getters ***************/

// return the i-th coefficient
double Polynomial::getCoef(int i) const
{
  if(i <= this->order())
    return (this->coef_)[i];
  else return 0.0;
}

// set the i-th coefficient
void Polynomial::setCoef(int i,double c)
{
  (this->coef_)[i] = c;
}

// return the value of the polynomial at a given parameter t
double Polynomial::getVal(double t) const
{
  //cout << "t : " << t << " ";
  if(t >= binf_ && t <= bsup_)
    {
      double res = 0;
      for(int i=0; i<=this->order(); i++)
	{
// 	  res += (coef_[i] * pow(t - t0_,i));
	  res += (coef_[i] * pow(t, i));
	}
      return res;
    }
  else return 0;
}

// // return the order of the polynomial
// int Polynomial::order() const
// {
//   return this->coef_.size() - 1;
// }


/******************** Operators *******************/

// return the sum of a Polynomial and a constant
Function* Polynomial::operator+(double v) const
{
  Polynomial* pres = new Polynomial(*this);
  pres->setCoef(0,this->getCoef(0) + v);
  return pres;
}

// return the sum of two Polynomials
Function* Polynomial::operator+(const Polynomial& p2) const
{
  int maxorder = max(this->order(),p2.order());
  Polynomial* pres = new Polynomial(maxorder);
  for(int i=0; i<=maxorder; i++)
    {
      pres->setCoef(i,this->getCoef(i) + p2.getCoef(i));
    }
//   pres->t0_ = this->t0_;

  return pres;
}

// return the sum of a Polynomial and a MultiItvFunction
Function* Polynomial::operator+(const MultiItvFunction& f2) const
{
  cerr << "Polynomial::operator+(const MultiItvFunction& f2) : Not implemented yet!!!!" << endl;
  exit(1);

  return new MultiItvFunction();
}

// return the product of a Polynomial and a constant
Function* Polynomial::operator*(double v) const
{
  if(v == 0) return new Polynomial();
  else
    {
      Polynomial* pres = new Polynomial(this->order());
      for(int i=0; i<=this->order(); i++)
	{
	  pres->setCoef(i,this->getCoef(i) * v);
	}
//       pres->t0_ = this->t0_;

      return pres;
    }
}

// return the product of two Polynomials
Function* Polynomial::operator*(const Polynomial& p2) const
{
//   cout << "Polynomial::operator*(const Polynomial& p2) : Does not work anymore... Must be updated..." << endl;
//   exit(1);

  Polynomial* pres = new Polynomial(this->order() + p2.order());
  for(int i=0; i<=this->order(); i++)
    {
      for(int j=0; j<=p2.order(); j++)
	{
	  pres->setCoef(i+j, pres->getCoef(i+j) + (this->getCoef(i) * p2.getCoef(j)));
	}
    }
  return pres;
}

// return the product of a Polynomial and a MultiItvFunction
Function* Polynomial::operator*(const MultiItvFunction& f2) const
{
  cerr << "Polynomial::operator*(const MultiItvFunction& f2) : Not implemented yet!!!!" << endl;
  exit(1);

  return new MultiItvFunction();
}

// return the difference of a Polynomial and a constant
Function* Polynomial::operator-(double v) const
{
  Polynomial* pres = new Polynomial(*this);
  pres->setCoef(0,this->getCoef(0) - v);
  return pres;
}

// return the difference of two Polynomials  
Function* Polynomial::operator-(const Polynomial& p2) const
{
//   cout << "Polynomial::operator-(const Polynomial& p2) : Does not work anymore... Must be updated..." << endl;
//   exit(1);

  int maxorder = max(this->order(),p2.order());
  Polynomial* pres = new Polynomial(maxorder);
  for(int i=0; i<=maxorder; i++)
    {
      pres->setCoef(i,this->getCoef(i) - p2.getCoef(i));
    }
  return pres;
}

// return the difference of a Polynomial and a MultiItvFunction
Function* Polynomial::operator-(const MultiItvFunction& f2) const
{
  cerr << "Polynomial::operator-(const MultiItvFunction& f2) : Not implemented yet!!!!" << endl;
  exit(1);

  return new MultiItvFunction();
}

// multiply the current Polynomial by a constante
void Polynomial::operator*=(double v)
{
  if(v == 0)
    {
      this->coef_.clear();
      this->coef_.push_back(0);
    }
  else
    {
      for(int i=0; i<=this->order(); i++)
	{
	  this->setCoef(i,this->getCoef(i) * v);
	}
    }
}

// multiply the current Polynomial by an other Polynomial
void Polynomial::operator*=(const Polynomial& f2)
{
  // build a copy of the current polynomial
  Polynomial ptmp =  Polynomial(*this);

  // reinitialize the current polynomial
  this->coef_.clear();
  int newOrder = ptmp.order() + f2.order();
  for(int i=0; i<=newOrder; i++)
      this->coef_.push_back(0);

  for(int i=0; i<=ptmp.order(); i++)
    {
      for(int j=0; j<=f2.order(); j++)
	{
	  this->setCoef(i+j, this->getCoef(i+j) + (ptmp.getCoef(i) * f2.getCoef(j)));
	}
    }
}

// multiply the current Polynomial by a MultiItvFunction
void Polynomial::operator*=(const MultiItvFunction& f2)
{
  cout << "This cannot be done : The result is not a Polynomial" << endl;
  exit(1);
}


/************ Integration & Derivation ************/

// return the derivation of the current polynome
Function* Polynomial::derivation() const 
{
//   cout << "poly::deriv 1" << endl;
  if(this->order() == 0)
    {
//       cout << "poly::deriv 2" << endl;
      return new Polynomial();
    }
  else
    {
      Polynomial* deriv = new Polynomial(this->order()-1);
      for(int i=1; i<=this->order(); i++)
	{
	  deriv->setCoef(i-1, this->getCoef(i) * i);
	}
//       cout << "poly::deriv 3" << endl;
      return deriv;
    }
}

// return the integration of the current polynome
Function* Polynomial::integration() const 
{
  if(this->order() == 0 && this->getCoef(0) == 0)
    return new Polynomial();
  else
    {
      Polynomial* itg = new Polynomial(this->order()+1);
      for(int i=0; i<=this->order(); i++)
	{
	  itg->setCoef(i+1, this->getCoef(i) / (i+1));
	}
      return itg;
    }
}

// return the integration of the current polynome with the integration constant
Function* Polynomial::integration(double cste) const 
{
  if(this->order() == 0 && this->getCoef(0) == 0)
    return new Polynomial();
  else
    {
      Polynomial* itg = new Polynomial(this->order()+1);
      itg->setCoef(0, cste);
      for(int i=0; i<=this->order(); i++)
	{
	  itg->setCoef(i+1, this->getCoef(i) / (i+1));
	}
      return itg;
    }
}

// return the integration of the current polynome with the integration constant at time binf
Function* Polynomial::integration(double binf,double cste) const 
{
  if(this->order() == 0 && this->getCoef(0) == 0)
    {
      Polynomial* itg = new Polynomial(0);
      itg->setCoef(0, cste);
      return itg;
    }
  else
    {
      Polynomial* itg = new Polynomial(this->order()+1);
      for(int i=0; i<=this->order(); i++)
	{
	  itg->setCoef(i+1, this->getCoef(i) / (i+1));
	}
      
      double valBinf = itg->getVal(binf);
      itg->setCoef(0, cste - valBinf);

      //cout << "cste : " << cste << "   binf : " << binf << "   valBinf : " << valBinf << endl;

      return itg;
    }
}


/****************** Roots search ******************/

// return the real roots of this function
int Polynomial::realRoots(std::vector<double>& r) const
{
  if(this->order() == 1)
    {
      double r1;
      int res = rootPoly1st(this->getCoef(1), this->getCoef(0), r1);

      if(res == 1) r.push_back(r1);
      return res;
    }
  else if(this->order() == 2)
    {
      double r1,r2;
      int res = realRootPoly2nd(this->getCoef(2), this->getCoef(1), this->getCoef(0), r1, r2);

      if(res >= 1) r.push_back(r1);
      if(res >= 2) r.push_back(r2);

      // sort the vector of roots
      if(res!=0) sortVectDble(r);

      return res;
    }
  else if(this->order() == 3)
    {
      double r1,r2,r3;
      int res = realRootPoly3rd(this->getCoef(3), this->getCoef(2), this->getCoef(1), this->getCoef(0), r1, r2, r3);

      if(res >= 1) r.push_back(r1);
      if(res >= 2) r.push_back(r2);
      if(res >= 3) r.push_back(r3);

      // sort the vector of roots
      if(res!=0) sortVectDble(r);

      return res;
    }
  else return 0;
}

// return the roots of the equation eq - v = 0
int Polynomial::realRoots(std::vector<double>& r,double v) const
{
  Polynomial p2(this->order());
  p2.setCoef(0, this->getCoef(0) - v);
  for(int i=1; i<=this->order(); i++)
    {
      p2.setCoef(i, this->getCoef(i));
    }

  return p2.realRoots(r);
}


/**************** Utility functions ***************/

// display the polynomial
void Polynomial::display() const
{
  cout << "[" << binf_ << ";" << bsup_ << "] -> R : ";
  for(int i=this->order(); i>=0; i--)
    {
      cout << this->getCoef(i);
      if(i == 0) 
	cout << endl;
      else
	cout << "t^" << i << " + ";
    }
}

// shift the function on Y axis from value v
void Polynomial::shiftY(double v)
{
  this->setCoef(0, this->getCoef(0) + v);
}

// shift the reference time value (t0)
void Polynomial::shiftTime(double t)
{
//   cout << "Polynomial::shiftTime : Not implemented anymore. Waiting for Pascal's triangle implementation" << endl;
//   exit(1);

  int order = this->order();

  PascalTriangle::buildUpToOrder(order);
  double* tempValues = new double[order+1];

  for(int j=0; j<=order; j++)
    {
      tempValues[j] = 0;
//       cout << "order : " << j;
      for(int i=0; i<=order-j; i++)
	{
	  tempValues[j] += this->getCoef(order-i) * PascalTriangle::getVal(order-i, j) * pow(-t,order-i-j);
// 	  cout << " + " << "coef:" << this->getCoef(i) << " * Pasc(" << order-i << "," << j << ")=" << PascalTriangle::getVal(order-i, j) << " * (-t=" << -t << ")^" <<  order-i-j << "=" << pow(-t,order-i-j) << " = " << this->getCoef(i) * PascalTriangle::getVal(order-i, j) * pow(-t,order-i-j) << endl;
	}
//       cout << " = " << tempValues[j] << endl;
    }

  // replace the old coefficients
  for(int i=0; i<=order; i++) this->setCoef(i, tempValues[i]);

  this->binf_ += t;
  this->bsup_ += t;

  delete [] tempValues;
}

/**
 *  \brief search for the discontinuities of the function
 *  \param discPoints : resulting discontinuity points
 */
void Polynomial::getVecDiscPoints(std::vector<Disc_Point>& discPoints) const
{
  // no discontinuities -> nothing to do
}
