#include "math/inc/SinFunc.hpp"
using namespace mpc;

#include <iostream>
#include <stdlib.h>
#include <cmath>
using namespace std;

                                    /**************************************************/
                                    /****************** Class SinFunc *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SinFunc::SinFunc() : Operation(1)
{
  this->op_ = SIN_OP;

  this->setBounds(0.0,-1.0);

  this->type_ = SIN_FUNC;
}

/**
 *  \brief init constructor
 *  \param f : the operand
 *
 *  Input pointer is directly assigned to the structure
 */
SinFunc::SinFunc(Function* f) : Operation(1)
{
  this->op_ = SIN_OP;
  this->tabOperand_[0] = f;

  this->setBounds(f->getBInf(),f->getBSup());

  this->type_ = SIN_FUNC;
}

/**
 *  \brief copy constructor
 */
SinFunc::SinFunc(const SinFunc& op) : Operation(op)
{
}

/**
 *  \brief destructor
 */
SinFunc::~SinFunc()
{
}

/**
 *  \brief return a copy of the current function
 *  \return a copy of the current function
 *
 *  A new instance of function is allocated and returned here
 */
SinFunc* SinFunc::copy() const
{
  return new SinFunc(*this);
}


/**************** Setters & Getters ***************/

/**
 *  \brief return the value of the polynomial at a given parameter t
 *  \param t : the input parameter value
 *  \return the value of the polynomial at a given parameter t
 */
double SinFunc::getVal(double t) const
{
  if(t >= binf_ && t <= bsup_)
    {
      if(tabOperand_[0] != NULL)
	{
	  return sin(tabOperand_[0]->getVal(t));
	}
      else 
	{
	  cerr << "SinFunc::getVal : Operand function is not defined." << endl;
	  exit(1);
	}
   }
  else return 0;
}

  
/******************** Operators *******************/

/**
 *  \brief multiply the current function by a constant value
 *  \param v : the input multiplication coefficient
 */
void SinFunc::operator*=(double v)
{
  cerr << "SinFunc::operator*= : Not defined yet!" << endl;
  exit(1);
}


/************ Integration & Derivation ************/

/**
 *  \brief return the derivation of the current function
 *  \return the derivation of the current function
 */
Function* SinFunc::derivation() const
{
  cerr << "SinFunc::derivation : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \return the integration of the current function
 */
Function* SinFunc::integration() const
{
  cerr << "SinFunc::integration : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \param cste : the integration constant at t=0
 *  \return the integration of the current function
 */
Function* SinFunc::integration(double cste) const
{
  cerr << "SinFunc::integration : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \param cste,binf : cste is the integration constant value defined at t=binf
 *  \return the integration of the current function
 */
Function* SinFunc::integration(double binf,double cste) const
{
  cerr << "SinFunc::integration : Not defined yet!" << endl;
  exit(1);
}


/****************** Roots search ******************/

/**
 *  \brief return the real roots of this function
 *  \param r : the list of roots of the function
 *  \return the number of roots of the function
 */
int SinFunc::realRoots(std::vector<double>& r) const
{
  cerr << "SinFunc::realRoots : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the roots of the equation eq - v = 0
 *  \param r : the list of roots of the function
 *  \param v : the input constant value
 *  \return the number of roots of the function
 */
int SinFunc::realRoots(std::vector<double>& r,double v) const
{
  cerr << "SinFunc::realRoots : Not defined yet!" << endl;
  exit(1);
}


/**************** Utility functions ***************/

/**
 *  \brief display the polynomial
 */
void SinFunc::display() const
{
  cout << "[" << binf_ << ";" << bsup_ << "] -> Sin : " << endl;
  if(tabOperand_ == NULL) cout << "tabOperand nul" << endl;
  else
    {
      if(tabOperand_[0] == NULL) cout << "tabOperand[0] nul" << endl;
      else tabOperand_[0]->display();
    }
}

/**
 *  \brief shift the function on Y axis from value v
 *  \param v : the shifting value
 */
void SinFunc::shiftY(double v)
{
  cerr << "SinFunc::shiftY : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief shift the time parameter of the function
 *  \param t : the shifting value
 */
void SinFunc::shiftTime(double t)
{
  if(tabOperand_ != NULL && tabOperand_[0] != NULL) tabOperand_[0]->shiftTime(t);
}


/***************** Discontinuities ****************/

/**
 *  \brief search for the discontinuities of the function
 *  \param discPoints : resulting discontinuity points
 */
void SinFunc::getVecDiscPoints(std::vector<Disc_Point>& discPoints) const
{
  if(tabOperand_ != NULL && tabOperand_[0] != NULL)tabOperand_[0]->getVecDiscPoints(discPoints);
}
