#ifndef TCP_SERVER_H
#define TCP_SERVER_H

#include <boost/bind.hpp>
#include <boost/smart_ptr.hpp>
#include <boost/asio.hpp>
#include <boost/thread.hpp>
using boost::asio::ip::tcp;

#include <vector>

namespace mpc
{

  class TCP_server
  {
  protected : 
    int port_;
    double* message_;
    int length_;

    int waiting_time_ms_;

    typedef boost::shared_ptr<tcp::socket> socket_ptr;
    std::vector<socket_ptr> sock;

    boost::mutex mutexMessage_;

  protected : 
    /**
     *  \brief default constructor
     */
    TCP_server();

  public : 
    /**
     *  \brief init constructor
     */
    TCP_server(int length, int port=40000, int wtime=10);

    /**
     *  \brief destructor
     */
    ~TCP_server();

  private : 
    /*------------------------------------------------------------------------
     *  TCP server session
     */
    void session();

    /*------------------------------------------------------------------------
     *  TCP connection server
     */
    void server(boost::asio::io_service& io_service);

    /*------------------------------------------------------------------------
     *  Start the server thread
     */
    void startThreadServer();

  public : 
    /*------------------------------------------------------------------------
     *  Start the TCP Communication
     */
    void start();

    /*------------------------------------------------------------------------
     *  Stop the TCP Communication
     */
    void stop();

    /*------------------------------------------------------------------------
     *  Set the sent message
     */
    void setMessage(const double* msg, int length);
  };

};
#endif
