#ifndef POINTMASS3DSYS_HPP
#define POINTMASS3DSYS_HPP

#include "system/inc/System.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /************** Class PointMass3DSys **************/
                                    /**************************************************/

  /** \class PointMass3DSys
   * \brief 3D Point mass system class
   */
  class PointMass3DSys : public System
  {
  protected : 
    double radius_;                                   /**< radius of the cirle which include the robot */

    double vxMax_;                                    /**< maximal linear speed on X */
    double vyMax_;                                    /**< maximal linear speed on Y */
    double vzMax_;                                    /**< maximal linear speed on Z */
  
    double axMax_;                                    /**< maximal linear acceleration on X */
    double ayMax_;                                    /**< maximal linear acceleration on Y */
    double azMax_;                                    /**< maximal linear acceleration on Z */

  public : 

    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PointMass3DSys();
    /**
     *  \brief destructor
     */
    virtual ~PointMass3DSys();

    /************** Getters and setters ***************/

    /**
     *  \brief getter : return the radius of the circle including the robot
     *  \return radius of the circle including the robot
     */
    inline double getRadius() const { return this->radius_; }

    /**
     *  \brief setter : set the radius of the circle including the robot
     *  \param r : radius of the circle including the robot
     */
    inline void setRadius(const double& r) { this->radius_ = r; }
  
    /**
     *  \brief getter : return the maximal linear speed on X
     *  \return maximal linear speed on X
     */
    inline double getVXMax() const { return this->vxMax_; }

    /**
     *  \brief setter : set the maximal linear speed on X
     *  \param v : new maximal linear speed on X
     */
    inline void setVXMax(const double& v) { this->vxMax_ = v; }

    /**
     *  \brief getter : return the maximal linear speed on Y
     *  \return maximal linear speed on Y
     */
    inline double getVYMax() const { return this->vyMax_; }

    /**
     *  \brief setter : set the maximal linear speed on Y
     *  \param v : new maximal linear speed on Y
     */
    inline void setVYMax(const double& v) { this->vyMax_ = v; }

    /**
     *  \brief getter : return the maximal linear speed on Z
     *  \return maximal linear speed on Z
     */
    inline double getVZMax() const { return this->vzMax_; }

    /**
     *  \brief setter : set the maximal linear speed on Z
     *  \param v : new maximal linear speed on Z
     */
    inline void setVZMax(const double& v) { this->vzMax_ = v; }

    /**
     *  \brief getter : return the maximal linear acceleration on X
     *  \return maximal linear acceleration on X
     */
    inline double getAXMax() const { return this->axMax_; }

    /**
     *  \brief setter : set the maximal linear acceleration on X
     *  \param a : new maximal linear acceleration on X
     */
    inline void setAXMax(const double& a) { this->axMax_ = a; }

    /**
     *  \brief getter : return the maximal linear acceleration on Y
     *  \return maximal linear acceleration on Y
     */
    inline double getAYMax() const { return this->ayMax_; }

    /**
     *  \brief setter : set the maximal linear acceleration on Y
     *  \param a : new maximal linear acceleration on Y
     */
    inline void setAYMax(const double& a) { this->ayMax_ = a; }

    /**
     *  \brief getter : return the maximal linear acceleration on Z
     *  \return maximal linear acceleration on Z
     */
    inline double getAZMax() const { return this->azMax_; }

    /**
     *  \brief setter : set the maximal linear acceleration on Z
     *  \param a : new maximal linear acceleration on Z
     */
    inline void setAZMax(const double& a) { this->azMax_ = a; }

    /**
     *  \brief getter : maximal speed
     *  return the maximal speed wrt. its environment
     *  \return maximal speed
     */
    virtual double getVMax() const;
    /**
     *  \brief getter : maximal acceleration
     *  return the maximal acceleration wrt. its environment
     *  \return acceleration speed
     */
    virtual double getAMax() const;

    /*************** Display functions ****************/

    /**
     *  \brief display the features of the car-like system
     */
    virtual void display() const;

    /**
     *  \brief draw the car-like system in an openGL viewer
     */
    virtual void toGL() const;

    /**
     *  \brief convert the system into a string
     *  \return the resulting string
     */
    virtual const char* toString() const;
  };

};
#endif // POINTMASS3DSYS_HPP
