#ifndef SPACESHIPSYS_HPP
#define SPACESHIPSYS_HPP

#include "system/inc/System.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class SpaceShipSys ***************/
                                    /**************************************************/

  /** \class SpaceShipSys
   * \brief Space ship system class
   */
  class SpaceShipSys : public System
  {
  protected : 
    double radius_;                                   /**< radius of the cirle which include the robot */

    double vxMax_;                                    /**< maximal linear speed on X */
    double vyMax_;                                    /**< maximal linear speed on Y */
    double vthMax_;                                   /**< maximal angular speed */
  
    double axMax_;                                    /**< maximal linear acceleration on X */
    double ayMax_;                                    /**< maximal linear acceleration on Y */
    double athMax_;                                    /**< maximal angular acceleration */

  public : 

    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    SpaceShipSys();
    /**
     *  \brief destructor
     */
    virtual ~SpaceShipSys();

    /************** Getters and setters ***************/

    /**
     *  \brief getter : return the radius of the circle including the robot
     *  \return radius of the circle including the robot
     */
    inline double getRadius() const { return this->radius_; }

    /**
     *  \brief setter : set the radius of the circle including the robot
     *  \param r : radius of the circle including the robot
     */
    inline void setRadius(const double& r) { this->radius_ = r; }
  
    /**
     *  \brief getter : return the maximal linear speed on X
     *  \return maximal linear speed on X
     */
    inline double getVXMax() const { return this->vxMax_; }

    /**
     *  \brief setter : set the maximal linear speed on X
     *  \param v : new maximal linear speed on X
     */
    inline void setVXMax(const double& v) { this->vxMax_ = v; }

    /**
     *  \brief getter : return the maximal linear speed on Y
     *  \return maximal linear speed on Y
     */
    inline double getVYMax() const { return this->vyMax_; }

    /**
     *  \brief setter : set the maximal linear speed on Y
     *  \param v : new maximal linear speed on Y
     */
    inline void setVYMax(const double& v) { this->vyMax_ = v; }

    /**
     *  \brief getter : return the maximal angular speed
     *  \return maximal angular speed
     */
    inline double getVThMax() const { return this->vthMax_; }

    /**
     *  \brief setter : set the maximal angular speed
     *  \param v : new maximal angular speed
     */
    inline void setVThMax(const double& v) { this->vthMax_ = v; }

    /**
     *  \brief getter : return the maximal linear acceleration on X
     *  \return maximal linear acceleration on X
     */
    inline double getAXMax() const { return this->axMax_; }

    /**
     *  \brief setter : set the maximal linear acceleration on X
     *  \param a : new maximal linear acceleration on X
     */
    inline void setAXMax(const double& a) { this->axMax_ = a; }

    /**
     *  \brief getter : return the maximal linear acceleration on Y
     *  \return maximal linear acceleration on Y
     */
    inline double getAYMax() const { return this->ayMax_; }

    /**
     *  \brief setter : set the maximal linear acceleration on Y
     *  \param a : new maximal linear acceleration on Y
     */
    inline void setAYMax(const double& a) { this->ayMax_ = a; }

    /**
     *  \brief getter : return the maximal angular acceleration
     *  \return maximal angular acceleration
     */
    inline double getAThMax() const { return this->athMax_; }

    /**
     *  \brief setter : set the maximal angular acceleration
     *  \param a : new maximal angular acceleration
     */
    inline void setAThMax(const double& a) { this->athMax_ = a; }

    /**
     *  \brief getter : return the maximal linear velocity
     *  \return maximal linear velocity
     */
    double getVMax() const;

    /**
     *  \brief getter : return the maximal linear acceleration
     *  \return maximal linear acceleration
     */
    double getAMax() const;

    /*************** Display functions ****************/

    /**
     *  \brief display the features of the car-like system
     */
    virtual void display() const;

    /**
     *  \brief draw the car-like system in an openGL viewer
     */
    virtual void toGL() const;

    /**
     *  \brief convert the system into a string
     *  \return the resulting string
     */
    virtual const char* toString() const;
  };

};
#endif // SPACESHIPSYS_HPP
