#include "DiffDrive2Sys.hpp"
using namespace mpc;

#include <cmath>

                                    /**************************************************/
                                    /*************** Class DiffDrive2Sys **************/
                                    /**************************************************/
  
/********** Constructors and destructors **********/
    
/**
 *  \brief default constructor
 */
DiffDrive2Sys::DiffDrive2Sys()
{
  this->max_speed          = 1.5;
  this->max_acc            = 0.3;
  this->max_rotation_speed = 0.3*M_PI;
  this->max_rotation_acc   = 0.1*M_PI;
  this->length             = 1.0;
  this->width              = 1.0;
  this->canMoveBack_       = false;

  this->type_              = DIFFDRIVE2_SYS;
}

/**
 *  \brief copy constructor
 *  \param sys : the wheelchair system constraints to copy
 */
DiffDrive2Sys::DiffDrive2Sys(const DiffDrive2Sys& sys)
{
  this->max_speed          = sys.max_speed;
  this->max_acc            = sys.max_acc;
  this->max_rotation_speed = sys.max_rotation_speed;
  this->max_rotation_acc   = sys.max_rotation_acc;
  this->length             = sys.length;
  this->width              = sys.width;
  this->canMoveBack_       = sys.canMoveBack_;

  this->type_              = sys.type_;
}

/**
 *  \brief init constructor
 *  \param wr : radius of the wheels (m)
 *  \param wb : distance between the 2 wheels
 *  \param v : maximal linear speed (m/s)
 *  \param a : maximal linear acceleration (m/s^2)
 *  \param av : maximal rotation speed (rad/s)
 *  \param aa : maximal rotation acceleration (rad/s^2)
 *  \param bd : braking distance from max speed (m)
 *  \param l : lenght size (m)
 *  \param w : width size (m)
 *  \param cmb : true if the wheelchair can move backward
 */
DiffDrive2Sys::DiffDrive2Sys(double v,double a,double av,double aa,double l,double w,bool cmb)
{
  this->max_speed          = v;
  this->max_acc            = a;
  this->max_rotation_speed = av;
  this->max_rotation_acc   = aa;
  this->length             = l;
  this->width              = w;
  this->canMoveBack_       = cmb;

  this->type_              = DIFFDRIVE2_SYS;
}

/**
 *  \brief destructor
 */
DiffDrive2Sys::~DiffDrive2Sys()
{
}
