#include "DiffDriveSys.hpp"
using namespace mpc;

#include <cmath>
#include <stdio.h>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /*************** Class DiffDriveSys ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DiffDriveSys::DiffDriveSys() : System()
{
  this->length_       = 0;
  this->wheelRad_     = 0;
  this->radius_       = 0;
  this->vmax_         = 0;
  this->amax_         = 0;
  this->type_         = DIFFDRIVE_SYS;
}

/**
 *  \brief default constructor
 *  \param maxWVel : maximal wheel velocity
 *  \param maxWAcc : maximal wheel acceleration
 *  \param R : radius of the differential drive
 */
DiffDriveSys::DiffDriveSys(double maxWVel,double maxWAcc,double R)
{
  this->length_       = R;
  this->wheelRad_     = 0;
  this->radius_       = R;
  this->vmax_         = maxWVel;
  this->amax_         = maxWAcc;
  this->type_         = DIFFDRIVE_SYS;
}

/**
 *  \brief destructor
 */
DiffDriveSys::~DiffDriveSys()
{
}


/************** Getters and setters ***************/

/**
 *  \brief getter : return the length of the robot (between the center and the wheels)
 *  \return the length of the robot (between the center and the wheels)
 */
double DiffDriveSys::getLength() const
{
  return this->length_;
}

/**
 *  \brief setter : set the length of the robot (between the center and the wheels)
 *  \param l : the new length of the robot (between the center and the wheels)
 */
void DiffDriveSys::setLength(const double& l)
{
  this->length_ = l;
}

/**
 *  \brief getter : return the radius of the wheels
 *  \return the radius of the wheels
 */
double DiffDriveSys::getWheelRad() const
{
  return this->wheelRad_;
}

/**
 *  \brief setter : set the radius of the wheels
 *  \param l : the new radius of the wheels
 */
void DiffDriveSys::setWheelRad(const double& r)
{
  this->wheelRad_ = r;
}

/**
 *  \brief getter : return the radius of the circle including the robot
 *  \return radius of the circle including the robot
 */
double DiffDriveSys::getRadius() const
{
  return this->radius_;
}

/**
 *  \brief setter : set the radius of the circle including the robot
 *  \param r : radius of the circle including the robot
 */
void DiffDriveSys::setRadius(const double& r)
{
  this->radius_ = r;
}

/**
 *  \brief getter : return the maximal speed
 *  \return maximal speed
 */
double DiffDriveSys::getWVMax() const
{
  return this->vmax_;
}

/**
 *  \brief setter : set the maximal speed
 *  \param v : new maximal speed
 */
void DiffDriveSys::setWVMax(const double& v)
{
  this->vmax_ = v;
}

/**
 *  \brief getter : return the maximal acceleration
 *  \return maximal acceleration
 */
double DiffDriveSys::getWAMax() const
{
  return this->amax_;
}

/**
 *  \brief setter : set the maximal acceleration
 *  \param v : new maximal acceleration
 */
void DiffDriveSys::setWAMax(const double& a)
{
  this->amax_ = a;
}


/*************** Display functions ****************/

/**
 *  \brief display the features of the differential drive system
 */
void DiffDriveSys::display() const
{
  cout << "DiffDriveSys : " << endl;
  cout << "\t Length (R) : " << length_ << endl;
  cout << "\t Wheel radius : " << wheelRad_ << endl;
  cout << "\t Speed Max : " << vmax_ << endl;
  cout << "\t Acc Max : " << amax_ << endl;
  cout << endl;
}

/**
 *  \brief draw the car-like system in an openGL viewer
 */
void DiffDriveSys::toGL() const
{
}
