#include "WheelChairSys.hpp"

#include "Utils.hpp"
using namespace mpc;

#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

                                    /**************************************************/
                                    /*************** Class WheelChairSys **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
WheelChairSys::WheelChairSys() : DiffDrive2Sys()
{
  this->wheel_radius       = 0.7;
  this->wheel_base         = 1.0;
//   this->max_speed          = 1.5;
//   this->max_acc            = 0.3;
//   this->max_rotation_speed = 0.3*M_PI;
//   this->max_rotation_acc   = 0.1*M_PI;
  this->braking_dist       = 1.0;
//   this->length             = 1.0;
//   this->width              = 1.0;
//   this->canMoveBack_       = false;

  this->type_              = WHEELCHAIR_SYS;
}

/**
 *  \brief copy constructor
 *  \param sys : the wheelchair system constraints to copy
 */
WheelChairSys::WheelChairSys(const WheelChairSys& sys) : DiffDrive2Sys(sys)
{
  this->wheel_radius       = sys.wheel_radius;
  this->wheel_base         = sys.wheel_base;
//   this->max_speed          = sys.max_speed;
//   this->max_acc            = sys.max_acc;
//   this->max_rotation_speed = sys.max_rotation_speed;
//   this->max_rotation_acc   = sys.max_rotation_acc;
  this->braking_dist       = sys.braking_dist;
//   this->length             = sys.length;
//   this->width              = sys.width;
//   this->canMoveBack_       = sys.canMoveBack_;

  this->type_              = WHEELCHAIR_SYS;
}

/**
 *  \brief init constructor
 *  \param wr : radius of the wheels (m)
 *  \param wb : distance between the 2 wheels
 *  \param v : maximal linear speed (m/s)
 *  \param a : maximal linear acceleration (m/s^2)
 *  \param av : maximal rotation speed (rad/s)
 *  \param aa : maximal rotation acceleration (rad/s^2)
 *  \param bd : braking distance from max speed (m)
 *  \param l : lenght size (m)
 *  \param w : width size (m)
 */
WheelChairSys::WheelChairSys(double wr,double wb,double v,double a,double av,double aa,double bd,double l,double w,bool cmb) : DiffDrive2Sys(v,a,av,aa,l,w,cmb)
{
  this->wheel_radius       = wr;
  this->wheel_base         = wb;
//   this->max_speed          = v;
//   this->max_acc            = a;
//   this->max_rotation_speed = av;
//   this->max_rotation_acc   = aa;
  this->braking_dist       = bd;
//   this->length             = l;
//   this->width              = w;
//   this->canMoveBack_       = cmb;

  this->type_              = WHEELCHAIR_SYS;
}

// /**
//  *  \brief default constructor
//  */
// WheelChairSys::WheelChairSys()
// {
//   this->wheel_radius       = 0.7;
//   this->wheel_base         = 1.0;
//   this->max_speed          = 1.5;
//   this->max_acc            = 0.3;
//   this->max_rotation_speed = 0.3*M_PI;
//   this->max_rotation_acc   = 0.1*M_PI;
//   this->braking_dist       = 1.0;
//   this->length             = 1.0;
//   this->width              = 1.0;
//   this->canMoveBack_       = false;

//   this->type_              = WHEELCHAIR_SYS;
// }

// /**
//  *  \brief copy constructor
//  *  \param sys : the wheelchair system constraints to copy
//  */
// WheelChairSys::WheelChairSys(const WheelChairSys& sys)
// {
//   this->wheel_radius       = sys.wheel_radius;
//   this->wheel_base         = sys.wheel_base;
//   this->max_speed          = sys.max_speed;
//   this->max_acc            = sys.max_acc;
//   this->max_rotation_speed = sys.max_rotation_speed;
//   this->max_rotation_acc   = sys.max_rotation_acc;
//   this->braking_dist       = sys.braking_dist;
//   this->length             = sys.length;
//   this->width              = sys.width;
//   this->canMoveBack_       = sys.canMoveBack_;

//   this->type_              = WHEELCHAIR_SYS;
// }

// /**
//  *  \brief init constructor
//  *  \param wr : radius of the wheels (m)
//  *  \param wb : distance between the 2 wheels
//  *  \param v : maximal linear speed (m/s)
//  *  \param a : maximal linear acceleration (m/s^2)
//  *  \param av : maximal rotation speed (rad/s)
//  *  \param aa : maximal rotation acceleration (rad/s^2)
//  *  \param bd : braking distance from max speed (m)
//  *  \param l : lenght size (m)
//  *  \param w : width size (m)
//  */
// WheelChairSys::WheelChairSys(double wr,double wb,double v,double a,double av,double aa,double bd,double l,double w,bool cmb)
// {
//   this->wheel_radius       = wr;
//   this->wheel_base         = wb;
//   this->max_speed          = v;
//   this->max_acc            = a;
//   this->max_rotation_speed = av;
//   this->max_rotation_acc   = aa;
//   this->braking_dist       = bd;
//   this->length             = l;
//   this->width              = w;
//   this->canMoveBack_       = cmb;

//   this->type_              = WHEELCHAIR_SYS;
// }

/**
 *  \brief destructor
 */
WheelChairSys::~WheelChairSys()
{
}


/*************** Display functions ****************/

/**
 *  \brief display the features of the differential drive system
 */
void WheelChairSys::display() const
{
  cout << "WheelChairSys : " << endl;
  cout << "\t wheel_radius : " << wheel_radius << endl;
  cout << "\t wheel_base  : " << wheel_base << endl;
  cout << "\t max_speed : " << max_speed << endl;
  cout << "\t max_acc : " << max_acc << endl;
  cout << "\t max_rotation_speed : " << max_rotation_speed << endl;
  cout << "\t max_rotation_acc : " << max_rotation_acc << endl;
  cout << "\t braking_dist : " << braking_dist << endl;
  cout << "\t length : " << length << endl;
  cout << "\t width : " << width << endl;  
  cout << endl;
}

/**
 *  \brief convert the system into a string
 *  \return the resulting string
 */
const char* WheelChairSys::toString() const
{
  char c1[50]; double2String(wheel_radius,c1,2);
  char c2[50]; double2String(wheel_base,c2,2);
  char c3[50]; double2String(max_speed,c3,2);
  char c4[50]; double2String(max_acc,c4,2);
  char c5[50]; double2String(max_rotation_speed,c5,2);
  char c6[50]; double2String(max_rotation_acc,c6,2);
  char c7[50]; double2String(braking_dist,c7,2);
  char c8[50]; double2String(length,c8,2);
  char c9[50]; double2String(width,c9,2);

  std::string str("WheelChairSys :");
  str += "\n";
  str += "wheel_radius : ";
  str += c1;
  str += "\n";
  str += "wheel_base  : ";
  str += c2;
  str += "\n";
  str += "max_speed : ";
  str += c3;
  str += "\n";
  str += "max_acc : ";
  str += c4;
  str += "\n";
  str += "max_rotation_speed : ";
  str += c5;
  str += "\n";
  str += "max_rotation_acc : ";
  str += c6;
  str += "\n";
  str += "braking_dist : ";
  str += c7;
  str += "\n";
  str += "length : ";
  str += c8;
  str += "\n";
  str += "width : ";
  str += c9;
  str += "\n";

  //cout << "ARGH : "<< endl;
  //cout << str.c_str() << endl;
  //cout << ":-(" << endl;

  return str.c_str();
}

/**
 *  \brief draw the differential drive system in an openGL viewer
 */
void WheelChairSys::toGL() const
{
  // differential drive robot drawing
  double heading = 0.0;
  double robotRadius = 0.5*length;

  const GLfloat blueMaterial[]   = {0.,0.,1.,1.};
  const GLfloat paleblueMaterial[] = {0.1,0.1,0.7,1.};
  const GLfloat redMaterial[]    = {1.,0.,0.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
  
  // draw the robot itself
  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
  gluCylinder(gluNewQuadric(), robotRadius, robotRadius, robotRadius/100, 20, 20);
  glTranslated(0,0,robotRadius/100);
  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
  
  // draw the wheels of the robot
  glMaterialfv(GL_FRONT, GL_AMBIENT, paleblueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, paleblueMaterial);
  
  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading + 0.5*M_PI),
	       1.0*robotRadius*sin(heading + 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 0.2*robotRadius, 20, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading - 0.5*M_PI),
	       1.0*robotRadius*sin(heading - 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 0.2*robotRadius, 20, 1);
  glPopMatrix();
  
  // draw the orientation of the robot
  glLineWidth(((GLfloat)(3.0)));
  glBegin(GL_LINES);
  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
  glVertex3f(0.0,0.0,0.0);
  glVertex3f(2.2*robotRadius*cos(heading), 
	     2.2*robotRadius*sin(heading), 
	     0.0);
  glEnd();
  glLineWidth(((GLfloat)(1.0)));
}
