#include "CLTrajGen.hpp"
#include "WCTrajGen.hpp"
#include "SSTrajGen.hpp"
using namespace mpc;

int main()
{
  // car-like initialization
  CarLikeSys carlike;
  carlike.setLength(1.0);                     // in meters
  carlike.setRadius(1.0);                     // in meters (=length)
  carlike.setVMax(30.0);                      // in m/s
  carlike.setAMax(5.0);                       // in m/s^2
  carlike.setMaxSteering(M_PI*0.25); // in rad/s
  carlike.setMaxSteerSpd(M_PI*0.05); // in rad/s^2


  // car-like trajectory generation initialization
  CLTrajGen trajGenCL(&carlike);
  
  /// car-like trajgen options

  // car-like constraints
  trajGenCL.setUseVelCst(true);               // default = true
  trajGenCL.setUseAccCst(true);               // default = true
  trajGenCL.setUseAVelCst(true);              // default = true
  trajGenCL.setUseAAccCst(false);             // default = false

  // final state constraints
  trajGenCL.fixFinalXY(true);                 // default = true
  trajGenCL.fixFinalTH(true);                 // default = true
  trajGenCL.fixFinalV(true);                  // default = true
  trajGenCL.fixFinalPH(true);                 // default = true


  /// common trajgen options

  // final time constraint
  timeCstt_t tcStatus = TIME_CSTD;            // time constraint status (TIME_FIXED,TIME_FREE,TIME_CSTD), default = TIME_CSTD
  double minTime = 15.0;                       // default = 0.0
  double maxTime = 25.0;                       // default = 999999999.9
  trajGenCL.setTcStatus(tcStatus,minTime,maxTime);

  // init options of the trajectory generation                                    
  trajGenCL.setMaxNbStep(20);                 // default = 20
  trajGenCL.setMaxResidual(0.05);             // default = 0.05
  
  // init the debug level mode
  trajGenCL.setDBG(1);                        // default = 0

  // init the IG database use
  trajGenCL.setIGdb(NULL);                    // default = NULL

  // init the finite difference partial derivative
  trajGenCL.setFDpartDeriv(false);            // default = false


  // initial and final state-times initialization
  double x0  = 0;                             // init position
  double y0  = 0;                             // init position
  double th0 = 0;                             // init heading
  double ph0 = 0;                             // init steering angle
  double v0  = 0;                             // init speed
  double t0  = 0;                             // init time

  double xf  = -10;                            // final position
  double yf  = -10;                            // final position
  double thf = -3.14;                             // final heading
  double phf = 0;                             // final steering angle
  double vf  = 0;                             // final speed
  double tf  = 3;                            // final time

  CLStateT s0(x0,y0,th0,ph0,v0,t0);
  CLStateT sf(xf,yf,thf,phf,vf,tf);

  
  // generation of a trajectory between s0 and sf
  // return 0 if the process succeed, -1 else
  int res = trajGenCL.ComputeTrajectory(s0,sf);

  // whatever the result, you can ask the following request
  
  // get back the final state-time of the trajectory
  CLStateT finalST = trajGenCL.getFinalCLStateTReached();
  cout << "finalST : "; finalST.display(); cout << endl;

  // get back the state-time of the trajectory defined at time tReq
  double tReq = 1.0;
  CLStateT ST = trajGenCL.getCLStateTReachedAt(tReq);
  cout << "ST : "; ST.display(); cout << endl;

  // get back the final residual
  double fresidual = trajGenCL.finalResidual();
  cout << "fresidual : " << fresidual << endl;

  // display the different profiles of the trajectory
  trajGenCL.displayProfiles();

  // sample the resulting state trajectory
  double timeStep = 0.1;
  CLSTraj* sampledTraj = trajGenCL.sampledTraj(timeStep);
  cout << "sampledTraj : " << endl;
  sampledTraj->display();
  cout << endl;

  return 0;
}
