#include "CLInitialGuess.hpp"
using namespace mpc;

#include "Timer.hpp"

#include <iostream>
#include <string>
#include <fstream>

#define SPG_T   sampling_[0]
#define SPG_VI  sampling_[1]
#define SPG_VG  sampling_[2]
#define SPG_DX  sampling_[3]
#define SPG_DY  sampling_[4]
#define SPG_DTH sampling_[5]
// #define SPG_S   samplingS_[6]

#define MIN_SPG_T   minSampS_[0]
#define MIN_SPG_VI  minSampS_[1]
#define MIN_SPG_VG  minSampS_[2]
#define MIN_SPG_DX  minSampS_[3]
#define MIN_SPG_DY  minSampS_[4]
#define MIN_SPG_DTH minSampS_[5]
// #define MIN_SPG_S   minSampS_[6]

#define MAX_SPG_T   maxSampS_[0]
#define MAX_SPG_VI  maxSampS_[1]
#define MAX_SPG_VG  maxSampS_[2]
#define MAX_SPG_DX  maxSampS_[3]
#define MAX_SPG_DY  maxSampS_[4]
#define MAX_SPG_DTH maxSampS_[5]
// #define MAX_SPG_S   maxSampS_[6]

#define USE_ACC_CST  true
#define USE_VEL_CST  true
#define USE_AACC_CST false
#define USE_AVEL_CST true

#define FIXED_XY true
#define FIXED_TH true
#define FIXED_V  false
#define FIXED_PH false
#define FIXED_T  true

                                    /**************************************************/
                                    /******************* Class CL_N2R *****************/
                                    /**************************************************/

double CL_N2R::wX_ = 1.0;                            /**< weight used to compute the distance between two nodes (on X) */
double CL_N2R::wY_ = 1.0;                            /**< weight used to compute the distance between two nodes (on Y) */
double CL_N2R::wTH_ = 10.0;                          /**< weight used to compute the distance between two nodes (on TH) */
double CL_N2R::wVI_ = 10.0;                          /**< weight used to compute the distance between two nodes (on VI) */
double CL_N2R::wVG_ = 10.0;                          /**< weight used to compute the distance between two nodes (on VG) */
double CL_N2R::wT_ = 1000.0;                         /**< weight used to compute the distance between two nodes (on T) */

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CL_N2R::CL_N2R() : N2R()
{
}

/**
 *  \brief init constructor (of an ig node)
 *  \param initST : the initial state of the trajectory
 *  \param goalST : the final state to reached
 *  \param param : the parametric representation of the control
 */
CL_N2R::CL_N2R(const CLStateT& initST,const CLStateT& goalST,double* param) : N2R()
{
  dimP_ = 7;
  param_ = new double[dimP_];
  for(int i=0; i<dimP_; i++) param_[i] = param[i];

  initST_ = new CLStateT(initST);
  goalST_ = new CLStateT(goalST);

  pos_ = NULL;
  
//   lnear_ = NULL;

  status_ = IG_N2R;
}

/**
 *  \brief init constructor (of an explored node)
 *  \param initST : the initial state of the trajectory
 *  \param goalST : the final state to reached
 *  \param i,j,k,l,m,n : position of the node in the grid
 *  \param node : the node from which the current one may be derivated
 */
CL_N2R::CL_N2R(const CLStateT& initST,const CLStateT& goalST,
	       int i,int j,int k,int l,int m,int n,N2R* node) : N2R()
{
  dimP_ = 7;
  param_ = NULL;

  initST_ = new CLStateT(initST);
  goalST_ = new CLStateT(goalST);

  pos_ = new int[6];
  pos_[0] = i;
  pos_[1] = j;
  pos_[2] = k;
  pos_[3] = l;
  pos_[4] = m;
  pos_[5] = n;
  
//   lnear_ = new std::list<N2R*>;
//   this->addNearNode(node);

  status_ = EXPLORED_N2R;
}

/**
 *  \brief init constructor (of a reached node)
 *  \param initST : the initial state of the trajectory
 *  \param goalST : the final state to reached
 *  \param i,j,k,l,m,n : position of the node in the grid
 *  \param param : the parametric representation of the control
 */
CL_N2R::CL_N2R(const CLStateT& initST,const CLStateT& goalST,
       int i,int j,int k,int l,int m,int n,double* param) : N2R()
{
  dimP_ = 7;
  param_ = new double[dimP_];
  for(int p=0; p<dimP_; p++) param_[p] = param[p];

  initST_ = new CLStateT(initST);
  goalST_ = new CLStateT(goalST);

  pos_ = new int[6];
  pos_[0] = i;
  pos_[1] = j;
  pos_[2] = k;
  pos_[3] = l;
  pos_[4] = m;
  pos_[5] = n;
  
//   lnear_ = NULL;

  status_ = REACHED_N2R;
}

// /**
//  *  \brief init constructor
//  *  \param i,j,k,l,m,n : position of the node in the grid
//  *  \param node : the node from which the current one may be derivated
//  */
// CL_N2R::CL_N2R(int i,int j,int k,int l,int m,int n,N2R* node) : N2R(6)
// {
//   pos_[0] = i;
//   pos_[1] = j;
//   pos_[2] = k;
//   pos_[3] = l;
//   pos_[4] = m;
//   pos_[5] = n;

//   this->addNearNode(node);
// }

/**
 *  \brief destructor
 */
CL_N2R::~CL_N2R()
{
//   this->clearNearNodes();
}


/*************** Getters and Setters **************/

// /**
//  *  \brief return the nearest neighbour of the current node
//  *  \return the nearest neighbour of the current node if one exists, NULL else
//  */
// CL_N2R* CL_N2R::getNearNeighbour() const
// {
//   cerr << "CL_N2R::getNearNeighbour() : Not implemented yet!" << endl;
//   exit(1);
// }


/**************** Utility functions ***************/

/**
 *  \brief return a weighted distance between the current node and the input node
 *  \param n2 : the second node
 *  \return a weighted distance between both nodes
 */
double CL_N2R::computeWDistTo(CL_N2R* n2) const
{
  const CLStateT& ist = this->initST();
  const CLStateT& gst = this->goalST();
  const CLStateT& ist2 = n2->initST();
  const CLStateT& gst2 = n2->goalST();

  double dx = gst2.x() - gst.x();
  double dy = gst2.y() - gst.y();
  double dth = computeMinDiffAngle(gst2.th(),gst.th());
  double dvi = ist2.s() - ist.s();
  double dvg = gst2.s() - gst.s();
  double dt = gst2.t() - gst.t();

  return ((CL_N2R::wX_*fabs(dx)) + 
	  (CL_N2R::wY_*fabs(dy)) + 
	  (CL_N2R::wTH_*fabs(dth)) + 
	  (CL_N2R::wVI_*fabs(dvi)) + 
	  (CL_N2R::wVG_*fabs(dvg)) + 
	  (CL_N2R::wT_*fabs(dt)));
}


                                    /**************************************************/
                                    /***************** Class CL_N2RGrid ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CL_N2RGrid::CL_N2RGrid() : N2RGrid()
{
}

/**
 *  \brief init constructor
 */
CL_N2RGrid::CL_N2RGrid(int* sampling,double* minSampS,double* maxSampS)
  : N2RGrid(6,sampling,minSampS,maxSampS)
{
}


/********* Access/Modification of the grid ********/

/**
 *  \brief return an element of the grid at pos = (i,j,k,l,m,n)
 *  \param i : index of the time dimension
 *  \param j : index of the initial velocity dimension
 *  \param k : index of the goal velocity dimension
 *  \param l : index of the x position dimension
 *  \param m : index of the y position dimension
 *  \param n : index of the heading dimension
 *  \return the element of the grid at pos = (i,j,k,l,m,n), NULL if the input coordinates are outside of the grid
 */
N2R* CL_N2RGrid::gridAt(int i,int j,int k,int l,int m,int n) const
{
  int pos[6] = {i,j,k,l,m,n};
  return this->N2RGrid::gridAt(pos);
}

/**
 *  \brief return an element of the grid at pos = (i,j,k,l,m,n)
 *  \param i : index of the time dimension
 *  \param j : index of the initial velocity dimension
 *  \param k : index of the goal velocity dimension
 *  \param l : index of the x position dimension
 *  \param m : index of the y position dimension
 *  \param n : index of the heading dimension
 *  \return the element of the grid at pos = (i,j,k,l,m,n), NULL if the input coordinates are outside of the grid
 */
N2R*& CL_N2RGrid::gridAt(int i,int j,int k,int l,int m,int n)
{
  int pos[6] = {i,j,k,l,m,n};
  return this->N2RGrid::gridAt(pos);
}

/**
 *  \brief compute the initial state-time of a cell of the grid
 *  \param i : index of the time dimension
 *  \param j : index of the initial velocity dimension
 *  \param k : index of the goal velocity dimension
 *  \param l : index of the x position dimension
 *  \param m : index of the y position dimension
 *  \param n : index of the heading dimension
 *  \param initST : resulting initial state-time
 */
void CL_N2RGrid::computeInitST(int i,int j,int k,int l,int m,int n,
		   CLStateT& initST) const
{
  initST.x() = 0;
  initST.y() = 0;
  initST.th() = 0;
  initST.ph() = 0;
  initST.s() = MIN_SPG_VI + j*(MAX_SPG_VI-MIN_SPG_VI)/((double)(SPG_VI-1));
  initST.t() = 0;
}

/**
 *  \brief compute the goal state-times of a cell of the grid
 *  \param i : index of the time dimension
 *  \param j : index of the initial velocity dimension
 *  \param k : index of the goal velocity dimension
 *  \param l : index of the x position dimension
 *  \param m : index of the y position dimension
 *  \param n : index of the heading dimension
 *  \param goalST : resulting goal state-time
 */
void CL_N2RGrid::computeGoalST(int i,int j,int k,int l,int m,int n,
		   CLStateT& goalST) const
{
  goalST.x() = MIN_SPG_DX + l*(MAX_SPG_DX-MIN_SPG_DX)/((double)(SPG_DX-1));
  goalST.y() = MIN_SPG_DY + m*(MAX_SPG_DY-MIN_SPG_DY)/((double)(SPG_DY-1));
  goalST.th() =MIN_SPG_DTH + n*(MAX_SPG_DTH-MIN_SPG_DTH)/((double)(SPG_DTH-1));
  goalST.ph() = 0;
  goalST.s() = MIN_SPG_VG + k*(MAX_SPG_VG-MIN_SPG_VG)/((double)(SPG_VG-1));
  goalST.t() = MIN_SPG_T + i*(MAX_SPG_T-MIN_SPG_T)/((double)(SPG_T-1));
}

/**
 *  \brief search for reachable nodes on the grid around a reached node (and mark them as reachable on the grid)
 *  \param node : the input already reached node
 *  \return a list containing the reachable nodes on the grid around the input node
 *
 *  An new instance of list is allocated and returned here
 */
std::list<N2R*>* CL_N2RGrid::reachableNodesAround(N2R* node)
{
  std::list<N2R*>* lres = new std::list<N2R*>;

  int indMinT, indMinVI, indMinVG, indMinDX, indMinDY, indMinDTH;
  int indMaxT, indMaxVI, indMaxVG, indMaxDX, indMaxDY, indMaxDTH;

  int* pos = node->pos();
  
  if(pos != NULL)
    {
      if(pos[0] == 0) indMinT = 0; 
      else indMinT = pos[0]-1;
      if(pos[1] == 0) indMinVI = 0; 
      else indMinVI = pos[1]-1;  
      if(pos[2] == 0) indMinVG = 0; 
      else indMinVG = pos[2]-1;  
      if(pos[3] == 0) indMinDX = 0; 
      else indMinDX = pos[3]-1;  
      if(pos[4] == 0) indMinDY = 0; 
      else indMinDY = pos[4]-1;  
      if(pos[5] == 0) indMinDTH = 0; 
      else indMinDTH = pos[5]-1;  
      
      if(pos[0] == SPG_T-1) indMaxT = SPG_T-1;
      else indMaxT = pos[0]+1;
      if(pos[1] == SPG_VI-1) indMaxVI = SPG_VI-1;
      else indMaxVI = pos[1]+1;  
      if(pos[2] == SPG_VG-1) indMaxVG = SPG_VG-1;
      else indMaxVG = pos[2]+1;  
      if(pos[3] == SPG_DX-1) indMaxDX = SPG_DX-1;
      else indMaxDX = pos[3]+1;  
      if(pos[4] == SPG_DY-1) indMaxDY = SPG_DY-1; 
      else indMaxDY = pos[4]+1;  
      if(pos[5] == SPG_DTH-1) indMaxDTH = SPG_DTH-1; 
      else indMaxDTH = pos[5]+1;
    }
  else
    {
      // search for the nearest node of the grid from the 
      // position of the final state of the current one

      // get back the final state-time of the node
      CLStateT initST = ((CL_N2R*)node)->initST();
      CLStateT goalST = ((CL_N2R*)node)->goalST();
      
      searchForItvSampling(goalST.t(),MIN_SPG_T,MAX_SPG_T,SPG_T,indMinT,indMaxT);
      searchForItvSampling(initST.s(),MIN_SPG_VI,MAX_SPG_VI,SPG_VI,indMinVI,indMaxVI);
      searchForItvSampling(goalST.s(),MIN_SPG_VG,MAX_SPG_VG,SPG_VG,indMinVG,indMaxVG);
      searchForItvSampling(goalST.x(),MIN_SPG_DX,MAX_SPG_DX,SPG_DX,indMinDX,indMaxDX);
      searchForItvSampling(goalST.y(),MIN_SPG_DY,MAX_SPG_DY,SPG_DY,indMinDY,indMaxDY);
      searchForItvSampling(goalST.th(),MIN_SPG_DTH,MAX_SPG_DTH,SPG_DTH,indMinDTH,indMaxDTH);

//       cout << "SPG_T:[" << indMinT << ";" << indMaxT << "]  ";
//       cout << "SPG_VI:[" << indMinVI << ";" << indMaxVI << "]  ";
//       cout << "SPG_VG:[" << indMinVG << ";" << indMaxVG << "]  ";
//       cout << "SPG_DX:[" << indMinDX << ";" << indMaxDX << "]  ";
//       cout << "goalST : "; goalST.display(); cout << "  MIN_SPG_DY : " << MIN_SPG_DY << "  MAX_SPG_DY : " << MAX_SPG_DY << "  SPG_DY : " << SPG_DY << "  ";
// 	cout << "SPG_DY:[" << indMinDY << ";" << indMaxDY << "]  "; cout << endl;
//       cout << "SPG_DTH:[" << indMinDTH << ";" << indMaxDTH << "]  " << endl;

    }

  for(int i=indMinT; i<=indMaxT; i++)
    for(int j=indMinVI; j<=indMaxVI; j++)
      for(int k=indMinVG; k<=indMaxVG; k++)
	for(int l=indMinDX; l<=indMaxDX; l++)
	  for(int m=indMinDY; m<=indMaxDY; m++)
	    for(int n=indMinDTH; n<=indMaxDTH; n++)
	      {
		// get back the current neighbour node
		N2R* n2r = this->gridAt(i,j,k,l,m,n);
		
		// check if the node at position (i,j,k,l,m,n) has already been created
		if(n2r == NULL)
		  {
		    // create a new node at position (i,j,k,l,m,n)
		    int pos[6] = {i,j,k,l,m,n};
		    n2r = this->addN2Rfrom(pos,node);
		  }

		// add it to the resulting list of neighbours
		lres->push_back(n2r);
	      }
  
  return lres;
}

/**
 *  \brief search for already reached nodes on the grid around a node to reach
 *  \param node : the input node to reach
 *  \return a list containing the already reached nodes on the grid around the input node
 *
 *  An new instance of list is allocated and returned here
 */
std::list<N2R*>* CL_N2RGrid::reachedNodesAround(N2R* node) const
{
  std::list<N2R*>* lres = new std::list<N2R*>;

  int indMinT, indMinVI, indMinVG, indMinDX, indMinDY, indMinDTH;
  int indMaxT, indMaxVI, indMaxVG, indMaxDX, indMaxDY, indMaxDTH;

  int* pos = node->pos();
  
  if(pos != NULL)   
    {
      if(pos[0] == 0) indMinT = 0; 
      else indMinT = pos[0]-1;
      if(pos[1] == 0) indMinVI = 0; 
      else indMinVI = pos[1]-1;  
      if(pos[2] == 0) indMinVG = 0; 
      else indMinVG = pos[2]-1;  
      if(pos[3] == 0) indMinDX = 0; 
      else indMinDX = pos[3]-1;  
      if(pos[4] == 0) indMinDY = 0; 
      else indMinDY = pos[4]-1;  
      if(pos[5] == 0) indMinDTH = 0; 
      else indMinDTH = pos[5]-1;  

      if(pos[0] == SPG_T-1) indMaxT = SPG_T-1;
      else indMaxT = pos[0]+1;
      if(pos[1] == SPG_VI-1) indMaxVI = SPG_VI-1;
      else indMaxVI = pos[1]+1;  
      if(pos[2] == SPG_VG-1) indMaxVG = SPG_VG-1;
      else indMaxVG = pos[2]+1;  
      if(pos[3] == SPG_DX-1) indMaxDX = SPG_DX-1;
      else indMaxDX = pos[3]+1;  
      if(pos[4] == SPG_DY-1) indMaxDY = SPG_DY-1; 
      else indMaxDY = pos[4]+1;  
      if(pos[5] == SPG_DTH-1) indMaxDTH = SPG_DTH-1; 
      else indMaxDTH = pos[5]+1;
    }
  else
    {
      cerr << "CL_N2RGrid::reachedNodesAround : Should not arrive here!" << endl;
      exit(1);
    }

  for(int i=indMinT; i<=indMaxT; i++)
    for(int j=indMinVI; j<=indMaxVI; j++)
      for(int k=indMinVG; k<=indMaxVG; k++)
	for(int l=indMinDX; l<=indMaxDX; l++)
	  for(int m=indMinDY; m<=indMaxDY; m++)
	    for(int n=indMinDTH; n<=indMaxDTH; n++)
	      {
		// get back the current neighbour node
		N2R* n2r = this->gridAt(i,j,k,l,m,n);
		
		// check if the node at position (i,j,k,l,m,n) has already been created
		if(n2r != NULL && n2r->status() == REACHED_N2R) 
		  {
		    // add it to the resulting list of neighbours
		    lres->push_back(n2r);
		  }
	      }
  
  return lres;
}

// /**
//  *  \brief create a N2R from its coordinates and a a node from which it can be derivated
//  *  \param coord : coordinates of the N2R
//  *  \param node : node from which the N2R may be reachable
//  *
//  *  An new instance of N2R is allocated here
//  */
// N2R* CL_N2RGrid::createN2R(int* coord,N2R* node)
// {
//   N2R* n2r = new CL_N2R(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],node);
//   return n2r;
// }

/**
 *  \brief create a N2R from its coordinates and a a node from which it can be derivated
 *  \param coord : coordinates of the N2R
 *  \param node : node from which the N2R may be reachable
 *
 *  An new instance of N2R is allocated here
 *  Put directly the new node in the grid
 */
N2R* CL_N2RGrid::createExploredN2R(int* coord,N2R* node)
{
  // compute the init and goal state-times of the node
  CLStateT initST;
  CLStateT goalST;
  computeInitST(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],initST);
  computeGoalST(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],goalST);
  
  // create the new node
  CL_N2R* n2r = new CL_N2R(initST,goalST,coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],node);

  // put the new node in the grid
  this->gridAt(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5]) = n2r;

  return n2r;
}

/**
 *  \brief create a N2R from its coordinates and a a node from which it can be derivated
 *  \param coord : coordinates of the N2R
 *  \param node : node from which the N2R may be reachable
 *
 *  An new instance of N2R is allocated here
 *  Both state-times and the control parameterization are copied here
 */
N2R* CL_N2RGrid::createIGN2R(const StateTime* initST,const StateTime* goalST,double* param)
{
  CL_N2R* n2r = new CL_N2R(*((CLStateT*)initST),*((CLStateT*)goalST),param);
  return n2r;
}

/**
 *  \brief create a reached node from an input file thanks to its position and parameters
 *  \param pos : the position of the node in the grid
 *  \param params : the parameterized control defining the trajectory
 *  \return a pointer to the created node
 */
N2R* CL_N2RGrid::createReachedNode(int* coord,double* params)
{
  // compute the initial and goal state-times from their positions in the grid
  CLStateT initST;
  CLStateT goalST;
  this->computeInitST(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],initST);
  this->computeGoalST(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],goalST);

  // create the node
  N2R* node = new CL_N2R(initST,goalST,coord[0],coord[1],coord[2],coord[3],coord[4],coord[5],params);

  // put the new node in the grid
  this->gridAt(coord[0],coord[1],coord[2],coord[3],coord[4],coord[5]) = node;

  // add the node to the list of reached nodes
  this->listNReached_->push_back(node);

  return node;
}

/**
 *  \brief compute the number of nodes noted as "nodes to reach" on the grid
 *  \return the number of nodes noted as "nodes to reach" on the grid
 */
int CL_N2RGrid::N2RonGrid() const
{
  int res = 0;

  for(int i=0; i<SPG_T; i++)
    for(int j=0; j<SPG_VI; j++)
      for(int k=0; k<SPG_VG; k++)
	for(int l=0; l<SPG_DX; l++)
	  for(int m=0; m<SPG_DY; m++)
	    for(int n=0; n<SPG_DTH; n++)
	      {
		N2R* node = this->gridAt(i,j,k,l,m,n);
		//if(node != NULL && node->status() == EXPLORED_N2R) res++;
		if(node != NULL) res++;
	      }

  return res;
}

/**
 *  \brief compute the number of nodes noted as reached on the grid
 *  \return the number of nodes noted as reached on the grid
 */
int CL_N2RGrid::NReachedonGrid() const
{
  int res = 0;

  for(int i=0; i<SPG_T; i++)
    for(int j=0; j<SPG_VI; j++)
      for(int k=0; k<SPG_VG; k++)
	for(int l=0; l<SPG_DX; l++)
	  for(int m=0; m<SPG_DY; m++)
	    for(int n=0; n<SPG_DTH; n++)
	      {
		N2R* node = this->gridAt(i,j,k,l,m,n);
		if(node != NULL && node->status() == REACHED_N2R) res++;
	      }

  return res;
}

/**
 *  \brief compute the number of nodes noted as used on the grid
 *  \return the number of nodes noted as used on the grid
 */
int CL_N2RGrid::NUsedonGrid() const
{
  int res = 0;

  for(int i=0; i<SPG_T; i++)
    for(int j=0; j<SPG_VI; j++)
      for(int k=0; k<SPG_VG; k++)
	for(int l=0; l<SPG_DX; l++)
	  for(int m=0; m<SPG_DY; m++)
	    for(int n=0; n<SPG_DTH; n++)
	      {
		N2R* node = this->gridAt(i,j,k,l,m,n);
		if(node != NULL && node->status() == USED_N2R) res++;
	      }

  return res;
}

/**
 *  \brief display the nodes set as reached on the grid
 */
void CL_N2RGrid::displayReachedOnGrid() const
{
  cout << "Reached nodes on grid : " << endl;
  for(int i=0; i<SPG_T; i++)
    for(int j=0; j<SPG_VI; j++)
      for(int k=0; k<SPG_VG; k++)
	for(int l=0; l<SPG_DX; l++)
	  for(int m=0; m<SPG_DY; m++)
	    for(int n=0; n<SPG_DTH; n++)
	      {
		N2R* node = this->gridAt(i,j,k,l,m,n);
		if(node != NULL && node->status() == REACHED_N2R)
		  {
		    cout << "(" << i << "," << j << "," << k << "," << l << "," << m << "," << n << ") -> ";
		    displayTab(node->params(),7); cout << endl;
		  }
	      }
}

/**
 *  brief display the sampling of the grid
 */
void CL_N2RGrid::displayGridSampling() const
{
  cout << "T : " << SPG_T << "  [";
  for(int i=0; i<SPG_T-1; i++) cout << (MIN_SPG_T + (double)i * (MAX_SPG_T - MIN_SPG_T) / (SPG_T-1)) << ";";
  cout << MAX_SPG_T << "]" << endl;

  cout << "VI : " << SPG_VI << "  [";
  for(int i=0; i<SPG_VI-1; i++) cout << (MIN_SPG_VI + (double)i * (MAX_SPG_VI - MIN_SPG_VI) / (SPG_VI-1)) << ";";
  cout << MAX_SPG_VI << "]" << endl;

  cout << "VG : " << SPG_VG << "  [";
  for(int i=0; i<SPG_VG-1; i++) cout << (MIN_SPG_VG + (double)i * (MAX_SPG_VG - MIN_SPG_VG) / (SPG_VG-1)) << ";";
  cout << MAX_SPG_VG << "]" << endl;

  cout << "DX : " << SPG_DX << "  [";
  for(int i=0; i<SPG_DX-1; i++) cout << (MIN_SPG_DX + (double)i * (MAX_SPG_DX - MIN_SPG_DX) / (SPG_DX-1)) << ";";
  cout << MAX_SPG_DX << "]" << endl;

  cout << "DY : " << SPG_DY << "  [";
  for(int i=0; i<SPG_DY-1; i++) cout << (MIN_SPG_DY + (double)i * (MAX_SPG_DY - MIN_SPG_DY) / (SPG_DY-1)) << ";";
  cout << MAX_SPG_DY << "]" << endl;

  cout << "DTH : " << SPG_DTH << "  [";
  for(int i=0; i<SPG_DTH-1; i++) cout << (MIN_SPG_DTH + (double)i * (MAX_SPG_DTH - MIN_SPG_DTH) / (SPG_DTH-1)) << ";";
  cout << MAX_SPG_DTH << "]" << endl;
}

#define SPG_T   sampling_[0]
#define SPG_VI  sampling_[1]
#define SPG_VG  sampling_[2]
#define SPG_DX  sampling_[3]
#define SPG_DY  sampling_[4]
#define SPG_DTH sampling_[5]

#define MIN_SPG_T   minSampS_[0]
#define MIN_SPG_VI  minSampS_[1]
#define MIN_SPG_VG  minSampS_[2]
#define MIN_SPG_DX  minSampS_[3]
#define MIN_SPG_DY  minSampS_[4]
#define MIN_SPG_DTH minSampS_[5]

#define MAX_SPG_T   maxSampS_[0]
#define MAX_SPG_VI  maxSampS_[1]
#define MAX_SPG_VG  maxSampS_[2]
#define MAX_SPG_DX  maxSampS_[3]
#define MAX_SPG_DY  maxSampS_[4]
#define MAX_SPG_DTH maxSampS_[5]


/**
 *  \brief search for the nearest defined node of the grid to compute a trajectory between both input state-times
 *  \param is : the input init state-time
 *  \param gs : the input goal state-time
 *  \return the nearest defined node of the grid to compute a trajectory between both input state-times
 */
N2R* CL_N2RGrid::nearestNodeToComputeTrajBetween(const CLStateT& is,const CLStateT& gs) const
{
  N2R* selectedNode = NULL;

//   cout << "1" << endl;
  double dx = gs.x() - is.x();
  double dy = gs.y() - is.y();
  double dth = computeMinDiffAngle(gs.th(),is.th());
  double vi = is.s();
  double vg = gs.s();
  double t = gs.t() - is.t();

  cout << "dx: " << dx << "  dy:" << dy << "  dth:" << dth << "  vi:" << vi << "  vg:" << vg << "  t:" << t << endl;

  int indMinT, indMinVI, indMinVG, indMinDX, indMinDY, indMinDTH;
  int indMaxT, indMaxVI, indMaxVG, indMaxDX, indMaxDY, indMaxDTH;

  int* indMin = new int[n_];
  int* indMax = new int[n_];

//   cout << "2" << endl;
  searchForItvSampling(t,MIN_SPG_T,MAX_SPG_T,SPG_T,indMin[0],indMax[0]);
  searchForItvSampling(vi,MIN_SPG_VI,MAX_SPG_VI,SPG_VI,indMin[1],indMax[1]);
  searchForItvSampling(vg,MIN_SPG_VG,MAX_SPG_VG,SPG_VG,indMin[2],indMax[2]);
  searchForItvSampling(dx,MIN_SPG_DX,MAX_SPG_DX,SPG_DX,indMin[3],indMax[3]);
  searchForItvSampling(dy,MIN_SPG_DY,MAX_SPG_DY,SPG_DY,indMin[4],indMax[4]);
  searchForItvSampling(dth,MIN_SPG_DTH,MAX_SPG_DTH,SPG_DTH,indMin[5],indMax[5]);
//   cout << "3" << endl;

  cout << "indT:[" << indMin[0] << ";" << indMax[0] << "]  ";
  cout << "indVI:[" << indMin[1] << ";" << indMax[1] << "]  ";
  cout << "indVG:[" << indMin[2] << ";" << indMax[2] << "]  ";
  cout << "indDX:[" << indMin[3] << ";" << indMax[3] << "]  ";
  cout << "indDY:[" << indMin[4] << ";" << indMax[4] << "]  ";
  cout << "indDTH:[" << indMin[5] << ";" << indMax[5] << "]  ";
  cout << endl;

  bool trouve = false;
  int maxDist = 0;
  while(!trouve && (maxDist < SPG_T && maxDist < SPG_VI && maxDist < SPG_VG && maxDist < SPG_DX && maxDist < SPG_DY && maxDist < SPG_DTH))
    {
      cout << "cur maxDist : " << maxDist << endl;
      int cpt = 0;

      // compute a list of all reached nodes at a distance maxDist from 
      // the nodes surronding the position of the node to reach in the grid
      std::list<N2R*> lDefinedNodes;
      for(int i=0; i<n_; i++)
	{
	  // compute the lower index, upper index and step of each dimension
	  int* minInd = new int[n_];
	  int* maxInd = new int[n_];
	  int* stepInd = new int[n_];

	  for(int j=0; j<n_; j++)
	    {
	      if(j == i)
		{
		  minInd[j] = indMin[j] - maxDist;
		  maxInd[j] = indMax[j] + maxDist;
		  stepInd[j] = (indMax[j] - indMin[j]) + 2*maxDist;
		}
	      else if(j < i)
		{
		  minInd[j] = indMin[j] - maxDist + 1;
		  maxInd[j] = indMax[j] + maxDist - 1;
		  stepInd[j] = 1;
		}
	      else // (j > i)
		{
		  minInd[j] = indMin[j] - maxDist;
		  maxInd[j] = indMax[j] + maxDist;
		  stepInd[j] = 1;
		}
	      if(stepInd[j] == 0) stepInd[j] = 1;
	      if(minInd[j] < 0) minInd[j] = 0;
	      if(maxInd[j] >= sampling_[j]) maxInd[j] = sampling_[j]-1;
	    }

	  cout << "indT2:[" << minInd[0] << ";" << maxInd[0] << "]  ";
	  cout << "indVI2:[" << minInd[1] << ";" << maxInd[1] << "]  ";
	  cout << "indVG2:[" << minInd[2] << ";" << maxInd[2] << "]  ";
	  cout << "indDX2:[" << minInd[3] << ";" << maxInd[3] << "]  ";
	  cout << "indDY2:[" << minInd[4] << ";" << maxInd[4] << "]  ";
	  cout << "indDTH2:[" << minInd[5] << ";" << maxInd[5] << "]  ";
	  cout << endl;

// 	  cout << "minInd : "; displayTab(minInd,6);
// 	  cout << "  maxInd : "; displayTab(maxInd,6);
// 	  cout << "  stepInd : "; displayTab(stepInd,6);
// 	  cout << endl;

	  for(int j=minInd[0]; j<maxInd[0]; j+=stepInd[0])
	    {
	      for(int k=minInd[1]; k<maxInd[1]; k+=stepInd[1])
		{
		  for(int l=minInd[2]; l<maxInd[2]; l+=stepInd[2])
		    {
		      for(int m=minInd[3]; m<maxInd[3]; m+=stepInd[3])
			{
			  for(int n=minInd[4]; n<maxInd[4]; n+=stepInd[4])
			    {
			      for(int p=minInd[5]; p<maxInd[5]; p+=stepInd[5])
				{
				  // get back the node defined at the current position if it exists
				  N2R* node = this->gridAt(j,k,l,m,n,p);
				  
				  // add it to the list of candidates nodes if it has been defined
				  if(node != NULL && node->status() == REACHED_N2R) 
				    {
				      lDefinedNodes.push_back(node);
				      cout << "candidate node: ";
				      cout << "(" << j << "," << k << "," << l << "," << m << "," << n << "," << p << ") ";
				      displayTab(node->pos(),6); 
				      cout << " -> initST : ";
				      node->initST().display();
				      cout << " -> goalST : ";
				      node->goalST().display();
				      cout << endl;
				    }
				}
			    }
			}
		    }
		}
	    }

	  delete [] minInd;
	  delete [] maxInd;
	  delete [] stepInd;
	}

      // check if at least one defined node has been found
      if(lDefinedNodes.size() != 0)
	{
	  cout << "maxDist : " << maxDist << endl;

	  // select a random node from the list of candidates
	  int indCand = getIntRandom(0,lDefinedNodes.size()-1);

	  // move up to the selected node
	  std::list<N2R*>::iterator it = lDefinedNodes.begin();
	  for(int i=0; i<indCand; i++) it++;

	  selectedNode = *it;
	  trouve = true;
	}
      else
	{
	  maxDist++;
	}
    }

//   cout << "4" << endl;
  delete [] indMin;
  delete [] indMax;

  if(selectedNode != NULL) 
    { 
      cout << "selected node : "; 
      displayTab(selectedNode->pos(),6); 
      cout << " -> initST : ";
      selectedNode->initST().display();
      cout << " -> goalST : ";
      selectedNode->goalST().display();
      cout << endl;
    }
  else cout << "No node found in the grid..." << endl;

//   cout << "selected node params : "; displayTab(selectedNode->params(),7); cout << endl;

  return selectedNode;
}

/**
 *  \brief search for the nearest defined node of the grid to compute a trajectory between both input state-times (2nd version)
 *  \param is : the input init state-time
 *  \param gs : the input goal state-time
 *  \return the nearest defined node of the grid to compute a trajectory between both input state-times
 */
N2R* CL_N2RGrid::nearestNodeToComputeTrajBetweenV2(const CLStateT& is,const CLStateT& gs) const
{
  N2R* selectedNode = NULL;

  double dxWRot = gs.x() - is.x();
  double dyWRot = gs.y() - is.y();
  PositionXY posXY(dxWRot,dyWRot);
  posXY.rotate(PositionXY(0,0),-is.th());
  double dx = posXY.x();
  double dy = posXY.y();
  double ddist = sqrt(dx*dx + dy*dy);

  double dth = computeMinDiffAngle(gs.th(),is.th());
  double vi = is.s();
  double vg = gs.s();
  double t = gs.t() - is.t();

  cout << "dx: " << dx << "  dy:" << dy << "  dth:" << dth << "  vi:" << vi << "  vg:" << vg << "  t:" << t << endl;

  int indMinT, indMinVI, indMinVG, indMinDX, indMinDY, indMinDTH;
  int indMaxT, indMaxVI, indMaxVG, indMaxDX, indMaxDY, indMaxDTH;

  int* indMin = new int[n_];
  int* indMax = new int[n_];

  searchForItvSampling(t,MIN_SPG_T,MAX_SPG_T,SPG_T,indMin[0],indMax[0]);
  searchForItvSampling(vi,MIN_SPG_VI,MAX_SPG_VI,SPG_VI,indMin[1],indMax[1]);
  searchForItvSampling(vg,MIN_SPG_VG,MAX_SPG_VG,SPG_VG,indMin[2],indMax[2]);
  searchForItvSampling(dx,MIN_SPG_DX,MAX_SPG_DX,SPG_DX,indMin[3],indMax[3]);
  searchForItvSampling(dy,MIN_SPG_DY,MAX_SPG_DY,SPG_DY,indMin[4],indMax[4]);
  searchForItvSampling(dth,MIN_SPG_DTH,MAX_SPG_DTH,SPG_DTH,indMin[5],indMax[5]);

  cout << "indT:[" << indMin[0] << ";" << indMax[0] << "]  ";
  cout << "indVI:[" << indMin[1] << ";" << indMax[1] << "]  ";
  cout << "indVG:[" << indMin[2] << ";" << indMax[2] << "]  ";
  cout << "indDX:[" << indMin[3] << ";" << indMax[3] << "]  ";
  cout << "indDY:[" << indMin[4] << ";" << indMax[4] << "]  ";
  cout << "indDTH:[" << indMin[5] << ";" << indMax[5] << "]  ";
  cout << endl;

  bool trouve = false;
  int maxDist = 0;
  while(!trouve && (maxDist < SPG_T && maxDist < SPG_VI && maxDist < SPG_VG && maxDist < SPG_DX && maxDist < SPG_DY && maxDist < SPG_DTH))
    {
      cout << "cur maxDist : " << maxDist << endl;

      // compute the lower index, upper index and step of each dimension
      int* minInd = new int[n_];
      int* maxInd = new int[n_];

      for(int i=0; i<n_; i++)
	{
	  if(indMin[i] - maxDist >= 0) minInd[i] = indMin[i] - maxDist;
	  else minInd[i] = 0;
	  if(indMax[i] + maxDist < sampling_[i]) maxInd[i] = indMax[i] + maxDist;
	  else maxInd[i] = sampling_[i]-1;
	}

      cout << "indT2:[" << minInd[0] << ";" << maxInd[0] << "]  ";
      cout << "indVI2:[" << minInd[1] << ";" << maxInd[1] << "]  ";
      cout << "indVG2:[" << minInd[2] << ";" << maxInd[2] << "]  ";
      cout << "indDX2:[" << minInd[3] << ";" << maxInd[3] << "]  ";
      cout << "indDY2:[" << minInd[4] << ";" << maxInd[4] << "]  ";
      cout << "indDTH2:[" << minInd[5] << ";" << maxInd[5] << "]  ";
      cout << endl;
      
      // compute a list of all reached nodes at a distance maxDist from 
      // the nodes surronding the position of the node to reach in the grid
      std::list<N2R*> lDefinedNodes;
      N2R* bestNode = NULL;
      double bestDist = -1;

      for(int j=minInd[0]; j<maxInd[0]; j++)
	{
	  for(int k=minInd[1]; k<maxInd[1]; k++)
	    {
	      for(int l=minInd[2]; l<maxInd[2]; l++)
		{
		  for(int m=minInd[3]; m<maxInd[3]; m++)
		    {
		      for(int n=minInd[4]; n<maxInd[4]; n++)
			{
			  for(int p=minInd[5]; p<maxInd[5]; p++)
			    {
			      // get back the node defined at the current position if it exists
			      CL_N2R* node = (CL_N2R*)(this->gridAt(j,k,l,m,n,p));
			      
			      // add it to the list of candidates nodes if it has been defined
			      if(node != NULL && node->status() == REACHED_N2R) 
				{
				  cout << "candidate node: ";
				  cout << "(" << j << "," << k << "," << l << "," << m << "," << n << "," << p << ") ";
				  displayTab(node->pos(),6); 
				  cout << " -> initST : ";
				  node->initST().display();
				  cout << " -> goalST : ";
				  node->goalST().display();
				  cout << endl;

				  double dxN = node->goalST().x() - node->initST().x() - dx;
				  double dyN = node->goalST().y() - node->initST().y() - dy;
				  double ddistN = sqrt(dxN*dxN + dyN*dyN);
				  double dthN = computeMinDiffAngle(gs.th(),is.th()) - dth;
				  double dviN = node->initST().s() - vi;
				  double dvgN = node->goalST().s() - vg;
				  double dtN = node->goalST().t() - t;
				  
				  double dist = ddistN + fabs(dthN) + fabs(dviN*t) + fabs(dvgN*t) + fabs(0.5*dtN*(vi+vg));

				  cout << "dxN:" << dxN << "  dyN:" << dyN << "  ddistN:" << ddistN << "  dthN:" << dthN 
				       << "  dviN:" << dviN << "  dvgN:" << dvgN << "  dtN:" << dtN;
				  cout << "  dist : " << dist << endl;

				  if(bestNode == NULL || dist < bestDist)
				    {
				      bestNode = node;
				      bestDist = dist;
				    }
				}
			    }
			}
		    }
		}
	    }
	}      
      
      delete minInd;
      delete maxInd;

      // check if at least one defined node has been found
      if(bestNode != NULL)
	{
	  cout << "maxDist : " << maxDist << endl;
	  cout << "bestDist:" << bestDist << "\t";

	  selectedNode = bestNode;
	  trouve = true;
	}
      else
	{
	  maxDist++;
	}
    }

  delete [] indMin;
  delete [] indMax;

  if(selectedNode != NULL) 
    { 
      cout << "selected node : "; 
      displayTab(selectedNode->pos(),6); 
      cout << " -> initST : ";
      selectedNode->initST().display();
      cout << " -> goalST : ";
      selectedNode->goalST().display();
      cout << endl;
    }
  else cout << "No node found in the grid..." << endl;

  return selectedNode;
}


                                    /**************************************************/
                                    /*************** Class CLIitialGuess *************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */  
CLInitialGuess::CLInitialGuess()
{
  firstIGMode_ = SCP_FIRSTIG; // (0=random, 1=sampled control, 2=sampled param control, 3=rstar)

  this->dimS_ = 6;
  this->dimP_ = 7;

  this->dimSampS_ = 6;

  this->sampling_ = new int [dimSampS_];
  this->minSampS_ = new double [dimSampS_];
  this->maxSampS_ = new double [dimSampS_];
  for(int i=0; i<dimSampS_; i++) sampling_[i] = 0;
  for(int i=0; i<dimSampS_; i++) minSampS_[i] = 0;
  for(int i=0; i<dimSampS_; i++) maxSampS_[i] = 0;

//   this->spg_ = NULL;
  this->grid_ = NULL;

  this->SC_AccSp_ = 8;
  this->SC_StSpdSp_ = 8;
}

/**
 *  \brief init constructor
 */  
CLInitialGuess::CLInitialGuess(CarLikeSys* sys) : InitialGuess(sys)
{
  firstIGMode_ = SCP_FIRSTIG; // (0=random, 1=sampled control, 2=sampled param control, 3=rstar)

  this->dimS_ = 6;
  this->dimP_ = 7;

  this->dimSampS_ = 6;

  this->sampling_ = new int [dimSampS_];
  this->minSampS_ = new double [dimSampS_];
  this->maxSampS_ = new double [dimSampS_];
  this->setDfltSampS();
  this->setDfltMinMaxSampS();

  this->trajGen_ = NULL; // ... Should be changed once the trajgen class modified ...  = new CLTrajGen(sys)

//   this->spg_ = NULL;
  this->initGrid();

  this->SC_AccSp_ = 8;
  this->SC_StSpdSp_ = 8;
}

/**
 *  \brief init constructor
 *  \param sys : the wheelchair features
 *  \param tmin : the minimal time
 *  \param tmax : the maximal time
 *  \param spg_t : the time sampling
 *  \param spg_vi : the initial speed sampling
 *  \param spg_vg : the goal speed sampling
 *  \param spg_dx : the position sampling
 *  \param spg_dy : the position sampling
 *  \param spg_dth : the heading sampling
 */
CLInitialGuess::CLInitialGuess(CarLikeSys* sys,
			       int tmin,int tmax,
			       int spg_t,int spg_vi,int spg_vg,int spg_dx,int spg_dy,int spg_dth) : InitialGuess(sys)
{
  firstIGMode_ = SCP_FIRSTIG; // (0=random, 1=sampled control, 2=sampled param control, 3=rstar)

  this->dimS_ = 6;
  this->dimP_ = 7;

  this->dimSampS_ = 6;

  this->sampling_ = new int [dimSampS_];
  this->minSampS_ = new double [dimSampS_];
  this->maxSampS_ = new double [dimSampS_];
  this->initSampS(spg_t,spg_vi,spg_vg,spg_dx,spg_dy,spg_dth);
  this->setDfltMinMaxSampS();
  MIN_SPG_T = tmin;
  MAX_SPG_T = tmax;

  this->trajGen_ = NULL; // ... Should be changed once the trajgen class modified ...  = new CLTrajGen(sys)

//   this->spg_ = NULL;
  this->initGrid();

  this->SC_AccSp_ = 8;
  this->SC_StSpdSp_ = 8;
}

/**
 *  \brief destructor
 */
CLInitialGuess::~CLInitialGuess()
{
  this->clearGrid();
}


/*************** Getters and Setters **************/

/**
 *  \brief use the sampled control mode to determine the first initial guess 
 *  \param sampAcc : number of samples of the acceleration control
 *  \param sampStSpd : number of samples of the steering speed control
 */
void CLInitialGuess::use_SC_FIGMode(double sampAcc,double sampStSpd)
{
  this->firstIGMode_ = SC_FIRSTIG;

  this->SC_AccSp_ = sampAcc;
  this->SC_StSpdSp_ = sampStSpd;
}

/**
 *  \brief use the sampled parametric control mode to determine the first initial guess 
 *  \param sampAcc : number of samples of the acceleration control
 *  \param sampStSpd : number of samples of the steering speed control
 */
void CLInitialGuess::use_SCP_FIGMode(double sampAcc,double sampStSpd)
{
  this->firstIGMode_ = SCP_FIRSTIG;

  this->SC_AccSp_ = sampAcc;
  this->SC_StSpdSp_ = sampStSpd;
}


/***************** Initialization *****************/

/**
 *  \brief initialize the sampling of the state space
 *  \param spg_t : sampling over the time
 *  \param spg_vi : sampling over the initial speed
 *  \param spg_vg : sampling over the goal speed
 *  \param spg_dx : sampling over the position x
 *  \param spg_dy : sampling over the position y
 *  \param spg_dth : sampling over the heading
 */
void CLInitialGuess::initSampS(int spg_t,int spg_vi,int spg_vg,int spg_dx,int spg_dy,int spg_dth)
{
  SPG_T = spg_t;
  SPG_VI = spg_vi;
  SPG_VG = spg_vg;
  SPG_DX = spg_dy;
  SPG_DY = spg_dx;
  SPG_DTH = spg_dth;
}

/**
 *  \brief set the default state sampling parameters
 */
void CLInitialGuess::setDfltSampS()
{
  SPG_T = 9;
  SPG_VI = 9;
  SPG_VG = 9;
  SPG_DX = 12;
  SPG_DY = 12;
  SPG_DTH = 12;

//   SPG_T = 9;
//   SPG_VI = 9;
//   SPG_VG = 9;
//   SPG_DX = 24;
//   SPG_DY = 24;
//   SPG_DTH = 12;
}

/**
 *  \brief set the default state sampling minimal values
 */
void CLInitialGuess::setDfltMinMaxSampS()
{
  MIN_SPG_T = 0.0;
  if(sys()->canMoveBack())
    {
      MIN_SPG_VI = -sys()->getVMax();
      MIN_SPG_VG = -sys()->getVMax();
    }
  else
    {
      MIN_SPG_VI = 0.0;
      MIN_SPG_VG = 0.0;
    }

  MAX_SPG_T = 5.0;
  //  MAX_SPG_T = 1.0;
  MAX_SPG_VI = sys()->getVMax();
  MAX_SPG_VG = sys()->getVMax();

  MIN_SPG_DX = -MAX_SPG_T * sys()->getVMax();
  MIN_SPG_DY = -MAX_SPG_T * sys()->getVMax();
  MIN_SPG_DTH = -M_PI;

  MAX_SPG_DX = MAX_SPG_T * sys()->getVMax();
  MAX_SPG_DY = MAX_SPG_T * sys()->getVMax();
  MAX_SPG_DTH = M_PI;
}

/**
 *  \brief initialize the sampling grid
 */
void CLInitialGuess::initGrid()
{
  // delete the old grid if needed
  this->clearGrid();

  // create the new grid with the actual sampling parameters
  this->grid_ = new CL_N2RGrid(sampling_,minSampS_,maxSampS_);
}

// /**
//  *  \brief initialize the sampling matrix
//  */
// void CLInitialGuess::initSampMatrix()
// {
//   this->clearSampMatrix();

//   spg_ = new CLNodeIG* ***** [SPG_T];
//   for(int i=0; i<SPG_T; i++)
//     {
//       spg_[i] = new CLNodeIG* **** [SPG_VI];
//       for(int j=0; j<SPG_VI; j++)
// 	{
// 	  spg_[i][j] = new CLNodeIG* *** [SPG_VG];
// 	  for(int k=0; k<SPG_VG; k++)
// 	    {
// 	      spg_[i][j][k] = new CLNodeIG* ** [SPG_DX];
// 	      for(int l=0; l<SPG_DX; l++)
// 		{
// 		  spg_[i][j][k][l] = new CLNodeIG* * [SPG_DY];
// 		  for(int m=0; m<SPG_DY; m++)
// 		    {
// 		      spg_[i][j][k][l][m] = new CLNodeIG* [SPG_DTH];
// 		      for(int n=0; n<SPG_DTH; n++)
// 			{
// 			  CLStateT initST;
// 			  CLStateT goalST;
// 			  computeInitST(i,j,k,l,m,n,initST);
// 			  computeGoalST(i,j,k,l,m,n,goalST);
			    
//  			  spg_[i][j][k][l][m][n] = new CLNodeIG(initST,goalST);
// 			}
// 		    }
// 		}
// 	    }
// 	}
//     }
// }

// /**
//  *  \brief clear the sampling matrix
//  */
// void CLInitialGuess::clearSampMatrix()
// {
//   if(spg_ != NULL)
//     {
//       for(int i=0; i<SPG_T; i++)
// 	{
// 	  for(int j=0; j<SPG_VI; j++)
// 	    {
// 	      for(int k=0; k<SPG_VG; k++)
// 		{
// 		  for(int l=0; l<SPG_DX; l++)
// 		    {
// 		      for(int m=0; m<SPG_DY; m++)
// 			{
// 			  for(int n=0; n<SPG_DTH; n++)
// 			    {
// 			      delete spg_[i][j][k][l][m][n];
// 			    }
// 			  delete spg_[i][j][k][l][m];
// 			}
// 		      delete spg_[i][j][k][l];
// 		    }		  
// 		  delete spg_[i][j][k];
// 		}
// 	      delete spg_[i][j];
// 	    }	  
// 	  delete spg_[i];
// 	}
//       delete spg_;
//     }
// }


/*************** Matrix State-times ***************/

// /**
//  *  \brief acces to a sample of the matrix
//  *  \param i : index of the time dimension
//  *  \param j : index of the initial velocity dimension
//  *  \param k : index of the goal velocity dimension
//  *  \param l : index of the x position dimension
//  *  \param m : index of the y position dimension
//  *  \param n : index of the heading dimension
//  *  \return a sample of the matrix
//  */
// CLNodeIG* CLInitialGuess::getSample(int i,int j,int k,int l,int m,int n)
// {
//   return spg_[i][j][k][l][m][n];
// }

// /**
//  *  \brief compute the initial state-time of a cell of the matrix
//  *  \param i : index of the time dimension
//  *  \param j : index of the initial velocity dimension
//  *  \param k : index of the goal velocity dimension
//  *  \param l : index of the x position dimension
//  *  \param m : index of the y position dimension
//  *  \param n : index of the heading dimension
//  *  \param initST : resulting initial state-time
//  */
// void CLInitialGuess::computeInitST(int i,int j,int k,int l,int m,int n,
// 				   CLStateT& initST) const
// {
//   initST.x() = 0;
//   initST.y() = 0;
//   initST.th() = 0;
//   initST.ph() = 0;
//   initST.s() = MIN_SPG_VI + j*(MAX_SPG_VI-MIN_SPG_VI)/((double)(SPG_VI-1));
//   initST.t() = 0;
// }


// /**
//  *  \brief compute the extremal available goal state-times of a cell of the matrix
//  *  \param i : index of the time dimension
//  *  \param j : index of the initial velocity dimension
//  *  \param k : index of the goal velocity dimension
//  *  \param l : index of the x position dimension
//  *  \param m : index of the y position dimension
//  *  \param n : index of the heading dimension
//  *  \param goalST : resulting goal state-time
//  */
// void CLInitialGuess::computeGoalST(int i,int j,int k,int l,int m,int n,
// 				   CLStateT& goalST) const
// {
//   goalST.x() = MIN_SPG_DX + l*(MAX_SPG_DX-MIN_SPG_DX)/((double)(SPG_DX-1));
//   goalST.y() = MIN_SPG_DY + m*(MAX_SPG_DY-MIN_SPG_DY)/((double)(SPG_DY-1));
//   goalST.th() =MIN_SPG_DTH + n*(MAX_SPG_DTH-MIN_SPG_DTH)/((double)(SPG_DTH-1));
//   goalST.ph() = 0;
//   goalST.s() = MIN_SPG_VG + k*(MAX_SPG_VG-MIN_SPG_VG)/((double)(SPG_VG-1));
//   goalST.t() = MIN_SPG_T + i*(MAX_SPG_T-MIN_SPG_T)/((double)(SPG_T-1));
// }


/******************** Learning ********************/

/**
 *  \brief compute a random initial guess for a sample of the state space
 *  \param ig : the resulting initial guess (must be allocated)
 *  \param t : time the control generated from the set of parameters will last
 */
void CLInitialGuess::computeRandFirstIG(double* ig,double t)
{
  cerr << "CLInitialGuess::computeRandFirstIG : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief compute random initial guesses for all samples of the state space
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* CLInitialGuess::computeRandFirstIG()
{
  cerr << "CLInitialGuess::computeRandFirstIG : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief compute a regular sampled control to determine the first initial guess
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* CLInitialGuess::computeSCFirstIG()
{
  std::list<N2R*>* lfig = new std::list<N2R*>;

  double minAcc = -sys()->getAMax();
  double maxAcc = sys()->getAMax();
  double minStSpd = -sys()->getMaxSteerSpd();
  double maxStSpd = sys()->getMaxSteerSpd();

  // compute the set of final times
  double* timeSp = new double[SPG_T];
  for(int i=0; i<SPG_T; i++) timeSp[i] = MIN_SPG_T + (((double)i) * (MAX_SPG_T - MIN_SPG_T)/((double)(SPG_T-1)));

  // compute the set of initial velocity
  double* ivelSp = new double[SPG_VI];
  for(int j=0; j<SPG_VI; j++) ivelSp[j] = MIN_SPG_VI + (((double)j) * (MAX_SPG_VI - MIN_SPG_VI)/((double)(SPG_VI-1)));
  
  // compute the set of available acceleration controls
  double* accSp = new double[SC_AccSp_];
  for(int k=0; k<SC_AccSp_; k++) accSp[k] = minAcc + (((double)k) * (maxAcc - minAcc)/((double)(SC_AccSp_-1)));

  // compute the set of available steering speed controls
  double* stspdSp = new double[SC_StSpdSp_];
  for(int l=0; l<SC_StSpdSp_; l++) stspdSp[l] = minStSpd + (((double)l) * (maxStSpd - minStSpd)/((double)(SC_StSpdSp_-1)));

  // add an IG for each 4-uplet of (initial velocity, acceleration, steering speed control, time)
  for(int j=0; j<SPG_VI; j++)
    {
      double ivel = ivelSp[j];
	  
      for(int k=0; k<SC_AccSp_; k++)
	{
	  double acc = accSp[k];
	      
	  for(int l=0; l<SC_StSpdSp_; l++)
	    {
	      double stspd = stspdSp[l];

	      // compute the initial state-time
	      CLStateT initST(0,0,0,0,ivel,0);
		  
	      // compute the control to apply
	      CLControlT ctrlt(acc,stspd,MAX_SPG_T);

	      // generate a trajectory from initST by applying the control ctrl
	      CLTrajGen trajgen(sys());

	      trajgen.setUseVelCst(USE_VEL_CST);
	      trajgen.setUseAccCst(USE_ACC_CST);
	      trajgen.setUseAVelCst(USE_AVEL_CST);
	      trajgen.setUseAAccCst(USE_AACC_CST);

	      trajgen.setDBG(0);
	      trajgen.ComputeFromCtrl(initST,ctrlt);

	      // get the parameters of the trajectory generator corresponding to the control
	      double params[7];
	      trajgen.finalParams(params);

	      // get back the state reached at each time sample 
	      for(int i=0; i<SPG_T; i++)
		{
		  double t = timeSp[i];

		  // check if the trajectory is interesting enough to be kept for the IG database
		  cout << "fabs(trajgen.funcTheta(t)) : " << fabs(trajgen.funcTheta(t));
		  if(fabs(trajgen.funcTheta(t)) < 1.55*M_PI)
		    {
		      double curParams[7];
		      for(int indP=0; indP<6; indP++) curParams[indP] = params[indP];
		      curParams[6] = t;

		      CLStateT finalST = trajgen.getCLStateTReachedAt(t);
		      
		      // create the node corresponding to this sample
		      CL_N2R* newNodeIG = new CL_N2R(initST,finalST,curParams);
		      
		      // append it to the list of nodes
		      lfig->push_back(newNodeIG);

		      cout << "  kept" << endl;
		    }
		  else cout << "  discarded" << endl;
		}
	    }
	}
    }

  // delete the sampling vectors
  delete [] timeSp;
  delete [] ivelSp;
  delete [] accSp;
  delete [] stspdSp;

  cout << "fin SC FIG" << endl;

  return lfig;
}

/**
 *  \brief compute a regular sampled parameterization of the control to determine the first initial guess
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* CLInitialGuess::computeSCPFirstIG()
{
  std::list<N2R*>* lfig = new std::list<N2R*>;

  double minAcc = -sys()->getAMax();
  double maxAcc = sys()->getAMax();
  double minStSpd = -sys()->getMaxSteerSpd();
  double maxStSpd = sys()->getMaxSteerSpd();

  // compute the set of final times
  double* timeSp = new double[SPG_T];
  for(int i=0; i<SPG_T; i++) timeSp[i] = MIN_SPG_T + (((double)i) * (MAX_SPG_T - MIN_SPG_T)/((double)(SPG_T-1)));

  // compute the set of initial velocity
  double* ivelSp = new double[SPG_VI];
  for(int j=0; j<SPG_VI; j++) ivelSp[j] = MIN_SPG_VI + (((double)j) * (MAX_SPG_VI - MIN_SPG_VI)/((double)(SPG_VI-1)));
  
  // compute the set of available acceleration controls
  double* accSp = new double[SC_AccSp_];
  for(int k=0; k<SC_AccSp_; k++) accSp[k] = minAcc + (((double)k) * (maxAcc - minAcc)/((double)(SC_AccSp_-1)));

  // compute the set of available steering speed controls
  double* stspdSp = new double[SC_StSpdSp_];
  for(int l=0; l<SC_StSpdSp_; l++) stspdSp[l] = minStSpd + (((double)l) * (maxStSpd - minStSpd)/((double)(SC_StSpdSp_-1)));

  // add an IG for each 4-uplet of (initial velocity, acceleration, steering speed control, time)
  for(int j=0; j<SPG_VI; j++)
    {
      double ivel = ivelSp[j];
	  
      for(int k=0; k<SC_AccSp_; k++)
	{
	  double acc = accSp[k];

	  double jMin = ((double)k)/((double)(SC_AccSp_-1))*(minAcc/((double)MAX_SPG_T));
	  double jMax = ((double)(SC_AccSp_-1-k))/((double)(SC_AccSp_-1))*(maxAcc/((double)MAX_SPG_T));
	      
	  for(int m=0; m<SC_AccSp_; m++)
	    {
	      double jerk = jMin + ((double)m)/((double)(SC_AccSp_-1)) * (jMax-jMin);

	      for(int l=0; l<SC_StSpdSp_; l++)
		{
		  double stspd = stspdSp[l];

		  double L = this->sys()->getLength();
		  double vI = ivel;
		  double vG = vI + acc*MAX_SPG_T;
		  if(sys()->canMoveBack() && vG < -sys()->getVMax()) vG = -sys()->getVMax();
		  if(!(sys()->canMoveBack()) && vG < 0.0) vG = 0.0;
		  if(vG > sys()->getVMax()) vG = sys()->getVMax();
		  double phiI = 0;
		  double phiG = stspd*MAX_SPG_T;
		  if(phiG < -sys()->getMaxSteering()) phiG = -sys()->getMaxSteering();
		  if(phiG > sys()->getMaxSteering()) phiG = sys()->getMaxSteering();
		  double tanPhiI = tan(phiI);
		  double tanPhiG = tan(phiG);

		  double aaccI = (1.0 / L) * (acc*tanPhiI + vI*stspd*(1.0 + tanPhiI*tanPhiI));
		  double aaccF = (1.0 / L) * (acc*tanPhiG + vG*stspd*(1.0 + tanPhiG*tanPhiG));
		  double aaccMoy = 0.5*(aaccI+aaccF);

		  double minAAcc = aaccMoy - (aaccF - aaccI);
		  double maxAAcc = aaccMoy + (aaccF - aaccI);

		  double ajMin = ((double)l)/((double)(SC_StSpdSp_-1))*(minAAcc/((double)MAX_SPG_T));
		  double ajMax = ((double)(SC_StSpdSp_-1-l))/((double)(SC_StSpdSp_-1))*(maxAAcc/((double)MAX_SPG_T));

		  for(int n=0; n<SC_StSpdSp_; n++)
		    {
		      double ajerk = ajMin + ((double)n)/((double)(SC_StSpdSp_-1)) * (ajMax-ajMin);

		      // compute the initial state-time
		      CLStateT initST(0,0,0,0,ivel,0);
		  
		      // compute the parametric control to apply
		      double params[7];
		      params[0] = aaccI;
		      if(MAX_SPG_T != 0) params[1] = (aaccF-aaccI)/((double)MAX_SPG_T);
		      else params[1] = 0;
		      params[2] = ajerk;
		      params[3] = acc;
		      params[4] = jerk;
		      params[5] = 0.0;
		      params[6] = MAX_SPG_T;


// 		      // compute the control to apply
// 		      CLControlT ctrlt(acc,stspd,MAX_SPG_T);

		      // generate a trajectory from initST by applying the control ctrl
		      CLTrajGen trajgen(sys());

		      trajgen.setUseVelCst(USE_VEL_CST);
		      trajgen.setUseAccCst(USE_ACC_CST);
		      trajgen.setUseAVelCst(USE_AVEL_CST);
		      trajgen.setUseAAccCst(USE_AACC_CST);

		      trajgen.setDBG(0);
// 		      trajgen.ComputeFromCtrl(initST,ctrlt);
		      trajgen.ComputeFromIG(initST,params);

// 		      // get the parameters of the trajectory generator corresponding to the control
// 		      double params[7];
// 		      trajgen.finalParams(params);

		      // get back the state reached at each time sample 
		      for(int i=0; i<SPG_T; i++)
			{
			  double t = timeSp[i];

			  // check if the trajectory is interesting enough to be kept for the IG database
// 			  cout << "fabs(trajgen.funcTheta(t)) : " << fabs(trajgen.funcTheta(t));
			  if(fabs(trajgen.funcTheta(t)) < 1.25*M_PI)
			    {
			      double curParams[7];
			      for(int indP=0; indP<6; indP++) curParams[indP] = params[indP];
			      curParams[6] = t;

			      CLStateT finalST = trajgen.getCLStateTReachedAt(t);
		      
			      // create the node corresponding to this sample
			      CL_N2R* newNodeIG = new CL_N2R(initST,finalST,curParams);
		      
			      // append it to the list of nodes
			      lfig->push_back(newNodeIG);

// 			      cout << "  kept" << endl;
			    }
// 			  else cout << "  discarded" << endl;
			}
		    }
		}
	    }
	}
    }

  // delete the sampling vectors
  delete [] timeSp;
  delete [] ivelSp;
  delete [] accSp;
  delete [] stspdSp;

  cout << "fin SCP FIG" << endl;

  return lfig;
}

/**
 *  \brief compute a  to determine the first initial guess
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* CLInitialGuess::computeRStarFirstIG()
{
  cerr << "CLInitialGuess::computeRStarFirstIG : Not implemented yet!" << endl;
  exit(1);
}

/**
 *  \brief load the first initial guess from a file
 *  \return the resulting list of first initial guesses
 */
std::list<N2R*>* CLInitialGuess::loadFirstIGfromFile()
{
  std::list<N2R*>* lfig = new std::list<N2R*>;

  ifstream fic(FIGfilename_, ios::in);

  if(fic)
    {
      if(fic.eof())
	{
	  cerr << "CLInitialGuess::loadFirstIGfromFile : The input file is empty..." << endl;
	  fic.close();
	  throw string("The input file is empty...");

	  delete lfig;
	  return NULL;
	}

      int dimShere,dimPhere;
      // get back the dimension of the state-time vector
      fic >> dimShere;
      // get back the dimension of the parameters vector
      fic >> dimPhere;
      
      // check if these parameters correspond to the current IG database
      if(dimShere != dimS_ || dimPhere != dimP_)
	{
	  cerr << "CLInitialGuess::loadFirstIGfromFile : Wrong input list of trajectories..." << endl;
	  fic.close();
	  throw string("Wrong input list of trajectories...");

	  delete lfig;
	  return NULL;
	}

      while(!fic.eof())
	{
	  CLStateT initST, finalST;

	  // get back the initial state-time of the trajectory
	  fic >> initST.x();
	  fic >> initST.y();
	  fic >> initST.th();
	  fic >> initST.ph();
	  fic >> initST.s();
	  fic >> initST.t();

	  // get back the final state-time of the trajectory
	  fic >> finalST.x();
	  fic >> finalST.y();
	  fic >> finalST.th();
	  fic >> finalST.ph();
	  fic >> finalST.s();
	  fic >> finalST.t();

	  // get back the parameterized control of the node
	  double* params = new double[dimP_];
	  for(int i=0; i<dimP_; i++) fic >> params[i];

	  // create the node corresponding to this sample
	  CL_N2R* newNodeIG = new CL_N2R(initST,finalST,params);

	  // append it to the list of nodes
	  lfig->push_back(newNodeIG);

	  // delete the parameters vector
	  delete [] params;
	}
    }
  else
    {
      cout << "CLInitialGuess::loadFirstIGfromFile : Unable to open the input file" << endl;
      throw string("Unable to open the input file");

      delete lfig;
      return NULL;
    }

  cout << "List of trajectories loaded from file " << FIGfilename_ << endl;

  fic.close();

  return lfig;
}

/**
 *  \brief try to generate a trajectory to the input node n2r thanks to the parameters of the node near
 *  \param n2r : the input node to reach
 *  \param near : the nearest node chosen as initial guess to converge to n2r
 *  \param param : the resulting parametrized control used to reach the goal defined if the trajectory generation succeed
 *  \return 0 if the node n2r has been reached, -1 else
 *
 *  The node n2r if updated if it has been reached
 */
int CLInitialGuess::genTrajTo(N2R* n2r,N2R* near,double* param,bool argh) const
{
  // get back the initial and final states between which a trajectory must be compted
  CLStateT initST = ((CL_N2R*)n2r)->initST();
  CLStateT goalST = ((CL_N2R*)n2r)->goalST();
  double t_f = goalST.t();
  
  // initialize the table of parameters
  const double* paramsNear = near->params();
  double* params = new double[7];
  params[0] = paramsNear[0];
  params[1] = paramsNear[1];
  params[2] = paramsNear[2];
  params[3] = paramsNear[3];
  params[4] = paramsNear[4];
  params[5] = paramsNear[5];
  params[6] = t_f;
//   params[6] = paramsNear[6];
  
  CLTrajGen trajgen(sys());

  trajgen.setUseVelCst(USE_VEL_CST);
  trajgen.setUseAccCst(USE_ACC_CST);
  trajgen.setUseAVelCst(USE_AVEL_CST);
  trajgen.setUseAAccCst(USE_AACC_CST);

  trajgen.fixFinalXY(FIXED_XY);
  trajgen.fixFinalTH(FIXED_TH);
  trajgen.fixFinalV(FIXED_V);
  trajgen.fixFinalPH(FIXED_PH);
  trajgen.setTcStatus(TIME_FIXED,0,999999999);

  trajgen.setDBG(0);
  trajgen.setMaxNbStep(maxTGSteps_);
  trajgen.setMaxResidual(maxTGResidual_);

  int res = trajgen.ComputeTrajectory(initST,goalST,&params);
  
  if(res == 0)
    {
      double finalParams[7];
      trajgen.finalParams(finalParams);

//       cout << "paramsNear : "; displayTab(paramsNear,7); cout << "  -> params : ";displayTab(params,7); cout << endl;
      param[0] = finalParams[0];
      param[1] = finalParams[1];
      param[2] = finalParams[2];
      param[3] = finalParams[3];
      param[4] = finalParams[4];
      param[5] = finalParams[5];
      param[6] = finalParams[6];
    }
  
  delete [] params;
  return res;

//   param[0] = 0;
//   param[1] = 0;
//   param[2] = 0;
//   param[3] = 0;
//   param[4] = 0;
//   param[5] = 0;
//   param[6] = 0;

//   return 0;
}

// /**
//  *  \brief generate the trajectories reaching states-times of the grid surrounding the first IGs
//  *  \param lfig : the input list of the first initial guesses
//  */
// void CLInitialGuess::genTrajFIG(const std::list<N2R*>& l)
// {
// }

// /**
//  *  \brief learn the initial guess for the sample set
//  */
// void CLInitialGuess::learnIG()
// {
//   Timer timer;
//   if(debug_ >= 0) timer.start();

//   // initilialize the sampling grid
//   this->initGrid();

//   // compute the first initial guess
//   std::list<N2R*>* lfig;
//   lfig = this->computeFirstIG();
// //   if(debug_ >= 2) drawGPTraj(*lfig);

//   // expand the set of final nodes reached
//   int nntsr = 999999999; // number of new target state-times reached
//   int step = 0;
  
// //   while(nntsr > minNNTSR_ && step < 1)
//   while(nntsr > minNNTSR_)
//     {
//       // increase the number of the current step
//       step++;
//       // initialize the number of nodes reached
//       nntsr = 0;

//       std::list<N2R*> lN2R;
//       if(lfig == NULL)
// 	{
// 	  // determine the nodes to reach from the nodes already reached
// 	  lN2R = this->determineNodesToReach();

// 	  if(debug_ >= 2) drawGPGrid();
// 	}
//       else
// 	{
// 	  // determine the nodes to reach around the final state-times of initial guesses
// 	  lN2R = this->determineNodesToReach(*lfig);

// 	  if(debug_ >= 2) drawGPTraj(*lfig);

// 	  // delete initial guesses
// 	  if(lfig != NULL) delete lfig;
// 	  lfig = NULL;
// 	}
      
// //       if(debug_ >= 2) this->drawGPGrid();

//       int N = lN2R.size();
//       int cptBar = 0;
//       int stepBar = 1;
//       int nbStepBar = 20;
//       std::list<N2R*>::iterator it;

//       if(debug_ >= 0) cout << "[" << flush; //fflush(std::cout);

//       for(it = lN2R.begin(); it != lN2R.end(); it++)
// 	{
// 	  // get back the current node to try to reach
// 	  N2R* n2r = *it;
// // 	  cout << "n2r : "; n2r->display(); cout << endl;
	  
// 	  // select a random node from this set
// 	  CL_N2R* near = ((CL_N2R*)n2r)->getRandNeighbour();

// 	  // try to generate a trajectory to n2r thanks to the parameters of near
// 	  double* param = new double[dimP_];
// 	  int res = this->genTrajTo((CL_N2R*)n2r,near,param);

// 	  if(res == 0)
// 	    {
// 	      // mark the node as reached
// 	      grid_->setNodeAsReached(n2r->pos(),param);
	      
// 	      // update the number of reached nodes if needed
// 	      nntsr++;
// 	    }

// 	  delete [] param;

// 	  if(debug_ >= 0)
// 	    {
// 	      cptBar++;
// 	      if(cptBar >= stepBar * ((double)N)/((double)nbStepBar))
// 		{
// 		  cout << "=" << flush; //fflush(std::cout);
// 		  stepBar++;
// 		}
// 	    }
// 	}

//       if(debug_ >= 0) cout << "]" << endl;

//       double prct;
//       if(N != 0) prct = ((double)nntsr)/((double)N)*100.0;
//       else prct = 0.0;
//       if(debug_ >= 1) cout << "step " << step << " -> Reached : " << nntsr << "/" << N << " (" << prct << "\%)   ";
//       if(debug_ >= 1) cout << "Total reached : " << grid_->NReachedonGrid() << "(" << grid_->NReachedinList() << ")"
// 			   << "  Total used : " << grid_->NUsedonGrid() << "(" << grid_->NUsedinList() << ")" << endl;

//       if(debug_ >= 0)
// 	{
// 	  timer.stop();
// 	  cout << "Elapsed time : " << timer.totalElapsedTime() << endl;
// 	}

//       if(debug_ >= 2) this->drawGPGrid();
//     }

//   this->save("CL_IG.dat");
// }


/********************* Query **********************/

/**
 *  \brief query the initial guess base
 *  \param is : the initial state-time
 *  \param is : the goal state-time
 *  \return the best set of parameters found
 */
const double* CLInitialGuess::queryIG(const CLStateT& is,const CLStateT& gs) const
{
  // search for the nearest defined node of the grid to compute a trajectory between both input state-times
  N2R* node = ((CL_N2RGrid*)grid_)->nearestNodeToComputeTrajBetweenV2(is,gs);
  
  // check if a node has been defined at the selected position
  if(node != NULL)
    {
      cout << "node found" << endl;
      cout << "selected params : "; displayTab(node->params(),7); cout << endl;
      return node->params();
    }
  else 
    {
      cout << "no node found" << endl;
      return NULL;
    }
}


/************* Input/Output Functions *************/

// /**
//  *  \brief load the sampling of initial guesses from an input file
//  *  \param nomfic : the name of the input file
//  *  \return 0 if the process succeed, -1 else
//  */
// int CLInitialGuess::load(char* nomfic)
// {
//   ifstream fic(nomfic, ios::in);

//   if(fic)
//     {
//       if(fic.eof())
// 	{
// 	  cout << "CLInitialGuess::load : The input file is empty..." << endl;
// 	  fic.close();
// 	  throw string("The input file is empty...");
// 	  return -1;
// 	}

//       int dimShere,dimPhere,dimSampShere;
//       // get back the dimension of the state-time vector
//       fic >> dimShere;
//       // get back the dimension of the parameters vector
//       fic >> dimPhere;
//       // get back the dimension of the parameters vector
//       fic >> dimSampShere;
      
//       // check if these parameters correspond to the current IG database
//       if(dimShere != dimS_ || dimPhere != dimP_ || dimSampShere != dimSampS_)
// 	{
// 	  cout << "CLInitialGuess::load : Wrong input IG database..." << endl;
// 	  fic.close();
// 	  throw string("Wrong input IG database...");
// 	  return -1;
// 	}

//       // get back the sampling vectors
//       for(int i=0; i<dimSampS_; i++) fic >> sampling_[i];
//       for(int i=0; i<dimSampS_; i++) fic >> minSampS_[i];
//       for(int i=0; i<dimSampS_; i++) fic >> maxSampS_[i];

//       while(!fic.eof())
// 	{
// 	  // get back the position of the node in the grid
// 	  int* pos = new int[dimS_];
// 	  for(int i=0; i<dimS_; i++)fic >> pos[i];

// 	  // get back the parameterized control of the node
// 	  double* params = new double[dimP_];
// 	  for(int i=0; i<dimP_; i++) fic >> params[i];

// 	  // create the new reached node and put it in the grid
// 	  N2R* node = grid_->createReachedNode(pos,params);

// 	  // delete the position and parameters
// 	  delete [] pos;
// 	  delete [] params;
// 	}
//     }
//   else
//     {
//       cout << "CLInitialGuess::load : Unable to open the input file" << endl;
//       throw string("Unable to open the input file");
//       return -1;
//     }

//   if(debug_ >= 1) cout << "IG database loaded" << endl;
//   if(debug_ >= 2) cout << "Total reached : " << grid_->NReachedonGrid() << "(" << grid_->NReachedinList() << ")"
// 		       << "  Total used : " << grid_->NUsedonGrid() << "(" << grid_->NUsedinList() << ")" << endl;

//   cout << "Car-like IG database loaded from file " << nomfic << endl;

//   fic.close();

//   if(debug_ >= 2) grid_->displayReachedOnGrid();
  
//   //grid_->displayGridSampling();
//   return 0;
// }

// /**
//  *  \brief save the sampling of initial guesses into an output file
//  *  \param nomfic : the name of the output file
//  *  \return 0 if the process succeed, -1 else
//  */
// int CLInitialGuess::save(char* nomfic)
// {
//   ofstream fic(nomfic, ios::out | ios::trunc);
 
//   if(fic)
//     {
//       // write the dimension of the state-time vector
//       fic << dimS_ << " ";
//       // write the dimension of the parameters vector
//       fic << dimP_ << " ";
//       // write the dimension of the parameters vector
//       fic << dimSampS_ << endl;

//       // write the sampling vectors
//       for(int i=0; i<dimSampS_-1; i++) fic << sampling_[i] << " ";
//       if(dimSampS_ >= 1) fic << sampling_[dimSampS_-1] << endl;
//       for(int i=0; i<dimSampS_-1; i++) fic << minSampS_[i] << " ";
//       if(dimSampS_ >= 1) fic << minSampS_[dimSampS_-1] << endl;
//       for(int i=0; i<dimSampS_-1; i++) fic << maxSampS_[i] << " ";
//       if(dimSampS_ >= 1) fic << maxSampS_[dimSampS_-1] << endl;

//       // get back the list of reached and used nodes
//       std::list<N2R*> lNReached = grid_->listNReached();
//       std::list<N2R*> lNUsed = grid_->listNUsed();

//       // put the location and the parameters of each completed trajectory of the reached nodes list
//       std::list<N2R*>::iterator itR;
//       for(itR = lNReached.begin(); itR != lNReached.end(); itR++)
// 	{
// 	  int nS = grid_->dimS();
// 	  int nP = (*itR)->dimP();

// 	  int* coord = (*itR)->pos();
// 	  double* param = (*itR)->params();

// 	  for(int i=0; i<nS; i++) fic << coord[i] << " ";
// 	  for(int j=0; j<nP-1; j++) fic << param[j] << " ";
// 	  if(nP >=1) fic << param[nP-1] << endl;
// 	}

//       // put the location and the parameters of each completed trajectory of the reached nodes list
//       std::list<N2R*>::iterator itU;
//       for(itU = lNUsed.begin(); itU != lNUsed.end(); itU++)
// 	{
// 	  int nS = grid_->dimS();
// 	  int nP = (*itU)->dimP();

// 	  int* coord = (*itU)->pos();
// 	  double* param = (*itU)->params();

// 	  for(int i=0; i<nS; i++) fic << coord[i] << " ";
// 	  for(int j=0; j<nP-1; j++) fic << param[j] << " ";
// 	  if(nP >=1) fic << param[nP-1] << endl;
// 	}      
      
//       fic.close();
//     }
//   else
//     {
//       cout << "CLInitialGuess::save : Unable to open the output file" << endl;
//       return -1;
//     }

//   cout << "Car-like IG database saved in file " << nomfic << endl;
//   return 0;
// }


/*************** Utility Functions ****************/

/**
 *  \brief list all nodes of the grid
 *  \return a pointer to a vector containing all nodes of the grid
 */
std::vector<N2R*>* CLInitialGuess::listAllNodes() const
{
  int N = computeNbTargetST();
  std::vector<N2R*>* nodes = new std::vector<N2R*>(N);

//   int p=0;

//   for(int i=0; i<SPG_T; i++)
//     for(int j=0; j<SPG_VI; j++)
//       for(int k=0; k<SPG_VG; k++)
// 	for(int l=0; l<SPG_DX; l++)
// 	  for(int m=0; m<SPG_DY; m++)
// 	    for(int n=0; n<SPG_DTH; n++)
// 	      {
// 		(*nodes)[p] = spg_[i][j][k][l][m][n];
// 		p++;
// 	      }

  return nodes;
}

/**
 *  \brief list all reached nodes of the grid
 *  \return a pointer to a vector containing all reached nodes of the grid
 */
std::vector<N2R*>* CLInitialGuess::listAllReachedNodes() const
{
  int N = computeNbTargetST();
  std::vector<N2R*>* nodes = new std::vector<N2R*>(N);

  int p=0;

//   for(int i=0; i<SPG_T; i++)
//     for(int j=0; j<SPG_VI; j++)
//       for(int k=0; k<SPG_VG; k++)
// 	for(int l=0; l<SPG_DX; l++)
// 	  for(int m=0; m<SPG_DY; m++)
// 	    for(int n=0; n<SPG_DTH; n++)
// 	      {
// 		// add the node to the vector only if it has been reached
// 		if(spg_[i][j][k][l][m][n]->isReached())
// 		  {
// 		    (*nodes)[p] = spg_[i][j][k][l][m][n];
// 		    p++;
// 		  }
// 	      }

  nodes->resize(p);
  return nodes;
}

/**
 *  \brief sample the trajectory defined on the node into tables of position and time
 *  \param node : the input node containing a trajectory
 *  \param tabX : the resulting table containing the position X values
 *  \param tabY : the resulting table containing the position Y values
 *  \param tabT : the resulting table containing the time T values
 *  \param n : the number of elements defined into tables
 *
 *  Tables must be allocated first
 *  The parameterized control of the node must have been defined first
 */
void CLInitialGuess::sampleTrajNode(N2R* node,double* tabX,double* tabY,double *tabT,int n) const
{
  // get back the initial and final state-times of the trajectory
  CLStateT s0 = ((CL_N2R*)node)->initST();
  CLStateT sf = ((CL_N2R*)node)->goalST();

  // get back the parameterized control of the trajectory
  const double* param = node->params();

  // generate the trajectory from the initial state and the parameterized control
  CLTrajGen trajgen(sys());

  trajgen.setUseVelCst(USE_VEL_CST);
  trajgen.setUseAccCst(USE_ACC_CST);
  trajgen.setUseAVelCst(USE_AVEL_CST);
  trajgen.setUseAAccCst(USE_AACC_CST);

  trajgen.setDBG(0);
  trajgen.ComputeFromIG(s0,param);
  
  // sample the trajectory generated
  trajgen.SaveTrajectory(n);

  // copy the sampled trajectory into the resulting tables
  for(int i=0; i<n; i++)
    {
      tabX[i] = trajgen.posX->at(i);
      tabY[i] = trajgen.posY->at(i);
      tabT[i] = trajgen.timeT->at(i);
    }
}

// /**
//  *  \brief display in a gnuplot window the generated trajectories of the grid
//  */
// void CLInitialGuess::drawGPTraj()
// {
//   cerr << "CLInitialGuess::drawGPTraj : Not implemented yet!" << endl;
//   exit(1);
// }
