#include "CLTrajGen.hpp"
using namespace mpc;

#include "matrix.h"

#include "Polynomial.hpp"
#include "Utils.hpp"

#include <stdio.h>
#include <math.h>
#include <vector>
#include "utilsLuis.h"
#include <complex>

#include "InvMatrix.hpp"
#include "gnuplot/inc/gnuplot_i.hpp"

#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;

// Parameters
#define B      (*vecParameters_p)[0]
#define C      (*vecParameters_p)[1]
#define D      (*vecParameters_p)[2]
#define B2     (*vecParameters_p)[3]
#define C2     (*vecParameters_p)[4]
#define D2     (*vecParameters_p)[5]
#define T      (*vecParameters_p)[6]

// States 
#define X0     vecState_x0[0]
#define Y0     vecState_x0[1]
#define TH0    vecState_x0[2]
#define OM0    vecState_x0[3]
#define V0     vecState_x0[4]
#define T0     vecState_x0[5]

#define Xg     vecState_xf[0]
#define Yg     vecState_xf[1]
#define THg    vecState_xf[2]
#define OMg    vecState_xf[3]
#define Vg     vecState_xf[4]
#define Tg     vecState_xf[5]

#define Xsim   vecState_xsim[0]
#define Ysim   vecState_xsim[1]
#define THsim  vecState_xsim[2]
#define OMsim  vecState_xsim[3]
#define Vsim   vecState_xsim[4]
#define Tsim   vecState_xsim[5]

#define Xind   matStateInd_[0]
#define Yind   matStateInd_[1]
#define THind  matStateInd_[2]
#define OMind  matStateInd_[3]
#define Vind   matStateInd_[4]
#define Tind   matStateInd_[5]

#define Xerr   vecConstraints_g[Xind]
#define Yerr   vecConstraints_g[Yind]
#define THerr  vecConstraints_g[THind]
#define OMerr  vecConstraints_g[OMind]
#define Verr   vecConstraints_g[Vind]

#define L      L_

#define CONST_ACC_BOUND true
#define CONST_VEL_BOUND true

// #define TIME_FIXED true

#define COEF_COR 0.2
#define VARIABLE_COEF_COR true

#define DFLT_DERIV_VAL -999999999.99

#define FAST_PART_DERIV true

#define COEF_WHENBDD 0.2

#define GSL_USED false

#define MAXABS_POS_ERR    0.1
#define MAXREL_POS_ERR    0.01

bool MUST_DISPLAY;

double minBoundAcc;
double maxBoundAcc;
double minBoundVel;
double maxBoundVel;




CLTrajGen::CLTrajGen()
{
  xLength = 6;       // State length (x, y, th, ph, v, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 5;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, ph(t)=ph_f, v(t)=v_f)
  dLength = 6;       // Dynamic Constraints length (maxAcc, maxSt, ...)

  USE_VEL_CST = true;
  USE_ACC_CST = true;
  USE_AVEL_CST = true;
  USE_AACC_CST = false;

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);

  posX = NULL;
  posY = NULL;
  timeT = NULL;
  velV = NULL;
  steerS = NULL;

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  fctAcc_ = NULL;
  fctVel_ = NULL;
  fctAVel_ = NULL;
  fctAAcc_ = NULL;
  fctTH_ = NULL;
  fctAccWithoutBounds_ = NULL;
  fctVelWithoutBounds_ = NULL;
  fctAVelWithoutBounds_ = NULL;
  fctAAccWithoutBounds_ = NULL;
  fctTHWithoutBounds_ = NULL;

  xyFixed_ = true;
  thFixed_ = true;
  vFixed_ = true;
  phFixed_ = true;

  matStateInd_ = new int[xLength];

  CLweights_ = CLState(1,1,1,1,1);

  regX.clear();
  regY.clear();
  regTH.clear();
  regV.clear();
  regAV.clear();
  regT.clear();
}

/*
 *  \brief init constructor
 *  \param sys : the constraints of the car-like system
 */
CLTrajGen::CLTrajGen(const CarLikeSys* sys) : TrajGen(sys)
{
  xLength = 6;       // State length (x, y, th, ph, v, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 5;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, ph(t)=ph_f, v(t)=v_f)
  dLength = 6;       // Dynamic Constraints length (maxAcc, maxSt, ...)

  USE_VEL_CST = true;
  USE_ACC_CST = true;
  USE_AVEL_CST = true;
  USE_AACC_CST = false;

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);

  posX = NULL;
  posY = NULL;
  timeT = NULL;
  velV = NULL;
  steerS = NULL;

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  fctAcc_ = NULL;
  fctVel_ = NULL;
  fctAVel_ = NULL;
  fctAAcc_ = NULL;
  fctTH_ = NULL;
  fctAccWithoutBounds_ = NULL;
  fctVelWithoutBounds_ = NULL;
  fctAVelWithoutBounds_ = NULL;
  fctAAccWithoutBounds_ = NULL;
  fctTHWithoutBounds_ = NULL;

  xyFixed_ = true;
  thFixed_ = true;
  vFixed_ = true;
  phFixed_ = true;

  matStateInd_ = new int[6];

  CLweights_ = CLState(1,1,1,1,1);

  regX.clear();
  regY.clear();
  regTH.clear();
  regV.clear();
  regAV.clear();
  regT.clear();
}

// destructor
CLTrajGen::~CLTrajGen()
{
  if(posX != NULL) delete posX;
  if(posY != NULL) delete posY;
  if(timeT != NULL) delete timeT;
  if(velV != NULL) delete velV;
  if(steerS != NULL) delete steerS;  

  if(fctAcc_ != NULL) delete fctAcc_;
  if(fctVel_ != NULL) delete fctVel_;
  if(fctAVel_ != NULL) delete fctAVel_;
  if(fctAAcc_ != NULL) delete fctAAcc_;
  if(fctTH_ != NULL) delete fctTH_;
  
  if(fctAccWithoutBounds_ != NULL) delete fctAccWithoutBounds_;
  if(fctVelWithoutBounds_ != NULL) delete fctVelWithoutBounds_;
  if(fctAVelWithoutBounds_ != NULL) delete fctAVelWithoutBounds_;
  if(fctAAccWithoutBounds_ != NULL) delete fctAAccWithoutBounds_;
  if(fctTHWithoutBounds_ != NULL) delete fctTHWithoutBounds_;

  if(matStateInd_ != NULL) delete [] matStateInd_;
}


/*************** Getters and Setters **************/

// return true if linear velocity constraints are used
bool CLTrajGen::getUseVelCst()
{
  return this->USE_VEL_CST;
}

// set true if linear velocity constraints are used
void CLTrajGen::setUseVelCst(bool u)
{
  this->USE_VEL_CST = u;
}

// return true if linear acceleration constraints are used
bool CLTrajGen::getUseAccCst()
{
  return this->USE_ACC_CST;
}

// set true if linear acceleration constraints are used
void CLTrajGen::setUseAccCst(bool u)
{
  this->USE_ACC_CST = u;
}

// return true if angular velocity constraints are used
bool CLTrajGen::getUseAVelCst()
{
  return this->USE_AVEL_CST;
}

// set true if angular velocity constraints are used
void CLTrajGen::setUseAVelCst(bool u)
{
  this->USE_AVEL_CST = u;
}

// return true if angular acceleration constraints are used
bool CLTrajGen::getUseAAccCst()
{
  return this->USE_AACC_CST;
}

// set true if angular acceleration constraints are used
void CLTrajGen::setUseAAccCst(bool u)
{
  this->USE_AACC_CST = u;
}

/*
 *  \brief fix or free the final position
 *  \param fixed : true if the final position must be fixed
 */
void CLTrajGen::fixFinalXY(bool fixed)
{
  xyFixed_ = fixed;
}

/*
 *  \brief fix or free the final heading
 *  \param fixed : true if the final heading must be fixed
 */
void CLTrajGen::fixFinalTH(bool fixed)
{
  thFixed_ = fixed;
}

/*
 *  \brief fix or free the final velocity
 *  \param fixed : true if the final velocity must be fixed
 */
void CLTrajGen::fixFinalV(bool fixed)
{
  vFixed_ = fixed;
}

/*
 *  \brief fix or free the final steering angle
 *  \param fixed : true if the final steering angle must be fixed
 */
void CLTrajGen::fixFinalPH(bool fixed)
{
  phFixed_ = fixed;
}

/**
 *  \brief set the car-like state weights for the cost function
 *  \param wXY : position weight
 *  \param wTH : heading weight
 *  \param wV  : speed weight
 *  \param wPH : steering weight
 */
void CLTrajGen::setCLweights(double wXY,double wTH,double wV,double wPH)
{
  CLweights_.x()  = wXY;
  CLweights_.y()  = wXY;
  CLweights_.th() = wTH;
  CLweights_.s()  = wV;
  CLweights_.ph() = wPH;
}

/*
 *  \brief compute the residual between the goal state and the stae reached
 *  \return the residual between the goal state and the stae reached
 */
double CLTrajGen::ComputeResidual() const
{
  double sqRes = 0;
  
  if(xyFixed_)
    {
      sqRes += (Xerr * Xerr);
      sqRes += (Yerr * Yerr);
    }
  if(thFixed_) sqRes += (50.0 * (THerr * THerr));
  if(phFixed_) sqRes += (10.0 * (OMerr * OMerr));
  if(vFixed_) sqRes += (10.0 * (Verr * Verr));
  
  return sqrt(sqRes);
}

// Quadrature function
double CLTrajGen::qgaus(double (CLTrajGen::*func)(const double) const, const double a, const double b) const
{
    static const int n = 20;
    double x[n];
    double w[n];
    int j;
    double s;
    
    gauleg(a, b, x, w, n);      
    s=0;
    for (j=0;j<n;j++) {        
        s += w[j]*((this->*func)(x[j]));
    }
    return s;
}

double CLTrajGen::funcItgX(const double t) const
{
  return funcVel(t) * cos(normalizedAngle(funcTheta(t)));
}

double CLTrajGen::funcItgY(const double t) const
{
  return funcVel(t) * sin(normalizedAngle(funcTheta(t)));
}

double CLTrajGen::funcX(const double t) const
{
  return vecState_x0[0] + qgaus(&CLTrajGen::funcItgX,0,t);
}

double CLTrajGen::funcY(const double t) const
{
  return vecState_x0[1] + qgaus(&CLTrajGen::funcItgY,0,t);
}

double CLTrajGen::funcTheta(const double t) const
{
  return fctTH_->getVal(t);
}

double CLTrajGen::funcAVel(const double t) const
{
  return fctAVel_->getVal(t);
}

double CLTrajGen::funcSteer(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;

  double linSpd = funcVel(t2);
  double angSpd = funcAVel(t2);
  
  if(fabs(linSpd) < 1e-10)
    return 0;
  else
    return atan(L * angSpd / linSpd);
}

double CLTrajGen::funcSteerSpd(const double t) const
{
  double linAcc   = funcAcc(t);
  double angAcc   = funcAAcc(t);
  double steer    = funcSteer(t);
  double tanSteer = tan(steer);
  double linVel   = funcVel(t);
  
  if(linVel == 0)
    return 0;
  else
    return (L*angAcc - linAcc*tanSteer) / (linVel * (1.0 + tanSteer*tanSteer));
}

double CLTrajGen::funcAAcc(const double t) const
{
  return this->fctAAcc_->getVal(t);
}

double CLTrajGen::funcAcc(const double t) const
{
  return fctAcc_->getVal(t);
}

double CLTrajGen::funcVel(const double t) const
{
  return fctVel_->getVal(t);
}

double CLTrajGen::funcItgXwb(const double t) const
{
  return funcVelwb(t) * cos(normalizedAngle(funcThetawb(t)));
}

double CLTrajGen::funcItgYwb(const double t) const
{
  return funcVelwb(t) * sin(normalizedAngle(funcThetawb(t)));
}

double CLTrajGen::funcXwb(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;
  return vecState_x0[0] + qgaus(&CLTrajGen::funcItgXwb,0,t2);
}

double CLTrajGen::funcYwb(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;
  return vecState_x0[1] + qgaus(&CLTrajGen::funcItgYwb,0,t2);
}

double CLTrajGen::funcThetawb(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;
  return fctTHWithoutBounds_->getVal(t2);
}

double CLTrajGen::funcAVelwb(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;

  return (OM0 + B2*t2 + C2*t2*t2 + D2*t2*t2*t2);
}

double CLTrajGen::funcAAccwb(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;

  return (B + 2.0*C*t2 + 3.0*D*t2*t2);
}

double CLTrajGen::funcVelwb(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;

  return (V0 + B2*t2 + C2*t2*t2 + D2*t2*t2*t2);
}

double CLTrajGen::funcAccwb(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;

  return (B2 + 2.0*C2*t2 + 3.0*D2*t2*t2);
}

double CLTrajGen::funcSteerwb(const double t) const
{
  double t2;
  if(t < 0) t2 = 0;
  else t2 = t;

  double linSpd = funcVelwb(t2);
  double angSpd = funcAVelwb(t2);
  
  if(fabs(linSpd) < 1e-10)
    return 0;
  else
    return atan(L * angSpd / linSpd);
}

double CLTrajGen::funcSteerSpdwb(const double t) const
{
  double linAcc   = funcAccwb(t);
  double angAcc   = funcAAccwb(t);
  double steer    = funcSteerwb(t);
  double tanSteer = tan(steer);
  double linVel   = funcVelwb(t);

  return (L*angAcc - linAcc*tanSteer) / (linVel * (1.0 + tanSteer*tanSteer));
}

void CLTrajGen::computeFctAcc()
{
  B2a = B2; C2a = C2; D2a = D2;

  // delete the previous acceleration function
  if(fctAcc_ != NULL) delete fctAcc_;
  if(fctAccWithoutBounds_ != NULL) delete fctAccWithoutBounds_;

  // create the polynomial acceleration function
  fctAccWithoutBounds_ = new Polynomial(3.0*D2,2.0*C2,B2);
  fctAcc_ = new MultiItvFunction(fctAccWithoutBounds_, 0,INFINITY);

//   MUST_DISPLAY = true;
//   if(MUST_DISPLAY)
//     {
//       cout << "1111111111111111111111111111111111111111111111111111111111111"<<endl;
//       cout << "fctVel avant bounding" << endl;
//       fctVel_ = (MultiItvFunction*) fctAcc_->integration(0.0,vecState_x0[4]);
//       fctVel_->display();
      
//       //Computing Linear Velocity
//       vector<double> acc,vel,minaccb,maxaccb,minvelb,maxvelb;
//       vector<double> distance;
//       for(int j=0;j<=200;j++){
// 	acc.push_back((double)this->funcAcc(j*(S/200)));
// 	minaccb.push_back(-maxAcc);
// 	maxaccb.push_back(maxAcc);
// 	vel.push_back((double)this->funcVel(j*(S/200)));
// 	minvelb.push_back(0.0);
// 	maxvelb.push_back(maxSpd);
// 	distance.push_back((j*(S/200)));
//       }

//       Gnuplot gAcc;
//       gAcc.reset_plot();
//       gAcc.set_style("lines").set_yrange(-14,8).plot_xy(distance,acc,"Linear Acceleration");
//       gAcc.set_style("points").set_yrange(-14,8).plot_xy(distance,minaccb,"Minimal Bound");
//       gAcc.set_style("points").set_yrange(-14,8).plot_xy(distance,maxaccb,"Maximal Bound");  

//       Gnuplot gVel;
//       gVel.reset_plot();
//       gVel.set_style("lines").set_yrange(-5,85).plot_xy(distance,vel,"Linear Velocity");
//       gVel.set_style("points").set_yrange(-5,85).plot_xy(distance,minvelb,"Minimal Bound");
//       gVel.set_style("points").set_yrange(-5,85).plot_xy(distance,maxvelb,"Maximal Bound");  
      
//       cout << "2222222222222222222222222222222222222222222222222222222222222"<<endl;
//       wait_for_key();
//     }
//   MUST_DISPLAY = false;

  if(USE_ACC_CST)
    {
//       if(!CONST_ACC_BOUND && CNT < 0.2*maxNbStep_)
// 	{
// 	  minBoundAcc = -maxAcc - (40.0*maxAcc*(0.2*maxNbStep_ - CNT)/(0.2*maxNbStep_));
// 	  maxBoundAcc = maxAcc + (40.0*maxAcc*(0.2*maxNbStep_ - CNT)/(0.2*maxNbStep_));
// 	}
//       else
// 	{
	  minBoundAcc = -maxAcc;
	  maxBoundAcc = maxAcc;
// 	}
      
      // add the bounds over the acceleration
	  list<ItvFunction> lMinBItv;
	  fctAcc_->addMinBound(minBoundAcc,&lMinBItv);
	  list<ItvFunction> lMaxBItv;
	  fctAcc_->addMaxBound(maxBoundAcc,&lMaxBItv);

//       MUST_DISPLAY = true;
//       if(MUST_DISPLAY)
// 	{
// 	  cout << "3333333333333333333333333333333333333333333333333333333333333"<<endl;
// 	  cout << "fctVel avant bounding" << endl;
// 	  fctVel_ = (MultiItvFunction*) fctAcc_->integration(0.0,vecState_x0[4]);
// 	  fctVel_->display();
	  
// 	  //Computing Linear Velocity
// 	  vector<double> acc,vel,minaccb,maxaccb,minvelb,maxvelb;
// 	  vector<double> distance;
// 	  for(int j=0;j<=200;j++){
// 	    acc.push_back((double)this->funcAcc(j*(S/200)));
// 	    minaccb.push_back(-maxAcc);
// 	    maxaccb.push_back(maxAcc);
// 	    vel.push_back((double)this->funcVel(j*(S/200)));
// 	    minvelb.push_back(0.0);
// 	    maxvelb.push_back(maxSpd);
// 	    distance.push_back((j*(S/200)));
// 	  }
	  
// 	  Gnuplot gAcc;
// 	  gAcc.reset_plot();
// 	  gAcc.set_style("lines").set_yrange(-14,8).plot_xy(distance,acc,"Linear Acceleration");
// 	  gAcc.set_style("points").set_yrange(-14,8).plot_xy(distance,minaccb,"Minimal Bound");
// 	  gAcc.set_style("points").set_yrange(-14,8).plot_xy(distance,maxaccb,"Maximal Bound");  
	  
// 	  Gnuplot gVel;
// 	  gVel.reset_plot();
// 	  gVel.set_style("lines").set_yrange(-5,85).plot_xy(distance,vel,"Linear Velocity");
// 	  gVel.set_style("points").set_yrange(-5,85).plot_xy(distance,minvelb,"Minimal Bound");
// 	  gVel.set_style("points").set_yrange(-5,85).plot_xy(distance,maxvelb,"Maximal Bound");
	  
// 	  cout << "44444444444444444444444444444444444444444444444444444444444444"<<endl;
// 	  wait_for_key();
// 	}
//       MUST_DISPLAY = false;

      if(ItvFunction::isIncludeIn(T,&lMinBItv)) lastItvAccBdd_ = -1;
      else if(ItvFunction::isIncludeIn(T,&lMaxBItv)) lastItvAccBdd_ = 1;
      else lastItvAccBdd_ = 0;
    }
}

void CLTrajGen::computeFctVel()
{
  B2v = B2; C2v = C2; D2v = D2;

  // delete the previous velocity function
  if(fctVel_ != NULL) delete fctVel_;

  // compute the velocity function by integrating the acceleration
  //fctVelWithoutBounds_ = new Polynomial(D2,C2,B2,vecState_x0[4]);
  fctVel_ = (MultiItvFunction*) fctAcc_->integration(0.0,vecState_x0[4]);

  if(USE_VEL_CST)
    {
      if(sys()->canMoveBack()) minBoundVel = -maxSpd;
      else minBoundVel = 0.0;
      maxBoundVel = maxSpd;
      
      list<ItvFunction> lMinBItv;
      fctVel_->addMinBound(minBoundVel,&lMinBItv);
      list<ItvFunction> lMaxBItv;
      fctVel_->addMaxBound(maxBoundVel,&lMaxBItv);

//       if(ItvFunction::isIncludeIn(T,&lMinBItv)) lastItvVelBdd_ = -1;
//       else if(ItvFunction::isIncludeIn(T,&lMaxBItv)) lastItvVelBdd_ = 1;
//       else lastItvVelBdd_ = 0;      

      if(DBG_ >= 1) {
	// recompute the acceleration if the velocity changed
	delete fctAcc_;
	fctAcc_ = (MultiItvFunction*) fctVel_->derivation();
      }
    }
}

void CLTrajGen::computeFctAAcc()
{
  Baa = B; Caa = C; Daa = D;

  if(fctAAccWithoutBounds_ != NULL) delete fctAAccWithoutBounds_;
  if(fctAAcc_ != NULL) delete fctAAcc_;

  fctAAccWithoutBounds_ = new Polynomial(3.0*D,2.0*C,B);
  fctAAcc_ = new MultiItvFunction(fctAAccWithoutBounds_, 0,INFINITY);

  if(USE_AACC_CST)
    {
      // compute the piecewise polynomial bound to apply on the angular acceleration profile
      double tanMaxSt = tan(maxSt);
      double coef1 = tanMaxSt / L;
      double coef2 = (tanMaxSt*tanMaxSt + 1.0) * maxStSpd / L;

//       cout << "coef1 : " << coef1 << endl;
//       cout << "coef2 : " << coef2 << endl;

      MultiItvFunction aaccBound1(*fctAcc_);
      aaccBound1 *= coef1;
      aaccBound1.addAbsoluteValue();
      MultiItvFunction aaccBound2(*fctVel_);
      aaccBound2 *= coef2;
      aaccBound2.addAbsoluteValue();
      aaccBound1 += aaccBound2;

      MultiItvFunction aaccMinBound(aaccBound1);
      aaccMinBound *= (-1.0);

//       double tanMaxSt2 = tan(-maxSt);
//       double coef3 = tanMaxSt2 / L;
//       double coef4 = (tanMaxSt2*tanMaxSt2 + 1.0) * (-maxStSpd) / L;

// //       cout << "coef1 : " << coef1 << endl;
// //       cout << "coef2 : " << coef2 << endl;

//       MultiItvFunction aaccBound3(*fctAcc_);
//       aaccBound3 *= coef3;
//       MultiItvFunction aaccBound4(*fctVel_);
//       aaccBound4 *= coef4;
//       aaccBound3 += aaccBound4;


//       MUST_DISPLAY = false;
//       if(MUST_DISPLAY)
// 	{
// 	  cout << "1111111111111111111111111111111111111111111111111111111111111"<<endl;
// 	  cout << "fctAAcc avant bounding" << endl;
// 	  fctAAcc_->display();
	  
// 	  //Computing Angular Acceleration
// 	  vector<double> aacc,aacc2,minb,maxb;
// 	  vector<double> distanceAA;
// 	  for(int j=0;j<=30;j++){
// 	    aacc.push_back((double)this->funcAAcc(j*(S/30)));
// 	    aacc2.push_back((double)this->fctAAcc_->getVal(j*(S/30)));
// 	    minb.push_back(aaccMinBound.getVal(j*(S/30)));
// 	    //minb.push_back(aaccBound3.getVal(j*(S/30)));
// 	    maxb.push_back(aaccBound1.getVal(j*(S/30)));
// 	    distanceAA.push_back((j*(S/30)));
// 	  }    
// 	  Gnuplot gAAcc;
// 	  gAAcc.reset_plot();
// 	  gAAcc.set_style("lines").plot_xy(distanceAA,aacc,"Angular Acceleration");
// 	  gAAcc.set_style("lines").plot_xy(distanceAA,aacc2,"Angular Acceleration2");
// 	  gAAcc.set_style("points").plot_xy(distanceAA,minb,"Minimal Bound");
// 	  gAAcc.set_style("points").plot_xy(distanceAA,maxb,"Maximal Bound");  
	  
// 	  cout << "2222222222222222222222222222222222222222222222222222222222222"<<endl;
// 	  wait_for_key();
// 	}
//       MUST_DISPLAY = false;
      
//       cout << "borne min sur aacc : " << endl;
//       aaccMinBound.display();
//       cout << "borne max sur aacc : " << endl;
//       aaccBound1.display();
      
      // add the minimal and maximal bounds over the angular acceleration profile
      list<ItvFunction> lMinBItv;
      fctAAcc_->addMinBound(&aaccMinBound,&lMinBItv);
      
//       if(lMinBItv->size() == 0) 
// 	cout << "pas de borne min sur aacc" << endl;
//       else 
// 	{
// 	  cout << "borne min sur aacc" << endl;
// 	  list<ItvFunction>::iterator it = lMinBItv->begin();
// 	  while(it != lMinBItv->end())
// 	    {
// 	      it->display();
// 	      it++;
// 	    }
// 	}
      
//       MUST_DISPLAY = false;
//       if(MUST_DISPLAY)
// 	{
// 	  cout << "3333333333333333333333333333333333333333333333333333333333333"<<endl;
// 	  cout << "fctAAcc apres bmin" << endl;
// 	  fctAAcc_->display();

// 	  //Computing Angular Acceleration
// 	  vector<double> aacc,aacc2,minb,maxb;
// 	  vector<double> distanceAA;
// 	  for(int j=0;j<=30;j++){
// 	    aacc.push_back((double)this->funcAAcc(j*(S/30)));
// 	    aacc2.push_back((double)this->fctAAcc_->getVal(j*(S/30)));
// 	    minb.push_back(aaccMinBound.getVal(j*(S/30)));
// 	    //minb.push_back(aaccBound3.getVal(j*(S/30)));
// 	    maxb.push_back(aaccBound1.getVal(j*(S/30)));
// 	    distanceAA.push_back((j*(S/30)));
// 	  }    
// 	  Gnuplot gAAcc;
// 	  gAAcc.reset_plot();
// 	  gAAcc.set_style("lines").plot_xy(distanceAA,aacc,"Angular Acceleration");
// 	  gAAcc.set_style("lines").plot_xy(distanceAA,aacc2,"Angular Acceleration2");
// 	  gAAcc.set_style("points").plot_xy(distanceAA,minb,"Minimal Bound");
// 	  gAAcc.set_style("points").plot_xy(distanceAA,maxb,"Maximal Bound");  
// 	  cout << "4444444444444444444444444444444444444444444444444444444444444"<<endl;
// 	  wait_for_key();
// 	}
//       MUST_DISPLAY = false;
      
      list<ItvFunction> lMaxBItv;
      fctAAcc_->addMaxBound(&aaccBound1,&lMaxBItv);

//       if(lMaxBItv->size() == 0) 
// 	cout << "pas de borne max sur aacc" << endl;
//       else 
// 	{
// 	  cout << "borne max sur aacc" << endl;
// 	  list<ItvFunction>::iterator it = lMaxBItv->begin();
// 	  while(it != lMaxBItv->end())
// 	    {
// 	      it->display();
// 	      it++;
// 	    }
// 	}
      
//       MUST_DISPLAY = false;
//       if(MUST_DISPLAY)
// 	{
// 	  cout << "5555555555555555555555555555555555555555555555555555555555555"<<endl;	  
// 	  cout << "fctAAcc apres bmax" << endl;
// 	  fctAAcc_->display();

// 	  //Computing Angular Acceleration
// 	  vector<double> aacc,aacc2,minb,maxb;
// 	  vector<double> distanceAA;
// 	  for(int j=0;j<=30;j++){
// 	    aacc.push_back((double)this->funcAAcc(j*(S/30)));
// 	    aacc2.push_back((double)this->fctAAcc_->getVal(j*(S/30)));
// 	    minb.push_back(aaccMinBound.getVal(j*(S/30)));
// 	    //minb.push_back(aaccBound3.getVal(j*(S/30)));
// 	    maxb.push_back(aaccBound1.getVal(j*(S/30)));
// 	    distanceAA.push_back((j*(S/30)));
// 	  }    
// 	  Gnuplot gAAcc;
// 	  gAAcc.reset_plot();
// 	  gAAcc.set_style("lines").plot_xy(distanceAA,aacc,"Angular Acceleration");
// 	  gAAcc.set_style("lines").plot_xy(distanceAA,aacc2,"Angular Acceleration2");
//  	  gAAcc.set_style("points").plot_xy(distanceAA,minb,"Minimal Bound");
// 	  gAAcc.set_style("points").plot_xy(distanceAA,maxb,"Maximal Bound");  
// 	  cout << "6666666666666666666666666666666666666666666666666666666666666"<<endl;
// 	  wait_for_key();
// 	}
//       MUST_DISPLAY = false;
      
      // check if the last interval has been bounded
      if(ItvFunction::isIncludeIn(T,&lMinBItv)) lastItvAAccBdd_ = -1;
      else if(ItvFunction::isIncludeIn(T,&lMaxBItv)) lastItvAAccBdd_ = 1;
      else lastItvAAccBdd_ = 0;
    }
}

void CLTrajGen::computeFctAVel()
{
  Bav = B; Cav = C; Dav = D;

  // delete the previous angular velocity function
  if(fctAVel_ != NULL) delete fctAVel_;
  if(fctAVelWithoutBounds_ != NULL) delete fctAVelWithoutBounds_;

  //double OM0 = V0 * tan(PH0) / L;

  fctAVelWithoutBounds_ = new Polynomial(D,C,B,OM0);

//   if(USE_AACC_CST)
//     {
      fctAVel_ = (MultiItvFunction*) (fctAAcc_->integration(0.0,OM0));
//     }
//   else
//     {
//       fctAVel_ = new MultiItvFunction(0,INFINITY,fctAVelWithoutBounds_);
//     }

  if(USE_AVEL_CST)
    {
      // compute the constant coef tan(Phi_max)/L
      //double coef = lambda_; 
      double coef = fabs(tan(maxSt) / L);
      
//       cout << "0000000000000000000000000000000000000000000000000000000000000"<<endl;
//       cout << "L : " << L << endl;
//       cout << "maxSt : " << maxSt << endl;
//       cout << "coef : " << coef << endl;
//       cout << "fctVel : " << endl;
//       fctVel_->display();
//       cout << "0000000000000000000000000000000000000000000000000000000000000"<<endl;

      // compute the bound to apply to the angular velocity : v(t) * tan(Phi_max)/L

      MultiItvFunction absVel(*fctVel_);
      absVel.addAbsoluteValue();

      MultiItvFunction minBoundAVel(absVel);
      minBoundAVel *= (-coef);
      MultiItvFunction maxBoundAVel(absVel);
      maxBoundAVel *= coef;
      
//       cout << "##################### BOUNDS  : #####################" << endl;
//       cout << "min bound : " << endl;
//       minBoundAVel.display();
//       cout << "max bound : " << endl;
//       maxBoundAVel.display();
//       cout << "#####################################################" << endl;


//       if(MUST_DISPLAY)
// 	{
// 	  cout << "1111111111111111111111111111111111111111111111111111111111111"<<endl;
// 	  cout << "fctAVel_ : " << endl;
// 	  fctAVel_->display();
// 	  cout << "minBoundAVel : " << endl;
// 	  minBoundAVel.display();
// 	  cout << "maxBoundAVel : " << endl;
// 	  maxBoundAVel.display();
	  
// 	  //Computing Angular Velocity
// 	  vector<double> avel,minb,maxb;
// 	  vector<double> distanceAV;
// 	  for(int j=0;j<=30;j++){
// 	    avel.push_back((double)this->funcAVel(j*(S/30)));
// 	    minb.push_back(minBoundAVel.getVal(j*(S/30)));
// 	    maxb.push_back(maxBoundAVel.getVal(j*(S/30)));
// 	    distanceAV.push_back((j*(S/30)));
// 	  }    
// 	  Gnuplot gAVel;
// 	  gAVel.reset_plot();
// 	  gAVel.set_style("lines").plot_xy(distanceAV,avel,"Angular Velocity");
// 	  gAVel.set_style("lines").plot_xy(distanceAV,minb,"Minimal Bound");
// 	  gAVel.set_style("lines").plot_xy(distanceAV,maxb,"Maximal Bound");  
	  
// 	  cout << "2222222222222222222222222222222222222222222222222222222222222"<<endl;
// 	  wait_for_key();
// 	}
      
      list<ItvFunction> lMinBItv;
      fctAVel_->addMinBound(&minBoundAVel,&lMinBItv);
      
//       if(MUST_DISPLAY)
// 	{
// 	  cout << "3333333333333333333333333333333333333333333333333333333333333"<<endl;
// 	  cout << "fctAVel_ : " << endl;
// 	  fctAVel_->display();
// 	  cout << "minBoundAVel : " << endl;
// 	  minBoundAVel.display();
// 	  cout << "maxBoundAVel : " << endl;
// 	  maxBoundAVel.display();
	  
// 	  //Computing Angular Velocity
// 	  vector<double> avel,minb,maxb;
// 	  vector<double> distanceAV;
// 	  for(int j=0;j<=30;j++){
// 	    avel.push_back((double)this->funcAVel(j*(S/30)));
// 	    minb.push_back(minBoundAVel.getVal(j*(S/30)));
// 	    maxb.push_back(maxBoundAVel.getVal(j*(S/30)));
// 	    distanceAV.push_back((j*(S/30)));
// 	  }    
// 	  Gnuplot gAVel;
// 	  gAVel.reset_plot();
// 	  gAVel.set_style("lines").plot_xy(distanceAV,avel,"Angular Velocity apres1");
// 	  gAVel.set_style("lines").plot_xy(distanceAV,minb,"Minimal Bound");
// 	  gAVel.set_style("lines").plot_xy(distanceAV,maxb,"Maximal Bound");  
	  
// 	  cout << "4444444444444444444444444444444444444444444444444444444444444"<<endl;
// 	  wait_for_key();
// 	}
      
      list<ItvFunction> lMaxBItv;
      fctAVel_->addMaxBound(&maxBoundAVel,&lMaxBItv);
      
//       if(MUST_DISPLAY)
// 	{
// 	  cout << "5555555555555555555555555555555555555555555555555555555555555"<<endl;
// 	  cout << "fctAVel_ : " << endl;
// 	  fctAVel_->display();
// 	  cout << "minBoundAVel : " << endl;
// 	  minBoundAVel.display();
// 	  cout << "maxBoundAVel : " << endl;
// 	  maxBoundAVel.display();
	  
// 	  //Computing Angular Velocity
// 	  vector<double> avel,minb,maxb;
// 	  vector<double> distanceAV;
// 	  for(int j=0;j<=30;j++)
// 	    {
// 	      avel.push_back((double)this->funcAVel(j*(S/30)));
// 	      minb.push_back(minBoundAVel.getVal(j*(S/30)));
// 	      maxb.push_back(maxBoundAVel.getVal(j*(S/30)));
// 	      distanceAV.push_back((j*(S/30)));
// 	    }
// 	  Gnuplot gAVel;
// 	  gAVel.reset_plot();
// 	  gAVel.set_style("lines").plot_xy(distanceAV,avel,"Angular Velocity apres2");
// 	  gAVel.set_style("lines").plot_xy(distanceAV,minb,"Minimal Bound");
// 	  //gAVel.set_style("lines").plot_xy(distanceAV,maxb,"Maximal Bound");  
	  
// 	  cout << "6666666666666666666666666666666666666666666666666666666666666"<<endl;
// 	  wait_for_key();
// 	}

      if(ItvFunction::isIncludeIn(T,&lMinBItv)) lastItvAVelBdd_ = -1;
      else if(ItvFunction::isIncludeIn(T,&lMaxBItv)) lastItvAVelBdd_ = 1;
      else lastItvAVelBdd_ = 0;
    }

  if(DBG_ >= 1) {
    // recompute the angular acceleration if the angular velocity changed
    delete fctAAcc_;
    fctAAcc_ = (MultiItvFunction*) fctAVel_->derivation();
    //cout << "derivation" << endl;
  }
}

void CLTrajGen::computeFctTH()
{
  // delete the previous heading functions
  if(fctTH_ != NULL) delete fctTH_;
  if(fctTHWithoutBounds_ != NULL) delete fctTHWithoutBounds_;

//   Polynomial itgTHwb(*fctVelWithoutBounds_);
//   itgTHwb *= (*fctAVelWithoutBounds_);
  Polynomial itgTHwb(*fctAVelWithoutBounds_);
  fctTHWithoutBounds_ = ((Polynomial*) (itgTHwb.integration(0.0,vecState_x0[2])));

  // th(t) = \int_0^TMAX v(t) om(t) dt
//   MultiItvFunction itgTH(*fctVel_);
//   itgTH *= (*fctAVel_);
//   fctTH_ = ((MultiItvFunction*) (itgTH.integration(0.0,vecState_x0[2])));
  fctTH_ = ((MultiItvFunction*) (fctAVel_->integration(0.0,vecState_x0[2])));
}

void CLTrajGen::computeAngularFct()
{
  computeFctAAcc();
  computeFctAVel();
  computeFctTH();
}

void CLTrajGen::computeAllFct()
{
  computeFctAcc();
  computeFctVel();
  computeFctAAcc();
  computeFctAVel();
  computeFctTH();
}

void CLTrajGen::ForwardSolution()
{
  //cout << "B : " << B << "  C : " << C << "  D : " << D << "  B2 : " << B2 << "  C2 : " << C2 << "  D2 : " << D2 << endl;

  // initialize the partial derivatives
  for(int i=0; i<gLength; i++)
//     for(int j=0; j<pLength-1; j++)
    for(int j=0; j<pLengthFree(); j++)
      matJacobianG[i][j] = DFLT_DERIV_VAL;

  lastItvAccBdd_ = 0;
  lastItvVelBdd_ = 0;
  lastItvAVelBdd_ = 0;

  // compute the different functions' profiles
  //MUST_DISPLAY = true;
  computeAllFct();
  //MUST_DISPLAY = false;

  // compute the simulated state
//   vecState_xsim[2] = funcTheta(S);
//   vecState_xsim[3] = funcAVel(S);
//   vecState_xsim[0] = funcX(S);
//   vecState_xsim[1] = funcY(S);
//   vecState_xsim[4] = funcVel(S);
  THsim = funcTheta(T);
  OMsim = funcAVel(T);
  Xsim = funcX(T);
  Ysim = funcY(T);
  Vsim = funcVel(T);
  if(tcStatus() != TIME_FIXED) Tsim = T;

  //cout << "fm : (" << vecState_xsim[0] << "," << vecState_xsim[1] << "," << vecState_xsim[2] << "," << vecState_xsim[3] << "," << vecState_xsim[4] << ")" << endl;

  //wait_for_key();

  //this->displayProfiles();
}

void CLTrajGen::setInitialGuess(double b,double c,double d,double b2,double c2,double d2,double t)
{
  if(vecParameters_p != NULL) delete vecParameters_p;
  vecParameters_p = new Vec_DP(pLength);

  B = b;
  C = c;
  D = d;
  B2 = b2;
  C2 = c2;
  D2 = d2;
  T = t;
}

/**
 *  \brief compute an initial guess of parameters to perform maneuvers
 */
void CLTrajGen::ComputeManeuverIG()
{
  //Inital guess of parameters
  double dist = sqrt(pow(vecState_xf[0]-vecState_x0[0],2) + pow(vecState_xf[1]-vecState_x0[1],2));
  double dir = atan2(vecState_xf[1]-vecState_x0[1],vecState_xf[0]-vecState_x0[0]);
  double difDir = computeMinDiffAngle(normalizedAngle(dir),normalizedAngle(vecState_x0[2]));
  
  //cout << "difDir : " << difDir << endl;
  
  if(DBG_ >= 1) {
    cout << "vecState_xf[5] : " << vecState_xf[5] << endl;
    cout << "vecState_x0[5] : " << vecState_x0[5] << endl;
  }
  
  T = Tg - T0;
  
  if(T <= 0)
    {
      B = 0;
      C = 0;
      D = 0;
      B2 = 0;
      C2 = 0;
      D2 = 0;
    }
  else
    {
      double coefDist = 1.0 + (0.5*M_PI - 1)*(fabs(difDir)/M_PI);
      //double vmoy = (dist*coefDist)/T;
      double vmoy = dist/T;
      
//       B = (-2.0*OM0 + 0.2*difDir) / (2.0*T);
//       C = 0.5*OM0 / (T*T);
//       D = 0;
      
      B = 0;
      C = 0;
      D = 0;
      
      if(DBG_ >= 1) {
	printf ("Initial Guess Parameters K = a=%f,b=%f,c=%f,d=%f,s=%f)\n", 
		vecState_x0[3],B,C,D,T);
      }
      
      B2 = -0.15*sys()->getAMax();  //-5.0*(vmoy - V0) / T;
      C2 = -1.2*B2/T;
      D2 = 0.0;
      
      if(DBG_ >= 1) {    
	printf ("Initial Guess Parameters V = a=%f,b=%f,c=%f,d=%f,s=%f)\n", 
		vecState_x0[4],B2,C2,D2,T);
      }
    }
}

/**
 *  \brief compute a default initial guess of parameters
 */
void CLTrajGen::ComputeDfltIG()
{
//   this->ComputeManeuverIG();
//   return;

  //Inital guess of parameters
  double dist = sqrt(pow(vecState_xf[0]-vecState_x0[0],2) + pow(vecState_xf[1]-vecState_x0[1],2));
  double dir = atan2(vecState_xf[1]-vecState_x0[1],vecState_xf[0]-vecState_x0[0]);
  double difDir = computeMinDiffAngle(normalizedAngle(dir),normalizedAngle(vecState_x0[2]));
  
  //cout << "difDir : " << difDir << endl;
  
  if(DBG_ >= 1) {
    cout << "vecState_xf[5] : " << vecState_xf[5] << endl;
    cout << "vecState_x0[5] : " << vecState_x0[5] << endl;
  }
  
  T = Tg - T0;
  
  if(T <= 0)
    {
      B = 0;
      C = 0;
      D = 0;
      B2 = 0;
      C2 = 0;
      D2 = 0;
    }
  else
    {
      double coefDist = 1.0 + (0.5*M_PI - 1)*(fabs(difDir)/M_PI);
      //double vmoy = (dist*coefDist)/T;
      double vmoy = dist/T;
      
//       B = (-2.0*OM0 + 0.2*difDir) / (2.0*T);
//       C = 0.5*OM0 / (T*T);
//       D = 0;
      
      B = 0;
      C = 0;
      D = 0;
      
      if(DBG_ >= 1) {
	printf ("Initial Guess Parameters K = a=%f,b=%f,c=%f,d=%f,s=%f)\n", 
		vecState_x0[3],B,C,D,T);
      }
      
      B2 = 2.0*(vmoy - V0) / T;
      C2 = 0.0;
      D2 = 0.0;
      
      if(DBG_ >= 1) {    
	printf ("Initial Guess Parameters V = a=%f,b=%f,c=%f,d=%f,s=%f)\n", 
		vecState_x0[4],B2,C2,D2,T);
      }
    }
}

/**
 *  \brief select an initial guess of parameters from an IG database
 *  \return 0 if a set of parameters has been found, -1 else
 */
int CLTrajGen::ComputeIGfromDB()
{
  //       cout << "An initial guess database has been found" << endl;
  double PH0;
  if(V0 != 0) PH0 = atan(L*OM0/V0);
  else PH0 = 0;
  double PHg;
  if(Vg != 0) PHg = atan(L*OMg/Vg);
  else PHg = 0;
  CLStateT initST(X0,Y0,TH0,PH0,V0,T0);
  CLStateT goalST(Xg,Yg,THg,PHg,Vg,Tg);
  const double* param = ((CLInitialGuess*)ig_)->queryIG(initST,goalST);
  
  if(param != NULL)
    {
      B = param[0];
      C = param[1];
      D = param[2];
      B2 = param[3];
      C2 = param[4];
      D2 = param[5];
      T = param[6];
      
      //delete [] param;
      return 0;
    }
  else return -1;
}

void CLTrajGen::ComputeG()
{
  if(xyFixed_) Xerr  = Xsim - Xg;
  if(xyFixed_) Yerr  = Ysim - Yg;
  if(thFixed_) THerr = computeMinDiffAngle(normalizedAngle(THsim),normalizedAngle(THg));
  if(vFixed_) Verr  = Vsim - Vg;
  if(phFixed_) OMerr = OMsim - OMg;

//   if(xyFixed_) Xerr  = Xg - Xsim;
//   if(xyFixed_) Yerr  = Yg - Ysim;
//   if(thFixed_) THerr = computeMinDiffAngle(normalizedAngle(THg),normalizedAngle(THsim));
//   if(vFixed_) Verr  = Vg - Vsim;
//   if(phFixed_) OMerr = OMg - OMsim;
}


// Partial derivative computations

// partial derivative dS/dP (2nd version)
void CLTrajGen::derivSwrtP()
{
//   double lambda = tan(maxSt)/L;

  if(vFixed_)
    {
      // computation of dV/dP
      matJacobianG[Vind][0] = 0.0; // dV/dB
      matJacobianG[Vind][1] = 0.0; // dV/dC
      matJacobianG[Vind][2] = 0.0; // dV/dD
      matJacobianG[Vind][3] = T; // dV/dB2
      matJacobianG[Vind][4] = T*T; // dV/dC2
      matJacobianG[Vind][5] = T*T*T; // dV/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[Vind][6] = funcAcc(T) * T; // dV/dT
    }

  if(phFixed_)
    {
      // computation of dOM/dP
      matJacobianG[OMind][0] = T; // dOM/dB
      matJacobianG[OMind][1] = T*T; // dOM/dC
      matJacobianG[OMind][2] = T*T*T; // dOM/dD
      matJacobianG[OMind][3] = 0.0; // dOM/dB2
      matJacobianG[OMind][4] = 0.0; // dOM/dC2
      matJacobianG[OMind][5] = 0.0; // dOM/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[OMind][6] = funcAAcc(T) * T; // dOM/dT
    }

  if(thFixed_)
    {
      // computation of dTH/dP
      matJacobianG[THind][0] = 0.5*T*T; // dTH/dB
      matJacobianG[THind][1] = 0.333333333*T*T*T; // dTH/dC
      matJacobianG[THind][2] = 0.25*T*T*T*T; // dTH/dD
      matJacobianG[THind][3] = 0.0; // dTH/dB2
      matJacobianG[THind][4] = 0.0; // dTH/dC2
      matJacobianG[THind][5] = 0.0; // dTH/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[THind][6] = funcAVel(T) * T; // dTH/dT
    }

  if(xyFixed_)
    {
      // computation of dX/dP
      matJacobianG[Xind][0] = qgaus(&CLTrajGen::itgdXdB,0,T); // dX/dB
      matJacobianG[Xind][1] = qgaus(&CLTrajGen::itgdXdC,0,T); // dX/dC
      matJacobianG[Xind][2] = qgaus(&CLTrajGen::itgdXdD,0,T); // dX/dD
      matJacobianG[Xind][3] = qgaus(&CLTrajGen::itgdXdB2,0,T); // dX/dB2
      matJacobianG[Xind][4] = qgaus(&CLTrajGen::itgdXdC2,0,T); // dX/dC2
      matJacobianG[Xind][5] = qgaus(&CLTrajGen::itgdXdD2,0,T); // dX/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[Xind][6] = funcItgX(T) * T; // dX/dT
      
      // computation of dY/dP
      matJacobianG[Yind][0] = qgaus(&CLTrajGen::itgdYdB,0,T); // dY/dB
      matJacobianG[Yind][1] = qgaus(&CLTrajGen::itgdYdC,0,T); // dY/dC
      matJacobianG[Yind][2] = qgaus(&CLTrajGen::itgdYdD,0,T); // dY/dD
      matJacobianG[Yind][3] = qgaus(&CLTrajGen::itgdYdB2,0,T); // dY/dB2
      matJacobianG[Yind][4] = qgaus(&CLTrajGen::itgdYdC2,0,T); // dY/dC2
      matJacobianG[Yind][5] = qgaus(&CLTrajGen::itgdYdD2,0,T); // dY/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[Yind][6] = funcItgY(T) * T; // dY/dT
    }
}

double CLTrajGen::itgdXdB(const double t) const
{
  return -0.5*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t;
}

double CLTrajGen::itgdXdC(const double t) const
{
  return -0.333333333*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t;
}

double CLTrajGen::itgdXdD(const double t) const
{
  return -0.25*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t*t;
}

double CLTrajGen::itgdXdB2(const double t) const
{
  return t * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double CLTrajGen::itgdXdC2(const double t) const
{
  return t*t * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double CLTrajGen::itgdXdD2(const double t) const
{
  return t*t*t * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double CLTrajGen::itgdYdB(const double t) const
{
  return 0.5*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t;
}

double CLTrajGen::itgdYdC(const double t) const
{
  return 0.333333333*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t;
}

double CLTrajGen::itgdYdD(const double t) const
{
  return 0.25*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t*t;
}

double CLTrajGen::itgdYdB2(const double t) const
{
  return t * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double CLTrajGen::itgdYdC2(const double t) const
{
  return t*t * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double CLTrajGen::itgdYdD2(const double t) const
{
  return t*t*t * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double CLTrajGen::TH(const double t) const
{
  return TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t;
}
double CLTrajGen::V(const double t) const
{
  return V0 + B2*t + C2*t*t + D2*t*t*t;
}
double CLTrajGen::AV(const double t) const
{
  return OM0 + B*t + C*t*t + D*t*t*t;
}
double CLTrajGen::A(const double t) const
{
  return B2 + 2.0*C2*t + 3.0*D2*t*t;
}
double CLTrajGen::AA(const double t) const
{
  return B + 2.0*C*t + 3.0*D*t*t;
}
double CLTrajGen::J(const double t) const
{
  return 2.0*C2 + 6.0*D2*t;
}
double CLTrajGen::AJ(const double t) const
{
  return 2.0*C + 6.0*D*t;
}

/***************************
 ********* d2X/dP2 *********
 ***************************/

double CLTrajGen::itgd2X_dBdB(const double t) const
{
  return -0.25*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2X_dBdC(const double t) const
{
  return -0.166666667*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2X_dBdD(const double t) const
{
  return -0.125*T*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2X_dBdB2(const double t) const
{
  return -0.5*T*T*T*sin(TH(T));
}
double CLTrajGen::itgd2X_dBdC2(const double t) const
{
  return -0.5*T*T*T*T*sin(TH(T));
}
double CLTrajGen::itgd2X_dBdD2(const double t) const
{
  return -0.5*T*T*T*T*T*sin(TH(T));
}
double CLTrajGen::itgd2X_dCdC(const double t) const
{
  return -0.111111111*T*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2X_dCdD(const double t) const
{
  return -0.083333333*T*T*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2X_dCdB2(const double t) const
{
  return -0.333333333*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2X_dCdC2(const double t) const
{
  return -0.333333333*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2X_dCdD2(const double t) const
{
  return -0.333333333*T*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2X_dDdD(const double t) const
{
  return -0.0625*T*T*T*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2X_dDdB2(const double t) const
{
  return -0.25*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2X_dDdC2(const double t) const
{
  return -0.25*T*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2X_dDdD2(const double t) const
{
  return -0.25*T*T*T*T*T*T*T*V(T)*sin(TH(T));
}

/***************************
 ********* d2Y/dP2 *********
 ***************************/

double CLTrajGen::itgd2Y_dBdB(const double t) const
{
  return -0.25*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2Y_dBdC(const double t) const
{
  return -0.166666667*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2Y_dBdD(const double t) const
{
  return -0.125*T*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2Y_dBdB2(const double t) const
{
  return 0.5*T*T*T*cos(TH(T));
}
double CLTrajGen::itgd2Y_dBdC2(const double t) const
{
  return 0.5*T*T*T*T*cos(TH(T));
}
double CLTrajGen::itgd2Y_dBdD2(const double t) const
{
  return 0.5*T*T*T*T*T*cos(TH(T));
}
double CLTrajGen::itgd2Y_dCdC(const double t) const
{
  return -0.111111111*T*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2Y_dCdD(const double t) const
{
  return -0.083333333*T*T*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2Y_dCdB2(const double t) const
{
  return 0.333333333*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2Y_dCdC2(const double t) const
{
  return 0.333333333*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2Y_dCdD2(const double t) const
{
  return 0.333333333*T*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2Y_dDdD(const double t) const
{
  return -0.0625*T*T*T*T*T*T*T*T*V(T)*sin(TH(T));
}
double CLTrajGen::itgd2Y_dDdB2(const double t) const
{
  return 0.25*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2Y_dDdC2(const double t) const
{
  return 0.25*T*T*T*T*T*T*V(T)*cos(TH(T));
}
double CLTrajGen::itgd2Y_dDdD2(const double t) const
{
  return 0.25*T*T*T*T*T*T*T*V(T)*cos(TH(T));
}
// double CLTrajGen::itgd2X_dB2dB2(const double t) const
// {
// }
// double CLTrajGen::itgd2X_dB2dC2(const double t) const
// {
// }
// double CLTrajGen::itgd2X_dB2dD2(const double t) const
// {
// }
// double CLTrajGen::itgd2X_dC2dC2(const double t) const
// {
// }
// double CLTrajGen::itgd2X_dC2dD2(const double t) const
// {
// }
// double CLTrajGen::itgd2X_dD2dD2(const double t) const
// {
// }


// double CLTrajGen::dXdB(const double t)
// {
//   return -0.166666667*V0*sin(TH0)*t*t*t + (-0.125*V0*cos(TH0)*OM0 - 0.125*B2*sin(TH0))*t*t*t*t;
// }

// double CLTrajGen::dXdC(const double t)
// {
//   return -0.083333333*V0*sin(TH0)*t*t*t*t;
// }

// double CLTrajGen::dXdD(const double t)
// {
//   return 0.0;
// }

// double CLTrajGen::dXdB2(const double t)
// {
//   return 0.5*cos(TH0)*t*t - 0.333333333*sin(TH0)*OM0*t*t*t + (-0.125*cos(TH0)*OM0*OM0 - 0.125*sin(TH0)*B)*t*t*t*t;
// }

// double CLTrajGen::dXdC2(const double t)
// {
//   return 0.333333333*cos(TH0)*t*t*t -0.25*sin(TH0)*OM0*t*t*t*t;
// }

// double CLTrajGen::dXdD2(const double t)
// {
//   return 0.25*cos(TH0)*t*t*t*t;
// }

// double CLTrajGen::dYdB(const double t)
// {
//   return 0.166666666*V0*cos(TH0)*t*t*t + (-0.125*V0*sin(TH0)*OM0 + 0.125*B2*cos(TH0))*t*t*t*t;
// }

// double CLTrajGen::dYdC(const double t)
// {
//   return 0.083333333*V0*cos(TH0)*t*t*t*t;
// }

// double CLTrajGen::dYdD(const double t)
// {
//   return 0.0;
// }

// double CLTrajGen::dYdB2(const double t)
// {
//   return 0.5*sin(TH0)*t*t + 0.333333333*cos(TH0)*OM0*t*t*t + (-0.125*sin(TH0)*OM0*OM0 + 0.125*cos(TH0)*B)*t*t*t*t;
// }

// double CLTrajGen::dYdC2(const double t)
// {
//   return 0.333333333*sin(TH0)*t*t*t + 0.25*cos(TH0)*OM0*t*t*t*t;
// }

// double CLTrajGen::dYdD2(const double t)
// {
//   return 0.25*sin(TH0)*t*t*t*t;
// }

void CLTrajGen::ComputeJacobianS()
{
#if FAST_PART_DERIV
  // compute the fast partial derivative dS/dP
  derivSwrtP();  
#else
  // compute the partial derivative dS/dB
  derivSwrtB();
  // compute the partial derivative dS/dC
  derivSwrtC();
  // compute the partial derivative dS/dD
  derivSwrtD();
  // compute the partial derivative dS/dB2
  derivSwrtB2();
  // compute the partial derivative dS/dC2
  derivSwrtC2();
  // compute the partial derivative dS/dD2
  derivSwrtD2();

  // recompute the forward method with the initial parameters
  computeAllFct();
#endif
}

/**
 *  \brief compute the second order partial derivatives of the state wrt. the parameters
 */
void CLTrajGen::ComputeHessianS()
{
  if(vFixed_)
    {
      // computation of d2V/dBdP                              --- checked
      setHessianS(Vind,0,0,0.0);   // d2V/dBdB
      setHessianS(Vind,0,1,0.0);   // d2V/dBdC
      setHessianS(Vind,0,2,0.0);   // d2V/dBdD
      setHessianS(Vind,0,3,0.0);   // d2V/dBdB2
      setHessianS(Vind,0,4,0.0);   // d2V/dBdC2
      setHessianS(Vind,0,5,0.0);   // d2V/dBdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Vind,0,6,0.0);   // d2V/dBdT

      // computation of d2V/dCdP                              --- checked
      setHessianS(Vind,1,0,0.0);   // d2V/dCdB
      setHessianS(Vind,1,1,0.0);   // d2V/dCdC
      setHessianS(Vind,1,2,0.0);   // d2V/dCdD
      setHessianS(Vind,1,3,0.0);   // d2V/dCdB2
      setHessianS(Vind,1,4,0.0);   // d2V/dCdC2
      setHessianS(Vind,1,5,0.0);   // d2V/dCdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Vind,1,6,0.0);   // d2V/dCdT

      // computation of d2V/dDdP                              --- checked
      setHessianS(Vind,2,0,0.0);   // d2V/dDdB
      setHessianS(Vind,2,1,0.0);   // d2V/dDdC
      setHessianS(Vind,2,2,0.0);   // d2V/dDdD
      setHessianS(Vind,2,3,0.0);   // d2V/dDdB2
      setHessianS(Vind,2,4,0.0);   // d2V/dDdC2
      setHessianS(Vind,2,5,0.0);   // d2V/dDdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Vind,2,6,0.0);   // d2V/dDdT

      // computation of d2V/dB2dP                              --- checked
      setHessianS(Vind,3,0,0.0);   // d2V/dB2dB
      setHessianS(Vind,3,1,0.0);   // d2V/dB2dC
      setHessianS(Vind,3,2,0.0);   // d2V/dB2dD
      setHessianS(Vind,3,3,0.0);   // d2V/dB2dB2
      setHessianS(Vind,3,4,0.0);   // d2V/dB2dC2
      setHessianS(Vind,3,5,0.0);   // d2V/dB2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(Vind,3,6,1.0);   // d2V/dB2dT

      // computation of d2V/dC2dP                              --- checked
      setHessianS(Vind,4,0,0.0);   // d2V/dC2dB
      setHessianS(Vind,4,1,0.0);   // d2V/dC2dC
      setHessianS(Vind,4,2,0.0);   // d2V/dC2dD
      setHessianS(Vind,4,3,0.0);   // d2V/dC2dB2
      setHessianS(Vind,4,4,0.0);   // d2V/dC2dC2
      setHessianS(Vind,4,5,0.0);   // d2V/dC2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(Vind,4,6,2.0*T);   // d2V/dC2dT

      // computation of d2V/dD2dP                              --- checked
      setHessianS(Vind,5,0,0.0);   // d2V/dD2dB
      setHessianS(Vind,5,1,0.0);   // d2V/dD2dC
      setHessianS(Vind,5,2,0.0);   // d2V/dD2dD
      setHessianS(Vind,5,3,0.0);   // d2V/dD2dB2
      setHessianS(Vind,5,4,0.0);   // d2V/dD2dC2
      setHessianS(Vind,5,5,0.0);   // d2V/dD2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(Vind,5,6,3.0*T*T);   // d2V/dD2dT

      if(tcStatus() != TIME_FIXED) 
	{
	  // computation of d2V/dTdP                              --- checked
	  setHessianS(Vind,6,0,0.0);     // d2V/dTdB
	  setHessianS(Vind,6,1,0.0);     // d2V/dTdC
	  setHessianS(Vind,6,2,0.0);     // d2V/dTdD
	  setHessianS(Vind,6,3,1.0);     // d2V/dTdB2
	  setHessianS(Vind,6,4,2.0*T);   // d2V/dTdC2
	  setHessianS(Vind,6,5,3.0*T*T); // d2V/dTdD2
	  setHessianS(Vind,6,6,J(T));    // d2V/dTdT
	}
    }

  if(phFixed_)
    {
      // computation of d2OM/dBdP                              --- checked
      setHessianS(OMind,0,0,0.0);   // d2OM/dBdB
      setHessianS(OMind,0,1,0.0);   // d2OM/dBdC
      setHessianS(OMind,0,2,0.0);   // d2OM/dBdD
      setHessianS(OMind,0,3,0.0);   // d2OM/dBdB2
      setHessianS(OMind,0,4,0.0);   // d2OM/dBdC2
      setHessianS(OMind,0,5,0.0);   // d2OM/dBdD2
      if(tcStatus() != TIME_FIXED) setHessianS(OMind,0,6,1.0);   // d2OM/dBdT

      // computation of d2OM/dCdP                              --- checked
      setHessianS(OMind,1,0,0.0);   // d2OM/dCdB
      setHessianS(OMind,1,1,0.0);   // d2OM/dCdC
      setHessianS(OMind,1,2,0.0);   // d2OM/dCdD
      setHessianS(OMind,1,3,0.0);   // d2OM/dCdB2
      setHessianS(OMind,1,4,0.0);   // d2OM/dCdC2
      setHessianS(OMind,1,5,0.0);   // d2OM/dCdD2
      if(tcStatus() != TIME_FIXED) setHessianS(OMind,1,6,2.0*T);   // d2OM/dCdT

      // computation of d2OM/dDdP                              --- checked
      setHessianS(OMind,2,0,0.0);   // d2OM/dDdB
      setHessianS(OMind,2,1,0.0);   // d2OM/dDdC
      setHessianS(OMind,2,2,0.0);   // d2OM/dDdD
      setHessianS(OMind,2,3,0.0);   // d2OM/dDdB2
      setHessianS(OMind,2,4,0.0);   // d2OM/dDdC2
      setHessianS(OMind,2,5,0.0);   // d2OM/dDdD2
      if(tcStatus() != TIME_FIXED) setHessianS(OMind,2,6,3.0*T*T);   // d2OM/dDdT

      // computation of d2OM/dB2dP                              --- checked
      setHessianS(OMind,3,0,0.0);   // d2OM/dB2dB
      setHessianS(OMind,3,1,0.0);   // d2OM/dB2dC
      setHessianS(OMind,3,2,0.0);   // d2OM/dB2dD
      setHessianS(OMind,3,3,0.0);   // d2OM/dB2dB2
      setHessianS(OMind,3,4,0.0);   // d2OM/dB2dC2
      setHessianS(OMind,3,5,0.0);   // d2OM/dB2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(OMind,3,6,0.0);   // d2OM/dB2dT

      // computation of d2OM/dC2dP                              --- checked
      setHessianS(OMind,4,0,0.0);   // d2OM/dC2dB
      setHessianS(OMind,4,1,0.0);   // d2OM/dC2dC
      setHessianS(OMind,4,2,0.0);   // d2OM/dC2dD
      setHessianS(OMind,4,3,0.0);   // d2OM/dC2dB2
      setHessianS(OMind,4,4,0.0);   // d2OM/dC2dC2
      setHessianS(OMind,4,5,0.0);   // d2OM/dC2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(OMind,4,6,0.0);   // d2OM/dC2dT

      // computation of d2OM/dD2dP                              --- checked
      setHessianS(OMind,5,0,0.0);   // d2OM/dD2dB
      setHessianS(OMind,5,1,0.0);   // d2OM/dD2dC
      setHessianS(OMind,5,2,0.0);   // d2OM/dD2dD
      setHessianS(OMind,5,3,0.0);   // d2OM/dD2dB2
      setHessianS(OMind,5,4,0.0);   // d2OM/dD2dC2
      setHessianS(OMind,5,5,0.0);   // d2OM/dD2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(OMind,5,6,0.0);   // d2OM/dD2dT

      if(tcStatus() != TIME_FIXED) 
	{
	  // computation of d2OM/dTdP                              --- checked
	  setHessianS(OMind,6,0,1.0);     // d2OM/dTdB
	  setHessianS(OMind,6,1,2.0*T);     // d2OM/dTdC
	  setHessianS(OMind,6,2,3.0*T*T);     // d2OM/dTdD
	  setHessianS(OMind,6,3,0.0);     // d2OM/dTdB2
	  setHessianS(OMind,6,4,0.0);   // d2OM/dTdC2
	  setHessianS(OMind,6,5,0.0); // d2OM/dTdD2
	  setHessianS(OMind,6,6,AJ(T));    // d2OM/dTdT
	}
    }

  if(thFixed_)
    {
      // computation of d2TH/dBdP                              --- checked
      setHessianS(THind,0,0,0.0);   // d2TH/dBdB
      setHessianS(THind,0,1,0.0);   // d2TH/dBdC
      setHessianS(THind,0,2,0.0);   // d2TH/dBdD
      setHessianS(THind,0,3,0.0);   // d2TH/dBdB2
      setHessianS(THind,0,4,0.0);   // d2TH/dBdC2
      setHessianS(THind,0,5,0.0);   // d2TH/dBdD2
      if(tcStatus() != TIME_FIXED) setHessianS(THind,0,6,T);   // d2TH/dBdT

      // computation of d2TH/dCdP                              --- checked
      setHessianS(THind,1,0,0.0);   // d2TH/dCdB
      setHessianS(THind,1,1,0.0);   // d2TH/dCdC
      setHessianS(THind,1,2,0.0);   // d2TH/dCdD
      setHessianS(THind,1,3,0.0);   // d2TH/dCdB2
      setHessianS(THind,1,4,0.0);   // d2TH/dCdC2
      setHessianS(THind,1,5,0.0);   // d2TH/dCdD2
      if(tcStatus() != TIME_FIXED) setHessianS(THind,1,6,T*T);   // d2TH/dCdT

      // computation of d2TH/dDdP                              --- checked
      setHessianS(THind,2,0,0.0);   // d2TH/dDdB
      setHessianS(THind,2,1,0.0);   // d2TH/dDdC
      setHessianS(THind,2,2,0.0);   // d2TH/dDdD
      setHessianS(THind,2,3,0.0);   // d2TH/dDdB2
      setHessianS(THind,2,4,0.0);   // d2TH/dDdC2
      setHessianS(THind,2,5,0.0);   // d2TH/dDdD2
      if(tcStatus() != TIME_FIXED) setHessianS(THind,2,6,T*T*T);   // d2TH/dDdT

      // computation of d2TH/dB2dP                              --- checked
      setHessianS(THind,3,0,0.0);   // d2TH/dB2dB
      setHessianS(THind,3,1,0.0);   // d2TH/dB2dC
      setHessianS(THind,3,2,0.0);   // d2TH/dB2dD
      setHessianS(THind,3,3,0.0);   // d2TH/dB2dB2
      setHessianS(THind,3,4,0.0);   // d2TH/dB2dC2
      setHessianS(THind,3,5,0.0);   // d2TH/dB2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(THind,3,6,0.0);   // d2TH/dB2dT

      // computation of d2TH/dC2dP                              --- checked
      setHessianS(THind,4,0,0.0);   // d2TH/dC2dB
      setHessianS(THind,4,1,0.0);   // d2TH/dC2dC
      setHessianS(THind,4,2,0.0);   // d2TH/dC2dD
      setHessianS(THind,4,3,0.0);   // d2TH/dC2dB2
      setHessianS(THind,4,4,0.0);   // d2TH/dC2dC2
      setHessianS(THind,4,5,0.0);   // d2TH/dC2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(THind,4,6,0.0);   // d2TH/dC2dT

      // computation of d2TH/dD2dP                              --- checked
      setHessianS(THind,5,0,0.0);   // d2TH/dD2dB
      setHessianS(THind,5,1,0.0);   // d2TH/dD2dC
      setHessianS(THind,5,2,0.0);   // d2TH/dD2dD
      setHessianS(THind,5,3,0.0);   // d2TH/dD2dB2
      setHessianS(THind,5,4,0.0);   // d2TH/dD2dC2
      setHessianS(THind,5,5,0.0);   // d2TH/dD2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(THind,5,6,0.0);   // d2TH/dD2dT

      if(tcStatus() != TIME_FIXED) 
	{
	  // computation of d2TH/dTdP                              --- checked
	  setHessianS(THind,6,0,T);     // d2TH/dTdB
	  setHessianS(THind,6,1,T*T);     // d2TH/dTdC
	  setHessianS(THind,6,2,T*T*T);     // d2TH/dTdD
	  setHessianS(THind,6,3,0.0);     // d2TH/dTdB2
	  setHessianS(THind,6,4,0.0);   // d2TH/dTdC2
	  setHessianS(THind,6,5,0.0); // d2TH/dTdD2
	  setHessianS(THind,6,6,AA(T));    // d2TH/dTdT
	}
    }

  if(xyFixed_)
    {
      // computation of d2X/dBdP                              --- checked
      setHessianS(Xind,0,0, qgaus(&CLTrajGen::itgd2X_dBdB,0,T)); // d2X/dBdB
      setHessianS(Xind,0,1, qgaus(&CLTrajGen::itgd2X_dBdC,0,T)); // d2X/dBdC
      setHessianS(Xind,0,2, qgaus(&CLTrajGen::itgd2X_dBdD,0,T)); // d2X/dBdD
      setHessianS(Xind,0,3, qgaus(&CLTrajGen::itgd2X_dBdB2,0,T)); // d2X/dBdB2
      setHessianS(Xind,0,4, qgaus(&CLTrajGen::itgd2X_dBdC2,0,T)); // d2X/dBdC2
      setHessianS(Xind,0,5, qgaus(&CLTrajGen::itgd2X_dBdD2,0,T)); // d2X/dBdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Xind,0,6, -0.5*T*T*V(T)*sin(TH(T))); // d2X/dBdT

      // computation of d2X/dCdP                              --- checked
      setHessianS(Xind,1,0, getHessianS(Xind,0,1)); // d2X/dCdB
      setHessianS(Xind,1,1, qgaus(&CLTrajGen::itgd2X_dCdC,0,T)); // d2X/dCdC
      setHessianS(Xind,1,2, qgaus(&CLTrajGen::itgd2X_dCdD,0,T)); // d2X/dCdD
      setHessianS(Xind,1,3, qgaus(&CLTrajGen::itgd2X_dCdB2,0,T)); // d2X/dCdB2
      setHessianS(Xind,1,4, qgaus(&CLTrajGen::itgd2X_dCdC2,0,T)); // d2X/dCdC2
      setHessianS(Xind,1,5, qgaus(&CLTrajGen::itgd2X_dCdD2,0,T)); // d2X/dCdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Xind,1,6, -0.333333333*T*T*T*V(T)*sin(TH(T))); // d2X/dCdT

      // computation of d2X/dDdP                              --- checked
      setHessianS(Xind,2,0, getHessianS(Xind,0,2)); // d2X/dDdB
      setHessianS(Xind,2,1, getHessianS(Xind,1,2)); // d2X/dDdC
      setHessianS(Xind,2,2, qgaus(&CLTrajGen::itgd2X_dDdD,0,T)); // d2X/dDdD
      setHessianS(Xind,2,3, qgaus(&CLTrajGen::itgd2X_dDdB2,0,T)); // d2X/dDdB2
      setHessianS(Xind,2,4, qgaus(&CLTrajGen::itgd2X_dDdC2,0,T)); // d2X/dDdC2
      setHessianS(Xind,2,5, qgaus(&CLTrajGen::itgd2X_dDdD2,0,T)); // d2X/dDdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Xind,2,6, -0.25*T*T*T*T*V(T)*sin(TH(T))); // d2X/dDdT

      // computation of d2X/dB2dP                              --- checked
      setHessianS(Xind,3,0, getHessianS(Xind,0,3)); // d2X/dB2dB
      setHessianS(Xind,3,1, getHessianS(Xind,1,3)); // d2X/dB2dC
      setHessianS(Xind,3,2, getHessianS(Xind,2,3)); // d2X/dB2dD
      setHessianS(Xind,3,3, 0.0); // d2X/dB2dB2
      setHessianS(Xind,3,4, 0.0); // d2X/dB2dC2
      setHessianS(Xind,3,5, 0.0); // d2X/dB2dD2
		  if(tcStatus() != TIME_FIXED) setHessianS(Xind,3,6, T*cos(TH(T))); // d2X/dB2dT

      // computation of d2X/dC2dP                              --- checked
      setHessianS(Xind,4,0, getHessianS(Xind,0,4)); // d2X/dC2dB
      setHessianS(Xind,4,1, getHessianS(Xind,1,4)); // d2X/dC2dC
      setHessianS(Xind,4,2, getHessianS(Xind,2,4)); // d2X/dC2dD
      setHessianS(Xind,4,3, getHessianS(Xind,3,4)); // d2X/dC2dB2
      setHessianS(Xind,4,4, 0.0); // d2X/dC2dC2
      setHessianS(Xind,4,5, 0.0); // d2X/dC2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(Xind,4,6, T*T*cos(TH(T))); // d2X/dC2dT

      // computation of d2X/dD2dP                              --- checked
      setHessianS(Xind,5,0, getHessianS(Xind,0,5)); // d2X/dD2dB
      setHessianS(Xind,5,1, getHessianS(Xind,1,5)); // d2X/dD2dC
      setHessianS(Xind,5,2, getHessianS(Xind,2,5)); // d2X/dD2dD
      setHessianS(Xind,5,3, getHessianS(Xind,3,5)); // d2X/dD2dB2
      setHessianS(Xind,5,4, getHessianS(Xind,4,5)); // d2X/dD2dC2
      setHessianS(Xind,5,5, 0.0); // d2X/dD2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(Xind,5,6, T*T*T*cos(TH(T))); // d2X/dD2dT

      if(tcStatus() != TIME_FIXED) 
	{
	  // computation of d2X/dTdP                              --- checked
	  setHessianS(Xind,6,0, getHessianS(Xind,0,6)); // d2X/dTdB
	  setHessianS(Xind,6,1, getHessianS(Xind,1,6)); // d2X/dTdC
	  setHessianS(Xind,6,2, getHessianS(Xind,2,6)); // d2X/dTdD
	  setHessianS(Xind,6,3, getHessianS(Xind,3,6)); // d2X/dTdB2
	  setHessianS(Xind,6,4, getHessianS(Xind,4,6)); // d2X/dTdC2
	  setHessianS(Xind,6,5, getHessianS(Xind,5,6)); // d2X/dTdD2
	  setHessianS(Xind,6,6, A(T)*cos(TH(T)) - V(T)*AV(T)*sin(TH(T))); // d2X/dTdT
	}

      // computation of d2Y/dBdP
      setHessianS(Yind,0,0, qgaus(&CLTrajGen::itgd2Y_dBdB,0,T)); // d2Y/dBdB
      setHessianS(Yind,0,1, qgaus(&CLTrajGen::itgd2Y_dBdC,0,T)); // d2Y/dBdC
      setHessianS(Yind,0,2, qgaus(&CLTrajGen::itgd2Y_dBdD,0,T)); // d2Y/dBdD
      setHessianS(Yind,0,3, qgaus(&CLTrajGen::itgd2Y_dBdB2,0,T)); // d2Y/dBdB2
      setHessianS(Yind,0,4, qgaus(&CLTrajGen::itgd2Y_dBdC2,0,T)); // d2Y/dBdC2
      setHessianS(Yind,0,5, qgaus(&CLTrajGen::itgd2Y_dBdD2,0,T)); // d2Y/dBdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Yind,0,6, 0.5*T*T*V(T)*cos(TH(T))); // d2Y/dBdT

      // computation of d2Y/dCdP
      setHessianS(Yind,1,0, getHessianS(Yind,0,1)); // d2Y/dCdB
      setHessianS(Yind,1,1, qgaus(&CLTrajGen::itgd2Y_dCdC,0,T)); // d2Y/dCdC
      setHessianS(Yind,1,2, qgaus(&CLTrajGen::itgd2Y_dCdD,0,T)); // d2Y/dCdD
      setHessianS(Yind,1,3, qgaus(&CLTrajGen::itgd2Y_dCdB2,0,T)); // d2Y/dCdB2
      setHessianS(Yind,1,4, qgaus(&CLTrajGen::itgd2Y_dCdC2,0,T)); // d2Y/dCdC2
      setHessianS(Yind,1,5, qgaus(&CLTrajGen::itgd2Y_dCdD2,0,T)); // d2Y/dCdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Yind,1,6, 0.333333333*T*T*T*V(T)*cos(TH(T))); // d2Y/dCdT

      // computation of d2Y/dDdP
      setHessianS(Yind,2,0, getHessianS(Yind,0,2)); // d2Y/dDdB
      setHessianS(Yind,2,1, getHessianS(Yind,1,2)); // d2Y/dDdC
      setHessianS(Yind,2,2, qgaus(&CLTrajGen::itgd2Y_dDdD,0,T)); // d2Y/dDdD
      setHessianS(Yind,2,3, qgaus(&CLTrajGen::itgd2Y_dDdB2,0,T)); // d2Y/dDdB2
      setHessianS(Yind,2,4, qgaus(&CLTrajGen::itgd2Y_dDdC2,0,T)); // d2Y/dDdC2
      setHessianS(Yind,2,5, qgaus(&CLTrajGen::itgd2Y_dDdD2,0,T)); // d2Y/dDdD2
      if(tcStatus() != TIME_FIXED) setHessianS(Yind,2,6, 0.25*T*T*T*T*V(T)*cos(TH(T))); // d2Y/dDdT

      // computation of d2Y/dB2dP
      setHessianS(Yind,3,0, getHessianS(Yind,0,3)); // d2Y/dB2dB
      setHessianS(Yind,3,1, getHessianS(Yind,1,3)); // d2Y/dB2dC
      setHessianS(Yind,3,2, getHessianS(Yind,2,3)); // d2Y/dB2dD
      setHessianS(Yind,3,3, 0.0); // d2Y/dB2dB2
      setHessianS(Yind,3,4, 0.0); // d2Y/dB2dC2
      setHessianS(Yind,3,5, 0.0); // d2Y/dB2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(Yind,3,6, T*sin(TH(T))); // d2Y/dB2dT

      // computation of d2Y/dC2dP
      setHessianS(Yind,4,0, getHessianS(Yind,0,4)); // d2Y/dC2dB
      setHessianS(Yind,4,1, getHessianS(Yind,1,4)); // d2Y/dC2dC
      setHessianS(Yind,4,2, getHessianS(Yind,2,4)); // d2Y/dC2dD
      setHessianS(Yind,4,3, getHessianS(Yind,3,4)); // d2Y/dC2dB2
      setHessianS(Yind,4,4, 0.0); // d2Y/dC2dC2
      setHessianS(Yind,4,5, 0.0); // d2Y/dC2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(Yind,4,6, T*T*sin(TH(T))); // d2Y/dC2dT

      // computation of d2Y/dD2dP
      setHessianS(Yind,5,0, getHessianS(Yind,0,5)); // d2Y/dD2dB
      setHessianS(Yind,5,1, getHessianS(Yind,1,5)); // d2Y/dD2dC
      setHessianS(Yind,5,2, getHessianS(Yind,2,5)); // d2Y/dD2dD
      setHessianS(Yind,5,3, getHessianS(Yind,3,5)); // d2Y/dD2dB2
      setHessianS(Yind,5,4, getHessianS(Yind,4,5)); // d2Y/dD2dC2
      setHessianS(Yind,5,5, 0.0); // d2Y/dD2dD2
      if(tcStatus() != TIME_FIXED) setHessianS(Yind,5,6, T*T*T*sin(TH(T))); // d2Y/dD2dT

      if(tcStatus() != TIME_FIXED) 
	{
	  // computation of d2Y/dTdP
	  setHessianS(Yind,6,0, getHessianS(Yind,0,6)); // d2Y/dTdB
	  setHessianS(Yind,6,1, getHessianS(Yind,1,6)); // d2Y/dTdC
	  setHessianS(Yind,6,2, getHessianS(Yind,2,6)); // d2Y/dTdD
	  setHessianS(Yind,6,3, getHessianS(Yind,3,6)); // d2Y/dTdB2
	  setHessianS(Yind,6,4, getHessianS(Yind,4,6)); // d2Y/dTdC2
	  setHessianS(Yind,6,5, getHessianS(Yind,5,6)); // d2Y/dTdD2
	  setHessianS(Yind,6,6, A(T)*sin(TH(T)) + V(T)*AV(T)*cos(TH(T))); // d2Y/dTdT
	}
    }
}

/*
 *  \brief set the system parameters
 *  \param constraints : constraints defined over the system
 */
void CLTrajGen::setSysParams(const System* constraints)
{
  if(constraints->getType() != CARLIKE_SYS)
    {
      throw string("CLTrajGen::setSysParams : Wrong system type!");
    }
  maxAcc = ((CarLikeSys*)constraints)->getAMax();
  maxSpd = ((CarLikeSys*)constraints)->getVMax();
  L = ((CarLikeSys*)constraints)->getLength();
  maxSt = ((CarLikeSys*)constraints)->getMaxSteering();
  maxStSpd = ((CarLikeSys*)constraints)->getMaxSteerSpd();
}

/*
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void CLTrajGen::setInitialGuess(const double* p)
{
  this->setInitialGuess(p[0],p[1],p[2],p[3],p[4],p[5],p[6]);
}

/*
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void CLTrajGen::setInitialConfig(const double* x0)
{
  // Set initial and final states
  vecState_x0 = Vec_DP(xLength);
  vecState_x0[0] = x0[0];
  vecState_x0[1] = x0[1];
  vecState_x0[2] = x0[2];
  vecState_x0[3] = x0[3];
  vecState_x0[4] = x0[4];
  vecState_x0[5] = x0[5];
  
  vecState_xf = vecState_x0;
}

/*
 *  \brief init a vector of the constraints defined over the system
 *  \param sys : the robotic system features
 *  \param vcstt : the resulting vector of constraints
 */
void CLTrajGen::initCstt(System* sys,Vec_DP& vcstt)
{
  vcstt[0] = 0;
  vcstt[1] = ((CarLikeSys*)sys)->getAMax();
  vcstt[2] = ((CarLikeSys*)sys)->getVMax();
  vcstt[3] = ((CarLikeSys*)sys)->getLength();
  vcstt[4] = ((CarLikeSys*)sys)->getMaxSteering();
  vcstt[5] = ((CarLikeSys*)sys)->getMaxSteerSpd();
}

/*
 *  \brief init the fixed state parameters
 *  \return the size of the vector of free parameters
 */
int CLTrajGen::initFixedSParam()
{
  int cpt = 0;

  if(xyFixed_)
    {
      Xind = 0;
      Yind = 1;
      cpt += 2;
    }
  if(thFixed_)
    {
      THind = cpt;
      cpt++;
    }
  if(phFixed_)
    {
      OMind = cpt;
      cpt++;
    }
  if(vFixed_)
    {
      Vind = cpt;
      cpt++;
    }

  return cpt;
}

// //try to generate a trajectory by constraints optimiation
// int CLTrajGen::ComputeTrajectory(Vec_DP x0,Vec_DP xf, Vec_DP constraints,Vec_DP** parameters)
// {
//   MUST_DISPLAY = false;

//   // initialize the correction coefficient
//   double coef_cor = COEF_COR;
//   int nbDeacreaseStep = 0;
//   double oldResidual = 0;

//   // Init the initial best parameters and residual
//   double bestB=0, bestC=0, bestD=0, bestB2=0, bestC2=0, bestD2=0;
//   double bestResidual = -1;
//   int bestCNT = -1;

//   //Set initial, final states and curvature constraint
//   setInitialParam(x0,xf,constraints);

//   if(DBG_ >= 1) {
//     cout << "trajGen : (" << vecState_x0[0] << "," << vecState_x0[1] << "," << vecState_x0[2] << "," 
// 	 << vecState_x0[3] << "," << vecState_x0[4] << "," << vecState_x0[5] << ") -> ";
//     cout << "(" << vecState_xf[0] << "," << vecState_xf[1] << "," << vecState_xf[2] << "," 
// 	 << vecState_xf[3] << "," << vecState_xf[4] << "," << vecState_xf[5] << ")" << endl;
//   }

//   //Heuristic guess
//   if(parameters == NULL || *parameters == NULL)
//     {
//       //cout << "param nuls" << endl;
//       InitialGuess(parameters);
//     }
//   else
//     {
//       //cout << "param non nuls  (" << (**parameters)[0] << "," << (**parameters)[1] << "," << (**parameters)[2] << "," << 
//       //(**parameters)[3] << "," << (**parameters)[4] << "," << (**parameters)[5] << "," << (**parameters)[6] << ")" << endl;
//       vecParameters_p = *parameters;
//       S = vecState_xf[5] - vecState_x0[5];
//     }

//   if(DBG_ >= 1) {
//     cout << "Initial Guess :  ";
//     cout << "B:" << B << "  C:" << C << "  D:" << D << "  S:" << S << "  B2:" << B2 << "  C2:" << C2 << "  D2:" << D2 << "  T:" << S << endl;
//   }

//   double residual = 1000000;  
//   int cnt=0;
//   CNT_ = 0;
  
//   vecCorrParameters_p = 0;  

//   while(residual > maxResidual_)
//     {
//       //cout << "CNT : " << CNT << endl;

//       // save the best parameters
//       if(CNT_ == 0 || residual < bestResidual)
// 	{
// 	  bestB = B;
// 	  bestC = C;
// 	  bestD = D;
// 	  bestB2 = B2;
// 	  bestC2 = C2;
// 	  bestD2 = D2;
// 	  bestResidual = residual;
// 	  bestCNT = CNT_;
// 	}

// #if VARIABLE_COEF_COR
//       // update the correction coefficient
//       if(residual <= oldResidual)
// 	{
// 	  nbDeacreaseStep++;
	  
// 	  if(nbDeacreaseStep >= 10) coef_cor = 1.0;
// 	  else if(nbDeacreaseStep >= 9) coef_cor = 0.8;
// 	  else if(nbDeacreaseStep >= 8) coef_cor = 0.7;
// 	  else if(nbDeacreaseStep >= 7) coef_cor = 0.6;
// 	  else if(nbDeacreaseStep >= 6) coef_cor = 0.5;
// 	  else if(nbDeacreaseStep >= 5) coef_cor = 0.4;
// 	  else if(nbDeacreaseStep >= 3) coef_cor = 0.3;	
// 	}
//       else
// 	{
// 	  coef_cor = COEF_COR;
// 	  nbDeacreaseStep = 0;
// 	}
//       oldResidual = residual;
// #endif

//       //Apply correction
//       int pl = 0;
//       if (tcStatus_ == TIME_FIXED) pl = pLength-1;
//       else pl = pLength;
      
//       for(int i=0; i<pl;i++)
// 	(*vecParameters_p)[i] += coef_cor*vecCorrParameters_p[i];            
      
//       //MUST_DISPLAY = false;
//       ForwardSolution();
//       //MUST_DISPLAY = false;
//       CNT_++;

//       if(isnan(vecState_xsim[0]) || isnan(vecState_xsim[1]) || isnan(vecState_xsim[2]) || isnan(vecState_xsim[3]) || isnan(vecState_xsim[4]))
// 	{
// 	  finalResidual_ = 99999999;
// 	  finalNbSteps_ = CNT_;
// 	  return -2;

// 	  cout << "JacobianG : " << endl;
// 	  for(int i2=0; i2<gLength; i2++)
// 	    {
// 	      for(int j2=0; j2<pLength-1; j2++)
// 		{
// 		  cout << matJacobianG[i2][j2] << " ";
// 		}
// 	      cout << endl;
// 	    }
// 	  cout << endl;

// 	  cout << "COR_B:" << (*vecParameters_p)[0] << "  COR_C:" << (*vecParameters_p)[1] << "  COR_D:" << (*vecParameters_p)[2] << "  COR_B2:" << (*vecParameters_p)[3] << "  COR_C2:" << (*vecParameters_p)[4] << "  COR_D2:" << (*vecParameters_p)[5] << endl;
// 	  cout << "B:" << B << "  C:" << C << "  D:" << D << "  S:" << S << "  B2:" << B2 << "  C2:" << C2 << "  D2:" << D2 << endl;
// 	}

//       if(DBG_ >= 1) {
// 	printf ("Simulated state = (%f,%f,%f,%f,%f)\n", 
// 		vecState_xsim[0],vecState_xsim[1],vecState_xsim[2],vecState_xsim[3],vecState_xsim[4]);    
//       }
//       ComputeG();
//       if(DBG_ >= 1) {
// 	printf ("Errors = (%f,%f,%f,%f,%f)\n", 
// 		vecConstraints_g[0],vecConstraints_g[1],vecConstraints_g[2],
// 		vecConstraints_g[3],vecConstraints_g[4]);
// 	printf ("Parameters K = (a=%f,b=%f,c=%f,d=%f,s=%f)\n", 
// 		vecState_x0[3],B,C,D,S);
// 	printf ("Parameters V = (a=%f,b=%f,c=%f,d=%f,s=%f)\n", 
// 		vecState_x0[4],B2,C2,D2,S);
//       }

//       ComputeJacobianG();
//       ComputeInvJacobianG();

//       //this->displayJacobian();
      
//       //Compute the correction
//       for(int i=0; i<pl;i++)
// 	{
// 	  vecCorrParameters_p[i] = 0;
	  
// 	  for(int j=0;j<gLength;j++)
// 	    {
// 	      vecCorrParameters_p[i] += (-1.0)*vecConstraints_g[j]*matInvJacobianG[i][j]; 
// 	    }	    
// 	}

//       if(DBG_ >= 1) {
// 	printf ("Correction = (b=%f,c=%f,d=%f,b2=%f,c2=%f,d2=%f)\n", 
// 		vecCorrParameters_p[0],vecCorrParameters_p[1],vecCorrParameters_p[2],
// 		vecCorrParameters_p[3],vecCorrParameters_p[4],vecCorrParameters_p[5]);
//       }
      
//       residual = sqrt( pow(vecConstraints_g[0],2) + 
// 		       pow(vecConstraints_g[1],2) +
// 		       pow(vecConstraints_g[2],2) +
// 		       pow(vecConstraints_g[3],2) +
// 		       pow(vecConstraints_g[4],2));

//       if(DBG_ >= 1) {        
// 	printf ("Residual = %f\t CNT : %d\t COEF_COR : %f\n", 
// 		residual,CNT_,coef_cor); 
	
// 	cout << "B:" << B << "  C:" << C << "  D:" << D << "  S:" << S << "  B2:" << B2 << "  C2:" << C2 << "  D2:" << D2 << "  T:" << S << endl;
//       }

//       if(DBG_ >= 2) {
// 	displayProfiles();
// 	if(DBG_ >= 3) {
// 	  displayProfilesWB();
// 	}
//       }

//       cnt++;
//       if(cnt >= maxNbStep_)
// 	{
// 	  if(DBG_ >= 1) {
// 	    cout << "  No traj found in " << maxNbStep_ << "iterations" << endl;
// 	  }

// 	  // get back the best parameters found
// 	  B = bestB;
// 	  C = bestC;
// 	  D = bestD;
// 	  B2 = bestB2;
// 	  C2 = bestC2;
// 	  D2 = bestD2;

// 	  // compute the forward method with these parameters
// 	  ForwardSolution();
// 	  // compute the associate error
// 	  ComputeG();

// 	  // get back errors on X, Y and TH
// 	  finalResidual_ = bestResidual;
// 	  finalNbSteps_ = CNT_;
// 	  finalErrX_ = vecConstraints_g[0];
// 	  finalErrY_ = vecConstraints_g[1];
// 	  finalErrTH_ = vecConstraints_g[2];

// 	  if(DBG_ >= 1) {
// 	    cout << "Best Solution -> B:" << B << "  C:" << C << "  D:" << D << "  S:" << S << "  B2:" << B2 << "  C2:" << C2 << "  D2:" << D2 << "  T:" << S << endl;
// 	    cout << "Best Residual : " << residual << "   Best CNT : " << bestCNT << endl;
// 	  }
// 	  return -1;
// 	}
//     }
//   if(DBG_ >= 1) {
//     cout << "  Nb iterations : " << CNT_ << "  ";
//   }
//   finalResidual_ = residual;
//   finalNbSteps_ = CNT_;
//   finalErrX_ = vecConstraints_g[0];
//   finalErrY_ = vecConstraints_g[1];
//   finalErrTH_ = vecConstraints_g[2];
//   return 0;
// }

/*
 *  \brief generate a trajectory by constraints optimiation
 *  \param s0 : initial state
 *  \param sf : final state
 *  \param constraints : constraints defined over the system
 *  \param parameters : initial guess of parameters
 */
int CLTrajGen::ComputeTrajectory(const CLStateT& s0,const CLStateT& sf,double** parameters)
{
  Vec_DP x0(6);
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.s()*tan(s0.ph())/sys()->getLength();
  x0[4] = s0.s();
  x0[5] = 0;
  
  Vec_DP xf(6);
  xf[0] = sf.x();
  xf[1] = sf.y();
  xf[2] = sf.th();
  xf[3] = sf.s()*tan(sf.ph())/sys()->getLength();
  xf[4] = sf.s();
  xf[5] = sf.t() - s0.t();

  Vec_DP** param = NULL;
  if(parameters == NULL) param = NULL;
  else
    {
      param = new Vec_DP*;
      (*param) = new Vec_DP(7);
      (**param)[0] = (*parameters)[0];
      (**param)[1] = (*parameters)[1];
      (**param)[2] = (*parameters)[2];
      (**param)[3] = (*parameters)[3];
      (**param)[4] = (*parameters)[4];
      (**param)[5] = (*parameters)[5];
      (**param)[6] = (*parameters)[6];
    }

  int res = this->TrajGen::ComputeTrajectory(x0,xf,param);

  if(parameters == NULL)
    {
      parameters = new double*;
      *parameters = new double[7];
    }

//   if(param == NULL) cout << "param NULL" << endl;
//   else if(*param == NULL) cout << "*param NULL" << endl;
//   else cout << "param not NULL" << endl;
  
//   cout << "11?" << endl;
//   (*parameters)[0] = (**param)[0];
//   (*parameters)[1] = (**param)[1];
//   (*parameters)[2] = (**param)[2];
//   (*parameters)[3] = (**param)[3];
//   (*parameters)[4] = (**param)[4];
//   (*parameters)[5] = (**param)[5];
//   (*parameters)[6] = (**param)[6];

//   cout << "2?" << endl;

//   if(param != NULL) 
//     {
//       if(*param != NULL) delete (*param);
//       delete param;
//     }

  return res;
}

/*
 *  \brief generate a trajectory by constraints optimiation
 *  \param s0 : initial state
 *  \param sf : final state
 *  \param constraints : constraints defined over the system
 *  \param parameters : initial guess of parameters
 *  \return 0 if a solution has been found, -1 else
 */
int CLTrajGen::ComputeTrajectory(const CLStateT& s0,const CLStateT& sf,Vec_DP** parameters)
{
  Vec_DP x0(6);
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.s()*tan(s0.ph())/sys()->getLength();
  x0[4] = s0.s();
  x0[5] = 0;
  
  Vec_DP xf(6);
  xf[0] = sf.x();
  xf[1] = sf.y();
  xf[2] = sf.th();
  xf[3] = sf.s()*tan(sf.ph())/sys()->getLength();
  xf[4] = sf.s();
  xf[5] = sf.t() - s0.t();

  int res = this->TrajGen::ComputeTrajectory(x0,xf,parameters);
  return res;
}

/*
 *  \brief generate a trajectory by constraints optimiation
 *  \param s0 : initial state
 *  \param sys : constraints defined over the system
 *  \param param : initial guess of parameters
 */
void CLTrajGen::ComputeFromIG(const CLStateT& s0,const double* param)
{
  double x0[6];
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.s()*tan(s0.ph())/sys()->getLength();
  x0[4] = s0.s();
  x0[5] = 0;

  this->TrajGen::ComputeFromIG(x0,param);
}

/*
 *  \brief generate a trajectory by constraints optimiation
 *  \param statet : initial state-time
 *  \param sys : constraints defined over the robotic system
 *  \param ctrlt : control-time applied on the system
 */
void CLTrajGen::ComputeFromCtrl(const StateTime& statet,const ControlTime& ctrlt)
{
//   double maxA = sys()->getAMax();
  double maxV = sys()->getVMax();
  double maxPh = sys()->getMaxSteering();
//   double maxMu = sys()->getMaxSteerSpd();

  CLStateT clST(*((CLStateT*)&statet));
  double x0[6];
  x0[0] = clST.x();
  x0[1] = clST.y();
  x0[2] = clST.th();
  x0[3] = clST.s()*tan(clST.ph())/(sys()->getLength());
  x0[4] = clST.s();
  x0[5] = 0;
  
  CLControlT clCT(*((CLControlT*)(&ctrlt)));
  
  // compute the angular velocity corresponding to the input steering speed
  double t       = clCT.t();
  double L       = this->sys()->getLength();
  double acc     = clCT.acc();
  double sts     = clCT.ss();
  double velI    = clST.s();
  double staI    = clST.ph();
  double tanStaI = tan(staI);
  double velF    = clST.s() + t*acc;
  if(velF < -maxV) velF = -maxV;
  if(velF > maxV) velF = maxV;
  double staF    = clST.ph() + t*sts;
  if(staF < -maxPh) staF = -maxPh;
  if(staF > maxPh) staF = maxPh;
  double tanStaF = tan(staF);
  
  double aaccI = (1.0 / L) * (acc*tanStaI + velI*sts*(1.0 + tanStaI*tanStaI));
  double aaccF = (1.0 / L) * (acc*tanStaF + velF*sts*(1.0 + tanStaF*tanStaF));
    
  double param[7];
  param[0] = aaccI;
  if(t != 0) param[1] = (aaccF-aaccI)/t;
  else param[1] = 0;
  param[2] = 0.0;
  param[3] = clCT.acc();
  param[4] = 0.0;
  param[5] = 0.0;
  param[6] = clCT.t();

  cout << "ctrlt : "; ctrlt.display(); cout << "   clCT : "; clCT.display(); cout << endl;
  cout << "param : (" << param[0] << "," << param[1] << "," << param[2] << "," << param[3] << "," << param[4] << "," << param[5] << "," << param[6] << ")" << endl;
  
  this->TrajGen::ComputeFromIG(x0,param);
}


/****** Trajectory Generation (2nd version) ******/

/**
 *  \brief init the cot function weights
 */
void CLTrajGen::initWeights()
{ 
  if(xyFixed_)
    {
      weightsCF_[Xind] = CLweights_.x();
      weightsCF_[Yind] = CLweights_.y();
    }
  if(thFixed_)
    {
      weightsCF_[THind] = CLweights_.th();
    }
  if(phFixed_)
    {
      weightsCF_[OMind] = CLweights_.ph();
    }
  if(vFixed_)
    {
      weightsCF_[Vind] = CLweights_.s();
    }
}

/**
 *  \brief generate a trajectory between two states
 *  \param s0 : initial state-time
 *  \param sf : initial state-time
 *  \param p : initial guess of parameters
 */
int CLTrajGen::genTraj(const CLStateT& s0,const CLStateT& sf,const double* p)
{
  regX.clear();
  regY.clear();
  regTH.clear();
  regV.clear();
  regAV.clear();
  regT.clear();  

  double x0[6];
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.s()*tan(s0.ph())/sys()->getLength();
  x0[4] = s0.s();
  x0[5] = 0;
  
  double xf[6];
  xf[0] = sf.x();
  xf[1] = sf.y();
  xf[2] = sf.th();
  xf[3] = sf.s()*tan(sf.ph())/sys()->getLength();
  xf[4] = sf.s();
  xf[5] = sf.t() - s0.t();

  return this->TrajGen::genTraj(x0,xf,p);
}


/**************** Output Functions ****************/

/*
 *  \brief get back the state reached at time t
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,ph,v,t)
 */
void CLTrajGen::getStateTReachedAt(double* rstatet,double t) const
{
  rstatet[0] = funcX(t);
  rstatet[1] = funcY(t);
  rstatet[2] = funcTheta(t);
  rstatet[3] = funcSteer(t); //atan(L * funcAVel(t));
  rstatet[4] = funcVel(t);
  rstatet[5] = t;
}

/*
 *  \brief get back the state reached at time t
 *  \param t : time at which we evaluate the reached state
 *  \result the reached state
 */
CLStateT CLTrajGen::getCLStateTReachedAt(double t) const
{
  return CLStateT(funcX(t),
		  funcY(t),
		  funcTheta(t),
		  funcSteer(t),
		  funcVel(t),
		  t);
}

/*
 *  \brief get back the state reached at time t
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,ph,v,t)
 */
void CLTrajGen::getFinalStateTReached(double* rstatet) const
{
  rstatet[0] = funcX(T);
  rstatet[1] = funcY(T);
  rstatet[2] = funcTheta(T);
  rstatet[3] = funcSteer(T); //atan(L * funcAVel(S));
  rstatet[4] = funcVel(T);
  rstatet[5] = T;
}

/*
 *  \brief get back the state reached at time t
 *  \param t : time at which we evaluate the reached state
 *  \result the reached state
 */
CLStateT CLTrajGen::getFinalCLStateTReached() const
{
  return CLStateT(funcX(T),
		  funcY(T),
		  funcTheta(T),
		  funcSteer(T),
		  funcVel(T),
		  T);
}

/*
 *  \brief return a sample trajectory from the parameterized one
 *  \param ts : the sampling time step
 *  \return a sample trajectory from the parameterized one
 *
 *  A new instance of sample trajectory is returned here
 */
CLSTraj* CLTrajGen::sampledTraj(double ts) const
{
  // compute the number of intervals required
  int nbItv = ((int)(T/ts));
  if(nbItv == 0) nbItv = 1;

  // compute the true time step used
  double trueTS = T/(nbItv);

  // initialize the output trajectory
  CLSTraj* traj = new CLSTraj();

  // compute each state of the trajectory
  for(int i=0; i<=nbItv; i++)
    {
      double t = i*trueTS;
      CLStateT curST(funcX(t),
		     funcY(t),
		     funcTheta(t),
		     funcSteer(t),
		     funcVel(t),
		     t);
      traj->append(curST);
    }
  
  // return the trajectory computed
  return traj;
}


/*************** Utility Functions ****************/

/**
 *  \brief convert the initial and final input state-times into tables of double
 *  \param initST : the input initial state-time
 *  \param goalST : the input goal state-time
 *  \param initTD : the resulting initial table
 *  \param goalTD : the resulting goal table
 *
 *  Tables must be allocated first
 */
void CLTrajGen::InitGoalST2Tab(const StateTime& initST,const StateTime& goalST,double* initTD,double* goalTD) const
{
  CLStateT s0(*((CLStateT*)(&initST)));
  CLStateT sf(*((CLStateT*)(&goalST)));

  initTD[0] = s0.x();
  initTD[1] = s0.y();
  initTD[2] = s0.th();
  initTD[3] = s0.s()*tan(s0.ph())/sys()->getLength();
  initTD[4] = s0.s();
  initTD[5] = 0;
  
  goalTD[0] = sf.x();
  goalTD[1] = sf.y();
  goalTD[2] = sf.th();
  goalTD[3] = sf.s()*tan(sf.ph())/sys()->getLength();
  goalTD[4] = sf.s();
  goalTD[5] = sf.t() - s0.t();
}

void CLTrajGen::SaveTrajectory(int nbp)
{
//   if(posX != NULL) delete posX;
//   if(posY != NULL) delete posY;
//   if(timeT != NULL) delete timeT;
//   if(velV != NULL) delete velV;
//   if(steerS != NULL) delete steerS;  

  posX = new vector<double>;
  posY = new vector<double>;
  timeT = new vector<double>;
  velV = new vector<double>;
  steerS = new vector<double>;

  posX->clear();
  posY->clear();
  timeT->clear();
  velV->clear();
  steerS->clear();

//   // check if the time is well defined
//   if(S < 0)
//     {
//       cout << "41" << endl;
//       cerr << "CLTrajGen::SaveTrajectory : T < 0" << endl;
//       return;
//     }
//   else cout << "42" << endl;

  //Computing trajectory
  double step = 1.0 / (((double)nbp)-1.0);
  for(int j=0;j<nbp;j++)
    {
      if(fctAcc_ != NULL && fctVel_ != NULL && fctAVel_ != NULL && fctAAcc_ != NULL && fctTH_ != NULL)
	{
	  posX->push_back(funcX(j*(step*T)));
	  posY->push_back(funcY(j*(step*T)));
	  timeT->push_back(j*step*T);
	  velV->push_back(funcVel(j*step*T));
	  steerS->push_back(funcSteer(j*step*T));
	}
      else
	{
	  posX->push_back(0.0);
	  posY->push_back(0.0);
	  timeT->push_back(0.0);
	  velV->push_back(0.0);
	  steerS->push_back(0.0);
	}
    }

  regX.push_back(posX);
  regY.push_back(posY);
  regV.push_back(velV);
  regT.push_back(timeT);
}

// trace the resulting trajectory on a map
void CLTrajGen::traceOn(Map *map,int nbp)
{
  if(map == NULL)
    {
      cout << "CLTrajGen:traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }
  else
    {
      int pathCol = (int) (map->getDepth() / 2);
      for(int i=0; i<=nbp; i++)
	{
	  double x = funcX(i*(1.0/((double)nbp)*T));
	  double y = funcY(i*(1.0/((double)nbp)*T));

// 	  cout << "(" << x << "," << y ") ";

	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
	}
    }
}

/*
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 *  \param initPos : initial position
 */  
void CLTrajGen::traceOn(Map *map,int nbp,const PositionXYTh& initPos)
{
  if(map == NULL)
    {
      cout << "DDTrajGen::traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }
  else
    {
      int pathCol = (int) (map->getDepth() / 2);
      for(int i=0; i<=nbp; i++)
	{
	  double x1 = funcX(i*(1.0/((double)nbp)*T));
	  double y1 = funcY(i*(1.0/((double)nbp)*T));
	  double th1 = funcTheta(i*(1.0/((double)nbp)*T));
	  PositionXYTh pos1(x1,y1,th1);

	  PositionXYTh pos = coordInFrame(pos1,initPos);
	  double x = pos.x();
	  double y = pos.y();

	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
	}
    }
}

// display linear acceleration and velocity, angular velocity, 
// steering angle, heading and position profiles
void CLTrajGen::displayProfiles()
{
//   // compute the minimal curvature circles

//   // compute the radius of the minimal curvature circles
//   double L = ((CarLikeSys*)sys_)->getLength();
//   double phiMax = ((CarLikeSys*)sys_)->getMaxSteering();
//   double rad = L / tan(phiMax);

//   // compute the centers of these circles
//   PositionXY leftCenter(X0 - rad*sin(TH0),
// 			Y0 + rad*cos(TH0));
//   PositionXY rightCenter(X0 + rad*sin(TH0),
// 			 Y0 - rad*cos(TH0));

  vector<double> xIGpos, yIGpos;
//   xIGpos.push_back(leftCenter.x());
//   xIGpos.push_back(rightCenter.x());
//   yIGpos.push_back(leftCenter.y());
//   yIGpos.push_back(rightCenter.y());

//   // compute both circles
//   vector<double> LC_X;
//   vector<double> LC_Y;
//   for(int i=0; i<=100; i++)
//     {
//       double X = leftCenter.x() + rad*cos(2.0*M_PI*0.01*i);
//       double Y = leftCenter.y() + rad*sin(2.0*M_PI*0.01*i);
//       LC_X.push_back(X);
//       LC_Y.push_back(Y);
//     }

//   vector<double> RC_X;
//   vector<double> RC_Y;
//   for(int i=0; i<=100; i++)
//     {
//       double X = rightCenter.x() + rad*cos(2.0*M_PI*0.01*i);
//       double Y = rightCenter.y() + rad*sin(2.0*M_PI*0.01*i);
//       RC_X.push_back(X);
//       RC_Y.push_back(Y);
//     }

  // get back the initial and final positions
  xIGpos.push_back(X0);
  yIGpos.push_back(Y0);
  xIGpos.push_back(Xg);
  yIGpos.push_back(Yg);

  Gnuplot gPosition;
  gPosition.reset_plot();
  gPosition.set_style("points").plot_xy(xIGpos,yIGpos,"Initial and Final Pos");
//   gPosition.set_style("lines").plot_xy(LC_X,LC_Y,"");
//   gPosition.set_style("lines").plot_xy(RC_X,RC_Y,"");

  for(int i=0; i<regX.size(); i++)
    {
      vector<double>* curRegX = regX.at(i);
      vector<double>* curRegY = regY.at(i);

      gPosition.set_style("lines").plot_xy(*curRegX,*curRegY,"");
    }

//   vector<vector<double>* > regV;
//   vector<vector<double>* > regT;

  //Computing Velocity
  Gnuplot gVel;
  gVel.reset_plot();

  vector<double> realMinboundVel;
  vector<double> realMaxboundVel;

  if(regT.size() != 0)
    {
      for(int j=0;j<regT.at(0)->size();j++){
	if(sys()->canMoveBack()) realMinboundVel.push_back(-maxSpd);
	else realMinboundVel.push_back(0.0);
	realMaxboundVel.push_back(maxSpd);
      }

      gVel.set_style("points").plot_xy(*(regT.at(0)),realMinboundVel,"Min bound");
      gVel.set_style("points").plot_xy(*(regT.at(0)),realMaxboundVel,"Max bound");
    }

  for(int i=0; i<regV.size(); i++)
    {
      vector<double>* curRegV = regV.at(i);
      vector<double>* curRegT = regT.at(i);

      gVel.set_style("lines").plot_xy(*curRegT,*curRegV,"");
    }

  wait_for_key();
}

// display linear acceleration and velocity, angular velocity, 
// steering angle, heading and position profiles
void CLTrajGen::displayProfiles2()
{
//   cout << "functions (avant) : " << endl;
//   cout << "Parameters :  "; this->displayParameters(); cout << endl;
//   this->displayAllFct();

  // compute the forward solution with the new parameters
  this->ForwardSolution();

//   cout << "functions (apres) : " << endl;
//   cout << "Parameters :  "; this->displayParameters(); cout << endl;
//   this->displayAllFct();
  
  //Computing Position
  this->SaveTrajectory(100);

  vector<double> vecX;
  vector<double> vecY;
  for(int j=0; j<50; j++)
    {
      vecX.push_back(this->funcX(j*(T/50)));
      vecY.push_back(this->funcY(j*(T/50)));
    }

  // compute the minimal curvature circles

  // compute the radius of the minimal curvature circles
  double L = ((CarLikeSys*)sys_)->getLength();
  double phiMax = ((CarLikeSys*)sys_)->getMaxSteering();
  double rad = L / tan(phiMax);

  // compute the centers of these circles
  PositionXY leftCenter(X0 - rad*sin(TH0),
			Y0 + rad*cos(TH0));
  PositionXY rightCenter(X0 + rad*sin(TH0),
			 Y0 - rad*cos(TH0));

  vector<double> xIGpos, yIGpos;
  xIGpos.push_back(leftCenter.x());
  xIGpos.push_back(rightCenter.x());
  yIGpos.push_back(leftCenter.y());
  yIGpos.push_back(rightCenter.y());

  // compute both circles
  vector<double> LC_X;
  vector<double> LC_Y;
  for(int i=0; i<=100; i++)
    {
      double X = leftCenter.x() + rad*cos(2.0*M_PI*0.01*i);
      double Y = leftCenter.y() + rad*sin(2.0*M_PI*0.01*i);
      LC_X.push_back(X);
      LC_Y.push_back(Y);
    }

  vector<double> RC_X;
  vector<double> RC_Y;
  for(int i=0; i<=100; i++)
    {
      double X = rightCenter.x() + rad*cos(2.0*M_PI*0.01*i);
      double Y = rightCenter.y() + rad*sin(2.0*M_PI*0.01*i);
      RC_X.push_back(X);
      RC_Y.push_back(Y);
    }

  // get back the initial and final positions
  xIGpos.push_back(X0);
  yIGpos.push_back(Y0);
  xIGpos.push_back(Xg);
  yIGpos.push_back(Yg);

//   Gnuplot& set_xrange(const int iFrom,
// 		      const int iTo);
//   Gnuplot& set_yrange(const int iFrom,
// 		      const int iTo);

  double minX = min(xIGpos[0],xIGpos[1]);
  double minY = min(yIGpos[0],yIGpos[1]);
  double maxX = max(xIGpos[0],xIGpos[1]);
  double maxY = max(yIGpos[0],yIGpos[1]);

//   cout << "xIGpos[0] : " << xIGpos[0] << endl;
//   cout << "xIGpos[1] : " << xIGpos[1] << endl;
//   cout << "yIGpos[0] : " << yIGpos[0] << endl;
//   cout << "yIGpos[1] : " << yIGpos[1] << endl;

//   cout << "minX : " << minX << endl;
//   cout << "maxX : " << maxX << endl;
//   cout << "minY : " << minY << endl;
//   cout << "maxY : " << maxY << endl;

  double maxDist = max(fabs(xIGpos[0] - xIGpos[1]),fabs(yIGpos[0] - yIGpos[1]));
  if(maxDist == 0) maxDist = 6.0*rad;
  
  Gnuplot gPosition;
  gPosition.reset_plot();
  gPosition.set_style("lines")
    .set_xrange((int)(minX - 0.6*maxDist),(int)(maxX + 0.6*maxDist))
    .set_yrange((int)(minY - 0.6*maxDist),(int)(maxY + 0.6*maxDist))
    .plot_xy(vecX,vecY,"Position");
  gPosition.set_style("points").plot_xy(xIGpos,yIGpos,"Initial and Final Pos");
  gPosition.set_style("lines").plot_xy(LC_X,LC_Y,"");
  gPosition.set_style("lines").plot_xy(RC_X,RC_Y,"");
  
  //Computing Velocity
  vector<double> vel;
  vector<double> realMinboundVel;
  vector<double> realMaxboundVel;
  vector<double> distanceV;
  for(int j=0;j<=50;j++){
    vel.push_back((double)this->funcVel(j*(T/50)));
    if(sys()->canMoveBack()) realMinboundVel.push_back(-maxSpd);
    else realMinboundVel.push_back(0.0);
    realMaxboundVel.push_back(maxSpd);
    distanceV.push_back((j*(T/50)));
  }    
  Gnuplot gVel;
  gVel.reset_plot();
  gVel.set_style("lines").plot_xy(distanceV,vel,"Velocity");	
  gVel.set_style("points").plot_xy(distanceV,realMinboundVel,"Min bound");
  gVel.set_style("points").plot_xy(distanceV,realMaxboundVel,"Max bound");
  
  //Computing Acceleration
  vector<double> acc;
  vector<double> realMinboundAcc;
  vector<double> realMaxboundAcc;
  vector<double> distanceA;
  for(int j=0;j<=50;j++){
    acc.push_back((double)this->funcAcc(j*(T/50)));
    realMinboundAcc.push_back(-maxAcc);
    realMaxboundAcc.push_back(maxAcc);
    distanceA.push_back((j*(T/50)));
  }    
  Gnuplot gAcc;
  gAcc.reset_plot();
  gAcc.set_style("lines").plot_xy(distanceA,acc,"Acceleration");
  gAcc.set_style("points").plot_xy(distanceA,realMinboundAcc,"Real min bound");
  gAcc.set_style("points").plot_xy(distanceA,realMaxboundAcc,"Real max bound");
  
//   //Computing Angular Acceleration
//   double tanMaxSt = tan(maxSt);
//   double coef1 = tanMaxSt / L;
//   double coef2 = (tanMaxSt*tanMaxSt + 1.0) * maxStSpd / L;
  
//   MultiItvFunction aaccBound1(*fctAcc_);
//   aaccBound1 *= coef1;
//   MultiItvFunction aaccBound2(*fctVel_);
//   aaccBound2 *= coef2;
//   aaccBound1 += aaccBound2;
//   MultiItvFunction aaccMinBound(aaccBound1);
//   aaccMinBound *= (-1.0);
  
//   cout << "fctAAcc_ : " << endl;
//   fctAAcc_->display();
//   cout << "fctAVel_ : " << endl;
//   fctAVel_->display();

//   vector<double> aacc;
//   vector<double> minboundAAcc;
//   vector<double> maxboundAAcc;
//   vector<double> distanceAA;
//   for(int j=0;j<=50;j++){
//     aacc.push_back((double)this->fctAAcc_->getVal(((double)j)*(S/50.0)));
//     minboundAAcc.push_back(aaccMinBound.getVal(((double)j)*(S/50.0)));
//     maxboundAAcc.push_back(aaccBound1.getVal(((double)j)*(S/50.0)));
//     distanceAA.push_back((j*(S/50)));
//   }    
//   Gnuplot gAAcc;
//   gAAcc.reset_plot();
//   gAAcc.set_style("lines").plot_xy(distanceAA,aacc,"Angular Acceleration");
//   gAAcc.set_style("points").plot_xy(distanceAA,minboundAAcc,"Min bound");
//   gAAcc.set_style("points").plot_xy(distanceAA,maxboundAAcc,"Max bound");
  
  //Computing Angular Velocity
  double coef = fabs(tan(maxSt) / L);
//   cout << "maxSt : " << maxSt << endl;
//   cout << "tan(maxSt) : " << tan(maxSt) << endl;
//   cout << "L : " << L << endl;
//   cout << "coef : " << coef << endl;

  computeFctVel();
  MultiItvFunction absVel(*fctVel_);
  absVel.addAbsoluteValue();

  MultiItvFunction minBoundAVel(absVel);
  minBoundAVel *= (-coef);
  MultiItvFunction maxBoundAVel(absVel);
  maxBoundAVel *= coef;
  
  vector<double> avel;
  vector<double> avelwb;
  vector<double> minboundAVel;
  vector<double> maxboundAVel;
  vector<double> distanceAV;
  for(int j=0;j<=50;j++){
    double val = this->funcAVel(j*(T/50));
    avel.push_back(val);
    avelwb.push_back(fctAVelWithoutBounds_->getVal(j*(T/50)));
    minboundAVel.push_back(minBoundAVel.getVal(j*(T/50)));
    maxboundAVel.push_back(maxBoundAVel.getVal(j*(T/50)));
    distanceAV.push_back((j*(T/50)));
  }    
  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(distanceAV,avelwb,"Angular Velocity WB");
  gAVel.set_style("lines").plot_xy(distanceAV,avel,"Angular Velocity");
  gAVel.set_style("points").plot_xy(distanceAV,minboundAVel,"Min bound");
  gAVel.set_style("points").plot_xy(distanceAV,maxboundAVel,"Max bound");
  
  //Computing Heading
  vector<double> dir2;
  vector<double> dir2n;
  vector<double> cosdir2n;
  vector<double> sindir2n;
  vector<double> distance2;
  for(int j=0;j<=50;j++){
    double thtmp = (double)this->funcTheta(j*(T/50.0));
    double thtmpn = normalizedAngle(thtmp);
    double disttmp = (j*(T/50.0));
    dir2.push_back(thtmp);
    dir2n.push_back(thtmpn);
    cosdir2n.push_back(cos(thtmpn));
    sindir2n.push_back(sin(thtmpn));
    distance2.push_back(disttmp);
  } 
  Gnuplot gTH;
  gTH.reset_plot();
  gTH.set_style("lines").plot_xy(distance2,dir2,"Heading");
  gTH.set_style("lines").plot_xy(distance2,dir2n,"Heading Nmzd");
  gTH.set_style("lines").plot_xy(distance2,cosdir2n,"Cos");
  gTH.set_style("lines").plot_xy(distance2,sindir2n,"Sin");
  
  //Computing Steering angle
  vector<double> steer;
  vector<double> distanceSteer;
  vector<double> minSteer;
  vector<double> maxSteer;
  for(int j=0;j<=50;j++){
    double st = (double)this->funcSteer(j*(T/50.0));
    double distst = (j*(T/50.0));
    steer.push_back(st);
    minSteer.push_back(-maxSt);
    maxSteer.push_back(maxSt);
    distanceSteer.push_back(distst);
  } 
  Gnuplot gSteer;
  gSteer.reset_plot();
  gSteer.set_style("lines").plot_xy(distanceSteer,steer,"Steering angle");
  gSteer.set_style("points").plot_xy(distanceSteer,minSteer,"Min bound");
  gSteer.set_style("points").plot_xy(distanceSteer,maxSteer,"Max bound"); 

//   //Computing Steering speed
//   vector<double> steerSpd;
//   vector<double> distanceSteerSpd;
//   vector<double> minSteerSpd;
//   vector<double> maxSteerSpd;
//   for(int j=0;j<=50;j++){
//     double stspd = (double)this->funcSteerSpd(j*(T/50.0));
//     double distst = (j*(T/50.0));
//     steerSpd.push_back(stspd);
//     minSteerSpd.push_back(-maxStSpd);
//     maxSteerSpd.push_back(maxStSpd);
//     distanceSteerSpd.push_back(distst);
//   } 
//   Gnuplot gSteerSpd;
//   gSteerSpd.reset_plot();
//   gSteerSpd.set_style("lines").plot_xy(distanceSteerSpd,steerSpd,"Steering speed");
//   gSteerSpd.set_style("points").plot_xy(distanceSteerSpd,minSteerSpd,"Min bound");
//   gSteerSpd.set_style("points").plot_xy(distanceSteerSpd,maxSteerSpd,"Max bound"); 

  wait_for_key();
}

// display linear acceleration and velocity, angular velocity 
// and steering angle profiles without bounds
void CLTrajGen::displayProfilesWB()
{
  //Computing Velocity without bounds
  vector<double> velwb;
  vector<double> realMinboundVelwb;
  vector<double> realMaxboundVelwb;
  vector<double> distanceVwb;
  for(int j=0;j<=50;j++){
    velwb.push_back((double)this->funcVelwb(j*(T/50)));
    realMinboundVelwb.push_back(0.0);
    realMaxboundVelwb.push_back(maxSpd);
    distanceVwb.push_back((j*(T/50)));
  }    
  Gnuplot gVelwb;
  gVelwb.reset_plot();
  gVelwb.set_style("lines").plot_xy(distanceVwb,velwb,"Velocity without bounds");	
  gVelwb.set_style("points").plot_xy(distanceVwb,realMinboundVelwb,"Min bound");
  gVelwb.set_style("points").plot_xy(distanceVwb,realMaxboundVelwb,"Max bound");
  
  //Computing Acceleration without bounds
  vector<double> accwb;
  vector<double> realMinboundAccwb;
  vector<double> realMaxboundAccwb;
  vector<double> distanceAwb;
  for(int j=0;j<=50;j++){
    accwb.push_back((double)this->funcAccwb(j*(T/50)));
    realMinboundAccwb.push_back(-maxAcc);
    realMaxboundAccwb.push_back(maxAcc);
    distanceAwb.push_back((j*(T/50)));
  }    
  Gnuplot gAccwb;
  gAccwb.reset_plot();
  gAccwb.set_style("lines").plot_xy(distanceAwb,accwb,"Acceleration without bounds");
  gAccwb.set_style("points").plot_xy(distanceAwb,realMinboundAccwb,"Min bound");
  gAccwb.set_style("points").plot_xy(distanceAwb,realMaxboundAccwb,"Max bound");
  
  //Computing Angular Velocity without bounds
  double coef = tan(maxSt) / L;
  vector<double> avelwb;
  vector<double> minboundAVelwb;
  vector<double> maxboundAVelwb;
  vector<double> distanceAVwb;
  for(int j=0;j<=50;j++){
    avelwb.push_back((double)this->funcAVelwb(j*(T/50)));
    minboundAVelwb.push_back(this->funcVelwb(j*(T/50))*-coef);
    maxboundAVelwb.push_back(this->funcVelwb(j*(T/50))*coef);
    distanceAVwb.push_back((j*(T/50)));
  }
  Gnuplot gAVelwb;
  gAVelwb.reset_plot();
  gAVelwb.set_style("lines").plot_xy(distanceAVwb,avelwb,"Angular Velocity without bounds");
  gAVelwb.set_style("points").plot_xy(distanceAVwb,minboundAVelwb,"Min bound");
  gAVelwb.set_style("points").plot_xy(distanceAVwb,maxboundAVelwb,"Max bound");
  
  //Computing Steering angle without bounds
  vector<double> steerwb;
  vector<double> distanceSteerwb;
  vector<double> minSteerwb;
  vector<double> maxSteerwb;
  for(int j=0;j<=50;j++){
    double stwb = (double)this->funcSteerwb(j*(T/50.0));
    double diststwb = (j*(T/50.0));
    steerwb.push_back(stwb);
    minSteerwb.push_back(-maxSt);
    maxSteerwb.push_back(maxSt);
    distanceSteerwb.push_back(diststwb);
  } 
  Gnuplot gSteerwb;
  gSteerwb.reset_plot();
  gSteerwb.set_style("lines").plot_xy(distanceSteerwb,steerwb,"Steering angle without bounds");
  gSteerwb.set_style("points").plot_xy(distanceSteerwb,minSteerwb,"Min bound");
  gSteerwb.set_style("points").plot_xy(distanceSteerwb,maxSteerwb,"Max bound");

  //Computing Steering speed without bounds
  vector<double> steerSpd;
  vector<double> distanceSteerSpd;
  vector<double> minSteerSpd;
  vector<double> maxSteerSpd;
  for(int j=0;j<=50;j++){
    double stspd = (double)this->funcSteerSpdwb(j*(T/50.0));
    double distst = (j*(T/50.0));
    steerSpd.push_back(stspd);
    minSteerSpd.push_back(-maxStSpd);
    maxSteerSpd.push_back(maxStSpd);
    distanceSteerSpd.push_back(distst);
  } 
  Gnuplot gSteerSpdwb;
  gSteerSpdwb.reset_plot();
  gSteerSpdwb.set_style("lines").plot_xy(distanceSteerSpd,steerSpd,"Steering speed without bounds");
  gSteerSpdwb.set_style("points").plot_xy(distanceSteerSpd,minSteerSpd,"Min bound");
  gSteerSpdwb.set_style("points").plot_xy(distanceSteerSpd,maxSteerSpd,"Max bound"); 

  wait_for_key();
}

// display linear acceleration and velocity, angular velocity, 
// steering angle, heading and position profiles with and 
// without bounds
void CLTrajGen::displayAllProfiles()
{
  displayProfiles();
  displayProfilesWB();
}

/*
 *  \brief display all functions
 */
void CLTrajGen::displayAllFct()
{
  cout << "Linear Acceleration : " << endl;
  fctAcc_->display();
  cout << "Linear Velocity : " << endl;
  fctVel_->display();
  cout << "Angular Acceleration : " << endl;
  fctAAcc_->display();
  cout << "Angular Velocity : " << endl;
  fctAVel_->display();
  cout << "Heading : " << endl;
  fctTH_->display();
}

// display in an openGL viewer
void CLTrajGen::toGL(double tmin)
{
  int nbPoint = 50;
  
//   // compute the forward solution with the new parameters
//   this->ForwardSolution();
  
//   glLineWidth(5.0);
//   glBegin(GL_LINE_STRIP);
  
//   for(int i=0; i<=nbPoint; i++)
//     {
//       double curT = T * ((double)i)/((double)nbPoint);
      
//       double x = this->funcX(curT);
//       double y = this->funcY(curT);
      
//       glVertex3f(x, y, tmin+curT);
//     }
//   glEnd();

  //glColor3f(1.0,0.0,0.0);
  glPointSize(20.0);

  glBegin(GL_POINTS);
  for(int i=0; i<=nbPoint; i++)
    {
      double curT = T * ((double)i)/((double)nbPoint);
      
      double x = this->funcX(curT);
      double y = this->funcY(curT);
      glVertex3f(x, y, tmin+curT);
      //glVertex3f(0.0, 0.0, 0.0);
    }
  glEnd();

//   const GLfloat redMaterial[]={1.,0.,0.,1.};
//   glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
//   glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);

//   gluDisk(gluNewQuadric(), 0, 0.5, 200, 1);
//   gluCylinder(gluNewQuadric(), 0.5, 0.5, 0.02, 20, 5);
//   glTranslated(0,0,0.02);
//   gluDisk(gluNewQuadric(), 0, 0.5, 200, 1);

//   for(int i=0; i<=nbPoint; i++)
//     {
//       double curT = T * ((double)i)/((double)nbPoint);
      
//       double x = this->funcX(curT);
//       double y = this->funcY(curT);

//       glPushMatrix();
//       glTranslated(x, y, curT);
      
//       gluDisk(gluNewQuadric(), 0, 0.5, 200, 1);
//       gluCylinder(gluNewQuadric(), 0.5, 0.5, 0.02, 20, 5);
//       glTranslated(0,0,0.02);
//       gluDisk(gluNewQuadric(), 0, 0.5, 200, 1);
      
//       glPopMatrix();

//       cout << "(" << x << "," << y << "," << tmin+curT << ")  "; 
//     }
}

// sample the X,Y,T coord of the trajectory
vector<vector<double> > CLTrajGen::sampleXYT(double tmin)
{
  int nbPoint = 50;

  // compute the forward solution with the new parameters
  this->ForwardSolution();

  vector<vector<double> > vres;

  for(int i=0; i<=nbPoint; i++)
    {
      double curT = T * ((double)i)/((double)nbPoint);
      double x = this->funcX(curT);
      double y = this->funcY(curT);

      vector<double> vec;
      vec.push_back(x);
      vec.push_back(y);
      vec.push_back(tmin + curT);

      vres.push_back(vec);
    }

  return vres;
}
