#include "InitialGuess.hpp"
using namespace mpc;

#include "common/inc/Timer.hpp"

#include "gnuplot/inc/gnuplot_i.hpp"
#include "Utils.hpp"
using namespace mpc;

#include <cstring>


                                    /**************************************************/
                                    /******************** Class N2R *******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
N2R::N2R()
{
  dimP_ = 0;
  param_ = NULL;
  initST_ = NULL;
  goalST_ = NULL;

  pos_ = NULL;
//   lnear_ = NULL;

  status_ = ND_N2R;
}

/*
 *  \brief destructor
 */
N2R::~N2R()
{
  if(param_ != NULL) delete [] param_;
  if(initST_ != NULL) delete initST_;
  if(goalST_ != NULL) delete goalST_;

  if(pos_ != NULL) delete [] pos_;
//   if(lnear_ != NULL) delete lnear_;
}


/*************** Getters and Setters **************/

/*
 *  \brief return the i-th parameter
 *  \param i : number of the parameter we want to get back
 *  \return i-th parameter
 */
double N2R::param(int i) const
{
  if(i < 0 || i >= dimP_) return 0;
  else return param_[i];
}

/*
 *  \brief get back the position of the N2R in the associated grid
 */
int* N2R::pos() const
{
  return this->pos_;
}

// /*
//  *  \brief add a new candidate node to try to use the IG parameters from
//  *  \param n : the new candidate node
//  */
// void N2R::addNearNode(N2R* n)
// {
//   if(lnear_ != NULL) lnear_->push_back(n);
// }

// /*
//  *  \brief clear the list of candiates nodes
//  */
// void N2R::clearNearNodes()
// {
//   if(lnear_ != NULL) lnear_->clear();
// }

/*
 *  \brief set the current node as reached and define its parameters
 *  \param param : the parameterized control
 *  
 *  The node must have been set to explored before
 */
void N2R::setAsReached(double* param)
{
  if(param != NULL)
    {
      int N = this->dimP();
      param_ = new double[N];
      
      for(int i=0; i<N; i++) param_[i] = param[i];

      status_ = REACHED_N2R;
    }
}

/*
 *  \brief set the current node as used
 *
 *  The node must have been set to reached before
 */
void N2R::setAsUsed()
{
  status_ = USED_N2R;
}

/*
 *  \brief append the current node in a list
 *  \param it : the list in which the node must be appended
 */
void N2R::appendInList(std::list<N2R*>* list)
{
  // check if the input list is not null
  if(list != NULL)
    {
      // append the node at the end of the list
      list->push_back(this);

      // save in the current node its position in the list
      std::list<N2R*>::iterator it;
      it = list->end();
      it--;
      this->it_ = it;
    }
}

/*
 *  \brief remove the current node from the input list
 *  \param list : the list from which the node must be removed
 *
 *  The user must guarantee that the current node belongs to the input list
 */
void N2R::removeFromList(std::list<N2R*>* list)
{
  // check if the input list is not null
  if(list != NULL)
    {
      list->erase(it_);
    }
}

// /*
//  *  \brief return a random neighbour of the current node
//  *  \return a random neighbour of the current node if one exists, NULL else
//  */
// N2R* N2R::getRandNeighbour() const
// {
//   // check if the list of neighbours has been defined and is not empty
//   if(lnear_ != NULL && (!(lnear_->empty()))) 
//     {
//       int N = lnear_->size();
//       int rand = getIntRandom(0,N-1);

//       // return the node at index rand
//       std::list<N2R*>::iterator it = lnear_->begin();
//       for(int i=0; i<rand; i++) it++;
//       return *it;
//     }
//   else 
//     {
//       return NULL;
//     }
// }

// /**
//  *  \brief return the nearest neighbour of the current node
//  *  \return the nearest neighbour of the current node if one exists, NULL else
//  */
// N2R* N2R::getNearNeighbour() const
// {
//   // check if the list of neighbours has been defined and is not empty
//   if(lnear_ != NULL && (!(lnear_->empty()))) 
//     {
//       N2R* bestNode = NULL;
//       double bestDist = 999999999;

//       std::list<N2R*>::iterator it;
//       for(it = lnear_->begin(); it != lnear_->end(); it++)
// 	{
// 	  N2R* node = *it;
// 	  double dist = this->computeWDistTo(node);

// 	  // check if the current node if the closest node
// 	  if(bestNode == NULL || dist < bestDist)
// 	    {
// 	      bestNode = node;
// 	      bestDist = dist;
// 	    }
// 	}

//       // return the closest node found
//       return bestNode;
//     }
//   else 
//     {
//       return NULL;
//     }
// }


/**************** Display functions ***************/

/*
 *  \brief display the current node
 */
void N2R::display() const
{
  switch(status_)
    {
    case ND_N2R : 
      cout << "Undefined node." << endl; 
      break;
    case IG_N2R : 
      cout << "IG node.";
      cout << "  init st: "; initST_->display();
      cout << "  goal st: "; goalST_->display();
      cout << "  ctrl params: "; mpc::displayTab(param_,dimP_);
      cout << endl;
      break;
    case EXPLORED_N2R : 
      cout << "Node to reach.";
      cout << "  init st: "; initST_->display();
      cout << "  goal st: "; goalST_->display();
//       cout << "  nb neighbours : " << lnear_->size();
      cout << endl;
      break;
    case REACHED_N2R : 
      cout << "Reached node.";
      cout << "  init st: "; initST_->display();
      cout << "  goal st: "; goalST_->display();
      cout << "  ctrl params: "; mpc::displayTab(param_,dimP_);
      cout << endl;
      break;
    case USED_N2R : 
      cout << "Used node.";
      cout << "  init st: "; initST_->display();
      cout << "  goal st: "; goalST_->display();
      cout << "  ctrl params: "; mpc::displayTab(param_,dimP_);
      cout << endl;
      break;
    }
}


/**************** Utility functions ***************/

/*
 *  \brief save a list of parameterized trajectories into an output file
 *  \param list : the list of trajectories to save
 *  \param filename : the path name of the output file
 *  \return 0 if the process succeed, -1 else
 *
 *  The features saved are the initial state-time, the final state-time and the parameterized control
 */
int N2R::saveListN2R(std::list<N2R*>* list,const char* filename)
{
  // check if the input list is not nul or empty
  if(list == NULL || list->empty())
    {
      cerr << "N2R::saveListN2R : the input list is null or empty" << endl;
      //throw string("N2R::saveListN2R : the input list is null or empty");
      return -1;
    }

  // get back the first node of the list
  std::list<N2R*>::iterator it = list->begin();

  // get back the dimension of the state-time vector
  int dimS = (*it)->initST().size();
  int dimP = (*it)->dimP();

  ofstream fic(filename, ios::out | ios::trunc);
 
  if(fic)
    {
      // write the dimension of the state-time vector
      fic << dimS << " ";
      // write the dimension of the parameters vector
      fic << dimP << endl;

      // write the features of each node in the output file
      for(it = list->begin(); it != list->end(); it++)
	{
	  // write the initial state-time into the file
	  for(int i=0; i<dimS; i++) fic << (*it)->initST().at(i) << " ";

	  // write the final state-time into the file
	  for(int i=0; i<dimS; i++) fic << (*it)->goalST().at(i) << " ";

	  // write the control trajectory parameters into the file
	  const double* param = (*it)->params();
	  for(int i=0; i<dimP-1; i++) fic << param[i] << " ";
	  fic << param[dimP-1] << endl;
	}
      fic << endl;
      
      fic.close();
    }
  else
    {
      cerr << "N2R::saveListN2R : cannot open the output file" << endl;
      //throw string("N2R::saveListN2R : cannot open the output file");
      return -1;
    }

  cout << "List of trajectories saved in file " << filename << endl;
  return 0;
}


                                    /**************************************************/
                                    /****************** Class N2RGrid *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
N2RGrid::N2RGrid()
{
  tabGrid_      = NULL;

  listN2R_      = NULL;
  listNReached_ = NULL;
  listNUsed_    = NULL;

  n_            = 0;
  sampling_     = NULL;
  minSampS_     = NULL;
  maxSampS_     = NULL;
}

/*
 *  \brief init constructor
 *  \param n : dimension of the grid space
 *  \param sampling : sampling of all dimensions of the grid
 *  \param minSampS : minimal values taken by the sampling
 *  \param maxSampS : maximal values taken by the sampling
 */
N2RGrid::N2RGrid(int n,int* sampling,double* minSampS,double* maxSampS)
{
  n_ = n;
  sampling_ = new int[n];
  minSampS_ = new double[n];
  maxSampS_ = new double[n];
  for(int i=0; i<n; i++) sampling_[i] = sampling[i];
  for(int i=0; i<n; i++) minSampS_[i] = minSampS[i];
  for(int i=0; i<n; i++) maxSampS_[i] = maxSampS[i];

  tabGrid_ = NULL;
  listN2R_ = new std::list<N2R*>;
  listNReached_ = new std::list<N2R*>;
  listNUsed_ = new std::list<N2R*>;

  this->initGrid();
}

/*
 *  \brief destructor
 */
N2RGrid::~N2RGrid()
{
  this->clearGrid();
  if(listN2R_ != NULL) delete listN2R_;
  if(listNReached_ != NULL) delete listNReached_;
  if(listNUsed_ != NULL) delete listNUsed_;

  if(sampling_ != NULL) delete [] sampling_;
  if(minSampS_ != NULL) delete [] minSampS_;
  if(maxSampS_ != NULL) delete [] maxSampS_;
}


/*************** Getters and Setters **************/

/*
 *  \brief return the size of the grid (computed from the sampling dimensions)
 *  \return the size of the grid
 */
int N2RGrid::size() const
{
  if(sampling_ != NULL)
    {
      int res = 1;
      for(int i=0; i<n_; i++) res *= sampling_[i];
      return res;
    }
  else return 0;
}

/*
 *  \brief return the list of nodes to reach
 *  \return a copy of the list of nodes to reach
 *
 *  A new instance of list is allocated here
 */
std::list<N2R*> N2RGrid::listN2R() const
{
  std::list<N2R*> list;
  std::list<N2R*>::iterator it;
  for(it = listN2R_->begin(); it != listN2R_->end(); it++) list.push_back(*it);
  return list;
}

/*
 *  \brief return the list of reached nodes
 *  \return a copy of the list of reached nodes
 *
 *  A new instance of list is allocated here
 */
std::list<N2R*> N2RGrid::listNReached() const
{
  std::list<N2R*> list;
  std::list<N2R*>::iterator it;
  for(it = listNReached_->begin(); it != listNReached_->end(); it++) list.push_back(*it);
  return list;
}

/*
 *  \brief return the list of used nodes
 *  \return a copy of the list of used nodes
 *
 *  A new instance of list is allocated here
 */
std::list<N2R*> N2RGrid::listNUsed() const
{
  std::list<N2R*> list;
  std::list<N2R*>::iterator it;
  for(it = listNUsed_->begin(); it != listNUsed_->end(); it++) list.push_back(*it);
  return list;
}


/********* Access/Modification of the grid ********/

/*
 *  \brief init the grid from the current sampling parameters
 */
void N2RGrid::initGrid()
{
  // clear the actual grid if needed
  this->clearGrid();

  // compute the dimension of the grid
  int res;
  if(n_ == 0) res = 0;
  else
    {
      res = 1;
      for(int i=0; i<n_; i++) res *= sampling_[i];
    }

  // allocate the grid
  tabGrid_ = new N2R* [res];
  for(int i=0; i<res; i++) tabGrid_[i] = NULL;
}

/*
 *  \brief clear the grid
 */
void N2RGrid::clearGrid()
{
  if(tabGrid_ != NULL)
    {
      int N = this->size();
      for(int i=0; i<N; i++)
	if(tabGrid_[i] != NULL) delete tabGrid_[i];

      delete [] tabGrid_;
      tabGrid_ = NULL;
    }
}

/*
 *  \brief return an element of the grid at coord
 *  \param coord : coordinates of the N2R
 *  \return the element of the grid at coord, NULL if the input coordinates are outside of the grid
 */
N2R* N2RGrid::gridAt(int* coord) const
{
  // check if the input position is inside of the gird
  bool outside = false;
  for(int i=0; i<n_; i++)
    {
      if(coord[i] < 0 || coord[i] >= sampling_[i]) outside = true;
    }
  if(outside) return NULL;

  // compute the index of the element in the 1D tab
  int ind1D;
  indMD_2_ind1D(n_,sampling_,coord,ind1D);

  return tabGrid_[ind1D];
}

/*
 *  \brief update the list of reachable nodes by considering the nodes already reached
 */
void N2RGrid::updateN2RfromReached()
{
  // Tables to save the minimal and maximal indices of each dimension 
  // of the search vector 
  int* minInd = new int[n_];
  int* maxInd = new int[n_];

  // search for the nodes surrounding each node already reached
  std::list<N2R*>::iterator it;
  for(it = listNReached_->begin(); it != listNReached_->end(); it++)
    {
      // search for the nodes surrounding each node already reached
      std::list<N2R*>* list = this->reachableNodesAround(*it);

      // check if the current node is used (ie. if the list is empty)
      if(list->empty())
	{
	  // set the current node as used
	  this->setNodeAsUsed(*it);
	}
      else
	{
	  // set each neighbour node as a node to reach
	  std::list<N2R*>::iterator it2;
	  for(it2 = list->begin(); it2 != list->end(); it2++) this->addN2Rfrom((*it2)->pos(),*it);
	}
	  
      // delete the created list of neighbours
      list->clear();
      delete list;
    }

  // delete the tables
  delete [] minInd;
  delete [] maxInd;
}

/*
 *  \brief update the list of reachable nodes from the initial guess of trajectories
 *  \param lig : the list of nodes provided by the initial guess
 */
void N2RGrid::updateN2RfromIG(std::list<N2R*>& lig)
{
  // search for nodes around each nodes included in the input list
  std::list<N2R*>::iterator it;
  for(it = lig.begin(); it != lig.end(); it++)
    {
      std::list<N2R*>* l = this->reachableNodesAround(*it);

      // concat each element of the current list to the resulting list
      std::list<N2R*>::iterator it2;
      for(it2 = l->begin(); it2 != l->end(); it2++) this->addN2Rfrom((*it2)->pos(),*it);

      // delete the current list
      l->clear();
      delete l;
    }
}

/*
 *  \brief set the node at position coord as reached
 *  \param coord : the input position in the grid of the reached node
 *  \param param : the parameterized control
 *
 *  The vector of parameterized control will be copied
 */
void N2RGrid::setNodeAsReached(int* coord,double* param)
{
  // get back the node at position coord
  N2R* node = this->gridAt(coord);
  
  // check if the node has already been defined
  if(node == NULL)
    {
      cerr << "N2RGrid::setNodeAsReached : The node cannot be set as reached because it does not have been defined..." << endl;
      return;
    }  

  // remove it from the explored list
  this->removeNodeFromLN2R(node);
  
  // add it to the reached list
  this->appendNodeToLReached(node);

  // set the node as reached
  node->setAsReached(param);
}

/*
 *  \brief set the input node as used
 *  \param node : the input node
 */
void N2RGrid::setNodeAsUsed(N2R* node)
{
  // check if the input node is not NULL
  if(node == NULL)
    {
      cerr << "N2RGrid::setNodeAsUsed : The node cannot be set as used because it does not have been defined..." << endl;
      return;
    }

  // remove the node from the reached list
  this->removeNodeFromLReached(node);
  
  // add it to the used list
  this->appendNodeToLUsed(node);

  // set the node as reached
  node->setAsUsed();
}

/*
 *  \brief return a reference to an element of the grid at coord
 *  \param coord : coordinates of the N2R
 *  \return a reference to the element of the grid at coord
 */
N2R*& N2RGrid::gridAt(int* coord)
{
  // compute the index of the element in the 1D tab
  int ind1D;
  indMD_2_ind1D(n_,sampling_,coord,ind1D);
  
  return tabGrid_[ind1D];
}

/**
 *  \brief select randomly an already reached node around the input node
 *  \param node : the input node
 *  \param lfig : list of first initial guesses
 */
N2R* N2RGrid::selectReachedNodeAround(N2R* node,std::list<N2R*>* lfig) const
{
  N2R* res = NULL;

  if(lfig != NULL)
    {
      double minDist = 0;
      res = NULL;
      std::list<N2R*>::iterator it;

      for(it = lfig->begin(); it != lfig->end(); it++)
	{
	  double dist = node->computeWDistTo(*it);

	  // check if the current node is the closest to the input node
	  if(res == NULL || dist < minDist)
	    {
	      minDist = dist;
	      res = *it;
	    }
	}
    }
  else
    {
//       cout << "selectReachedNodeAround 1" << endl;
      // get back the list of reached nodes around the input node
      std::list<N2R*>* lReached = this->reachedNodesAround(node);
      
//       cout << "selectReachedNodeAround 2" << endl;
      // get back the number of reached nodes around the input node
      int N=0;
      if(lReached == NULL) N=0;
      else N = lReached->size();
      
//       cout << "selectReachedNodeAround 3" << endl;
      if(N == 0) res = NULL;
      else
	{
	  int rand = getIntRandom(0,N-1);
	  
	  // get back the node at index rand
	  std::list<N2R*>::iterator it = lReached->begin();
	  for(int i=0; i<rand; i++) it++;
	  res = *it;
	}
      
//       cout << "selectReachedNodeAround 4" << endl;
      // delete the list of reached nodes
      if(lReached != NULL) delete lReached;
//       cout << "selectReachedNodeAround 5" << endl;
    }
  return res;
}

/*
 *  \brief note a N2R of the grid as possibly reachable from node
 *  \param coord : coordinates of the N2R
 *  \param node : node from which the N2R may be reachable
 *  \return a pointer to the N2R
 */
N2R* N2RGrid::addN2Rfrom(int* coord,N2R* node)
{
  // get back the N2R cell in the grid
  N2R* n2r = this->gridAt(coord);

  // check if the node has already been noted as potentially reachable
  if(n2r != NULL)
    {
      // we add the input node as an other possibility to reach the n2r
//       n2r->addNearNode(node);
      // nothing to do? I don't think so
    }
  else 
    {
      // we create a new node to reach
      n2r = createExploredN2R(coord,node);

      // add it to the list of possibly reachable nodes
      this->appendNodeToLN2R(n2r);
    }

  return n2r;
}


/******** Access/Modification of the lists ********/

/*
 *  \brief append the input node to the list of used nodes
 *  \param node : the input node
 */
void N2RGrid::appendNodeToLUsed(N2R* node)
{
  if(node != NULL) node->appendInList(listNUsed_);
}

/*
 *  \brief append the input node to the list of reached nodes
 *  \param node : the input node
 */
void N2RGrid::appendNodeToLReached(N2R* node)
{
  if(node != NULL) node->appendInList(listNReached_);
}

/*
 *  \brief append the input node to the list of explored nodes (nodes to reach)
 *  \param node : the input node
 */
void N2RGrid::appendNodeToLN2R(N2R* node)
{
  if(node != NULL) node->appendInList(listN2R_);
}

/*
 *  \brief remove the input node from the list of reached nodes
 *  \param node : the input node
 */
void N2RGrid::removeNodeFromLReached(N2R* node)
{
  if(node != NULL) node->removeFromList(listNReached_);
}

/*
 *  \brief remove the input node from the list of explored nodes (nodes to reach)
 *  \param node : the input node
 */
void N2RGrid::removeNodeFromLN2R(N2R* node)
{
  if(node != NULL) node->removeFromList(listN2R_);
}

/*
 *  \brief clear the list of nodes to reach
 */
void N2RGrid::clearListN2R()
{
  // check if the list of N2R is defined
  if(listN2R_ != NULL)
    {
      // remove the previous links of each node to its neighbours
//       std::list<N2R*>::iterator it;
//       for(it = listN2R_->begin(); it != listN2R_->end(); it++)
// 	(*it)->clearNearNodes();
      
      // clear the list of nodes to reach
      listN2R_->clear();
    }
}

/*
 *  \brief display the list of explored nodes (nodes to reach)
 */
void N2RGrid::displayLN2R() const
{
  cout << "listN2R_->size() : " << listN2R_->size() << endl;
  cout << "List of explored nodes: " << endl;
  std::list<N2R*>::iterator it;
  for(it = listN2R_->begin(); it != listN2R_->end(); it++)
    (*it)->display();
}

/*
 *  \brief display the list of reached nodes
 */
void N2RGrid::displayLReached() const
{
  cout << "List of reached nodes: " << endl;
  cout << "listNReached_->size() : " << listNReached_->size() << endl;
  std::list<N2R*>::iterator it;
  for(it = listNReached_->begin(); it != listNReached_->end(); it++)
    (*it)->display();
  cout << "fin...!" << endl;
}

/*
 *  \brief display the list of used nodes
 */
void N2RGrid::displayLUsed() const
{
  cout << "List of used nodes: " << endl;
  std::list<N2R*>::iterator it;
  for(it = listNUsed_->begin(); it != listNUsed_->end(); it++)
    (*it)->display();
}

/*
 *  \brief display the nodes set as reached on the grid
 */
void N2RGrid::displayReachedOnGrid() const
{
}


                                    /**************************************************/
                                    /**************** Class InitialGuess **************/
                                    /**************************************************/


/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */  
InitialGuess::InitialGuess()
{
  grid_ = NULL;

  dimSampS_ = 0;

  sampling_ = NULL;
  minSampS_ = NULL;
  maxSampS_ = NULL;

  trajGen_ = NULL;
  sys_ = NULL;

  firstIGMode_ = 1; // (0=random, 1=sampled control, 2=sampled param control, 3=rstar)

  minNNTSR_ = 1;

  debug_ = 0; // (0=no debug, 1=save traj of final state-times reached, 2=save traj of intermediate steps)

  tabX_ = NULL;
  tabY_ = NULL;
  tabV_ = NULL;
  tabT_ = NULL;

  maxTGResidual_ = 0.5;
  maxTGSteps_ = 20;

  FIGfilename_ = NULL;
}

/*
 *  \brief init constructor
 */
InitialGuess::InitialGuess(System* sys)
{
  grid_ = NULL;

  dimSampS_ = 0;

  sampling_ = NULL;
  minSampS_ = NULL;
  maxSampS_ = NULL;

  trajGen_ = NULL;
  sys_ = sys;

  firstIGMode_ = 1; // (0=random, 1=sampled control, 2=sampled param control, 3=rstar)

  minNNTSR_ = 1;

  debug_ = 0; // (0=no debug, 1=save traj of final state-times reached, 2=save traj of intermediate steps)

  tabX_ = NULL;
  tabY_ = NULL;
  tabV_ = NULL;
  tabT_ = NULL;

  maxTGResidual_ = 0.5;
  maxTGSteps_ = 20;

  FIGfilename_ = NULL;
}

/*
 *  \brief destructor
 */  
InitialGuess::~InitialGuess()
{
  if(sampling_ != NULL) delete [] sampling_;
  if(minSampS_ != NULL) delete [] minSampS_;
  if(maxSampS_ != NULL) delete [] maxSampS_;

  if(tabX_ != NULL) delete tabX_;
  if(tabY_ != NULL) delete tabY_;
  if(tabV_ != NULL) delete tabV_;
  if(tabT_ != NULL) delete tabT_;
}


/*************** Getters and Setters **************/

/**
 *  \brief use the "load from file" mode to determine the first initial guess 
 */
void InitialGuess::use_LFF_FIGMode(const char* filename)
{
  if(filename != NULL)
    {
      if(FIGfilename_ != NULL) delete FIGfilename_;
      FIGfilename_ = new char[50];
      strcpy(FIGfilename_,filename);
      this->firstIGMode_ = LOADFROMFILE_FIRSTIG;
    }
  else
    {
      cerr << "InitialGuess::use_LFF_FIGMode : the input filename has not been defined. Default mode used." << endl;
      this->firstIGMode_ = SC_FIRSTIG;
    }
}

// /*
//  *  \brief set the mode used to determine the first initial guess 
//  *  \param mode : the new mode (0=random, 1=sampled control, 2=sampled param control, 3=rstar)
//  *  \param options : options of the different modes
//  *   - SC_FIRSTIG : table containing the sampling over the control
//  *   - LOADFROMFILE_FIRSTIG : input filename path
//  */
// void InitialGuess::setFirstIGMode(int mode,const void* options)
// { 
//   this->firstIGMode_ = mode; 

//   switch(mode)
//     {
//     case SC_FIRSTIG : 
//       {
// 	if(options != NULL)
// 	  {
// 	    this->setFIGCtrlSpg((int*)options);
// 	  }
//       }
//     case LOADFROMFILE_FIRSTIG : 
//       {
// 	if(options != NULL)
// 	  {
// 	    if(FIGfilename_ != NULL) delete FIGfilename_;
// 	    FIGfilename_ = new char[50];
// 	    strcpy(FIGfilename_,(char*)options);
// 	  }
// 	else
// 	  {
// 	    cerr << "InitialGuess::setFirstIGMode : the input filename has not been defined. Default mode used." << endl;
// 	    mode = SC_FIRSTIG;
// 	  }
//       }
//     }
// }


/***************** Initialization *****************/

/*
 *  \brief clear the sampling grid
 */
void InitialGuess::clearGrid()
{
  if(grid_ != NULL) delete grid_;
}


/******************** Learning ********************/

/*
 *  \brief compute the first initial guess from general functions
 *  \param lfig : the resulting list of first initial guesses
 */
std::list<N2R*>* InitialGuess::computeFirstIG()
{
  switch(firstIGMode_)
    {
    case RAND_FIRSTIG : return this->computeRandFirstIG(); break;
    case SC_FIRSTIG : return this->computeSCFirstIG(); break;
    case SCP_FIRSTIG : return this->computeSCPFirstIG(); break;
    case RSTAR_FIRSTIG : return this->computeRStarFirstIG(); break;
    case LOADFROMFILE_FIRSTIG : return this->loadFirstIGfromFile(); break;
    default : 
      cerr << "InitialGuess::computeFirstIG : unknown first ig mode" << endl;
      exit(1);
    }
}

/*
 *  \brief determine the list of potentially reachable nodes on the grid
 *  \return the resulting list of potentially reachable nodes
 */
std::list<N2R*> InitialGuess::determineNodesToReach()
{
  // clear the old list of nodes to reach on the grid
  grid_->clearListN2R();

  // compute the nodes to reach from the list of nodes already reached
  grid_->updateN2RfromReached();

  // return the list of nodes to reach defined by the grid
  return grid_->listN2R();
}

/*
 *  \brief determine the list of potentially reachable nodes from a list
 *  \param lig : the input list of nodes
 *  \return the resulting list of potentially reachable nodes
 */
std::list<N2R*> InitialGuess::determineNodesToReach(std::list<N2R*>& lig)
{
  // clear the old list of nodes to reach on the grid
  grid_->clearListN2R();

  // compute the nodes to reach from the list of nodes already reached
  grid_->updateN2RfromIG(lig);

  // return the list of nodes to reach defined by the grid
  return grid_->listN2R();
}

/*
 *  \brief learn the initial guess for the sample set
 */
void InitialGuess::learnIG()
{
  Timer timer;
  if(debug_ >= 0) timer.start();

  // initilialize the sampling grid
  this->initGrid();

  // compute the first initial guess
  std::list<N2R*>* lfig;
  lfig = this->computeFirstIG();
  if(debug_ >= 2) drawGPTraj(*lfig);

  // expand the set of final nodes reached
  int nntsr = 999999999; // number of new target state-times reached
  int step = 0;
  
  //while(nntsr > minNNTSR_ && step < 2)
  while(nntsr > minNNTSR_)
    {
      // increase the number of the current step
      step++;
      // initialize the number of nodes reached
      nntsr = 0;

      std::list<N2R*> lN2R;
      if(lfig == NULL)
	{
	  // determine the nodes to reach from the nodes already reached
	  lN2R = this->determineNodesToReach();

	  if(debug_ >= 2) drawGPGrid();
	}
      else
	{
	  // determine the nodes to reach around the final state-times of initial guesses
	  lN2R = this->determineNodesToReach(*lfig);

	  if(debug_ >= 2) drawGPTraj(*lfig);

// 	  // delete initial guesses
// 	  if(lfig != NULL) delete lfig;
// 	  lfig = NULL;
	}
      
//       if(debug_ >= 2) this->drawGPGrid();

      int N = lN2R.size();
      int cptBar = 0;
      int stepBar = 1;
      int nbStepBar = 20;
      std::list<N2R*>::iterator it;

      if(debug_ >= 0) cout << "[" << flush;

      for(it = lN2R.begin(); it != lN2R.end(); it++)
	{
	  // get back the current node to try to reach
	  N2R* n2r = *it;
// 	  cout << "n2r : "; n2r->display(); cout << endl;
	  
	  // select a random node from this set
// 	  N2R* near = n2r->getRandNeighbour();
	  N2R* near;
	  if(lfig == NULL) near = grid_->selectReachedNodeAround(n2r);
	  else near = grid_->selectReachedNodeAround(n2r,lfig);

	  // try to generate a trajectory to n2r thanks to the parameters of near
	  double* param = new double[dimP_];
	  int res = this->genTrajTo(n2r,near,param);
	  if(res == 0)
	    {
	      // mark the node as reached
	      grid_->setNodeAsReached(n2r->pos(),param);
	      
	      // update the number of reached nodes if needed
	      nntsr++;
	    }

	  delete [] param;

	  if(debug_ >= 0)
	    {
// 	      cout << "cptBar : " << cptBar << endl;
	      cptBar++;
	      if(cptBar >= stepBar * ((double)N)/((double)nbStepBar))
		{
		  cout << "=" << flush;
		  stepBar++;
		}
	    }
	}

      // delete initial guesses
      if(lfig != NULL) delete lfig;
      lfig = NULL;

      if(debug_ >= 0) cout << "]" << endl;
      cout << "//////////////////////////////////////////////////////////////////" << endl;

      double prct;
      if(N != 0) prct = ((double)nntsr)/((double)N)*100.0;
      else prct = 0.0;
      if(debug_ >= 1) cout << "step " << step << " -> Reached : " << nntsr << "/" << N << " (" << prct << "\%)   ";
      if(debug_ >= 1) cout << "Total reached : " << grid_->NReachedonGrid() << "(" << grid_->NReachedinList() << ")"
			   << "  Total used : " << grid_->NUsedonGrid() << "(" << grid_->NUsedinList() << ")" << endl;

      if(debug_ >= 0) cout << "Total elapsed time : " << timer.totalElapsedTime() << endl;
      if(debug_ >= 2) this->drawGPGrid();
    }
  if(debug_ >= 0) timer.stop();

  this->save("IG.dat");
}


/************* Input/Output Functions *************/

/*
 *  \brief load the sampling of initial guesses from an input file
 *  \param nomfic : the name of the input file
 *  \return 0 if the process succeed, -1 else
 */
int InitialGuess::load(char* nomfic)
{
  ifstream fic(nomfic, ios::in);

  if(fic)
    {
      if(fic.eof())
	{
	  cerr << "InitialGuess::load : The input file is empty..." << endl;
	  fic.close();
	  throw string("The input file is empty...");
	  return -1;
	}

      int dimShere,dimPhere,dimSampShere;
      // get back the dimension of the state-time vector
      fic >> dimShere;
      // get back the dimension of the parameters vector
      fic >> dimPhere;
      // get back the dimension of the parameters vector
      fic >> dimSampShere;
      
      // check if these parameters correspond to the current IG database
      if(dimShere != dimS_ || dimPhere != dimP_ || dimSampShere != dimSampS_)
	{
	  cerr << "InitialGuess::load : Wrong input IG database..." << endl;
	  fic.close();
	  throw string("Wrong input IG database...");
	  return -1;
	}

      // get back the sampling vectors
      for(int i=0; i<dimSampS_; i++) fic >> sampling_[i];
      for(int i=0; i<dimSampS_; i++) fic >> minSampS_[i];
      for(int i=0; i<dimSampS_; i++) fic >> maxSampS_[i];

      while(!fic.eof())
	{
	  // get back the position of the node in the grid
	  int* pos = new int[dimS_];
	  for(int i=0; i<dimS_; i++) fic >> pos[i];

	  // get back the parameterized control of the node
	  double* params = new double[dimP_];
	  for(int i=0; i<dimP_; i++) fic >> params[i];

	  // create the new reached node and put it in the grid
	  N2R* node = grid_->createReachedNode(pos,params);

	  // delete the position and parameters
	  delete [] pos;
	  delete [] params;
	}
    }
  else
    {
      cout << "InitialGuess::load : Unable to open the input file" << endl;
      throw string("Unable to open the input file");
      return -1;
    }

  if(debug_ >= 1) cout << "Total reached : " << grid_->NReachedonGrid() << "(" << grid_->NReachedinList() << ")"
		       << "  Total used : " << grid_->NUsedonGrid() << "(" << grid_->NUsedinList() << ")" << endl;

  cout << "IG database loaded from file " << nomfic << endl;

  fic.close();

  if(debug_ >= 1) grid_->displayReachedOnGrid();
  return 0;
}

/*
 *  \brief save the sampling of initial guesses into an output file
 *  \param nomfic : the name of the output file
 *  \return 0 if the process succeed, -1 else
 */
int InitialGuess::save(char* nomfic)
{
  ofstream fic(nomfic, ios::out | ios::trunc);
 
  if(fic)
    {
      // write the dimension of the state-time vector
      fic << dimS_ << " ";
      // write the dimension of the parameters vector
      fic << dimP_ << " ";
      // write the dimension of the parameters vector
      fic << dimSampS_ << endl;

      // write the sampling vectors
      for(int i=0; i<dimSampS_-1; i++) fic << sampling_[i] << " ";
      if(dimSampS_ >= 1) fic << sampling_[dimSampS_-1] << endl;
      for(int i=0; i<dimSampS_-1; i++) fic << minSampS_[i] << " ";
      if(dimSampS_ >= 1) fic << minSampS_[dimSampS_-1] << endl;
      for(int i=0; i<dimSampS_-1; i++) fic << maxSampS_[i] << " ";
      if(dimSampS_ >= 1) fic << maxSampS_[dimSampS_-1] << endl;

      // get back the list of reached and used nodes
      std::list<N2R*> lNReached = grid_->listNReached();
      std::list<N2R*> lNUsed = grid_->listNUsed();

      // put the location and the parameters of each completed trajectory of the reached nodes list
      std::list<N2R*>::iterator itR;
      for(itR = lNReached.begin(); itR != lNReached.end(); itR++)
	{
	  int nS = grid_->dimS();
	  int nP = (*itR)->dimP();

	  int* coord = (*itR)->pos();
	  const double* param = (*itR)->params();

	  for(int i=0; i<nS; i++) fic << coord[i] << " ";
	  for(int j=0; j<nP-1; j++) fic << param[j] << " ";
	  if(nP >=1) fic << param[nP-1] << endl;
	}

      // put the location and the parameters of each completed trajectory of the used nodes list
      std::list<N2R*>::iterator itU;
      for(itU = lNUsed.begin(); itU != lNUsed.end(); itU++)
	{
	  int nS = grid_->dimS();
	  int nP = (*itU)->dimP();

	  int* coord = (*itU)->pos();
	  const double* param = (*itU)->params();

	  for(int i=0; i<nS; i++) fic << coord[i] << " ";
	  for(int j=0; j<nP-1; j++) fic << param[j] << " ";
	  if(nP >=1) fic << param[nP-1] << endl;
	}      
      
      fic.close();
    }
  else
    {
      cout << "InitialGuess::save : Unable to open the output file" << endl;
      return -1;
    }

  cout << "IG database saved in file " << nomfic << endl;
  return 0;
}


/*************** Utility Functions ****************/

/*
 *  \brief compute the dimension of the grid (number of target state-times we try to reach)
 *  \return the dimension of the grid (number of target state-times we try to reach)
 */
int InitialGuess::computeNbTargetST() const
{
  if(sampling_ != NULL)
    {
      int res = 1;
      for(int i=0; i<dimSampS_; i++) res *= sampling_[i];
      return res;
    }
  else return 0;
}

/*
 *  \brief display the base of initial guesses
 */
void InitialGuess::display() const
{
  cout << "Base of initial guesses: " << endl;
//   cout << "dimS_ : " << dimS_ << endl;
  cout << "dimSampS_ : " << dimSampS_ << endl;
  cout << "dimP_ : " << dimP_ << endl;
//   cout << "dimC_ : " << dimC_ << endl;
  cout << endl;
}

/*
 *  \brief sample the target state-times into tables of position and time (WxT)
 *  \param tabX : the resulting table containing the position X values
 *  \param tabY : the resulting table containing the position Y values
 *  \param tabT : the resulting table containing the time T values
 *
 *  Tables must be allocated first
 */
void InitialGuess::sampleTgtST(double* tabX,double* tabY,double *tabT) const
{
  std::vector<N2R*>* l = this->listAllNodes();
  std::vector<N2R*>::iterator it = l->begin();

  int i=0;
  for(it = l->begin(); it != l->end(); it++)
    {
      // get back the position-time of the current target state-time
      PositionXYT posXYT = (*it)->goalST().posXYT();
      
      // fill the resulting tables with its coordinates
      tabX[i] = posXYT.x();
      tabY[i] = posXYT.y();
      tabT[i] = posXYT.t();
      i++;
    }
  delete l;
}

/*
 *  \brief display in a gnuplot window the trajectories of the input list
 *  \param l : list of the trajectories to display
 */
void InitialGuess::drawGPTraj(std::list<N2R*> l)
{
  // get back the lists of explored, reached and used nodes
  std::list<N2R*> lN2R = grid_->listN2R();
  std::list<N2R*> lNReached = grid_->listNReached();
  std::list<N2R*> lNUsed = grid_->listNUsed();

  int N_N2R = lN2R.size();
  std::vector<double> vecN2RX(N_N2R);
  std::vector<double> vecN2RY(N_N2R);
  std::vector<double> vecN2RT(N_N2R);

  std::list<N2R*>::iterator it1;
  int i = 0;
  for(it1=lN2R.begin(); it1 != lN2R.end(); it1++)
    {
      PositionXYT pos = (*it1)->goalST().posXYT();
      vecN2RX[i] = pos.x();
      vecN2RY[i] = pos.y();
      vecN2RT[i] = pos.t();
      i++;
    }

  int N_Reached = lNReached.size();
  std::vector<double> vecNReachedX(N_Reached);
  std::vector<double> vecNReachedY(N_Reached);
  std::vector<double> vecNReachedT(N_Reached);

  std::list<N2R*>::iterator it2;
  int j = 0;
  for(it2=lNReached.begin(); it2 != lNReached.end(); it2++)
    {
      PositionXYT pos = (*it2)->goalST().posXYT();
      vecNReachedX[j] = pos.x();
      vecNReachedY[j] = pos.y();
      vecNReachedT[j] = pos.t();
      j++;
    }

  int N_Used = lNUsed.size();
  std::vector<double> vecNUsedX(N_Used);
  std::vector<double> vecNUsedY(N_Used);
  std::vector<double> vecNUsedT(N_Used);

  std::list<N2R*>::iterator it3;
  int k = 0;
  for(it3=lNUsed.begin(); it3 != lNUsed.end(); it3++)
    {
      PositionXYT pos = (*it3)->goalST().posXYT();
      vecNReachedX[k] = pos.x();
      vecNReachedY[k] = pos.y();
      vecNReachedT[k] = pos.t();
      k++;
    }

  int NbP = 40;

  // reinit the tables containing the coordinates of trajectories
  if(tabX_ != NULL) delete tabX_;
  if(tabY_ != NULL) delete tabY_;
  if(tabV_ != NULL) delete tabV_;
  if(tabT_ != NULL) delete tabT_;
  tabX_ = new vector<double>;
  tabY_ = new vector<double>;
  tabV_ = new vector<double>;
  tabT_ = new vector<double>;
  
  // sample all generated trajectories
  int N = l.size();
  std::list<N2R*>::iterator it = l.begin();

  for(int i=0; i<N; i++)
    {
      // check if a trajectory has been found between the state-times of this sample
      if(*it != NULL && (*it)->params() != NULL)
	{
	  // sample the trajectory defined by the parameterized control
	  double tabX[NbP];
	  double tabY[NbP];
	  double tabT[NbP];
	  this->sampleTrajNode(*it,tabX,tabY,tabT,NbP);

	  // append these tables to the list of all points of the trajectories
	  for(int j=0; j<NbP; j++)
	    {
	      tabX_->push_back(tabX[j]);
	      tabY_->push_back(tabY[j]);
	      tabT_->push_back(tabT[j]);
	    }
	  for(int j=NbP-1; j>=0; j--)
	    {
	      tabX_->push_back(tabX[j]);
	      tabY_->push_back(tabY[j]);
	      tabT_->push_back(tabT[j]);
	    }
	}

      it++;
    }

  // draw the list of trajectories in GnuPlot windows
//   Gnuplot gPosXY;
//   gPosXY.reset_plot();
//   gPosXY.cmd("set linestyle 1");
//   gPosXY.set_style("lines").plot_xy(*tabX_,*tabY_,"Trajectories (XxY)");

  Gnuplot gPosXYT;
  gPosXYT.reset_plot();
  gPosXYT.cmd("set linestyle 1");
//   gPosXYT.set_style("lines").plot_xyz(*tabX_,*tabY_,*tabT_,"Trajectories (XxYxT)");

  if(N_N2R != 0) gPosXYT.set_style("points").plot_xyz(vecN2RX,vecN2RY,vecN2RT,"Nodes to reach");
  if(N_Reached != 0) gPosXYT.set_style("points").plot_xyz(vecNReachedX,vecNReachedY,vecNReachedT,"Reached nodes");
  if(N_Used) gPosXYT.set_style("points").plot_xyz(vecNUsedX,vecNUsedY,vecNUsedT,"Used nodes");

  wait_for_key();
}

/*
 *  \brief display in a gnuplot window the status of each node of the grid
 */
void InitialGuess::drawGPGrid()
{
  // get back the lists of explored, reached and used nodes
  std::list<N2R*> lN2R = grid_->listN2R();
  std::list<N2R*> lNReached = grid_->listNReached();
  std::list<N2R*> lNUsed = grid_->listNUsed();

//   cout << "Nodes reached : " << endl;
//   grid_->displayLReached();

  int N_N2R = lN2R.size();
  std::vector<double> vecN2RX(N_N2R);
  std::vector<double> vecN2RY(N_N2R);
  std::vector<double> vecN2RT(N_N2R);

  std::list<N2R*>::iterator it1;
  int i = 0;
  for(it1=lN2R.begin(); it1 != lN2R.end(); it1++)
    {
      PositionXYT pos = (*it1)->goalST().posXYT();
      vecN2RX[i] = pos.x();
      vecN2RY[i] = pos.y();
      vecN2RT[i] = pos.t();
      i++;
    }

  int N_Reached = lNReached.size();
  std::vector<double> vecNReachedX(N_Reached);
  std::vector<double> vecNReachedY(N_Reached);
  std::vector<double> vecNReachedT(N_Reached);

  std::list<N2R*>::iterator it2;
  int j = 0;
  for(it2=lNReached.begin(); it2 != lNReached.end(); it2++)
    {
      PositionXYT pos = (*it2)->goalST().posXYT();
      vecNReachedX[j] = pos.x();
      vecNReachedY[j] = pos.y();
      vecNReachedT[j] = pos.t();
      j++;
    }

  int N_Used = lNUsed.size();
  std::vector<double> vecNUsedX(N_Used);
  std::vector<double> vecNUsedY(N_Used);
  std::vector<double> vecNUsedT(N_Used);

  std::list<N2R*>::iterator it3;
  int k = 0;
  for(it3=lNUsed.begin(); it3 != lNUsed.end(); it3++)
    {
      PositionXYT pos = (*it3)->goalST().posXYT();
      vecNReachedX[k] = pos.x();
      vecNReachedY[k] = pos.y();
      vecNReachedT[k] = pos.t();
      k++;
    }


  int NbP = 40;

  // reinit the tables containing the coordinates of trajectories
  if(tabX_ != NULL) delete tabX_;
  if(tabY_ != NULL) delete tabY_;
  if(tabV_ != NULL) delete tabV_;
  if(tabT_ != NULL) delete tabT_;
  tabX_ = new vector<double>;
  tabY_ = new vector<double>;
  tabV_ = new vector<double>;
  tabT_ = new vector<double>;
  
  // get back the list of reached nodes
  std::list<N2R*> lReached = grid_->listNReached();

  // sample all generated trajectories
  int N = lReached.size();
  std::list<N2R*>::iterator it = lReached.begin();

  for(int i=0; i<N; i++)
    {
      // check if a trajectory has been found between the state-times of this sample
      if(*it != NULL && (*it)->params() != NULL)
	{
	  // sample the trajectory defined by the parameterized control
	  double tabX[NbP];
	  double tabY[NbP];
	  double tabT[NbP];
	  this->sampleTrajNode(*it,tabX,tabY,tabT,NbP);

	  // 	  cout << "last node of the traj : (" << tabX[NbP-1] << "," << tabY[NbP-1] << "," << tabT[NbP-1] << ")" << endl;

	  // append these tables to the list of all points of the trajectories
	  for(int j=0; j<NbP; j++)
	    {
	      tabX_->push_back(tabX[j]);
	      tabY_->push_back(tabY[j]);
	      tabT_->push_back(tabT[j]);
	    }
	  for(int j=NbP-1; j>=0; j--)
	    {
	      tabX_->push_back(tabX[j]);
	      tabY_->push_back(tabY[j]);
	      tabT_->push_back(tabT[j]);
	    }
	}

      it++;
    }

  Gnuplot gPosXYT;
  gPosXYT.reset_plot();
  gPosXYT.cmd("set linestyle 1");
//   gPosXYT.set_style("lines").plot_xyz(*tabX_,*tabY_,*tabT_,"Trajectories (XxYxT)");
  if(N_N2R != 0) gPosXYT.set_style("points").plot_xyz(vecN2RX,vecN2RY,vecN2RT,"Nodes to reach");
  if(N_Reached != 0) gPosXYT.set_style("points").plot_xyz(vecNReachedX,vecNReachedY,vecNReachedT,"Reached nodes");
//   if(N_Used != 0) gPosXYT.set_style("points").plot_xyz(vecNUsedX,vecNUsedY,vecNUsedT,"Used nodes");

  wait_for_key();
}
