#include "WCTrajGen.hpp"
using namespace mpc;

#include <stdio.h>
#include <math.h>
#include <vector>
#include "utilsLuis.h"
#include <complex>

#include "InvMatrix.hpp"
#include "gnuplot/inc/gnuplot_i.hpp"

#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;

// Parameters
#define B      (*vecParameters_p)[0]
#define C      (*vecParameters_p)[1]
#define D      (*vecParameters_p)[2]
#define B2     (*vecParameters_p)[3]
#define C2     (*vecParameters_p)[4]
#define D2     (*vecParameters_p)[5]
#define T      (*vecParameters_p)[6]

// States 
#define X0     vecState_x0[0]
#define Y0     vecState_x0[1]
#define TH0    vecState_x0[2]
#define V0     vecState_x0[3]
#define OM0    vecState_x0[4]
#define T0     vecState_x0[5]

#define Xf     vecState_xf[0]
#define Yf     vecState_xf[1]
#define THf    vecState_xf[2]
#define Vf     vecState_xf[3]
#define OMf    vecState_xf[4]
#define Tf     vecState_xf[5]

#define Xsim   vecState_xsim[0]
#define Ysim   vecState_xsim[1]
#define THsim  vecState_xsim[2]
#define Vsim   vecState_xsim[3]
#define OMsim  vecState_xsim[4]
#define Tsim   vecState_xsim[5]

#define Xind   matStateInd_[0]
#define Yind   matStateInd_[1]
#define THind  matStateInd_[2]
#define Vind   matStateInd_[3]
#define OMind  matStateInd_[4]
#define Tind   matStateInd_[5]

#define Xerr   vecConstraints_g[Xind]
#define Yerr   vecConstraints_g[Yind]
#define THerr  vecConstraints_g[THind]
#define Verr   vecConstraints_g[Vind]
#define OMerr  vecConstraints_g[OMind]

#define DFLT_DERIV_VAL -999999999.99

                                    /**************************************************/
                                    /***************** Class WCTrajGen ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 */
WCTrajGen::WCTrajGen()
{
  xLength = 6;       // State length (x, y, th, v, om, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 5;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, v(t)=v_f, om(t)=om_f)
  dLength = 4;       // Dynamic Constraints length (maxAcc, maxAAcc, maxVel, maxAVel)

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);

  posX = NULL;
  posY = NULL;
  timeT = NULL;
  velV = NULL;
  avelOM = NULL;

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  fctAcc_ = NULL;
  fctVel_ = NULL;
  fctAVel_ = NULL;
  fctAAcc_ = NULL;
  fctTH_ = NULL;
  fctAccWB_ = NULL;
  fctVelWB_ = NULL;
  fctAVelWB_ = NULL;
  fctAAccWB_ = NULL;
  fctTHWB_ = NULL;

  xyFixed_ = true;
  thFixed_ = true;
  vFixed_ = true;
  omFixed_ = true;

  USE_VEL_CST = true;
  USE_ACC_CST = true;
  USE_AVEL_CST = true;
  USE_AACC_CST = true;

  matStateInd_ = new int[xLength];

//   canMoveBack_ = false;

  WCweights_ = WCState(1,1,1,1,1);
}

/**
 *  \brief init constructor
 *  \param sys : the constraints of the wheelchair system
 */
WCTrajGen::WCTrajGen(const WheelChairSys* sys) : TrajGen(sys)
{
  xLength = 6;       // State length (x, y, th, v, om, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 5;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, v(t)=v_f, om(t)=om_f)
  dLength = 4;       // Dynamic Constraints length (maxAcc, maxAAcc, maxVel, maxAVel)

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);

  posX = NULL;
  posY = NULL;
  timeT = NULL;
  velV = NULL;
  avelOM = NULL;

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  fctAcc_ = NULL;
  fctVel_ = NULL;
  fctAVel_ = NULL;
  fctAAcc_ = NULL;
  fctTH_ = NULL;
  fctAccWB_ = NULL;
  fctVelWB_ = NULL;
  fctAVelWB_ = NULL;
  fctAAccWB_ = NULL;
  fctTHWB_ = NULL;

  xyFixed_ = true;
  thFixed_ = true;
  vFixed_ = true;
  omFixed_ = true;

  USE_VEL_CST = true;
  USE_ACC_CST = true;
  USE_AVEL_CST = true;
  USE_AACC_CST = true;

  matStateInd_ = new int[xLength];

  WCweights_ = WCState(1,1,1,1,1);
}

/**
 *  \brief destructor
 */  
WCTrajGen::~WCTrajGen()
{
  if(posX != NULL) delete posX;
  if(posY != NULL) delete posY;
  if(timeT != NULL) delete timeT;
  if(velV != NULL) delete velV;
  if(avelOM != NULL) delete avelOM;  

  if(fctAcc_ != NULL) delete fctAcc_;
  if(fctVel_ != NULL) delete fctVel_;
  if(fctAVel_ != NULL) delete fctAVel_;
  if(fctAAcc_ != NULL) delete fctAAcc_;
  if(fctTH_ != NULL) delete fctTH_;
  
  if(fctAccWB_ != NULL) delete fctAccWB_;
  if(fctVelWB_ != NULL) delete fctVelWB_;
  if(fctAVelWB_ != NULL) delete fctAVelWB_;
  if(fctAAccWB_ != NULL) delete fctAAccWB_;
  if(fctTHWB_ != NULL) delete fctTHWB_;

  if(matStateInd_ != NULL) delete [] matStateInd_;
}


/*************** Getters and Setters **************/

// return true if linear velocity constraints are used
bool WCTrajGen::getUseVelCst()
{
  return this->USE_VEL_CST;
}

// set true if linear velocity constraints are used
void WCTrajGen::setUseVelCst(bool u)
{
  this->USE_VEL_CST = u;
}

// return true if linear acceleration constraints are used
bool WCTrajGen::getUseAccCst()
{
  return this->USE_ACC_CST;
}

// set true if linear acceleration constraints are used
void WCTrajGen::setUseAccCst(bool u)
{
  this->USE_ACC_CST = u;
}

// return true if angular velocity constraints are used
bool WCTrajGen::getUseAVelCst()
{
  return this->USE_AVEL_CST;
}

// set true if angular velocity constraints are used
void WCTrajGen::setUseAVelCst(bool u)
{
  this->USE_AVEL_CST = u;
}

// return true if angular acceleration constraints are used
bool WCTrajGen::getUseAAccCst()
{
  return this->USE_AACC_CST;
}

// set true if angular acceleration constraints are used
void WCTrajGen::setUseAAccCst(bool u)
{
  this->USE_AACC_CST = u;
}

/**
 *  \brief fix or free the final position
 *  \param fixed : true if the final position must be fixed
 */
void WCTrajGen::fixFinalXY(bool fixed)
{
  xyFixed_ = fixed;
}

/**
 *  \brief fix or free the final heading
 *  \param fixed : true if the final heading must be fixed
 */
void WCTrajGen::fixFinalTH(bool fixed)
{
  thFixed_ = fixed;
}

/**
 *  \brief fix or free the final linear velocity
 *  \param fixed : true if the final linear velocity must be fixed
 */
void WCTrajGen::fixFinalV(bool fixed)
{
  vFixed_ = fixed;
}

/**
 *  \brief fix or free the final angular velocity
 *  \param fixed : true if the final angular velocity must be fixed
 */
void WCTrajGen::fixFinalOM(bool fixed)
{
  omFixed_ = fixed;
}


/********* Optimization related routines **********/

/**
 *  \brief compute a default initial guess of parameters
 */
void WCTrajGen::ComputeDfltIG()
{
  //Inital guess of parameters
  double dist = sqrt(pow(Xf-X0,2) + pow(Yf-Y0,2));
  double dir = atan2(Yf-Y0,Xf-X0);
  double difDir = computeMinDiffAngle(normalizedAngle(dir),normalizedAngle(TH0));

  if(DBG_ >= 1) {
    cout << "Tf : " << Tf << endl;
    cout << "T0 : " << T0 << endl;
  }

  T = Tf - T0;

  if(T <= 0)
    {
      B = 0;
      C = 0;
      D = 0;
      B2 = 0;
      C2 = 0;
      D2 = 0;
    }
  else
    {
      double coefDist = 1.0 + (0.5*M_PI - 1)*(fabs(difDir)/M_PI);
      double vmoy = (dist*coefDist)/T;
	
      B = (-2.0*OM0 + 0.2*difDir) / T;
      C = 0.5*OM0 / (T*T); // c
      D = 0; // d
	
      if(DBG_ >= 1) {
	printf ("Initial Guess Parameters K = a=%f,b=%f,c=%f,d=%f,t=%f)\n", 
		OM0,B,C,D,T);
      }
	
      B2 = 2.0*(vmoy - V0) / T;
      C2 = 0.0;
      D2 = 0.0;
	
      if(DBG_ >= 1) {    
	printf ("Initial Guess Parameters V = a=%f,b=%f,c=%f,d=%f,t=%f)\n", 
		V0,B2,C2,D2,T);
      }
    }
}

/**
 *  \brief select an initial guess of parameters from an IG database
 *  \return 0 if a set of parameters has been found, -1 else
 */
int WCTrajGen::ComputeIGfromDB()
{
  //       cout << "An initial guess database has been found" << endl;
  WCStateT initST(X0,Y0,TH0,V0,OM0,T0);
  WCStateT goalST(Xf,Yf,THf,Vf,OMf,Tf);
  const double* param = ((WCInitialGuess*)ig_)->queryIG(initST,goalST);
  
  if(param != NULL)
    {
      B = param[0];
      C = param[1];
      D = param[2];
      B2 = param[3];
      C2 = param[4];
      D2 = param[5];
      T = param[6];
      
      //delete [] param;
      return 0;
    }
  else return -1;
}

/**
 *  \brief compute the error between the state reached and the state we want to reach
 */
void WCTrajGen::ComputeG()
{
  if(xyFixed_) Xerr  = Xsim - Xf;
  if(xyFixed_) Yerr  = Ysim - Yf;
  if(thFixed_) THerr = computeMinDiffAngle(normalizedAngle(THsim),normalizedAngle(THf));
  if(vFixed_) Verr  = Vsim - Vf;
  if(omFixed_) OMerr = OMsim - OMf;
}

/**
 *  \brief compute the partial derivatives of the state wrt. the parameters by finite differences
 */
void WCTrajGen::ComputeJacobianG_FD()
{
}

/**
 *  \brief compute the partial derivatives of the state wrt. the parameters
 */
void WCTrajGen::ComputeJacobianS()
{
  if(vFixed_)
    {
      // computation of dV/dP
      matJacobianG[Vind][0] = 0.0; // dV/dB
      matJacobianG[Vind][1] = 0.0; // dV/dC
      matJacobianG[Vind][2] = 0.0; // dV/dD
      matJacobianG[Vind][3] = T; // dV/dB2
      matJacobianG[Vind][4] = T*T; // dV/dC2
      matJacobianG[Vind][5] = T*T*T; // dV/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[Vind][6] = funcAcc(T) * T; // dV/T
    }

  if(omFixed_)
    {
      // computation of dOM/dP
      matJacobianG[OMind][0] = T; // dOM/dB
      matJacobianG[OMind][1] = T*T; // dOM/dC
      matJacobianG[OMind][2] = T*T*T; // dOM/dD
      matJacobianG[OMind][3] = 0.0; // dOM/dB2
      matJacobianG[OMind][4] = 0.0; // dOM/dC2
      matJacobianG[OMind][5] = 0.0; // dOM/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[OMind][6] = funcAAcc(T) * T; // dOM/dT
    }

  if(thFixed_)
    {
      // computation of dTH/dP
      matJacobianG[THind][0] = 0.5*T*T; // dTH/dB
      matJacobianG[THind][1] = 0.333333333*T*T*T; // dTH/dC
      matJacobianG[THind][2] = 0.25*T*T*T*T; // dTH/dD
      matJacobianG[THind][3] = 0.0; // dTH/dB2
      matJacobianG[THind][4] = 0.0; // dTH/dC2
      matJacobianG[THind][5] = 0.0; // dTH/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[THind][6] = funcAVel(T) * T; // dTH/dT
    }
  
  if(xyFixed_)
    {
      matJacobianG[Xind][0] = qgaus(&WCTrajGen::itgdXdB,0,T); // dX/dB
      matJacobianG[Xind][1] = qgaus(&WCTrajGen::itgdXdC,0,T); // dX/dC
      matJacobianG[Xind][2] = qgaus(&WCTrajGen::itgdXdD,0,T); // dX/dD
      matJacobianG[Xind][3] = qgaus(&WCTrajGen::itgdXdB2,0,T); // dX/dB2
      matJacobianG[Xind][4] = qgaus(&WCTrajGen::itgdXdC2,0,T); // dX/dC2
      matJacobianG[Xind][5] = qgaus(&WCTrajGen::itgdXdD2,0,T); // dX/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[Xind][6] = funcItgX(T) * T; // dX/dT
      
      matJacobianG[Yind][0] = qgaus(&WCTrajGen::itgdYdB,0,T);  // dY/dB
      matJacobianG[Yind][1] = qgaus(&WCTrajGen::itgdYdC,0,T);  // dY/dC
      matJacobianG[Yind][2] = qgaus(&WCTrajGen::itgdYdD,0,T);  // dY/dD
      matJacobianG[Yind][3] = qgaus(&WCTrajGen::itgdYdB2,0,T); // dY/dB2
      matJacobianG[Yind][4] = qgaus(&WCTrajGen::itgdYdC2,0,T); // dY/dC2
      matJacobianG[Yind][5] = qgaus(&WCTrajGen::itgdYdD2,0,T); // dY/dD2
      if(tcStatus() != TIME_FIXED) matJacobianG[Yind][6] = funcItgY(T) * T; // dY/dT
    }
}

/**
 *  \brief compute the second order partial derivatives of the state wrt. the parameters
 */
void WCTrajGen::ComputeHessianS()
{
  cerr << "WCTrajGen::ComputeHessianS : Not implemented yet!" << endl;
  exit(1);
}

double WCTrajGen::itgdXdB(const double t) const
{
  return -0.5*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t;
}

double WCTrajGen::itgdXdC(const double t) const
{
  return -0.333333333*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t;
}

double WCTrajGen::itgdXdD(const double t) const
{
  return -0.25*(V0 + B2*t + C2*t*t + D2*t*t*t)* sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t*t;
}

double WCTrajGen::itgdXdB2(const double t) const
{
  return t * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double WCTrajGen::itgdXdC2(const double t) const
{
  return t*t * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double WCTrajGen::itgdXdD2(const double t) const
{
  return t*t*t * cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double WCTrajGen::itgdYdB(const double t) const
{
  return 0.5*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t;
}

double WCTrajGen::itgdYdC(const double t) const
{
  return 0.333333333*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t;
}

double WCTrajGen::itgdYdD(const double t) const
{
  return 0.25*(V0 + B2*t + C2*t*t + D2*t*t*t)* cos(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t) *t*t*t*t;
}

double WCTrajGen::itgdYdB2(const double t) const
{
  return t * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double WCTrajGen::itgdYdC2(const double t) const
{
  return t*t * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

double WCTrajGen::itgdYdD2(const double t) const
{
  return t*t*t * sin(TH0 + OM0*t + 0.5*B*t*t + 0.333333333*C*t*t*t + 0.25*D*t*t*t*t);
}

// double fct1(const double t) const;
// double fct2(const double t) const;

/**
 *  \brief compute the residual between the goal state and the stae reached
 *  \return the residual between the goal state and the stae reached
 */
double WCTrajGen::ComputeResidual() const
{
  double sqRes = 0;

  if(xyFixed_)
    {
      sqRes += (Xerr * Xerr);
      sqRes += (Yerr * Yerr);
    }
  if(thFixed_) sqRes += (50.0 * (THerr * THerr));
  if(vFixed_) sqRes += (10.0 * (OMerr * OMerr));
  if(omFixed_) sqRes += (10.0 * (Verr * Verr));
  
  return sqrt(sqRes);
}


/**************** Quadrature tools ****************/
  
/**
 *  \brief Quadrature tools...
 */
double WCTrajGen::qgaus(double (WCTrajGen::*func)(const double) const, const double a, const double b) const
{
  static const int n = 20;
  double x[n];
  double w[n];
  int j;
  double s;
  
  gauleg(a, b, x, w, n);      
  s=0;
  for (j=0;j<n;j++) {        
    s += w[j]*((this->*func)(x[j]));
    }
  return s;
}


/*********** State profiles computations **********/

/**
 *  \brief compute the linear acceleration profile
 */
void WCTrajGen::computeFctAcc()
{
  // delete the previous acceleration function
  if(fctAccWB_ != NULL) delete fctAccWB_;
  if(fctAcc_ != NULL) delete fctAcc_;

  // create the polynomial acceleration function
  fctAccWB_ = new Polynomial(3.0*D2,2.0*C2,B2);
  fctAcc_ = new MultiItvFunction(fctAccWB_, 0,INFINITY);

  if(USE_ACC_CST)
    {
      // add the bounds over the acceleration
      fctAcc_->addMinBound(-maxAcc);
      fctAcc_->addMaxBound(maxAcc);
    }
}

/**
 *  \brief compute the linear velocity profile
 */
void WCTrajGen::computeFctVel()
{
  // delete the previous velocity function
  if(fctVel_ != NULL) delete fctVel_;

  // compute the velocity function by integrating the acceleration
  fctVel_ = (MultiItvFunction*) fctAcc_->integration(0.0,V0);
  
  if(USE_VEL_CST)
    {
      if(sys()->canMoveBack()) fctVel_->addMinBound(-maxVel);
      else fctVel_->addMinBound(0.0);
      fctVel_->addMaxBound(maxVel);
    }
  else if(!(sys()->canMoveBack())) fctVel_->addMinBound(0.0);
  
  if(DBG_ >= 1) {
    // recompute the acceleration if the velocity changed
    delete fctAcc_;
    fctAcc_ = (MultiItvFunction*) fctVel_->derivation();
  }
}

/**
 *  \brief compute the angular acceleration profile
 */
void WCTrajGen::computeFctAAcc()
{
  // delete the previous acceleration function
  if(fctAAccWB_ != NULL) delete fctAAccWB_;
  if(fctAAcc_ != NULL) delete fctAAcc_;

  // create the polynomial acceleration function
  fctAAccWB_ = new Polynomial(3.0*D,2.0*C,B);
  fctAAcc_ = new MultiItvFunction(fctAAccWB_,0,INFINITY);

  // add the bounds over the angular acceleration
  if(USE_AACC_CST)
    {
      fctAAcc_->addMinBound(-maxAAcc);
      fctAAcc_->addMaxBound(maxAAcc);
    }
}

/**
 *  \brief compute the angular velocity profile
 */
void WCTrajGen::computeFctAVel()
{
  // delete the previous velocity function
  if(fctAVel_ != NULL) delete fctAVel_;

  // compute the velocity function by integrating the acceleration
  fctAVel_ = (MultiItvFunction*) fctAAcc_->integration(0.0,OM0);
  
  if(USE_AVEL_CST)
    {
      fctAVel_->addMinBound(-maxAVel);
      fctAVel_->addMaxBound(maxAVel);
    }
  
  if(DBG_ >= 1) {
    // recompute the acceleration if the velocity changed
    delete fctAAcc_;
    fctAAcc_ = (MultiItvFunction*) fctAVel_->derivation();
  }
}

/**
 *  \brief compute the heading profile
 */
void WCTrajGen::computeFctTH()
{
  // delete the previous heading profile
  if(fctTH_ != NULL) delete fctTH_;

  // compute the heading profile by integrating the angular velocity
  fctTH_ = ((MultiItvFunction*) (fctAVel_->integration(0.0,TH0)));
}

/**
 *  \brief compute all profiles
 */
void WCTrajGen::computeAllFct()
{
  computeFctAcc();
  computeFctVel();
  computeFctAAcc();
  computeFctAVel();
  computeFctTH();

//   cout << "functions : " << endl;
//   this->displayAllFct();
}

  
/********** State Parameters Computation **********/  

/**
 *  \brief get the value of the linear acceleration profile of the robot at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the linear acceleration profile of the robot at time t
 */
double WCTrajGen::funcAcc(const double t) const
{
  return fctAcc_->getVal(t);
}

/**
 *  \brief get the value of the angular acceleration profile of the robot at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular acceleration profile of the robot at time t
 */
double WCTrajGen::funcAAcc(const double t) const
{
  return fctAAcc_->getVal(t);
}

/**
 *  \brief get the value of the linear velocity profile of the robot at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the linear velocity profile of the robot at time t
 */
double WCTrajGen::funcVel(const double t) const
{
  return fctVel_->getVal(t);
}

/**
 *  \brief get the value of the angular velocity profile of the robot at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular velocity profile of the robot at time t
 */
double WCTrajGen::funcAVel(const double t) const
{
  return fctAVel_->getVal(t);
}

/**
 *  \brief get the value of the heading profile of the robot at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the heading profile of the robot at time t
 */
double WCTrajGen::funcTheta(const double t) const
{
  return fctTH_->getVal(t);
}

/**
 *  \brief integration function for the position X profile computation
 *  \param t : the time at which the function is evaluated
 *  \return the value at time t of the integration function
 */  
double WCTrajGen::funcItgX(const double t) const
{
  return funcVel(t) * cos(normalizedAngle(funcTheta(t)));
}

/**
 *  \brief integration function for the position Y profile computation
 *  \param t : the time at which the function is evaluated
 *  \return the value at time t of the integration function
 */  
double WCTrajGen::funcItgY(const double t) const
{
  return funcVel(t) * sin(normalizedAngle(funcTheta(t)));
}

/**
 *  \brief get the value of the position X profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the position X profile at time t
 */
double WCTrajGen::funcX(const double t) const
{
  return X0 + qgaus(&WCTrajGen::funcItgX,0,t);
}

/**
 *  \brief get the value of the position Y profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the position Y profile at time t
 */
double WCTrajGen::funcY(const double t) const
{
  return Y0 + qgaus(&WCTrajGen::funcItgY,0,t);
}


/*************** Forward Solution *****************/
  
/**
 *  \brief compute the state reached by the current set of parameters
 */
void WCTrajGen::ForwardSolution()
{
  for(int i=0; i<gLength; i++)
//     for(int j=0; j<pLength-1; j++)
    for(int j=0; j<pLengthFree(); j++)
      matJacobianG[i][j] = DFLT_DERIV_VAL;
  
  // compute the different functions' profiles
  computeAllFct();

  // compute the simulated state
  Xsim  = funcX(T);
  Ysim  = funcY(T);
  THsim = funcTheta(T);
  Vsim  = funcVel(T);
  OMsim = funcAVel(T);
  if(tcStatus() != TIME_FIXED) Tsim  = T;
}


/***************** Initialization *****************/

/**
 *  \brief set the initial guess of parameters
 *  \param b,c,d : coefficients of the polynomial angular acceleration profile
 *  \param b2,c2,d2 : coefficients of the polynomial linear acceleration profile
 *  \param t : final time to compute the trajectory (time of the final state)
 */
void WCTrajGen::setInitialGuess(double b,double c,double d,double b2,double c2,double d2,double t)
{
  vecParameters_p = new Vec_DP(pLength);

  B = b;
  C = c;
  D = d;
  B2 = b2;
  C2 = c2;
  D2 = d2;
  T = t;
}

// /**
//  *  \brief set the initial state, final state and constraints vectors
//  *  \param x0 : initial state
//  *  \param xf : final state
//  *  \param constraints : constraints defined over the system
//  *
//  *  constraints[0] : max linear acceleration
//  *  constraints[1] : max angular acceleration
//  *  constraints[2] : max linear velocity
//  *  constraints[3] : max angular velocity
//  */
// void WCTrajGen::setInitialParam(Vec_DP& x0,Vec_DP& xf,Vec_DP& constraints)
// {
//   vecState_x0 = x0;
//   vecState_xf = xf;
//   maxAcc = constraints[0];
//   maxAAcc = constraints[1];
//   maxVel = constraints[2];
//   maxAVel = constraints[3];
// }

/**
 *  \brief set the system parameters
 *  \param constraints : constraints defined over the system
 */
void WCTrajGen::setSysParams(const System* constraints)
{
  if(constraints->getType() != WHEELCHAIR_SYS && constraints->getType() != MANA_SYS)
    {
      throw string("WCTrajGen::setSysParams : Wrong system type!");
    }

  maxAcc = ((WheelChairSys*)constraints)->getAMax();
  maxVel = ((WheelChairSys*)constraints)->getVMax();
  maxAAcc = ((WheelChairSys*)constraints)->getAAMax();
  maxAVel = ((WheelChairSys*)constraints)->getAVMax();
}

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void WCTrajGen::setInitialGuess(const double* p)
{
  this->setInitialGuess(p[0],p[1],p[2],p[3],p[4],p[5],p[6]);
}

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void WCTrajGen::setInitialConfig(const double* x0)
{
  // Set initial and final states
  vecState_x0 = Vec_DP(xLength);
  vecState_x0[0] = x0[0];
  vecState_x0[1] = x0[1];
  vecState_x0[2] = x0[2];
  vecState_x0[3] = x0[3];
  vecState_x0[4] = x0[4];
  vecState_x0[5] = x0[5];
  
  vecState_xf = vecState_x0;
}


/************* Trajectory Generation **************/

/**
 *  \brief init a vector of the constraints defined over the system
 *  \param sys : the robotic system features
 *  \param vcstt : the resulting vector of constraints
 */
void WCTrajGen::initCstt(System* sys,Vec_DP& vcstt)
{
  vcstt[0] = ((WheelChairSys*)sys)->getAMax();
  vcstt[1] = ((WheelChairSys*)sys)->getAAMax();
  vcstt[2] = ((WheelChairSys*)sys)->getVMax();
  vcstt[3] = ((WheelChairSys*)sys)->getAVMax();
}

/**
 *  \brief init the fixed state parameters
 *  \return the size of the vector of free parameters
 */
int WCTrajGen::initFixedSParam()
{
  int cpt = 0;

  if(xyFixed_)
    {
      Xind = 0;
      Yind = 1;
      cpt += 2;
    }
  if(thFixed_)
    {
      THind = cpt;
      cpt++;
    }
  if(vFixed_)
    {
      Vind = cpt;
      cpt++;
    }
  if(omFixed_)
    {
      OMind = cpt;
      cpt++;
    }

  return cpt;
}

/**
 *  \brief generate a trajectory by constraints optimiation
 *  \param s0 : initial state
 *  \param sf : final state
 *  \param parameters : initial guess of parameters
 */
int WCTrajGen::ComputeTrajectory(const WCStateT& s0,const WCStateT& sf,double** parameters)
{
  Vec_DP x0(6);
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.s();
  x0[4] = s0.om();
  x0[5] = 0;
  
  Vec_DP xf(6);
  xf[0] = sf.x();
  xf[1] = sf.y();
  xf[2] = sf.th();
  xf[3] = sf.s();
  xf[4] = sf.om();
  xf[5] = sf.t() - s0.t();

  Vec_DP** param = NULL;
  if(parameters == NULL) param = NULL;
  else
    {
      param = new Vec_DP*;
      (*param) = new Vec_DP(7);
      (**param)[0] = (*parameters)[0];
      (**param)[1] = (*parameters)[1];
      (**param)[2] = (*parameters)[2];
      (**param)[3] = (*parameters)[3];
      (**param)[4] = (*parameters)[4];
      (**param)[5] = (*parameters)[5];
      (**param)[6] = (*parameters)[6];
    }

  int res = this->TrajGen::ComputeTrajectory(x0,xf,param);

  if(parameters == NULL)
    {
      parameters = new double*;
      *parameters = new double[7];
    }

//   if(param != NULL)
//     {
//       if((*param) != NULL) delete (*param);
//       delete param;
//     }

  return res;
}

/**
 *  \brief generate a trajectory by constraints optimiation
 *  \param s0 : initial state
 *  \param sf : final state
 *  \param parameters : initial guess of parameters
 *  \return 0 if a solution has been found, -1 else
 */
int WCTrajGen::ComputeTrajectory(const WCStateT& s0,const WCStateT& sf,Vec_DP** parameters)
{
  Vec_DP x0(6);
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.s();
  x0[4] = s0.om();
  x0[5] = 0;
  
  Vec_DP xf(6);
  xf[0] = sf.x();
  xf[1] = sf.y();
  xf[2] = sf.th();
  xf[3] = sf.s();
  xf[4] = sf.om();
  xf[5] = sf.t() - s0.t();

//   Vec_DP constraints(4);
//   constraints[0] = sys->getAMax();
//   constraints[1] = sys->getAAMax();
//   constraints[2] = sys->getVMax();
//   constraints[3] = sys->getAVMax();

  int res = this->TrajGen::ComputeTrajectory(x0,xf,parameters);
  return res;
}

/**
 *  \brief generate a trajectory by constraints optimiation
 *  \param s0 : initial state
 *  \param param : initial guess of parameters
 */
void WCTrajGen::ComputeFromIG(const WCStateT& s0,const double* param)
{
  double x0[6];
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.s();
  x0[4] = s0.om();
  x0[5] = 0;

  this->TrajGen::ComputeFromIG(x0,param);
}

/**
 *  \brief generate a trajectory by constraints optimiation
 *  \param statet : initial state-time
 *  \param ctrlt : control-time applied on the system
 */
void WCTrajGen::ComputeFromCtrl(const StateTime& statet,const ControlTime& ctrlt)
{
  WCStateT wcST(*((WCStateT*)&statet));
  WCControlT wcCT(*((WCControlT*)&ctrlt));

  double x0[6];
  x0[0] = wcST.x();
  x0[1] = wcST.y();
  x0[2] = wcST.th();
//   x0[3] = wcST.s();
//   x0[4] = wcST.om();
  x0[3] = wcCT.v();
  x0[4] = wcCT.av();
  x0[5] = 0;

  double param[7];
  param[0] = 0.0;
  param[1] = 0.0;
  param[2] = 0.0;
  param[3] = 0.0;
  param[4] = 0.0;
  param[5] = 0.0;
  param[6] = wcCT.t();

  this->TrajGen::ComputeFromIG(x0,param);
}

/**
 *  \brief generate a trajectory by constraints optimiation
 *  \param statet : initial state-time
 *  \param ctrlt : control-time applied on the system
 */
void WCTrajGen::ComputeFromCtrl(const WCStateT& statet,const WCControlT& ctrlt)
{
  double x0[6];
  x0[0] = statet.x();
  x0[1] = statet.y();
  x0[2] = statet.th();
  x0[3] = ctrlt.v();
  x0[4] = ctrlt.av();
  x0[5] = 0;

  double param[7];
  param[0] = 0.0;
  param[1] = 0.0;
  param[2] = 0.0;
  param[3] = 0.0;
  param[4] = 0.0;
  param[5] = 0.0;
  param[6] = ctrlt.t();

  this->TrajGen::ComputeFromIG(x0,param);
}


/****** Trajectory Generation (2nd version) ******/

/**
 *  \brief init the cot function weights
 */
void WCTrajGen::initWeights()
{
  if(xyFixed_)
    {
      weightsCF_[Xind] = WCweights_.x();
      weightsCF_[Yind] = WCweights_.y();
    }
  if(thFixed_)
    {
      weightsCF_[THind] = WCweights_.th();
    }
  if(omFixed_)
    {
      weightsCF_[OMind] = WCweights_.om();
    }
  if(vFixed_)
    {
      weightsCF_[Vind] = WCweights_.s();
    }
}

/**
 *  \brief generate a trajectory between two states
 *  \param s0 : initial state-time
 *  \param sf : initial state-time
 *  \param p : initial guess of parameters
 */
int WCTrajGen::genTraj(const WCStateT& s0,const WCStateT& sf,const double* p)
{
  double x0[6];
  x0[0] = s0.x();
  x0[1] = s0.y();
  x0[2] = s0.th();
  x0[3] = s0.s();
  x0[4] = s0.om();
  x0[5] = 0;
  
  double xf[6];
  xf[0] = sf.x();
  xf[1] = sf.y();
  xf[2] = sf.th();
  xf[3] = sf.s();
  xf[4] = sf.om();
  xf[5] = sf.t() - s0.t();

  return this->TrajGen::genTraj(x0,xf,p);
}


/**************** Output Functions ****************/

/**
 *  \brief get back the state reached at time t
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,v,om,t)
 */
void WCTrajGen::getStateTReachedAt(double* rstatet,double t) const
{
  rstatet[0] = funcX(t);
  rstatet[1] = funcY(t);
  rstatet[2] = funcTheta(t);
  rstatet[3] = funcVel(t);
  rstatet[4] = funcAVel(t);
  rstatet[5] = t;
}

/**
 *  \brief get back the state reached at time t
 *  \param t : time at which we evaluate the reached state
 *  \result the reached state
 */
WCStateT WCTrajGen::getWCStateTReachedAt(double t) const
{
  return WCStateT(funcX(t),
		  funcY(t),
		  funcTheta(t),
		  funcVel(t),
		  funcAVel(t),
		  t);
}

/**
 *  \brief get back the state reached at time t
 *  \param t : time at which we evaluate the reached state
 *  \result the reached state
 */
ManaStateT WCTrajGen::getManaStateTReachedAt(double t) const
{
  return ManaStateT(funcX(t),
		    funcY(t),
		    funcTheta(t),
		    funcVel(t),
		    funcAVel(t),
		    t);
}

/**
 *  \brief get back the control reached at time t
 *  \param t : time at which we evaluate the reached control
 *  \result the reached control
 */
WCControl WCTrajGen::getWCControlReachedAt(double t) const
{
  return WCControl(funcVel(t),
		   funcAVel(t));
}

/**
 *  \brief get back the control reached at time t
 *  \param t : time at which we evaluate the reached control
 *  \result the reached control
 */
ManaControl WCTrajGen::getManaControlReachedAt(double t) const
{
  return ManaControl(funcVel(t),
		     funcAVel(t));
}

/**
 *  \brief get back the state reached at time t
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,v,om,t)
 */
void WCTrajGen::getFinalStateTReached(double* rstatet) const
{
  rstatet[0] = funcX(T);
  rstatet[1] = funcY(T);
  rstatet[2] = funcTheta(T);
  rstatet[3] = funcVel(T);
  rstatet[4] = funcAVel(T);
  rstatet[5] = T;
}

/**
 *  \brief get back the state reached at time t
 *  \param t : time at which we evaluate the reached state
 *  \result the reached state
 */
WCStateT WCTrajGen::getFinalWCStateTReached() const
{
  return WCStateT(funcX(T),
		  funcY(T),
		  funcTheta(T),
		  funcVel(T),
		  funcAVel(T),
		  T);
}

/**
 *  \brief get back the state reached at time t
 *  \param t : time at which we evaluate the reached state
 *  \result the reached state
 */
ManaStateT WCTrajGen::getFinalManaStateTReached() const
{
  return ManaStateT(funcX(T),
		    funcY(T),
		    funcTheta(T),
		    funcVel(T),
		    funcAVel(T),
		    T);
}

/**
 *  \brief save a sampling of the trajectory profile
 *  \param nbp : number of sampling points
 */
void WCTrajGen::SaveTrajectory(int nbp)
{
  if(posX != NULL) delete posX;
  if(posY != NULL) delete posY;
  if(timeT != NULL) delete timeT;
  if(velV != NULL) delete velV;
  if(avelOM != NULL) delete avelOM;  

  posX = new vector<double>;
  posY = new vector<double>;
  timeT = new vector<double>;
  velV = new vector<double>;
  avelOM = new vector<double>;

  posX->clear();
  posY->clear();
  timeT->clear();
  velV->clear();
  avelOM->clear();

  //Computing trajectory
  double step = 1.0 / (((double)nbp)-1.0);
  for(int j=0;j<nbp;j++)
    {
      if(fctAcc_ != NULL && fctVel_ != NULL && fctAVel_ != NULL && fctAAcc_ != NULL && fctTH_ != NULL)
	{
	  posX->push_back(funcX(j*(step*T)));
	  posY->push_back(funcY(j*(step*T)));
	  timeT->push_back(j*step*T);
	  velV->push_back(funcVel(j*step*T));
	  avelOM->push_back(funcAVel(j*step*T));
	}
      else
	{
	  posX->push_back(0.0);
	  posY->push_back(0.0);
	  timeT->push_back(0.0);
	  velV->push_back(0.0);
	  avelOM->push_back(0.0);
	}
    }
}

/**
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 */
void WCTrajGen::traceOn(Map *map,int nbp)
{
  if(map == NULL)
    {
      cout << "WCTrajGen:traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }
  else
    {
      int pathCol = (int) (map->getDepth() / 2);
      for(int i=0; i<=nbp; i++)
	{
	  double x = funcX(i*(1.0/((double)nbp)*T));
	  double y = funcY(i*(1.0/((double)nbp)*T));

	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
	}
    }
}

/**
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 *  \param initPos : initial position
 */  
void WCTrajGen::traceOn(Map *map,int nbp,const PositionXYTh& initPos)
{
  if(map == NULL)
    {
      cout << "WCTrajGen::traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }
  else
    {
      int pathCol = (int) (map->getDepth() / 2);
      for(int i=0; i<=nbp; i++)
	{
	  double x1 = funcX(i*(1.0/((double)nbp)*T));
	  double y1 = funcY(i*(1.0/((double)nbp)*T));
	  double th1 = funcTheta(i*(1.0/((double)nbp)*T));
	  PositionXYTh pos1(x1,y1,th1);

	  PositionXYTh pos = coordInFrame(pos1,initPos);
	  double x = pos.x();
	  double y = pos.y();

	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
	}
    }
}

/**
 *  \brief return a sample trajectory from the parameterized one
 *  \param ts : the sampling time step
 *  \return a sample trajectory from the parameterized one
 *
 *  A new instance of sample trajectory is returned here
 */
WCSTraj* WCTrajGen::sampledTraj(double ts) const
{
  // compute the number of intervals required
  int nbItv = ((int)(T/ts));
  if(nbItv == 0) nbItv = 1;

  // compute the true time step used
  double trueTS = T/(nbItv);

  // initialize the output trajectory
  WCSTraj* traj = new WCSTraj();

  // compute each state of the trajectory
  for(int i=0; i<=nbItv; i++)
    {
      double t = i*trueTS;
      WCStateT curST(funcX(t),
		     funcY(t),
		     funcTheta(t),
		     funcVel(t),
		     funcAVel(t),
		     t);
      traj->append(curST);
    }
  
  // return the trajectory computed
  return traj;
}

/**
 *  \brief sample the X and Y coordinates of the trajectory into the input tables
 *  \param sizeTabs : size of the input tables
 *  \param tabX : resulting table containing the X coordinates
 *  \param tabY : resulting table containing the Y coordinates
 *  \return 0 if the sampling succeed, -1 else
 */
int WCTrajGen::sampleXY(int sizeTabs,double** tabX, double** tabY) const
{
  // check if the input table size is correct
  if(sizeTabs < 2) return -1;

  // compute the true time step used
  double TS = T/(sizeTabs-1);

  // compute each state of the trajectory
  for(int i=0; i<sizeTabs; i++)
    {
      double t = i*TS;
      
      cout << "i : " << i << "t : " << t << endl; 

      (*tabX)[i] = funcX(t);
      (*tabY)[i] = funcY(t);

      cout << "(*tabX)[i] : " << (*tabX)[i] << "(*tabY)[i] : " << (*tabY)[i] << endl;       
    }
  
  return 0;
}


/*************** Utility Functions ****************/

/**
 *  \brief convert the initial and final input state-times into tables of double
 *  \param initST : the input initial state-time
 *  \param goalST : the input goal state-time
 *  \param initTD : the resulting initial table
 *  \param goalTD : the resulting goal table
 *
 *  Tables must be allocated first
 */
void WCTrajGen::InitGoalST2Tab(const StateTime& initST,const StateTime& goalST,double* initTD,double* goalTD) const
{
  WCStateT s0(*((WCStateT*)(&initST)));
  WCStateT sf(*((WCStateT*)(&goalST)));

  initTD[0] = s0.x();
  initTD[1] = s0.y();
  initTD[2] = s0.th();
  initTD[3] = s0.s();
  initTD[4] = s0.om();
  initTD[5] = 0;
  
  goalTD[0] = sf.x();
  goalTD[1] = sf.y();
  goalTD[2] = sf.th();
  goalTD[3] = sf.s();
  goalTD[4] = sf.om();
  goalTD[5] = sf.t() - s0.t();
}

/**
 *  \brief display both wheels velocity, linear and angular velocity, heading and position profiles
 */
void WCTrajGen::displayProfiles()
{
  // compute the forward solution with the new parameters
  this->ForwardSolution();
  
  //Computing Position
  this->SaveTrajectory(100);
  vector<double> xIGpos, yIGpos;
  xIGpos.push_back(X0);
  yIGpos.push_back(Y0);
  xIGpos.push_back(Xf);
  yIGpos.push_back(Yf);

  Gnuplot gPosition;
  gPosition.reset_plot();
  gPosition.set_style("lines").plot_xy(*(this->posX),*(this->posY),"Position");
  gPosition.set_style("points").plot_xy(xIGpos,yIGpos,"Initial and Final Pos");

  //Computing Linear Acceleration
  vector<double> acc;
  vector<double> accBinf;
  vector<double> accBsup;
  vector<double> distanceacc;
  for(int j=0;j<=50;j++){
    acc.push_back((double)this->funcAcc(j*(T/50)));
    accBinf.push_back(-maxAcc);
    accBsup.push_back(maxAcc);
    distanceacc.push_back((j*(T/50)));
  }    
  Gnuplot gAcc;
  gAcc.reset_plot();
  gAcc.set_style("lines").plot_xy(distanceacc,acc,"Acceleration");
  gAcc.set_style("points").plot_xy(distanceacc,accBinf,"Min Bound");
  gAcc.set_style("points").plot_xy(distanceacc,accBsup,"Max Bound");

  //Computing Angular Acceleration
  vector<double> aacc;
  vector<double> aaccBinf;
  vector<double> aaccBsup;
  vector<double> distanceaacc;
  for(int j=0;j<=50;j++){
    aacc.push_back((double)this->funcAAcc(j*(T/50)));
    aaccBinf.push_back(-maxAAcc);
    aaccBsup.push_back(maxAAcc);
    distanceaacc.push_back((j*(T/50)));
  }    
  Gnuplot gAAcc;
  gAAcc.reset_plot();
  gAAcc.set_style("lines").plot_xy(distanceaacc,aacc,"Angular Acc");
  gAAcc.set_style("points").plot_xy(distanceaacc,aaccBinf,"Min Bound");
  gAAcc.set_style("points").plot_xy(distanceaacc,aaccBsup,"Max Bound");
  
  //Computing Velocity
  vector<double> vel;
  vector<double> velBinf;
  vector<double> velBsup;
  vector<double> distanceV;
  for(int j=0;j<=50;j++){
    vel.push_back((double)this->funcVel(j*(T/50)));
    velBinf.push_back(-maxVel);
    velBsup.push_back(maxVel);
    distanceV.push_back((j*(T/50)));
  }    
  Gnuplot gVel;
  gVel.reset_plot();
  gVel.set_style("lines").plot_xy(distanceV,vel,"Velocity");
  gVel.set_style("points").plot_xy(distanceV,velBinf,"Min Bound");
  gVel.set_style("points").plot_xy(distanceV,velBsup,"Max Bound");  

  //Computing Velocity
  vector<double> avel;
  vector<double> avelBinf;
  vector<double> avelBsup;
  vector<double> distanceAV;
  for(int j=0;j<=50;j++){
    avel.push_back((double)this->funcAVel(j*(T/50)));
    avelBinf.push_back(-maxAVel);
    avelBsup.push_back(maxAVel);
    distanceAV.push_back((j*(T/50)));
  }    
  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(distanceAV,avel,"Angular Vel");
  gAVel.set_style("points").plot_xy(distanceAV,avelBinf,"Min Bound");
  gAVel.set_style("points").plot_xy(distanceAV,avelBsup,"Max Bound");

  //Computing Heading
  vector<double> dir2;
  vector<double> distance2;
  for(int j=0;j<=50;j++){
    double thtmp = (double)this->funcTheta(j*(T/50.0));
    double disttmp = (j*(T/50.0));
    dir2.push_back(thtmp);
    distance2.push_back(disttmp);
  } 
  Gnuplot gTH;
  gTH.reset_plot();
  gTH.set_style("lines").plot_xy(distance2,dir2,"Heading");  

  wait_for_key();
}

/**
 *  \brief display all functions
 */
void WCTrajGen::displayAllFct()
{
  cout << "Linear Acceleration : " << endl;
  fctAcc_->display();
  cout << "Linear Velocity : " << endl;
  fctVel_->display();
  cout << "Angular Acceleration : " << endl;
  fctAAcc_->display();
  cout << "Angular Velocity : " << endl;
  fctAVel_->display();
  cout << "Heading : " << endl;
  fctTH_->display();
}







// /**
//  *  \brief generate a trajectory by constraints optimiation
//  *  \param x0 : initial state
//  *  \param xf : final state
//  *  \param constraints : constraints defined over the system
//  *  \param parameters : initial guess of parameters
//  *
//  *  x0 = (X0,Y0,TH0,VX0,VY0,OM0,T0)
//  *  xf = (Xf,Yf,THf,VXf,VYf,OMf,Tf)
//  *  constraints = (maxAcc,maxAAcc,maxVel,maxAVel)
//  */
// int WCTrajGen::ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP constraints,Vec_DP** parameters)
// {
//   // initialize the correction coefficient
//   double coef_cor = COEF_COR;
//   int nbDeacreaseStep = 0;
//   double oldResidual = 0;

//   // Init the initial best parameters and residual
//   double bestB=0, bestC=0, bestD=0, bestB2=0, bestC2=0, bestD2=0;
//   double bestResidual = -1;
//   int bestCNT = -1;

//   //Set initial state, final state and system constraints
//   setInitialParam(x0,xf,constraints);

//   if(DBG_ >= 1) {
//     cout << "trajGen : (X0=" << X0 << ",Y0=" << Y0 << ",TH0=" << TH0 << ",VX0=" << VX0 <<
//       ",VY0=" << VY0 << ",OM0=" << OM0 << ",T0=" << T0 << ") -> ";
//     cout << "(Xf=" << Xf << ",Yf=" << Yf << ",THf=" << THf << ",VXf=" << VXf <<
//       ",VYf=" << VYf << ",OMf=" << OMf << ",Tf=" << Tf << ")" << endl;
//   }

//   //Heuristic guess
//   if(parameters == NULL || *parameters == NULL)
//     InitialGuess(parameters);
//   else
//     {
//       vecParameters_p = *parameters;
//       T = vecState_xf[6] - vecState_x0[6];
//     }

//   if(DBG_ >= 1) {
//     cout << "Initial Guess :  ";
//     cout << "B:" << B << "  C:" << C << "  D:" << D << "  T:" << T << "  B2:" << B2 << "  C2:" << C2 << "  D2:" << D2 << "  T:" << T << endl;
//   }

//   // Init the number of the algorithm's step and the residual
//   double residual = 1000000;  
//   int cnt=0;
//   CNT_ = 0;
  
//   vecCorrParameters_p = 0;  

//   while(residual > maxResidual_)
//     {
//       // save the best parameters
//       if(CNT_ == 0 || residual < bestResidual)
// 	{
// 	  bestB = B;
// 	  bestC = C;
// 	  bestD = D;
// 	  bestB2 = B2;
// 	  bestC2 = C2;
// 	  bestD2 = D2;
// 	  bestResidual = residual;
// 	  bestCNT = CNT_;
// 	}

// #if VARIABLE_COEF_COR
//       // update the correction coefficient
//       if(residual <= oldResidual)
// 	{
// 	  nbDeacreaseStep++;
	  
// 	  if(nbDeacreaseStep >= 10) coef_cor = 1.0;
// 	  else if(nbDeacreaseStep >= 9) coef_cor = 0.8;
// 	  else if(nbDeacreaseStep >= 8) coef_cor = 0.7;
// 	  else if(nbDeacreaseStep >= 7) coef_cor = 0.6;
// 	  else if(nbDeacreaseStep >= 6) coef_cor = 0.5;
// 	  else if(nbDeacreaseStep >= 5) coef_cor = 0.4;
// 	  else if(nbDeacreaseStep >= 3) coef_cor = 0.3;	
// 	}
//       else
// 	{
// 	  coef_cor = COEF_COR;
// 	  nbDeacreaseStep = 0;
// 	}
// //       coef_cor = 0.1;
      
//       oldResidual = residual;
// #endif

//       //Apply correction
// #if TIME_FIXED
//       for(int i=0; i<pLength-1;i++)
// #else
//       for(int i=0; i<pLength;i++)
// #endif
// 	(*vecParameters_p)[i] += coef_cor*vecCorrParameters_p[i];            
      
//       //MUST_DISPLAY = false;
//       ForwardSolution();
//       //MUST_DISPLAY = false;
//       CNT_++;

//       if(DBG_ >= 1) {
// 	printf ("Simulated state = (Xsim=%f,Ysim=%f,VXsim=%f,VYsim=%f,OMsim=%f)\n", 
// 		Xsim,Ysim,THsim,VXsim,VYsim,OMsim);
//       }

//       ComputeG();

//       if(DBG_ >= 1) {
// 	printf ("Errors = (%f,%f,%f,%f,%f,%f)\n", 
// 		Xerr,Yerr,THerr,VXerr,VYerr,OMerr);
// 	printf ("Parameters K = (OM0=%f,B=%f,C=%f,D=%f,T=%f)\n", 
// 		OM0,B,C,D,T);
// 	printf ("Parameters A = (B2=%f,C2=%f,D2=%f,T=%f)\n", 
// 		B2,C2,D2,T);
//       }

//       //ComputeJacobianG();
//       ComputeJacobianG_FD();
//       ComputeInvJacobianG();

//       //Compute the correction
// #if TIME_FIXED
//       for(int i=0; i<pLength-1;i++)
// #else
// 	for(int i=0; i<pLength;i++)
// #endif
// 	  {
// 	    vecCorrParameters_p[i] = 0;
	    
// 	    for(int j=0;j<gLength;j++)
// 	      {
// 		vecCorrParameters_p[i] += (-1.0)*vecConstraints_g[j]*matInvJacobianG[i][j]; 
// 	      }	    
// 	  }
      
//       if(DBG_ >= 1) {
// 	printf ("Correction = (B=%f,C=%f,D=%f,B2=%f,C2=%f,D2=%f)\n", 
// 		vecCorrParameters_p[0],vecCorrParameters_p[1],vecCorrParameters_p[2],
// 		vecCorrParameters_p[3],vecCorrParameters_p[4],vecCorrParameters_p[5]);
//       }

//       residual = sqrt( pow(vecConstraints_g[0],2) + 
// 		       pow(vecConstraints_g[1],2) +
// 		       pow(vecConstraints_g[2],2) +
// 		       pow(vecConstraints_g[3],2) +
// 		       pow(vecConstraints_g[4],2) + 
// 		       pow(vecConstraints_g[5],2));
      
//       if(DBG_ >= 1) {        
// 	printf ("Residual = %f\t CNT : %d\t COEF_COR : %f\n", 
// 		residual,CNT_,coef_cor); 
//       } 

//       if(DBG_ >= 2) {
// 	displayProfiles();
//       }     

//       cnt++;
//       if(cnt >= maxNbStep_)
// 	{
// 	  if(DBG_ >= 1) {
// 	    cout << "  No traj found in " << maxNbStep_ << "iterations" << endl;
// 	  }
	  
// 	  // get back the best parameters found
// 	  B = bestB;
// 	  C = bestC;
// 	  D = bestD;
// 	  B2 = bestB2;
// 	  C2 = bestC2;
// 	  D2 = bestD2;
	  
// 	  // compute the forward method with these parameters
// 	  ForwardSolution();
// 	  // compute the associate error
// 	  ComputeG();
	  
// 	  // get back errors on X, Y and TH
// 	  finalResidual_ = bestResidual;
// 	  finalNbSteps_ = CNT_;
// 	  finalErrX_ = vecConstraints_g[0];
// 	  finalErrY_ = vecConstraints_g[1];
// 	  finalErrTH_ = vecConstraints_g[2];
	  
// 	  return -1;
// 	}
//     }
//   finalResidual_ = residual;
//   finalNbSteps_ = CNT_;
//   finalErrX_ = vecConstraints_g[0];
//   finalErrY_ = vecConstraints_g[1];
//   finalErrTH_ = vecConstraints_g[2];
//   return 0;
// }
