#ifndef MPC_HUGRIFC_HPP
#define MPC_HUGRIFC_HPP

#include "environment/inc/Map.hpp"
#include "motion/inc/STrajectory.hpp"

#include "hugr/HCommon/inc/HCommon.hpp"
#include "robot/inc/Robot.hpp"
#include "robot/inc/WheelChair.hpp"
#include "robot/inc/Cycab.hpp"

#include "hugr/hugrEnvironment/inc/HEnvironment.hpp"
#include "environment/inc/MapHugr.hpp"

#include "hugr.hpp"
#include "WheelchairStructs.hpp"

#include "planner/inc/TrajPlanner.hpp"

#include "foxExtension/inc/FXGLDynCanvas.hpp"
#include <fx.h>
#include <fx3d.h>
#include <stdio.h>
#include <iostream>
using namespace std;
using namespace mpc;

#include <boost/thread/thread.hpp>
//#include <boost/thread/thread_group.hpp>


class MPC_HugrIfc : public FXMainWindow {
  FXDECLARE(MPC_HugrIfc)

private : 
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////// Application components /////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // environment components
  Map                   *map;                     // current map
//   Map                   *mapThInc;                // map thresolded with increased obstacles

  FXdouble              minFieldX;                // the minimal X boundary of the field
  FXdouble              maxFieldX;                // the maximal X boundary of the field
  FXdouble              minFieldY;                // the minimal Y boundary of the field
  FXdouble              maxFieldY;                // the maximal Y boundary of the field

  FXString              stEnvFilename;            // name of the file containing a static environment to load
  FXString              samplingStEnvFilename;    // name of the file containing the sampling of the static environment

  bool                  dispRoadmap;              // true if the contructed roadmap must be displayed
  bool                  autoBuildRoadmap;         // true if a roadmap must be automatically built when a static env is loaded  
  FXint                 optRoadmap;               // roadmap option : 0 no roadmap used, 1 autobuild roadmap, 2 load roadmap from file
  FXString              RMfilename;               // roadmap filename

  FXdouble              distFilter;               // maximal distance from static obstacles to consider the dynamic obstacles perceived as noise
  FXdouble              securDist;                // security radius kept around dynamic obstacles

  // localisation components
  FXdouble              initX;                    // variable used to set the initial position of the robot
  FXdouble              initY;                    // variable used to set the initial position of the robot
  FXdouble              initTH;                   // variable used to set the initial heading of the robot

  bool                  locActive;                // true if the localization is active

  FXdouble              dtLoc;                    // time interval between two localization requests

  FXString              nameDynObs;               // hugr variable name of the dynamic obstacles

  std::vector<StateTime*>* vecTrajSt;             // vector of state-times of the trajectory that a robot has to follow
  std::vector<StateTime*>* vecLoc;                // vector of recorded localization state-times
  std::vector<StateTime*>* vecDev;                // vector of state deviation recorded during a trajectoy following process
  std::vector<ControlTime*>* vecCtrlT;            // vector of state control-times recorded during a trajectoy following process

  std::vector<std::vector<StateTime*>*> recordsTrajSt;    // records of all vectors of state-times of the trajectory that a robot has to follow
  std::vector<std::vector<StateTime*>*> recordsLoc;       // records of all vectors of recorded localization state-times
  std::vector<std::vector<StateTime*>*> recordsDev;       // records of all vectors of state deviation recorded during a trajectoy following process
  std::vector<std::vector<ControlTime*>*> recordsCtrlT;   // records of all vectors of state control-times recorded during a trajectoy following process

  // robot configuration components
  Robot*                rb_;                      // the robot managed
  system_t              typeSys_;                 // the type of the robotic system handle

  FXdouble              WC_x;                     // position of the wheelchair
  FXdouble              WC_y;                     // position of the wheelchair
  FXdouble              WC_th;                    // heading of the wheelchair
  FXdouble              WC_v;                     // linear velocity of the wheelchair
  FXdouble              WC_av;                    // angular velocity of the wheelchair

  FXdouble              CC_x;                     // position of the carlike
  FXdouble              CC_y;                     // position of the carlike
  FXdouble              CC_th;                    // heading of the carlike
  FXdouble              CC_v;                     // linear velocity of the carlike
  FXdouble              CC_ph;                    // steering angle of the carlike

  FXString              WC_namePos;               // name of the pose variable of the wheelchair in the hugrstore
  FXString              WC_nameCmd;               // name of the command variable of the wheelchair in the hugrstore

  FXString              CC_namePos;               // name of the pose variable of the cycab in the hugrstore
  FXString              CC_nameCmd;               // name of the command variable of the cycab in the hugrstore

  FXdouble              WC_wheel_radius;          // radius of the wheels of the wheelchair (m)
  FXdouble              WC_wheel_base;            // distance between the 2 wheels of the wheelchair (m)
  FXdouble              WC_max_speed;             // maximal linear speed of the wheelchair (m/s)
  FXdouble              WC_max_acc;               // maximal linear acceleration of the wheelchair (m/s^2)
  FXdouble              WC_max_rotation_speed;    // maximal rotation speed of the wheelchair (rad/s)
  FXdouble              WC_max_rotation_acc;      // maximal rotation acceleration of the wheelchair (rad/s^2)
  FXdouble              WC_braking_dist;          // braking distance from max speed of the wheelchair (m)
  FXdouble              WC_length;                // length  of the wheelchair (m)
  FXdouble              WC_width;                 // width  of the wheelchair (m)

  FXdouble              CC_length;                // length of the car-like (m)
  FXdouble              CC_max_speed;             // maximal linear speed of the car-like (m/s)
  FXdouble              CC_max_acc;               // maximal linear acceleration of the car-like (m/s^2)
  FXdouble              CC_max_steer;             // maximal steering angle of the car-like (rad)
  FXdouble              CC_max_stspd;             // maximal steering speed of the car-like (rad/s)

  FXdouble              Ctrl_v;                   // velocity control of the wheelchair/cycab
  FXdouble              Ctrl_avph;                // angular velocity control of the wheelchair / steering control of the cycab
  FXdouble              Coef_Ctrl_v;              // velocity control coefficient of the wheelchair/cycab (in [0;1])
  FXdouble              Coef_Ctrl_avph;           // angular velocity / steering control coefficient (in [0;1])

  FXdouble              CC_Ctrl_v;                // velocity control of the car-like
  FXdouble              CC_Ctrl_ph;               // steering control of the car-like

  // planner components
  FXdouble              planner_goalX;            // goal X position of the planner
  FXdouble              planner_goalY;            // goal Y position of the planner
  FXdouble              planner_goalTH;           // goal TH heading of the planner

  FXint                 plannerType;              // type of the planner used
                                                  // SIMPLE_TP -> Simple trajectory planner
                                                  // PVD_TP -> Path-velocity decomposition trajectory planner
                                                  // ROADMAP_TP -> Trajectory planner on a roadmap

  bool                  fixed_T;                  // true if the final time must be fixed
  FXdouble              planner_goalT;            // value of the fixed final time
  FXdouble              planner_DT;               // value of the delta time between each state-time of the computed trajectory

  FXdouble              planner_CoefSpd;          // coef \in [0;1] determining the ratio of the maximal speed used for the trajectory planning
  FXdouble              planner_CoefAcc;          // coef \in [0;1] determining the ratio of the maximal acceleration used for the trajectory planning

  // controller components
  FXint                 ctrlerType;               // type of the controller used
                                                  // 0 -> Simple controller
                                                  // 1 -> Trajectory follower
                                                  // 2 -> Teddy (Trajectory Deformation)

  FXdouble              dtCtrl;                   // time interval between two control requests
  
  FXint                 ctrlerDebugLevel;         // the controller debug level

  bool                  recordTraj;               // true if the real trajectory followed must be 
                                                  // recorded during the execution of the motion

  // deformer components
  FXdouble              k_int;                    // internal forces coefficient
  FXdouble              k_ext;                    // external forces coefficient (dynamic obstacles)
  FXdouble              k_stext;                  // external forces coefficient (static obstacles)
  FXdouble              dyn_med;                  // maximal effective distance (dynamic obstacles)
  FXdouble              st_med;                   // maximal effective distance (static obstacles)
  FXdouble              tmp_med;                  // maximal temporal effective distance

  // hugr store components
  hugr::Store           store;                    // the hugr store

  hugr::VariableId      WC_poseId;                // store id for the pose of the wheelchair
  hugr::VariableId      WC_commandId;             // store id for the command of the wheelchair
  hugr::VariableId      WC_constantsId;           // store id for the constants of the wheelchair

  hugr::VariableId      CC_poseId;                // store id for the pose of the cycab
  hugr::VariableId      CC_commandId;             // store id for the command of the cycab
  hugr::VariableId      CC_constantsId;           // store id for the constants of the cycab

  hugr::VariableId      Env_lObsId;               // store id for the list of dynamic obstacles

  WheelchairCommand     WC_command;               // pose of the wheelchair
  WheelchairPose        WC_pose;                  // command of the wheelchair
  WheelchairConstants   WC_constants;             // constants of the wheelchair

  // security components
  boost::mutex          mutexDrawScene;           // mutex used to prevent GL drawing conflicts


  // interface components
  bool                  orthoView;                // Orthogonal / Perspective view

  FXint                 wtI;                      // width of the image use for motion planning
  FXint                 htI;                      // height of the image use for motion planning
  bool                  stEnvLoaded;              // true if a static environment has been loaded

  FXint                 sampleX;                  // sampling of the static environment on X
  FXint                 sampleY;                  // sampling of the static environment on Y

  FXString              filename;                 // File being viewed

  bool                  dynamicView;              // true if the view move along the time

  FXHorizontalFrame     *contents;                // Content frame
  FXVerticalFrame       *imageFrame;              // Image frame
  FXVerticalFrame       *buttonFrame;             // Button frame

  FXGLDynCanvas         *glcanvas;                // GL Canvas to draw into
  FXGLVisual            *glvisual;                // OpenGL visual

  FXImageView           *imageview;               // Image viewer 

  bool                  mlbFlag;                  // true if the left button of the mouse is pressed
  bool                  mrbFlag;                  // true if the right button of the mouse is pressed

  FXColor               initColor;                // Color for the init configuration
  FXColor               goalColor;                // Color for the goal configuration

  FXLabel               *msgBox;                  // message box of the application

  FXLabel               *sysConfLabel;            // System configuration label
  FXLabel               *sysStateLabel;           // System state label

  FXLabel               *locStatusLabel;          // localization status label
  FXLabel               *locLabel;                // localization label

  FXLabel               *slammotStatusLabel;      // slammot status label

  FXLabel               *ctrlerStatusLabel;       // controller status label

  FXLabel               *currentSystemLabel;      // the current connected system
  FXLabel               *currentPlannerLabel;     // the current used planner
  FXLabel               *currentControllerLabel;  // the current used planner

  FXbool                dispFExt;                 // true if the external forces from dynamic obstacles must be displayed (during deformation process)
  FXbool                dispStFExt;               // true if the external forces from static obstacles must be displayed (during deformation process)
  FXbool                dispNearDynObs;           // true if the position of the nearest dynamic obstacles must be displayed (during deformation process)
  FXbool                dispNearStObs;            // true if the position of the nearest static obstacles must be displayed (during deformation process)

  FXLabel               *nbDeformLabel;           // label displaying the number of deformations
  FXLabel               *avgDeformDurLabel;       // label displaying the average deformation duration
  FXLabel               *DPSLabel;                // label displaying the number of deformations per second

  // link components
  FXDataTarget       initX_target;
  FXDataTarget       initY_target;
  FXDataTarget       initTH_target;

  FXDataTarget       WC_x_target;
  FXDataTarget       WC_y_target;
  FXDataTarget       WC_th_target;
  FXDataTarget       WC_v_target;
  FXDataTarget       WC_av_target;

  FXDataTarget       CC_x_target;
  FXDataTarget       CC_y_target;
  FXDataTarget       CC_th_target;
  FXDataTarget       CC_v_target;
  FXDataTarget       CC_ph_target;

  FXDataTarget       WC_namePos_target;
  FXDataTarget       WC_nameCmd_target;
  FXDataTarget       CC_namePos_target;
  FXDataTarget       CC_nameCmd_target;

  FXDataTarget       WC_wheel_radius_target;
  FXDataTarget       WC_wheel_base_target;
  FXDataTarget       WC_max_speed_target;
  FXDataTarget       WC_max_acc_target;
  FXDataTarget       WC_max_rotation_speed_target;
  FXDataTarget       WC_max_rotation_acc_target;
  FXDataTarget       WC_braking_dist_target;
  FXDataTarget       WC_length_target;
  FXDataTarget       WC_width_target;

  FXDataTarget       CC_length_target;
  FXDataTarget       CC_max_speed_target;
  FXDataTarget       CC_max_acc_target;
  FXDataTarget       CC_max_steer_target;
  FXDataTarget       CC_max_stspd_target;

  FXDataTarget       Ctrl_v_target;
  FXDataTarget       Ctrl_avph_target;
  FXDataTarget       Coef_Ctrl_v_target;
  FXDataTarget       Coef_Ctrl_avph_target;

//   FXDataTarget       CC_Ctrl_v_target;
//   FXDataTarget       CC_Ctrl_ph_target;

  FXDataTarget       planner_goalX_target;
  FXDataTarget       planner_goalY_target;
  FXDataTarget       planner_goalTH_target;
  FXDataTarget       planner_goalT_target;
  FXDataTarget       planner_DT_target;
  FXDataTarget       planner_CoefSpd_target;
  FXDataTarget       planner_CoefAcc_target;

  FXDataTarget       ctrlerType_target;

  FXDataTarget       plannerType_target;

  FXDataTarget       dtLoc_target;
  FXDataTarget       dtCtrl_target;

  FXDataTarget       k_int_target;
  FXDataTarget       k_ext_target;
  FXDataTarget       k_stext_target;
  FXDataTarget       dyn_med_target;
  FXDataTarget       st_med_target;
  FXDataTarget       tmp_med_target;

  FXDataTarget       nameDynObs_target;

  FXDataTarget       ctrlerDebugLevel_target;

  FXDataTarget       stEnvFilename_target;

  FXDataTarget       optRoadmap_target;
  FXDataTarget       RMfilename_target;

  FXDataTarget       distFilter_target;
  FXDataTarget       securDist_target;


protected : 
  MPC_HugrIfc();

public : 
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////// Constructors & Init functions //////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // constructors & destructors
  MPC_HugrIfc(FXApp* a);
  virtual ~MPC_HugrIfc();
  virtual void create();

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////// Signal functions //////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  
  // GL viewer signals
  long onConfigure(FXObject*,FXSelector,void*);
  long onExpose(FXObject*,FXSelector,void*);
  long repaint(FXObject*,FXSelector,void*);
  long timeoutDraw(FXObject*,FXSelector,void*);

  // Map loading signals
//   long loadMapFromFile(FXObject*,FXSelector,void*);
//   long loadHugrMap(FXObject*,FXSelector,void*);
  long loadStaticEnvBtn(FXObject*,FXSelector,void*);
  long loadRoadmapBtn(FXObject*,FXSelector,void*);
  long saveRoadmapBtn(FXObject*,FXSelector,void*);
  long changeStEnvFileName(FXObject*,FXSelector,void*);
  long changeRoadmapFileName(FXObject*,FXSelector,void*);

  // Map position choice signals
//   long onLeftMouseDown(FXObject*,FXSelector,void*);
//   long onLeftMouseUp(FXObject*,FXSelector,void*);
  long onRightMouseDown(FXObject*,FXSelector,void*);
  long onRightMouseUp(FXObject*,FXSelector,void*);
  long onMouseMove(FXObject*,FXSelector,void*);

  // Init of the robot config signals
  long initWCCfg(FXObject*,FXSelector,void*);
  long initCCCfg(FXObject*,FXSelector,void*);

  // Localization signals
  long startLoc(FXObject*,FXSelector,void*);
  long stopLoc(FXObject*,FXSelector,void*);
  long displayTrajDevRecBtn(FXObject*,FXSelector,void*);
  long displayAllTrajRecBtn(FXObject*,FXSelector,void*);
  long deleteTrajDevRecBtn(FXObject*,FXSelector,void*);
  long deleteAllTrajRecBtn(FXObject*,FXSelector,void*);

  // Controller signals
  long initCtrler(FXObject*,FXSelector,void*);
  long startCtrler(FXObject*,FXSelector,void*);
  long stopCtrler(FXObject*,FXSelector,void*);
  long updateCtrl(FXObject*,FXSelector,void*);
  long stopRobotMotion(FXObject*,FXSelector,void*);
  long updateRecordTrajBtn(FXObject*,FXSelector,void*);
  long moveForwardBtn(FXObject*,FXSelector,void*);
  long moveBackwardBtn(FXObject*,FXSelector,void*);
  long turnLeftBtn(FXObject*,FXSelector,void*);
  long turnRightBtn(FXObject*,FXSelector,void*);
  long moveFwdLeft(FXObject*,FXSelector,void*);
  long moveFwdRight(FXObject*,FXSelector,void*);
  long moveBwdLeft(FXObject*,FXSelector,void*);
  long moveBwdRight(FXObject*,FXSelector,void*);
  long stopMotionBtn(FXObject*,FXSelector,void*);

  // Planner signals
  long initGoalPosPlanner(FXObject*,FXSelector,void*);
  long initPlanner(FXObject*,FXSelector,void*);
  long startPlanner(FXObject*,FXSelector,void*);

  // SLAMMOT signals
  long startSLAMMOT(FXObject*,FXSelector,void*);
  long stopSLAMMOT(FXObject*,FXSelector,void*);

  // Environment signals
  long updateDispRoadmap(FXObject*,FXSelector,void*);
  long updateAutoBuildRoadmap(FXObject*,FXSelector,void*);

  // interface signals
  long updateDynView(FXObject*,FXSelector,void*);
  long centerViewBtn(FXObject*,FXSelector,void*);
  long updateFixedTime(FXObject*,FXSelector,void*);

  long updateDispFExt(FXObject*,FXSelector,void*);
  long updateDispStFExt(FXObject*,FXSelector,void*);
  long updateDispNearDynObs(FXObject*,FXSelector,void*);
  long updateDispNearStObs(FXObject*,FXSelector,void*);


  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////// HUGR store connection functions ///////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // HUGR store connection functions
  int connectSLAMMOT();                                                       // establish a connection to the hugr store 
                                                                              // and try to get back informations about the environment
                                                                              // return -1 if the connection has been established, -1 else
  long setWCSimuVar(FXObject*,FXSelector,void*);                              // set the wheelchair simulation variables for the hugr store
  long setWCRobotVar(FXObject*,FXSelector,void*);                             // set the wheelchair robot variables for the hugr store
  long connectHugrWheelChair(FXObject*,FXSelector,void*);                     // establish a connection to the hugr store 
                                                                              // and try to take control of the wheelchair
  long connectHugrCycab(FXObject*,FXSelector,void*);                          // establish a connection to the hugr store 
                                                                              // and try to take control of the cycab
  bool WheelChairExists();                                                    // check if the wheelchair exists in the HUGR store
  bool CycabExists();                                                         // check if the cycab exists in the HUGR store

  void loopStatus();                                                          // loop to update the localization and controller status of the robot

  void loopLocWC();                                                           // loop to update the wheelchair localization display
  void loopLocCC();                                                           // loop to update the cycab localization display

  // other functions
//   void drawInitPos();                                                         // draw the init position on the image
//   void drawInitDir();                                                         // draw the init direction on the image
//   void drawInitCfg();                                                         // draw the init configuration on the image
//   void drawGoalPos();                                                         // draw the goal position on the image
//   void drawGoalDir();                                                         // draw the goal direction on the image
//   void drawGoalCfg();                                                         // draw the goal configuration on the image  

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////// Map treatment functions ///////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  \brief display the current map in the viewer
   */
  void loadMapInViewer();

  FXbool loadimage(const FXString& file);                                     // load the image of the map
  void coordFrameToCoordMap(const int& xF, const int& yF, double& xM, double& yM);  // convert frame coordinates to map coordinates
  void coordMapToCoordFrame(const double& xM, const double& yM, int& xF, int& yF);  // convert map coordinates to frame coordinates

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////// OpenGL Viewer Drawing functions ///////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  void updDynamicView();                                                      // update the view of the dynamic viewer mode
  void drawRobot();                                                           // draw the robotic system
  void drawTraj();                                                            // draw the trajectory followed by the robot
  void drawFieldBounds();                                                     // draw the bounds of the field
  void drawObstacles();                                                       // draw the obstacles of the environment and their forecast models
  void drawControllerSpec();                                                  // draw the controller specifications if needed
  void drawScene();                                                           // draw all elements of the scene

  /**
   *  \brief center the openGL view on the scene
   */
  void centerView();

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////// Environment management ///////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  \brief load a static environment from an input file
   *  \param nomfic : the input filename
   *  \return 0 if the process succeed
   *         -1 if the file has not been found
   *         -2 if the localization has not been started
   *         -3 if the environment perception has not been initialized
   *         -4 if an unknown error occured
   *         -5 if no system has been initilialized
   */
  int loadStaticEnv(const char* filename);

  /**
   *  \brief load a roadmap from an input file
   *  \param nomfic : the input filename
   *  \return 0 if the process succeed
   *         -1 if the file has not been found
   *         -2 if the localization has not been started
   *         -3 if the environment perception has not been initialized
   *         -4 if the loading process failed
   *         -5 if no system has been initilialized
   */
  int loadRoadmap(const char* filename);

  /**
   *  \brief save a roadmap into an output file
   *  \param nomfic : the input filename
   *  \return 0 if the process succeed
   *         -1 if the file has not been found
   *         -2 if the localization has not been started
   *         -3 if the environment perception has not been initialized
   *         -4 if the loading process failed
   *         -5 if no system has been initilialized
   */
  int saveRoadmap(const char* filename);

  /**
   *  \brief sample the static environment to select easily the goal configuration
   *  \return the sampled map if the process succeed, NULL else
   */
  Map* sampleStaticEnv();

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////// Recording tools ///////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  \brief reinit the trajectory deviation record
   */
  void reinitTrajDevRec();

  /**
   *  \brief delete the current trajectory record
   */
  void deleteTrajDevRec();

  /**
   *  \brief delete all the trajectory records
   */
  void deleteAllTrajRec();

  /**
   *  \brief record the position deviation during a trajectory following process
   */
  void recordTrajDev();

  /**
   *  \brief display the trajectory deviation record
   */
  void displayTrajDevRec();

  /**
   *  \brief display all the trajectory records
   */
  void displayAllTrajRec();

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////// GUI functions ////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  void updateCurSystemLabel();                                  // update the current connected system label
  void updateCurPlannerLabel();                                 // update the current connected planner label
  void updateCurControllerLabel();                              // update the current connected controller label

  /**
   *  \brief init the gui from an input configuration file
   *  \param filename : the input filename
   */
  void initHugrIfcFromFile(const char* filename);


  // Messages for our class
  enum{
    ID_CANVAS=FXMainWindow::ID_LAST,
    ID_REPAINT,
    ID_TIMEOUT_DRAW,
    ID_LOAD_MAP,
    ID_LOAD_HUGRMAP,
    ID_LOAD_STATICENV,
    ID_LOAD_ROADMAP,
    ID_SAVE_ROADMAP,
    ID_CONNECT_HUGR_WC,
    ID_CONNECT_HUGR_CC,
    ID_INIT_WCCFG,
    ID_INIT_CCCFG,
    ID_CONNECT_ENV,
    ID_START_LOC,
    ID_STOP_LOC,
    ID_INIT_CTRLER,
    ID_START_CTRLER,
    ID_STOP_CTRLER,
    ID_UPDATE_CTRL,
    ID_INIT_PL_GOALPOS,
    ID_INIT_PLANNER,
    ID_START_PLANNER,
    ID_IMAGE,
    ID_START_SLAMMOT,
    ID_STOP_SLAMMOT,
    ID_DYNVIEW_CB,
    ID_CHANGE_STENV,
    ID_CHANGE_ROADMAP,
    ID_DISP_ROADMAP,
    ID_AUTOBUILD_ROADMAP,
    ID_STOP_ROBOT,
    ID_CENTER_VIEW,
    ID_WC_SIMUVAR,
    ID_WC_ROBOTVAR,
    ID_RECORD_TRAJ_BTN,
    ID_DISP_TRAJREC,
    ID_DISP_ALLTRAJREC,
    ID_EMPTY_DEVREC,
    ID_EMPTY_ALLREC,
    ID_FIXED_TIME_CB,
    ID_MOVE_FWD,
    ID_MOVE_BWD,
    ID_TURN_LEFT,
    ID_TURN_RIGHT,
    ID_MOVE_FWD_LEFT,
    ID_MOVE_FWD_RIGHT,
    ID_MOVE_BWD_LEFT,
    ID_MOVE_BWD_RIGHT,
    ID_STOP_MOTION,
    ID_DISP_FEXT_UPD,
    ID_DISP_STFEXT_UPD,
    ID_DISP_NEAR_DYNOBS_UPD,
    ID_DISP_NEAR_STOBS_UPD,
    ID_LAST
  };

};

FXDEFMAP(MPC_HugrIfc) MPC_HugrIfcMap[]={
  //________Message_Type_______________________ID________________________Message_Handler_______
  FXMAPFUNC(SEL_PAINT,                MPC_HugrIfc::ID_CANVAS,            MPC_HugrIfc::onExpose),
  FXMAPFUNC(SEL_CONFIGURE,            MPC_HugrIfc::ID_CANVAS,            MPC_HugrIfc::onConfigure),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_REPAINT,           MPC_HugrIfc::repaint),
  FXMAPFUNC(SEL_TIMEOUT,              MPC_HugrIfc::ID_TIMEOUT_DRAW,      MPC_HugrIfc::timeoutDraw),
//   FXMAPFUNC(SEL_LEFTBUTTONPRESS,      MPC_HugrIfc::ID_IMAGE,             MPC_HugrIfc::onLeftMouseDown),
//   FXMAPFUNC(SEL_LEFTBUTTONRELEASE,    MPC_HugrIfc::ID_IMAGE,             MPC_HugrIfc::onLeftMouseUp),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,     MPC_HugrIfc::ID_IMAGE,             MPC_HugrIfc::onRightMouseDown),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,   MPC_HugrIfc::ID_IMAGE,             MPC_HugrIfc::onRightMouseUp),
  FXMAPFUNC(SEL_MOTION,               MPC_HugrIfc::ID_IMAGE,             MPC_HugrIfc::onMouseMove),
//   FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_LOAD_MAP,          MPC_HugrIfc::loadMapFromFile),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_CONNECT_HUGR_WC,   MPC_HugrIfc::connectHugrWheelChair),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_CONNECT_HUGR_CC,   MPC_HugrIfc::connectHugrCycab),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_INIT_WCCFG,        MPC_HugrIfc::initWCCfg),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_INIT_CCCFG,        MPC_HugrIfc::initCCCfg),
//   FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_CONNECT_ENV,      MPC_HugrIfc::connectEnv),
//   FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_LOAD_HUGRMAP,      MPC_HugrIfc::loadHugrMap),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_LOAD_STATICENV,    MPC_HugrIfc::loadStaticEnvBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_LOAD_ROADMAP,      MPC_HugrIfc::loadRoadmapBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_SAVE_ROADMAP,      MPC_HugrIfc::saveRoadmapBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_START_LOC,         MPC_HugrIfc::startLoc),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_STOP_LOC,          MPC_HugrIfc::stopLoc),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_INIT_CTRLER,       MPC_HugrIfc::initCtrler),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_START_CTRLER,      MPC_HugrIfc::startCtrler),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_STOP_CTRLER,       MPC_HugrIfc::stopCtrler),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_UPDATE_CTRL,       MPC_HugrIfc::updateCtrl),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_INIT_PL_GOALPOS,   MPC_HugrIfc::initGoalPosPlanner),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_INIT_PLANNER,      MPC_HugrIfc::initPlanner),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_START_PLANNER,     MPC_HugrIfc::startPlanner),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_START_SLAMMOT,     MPC_HugrIfc::startSLAMMOT),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_STOP_SLAMMOT,      MPC_HugrIfc::stopSLAMMOT),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_DYNVIEW_CB,        MPC_HugrIfc::updateDynView),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_CHANGE_STENV,      MPC_HugrIfc::changeStEnvFileName),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_CHANGE_ROADMAP,    MPC_HugrIfc::changeRoadmapFileName),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_DISP_ROADMAP,      MPC_HugrIfc::updateDispRoadmap),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_AUTOBUILD_ROADMAP, MPC_HugrIfc::updateAutoBuildRoadmap),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_STOP_ROBOT,        MPC_HugrIfc::stopRobotMotion),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_CENTER_VIEW,       MPC_HugrIfc::centerViewBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_WC_SIMUVAR,        MPC_HugrIfc::setWCSimuVar),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_WC_ROBOTVAR,       MPC_HugrIfc::setWCRobotVar),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_RECORD_TRAJ_BTN,   MPC_HugrIfc::updateRecordTrajBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_DISP_TRAJREC,      MPC_HugrIfc::displayTrajDevRecBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_DISP_ALLTRAJREC,   MPC_HugrIfc::displayAllTrajRecBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_EMPTY_DEVREC,      MPC_HugrIfc::deleteTrajDevRecBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_EMPTY_ALLREC,      MPC_HugrIfc::deleteAllTrajRecBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_FIXED_TIME_CB,     MPC_HugrIfc::updateFixedTime),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_MOVE_FWD,          MPC_HugrIfc::moveForwardBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_MOVE_BWD,          MPC_HugrIfc::moveBackwardBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_TURN_LEFT,         MPC_HugrIfc::turnLeftBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_TURN_RIGHT,        MPC_HugrIfc::turnRightBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_STOP_MOTION,       MPC_HugrIfc::stopMotionBtn),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_MOVE_FWD_LEFT,     MPC_HugrIfc::moveFwdLeft),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_MOVE_FWD_RIGHT,    MPC_HugrIfc::moveFwdRight),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_MOVE_BWD_LEFT,     MPC_HugrIfc::moveBwdLeft),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_MOVE_BWD_RIGHT,    MPC_HugrIfc::moveBwdRight),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_DISP_FEXT_UPD,        MPC_HugrIfc::updateDispFExt),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_DISP_STFEXT_UPD,      MPC_HugrIfc::updateDispStFExt),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_DISP_NEAR_DYNOBS_UPD, MPC_HugrIfc::updateDispNearDynObs),
  FXMAPFUNC(SEL_COMMAND,              MPC_HugrIfc::ID_DISP_NEAR_STOBS_UPD,  MPC_HugrIfc::updateDispNearStObs),
};

FXIMPLEMENT(MPC_HugrIfc,FXMainWindow,MPC_HugrIfcMap,ARRAYNUMBER(MPC_HugrIfcMap))

#endif // MPC_HUGRIFC_HPP
