#include "Hugr_TrajDeform.hpp"

                                    /**************************************************/
                                    /************** Class Hugr_TrajDeform *************/
                                    /**************************************************/
  
/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Hugr_TrajDeform::Hugr_TrajDeform()
{
  trajRes_ = NULL;

  envReg_ = NULL;

  timer_ = NULL;

  threadDeform_ = NULL;

  mutexTraj_ = NULL;
  mutexLoc_ = NULL;
  mutexEnv_ = NULL;

  status_ = DEFORMER_STOPPED;
}

/**
 *  \brief init constructor
 *  \param timer : timer of the associated controller
 *  \param mutexTraj : mutex used to prevent trajectory access conflicts
 *  \param mutexLoc : mutex used to prevent localization access conflicts
 *  \param mutexEnv : mutex used to prevent environment model access conflicts
 */
Hugr_TrajDeform::Hugr_TrajDeform(Timer* timer,boost::mutex* mutexTraj,boost::mutex* mutexLoc,boost::mutex* mutexEnv)
{
  trajRes_ = NULL;

  envReg_ = NULL;

  timer_ = timer;

  threadDeform_ = NULL;

  mutexTraj_ = mutexTraj;
  mutexLoc_ = mutexLoc;
  mutexEnv_ = mutexEnv;

  status_ = DEFORMER_STOPPED;
}

/**
 *  \brief destructor
 */
Hugr_TrajDeform::~Hugr_TrajDeform()
{
  if(trajRes_ != NULL) delete trajRes_;
  if(envReg_ != NULL) delete envReg_;
}


/********************* Actions ********************/

/**
 *  \brief update the trajectory followed by the robot with the deformed trajectory
 */
void Hugr_TrajDeform::updateTraj()
{
  // block the access to the trajectory during its reinitialization
  if(MUTEX_DEBUG == 1) cout << "Hugr_TrajDeform::updateTraj : waiting for mutex Traj" << endl;
  boost::mutex::scoped_lock lock(*mutexTraj_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_TrajDeform::updateTraj : mutex Traj locked" << endl;

  if(trajRes_ != NULL) delete trajRes_;
  trajRes_ = traj();

  if(MUTEX_DEBUG == 1) cout << "Hugr_TrajDeform::updateTraj : mutex Traj released" << endl;
}

// /**
//  *  \brief execute the control loop
//  */
// void Hugr_TrajDeform::loop()
// {
//   while(this->status_ == DEFORMER_STARTED)
//     {
//       if(this->traj() == NULL) cout << "Hugr_TrajDeform::loop : traj() null" << endl;
//       else cout << "Hugr_TrajDeform::loop : traj() not null" << endl;
      
//       if(trajRes_ == NULL) cout << "Hugr_TrajDeform::loop : trajRes_ null" << endl;
//       else cout << "Hugr_TrajDeform::loop : trajRes_ not null" << endl;

//       // init the trajectory to deform
//       this->initTraj();

//       // replace the current environment model by the last 
//       // snapshot registered by the perception process
//       if(env_ != NULL) delete env_;
//       env_ = envReg_;

//       // retrieve the current time wrt the trajectory followed
//       double time = timer_->totalElapsedTime();

//       // deform the trajectory
//       int res = this->deform(time);

//       switch(res)
// 	{
// 	case 0 : // case where the deformation process ended successfully
// 	  // update the result trajectory
// 	  this->updateTraj();
// 	  break;
// 	case -1 : // case where no input trajectory has been defined
// 	  cerr << "Hugr_TrajDeform::loop. No input trajectory has been defined." << endl;
// 	  this->status_ == DEFORMER_STOPPED;
// 	  exit(1);
// 	  break;
// 	case -2 : // case where a collision may occur during the prediction horizon
// 	  cerr << "Hugr_TrajDeform::loop. Warning : A collision may occur during the prediction horizon." << endl;
// 	  break;
// 	case -3 : // case where the connectivity of the trajectory has been lost
// 	  cerr << "Hugr_TrajDeform::loop. Warning : The connectivity of the trajectory has been lost." << endl;
// 	  break;
// 	case -4 : // case where the final time of the trajectory has been already past
// 	  cout << "Hugr_TrajDeform::loop. End of the deformation process." << endl;
// 	  this->status_ == DEFORMER_STOPPED;
// 	  break;
// 	}
//     }
// }

// /**
//  *  \brief start the timer
//  *  \return 0 if the starting process succeed, -1 else
//  */
// int Hugr_TrajDeform::start()
// {
//   switch(status_)
//     {
//     case DEFORMER_NULL : 
//     case DEFORMER_DISCONNECTED : 
//       {
// 	cerr << "Hugr_TrajDeform::start : Warning. The deformation process is disconnected. Should never appear..." << endl;
// 	return -1;
//       }
//     case DEFORMER_STOPPED : 
//       {
// 	status_ = DEFORMER_STARTED;
// 	threadDeform_ = new boost::thread(boost::bind(&Hugr_TrajDeform::loop,this));
// 	return 0;
//       }
//     case DEFORMER_STARTED : 
//       {
// 	// nothing to do
// 	cerr << "Hugr_TrajDeform::start : Deformation process already started" << endl;
// 	return 0;
//       }
//     default : 
//       {
// 	cerr << "Hugr_TrajDeform::start : Unknown deformation process status" << endl;
// 	return -1;
//       }
//     }
// }

// /**
//  *  \brief stop the timer
//  *  \return 0 if the stoping process succeed, -1 else
//  */
// int Hugr_TrajDeform::stop()
// {
//   switch(status_)
//     {
//     case DEFORMER_NULL : 
//     case DEFORMER_DISCONNECTED : 
//       {
// 	cerr << "Hugr_TrajDeform::stop : Warning. The deformation process is disconnected. Should never appear..." << endl;
// 	return -1;
//       }
//     case DEFORMER_STOPPED : 
//       {
// 	// nothing to do
// 	return 0;
//       }
//     case DEFORMER_STARTED : 
//       {
// 	status_ = DEFORMER_STOPPED;
// 	if(threadDeform_ != NULL) delete threadDeform_;
// 	return 0;
//       }
//     }
// }
