#include "MPC_TeddyIfc.hpp"

#include "common/inc/Utils.hpp"

#include "planner/inc/PVDTrajPlanner.hpp"
#include "planner/inc/CTrajPlanner.hpp"
//#include "AccASTrajGen.hpp"
//#include "WAccTrajGen.hpp"
#include "planner/inc/RoadMapPP.hpp"
#include "planner/inc/RoadMapTP.hpp"

#include "trajGen/inc/CLTrajGen.hpp"
// #include "CLTrajGen2.hpp"
// #include "CL2TrajGen.hpp"
#include "trajGen/inc/DDTrajGen.hpp"
#include "trajGen/inc/UtilsTG.hpp"

#include "trajFollow/inc/CLTrajFollow.hpp"
#include "trajFollow/inc/CCTrajFollow.hpp"
#include "trajFollow/inc/WCTrajFollow.hpp"

#include <time.h>
#include <sys/time.h>

#include <GL/glut.h>
#include <list>

#include "gnuplot/inc/gnuplot_i.hpp"

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Constructors & Init functions //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// default constructor (only used by subclasses)
MPC_TeddyIfc::MPC_TeddyIfc(){}

// init constructor
MPC_TeddyIfc::MPC_TeddyIfc(FXApp* a):FXMainWindow(a,"Trajectory Deformation",NULL,NULL,DECOR_ALL,0,0,950,650)
{
  /***************************************************************************************************************
   ********************************************** initialisation *************************************************
   ***************************************************************************************************************/

  FXHorizontalFrame *mainFrame;
  FXVerticalFrame *glcanvasFrame;
  FXVerticalFrame *buttonFrame;
  FXComposite *glpanel;

  path_     = NULL;
  traj_     = NULL;
  initTraj_ = NULL;

  //rb_       = new Robot;
  //carlike_  = NULL;
  //diffdrive_= new DiffDriveSys;
  sys_      = NULL;

  deformer_ = NULL;
  follower_ = NULL;
  //mfollower_ = NULL;

  roadmap_  = NULL;
  map_      = NULL;

  env_ = new Environment;

  typeState = NULL;
//   tableFInt = NULL;
//   tableFExt = NULL;
  
  glcanvas = NULL;
  glvisual = NULL;

  robotStateTF_ = NULL;
  stateToReachTF_ = NULL;

  processType_ = TEDDY_PROC;

  maxSpd      = 45.0f;
  maxAcc      = 17.0f;
  maxWVel     = 8.0*M_PI;
  maxWAcc     = 1.5*M_PI;

  maxSteer    = 0.25*M_PI;
  maxSteerS   = 2.0*M_PI;
  maxSpdCoef  = 0.30f;
  maxAccCoef  = 0.30f;
  robotRadius = 10.0f;
  robotLength = 20.0f;
  //robotRadius = 15.0f;
  
  seqNum    = 0;
  optSeqNum = 0;
  
  idObs   = FXString("obs1");
  xObs    = 0;
  yObs    = 0;
  sxObs   = 0;
  syObs   = 0;
  rObs    = 0;
  initT   = 0;
  lifeT   = 0;
  randObs = FXString("no");
  
  genTrajType  = PM_TRAJ;
  trajType     = PM_TRAJ;
  trajTimeStep = 0.3;
  trajPosStep  = 12.0;
  trajStep     = 0;
  maxTrajStep  = 1;
  time         = 0.0;
  maxTime      = 1.0;
  running      = false;
  
  curStep      = 0;

  inputPathName = FXString("../path/path2.pth");
  mapName       = FXString("../map/OccGrid.pgm");
  occgridName   = FXString("../map/OccGrid.pgm");
  
  sampleXOccGrid = 400;
  sampleYOccGrid = 400;

  nbRandObs = 10;
  
  dispFuture         = false;
  alreadyElapsedTime = 0;
  oneStep            = false;
  reliabilityTime    = 12.0;
  epsilonRT          = 0.05;
  predHorizon        = 3.0;
  nbTimeout          = 0;
  totalNbTimeout     = 0;
  
  FPS                = 0;
  lastFPS            = 0;
  deformDur          = 0;
  nbST               = 0;
  nbSTtot            = 0;
  
  dispForces     = false;
  dispAllForces  = false;
  indDispForces  = 0;
  dispNearStates = false;
  dispNearStStates = false;
  
  optModeDeform = 0;
  
  K_INT           = 1.0;
  K_EXT_SPT       = 0.01;
  K_EXT_TMP       = 0.05;
  K_SPT_TMP       = 3.0;
  K_EXT           = 6.0;
  K_STEXT         = 5.0;
//   maxEffectDist   = 80.0;
  maxStEffectDist = 80.0;

  alphaMin        = 0.1;     // minimal spatial coefficient (for the metric)
  alphaMax        = 1.0;     // maximal spatial coefficient (for the metric)
  betaMin         = 0.0;     // minimal temporal coefficient (for the metric)
  betaMax         = 30.0;    // maximal temporal coefficient (for the metric)
  
  maxSpDist       = 100.0;
  maxTmpDist      = 5.0;
  
  useTemporalConsistency = true;
  useConnectivityChecker = false; //true;
  pauseColLoc = false;
  dynamicView = false;
  roadmapView = true;
  
  seuilChgt = 0.7;
  maxAngleChgt = 2*M_PI*0.05;
  maxNormChgt = 0.03;
  
  typeMotion = 0;
  
  initXRb  = 100;
  initYRb  = 100;
  initThRb = 0.0;
  initVRb  = 0.0;
  initPhRb = 0.0;  

  goalXRb  = 480;
  goalYRb  = 80;
  goalThRb = -M_PI;

  dfltXMap          = 400;
  dfltYMap          = 400;
  drawMapBoundaries = true;
  
  goalPosX = -1;
  goalPosY = -1;
  
  realTimeAlgo = true;
  
  orthoView = false;
  
  debugLevel = 0;
  maxNbStep = 20;
  maxResidual = 0.5;
  
  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;
  
  drawTrajGen = false;
  drawTrajGenTF = true;

  parallelize = false;
  nbThread = 5;

  noiseX = 2.0;
  noiseY = 2.0;
  noiseTH = 0.4*M_PI;
  noiseV = 12.0;
  noisePH = 0.05;

  noiseAcc = 2.0;
  noiseSS = 0.02*M_PI;

  TF_mode = CC_TF_MODE;
  TF_dt = 0.5;
  TF_timeMeeting = 10.0;

  initXTraj  = 95;
  initYTraj  = 100;
  initThTraj = 0.25*M_PI;
  initTTraj  = 0.0;
  
  trueLastElapsedT = 0.0;

  // true if the goal position for the motion planning if fixed (else random pos)
  fixRbGoalPos = true;

  TF_X = NULL;                    // position X of the trajectory followed
  TF_Y = NULL;                    // position Y of the trajectory followed
  TF_T = NULL;                    // time T of the trajectory followed
  TF_locX = NULL;                 // position X of the localization percieved during the traj following process
  TF_locY = NULL;                 // position Y of the localization percieved during the traj following process
  TF_rbX = NULL;                  // position X of the robot during the traj following process
  TF_rbY = NULL;                  // position Y of the robot during the traj following process

  TF_DerivLoc = NULL;             // derivation from the localization to the trajectory followed
  TF_ErrorLoc = NULL;             // error of the localization
  TF_DerivRb = NULL;              // derivation from the robot position to the trajectory followed
  TF_rbCtrl1 = NULL;              // first control applied on the robot
  TF_rbCtrl2 = NULL;              // second control applied on the robot
  TF_noisedCtrl1 = NULL;          // first control applied on the robot
  TF_noisedCtrl2 = NULL;          // second control applied on the robot

  II_ = 0;
  medDerivLoc_ = 0.0;
  medDerivRb_ = 0.0;

  maxSpd_target.connect(maxSpd);
  maxAcc_target.connect(maxAcc);
  maxWVel_target.connect(maxWVel);
  maxWAcc_target.connect(maxWAcc);

  maxSteer_target.connect(maxSteer);
  maxSteerS_target.connect(maxSteerS);
  maxSpdCoef_target.connect(maxSpdCoef);
  maxAccCoef_target.connect(maxAccCoef);
  robotRadius_target.connect(robotRadius);
  robotLength_target.connect(robotLength);
  
  seqNum_target.connect(seqNum);
  optSeqNum_target.connect(optSeqNum);
  
  idObs_target.connect(idObs);
  xObs_target.connect(xObs);
  yObs_target.connect(yObs);
  sxObs_target.connect(sxObs);
  syObs_target.connect(syObs);
  rObs_target.connect(rObs);
  initT_target.connect(initT);
  lifeT_target.connect(lifeT);
  randObs_target.connect(randObs);
  
  trajType_target.connect(trajType);
  trajTimeStep_target.connect(trajTimeStep);
  trajPosStep_target.connect(trajPosStep);

  trajStep_target.connect(trajStep);
  maxTrajStep_target.connect(maxTrajStep);
  time_target.connect(time);
  maxTime_target.connect(maxTime);
  
  inputPathName_target.connect(inputPathName);
  mapName_target.connect(mapName);
  occgridName_target.connect(occgridName);

  sampleXOccGrid_target.connect(sampleXOccGrid);
  sampleYOccGrid_target.connect(sampleYOccGrid);
  
  nbRandObs_target.connect(nbRandObs);
  
  reliabilityTime_target.connect(reliabilityTime);
  epsilonRT_target.connect(epsilonRT);
  predHorizon_target.connect(predHorizon);
  
  FPS_target.connect(FPS);
  lastFPS_target.connect(lastFPS);
  deformDur_target.connect(deformDur);
  nbST_target.connect(nbST);
  nbSTtot_target.connect(nbSTtot);

  indDispForces_target.connect(indDispForces);

  optModeDeform_target.connect(optModeDeform);

  K_INT_target.connect(K_INT);
  K_EXT_SPT_target.connect(K_EXT_SPT);
  K_EXT_TMP_target.connect(K_EXT_TMP);
  K_SPT_TMP_target.connect(K_SPT_TMP);
  K_EXT_target.connect(K_EXT);
  K_STEXT_target.connect(K_STEXT);
  maxStEffectDist_target.connect(maxStEffectDist);

  alphaMin_target.connect(alphaMin);
  alphaMax_target.connect(alphaMax);
  betaMin_target.connect(betaMin);
  betaMax_target.connect(betaMax);
  maxSpDist_target.connect(maxSpDist);
  maxTmpDist_target.connect(maxTmpDist);

  seuilChgt_target.connect(seuilChgt);
  maxAngleChgt_target.connect(maxAngleChgt);
  maxNormChgt_target.connect(maxNormChgt);

  typeMotion_target.connect(typeMotion);

  initXRb_target.connect(initXRb);
  initYRb_target.connect(initYRb);
  initThRb_target.connect(initThRb);
  initVRb_target.connect(initVRb);
  initPhRb_target.connect(initPhRb);

  goalXRb_target.connect(goalXRb);
  goalYRb_target.connect(goalYRb);
  goalThRb_target.connect(goalThRb);

  dfltXMap_target.connect(dfltXMap);
  dfltYMap_target.connect(dfltYMap);

  goalPosX_target.connect(goalPosX);
  goalPosY_target.connect(goalPosY);

  debugLevel_target.connect(debugLevel);
  maxNbStep_target.connect(maxNbStep);
  maxResidual_target.connect(maxResidual);

  nbThread_target.connect(nbThread);

  noiseX_target.connect(noiseX);
  noiseY_target.connect(noiseY);
  noiseTH_target.connect(noiseTH);
  noiseV_target.connect(noiseV);
  noisePH_target.connect(noisePH);

  noiseAcc_target.connect(noiseAcc);
  noiseSS_target.connect(noiseSS);

  TF_mode_target.connect(TF_mode);
  TF_dt_target.connect(TF_dt);
  TF_timeMeeting_target.connect(TF_timeMeeting);

  initXTraj_target.connect(initXTraj);
  initYTraj_target.connect(initYTraj);
  initThTraj_target.connect(initThTraj);
  initTTraj_target.connect(initTTraj);


  /***************************************************************************************************************
   ************************************************ main frame ***************************************************
   ***************************************************************************************************************/
  
  mainFrame = new FXHorizontalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  /***************************************************************************************************************
   *********************************************** canvas frame **************************************************
   ***************************************************************************************************************/

  glcanvasFrame = new FXVerticalFrame(mainFrame,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // Label above the buttons
  new FXLabel(glcanvasFrame,"Viewer",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  // Horizontal divider line
  new FXHorizontalSeparator(glcanvasFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Drawing glcanvas
  glpanel=new FXVerticalFrame(glcanvasFrame,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0, 0,0,0,0);

  // A Visual to drag OpenGL
  glvisual=new FXGLVisual(getApp(),VISUAL_DOUBLEBUFFER|VISUAL_STEREO);

  // Drawing glcanvas
  glcanvas=new FXGLDynCanvas(glpanel,glvisual,this,(FXSelector)ID_CANVAS,(FXuint)(LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT));
  glcanvas->setMainWin(this,ID_REPAINT);
  //glcanvas=new FXGLViewer(glpanel,glvisual,this,ID_GLVIEWER,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT);

  msgBox = new FXLabel(glpanel,"Select an environment and a motion behaviour",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);


  /***************************************************************************************************************
   *********************************************** button frame **************************************************
   ***************************************************************************************************************/
  
  buttonFrame = new FXVerticalFrame(mainFrame,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // Label above the buttons
  new FXLabel(buttonFrame,"Commands",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  
  // Tab book with switchable panels
  FXTabBook* panels=new FXTabBook(buttonFrame,NULL,0,LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);


  /***************************************************************************************************************
   ***************************************** system configuration tab ********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Sys config");

  FXVerticalFrame *tabSysCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabSysCfg,"System configuration");

  FXMatrix *matSysCfg=new FXMatrix(tabSysCfg,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matSysCfg,"Max speed :");
  new FXTextField(matSysCfg,6,&maxSpd_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max acceleration :");
  new FXTextField(matSysCfg,6,&maxAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max wvel :");
  new FXTextField(matSysCfg,6,&maxWVel_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max wacc :");
  new FXTextField(matSysCfg,6,&maxWAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max steering :");
  new FXTextField(matSysCfg,6,&maxSteer_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Max steer speed :");
  new FXTextField(matSysCfg,6,&maxSteerS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Robot radius :");
  new FXTextField(matSysCfg,6,&robotRadius_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matSysCfg,"Robot length :");
  new FXTextField(matSysCfg,6,&robotLength_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matCoef=new FXMatrix(tabSysCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
new FXLabel(matCoef,"Max speed coef :");
  new FXTextField(matCoef,6,&maxSpdCoef_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matCoef,"Max acc coef :");
  new FXTextField(matCoef,6,&maxAccCoef_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  //FXVerticalFrame *paramFrame = new FXVerticalFrame(tabSysCfg,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabSysCfg,"Deformation parameters");

  FXHorizontalFrame* deformParamFrame = new FXHorizontalFrame(tabSysCfg,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXMatrix *matOptionTraj=new FXMatrix(deformParamFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptionTraj,"reliability time :");
  new FXTextField(matOptionTraj,6,&reliabilityTime_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptionTraj,"epsilon RT :");
  new FXTextField(matOptionTraj,6,&epsilonRT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptionTraj,"prediction horizon :");
  new FXTextField(matOptionTraj,6,&predHorizon_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptionTraj,"max effective st dist :");
  new FXTextField(matOptionTraj,6,&maxEffectDist_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
//   FXCheckButton *connecChkBtn = new FXCheckButton(matOptionTraj,"Use connectivity checker",this,ID_CONNEC_CHK);
//   connecChkBtn->setCheck(true,false);
//   new FXLabel(matOptionTraj," ");
  FXCheckButton *tempCstcyBtn = new FXCheckButton(matOptionTraj,"Use temporal consistency",this,ID_TEMP_CSTCY);
  tempCstcyBtn->setCheck(true,false);
  new FXLabel(matOptionTraj," ");

  FXMatrix *matModeDeform=new FXMatrix(deformParamFrame,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matModeDeform,"Deformation Mode :");
  new FXRadioButton(matModeDeform,"Spatio-temporal deform",&optModeDeform_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matModeDeform,"Spatial deform",&optModeDeform_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matModeDeform,"Temporal deform",&optModeDeform_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  // Button to adjust parameters in function of the mode used
  new FXButton(matModeDeform,"Adjust parameters",NULL,this,ID_ADJ_PARAM,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matCoefForces =new FXMatrix(tabSysCfg,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matCoefForces,"internal forces :");
  new FXTextField(matCoefForces,6,&K_INT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"spt-tmp balance :");
  new FXTextField(matCoefForces,6,&K_SPT_TMP_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"external forces :");
  new FXTextField(matCoefForces,6,&K_EXT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"static ext forces :");
  new FXTextField(matCoefForces,6,&K_STEXT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"min spatial coef :");
  new FXTextField(matCoefForces,6,&alphaMin_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"max spatial coef :");
  new FXTextField(matCoefForces,6,&alphaMax_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"min temporal coef :");
  new FXTextField(matCoefForces,6,&betaMin_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"max temporal coef :");
  new FXTextField(matCoefForces,6,&betaMax_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"spatial max dist :");
  new FXTextField(matCoefForces,6,&maxSpDist_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCoefForces,"temporal max dist :");
  new FXTextField(matCoefForces,6,&maxTmpDist_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);  

  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  new FXHorizontalSeparator(tabSysCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabSysCfg,"Other options");

  FXMatrix *matRandMode=new FXMatrix(tabSysCfg,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  FXMatrix *matRandSeq=new FXMatrix(matRandMode,2,MATRIX_BY_COLUMNS|JUSTIFY_LEFT,0,0,0,0,0,0,0,0);
  new FXRadioButton(matRandSeq,"Random sequence",&optSeqNum_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  FXMatrix *matSeqNum=new FXMatrix(matRandMode,2,MATRIX_BY_COLUMNS|JUSTIFY_LEFT,0,0,0,0,0,0,0,0);
  new FXRadioButton(matSeqNum,"Sequence num : ",&optSeqNum_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXTextField(matSeqNum,10,&seqNum_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matRTAlgo=new FXMatrix(tabSysCfg,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXCheckButton *useRTAbtn = new FXCheckButton(matRTAlgo,"Use real-time algorithm",this,ID_USE_RTALGO);
  useRTAbtn->setCheck(TRUE);


  /***************************************************************************************************************
   ********************************************* Environment tab *************************************************
   ***************************************************************************************************************/
  
  new FXTabItem(panels,"Environment");

  FXVerticalFrame *tabEnv=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

//   FXMatrix *matMap=new FXMatrix(tabEnv,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
//   new FXLabel(matMap,"Map filename :");
//   new FXButton(matMap,"...",NULL,this,ID_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
//   new FXTextField(matMap,25,&mapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
 
//   // Button to load the map
//   new FXButton(tabEnv,"Load &Map",NULL,this,ID_LOAD_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matDfltMapDim=new FXMatrix(tabEnv,3,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matDfltMapDim,"Default map dimensions :");
  new FXTextField(matDfltMapDim,6,&dfltXMap_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXTextField(matDfltMapDim,6,&dfltYMap_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabEnv,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matObsList=new FXMatrix(tabEnv,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matObsList,"Obstacles list :");
  
  listHeader = new FXHeader(matObsList,this,ID_HEADER,HEADER_BUTTON|HEADER_RESIZE|FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X);

  listHeader->appendItem("ID Obs",NULL,45);
  listHeader->appendItem("X",NULL,45);
  listHeader->appendItem("Y",NULL,45);
  listHeader->appendItem("SX",NULL,45);
  listHeader->appendItem("SY",NULL,45);
  listHeader->appendItem("radius",NULL,45);
  listHeader->appendItem("init t",NULL,45);
  listHeader->appendItem("life t",NULL,45);
  listHeader->appendItem("rand",NULL,45);
  

  FXHorizontalFrame* listObsFame = new FXHorizontalFrame(matObsList,FRAME_SUNKEN|FRAME_THICK|LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0, 0,0);
  
  list[0]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[1]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[2]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[3]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[4]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[5]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[6]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[7]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[8]=new FXList(listObsFame,NULL,0,LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|LIST_SINGLESELECT,0,0,45,150);
  list[0]->setBackColor(FXRGB(255,240,240));
  list[1]->setBackColor(FXRGB(240,255,240));
  list[2]->setBackColor(FXRGB(240,240,255));
  list[3]->setBackColor(FXRGB(255,255,240));
  list[4]->setBackColor(FXRGB(255,240,255));
  list[5]->setBackColor(FXRGB(240,255,255));
  list[6]->setBackColor(FXRGB(255,255,255));
  list[7]->setBackColor(FXRGB(240,240,240));
  list[8]->setBackColor(FXRGB(255,240,240));

  new FXHorizontalSeparator(tabEnv,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabEnv,"New obstacle :");
  FXMatrix *matNewObs=new FXMatrix(tabEnv,5,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXTextField(matNewObs,6,&idObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Rand mode : ");
  new FXTextField(matNewObs,6,&randObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Rad :");
  new FXTextField(matNewObs,6,&rObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Position :");
  new FXLabel(matNewObs,"X :");
  new FXTextField(matNewObs,6,&xObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Y :");
  new FXTextField(matNewObs,6,&yObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Speed :");
  new FXLabel(matNewObs,"SX :");
  new FXTextField(matNewObs,6,&sxObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"SY :");
  new FXTextField(matNewObs,6,&syObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"Time :");
  new FXLabel(matNewObs,"init t :");
  new FXTextField(matNewObs,6,&initT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matNewObs,"life t :");
  new FXTextField(matNewObs,6,&lifeT_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);


  // Button to copy the map
  new FXButton(tabEnv,"&Add/Change Obstacle",NULL,this,ID_ADD_OBS,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(tabEnv,"&Remove Obstacle",NULL,this,ID_REM_OBS,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabEnv,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame *optObsFrame = new FXHorizontalFrame(tabEnv,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXVerticalFrame *randObsFrame = new FXVerticalFrame(optObsFrame,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(randObsFrame,"Obstacle random motion :");
  FXMatrix *matRandMotion=new FXMatrix(randObsFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matRandMotion,"Random threshold :");
  new FXTextField(matRandMotion,8,&seuilChgt_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matRandMotion,"Max angle change :");
  new FXTextField(matRandMotion,8,&maxAngleChgt_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matRandMotion,"Max norm change :");
  new FXTextField(matRandMotion,8,&maxNormChgt_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  FXVerticalFrame *initEnvFrame = new FXVerticalFrame(optObsFrame,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matInitEnv=new FXMatrix(initEnvFrame,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|JUSTIFY_LEFT,0,0,0,0,0,0,0,0);

  new FXLabel(matInitEnv,"Environment \n initialisation :");

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Dyn Obs",NULL,this,ID_INIT2,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Labyrinth",NULL,this,ID_INIT3,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

//   // Button to load the preconfigured obstacles
//   new FXButton(matInitEnv,"Init4",NULL,this,ID_INIT4,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Motorway",NULL,this,ID_INIT5,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Labth (RM)",NULL,this,ID_INIT6,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  

  // Button to load the preconfigured obstacles
  new FXButton(matInitEnv,"Intersection",NULL,this,ID_INIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matInitRandEnv=new FXMatrix(initEnvFrame,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|JUSTIFY_LEFT,0,0,0,0,0,0,0,0);
  
  // Button to load random obstacles
  new FXButton(matInitEnv,"InitRandObs",NULL,this,ID_INITRANDOBS,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5); 
  new FXTextField(matInitEnv,6,&nbRandObs_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matMap=new FXMatrix(tabEnv,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXLabel(matMap,"Map filename :");
  new FXButton(matMap,"...",NULL,this,ID_MAP_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matMap,25,&mapName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matMap,"");
  new FXLabel(matMap,"sample X : ");
  new FXTextField(matMap,8,&sampleXOccGrid_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matMap,"sample Y : ");
  new FXTextField(matMap,8,&sampleYOccGrid_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  

  // Button to convert the environment into an occupancy grid
  new FXButton(tabEnv,"Convert into Occupancy Grid",NULL,this,ID_CONVERT_OCCGRID,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to clear the environment
  new FXButton(tabEnv,"Clear Environment",NULL,this,ID_CLEAR_ENV,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   ******************************************** motion config tab ************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Motion config");

  FXVerticalFrame *tabMotionCfg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matTrajConfig=new FXMatrix(tabMotionCfg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTrajConfig,"Trajectory computing time step :");
  new FXTextField(matTrajConfig,6,&trajTimeStep_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTrajConfig,"Trajectory computing pos step :");
  new FXTextField(matTrajConfig,6,&trajPosStep_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  new FXRadioButton(tabMotionCfg,"Trajectory tracking",&typeMotion_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXVerticalFrame *trajPlngFrame=new FXVerticalFrame(tabMotionCfg,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matInputPath=new FXMatrix(trajPlngFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInputPath,"Input path filename :");
  new FXButton(matInputPath,"...",NULL,this,ID_PATH_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matInputPath,25,&inputPathName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);

  // Button to load the init path
  new FXButton(trajPlngFrame,"Load Path",NULL,this,ID_LOAD_PATH,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Button to compute the trajectory
  new FXButton(trajPlngFrame,"Compute Trajectory",NULL,this,ID_COMPUTE_TRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

   // Button to load a precomputed trajectory
  new FXButton(trajPlngFrame,"Load Precomputed Trajectory",NULL,this,ID_LOAD_PCPTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

//   // Button to compute the trajectory
//   new FXButton(trajPlngFrame,"Load Precomputed Trajectory",NULL,this,ID_LOAD_PCPPATH,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
//   // Button to compute a random trajectory by Runge Kutta
//   new FXButton(trajPlngFrame,"Compute CarLike Trajectory",NULL,this,ID_CPT_RANDRKTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

//   // Button to compute a trajectory for the intersection scenario
//   new FXButton(trajPlngFrame,"Compute Intersection Traj",NULL,this,ID_CPT_ITSCTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

//   // Button to compute a trajectory for the motorway scenario
//   new FXButton(trajPlngFrame,"Compute Motorway Traj",NULL,this,ID_CPT_MTWYTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

//   // Button to compute the diffdrive trajectory
//   new FXButton(trajPlngFrame,"Compute DiffDrive Trajectory",NULL,this,ID_CPT_DIFFDRIVETRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXLabel(trajPlngFrame,"Init trajectory cfg :");

  FXMatrix *matInitTrajCfg=new FXMatrix(trajPlngFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitTrajCfg,"X :");
  new FXTextField(matInitTrajCfg,6,&initXTraj_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitTrajCfg,"Y :");
  new FXTextField(matInitTrajCfg,6,&initYTraj_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitTrajCfg,"Th :");
  new FXTextField(matInitTrajCfg,6,&initThTraj_target,FXDataTarget::ID_VALUE,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitTrajCfg,"T :");
  new FXTextField(matInitTrajCfg,6,&initTTraj_target,FXDataTarget::ID_VALUE,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabMotionCfg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXRadioButton(tabMotionCfg,"Position tracking",&typeMotion_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXVerticalFrame *posTrckFrame=new FXVerticalFrame(tabMotionCfg,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXHorizontalFrame* rbPosFrame = new FXHorizontalFrame(posTrckFrame,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXVerticalFrame *rbIPosFrame=new FXVerticalFrame(rbPosFrame,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(rbIPosFrame,"Init robot position :");

  FXMatrix *matInitRbPos=new FXMatrix(rbIPosFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitRbPos,"X :");
  new FXTextField(matInitRbPos,6,&initXRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matInitRbPos,"Y :");
  new FXTextField(matInitRbPos,6,&initYRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  new FXLabel(matInitRbPos,"Th :");
  new FXTextField(matInitRbPos,6,&initThRb_target,FXDataTarget::ID_VALUE,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXVerticalFrame *rbGPosFrame=new FXVerticalFrame(rbPosFrame,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXCheckButton *rbGoalPosBtn = new FXCheckButton(rbGPosFrame,"Goal robot position :",this,ID_RB_GOALPOS);
  rbGoalPosBtn->setCheck(fixRbGoalPos,false);

  FXMatrix *matGoalRbPos=new FXMatrix(rbGPosFrame,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matGoalRbPos,"X :");
  new FXTextField(matGoalRbPos,6,&goalXRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(matGoalRbPos,"Y :");
  new FXTextField(matGoalRbPos,6,&goalYRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  new FXLabel(matGoalRbPos,"Th :");
  new FXTextField(matGoalRbPos,6,&goalThRb_target,FXDataTarget::ID_VALUE,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXRadioButton(posTrckFrame,"Point mass trajectories",&trajType_target,FXDataTarget::ID_OPTION+PM_TRAJ,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(posTrckFrame,"Car-like trajectories",&trajType_target,FXDataTarget::ID_OPTION+CL_TRAJ,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(posTrckFrame,"Wheelchair trajectories",&trajType_target,FXDataTarget::ID_OPTION+WC_TRAJ,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  // Button to init the position tracking mode
  new FXButton(posTrckFrame,"&Init Position Tracking",NULL,this,ID_INIT_POSTRACK,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
  // Button to display the reliabilty coefficient graph
  new FXButton(tabMotionCfg,"Reliability Coef Graph",NULL,this,ID_RT_GRAPH,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   ************************************ Trajectory Generation options tab ****************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Options");

  /***************************************** NH Path Planning options ********************************************/

  FXVerticalFrame *tabNHOptions=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabNHOptions,"NH Path Planning Options");

  FXMatrix *matOccGrid=new FXMatrix(tabNHOptions,2,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOccGrid,"Occ Grid filename :");
  new FXButton(matOccGrid,"...",NULL,this,ID_OCCGRID_NAME,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXTextField(matOccGrid,25,&occgridName_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);  

  /************************************** Trajectory Generation options ******************************************/

  new FXHorizontalSeparator(tabNHOptions,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabNHOptions,"Trajectory Generation Options");
  
  FXVerticalFrame *tabTGOptions=new FXVerticalFrame(tabNHOptions,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matOptCst=new FXMatrix(tabTGOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptCst,"Constraints used :");
  FXCheckButton *velCstBtn = new FXCheckButton(matOptCst,"lin velocity cst",this,ID_VEL_CST);
  velCstBtn->setCheck(true,false);
  FXCheckButton *accCstBtn = new FXCheckButton(matOptCst,"lin acceleration cst",this,ID_ACC_CST);
  accCstBtn->setCheck(true,false);
  FXCheckButton *avelCstBtn = new FXCheckButton(matOptCst,"ang velocity cst",this,ID_AVEL_CST);
  avelCstBtn->setCheck(true,false);
  FXCheckButton *aaccCstBtn = new FXCheckButton(matOptCst,"ang acceleration cst",this,ID_AACC_CST);

  FXMatrix *matOptPrec=new FXMatrix(tabTGOptions,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptPrec,"Speed / Precision :");
  new FXLabel(matOptPrec,"");
  new FXLabel(matOptPrec,"Max number of steps:");
  new FXTextField(matOptPrec,6,&maxNbStep_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptPrec,"Max residual:");
  new FXTextField(matOptPrec,6,&maxResidual_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  FXMatrix *matOptDbg=new FXMatrix(tabTGOptions,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptDbg,"Debugging mode :");
  new FXRadioButton(matOptDbg,"0",&debugLevel_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"1",&debugLevel_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"2",&debugLevel_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matOptDbg,"3",&debugLevel_target,FXDataTarget::ID_OPTION+3,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);


  /************************************ Parallelization Generation options ***************************************/

  new FXHorizontalSeparator(tabNHOptions,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabNHOptions,"Parallelization Options");
  
  FXMatrix *matParallelOpt=new FXMatrix(tabNHOptions,2,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  FXCheckButton *parallelBtn = new FXCheckButton(matParallelOpt,"Parallelization",this,ID_PARALLEL);
  //parallelBtn->setCheck(true,false);
  new FXLabel(matParallelOpt,"");
  new FXLabel(matParallelOpt,"Nb threads : ");
  new FXTextField(matParallelOpt,6,&nbThread_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  

  /***************************************************************************************************************
   **************************************** trajectory deformation tab *******************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Traj deformation");

  FXVerticalFrame *tabTrajPl=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matCmdTraj=new FXMatrix(tabTrajPl,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  // slider of the time
  slider = new FXRealSlider(matCmdTraj,&time_target,FXDataTarget::ID_VALUE,LAYOUT_CENTER_Y|LAYOUT_FILL_ROW|LAYOUT_FIX_WIDTH,0,0,250,0,0,0,0,0);
  slider->setRange(0.0,1.0);
  new FXTextField(matCmdTraj,6,&time_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  // Button to start the trajectory
  new FXButton(matCmdTraj,"Start",NULL,this,ID_START,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to compute just one step of the trajectory deformation
  new FXButton(matCmdTraj,"One step",NULL,this,ID_ONESTEP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to stop the trajectory
  new FXButton(matCmdTraj,"Stop",NULL,this,ID_STOP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to restart the trajectory
  new FXButton(matCmdTraj,"Restart",NULL,this,ID_RESTART,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabTrajPl,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame* viewFrame = new FXHorizontalFrame(tabTrajPl,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  FXMatrix *matViewTraj=new FXMatrix(viewFrame,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matViewTraj, "Point of view : ");
  new FXButton(matViewTraj,"Time view",NULL,this,ID_TIME_VIEW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matViewTraj,"Space view",NULL,this,ID_SPACE_VIEW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matViewTraj,"Space-Time view",NULL,this,ID_PERSP_VIEW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matPerspOrth=new FXMatrix(viewFrame,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matPerspOrth, "View Mode : ");
  viewModeBox = new FXLabel(matPerspOrth, "Perspective View");
  new FXButton(matPerspOrth,"Perspective view",NULL,this,ID_PERSP_MODE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matPerspOrth,"Orthogonal view",NULL,this,ID_ORTH_MODE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXMatrix *matOptViewTraj=new FXMatrix(tabTrajPl,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptViewTraj, "Display Options : ");
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Display future model",this,ID_DISP_FUTURE);
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Display forces",this,ID_DISP_FORCES);
  new FXTextField(matOptViewTraj,6,&indDispForces_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXCheckButton(matOptViewTraj,"Display all forces",this,ID_DISP_ALLFORCES);
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Display nearest dyn obs states",this,ID_DISP_NEAR_ST);
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Display nearest st obs states",this,ID_DISP_NEAR_STST);
  new FXLabel(matOptViewTraj, "");
  new FXCheckButton(matOptViewTraj,"Pause when collision/lose of connectivity",this,ID_PAUSE_COL_LOC);
  new FXLabel(matOptViewTraj, "");
  FXCheckButton *drawMBbtn = new FXCheckButton(matOptViewTraj,"Draw map boundaries",this,ID_DRAW_MAPBOUND);
  drawMBbtn->setCheck(drawMapBoundaries);
  new FXLabel(matOptViewTraj, "");
  FXCheckButton *dynViewBtn = new FXCheckButton(matOptViewTraj,"Dynamic view",this,ID_DYNVIEW);
  dynViewBtn->setCheck(dynamicView);
  new FXLabel(matOptViewTraj, "");
  FXCheckButton* rmviewBtn = new FXCheckButton(matOptViewTraj,"RoadMap view",this,ID_ROADMAPVIEW);
  rmviewBtn->setCheck(roadmapView);
  new FXLabel(matOptViewTraj, "");
  FXCheckButton* drawGenTrajBtn = new FXCheckButton(matOptViewTraj,"Draw generated trajectories",this,ID_DRAW_TRAJGEN);
  new FXLabel(matOptViewTraj, "");

  new FXHorizontalSeparator(tabTrajPl,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matFPS=new FXMatrix(tabTrajPl,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matFPS,"Medium FPS :");
  new FXTextField(matFPS,6,&FPS_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matFPS,"Last FPS :");
  new FXTextField(matFPS,6,&lastFPS_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matFPS,"Deformation duration :");
  new FXTextField(matFPS,6,&deformDur_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matFPS,"Nb of state-times :");
  new FXTextField(matFPS,6,&nbST_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matFPS,"Tot nb of state-times :");
  new FXTextField(matFPS,6,&nbSTtot_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);


  /***************************************************************************************************************
   ***************************************** trajectory following tab ********************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Traj following");
  
  FXVerticalFrame *tabTrajFlwg=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabTrajFlwg,"Traj following commands:");

  FXMatrix *matCmdTF=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  // slider of the time
  slider = new FXRealSlider(matCmdTF,&time_target,FXDataTarget::ID_VALUE,LAYOUT_CENTER_Y|LAYOUT_FILL_ROW|LAYOUT_FIX_WIDTH,0,0,250,0,0,0,0,0);
  slider->setRange(0.0,1.0);
  new FXTextField(matCmdTF,6,&time_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|FRAME_SUNKEN|FRAME_THICK);
  // Button to start the trajectory
  new FXButton(matCmdTF,"Start",NULL,this,ID_START_TF,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to compute just one step of the trajectory deformation
  new FXButton(matCmdTF,"One step",NULL,this,ID_ONESTEP_TF,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to stop the trajectory
  new FXButton(matCmdTF,"Stop",NULL,this,ID_STOP_TF,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  // Button to restart the trajectory
  new FXButton(matCmdTF,"Restart",NULL,this,ID_RESTART_TF,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabTrajFlwg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabTrajFlwg,"Traj following parameters:");

  FXMatrix *matTFparam=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTFparam,"DT :");
  new FXTextField(matTFparam,6,&TF_dt_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTFparam,"T meet :");
  new FXTextField(matTFparam,6,&TF_timeMeeting_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  FXCheckButton* dispTG_TF_Btn = new FXCheckButton(matTFparam,"Disp trajgen",this,ID_DRAW_TRAJGEN_TF);  
  dispTG_TF_Btn->setCheck(TRUE);
  new FXLabel(matTFparam,"");

  new FXLabel(tabTrajFlwg,"Traj Following Mode:");
  FXMatrix *matTFmode=new FXMatrix(tabTrajFlwg,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXRadioButton(matTFmode,"Car-like Traj Follow",&TF_mode_target,FXDataTarget::ID_OPTION+CL_TF_MODE,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matTFmode,"Cycab Traj Follow",&TF_mode_target,FXDataTarget::ID_OPTION+CC_TF_MODE,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matTFmode,"WheelChair Traj Follow",&TF_mode_target,FXDataTarget::ID_OPTION+WC_TF_MODE,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);  

  new FXButton(tabTrajFlwg,"Init Traj Following",NULL,this,ID_INIT_TFOLLOW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXButton(tabTrajFlwg,"Display TF Traj",NULL,this,ID_DISP_TFTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXButton(tabTrajFlwg,"Display Join Traj",NULL,this,ID_DISP_JOINTRAJ,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabTrajFlwg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabTrajFlwg,"Init Robot State:");

  FXMatrix *matInitRbState=new FXMatrix(tabTrajFlwg,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matInitRbState,"X :");
  new FXTextField(matInitRbState,6,&initXRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitRbState,"Y :");
  new FXTextField(matInitRbState,6,&initYRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitRbState,"Th :");
  new FXTextField(matInitRbState,6,&initThRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitRbState,"V :");
  new FXTextField(matInitRbState,6,&initVRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matInitRbState,"Ph :");
  new FXTextField(matInitRbState,6,&initPhRb_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabTrajFlwg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabTrajFlwg,"Max State Noise:");

  FXMatrix *matStateNoise=new FXMatrix(tabTrajFlwg,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matStateNoise,"XY noise:");
  new FXTextField(matStateNoise,6,&noiseX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matStateNoise,"Y noise:");
  new FXTextField(matStateNoise,6,&noiseY_target,FXDataTarget::ID_VALUE,TEXTFIELD_INTEGER|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matStateNoise,"Th noise:");
  new FXTextField(matStateNoise,6,&noiseTH_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matStateNoise,"V noise:");
  new FXTextField(matStateNoise,6,&noiseV_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matStateNoise,"Ph noise:");
  new FXTextField(matStateNoise,6,&noisePH_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabTrajFlwg,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabTrajFlwg,"Max Control Noise:");

  FXMatrix *matControlNoise=new FXMatrix(tabTrajFlwg,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matControlNoise,"Acc noise:");
  new FXTextField(matControlNoise,6,&noiseAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matControlNoise,"SS noise:");
  new FXTextField(matControlNoise,6,&noiseSS_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


  /***************************************************************************************************************
   ********************************************* general commands ************************************************
   ***************************************************************************************************************/

  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Exit button
  new FXButton(buttonFrame,"&Exit",NULL,getApp(),FXApp::ID_QUIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
}

// destructor
MPC_TeddyIfc::~MPC_TeddyIfc()
{
  if(env_ != NULL) delete env_;
  if(path_ != NULL) delete path_;
  if(traj_ != NULL) delete traj_;

  if(glcanvas != NULL) delete glcanvas;
  if(glvisual != NULL) delete glvisual; 
}

// Create and initialize
void MPC_TeddyIfc::create(){
  FXMainWindow::create();
  show(PLACEMENT_SCREEN);
}

// intersections environment
void MPC_TeddyIfc::initObs()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
      this->env_->clearObs2D();
      this->env_->clearFixedObs2D();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(1370);
  this->env_->setDfltY(900);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());
  
  // create the static obstacles

  BBoxObs2D stObs1(PositionXYThT(400,200,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs2(PositionXYThT(500,200,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs3(PositionXYThT(400,700,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs4(PositionXYThT(500,700,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);

  BBoxObs2D stObs5(PositionXYThT(215,400,0,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs6(PositionXYThT(685,400,0,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs7(PositionXYThT(215,500,0,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs8(PositionXYThT(685,500,0,0),SpeedXYTh(0,0,0),400,30);

  BBoxObs2D stObs9(PositionXYThT(870,200,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs10(PositionXYThT(970,200,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs11(PositionXYThT(870,700,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs12(PositionXYThT(970,700,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);

  BBoxObs2D stObs13(PositionXYThT(1155,400,0,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs14(PositionXYThT(1155,500,0,0),SpeedXYTh(0,0,0),400,30);

//   PMConfig pos1(400,200);
//   Obstacle2D* stObs1 = new ObsRect2D(pos1,400, 30.0, 0.5*M_PI);
//   PMConfig pos2(500,200);
//   Obstacle2D* stObs2 = new ObsRect2D(pos2,400, 30.0, 0.5*M_PI);
//   PMConfig pos3(400,700);
//   Obstacle2D* stObs3 = new ObsRect2D(pos3,400, 30.0, 0.5*M_PI);
//   PMConfig pos4(500,700);
//   Obstacle2D* stObs4 = new ObsRect2D(pos4,400, 30.0, 0.5*M_PI);

//   PMConfig pos5(215,400);
//   Obstacle2D* stObs5 = new ObsRect2D(pos5,400, 30.0, 0.0);
//   PMConfig pos6(685,400);
//   Obstacle2D* stObs6 = new ObsRect2D(pos6,400, 30.0, 0.0);
//   PMConfig pos7(215,500);
//   Obstacle2D* stObs7 = new ObsRect2D(pos7,400, 30.0, 0.0);
//   PMConfig pos8(685,500);
//   Obstacle2D* stObs8 = new ObsRect2D(pos8,400, 30.0, 0.0);

//   PMConfig pos9(870,200);
//   Obstacle2D* stObs9 = new ObsRect2D(pos9,400, 30.0, 0.5*M_PI);
//   PMConfig pos10(970,200);
//   Obstacle2D* stObs10 = new ObsRect2D(pos10,400, 30.0, 0.5*M_PI);
//   PMConfig pos11(870,700);
//   Obstacle2D* stObs11 = new ObsRect2D(pos11,400, 30.0, 0.5*M_PI);
//   PMConfig pos12(970,700);
//   Obstacle2D* stObs12 = new ObsRect2D(pos12,400, 30.0, 0.5*M_PI);

//   PMConfig pos13(1155,400);
//   Obstacle2D* stObs13 = new ObsRect2D(pos13,400, 30.0, 0.0);
//   PMConfig pos14(1155,500);
//   Obstacle2D* stObs14 = new ObsRect2D(pos14,400, 30.0, 0.0);

  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);
  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);
  this->addStObstacle(FXString("stObs6"),stObs6);
  this->addStObstacle(FXString("stObs7"),stObs7);
  this->addStObstacle(FXString("stObs8"),stObs8);

  this->addStObstacle(FXString("stObs9"),stObs9);
  this->addStObstacle(FXString("stObs10"),stObs10);
  this->addStObstacle(FXString("stObs11"),stObs11);
  this->addStObstacle(FXString("stObs12"),stObs12);
  this->addStObstacle(FXString("stObs13"),stObs13);
  this->addStObstacle(FXString("stObs14"),stObs14);

//   // create the dynamic obstacles
//   ObsDisk2D* obs1  = new ObsDisk2D(PMConfig(450,375), Speed2D(0,20),  0,  500, 20);
//   ObsDisk2D* obs2  = new ObsDisk2D(PMConfig(450,250), Speed2D(0,20),  0,  500, 20);
//   ObsDisk2D* obs3  = new ObsDisk2D(PMConfig(450,125),  Speed2D(0,20),  0,  500, 20);
//   ObsDisk2D* obs4  = new ObsDisk2D(PMConfig(450,0),   Speed2D(0,20),  3,  500, 20);
//   ObsDisk2D* obs5  = new ObsDisk2D(PMConfig(920,500), Speed2D(0,-20), 0,  500, 20);
//   ObsDisk2D* obs6  = new ObsDisk2D(PMConfig(920,650), Speed2D(0,-20), 0,  500, 20);
//   ObsDisk2D* obs7  = new ObsDisk2D(PMConfig(920,800), Speed2D(0,-20), 0,  500, 20);
//   ObsDisk2D* obs8  = new ObsDisk2D(PMConfig(920,800), Speed2D(0,-20), 6,  500, 20);
//   ObsDisk2D* obs9  = new ObsDisk2D(PMConfig(920,800), Speed2D(0,-20), 12, 500, 20);
//   ObsDisk2D* obs10 = new ObsDisk2D(PMConfig(920,800), Speed2D(0,-20), 18, 500, 20);

//   DiskObs2D obs1(PositionXYT(450,375,0),SpeedXY,20);

//   // append them to the list of obstacles
//   this->addObstacle(FXString("obs1"),obs1,true);
//   this->addObstacle(FXString("obs2"),obs2,true);
//   this->addObstacle(FXString("obs3"),obs3,true);
//   this->addObstacle(FXString("obs4"),obs4,true);
//   this->addObstacle(FXString("obs5"),obs5,true);
//   this->addObstacle(FXString("obs6"),obs6,true);
//   this->addObstacle(FXString("obs7"),obs7,true);
//   this->addObstacle(FXString("obs8"),obs8,true);
//   this->addObstacle(FXString("obs9"),obs9,true);
//   this->addObstacle(FXString("obs10"),obs10,true);

  // refresh the viewer
  this->drawScene();
  
  // init the random
  // initRandomDeform();

  msgBox->setText("Environment \"Intersection\" loaded.");

  // compute the intersection trajectory
  orthoView = true;
  dynamicView = true;
  this->computeIntersectionTraj();
}

// dynamic obstacles environment
void MPC_TeddyIfc::initObs2()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
//       this->env_->clearObs();
      this->env_->clearObs2D();
      this->env_->clearFixedObs2D();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(700);
  this->env_->setDfltY(700);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());

  // create the dynamic obstacles
  DiskObs2D obs1(PositionXYT(150,187,0), SpeedXY(2,-8), 10);
  DiskObs2D obs2(PositionXYT(75,0,0), SpeedXY(10,5), 13);
  DiskObs2D obs3(PositionXYT(263,38,0), SpeedXY(-12,7), 17);

  // append it to the list of obstacles
  this->addObstacle(FXString("obs1"),obs1,true);
  this->addObstacle(FXString("obs2"),obs2,true);
  this->addObstacle(FXString("obs3"),obs3,true);
  
  // init the random
  initRandomDeform();

  // refresh the viewer
  this->setLookAt(env_,NULL,NULL);
  this->drawScene();

  msgBox->setText("Environment 3 loaded (car-like system environment).");
}

// labyrinth with dynamic obstacles environment
void MPC_TeddyIfc::initObs3()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
//       this->env_->clearObs();
      this->env_->clearObs2D();
      this->env_->clearFixedObs2D();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(700);
  this->env_->setDfltY(700);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());

  // borders obstacles
  BBoxObs2D stObs1(PositionXYThT(-50.0,0.5*dfltYMap,0.5*M_PI,0),SpeedXYTh(0,0,0),(double)dfltYMap + 200.0, 100.0);
  BBoxObs2D stObs2(PositionXYThT((double)dfltXMap + 50.0,0.5*dfltYMap,0.5*M_PI,0),SpeedXYTh(0,0,0),(double)dfltYMap + 200.0, 100.0);
  BBoxObs2D stObs3(PositionXYThT(0.5*dfltXMap,-50.0,0.0,0),SpeedXYTh(0,0,0),(double)dfltXMap, 100.0);
  BBoxObs2D stObs4(PositionXYThT(0.5*dfltXMap,(double)dfltYMap + 50.0,0.0,0),SpeedXYTh(0,0,0),(double)dfltXMap, 100.0);

  // inside obstacles
  BBoxObs2D stObs5(PositionXYThT(100,175,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs6(PositionXYThT(350,200,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs7(PositionXYThT(250,350,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs8(PositionXYThT(450,200,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs9(PositionXYThT(225,600,0.5*M_PI,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs10(PositionXYThT(175,515,0.0,0),SpeedXYTh(0,0,0),100,30);
  BBoxObs2D stObs11(PositionXYThT(305,585,0.0,0),SpeedXYTh(0,0,0),150,30);
  BBoxObs2D stObs12(PositionXYThT(625,400,0.0,0),SpeedXYTh(0,0,0),150,30);
  BBoxObs2D stObs13(PositionXYThT(565,515,0.5*M_PI,0),SpeedXYTh(0,0,0),200,30);

  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);
  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);
  this->addStObstacle(FXString("stObs6"),stObs6);
  this->addStObstacle(FXString("stObs7"),stObs7);
  this->addStObstacle(FXString("stObs8"),stObs8);
  this->addStObstacle(FXString("stObs9"),stObs9);
  this->addStObstacle(FXString("stObs10"),stObs10);
  this->addStObstacle(FXString("stObs11"),stObs11);
  this->addStObstacle(FXString("stObs12"),stObs12);
  this->addStObstacle(FXString("stObs13"),stObs13);
  
  // init the random
  initRandomDeform();

  // refresh the viewer
  this->setLookAt(env_,NULL,NULL);
  this->drawScene();

  msgBox->setText("Environment 3 loaded (car-like system environment).");
}

// not used...
void MPC_TeddyIfc::initObs4()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
//       this->env_->clearObs();
      this->env_->clearObs2D();
      this->env_->clearFixedObs2D();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(700);
  this->env_->setDfltY(700);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());

  for(int i=0; i<40; i++)
    {
      double x = getIntRandom(0,dfltXMap);
      double y = getIntRandom(0,dfltYMap);
      double sx = getIntRandom(-30,30);
      double sy = getIntRandom(-30,30);
      double initTime = getIntRandom(0,15);
      double lifeTime = getIntRandom(50,200);
      double sizeObs = getIntRandom(10,20);
      
      // create the obstacles
      //ObsDisk2D* newObs = new ObsDisk2D(PMConfig(x,y), Speed2D(sx,sy), initTime, lifeTime, sizeObs);
      DiskObs2D newObs(PositionXYT(x,y,0),SpeedXY(sx,sy),sizeObs);
      
      // append it to the list of obstacles
      this->addObstacle(FXString("obs"),newObs,true);
    }

  // create the static obstacles
  // borders obstacles
//   PMConfig pos1(-5.0,0.5*dfltYMap);
//   Obstacle2D* stObs1 = new ObsRect2D(pos1,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
//   PMConfig pos2((double)dfltXMap + 5.0,0.5*dfltYMap);
//   Obstacle2D* stObs2 = new ObsRect2D(pos2,(double)dfltYMap + 20.0, 10.0, 0.5*M_PI);
//   PMConfig pos3(0.5*dfltXMap,-5.0);
//   Obstacle2D* stObs3 = new ObsRect2D(pos3,(double)dfltXMap, 10.0, 0.0);
//   PMConfig pos4(0.5*dfltXMap,(double)dfltYMap + 5.0);
//   Obstacle2D* stObs4 = new ObsRect2D(pos4,(double)dfltXMap, 10.0, 0.0);

//   // inside obstacles
//   PMConfig pos5(100,175);
//   Obstacle2D* stObs5 = new ObsRect2D(pos5,200, 30.0, 0.0);
//   PMConfig pos6(350,200);
//   Obstacle2D* stObs6 = new ObsRect2D(pos6,400, 30.0, 0.5*M_PI);
//   PMConfig pos7(250,350);
//   Obstacle2D* stObs7 = new ObsRect2D(pos7,200, 30.0, 0.0);
//   PMConfig pos8(450,200);
//   Obstacle2D* stObs8 = new ObsRect2D(pos8,200, 30.0, 0.0);
//   PMConfig pos9(225,600);
//   Obstacle2D* stObs9 = new ObsRect2D(pos9,200, 30.0, 0.5*M_PI);
//   PMConfig pos10(175,515);
//   Obstacle2D* stObs10 = new ObsRect2D(pos10,100, 30.0, 0.0);
//   PMConfig pos11(305,585);
//   Obstacle2D* stObs11 = new ObsRect2D(pos11,150, 30.0, 0.0);
//   PMConfig pos12(625,400);
//   Obstacle2D* stObs12 = new ObsRect2D(pos12,150, 30.0, 0.0);
//   PMConfig pos13(565,515);
//   Obstacle2D* stObs13 = new ObsRect2D(pos13,200, 30.0, 0.5*M_PI);

  // borders obstacles
  BBoxObs2D stObs1(PositionXYThT(-50.0,0.5*dfltYMap,0.5*M_PI,0),SpeedXYTh(0,0,0),(double)dfltYMap + 200.0, 100.0);
  BBoxObs2D stObs2(PositionXYThT((double)dfltXMap + 50.0,0.5*dfltYMap,0.5*M_PI,0),SpeedXYTh(0,0,0),(double)dfltYMap + 200.0, 100.0);
  BBoxObs2D stObs3(PositionXYThT(0.5*dfltXMap,-50.0,0.0,0),SpeedXYTh(0,0,0),(double)dfltXMap, 100.0);
  BBoxObs2D stObs4(PositionXYThT(0.5*dfltXMap,(double)dfltYMap + 50.0,0.0,0),SpeedXYTh(0,0,0),(double)dfltXMap, 100.0);

  // inside obstacles
  BBoxObs2D stObs5(PositionXYThT(100,175,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs6(PositionXYThT(350,200,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs7(PositionXYThT(250,350,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs8(PositionXYThT(450,200,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs9(PositionXYThT(225,600,0.5*M_PI,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs10(PositionXYThT(175,515,0.0,0),SpeedXYTh(0,0,0),100,30);
  BBoxObs2D stObs11(PositionXYThT(305,585,0.0,0),SpeedXYTh(0,0,0),150,30);
  BBoxObs2D stObs12(PositionXYThT(625,400,0.0,0),SpeedXYTh(0,0,0),150,30);
  BBoxObs2D stObs13(PositionXYThT(565,515,0.5*M_PI,0),SpeedXYTh(0,0,0),200,30);

  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);
  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);
  this->addStObstacle(FXString("stObs6"),stObs6);
  this->addStObstacle(FXString("stObs7"),stObs7);
  this->addStObstacle(FXString("stObs8"),stObs8);
  this->addStObstacle(FXString("stObs9"),stObs9);
  this->addStObstacle(FXString("stObs10"),stObs10);
  this->addStObstacle(FXString("stObs11"),stObs11);
  this->addStObstacle(FXString("stObs12"),stObs12);
  this->addStObstacle(FXString("stObs13"),stObs13);
  
  // init the random
  initRandomDeform();

  // compute the roadmap on the environment
  roadmap_ = new DRoadMap(this->env_);
  roadmap_->setNbNodesNeeded(700);
  roadmap_->setMaxDistNgh(50.0);
  roadmap_->construct();

  // refresh the viewer
  this->setLookAt(env_,NULL,NULL);
  this->drawScene();

  msgBox->setText("Environment 6 loaded.");
}

// motorway environment
void MPC_TeddyIfc::initObs5()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
//       this->env_->clearObs();
      this->env_->clearObs2D();
      this->env_->clearFixedObs2D();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(1420);
  this->env_->setDfltY(680);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());
  
  // create the static obstacles

//   PMConfig pos1(610,515);
//   Obstacle2D* stObs1 = new ObsRect2D(pos1,1220, 30.0, 0.0);
//   PMConfig pos2(920,400);
//   Obstacle2D* stObs2 = new ObsRect2D(pos2,600, 30.0, 0.01*M_PI);
//   PMConfig pos3(245,400);
//   Obstacle2D* stObs3 = new ObsRect2D(pos3,490, 30.0, 0.0);

//   PMConfig pos4(282,297);
//   Obstacle2D* stObs4 = new ObsRect2D(pos4,450, 30.0, 0.15*M_PI);
//   PMConfig pos5(360,258);
//   Obstacle2D* stObs5 = new ObsRect2D(pos5,600, 30.0, 0.15*M_PI);

  BBoxObs2D stObs1(PositionXYThT(610,515,0.0,0),SpeedXYTh(0,0,0),1220,30);
  BBoxObs2D stObs2(PositionXYThT(920,400,0.01*M_PI,0),SpeedXYTh(0,0,0),600,30);
  BBoxObs2D stObs3(PositionXYThT(245,400,0.0,0),SpeedXYTh(0,0,0),490,30);
  BBoxObs2D stObs4(PositionXYThT(282,297,0.15*M_PI,0),SpeedXYTh(0,0,0),450,30);
  BBoxObs2D stObs5(PositionXYThT(360,258,0.15*M_PI,0),SpeedXYTh(0,0,0),600,30);

  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);

  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);

//   // create the dynamic obstacles
//   ObsDisk2D* obs1  = new ObsDisk2D(PMConfig(330,436), Speed2D(32,0),  0,    27.5, 20);
//   ObsDisk2D* obs2  = new ObsDisk2D(PMConfig(180,436),  Speed2D(32,0),  0,    32,   20);
//   ObsDisk2D* obs3  = new ObsDisk2D(PMConfig(30,436),   Speed2D(32,0),  0,    37.5, 20);
//   ObsDisk2D* obs4  = new ObsDisk2D(PMConfig(0,  436), Speed2D(32,0),  3.5,  37.5,   20);
//   ObsDisk2D* obs5  = new ObsDisk2D(PMConfig(0,  436), Speed2D(32,0),  8,   37.5, 20);
//   ObsDisk2D* obs6  = new ObsDisk2D(PMConfig(350,479), Speed2D(40,0),  0,    28,   20);
//   ObsDisk2D* obs7  = new ObsDisk2D(PMConfig(200,479), Speed2D(40,0),  0,    32,   20);
//   ObsDisk2D* obs8  = new ObsDisk2D(PMConfig(50, 479), Speed2D(40,0),  0,    36,   20);
//   ObsDisk2D* obs9  = new ObsDisk2D(PMConfig(0,  479), Speed2D(40,0),  2.5,  37.5,   20);
//   ObsDisk2D* obs10 = new ObsDisk2D(PMConfig(0,  479), Speed2D(40,0),  6.5,  37.5,   20);

//   // append them to the list of obstacles
//   this->addObstacle(FXString("obs1"),obs1,true);
//   this->addObstacle(FXString("obs2"),obs2,true);
//   this->addObstacle(FXString("obs3"),obs3,true);
//   this->addObstacle(FXString("obs4"),obs4,true);
//   this->addObstacle(FXString("obs5"),obs5,true);
//   this->addObstacle(FXString("obs6"),obs6,true);
//   this->addObstacle(FXString("obs7"),obs7,true);
//   this->addObstacle(FXString("obs8"),obs8,true);
//   this->addObstacle(FXString("obs9"),obs9,true);
//   this->addObstacle(FXString("obs10"),obs10,true);

  // refresh the viewer
  this->setLookAt(env_,NULL,NULL);
  this->drawScene();
  
  // init the random
  // initRandomDeform();

  msgBox->setText("Environment \"Motorway\" loaded.");

  // compute the intersection trajectory
  orthoView = true;
  dynamicView = true;
  this->computeMotorwayTraj();
}

// labyrinth with static obstacles and roadmap environment
void MPC_TeddyIfc::initObs6()
{
  // clear the current environment
  if(this->env_ != NULL)
    {
//       this->env_->clearObs();
      this->env_->clearObs2D();
      this->env_->clearFixedObs2D();
    }

  // increase the dimensions of the environment
  this->env_->setDfltX(700);
  this->env_->setDfltY(700);

  // adjust the dimensions of the viewer
  dfltXMap = (int)(this->env_->getDfltX());
  dfltYMap = (int)(this->env_->getDfltY());

  // create the dynamic obstacles
  DiskObs2D obs1(PositionXYT(150,187,0), SpeedXY(12,-10), 10);
  DiskObs2D obs2(PositionXYT(75,0,0), SpeedXY(20,6), 13);
  DiskObs2D obs3(PositionXYT(263,38,0), SpeedXY(-16,10), 17);
  DiskObs2D obs4(PositionXYT(38,315,0), SpeedXY(20,-8), 10);
  DiskObs2D obs5(PositionXYT(170,263,0), SpeedXY(10,16), 13);
  DiskObs2D obs6(PositionXYT(300,150,0), SpeedXY(-20,-4), 17);
  DiskObs2D obs7(PositionXYT(38,150,0), SpeedXY(16,10), 10);
  DiskObs2D obs8(PositionXYT(187,105,0), SpeedXY(0,24), 13);
  DiskObs2D obs9(PositionXYT(285,338,0), SpeedXY(-14,-14), 17);
  DiskObs2D obs10(PositionXYT(338,225,0), SpeedXY(12,-12), 10);

  DiskObs2D obs11(PositionXYT(550,287,0), SpeedXY(-12,10), 10);
  DiskObs2D obs12(PositionXYT(75,600,0), SpeedXY(-6,-20), 13);
  DiskObs2D obs13(PositionXYT(363,538,0), SpeedXY(16,-10), 17);
  DiskObs2D obs14(PositionXYT(238,615,0), SpeedXY(-20,8), 10);
  DiskObs2D obs15(PositionXYT(570,563,0), SpeedXY(-10,-16), 13);
  DiskObs2D obs16(PositionXYT(650,150,0), SpeedXY(20,4), 17);
  DiskObs2D obs17(PositionXYT(438,450,0), SpeedXY(-16,-10), 10);
  DiskObs2D obs18(PositionXYT(287,405,0), SpeedXY(0,-24), 13);
  DiskObs2D obs19(PositionXYT(485,598,0), SpeedXY(14,14), 17);
  DiskObs2D obs20(PositionXYT(638,325,0), SpeedXY(-12,12), 10);

  // append it to the list of obstacles
  this->addObstacle(FXString("obs1"),obs1,true);
  this->addObstacle(FXString("obs2"),obs2,true);
  this->addObstacle(FXString("obs3"),obs3,true);
  this->addObstacle(FXString("obs4"),obs4,true);
  this->addObstacle(FXString("obs5"),obs5,true);
  this->addObstacle(FXString("obs6"),obs6,true);
  this->addObstacle(FXString("obs7"),obs7,true);
  this->addObstacle(FXString("obs8"),obs8,true);
  this->addObstacle(FXString("obs9"),obs9,true);
  this->addObstacle(FXString("obs10"),obs10,true);

  this->addObstacle(FXString("obs11"),obs11,true);
  this->addObstacle(FXString("obs12"),obs12,true);
  this->addObstacle(FXString("obs13"),obs13,true);
  this->addObstacle(FXString("obs14"),obs14,true);
  this->addObstacle(FXString("obs15"),obs15,true);
  this->addObstacle(FXString("obs16"),obs16,true);
  this->addObstacle(FXString("obs17"),obs17,true);
  this->addObstacle(FXString("obs18"),obs18,true);
  this->addObstacle(FXString("obs19"),obs19,true);
  this->addObstacle(FXString("obs20"),obs20,true);

  // create the static obstacles

  // borders obstacles
  BBoxObs2D stObs1(PositionXYThT(-50.0,0.5*dfltYMap,0.5*M_PI,0),SpeedXYTh(0,0,0),(double)dfltYMap + 200.0, 100.0);
  BBoxObs2D stObs2(PositionXYThT((double)dfltXMap + 50.0,0.5*dfltYMap,0.5*M_PI,0),SpeedXYTh(0,0,0),(double)dfltYMap + 200.0, 100.0);
  BBoxObs2D stObs3(PositionXYThT(0.5*dfltXMap,-50.0,0.0,0),SpeedXYTh(0,0,0),(double)dfltXMap, 100.0);
  BBoxObs2D stObs4(PositionXYThT(0.5*dfltXMap,(double)dfltYMap + 50.0,0.0,0),SpeedXYTh(0,0,0),(double)dfltXMap, 100.0);

  // inside obstacles
  BBoxObs2D stObs5(PositionXYThT(100,175,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs6(PositionXYThT(350,200,0.5*M_PI,0),SpeedXYTh(0,0,0),400,30);
  BBoxObs2D stObs7(PositionXYThT(250,350,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs8(PositionXYThT(450,200,0.0,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs9(PositionXYThT(225,600,0.5*M_PI,0),SpeedXYTh(0,0,0),200,30);
  BBoxObs2D stObs10(PositionXYThT(175,515,0.0,0),SpeedXYTh(0,0,0),100,30);
  BBoxObs2D stObs11(PositionXYThT(305,585,0.0,0),SpeedXYTh(0,0,0),150,30);
  BBoxObs2D stObs12(PositionXYThT(625,400,0.0,0),SpeedXYTh(0,0,0),150,30);
  BBoxObs2D stObs13(PositionXYThT(565,515,0.5*M_PI,0),SpeedXYTh(0,0,0),200,30);

  this->addStObstacle(FXString("stObs1"),stObs1);
  this->addStObstacle(FXString("stObs2"),stObs2);
  this->addStObstacle(FXString("stObs3"),stObs3);
  this->addStObstacle(FXString("stObs4"),stObs4);
  this->addStObstacle(FXString("stObs5"),stObs5);
  this->addStObstacle(FXString("stObs6"),stObs6);
  this->addStObstacle(FXString("stObs7"),stObs7);
  this->addStObstacle(FXString("stObs8"),stObs8);
  this->addStObstacle(FXString("stObs9"),stObs9);
  this->addStObstacle(FXString("stObs10"),stObs10);
  this->addStObstacle(FXString("stObs11"),stObs11);
  this->addStObstacle(FXString("stObs12"),stObs12);
  this->addStObstacle(FXString("stObs13"),stObs13);
  
  // init the random
  initRandomDeform();

//   // test of copy of the roadmap
//   RoadMap* copyRM = new RoadMap(this->env_);
//   copyRM->construct();
//   cout << "copyRM->getNbNodes() : " << copyRM->getNbNodes() << endl;
//   roadmap_ = new RoadMap(*copyRM);
//   cout << "roadmap_->getNbNodes() : " << roadmap_->getNbNodes() << endl;

  // compute the roadmap on the environment
  roadmap_ = new DRoadMap(this->env_);
  roadmap_->setNbNodesNeeded(700);
  roadmap_->setMaxDistNgh(50.0);
  roadmap_->construct();

  // refresh the viewer
  this->setLookAt(env_,NULL,NULL);
  this->drawScene();

  msgBox->setText("Environment 6 loaded (double integrator system environment) and road map constructed.");
}

void MPC_TeddyIfc::initObsRand(int nb)
{
  // clear the current environment
  if(this->env_ != NULL)
    {
//       this->env_->clearObs();
      this->env_->clearObs2D();
      this->env_->clearFixedObs2D();
    }

  for(int i=0; i<nb; i++)
    {
      double x = getIntRandom(0,dfltXMap);
      double y = getIntRandom(0,dfltYMap);
      double sx = getIntRandom(-15,15);
      double sy = getIntRandom(-15,15);
      double initTime = getIntRandom(0,15);
      double lifeTime = getIntRandom(20,200);
      double sizeObs = getIntRandom(10,30);
      
      // create the obstacles
      //ObsDisk2D* newObs = new ObsDisk2D(PMConfig(x,y), Speed2D(sx,sy), initTime, lifeTime, sizeObs);
      DiskObs2D newObs(PositionXYT(x,y,0),SpeedXY(sx,sy),sizeObs);
      // append it to the list of obstacles
      this->addObstacle(FXString("obs"),newObs,true);
    }

  // refresh the viewer
  this->setLookAt(env_,NULL,NULL);
  this->drawScene();

  // init the random
  initRandomDeform();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Signal functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Widget has been resized
long MPC_TeddyIfc::onConfigure(FXObject*,FXSelector,void*)
{
//   if(glcanvas->makeCurrent())
//     {
//       glViewport(0,0,glcanvas->getWidth(),glcanvas->getHeight());
//       glcanvas->makeNonCurrent();
//     }
  return 1;
}

long MPC_TeddyIfc::onExpose(FXObject*,FXSelector,void*)
{
  drawScene();
  return 1;
}

long MPC_TeddyIfc::loadPath(FXObject*,FXSelector,void*)
{
  // destroy the old path
  if(this->path_ != NULL)
    {
      delete this->path_;
    }

  // reinit the time already past
  alreadyElapsedTime = 0;
  FPS = 0;
  nbTimeout = 0;
  totalNbTimeout = 0;

//   // destroy the old trajectory
//   if(this->traj_ != NULL)
//     {
//       delete this->traj_;
//     }

  // check the type of the path
  if(isExtFilename((char*)(inputPathName.text()), (char*)("pth")))
    {
      // the path is a holonomic path : create a new holonomic path
      this->path_ = new PMPath;
      genTrajType = PM_TRAJ;
      trajType = PM_TRAJ;
    }
  else if(isExtFilename((char*)(inputPathName.text()), (char*)("nhp")))
    {
      // the path is a non holonomic path : create a new non holonomic path
      this->path_ = new CLPath;
      genTrajType = CL_TRAJ;
      trajType = CL_TRAJ;
    }
  else
    {
      // unknown path type
      FXMessageBox::error(this,MBOX_OK,"Error while loading the path","The input path filename extension is not recognized");
      return -1;
    }
  

  //load the path
  bool resLoad = this->path_->load((char*)(inputPathName.text()));

  if(!resLoad)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while loading the path","The input path filename isn't available");
      return -1;
    }

  // init the viewer
  PositionXY center = path_->center();
  double radius = path_->radius();
  double depth = 5*radius;
  double dt = 0;

  this->glcanvas->init(center.x(), center.y() + dt, depth, center.x(), center.y(), 0, 0, 1, 0);

  // draw the scene
  this->drawScene();
  return 1;
}

long MPC_TeddyIfc::loadPrecompPath(FXObject*,FXSelector,void*)
{
//   // compute a straight line path until this position
//   DPMConfig cfg1(50,100,0.5*M_PI);
//   DPMConfig cfg2(50,300,0.5*M_PI);
//   DPMConfig cfg3(50,300,0);
//   DPMConfig cfg4(150,300,0);
//   DPMConfig cfg5(150,300,-0.5*M_PI);
//   DPMConfig cfg6(150,100,-0.5*M_PI);
//   DPMConfig cfg7(150,100,0);
//   DPMConfig cfg8(250,100,0);
//   DPMConfig cfg9(250,100,0.5*M_PI);
//   DPMConfig cfg10(250,300,0.5*M_PI);
//   DPMConfig cfg11(250,300,0);
//   DPMConfig cfg12(350,300,0);
//   DPMConfig cfg13(350,300,-0.5*M_PI);
//   DPMConfig cfg14(350,100,-0.5*M_PI);
  
//   this->path_ = new Path();
  
//   GeomPathPlanner gpp1(NULL, &cfg1, &cfg2);
//   gpp1.initStrLineCptg();
//   int resPP1 = gpp1.computePath();
//   Path* path1 = gpp1.getPath()->getPath2D();
//   path_->concat(*path1);

//   GeomPathPlanner gpp2(NULL, &cfg3, &cfg4);
//   gpp2.initStrLineCptg();
//   int resPP2 = gpp2.computePath();
//   Path* path2 = gpp2.getPath()->getPath2D();
//   path_->concat(*path2);

//   GeomPathPlanner gpp3(NULL, &cfg5, &cfg6);
//   gpp3.initStrLineCptg();
//   int resPP3 = gpp3.computePath();
//   Path* path3 = gpp3.getPath()->getPath2D();
//   path_->concat(*path3);

//   GeomPathPlanner gpp4(NULL, &cfg7, &cfg8);
//   gpp4.initStrLineCptg();
//   int resPP4 = gpp4.computePath();
//   Path* path4 = gpp4.getPath()->getPath2D();
//   path_->concat(*path4);

//   GeomPathPlanner gpp5(NULL, &cfg9, &cfg10);
//   gpp5.initStrLineCptg();
//   int resPP5 = gpp5.computePath();
//   Path* path5 = gpp5.getPath()->getPath2D();
//   path_->concat(*path5);

//   GeomPathPlanner gpp6(NULL, &cfg11, &cfg12);
//   gpp6.initStrLineCptg();
//   int resPP6 = gpp6.computePath();
//   Path* path6 = gpp6.getPath()->getPath2D();
//   path_->concat(*path6);

//   GeomPathPlanner gpp7(NULL, &cfg13, &cfg14);
//   gpp7.initStrLineCptg();
//   int resPP7 = gpp7.computePath();
//   Path* path7 = gpp7.getPath()->getPath2D();
//   path_->concat(*path7);

  if(roadmap_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Path","No roadmap has been defined");
      return -1;
    }

  // compute a path on the road map
  this->path_ = new PMPath();
  PMConfig cfg1(50,100);
  PMConfig cfg2(350,100);

//   //////////////////////////////////////////////////////////////////////////////////
//   cout << "MPC_TeddyIfc::loadPrecompPath : Not finished yet!" << endl;
//   FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::loadPrecompPath","Function not available... :-(");
  //////////////////////////////////////////////////////////////////////////////////
  RoadMapPP rmpp(sys_,env_,roadmap_);
//   cout << "debut roadmap path planning" << endl;
  rmpp.computePath(cfg1,cfg2);
//   cout << "fin roadmap path planning" << endl;
//   Path* path = rmpp.getPath();
//   path_->concat(*path);
  this->path_ = new PMPath(*(rmpp.getPath()));

  // compute the trajectory until this position
  int res = this->computeTrajectory(NULL, 0, NULL);
  //////////////////////////////////////////////////////////////////////////////////

  if(res == TRAJ_CPTD) return 1;
  else return -1;
}

long MPC_TeddyIfc::loadPrecompTraj(FXObject*,FXSelector,void*)
{
  // delete the old trajectory if needed
  if(traj_ != NULL) delete traj_;

  // init the car-like system constraints
  this->initCarLikeSys();

  // initialize the trajectory planner
  CCCTrajPlanner ctp(sys_,NULL);

  // initialize the initial state
  CLState initS(initXTraj,initYTraj,initThTraj,0.0,0.0);

  // compute the trajectory
  ctp.computeDefinedTraj3(initS);

  this->traj_ = new CLSTraj(*(ctp.getTraj()));

  trajType = CL_TRAJ;
  processType_ = TRAJ_PLNG_PROC;
}

long MPC_TeddyIfc::loadMap(FXObject*,FXSelector,void*)
{
  // Load the map
  bool resLoad = env_->loadMap((char*)(mapName.text()));

  if(resLoad != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Loading the Map","Map's filename doesn't exist");
      return -1;
    }

  this->loadMapInViewer();
  return 1;
}

long MPC_TeddyIfc::computeTrajectory(FXObject*,FXSelector,void*)
{
  processType_ = TRAJ_PLNG_PROC;

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  if(this->path_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","No init path has been defined");
      return -1;
    }

  // init the general trajectory type
  if(typeMotion == 1)
    {
      if(trajType == PM_TRAJ) genTrajType = PM_TRAJ;
      else genTrajType = CL_TRAJ;
    }

  PVDTrajPlanner* tp;
  int resTP;

  if(genTrajType == PM_TRAJ)
    {
//       tp = PVDTrajPlanner(sys_,env_,StateR5D(((*(path_))[0])->x(),((*(path_))[0])->y(),0,0,0),StateR5D(((*(path_))[path_->getSize()-1])->x(),((*(path_))[path_->getSize()-1])->y(),0,0,0),path_);
      tp = new PM_PVDTrajPlanner((PointMassSys*)sys_);
//       tp = new PM_PVDTrajPlanner((PointMassSys*)sys_,env_);
      tp->setTimeStep(trajTimeStep);

      PMConfig initCfg = PMConfig(((PMPath*)path_)->firstCfg());
      PMConfig goalCfg = PMConfig(((PMPath*)path_)->lastCfg());
      PMState initS(initCfg.x(),initCfg.y(),0,0);
      PMState goalS(goalCfg.x(),goalCfg.y(),0,0);

      resTP = ((PM_PVDTrajPlanner*)tp)->computeTraj(initS,goalS,(PMPath*)path_);
    }
  else if(genTrajType == CL_TRAJ)
    {
      ////////////////////////////////////////////////////////////////////////////////////////////////////////
      cout << "MPC_TeddyIfc::computeTrajectory : Not finished yet!" << endl; 
      FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::computeTrajectory","Not finished yet!!!!!!!!!!!!!!!!!");
      ////////////////////////////////////////////////////////////////////////////////////////////////////////
//       tp = PVDTrajPlanner(sys_,env_,StateS6D(((*(path_))[0])->x(),((*(path_))[0])->y(),((*(path_))[0])->th(),0,0,0),StateS6D(((*(path_))[path_->getSize()-1])->x(),((*(path_))[path_->getSize()-1])->y(),((*(path_))[path_->getSize()-1])->th(),0,0,0),path_);
    }

  if(resTP != TRAJ_CPTD)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
      return -2;
    }
  else
    {
      if(this->traj_ != NULL) delete traj_;

//       cout << "traj cptd size : " << (((PM_PVDTrajPlanner*)tp)->getTraj())->size() << endl;
//       cout << "Trajectory computed (in PVD) : " << endl;
//       (((PM_PVDTrajPlanner*)tp)->getTraj())->display();

      if(genTrajType == PM_TRAJ) this->traj_ = new PMSTraj(*(((PM_PVDTrajPlanner*)tp)->getTraj()));
//       else if(genTrajType == CL_TRAJ) this->traj_ = new CLSTraj(*(((PM_PVDTrajPlanner*)tp)->getTraj()));

      this->traj_->shift(time);

      if(this->traj_->getSize() != 0) slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1].t());
      curStep = 0;
//       this->traj_->display();

//       // init the table of state types
//       typeState = new int*;
//       (*typeState) = new int[this->traj_->getSize()];
//       for(int i=0; i < this->traj_->getSize(); i++)
// 	{
// 	  (*typeState)[i] = 0;
// 	}

//       //init the table of nearest states
//       tableNearState = new PositionXYT**;
//       (*tableNearState) = new PositionXYT* [this->traj_->getSize()];
//       for(int i=0; i < this->traj_->getSize(); i++)
// 	{
// 	  (*tableNearState)[i] = NULL;
// 	}

//       // init the tables of internal and external forces
//       tableFInt = new StateTime*;
//       *tableFInt = NULL;
//       tableFExt = new StatTime*;
//       *tableFExt = NULL;

//       // init the trajectory deformer
//       if(deformer_ != NULL) delete deformer_;


//       if(genTrajType == PM_TRAJ)
// 	{
// 	  deformer_ = new STDeformation(this->traj_, this->env_, this->sys_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
// 	}
//       else
// 	{
// 	  cout << "input trajectory 111 : " << endl;
// 	  this->traj_->display();

// 	  deformer_ = new STNHDeformation(this->traj_, this->env_, this->sys_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
// 	  ((STNHDeformation*)deformer_)->setParallelExec(parallelize);
// 	  ((STNHDeformation*)deformer_)->setNbThread(nbThread);
// 	}

//       if(optModeDeform == 0) deformer_->setModeDeform(SPATIOTEMP_DEFORM);
//       else if(optModeDeform == 1) deformer_->setModeDeform(SPATONLY_DEFORM);
//       else if(optModeDeform == 2) deformer_->setModeDeform(TEMPONLY_DEFORM);
//       deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &K_SPT_TMP);
//       deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &alphaMin, &alphaMax, &betaMin, &betaMax);
//       deformer_->setMaxEffectDist(&maxEffectDist);
//       deformer_->setMaxStEffectDist(&maxStEffectDist);
//       deformer_->setMaxEffectDist(&maxSpDist,&maxTmpDist,&maxStEffectDist);
//       deformer_->setReliability(&predHorizon, &reliabilityTime, &epsilonRT);
//       deformer_->setTemporalConst(&useTemporalConsistency);
//       deformer_->setTimeStep(trajTimeStep);
//       //deformer_->setPosStep(trajTimeStep * maxSpd);
//       deformer_->setPosStep(trajPosStep);

//       // init the trajectory generation parameters
//       DDTrajGen::setMaxNbStep(maxNbStep);                // set the maximal number of steps
//       DDTrajGen::setMaxResidual(maxResidual);            // set the maximal residual allowed
//       DDTrajGen::setDBG(debugLevel);                     // set the debugging level
//       CLTrajGen::setMaxNbStep(maxNbStep);                // set the maximal number of steps
//       CLTrajGen::setMaxResidual(maxResidual);            // set the maximal residual allowed
//       CLTrajGen::setDBG(debugLevel);                     // set the debugging level
//       CLTrajGen::setUseVelCst(useVelCst);                // set true if linear velocity constraints are used
//       CLTrajGen::setUseAccCst(useAccCst);                // set true if linear acceleration constraints are used
//       CLTrajGen::setUseAVelCst(useAVelCst);              // set true if angular velocity constraints are used
//       CLTrajGen::setUseAAccCst(useAAccCst);              // set true if angular acceleration constraints are used

      if(typeMotion == 0) // trajectory tracking
	{
	  // init the viewer
	  PositionXY center;
	  double radius; 
	  if(env_ != NULL)
	    {
	      center = env_->getCenter();
	      radius = env_->getRadius();
	    }
	  else if(path_ != NULL)
	    {
	      center = path_->center();
	      radius = path_->radius();
	    }
	  else
	    {
	      cerr << "MPC_TeddyIfc::computeTrajectory : path and environment are not defined..." << endl;
	      exit(1);
	    }
	  double depth = 5*radius;
	  double dt = 3*(*traj_)[traj_->getSize()-1].t();
	  double upx,upy,upz;
	  
	  Vec v1(3);
	  v1.setVal(0,0);
	  v1.setVal(1,-dt);
	  v1.setVal(2,-depth);
	  double v1Norm = v1.norm();
	  Vec v2(3);
	  v2.setVal(0,0);
	  v2.setVal(1,-dt/v1Norm);
	  v2.setVal(2,-depth/v1Norm);
	  upx = 0;
	  upy = -v2[2];
	  upz = v2[1];	  

	  this->glcanvas->init(center.x(), center.y() + dt, depth, center.x(), center.y(), 0, upx, upy, upz);

	  // draw the scene
	  this->drawScene();
	}
//       cout << "deform:ctraj41" << endl;
//       cout << "traj[0] : "; (*(this->traj_))[0]->display(); cout << endl << endl; 
//       cout << "deform:ctraj42" << endl;

      return 1;
    }
}

long MPC_TeddyIfc::computeRandRKTraj(FXObject*,FXSelector,void*)
{
  CLStateT initCfg((double)initXRb,(double)initYRb,0,10.0,0,0);
  return this->computeNewCarLikeTraj(initCfg);
}

long MPC_TeddyIfc::addObstacle(FXObject*,FXSelector,void*)
{
  cerr << "MPC_TeddyIfc::addObstacle : Not implemented anymore..." << endl;
  exit(1);

//   int id;
  
//   if(lifeT < 0)
//     {
//       FXMessageBox::error(this,MBOX_OK,"Error while adding obstacle","The life of time must be positive");
//       return -2;
//     }

//   if(randObs != FXString("yes") && randObs != FXString("no"))
//     {
//       FXMessageBox::error(this,MBOX_OK,"Error while adding obstacle","The random mode must be 'yes' or 'no'");
//       return -3;
//     }

//   if((id = this->elemListExists(idObs)) != -1)
//     {
//       // change the display list parameters
//       list[3]->setItem(id,FXStringVal(sxObs,2,FALSE));
//       list[4]->setItem(id,FXStringVal(syObs,2,FALSE));
//       list[8]->setItem(id,randObs);

//       // change the obstacle parameters
//       time = time + 0.0001;
//       this->env_->getObs(id)->setLastSpd(new Speed2D(sxObs,syObs));
      
//       if(randObs == FXString("yes"))
// 	{
// 	  this->env_->getObs(id)->setRdmMvt(true);
// 	}
//       else
// 	{
// 	  this->env_->getObs(id)->setRdmMvt(false);
// 	}

//       return 1;
//     }

//   // create the obstacle list element
//   list[0]->appendItem(idObs);
//   list[1]->appendItem(FXStringVal(xObs,2,FALSE));
//   list[2]->appendItem(FXStringVal(yObs,2,FALSE));
//   list[3]->appendItem(FXStringVal(sxObs,2,FALSE));
//   list[4]->appendItem(FXStringVal(syObs,2,FALSE));
//   list[5]->appendItem(FXStringVal(rObs,2,FALSE));
//   list[6]->appendItem(FXStringVal(initT,2,FALSE));
//   list[7]->appendItem(FXStringVal(lifeT,2,FALSE));
//   list[8]->appendItem(randObs);

//   // create the obstacle
//   PMConfig cfgO(xObs,yObs);
//   Speed2D spdO(sxObs,syObs);
//   ObsDisk2D* obs = new ObsDisk2D(PMConfig(xObs,yObs), Speed2D(sxObs,syObs), initT, lifeT, rObs);

//   if(randObs == FXString("yes"))
// 	{
// 	  obs->setRdmMvt(true);
// 	}
//       else
// 	{
// 	  obs->setRdmMvt(false);
// 	}

//   // append it to the list of obstacles
//   this->env_->appendObs(obs);
  
//   // repaint the viewer
//   this->drawScene();

//   return 1;
}

// long MPC_TeddyIfc::addObstacle(FXString id, ObsDisk2D* obs, bool rand)
// {
//   // create the obstacle list element
//   list[0]->appendItem(id);
//   list[1]->appendItem(FXStringVal(obs->initCfg()->x(),2,FALSE));
//   list[2]->appendItem(FXStringVal(obs->initCfg()->y(),2,FALSE));
//   list[3]->appendItem(FXStringVal(obs->initSpd()->x(),2,FALSE));
//   list[4]->appendItem(FXStringVal(obs->initSpd()->y(),2,FALSE));
//   list[5]->appendItem(FXStringVal(((ObsDisk2D*)obs)->getRadius(),2,FALSE));
//   list[6]->appendItem(FXStringVal(obs->initTime(),2,FALSE));
//   list[7]->appendItem(FXStringVal(obs->lifeTime(),2,FALSE));
//   if(rand) list[8]->appendItem(FXString("yes"));
//   else list[8]->appendItem(FXString("no"));

//   if(rand) obs->setRdmMvt(true);
//   else obs->setRdmMvt(false);

//   // append it to the list of obstacles
//   this->env_->appendObs(obs);
  
//   // repaint the viewer
//   this->drawScene();

//   return 1;
// }

long MPC_TeddyIfc::addObstacle(FXString id, const DiskObs2D& obs, bool rand)
{
//   if(rand) obs->setRdmMvt(true);
//   else obs->setRdmMvt(false);

  // append it to the list of obstacles
  this->env_->appendObs2D(obs);
  
  // repaint the viewer
  this->drawScene();

  return 1;
}

// long MPC_TeddyIfc::addStObstacle(FXString id, Obstacle2D* obs)
// {
//   // append it to the list of obstacles
//   this->env_->appendStObs(obs);
  
//   // repaint the viewer
//   this->drawScene();
  
//   return 1;
// }

long MPC_TeddyIfc::addStObstacle(FXString id, const BBoxObs2D& obs)
{
  // append it to the list of obstacles
  this->env_->appendFixedObs2D(obs);
  
  // repaint the viewer
  this->drawScene();
  
  return 1;
}

long MPC_TeddyIfc::removeObstacle(FXObject*,FXSelector,void*)
{
  cerr << "MPC_TeddyIfc::removeObstacle : Not implemented anymore..." << endl;

//   int id;
//   if((id = this->elemListExists(idObs)) != -1)
//     {
//       for(int i=0; i<8; i++)
// 	{
// 	  list[i]->removeItem(id);
// 	}
//       this->env_->removeObs(id);

//       // repaint the viewer
//       this->drawScene();
    
//       return 1;
//     }
//   else
//     {
//       FXMessageBox::error(this,MBOX_OK,"Error while removing obstacle","The obstacle ID doesn't exists");
//       return -1;
//     }
}

long MPC_TeddyIfc::changeInputPathName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Input path");
  open.setFilename(inputPathName);
  if(open.execute()){
    inputPathName = open.getFilename();
  }
  return 1;
}

long MPC_TeddyIfc::changeMapName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Environment map");
  open.setFilename(mapName);
  if(open.execute()){
    mapName = open.getFilename();
  }
  return 1;
}

// Clicked a header button:- we highlight all in the list
long MPC_TeddyIfc::onCmdHeaderButton(FXObject*,FXSelector,void* /* ptr */){
//   FXint which=(FXint)(long)ptr;
//   FXint i;
//   FXASSERT(0<=which && which<5);
//   for(i=0; i<list[which]->getNumItems(); i++){
//     list[which]->selectItem(i);
//   }
  return 1;
}

long MPC_TeddyIfc::updateScene(FXObject*,FXSelector,void*)
{
  // update obstacles random motion parameters
//   DynamicObs::setSeuilChgt(seuilChgt);
//   DynamicObs::setMaxAngleChgt(maxAngleChgt);
//   DynamicObs::setMaxNormChgt(maxNormChgt);

  // compute the time before the deformation
  struct timeval t3;
  gettimeofday(&t3,NULL);

  // deform the trajectory
  int status = this->deformTraj();

//   if(this->traj_ != NULL)
//     {
//       cout << "traj_->getSize() : " << traj_->getSize() << endl;
//       this->traj_->display();
// //       cout << "last state-time : ";
// //       (*this->traj_)[this->traj_->getSize()-1].display();
//       cout << endl;
//     }

  // compute the time past
  gettimeofday(&t2,NULL);
  time_t elapsedTs = t2.tv_sec - t1.tv_sec;
  suseconds_t elapsedTus = t2.tv_usec - t1.tv_usec;

  double elapsedT = alreadyElapsedTime + ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);

  // cout << "timeout : \t Ts : "<< elapsedTs << "\t Tus : " << elapsedTus << " \t time : " << elapsedT << endl;

  // set the current time
  if(realTimeAlgo) time = elapsedT;
  else{
    if(elapsedT - time < 0.5)
      {
	cout << "inf 0.5sec" << endl;
	sleep((int)((0.5 - elapsedT + time)*1000.0));
      }
    time = time + 0.5;
  }
  // compute the deformation duration
  time_t deformTs = t2.tv_sec - t3.tv_sec;
  suseconds_t deformTus = t2.tv_usec - t3.tv_usec;
  deformDur = ((double)deformTs) + 0.000001 * ((double)deformTus);
  if(deformDur != 0) lastFPS = 1.0/deformDur;
  else lastFPS = 0.0;
  if(this->traj_ != NULL) nbST = this->traj_->getSize() - curStep - 1;
  else nbST = 0;
  if(this->traj_ != NULL) nbSTtot = this->traj_->getSize();
  else nbSTtot = 0;

  bool stop = false;

//   cout << "curStep : " << curStep << "  this->traj_->getSize() : " << this->traj_->getSize() << "  status : " << status << endl;

  switch(status)
    {
    case 0 : 
      // search the step of the trajectory reached at this time
      if(this->traj_ != NULL)
	{
	  if(curStep >= this->traj_->getSize()-1)
	    {
// 	      cout << "fin traj" << endl;
// 	      cout << "typeMotion : " << typeMotion << endl;
	      
	      if(typeMotion == 0) // trajectory tracking
		{
		  msgBox->setText("The trajectory deformation process achieved sucessfully.");
		  this->stopTraj(NULL,0,NULL);
		  stop = true;
		}
	      else // positions tracking
		{
		  if(trajType == PM_TRAJ)
		    {
// 		      cout << "new holonomic trajectory" << endl;
		      PMStateT newInitCfg(*((PMStateT*)(&(this->traj_->at(this->traj_->getSize()-1)))));
		      this->computeNewRandTrajFrom(newInitCfg);
		    }
		  else if(trajType == CL_TRAJ)
		    {
// 		      FXMessageBox::error(this,MBOX_OK,"Car-like position tracking error","This function has not been implemented yet :-(");
// 		      msgBox->setText("This function has not been implemented yet :-(");
// 		      return -1;

		      CLStateT newInitST = (*((CLStateT*)(&(this->traj_->at(this->traj_->getSize()-1)))));
		      this->computeNewRandTrajFrom(newInitST);
		    }
		  else if(trajType == WC_TRAJ)
		    {
		      WCStateT newInitST = (*((WCStateT*)(&(this->traj_->at(this->traj_->getSize()-1)))));
		      this->computeNewRandTrajFrom(newInitST);
		    }
		}
	    }
	  else
	    {
	      bool trouve = false;
	      double oldCurStep = curStep;
	      curStep = 0;
	      while(!trouve)
		{
		  if(curStep >= 0 && curStep < this->traj_->getSize()-1 && time >= (*(this->traj_))[(int)(curStep+1)].t())
		    {
		      curStep = curStep+1;
		      if((curStep > oldCurStep && pauseColLoc) && ((*typeState)[(int)curStep] == -1 || ((*typeState)[(int)curStep] == -2 && !useTemporalConsistency)))
			{
			  this->stopTraj(NULL,0,NULL);
			}
		    }
		  else
		    trouve = true;
		}
	    }
	}
      
      break;
    case -1 : 
      msgBox->setText("Warning. The input trajectory is null.");
      cout << "MPC_TeddyIfc::deformTraj() : Warning. The input trajectory is null." << endl;
      this->stopTraj(NULL, 0, NULL);
      break;
    case -2 : 
      msgBox->setText("Warning. A collision will occur during the next prediction horizon.");
      cout << "MPC_TeddyIfc::deformTraj() : Warning. A collision will occur during the next prediction horizon." << endl;
      this->stopTraj(NULL, 0, NULL);
      break;
    case -3 : 
      msgBox->setText("Warning. The connectivity has been lost during the next prediction horizon.");
      cout << "MPC_TeddyIfc::deformTraj() : Warning. The connectivity has been lost during the next prediction horizon." << endl;
      this->stopTraj(NULL, 0, NULL);
      break;    
    }
  
  // repaint the viewer
  this->drawScene();
  
  // rerun the timeout
  if(oneStep)
    {
      alreadyElapsedTime = time;
      oneStep = false;
    }
  else if(running)
    {
      getApp()->addTimeout(this,ID_TIMEOUT,TIMER_INTERVAL);
    }

  // update the FPS number
  nbTimeout++;
  FPS = ((double)(totalNbTimeout + nbTimeout))/time;

  return 1;
}

long MPC_TeddyIfc::startTraj(FXObject*,FXSelector,void*)
{
  // verifying that the trajectory is not NULL
  if(this->traj_ == NULL && typeMotion == 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Deformation","The initial trajectory hasn't been computed");
      return -1;
    }

  // saving the initial trajectory
  if(time == 0)
    {
      if(this->initTraj_ != NULL) delete this->initTraj_;
      if(this->path_->type() == PM_PATH)
	this->initTraj_ = new PMSTraj(*((PMSTraj*)(this->traj_)));
      else if(this->path_->type() == CL_PATH)
	this->initTraj_ = new CLSTraj(*((CLSTraj*)(this->traj_)));
    }

  cout << "starting timeout" << endl;
  // init the timer
  gettimeofday(&t1,NULL);
  running = true;
  getApp()->addTimeout(this,ID_TIMEOUT,TIMER_INTERVAL);
  return 1;
}

long MPC_TeddyIfc::oneStepTraj(FXObject* ,FXSelector ,void*)
{
  // verifying that the trajectory is not NULL
  if(this->traj_ == NULL && typeMotion == 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Deformation","The initial trajectory hasn't been computed");
      return -1;
    }
  
  // init the timer
  gettimeofday(&t1,NULL);
  oneStep = true;
  getApp()->addTimeout(this,ID_TIMEOUT,TIMER_INTERVAL);
  return 1;
}

long MPC_TeddyIfc::stopTraj(FXObject*,FXSelector,void*)
{
  cout << "stoping timeout" << endl;
  running = false;
  getApp()->removeTimeout(this,ID_TIMEOUT);
  alreadyElapsedTime = time;
  return 1;
}

long MPC_TeddyIfc::restartTraj(FXObject*,FXSelector,void*)
{
  cerr << "MPC_TeddyIfc::restartTraj : Not implemented anymore..." << endl;
  exit(1);

//   // reinit the trajectory
//   if(this->initTraj_ != NULL)
//     {
//       if(this->traj_ != NULL) delete this->traj_;
//       if(this->path_->type() == PM_PATH)
// 	this->traj_ = new PMSTraj(*((PMSTraj*)(this->initTraj_)));
//       else if(this->path_->type() == CL_PATH)
// 	this->traj_ = new CLSTraj(*((CLSTraj*)(this->initTraj_)));
//     }

//   // reinit the environment
//   if(this->env_ != NULL && this->env_->getNbObs() != -1)
//     {
//       for(int i=0; i<this->env_->getNbObs(); i++)
// 	{
// 	  this->env_->getObs(i)->reinit();
// 	}
//     }

//   // reinit the robot
  

//   // reinit the time already past
//   time = 0;
//   alreadyElapsedTime = 0;
//   FPS = 0;
//   totalNbTimeout = 0;
//   nbTimeout = 0;
  
//   // repaint the viewer
//   this->drawScene();

//   cout << "nbObs : " << this->env_->getNbObs() << endl;
//   cout << "stackSize : " << this->env_->getObs(0)->stackSize() << endl;
//   cout << "initTime : " << this->env_->getObs(0)->initTime() << endl;
//   cout << "stateStack : "; this->env_->getObs(0)->getStateStack(0)->display(); cout << endl;
//   cout << "stateStack : "; this->env_->getObs(0)->getStateStack(1)->display(); cout << endl;

//   return 1;
}

long MPC_TeddyIfc::repaint(FXObject*,FXSelector,void*)
{
  this->drawScene();
  return 1;
}

long MPC_TeddyIfc::timeView(FXObject*,FXSelector,void*)
{
  if(this->traj_ != NULL)
    {
      // init the viewer
      PositionXY center;
      double radius; 
      if(env_ != NULL)
	{
	  center = env_->getCenter();
	  radius = env_->getRadius();
	}
      else if(path_ != NULL)
	{
	  center = path_->center();
	  radius = path_->radius();
	}
      else
	{
	  cerr << "MPC_TeddyIfc::timeView : path and environment are not defined..." << endl;
	  exit(1);
	}
      double depth = 5*radius;
      double dt = traj_->at(traj_->getSize()-1).t();
      
      // change the point of view
      this->glcanvas->init(center.x(), -depth, dt/2, center.x(), center.y(), dt/2, 0, 0, 1);
      
      // repaint the viewer
      this->drawScene();

      return 1;
    }
  else return 0;
}

long MPC_TeddyIfc::spaceView(FXObject*,FXSelector,void*)
{
  if(this->traj_ != NULL)
    {
      // init the viewer
      PositionXY center;
      double radius; 
      if(env_ != NULL)
	{
	  center = env_->getCenter();
	  radius = env_->getRadius();
	}
      else if(path_ != NULL)
	{
	  center = path_->center();
	  radius = path_->radius();
	}
      else
	{
	  cerr << "MPC_TeddyIfc::spaceView : path and environment are not defined..." << endl;
	  exit(1);
	}
      double depth = 5*radius;
      
      // change the point of view
      this->glcanvas->init(center.x(), center.y(), depth, center.x(), center.y(), 0, 0, 1, 0);
      
      // repaint the viewer
      this->drawScene();
      
      return 1;
    }
  else return 0;
}

long MPC_TeddyIfc::perspView(FXObject*,FXSelector,void*)
{
  if(this->traj_ != NULL)
    {
      // init the viewer
      PositionXY center;
      double radius; 
      if(env_ != NULL)
	{
	  center = env_->getCenter();
	  radius = env_->getRadius();
	}
      else if(path_ != NULL)
	{
	  center = path_->center();
	  radius = path_->radius();
	}
      else
	{
	  cerr << "MPC_TeddyIfc::perspView : path and environment are not defined..." << endl;
	  exit(1);
	}
      double depth = 5*radius;
      double dt = 3*(*traj_)[traj_->getSize()-1].t();
      double upx,upy,upz;
      
      Vec v1(3);
      v1.setVal(0,0);
      v1.setVal(1,-dt);
      v1.setVal(2,-depth);
      double v1Norm = v1.norm();
      Vec v2(3);
      v2.setVal(0,0);
      v2.setVal(1,-dt/v1Norm);
      v2.setVal(2,-depth/v1Norm);
      upx = 0;
      upy = -v2[2];
      upz = v2[1];
      
      // change the point of view
      this->glcanvas->init(center.x(), center.y() + dt, depth, center.x(), center.y(), 0, upx, upy, upz);

      // repaint the viewer
      this->drawScene();
      
      return 1;
    }
  else return 0;
}

long MPC_TeddyIfc::updateDispFuture(FXObject*,FXSelector,void*)
{
  // update the disp future mode
  if(dispFuture == false)
    dispFuture = true;
  else
    dispFuture = false;

  // repaint the viewer
      this->drawScene();

  return 1;
}

long MPC_TeddyIfc::updateDispForces(FXObject*,FXSelector,void*)
{
  // update the disp forces mode
  if(dispForces == false)
    dispForces = true;
  else
    dispForces = false;

  // repaint the viewer
  //this->drawScene();

  return 1;
}

long MPC_TeddyIfc::updateDispAllForces(FXObject*,FXSelector,void*)
{
  // update the disp forces mode
  if(dispAllForces == false)
    dispAllForces = true;
  else
    dispAllForces = false;

  // repaint the viewer
  //this->drawScene();

  return 1;
}

long MPC_TeddyIfc::updateDispNearSt(FXObject*,FXSelector,void*)
{
  // update the disp nearest states mode
  if(dispNearStates == false)
    dispNearStates = true;
  else
    dispNearStates = false;
  return 1;
}

long MPC_TeddyIfc::updateDispNearStSt(FXObject*,FXSelector,void*)
{
  // update the disp nearest states mode
  if(dispNearStStates == false)
    dispNearStStates = true;
  else
    dispNearStStates = false;
  return 1;
}

long MPC_TeddyIfc::init(FXObject*,FXSelector,void*)
{
  initObs();
  return 1;
}

long MPC_TeddyIfc::init2(FXObject*,FXSelector,void*)
{
  initObs2();
  return 1;
}

long MPC_TeddyIfc::init3(FXObject*,FXSelector,void*)
{
  initObs3();
  return 1;
}

long MPC_TeddyIfc::init4(FXObject*,FXSelector,void*)
{
  initObs4();
  return 1;
}

long MPC_TeddyIfc::init5(FXObject*,FXSelector,void*)
{
  initObs5();
  return 1;
}

long MPC_TeddyIfc::init6(FXObject*,FXSelector,void*)
{
  initObs6();
  return 1;
}

long MPC_TeddyIfc::initRandObs(FXObject*,FXSelector,void*)
{
  initObsRand(nbRandObs);
  return 1;
}

long MPC_TeddyIfc::clearEnv(FXObject*,FXSelector,void*)
{
  // clear the current environment
  if(this->env_ != NULL)
    {
//       this->env_->clearObs();
//       this->env_->clearStObs();
      this->env_->clearObs2D();
      this->env_->clearFixedObs2D();
    }
}

long MPC_TeddyIfc::useConnecChk(FXObject*,FXSelector,void*)
{
  if(useConnectivityChecker == true) useConnectivityChecker = false;
  else useConnectivityChecker = true;
  return 1;
}

long MPC_TeddyIfc::useTempCstcy(FXObject*,FXSelector,void*)
{
  if(useTemporalConsistency == true) useTemporalConsistency = false;
  else useTemporalConsistency = true;
  return 1;
}

long MPC_TeddyIfc::usePauseColLoc(FXObject*,FXSelector,void*)
{
  if(pauseColLoc == true) pauseColLoc = false;
  else pauseColLoc = true;
  return 1;
}

long MPC_TeddyIfc::initPositionTracking(FXObject*,FXSelector,void*)
{
  // set the position tracking mode on
  typeMotion = 1;
  // set up the dynamic orthogonal view
  orthoView = true;
  dynamicView = true;

  // init the random
  initRandomDeform();

  // compute the first trajectory to follow
  if(trajType == PM_TRAJ)
    {
      PMStateT initSt(initXRb,initYRb,0,0,0);
      this->computeNewRandTrajFrom(initSt);
    }
  else if(trajType == CL_TRAJ)
    {
      CLStateT initSt(initXRb,initYRb,initThRb,0,0,0);
      this->computeNewRandTrajFrom(initSt);
    }
  else if(trajType == WC_TRAJ)
    {
      WCStateT initSt(initXRb,initYRb,initThRb,0,0,0);
      this->computeNewRandTrajFrom(initSt);
    }

//   this->traj_->display();

  // get back the dimensions of the map
  double dimMapX;
  double dimMapY;

//   if(this->env_ != NULL && this->env_->getMap() != NULL)
//     {
//       dimMapX = this->env_->getMap()->getWidth();
//       dimMapY = this->env_->getMap()->getHeight();
//     }
//   else 
  if(this->env_ != NULL)
    {
      dimMapX = this->env_->getDfltX();
      dimMapY = this->env_->getDfltY();
    }
  else
    {
      dimMapX = dfltXMap;
      dimMapY = dfltYMap;
    }

//   // init the viewer
//   PositionXY center(0.5*dimMapX, 0.5*dimMapY);
//   //double radius = path_->radius();
//   double radius = sqrt(dimMapX*dimMapX/4 * dimMapY*dimMapY/4);
//   double depth = 0.022 * radius;
//   double dt = 3*(*traj_)[traj_->getSize()-1].t();
//   double upx,upy,upz;

//   Vec v1(3);
//   v1.setVal(0,0);
//   v1.setVal(1,-dt);
//   v1.setVal(2,-depth);
//   double v1Norm = v1.norm();
//   Vec v2(3);
//   v2.setVal(0,0);
//   v2.setVal(1,-dt/v1Norm);
//   v2.setVal(2,-depth/v1Norm);
//   upx = 0;
//   upy = -v2[2];
//   upz = v2[1];

//   this->glcanvas->init(center.x(), center.y() + dt, depth, center.x(), center.y(), 0, upx, upy, upz);

  // draw the scene
  this->drawScene();

  return 1;
}

long MPC_TeddyIfc::updateDrawMapBoundaries(FXObject*,FXSelector,void*)
{
  // update the drawing of the boundaries of the map
  if(drawMapBoundaries == false)
    drawMapBoundaries = true;
  else
    drawMapBoundaries = false;
  return 1;
}

long MPC_TeddyIfc::updateUseRTAlgo(FXObject*,FXSelector,void*)
{
  // update the drawing of the boundaries of the map
  if(realTimeAlgo == false)
    realTimeAlgo = true;
  else
    realTimeAlgo = false;
  return 1;
}

long MPC_TeddyIfc::updateDynamicView(FXObject*,FXSelector,void*)
{
  // update the dynamic view option
  if(dynamicView == false)
    dynamicView = true;
  else
    dynamicView = false;

  this->drawScene();
  return 1;
}

long MPC_TeddyIfc::updateRoadMapView(FXObject*,FXSelector,void*)
{
  // update the roadmap view option
  if(roadmapView == false)
    roadmapView = true;
  else
    roadmapView = false;

  this->drawScene();
  return 1;
}

long MPC_TeddyIfc::adjustParameters(FXObject*,FXSelector,void*)
{
  // adjust the parameters of the algorithm in function of the mode used
  if(optModeDeform == 0) // spatio-temporal deformation
    {
      K_INT         = 1.0;
      K_EXT         = 2.0;
      K_STEXT       = 2.0;
      K_SPT_TMP     = 3.0;

      alphaMin      = 0.5;
      alphaMax      = 2.0;
      betaMin       = 1.0;
      betaMax       = 10.0;

      maxSpDist     = 80.0;
      maxTmpDist    = 5.0;
    }
  else if(optModeDeform == 1) // spatial deformation
    {
      K_INT         = 1.0;
      K_EXT         = 2.0;
      K_SPT_TMP     = 3.0;

      alphaMin      = 0.1;
      alphaMax      = 1.0;
      betaMin       = 0.0;
      betaMax       = 20.0;

      maxSpDist     = 90.0;
      maxTmpDist    = 5.0;
    }
  else // optModeDeform == 2 // temporal deformation
    {
      K_INT         = 0.1;
      K_EXT         = 8.0;
      K_SPT_TMP     = 3.0;

      alphaMin      = 0.1;
      alphaMax      = 1.0;
      betaMin       = 0.0;
      betaMax       = 20.0;

      maxSpDist     = 100.0;
      maxTmpDist    = 8.0;
    }

  return 1;
}

long MPC_TeddyIfc::perspectiveMode(FXObject*,FXSelector,void*)
{
  FXString msg("Perspective View");
  viewModeBox->setText(msg);

  orthoView = false;  

  return 1;
}

long MPC_TeddyIfc::orthogonalMode(FXObject*,FXSelector,void*)
{
  FXString msg("Orthogonal View");
  viewModeBox->setText(msg);

  orthoView = true;

  return 1;
}

long MPC_TeddyIfc::displayRTGraph(FXObject*,FXSelector,void*)
{
  if(this->deformer_ != NULL)
    this->deformer_->displayPredHorizonProfile(0.01);
  
  return 1;
}

long MPC_TeddyIfc::updateUseVelCst(FXObject*,FXSelector,void*)
{
  // update the use of linear velocity constraints
  if(useVelCst == false)
    useVelCst = true;
  else
    useVelCst = false;
  return 1;
}

long MPC_TeddyIfc::updateUseAccCst(FXObject*,FXSelector,void*)
{
  // update the use of linear acceleration constraints
  if(useAccCst == false)
    useAccCst = true;
  else
    useAccCst = false;
  return 1;
}

long MPC_TeddyIfc::updateUseAVelCst(FXObject*,FXSelector,void*)
{
  // update the use of angular velocity constraints
  if(useAVelCst == false)
    useAVelCst = true;
  else
    useAVelCst = false;
  return 1;
}

long MPC_TeddyIfc::updateUseAAccCst(FXObject*,FXSelector,void*)
{
  // update the use of angular acceleration constraints
  if(useAAccCst == false)
    useAAccCst = true;
  else
    useAAccCst = false;
  return 1;
}

long MPC_TeddyIfc::updateDrawTrajGen(FXObject*,FXSelector,void*)
{
  // update the drawing of generated trajectories
  if(drawTrajGen == false)
    drawTrajGen = true;
  else
    drawTrajGen = false;
  return 1;
}

// convert the current environment into an occupancy grid
long MPC_TeddyIfc::convertEnvOccGrid(FXObject*,FXSelector,void*)
{
  // check if the current environment is not null
  if(this->env_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Converting the Environment","No environment has been defined yet.");      
      return -1;
    }
  
  // delete the current occupancy grid if it exists
  if(this->map_ != NULL)
    delete this->map_;
  
  struct timeval t1;
  struct timeval t2;
  
  // init the timer
  gettimeofday(&t1,NULL);
  
  // convert the current environment into an occupancy grid
  cout << "sampleXOccGrid : " << sampleXOccGrid << endl;
  cout << "sampleYOccGrid : " << sampleYOccGrid << endl;
  //this->map_ = this->env_->sample(this->env_->getDfltX(), this->env_->getDfltY(), sampleXOccGrid, sampleYOccGrid);
  this->map_ = this->env_->sample(sampleXOccGrid, sampleYOccGrid);
  
  if(this->map_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Converting the Environment","The resulting occupancy grid is nul.");      
      return -2;
    }
  
  // compute the time past
  gettimeofday(&t2,NULL);
  time_t elapsedTs = t2.tv_sec - t1.tv_sec;
  suseconds_t elapsedTus = t2.tv_usec - t1.tv_usec;
  double elapsedTime = elapsedTs + 0.000001*elapsedTus;
  
  // save the occupancy grid in a pgm file
  if(this->map_->save(mapName.text()) != 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Converting the Environment","The ouput pgm file cannot be saved");      
      return -3;
    }
  
  FXString msg("The environment has been sucessfully converted into an occupancy grid in ");
  msg += FXStringVal(elapsedTime,2,FALSE);
  msg += FXString(" sec.");
  msgBox->setText(msg);
  return 0;
}

long MPC_TeddyIfc::changeOccGridName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Occupancy Grid");
  open.setFilename(occgridName);
  if(open.execute()){
    occgridName = open.getFilename();
  }
  return 1;
}

long MPC_TeddyIfc::updateParallel(FXObject*,FXSelector,void*)
{
  // update the parallelization of trajectory generation
  if(parallelize == false)
    parallelize = true;
  else
    parallelize = false;
  return 1;
}

long MPC_TeddyIfc::computeDiffDriveTraj(FXObject*,FXSelector,void*)
{
  DDStateT initCfg((double)initXRb,(double)initYRb,0,0,0,0);
  return this->computeNewDiffDriveTraj(initCfg);
}

long MPC_TeddyIfc::computeItscTraj(FXObject*,FXSelector,void*)
{
  return this->computeIntersectionTraj();
}

long MPC_TeddyIfc::computeMtwyTraj(FXObject*,FXSelector,void*)
{
  return this->computeMotorwayTraj();
}

long MPC_TeddyIfc::initTrajFollowing(FXObject*,FXSelector,void*)
{
  // check if a carlike trajectory has been computed
  if((trajType != CL_TRAJ && trajType != WC_TRAJ) || this->traj_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error while initializing the trajectory follower","No trajectory has been loaded or wrong trajectory type");
      return -1;
    }

  II_ = 0;
  medDerivLoc_ = 0.0;
  medDerivRb_ = 0.0;

  // init the process type
  processType_ = TF_PROC;

  // init the vectors to stock the trajectory followed
  if(TF_X != NULL) delete TF_X;
  if(TF_Y != NULL) delete TF_Y;
  if(TF_T != NULL) delete TF_T;
  if(TF_locX != NULL) delete TF_locX;
  if(TF_locY != NULL) delete TF_locY;
  if(TF_rbX != NULL) delete TF_rbX;
  if(TF_rbY != NULL) delete TF_rbY;
  if(TF_DerivLoc != NULL) delete TF_DerivLoc;
  if(TF_ErrorLoc != NULL) delete TF_ErrorLoc;
  if(TF_DerivRb != NULL) delete TF_DerivRb;
  if(TF_rbCtrl1 != NULL) delete TF_rbCtrl1;
  if(TF_rbCtrl2 != NULL) delete TF_rbCtrl2;
  if(TF_noisedCtrl1 != NULL) delete TF_noisedCtrl1;
  if(TF_noisedCtrl2 != NULL) delete TF_noisedCtrl2;

  TF_X = new vector<double>;
  TF_Y = new vector<double>;
  TF_T = new vector<double>;
  TF_locX = new vector<double>;
  TF_locY = new vector<double>;
  TF_rbX = new vector<double>;
  TF_rbY = new vector<double>;
  TF_DerivLoc = new vector<double>;
  TF_ErrorLoc = new vector<double>;
  TF_DerivRb = new vector<double>;
  TF_rbCtrl1 = new vector<double>;
  TF_rbCtrl2 = new vector<double>;
  TF_noisedCtrl1 = new vector<double>;
  TF_noisedCtrl2 = new vector<double>;


  if(TF_mode == CL_TF_MODE)
    {
      // init the car-like system constraints
      this->initCarLikeSys();
      
      // initialize the car-like state
      if(this->robotStateTF_ != NULL) delete this->robotStateTF_;
      this->robotStateTF_ = new CLStateT(initXRb,initYRb,initThRb,initVRb,initPhRb,0.0);

      // initialize the trajectory follower
      this->follower_ = new CLTrajFollow((CarLikeSys*)sys_,(CLSTraj*)traj_);
      
    }
  else if(TF_mode == CC_TF_MODE)
    {
      // init the car-like system constraints
      this->initCarLikeSys();
      
      // initialize the car-like state
      if(this->robotStateTF_ != NULL) delete this->robotStateTF_;
      this->robotStateTF_ = new CLStateT(initXRb,initYRb,initThRb,initVRb,initPhRb,0.0);

      // initialize the trajectory follower
      this->follower_ = new CCTrajFollow((CarLikeSys*)sys_,(CLSTraj*)traj_);
    }
  else if(TF_mode == WC_TF_MODE)
    {
      // init the car-like system constraints
      this->initWheelChairSys();
      
      // initialize the car-like state
      if(this->robotStateTF_ != NULL) delete this->robotStateTF_;
      this->robotStateTF_ = new WCStateT(initXRb,initYRb,initThRb,initVRb,0.0,0.0);

      // initialize the trajectory follower
      this->follower_ = new WCTrajFollow((WheelChairSys*)sys_,(WCSTraj*)traj_);
    }  

  // draw the scene
  this->drawScene();
  return 1;
}

long MPC_TeddyIfc::startTF(FXObject*,FXSelector,void*)
{
  // verifying that the trajectory and the follower are not NULL
  if(this->traj_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following","The initial trajectory hasn't been computed");
      return -1;
    }
  else if((TF_mode == CL_TF_MODE || TF_mode == CC_TF_MODE) && this->follower_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following","The trajectory follower must be initialized");
      return -2;
    }

  // saving the initial trajectory
  if(time == 0)
    {
      switch(TF_mode)
	{
	case CL_TF_MODE : 
	case CC_TF_MODE : 
	  this->initTraj_ = new CLSTraj(*((CLSTraj*)(this->traj_)));
	  break;
	case WC_TF_MODE : 
	  this->initTraj_ = new WCSTraj(*((WCSTraj*)(this->traj_)));
	  break;
	}
    }
  
  cout << "starting timeout" << endl;
  // init the timer
  gettimeofday(&t1,NULL);
  running = true;
  getApp()->addTimeout(this,ID_TIMEOUT_TF,TIMER_INTERVAL);

  oldTimeTs = t1.tv_sec;
  oldTimeTus = t1.tv_usec;
//   cout << "oldTime (start) : " << oldTimeTs << "," << oldTimeTus << endl;
  return 1;
}

long MPC_TeddyIfc::oneStepTF(FXObject*,FXSelector,void*)
{
  // verifying that the trajectory and the follower are not NULL
  if(this->traj_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following process","The initial trajectory hasn't been computed");
      return -1;
    }
  else if((TF_mode == CL_TF_MODE || TF_mode == CC_TF_MODE || TF_mode == WC_TF_MODE) && this->follower_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error Starting the Trajectory Following","The trajectory follower must be initialized");
      return -2;
    }
  
  // init the timer
  gettimeofday(&t1,NULL);
  oneStep = true;
  getApp()->addTimeout(this,ID_TIMEOUT_TF,TIMER_INTERVAL);

  oldTimeTs = t1.tv_sec;
  oldTimeTus = t1.tv_usec;
//   cout << "oldTime (one step) : " << oldTimeTs << "," << oldTimeTus << endl;
  return 1;
}

long MPC_TeddyIfc::stopTF(FXObject*,FXSelector,void*)
{
  cout << "stoping timeout" << endl;
  running = false;
  getApp()->removeTimeout(this,ID_TIMEOUT_TF);
  alreadyElapsedTime = time;

  this->displayTFTraj();
  return 1;
}

long MPC_TeddyIfc::restartTF(FXObject*,FXSelector,void*)
{
  // TODO... 
  FXMessageBox::error(this,MBOX_OK,"Error","TODO!!!");
  return -1;
}

long MPC_TeddyIfc::updateTF(FXObject*,FXSelector,void*)
{
  // update obstacles random motion parameters
//   DynamicObs::setSeuilChgt(seuilChgt);
//   DynamicObs::setMaxAngleChgt(maxAngleChgt);
//   DynamicObs::setMaxNormChgt(maxNormChgt);

  // compute the time before the following process
  struct timeval t3;
  gettimeofday(&t3,NULL);

  // compute the trajectory folowing process
  int status = this->followTraj();

  // compute the time past
  gettimeofday(&t2,NULL);
  time_t elapsedTs = t2.tv_sec - t1.tv_sec;
  suseconds_t elapsedTus = t2.tv_usec - t1.tv_usec;

  double elapsedT = alreadyElapsedTime + ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);

  time_t timeTs = t2.tv_sec - oldTimeTs;
  suseconds_t timeTus = t2.tv_usec - oldTimeTus;
  trueLastElapsedT = ((double)(timeTs)) + 0.000001 * ((double)(timeTus));
//   cout << "trueLastElapsedT : " << trueLastElapsedT << endl;

  oldTimeTs = t2.tv_sec;
  oldTimeTus = t2.tv_usec;
//   cout << "oldTime (update) : " << oldTimeTs << "," << oldTimeTus << endl;

  // set the current time
  time = elapsedT;

  // compute the trajectory following duration
  time_t deformTs = t2.tv_sec - t3.tv_sec;
  suseconds_t deformTus = t2.tv_usec - t3.tv_usec;
  deformDur = ((double)deformTs) + 0.000001 * ((double)deformTus);
  if(deformDur != 0) lastFPS = 1.0/deformDur;
  else lastFPS = 0.0;

  switch(status)
    {
    case 0 : 
      
      break;
    case -1 : 
//       cout << "MPC_TeddyIfc::updateTF() : Warning. The trajectory generation process failed at time << " << time << "." << endl;
      break;
    }

  // check if the end time of the trajectory has been reached
  if(time >= this->traj_->at(traj_->getSize()-1).t())
    {
      cout << "fin traj" << endl;
      
      // stop the following process
      this->stopTF(NULL,0,NULL);
    }
  
  // repaint the viewer
  this->drawScene();
  
  // rerun the timeout
  if(oneStep)
    {
      alreadyElapsedTime = time;
      oneStep = false;
    }
  else if(running)
    {
      getApp()->addTimeout(this,ID_TIMEOUT_TF,TIMER_INTERVAL);
    }

  // update the FPS number
  nbTimeout++;
  FPS = ((double)(totalNbTimeout + nbTimeout))/time;

  return 1;
}

long MPC_TeddyIfc::updateDrawTrajGenTF(FXObject*,FXSelector,void*)
{
  // update the drawing of generated trajectories
  if(drawTrajGenTF == false)
    drawTrajGenTF = true;
  else
    drawTrajGenTF = false;
  return 1;
}

long MPC_TeddyIfc::updateRbGoalPos(FXObject*,FXSelector,void*)
{
  // update the choice of a fix goal position for motion planning
  if(fixRbGoalPos == false)
    fixRbGoalPos = true;
  else
    fixRbGoalPos = false;
  return 1;
}

long MPC_TeddyIfc::dispTFTraj(FXObject*,FXSelector,void*)
{
  this->displayTFTraj();
  return 1;
}

long MPC_TeddyIfc::dispJoinTraj(FXObject*,FXSelector,void*)
{
  this->displayJoinTraj();
  return 1;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Other functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// initialize the deformation process
void MPC_TeddyIfc::initDeform()
{
  // check if the trajectory is not null
  if(this->traj_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error during deformation init","The input trajectory has not been defined");
      msgBox->setText("Error during deformation init. The input trajectory has not been defined");
      return;
    }
  
  // check if the environment is not null
  if(this->env_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error during deformation init","The environment has not been defined");
      msgBox->setText("Error during deformation init. The environment has not been defined");
      return;
    }

  // check if the system is not null
  if(this->sys_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Error during deformation init","The robotic system has not been defined");
      msgBox->setText("Error during deformation init. The robotic system has not been defined");
      return;
    }  

  processType_ = TEDDY_PROC;

  // init the table of state types
  typeState = new int*;
  (*typeState) = new int[this->traj_->getSize()];
  for(int i=0; i < this->traj_->getSize(); i++)
    {
      (*typeState)[i] = 0;
    }
  
//   //init the table of nearest states
//   tableNearState = new PositionXYT**;
//   (*tableNearState) = new PositionXYT* [this->traj_->getSize()];
//   for(int i=0; i < this->traj_->getSize(); i++)
//     {
//       (*tableNearState)[i] = NULL;
//     }
  
//   // init the tables of internal and external forces
//   tableFInt = new StateTime*;
//   *tableFInt = NULL;
//   tableFExt = new StateTime*;
//   *tableFExt = NULL;

  switch(trajType)
    {
    case PM_TRAJ : 
      {
	// reinit the deformer
	if(deformer_ != NULL) delete deformer_;
	//deformer_ = new PMTrajDeform((PMSTraj*)(this->traj_), this->env_, (PointMassSys*)(this->sys_), typeState, (PositionXYT***)tableNearState, (StateTime**)tableFInt, (StateTime**)tableFExt);
	deformer_ = new PMTrajDeform((PMSTraj*)(this->traj_), this->env_, (PointMassSys*)(this->sys_), 1);
	break;
      }
    case CL_TRAJ : 
      {
	// reinit the deformer
	if(deformer_ != NULL) delete deformer_;
	//deformer_ = new CLTrajDeform((CLSTraj*)(this->traj_), this->env_, (CarLikeSys*)(this->sys_), typeState, (PositionXYT***)tableNearState, (StateTime**)tableFInt, (StateTime**)tableFExt);
	deformer_ = new CLTrajDeform3((CLSTraj*)(this->traj_), this->env_, (CarLikeSys*)(this->sys_), 1);

// 	((CLTrajDeform3*)deformer_)->setMaxResidual(maxResidual);
// 	((CLTrajDeform3*)deformer_)->setMaxResidualAccepted(20.0);
// 	((CLTrajDeform3*)deformer_)->setMaxTGnbSteps(maxNbStep);
	break;
      }
    case WC_TRAJ : 
	// reinit the deformer
	if(deformer_ != NULL) delete deformer_;
	//deformer_ = new WCTrajDeform((WCSTraj*)(this->traj_), this->env_, (WheelChairSys*)(this->sys_), typeState, (PositionXYT***)tableNearState, (StateTime**)tableFInt, (StateTime**)tableFExt);
	deformer_ = new WCTrajDeform3((WCSTraj*)(this->traj_), this->env_, (WheelChairSys*)(this->sys_), 1);

// 	((WCTrajDeform3*)deformer_)->setMaxResidual(maxResidual);
// 	((WCTrajDeform3*)deformer_)->setMaxResidualAccepted(20.0);
// 	((WCTrajDeform3*)deformer_)->setMaxTGnbSteps(maxNbStep);
	break;
    case DD_TRAJ : 
      FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::initDeform DD_TRAJ","This function is no more handled :-(");
      msgBox->setText("This function is no more handled :-(");
      return;
      break;
    }

  // initialize the trajectory deformation options
  deformer_->setLastDeformDuration(deformDur);
  if(optModeDeform == 0) deformer_->setModeDeform(SPATIOTEMP_DEFORM);
  else if(optModeDeform == 1) deformer_->setModeDeform(SPATONLY_DEFORM);
  else if(optModeDeform == 2) deformer_->setModeDeform(TEMPONLY_DEFORM);
  deformer_->setForcesCoef(K_INT, K_EXT, K_STEXT, alphaMin, alphaMax, betaMin, betaMax);
  deformer_->setMaxEffectDist(maxSpDist,maxTmpDist,maxStEffectDist);
  deformer_->setReliability(predHorizon, reliabilityTime, epsilonRT);
  deformer_->setTemporalConst(&useTemporalConsistency);
  deformer_->setTimeStep(trajTimeStep);
  deformer_->setPosStep(trajPosStep);
  
  // initialize the trajectory generation options
  TrajGen::setMaxNbStep(maxNbStep);                // set the maximal number of steps
  TrajGen::setMaxResidual(maxResidual);            // set the maximal residual allowed
  TrajGen::setDBG(debugLevel);                     // set the debugging level

//       // init the trajectory generation parameters
//       DDTrajGen::setMaxNbStep(maxNbStep);                // set the maximal number of steps
//       DDTrajGen::setMaxResidual(maxResidual);            // set the maximal residual allowed
//       DDTrajGen::setDBG(debugLevel);                     // set the debugging level
//       CLTrajGen::setMaxNbStep(maxNbStep);                // set the maximal number of steps
//       CLTrajGen::setMaxResidual(maxResidual);            // set the maximal residual allowed
//       CLTrajGen::setDBG(debugLevel);                     // set the debugging level
//       CLTrajGen::setUseVelCst(useVelCst);                // set true if linear velocity constraints are used
//       CLTrajGen::setUseAccCst(useAccCst);                // set true if linear acceleration constraints are used
//       CLTrajGen::setUseAVelCst(useAVelCst);              // set true if angular velocity constraints are used
//       CLTrajGen::setUseAAccCst(useAAccCst);              // set true if angular acceleration constraints are used
}

int MPC_TeddyIfc::deformTraj()
{
  if(this->traj_ == NULL)
    {
      // if the trajectory hasn't been computed, the deformation must be stoped (this cas should'nt appear)
      this->stopTraj(NULL, 0, NULL);
      return -1;
    }
  else
    {
      // initialize the deformation process
      this->initDeform();

      // update the new knowledge of the environment
      env_->simuRandMotion(time);

      // launching the deformation algorithm
      glcanvas->makeCurrent();

      int status = deformer_->deform(time, curStep);
      //int status = 0;
      glcanvas->makeNonCurrent();

      if(this->path_ != NULL) delete this->path_;
      this->path_ = this->traj_->getPath();

      return status;
    }
}

// initialize the traj following process params that can change at each step
void MPC_TeddyIfc::initTF()
{
  processType_ = TF_PROC;
  if(this->TF_mode == CL_TF_MODE)
    {
      if(this->follower_ == NULL) return;

      // initialize the trajectory process parameters
      this->follower_->dtReach() = TF_dt;
      this->follower_->timeMeeting() = TF_timeMeeting;
    }
  else if(this->TF_mode == CC_TF_MODE)
    {
      if(this->follower_ != NULL) return;

      // initialize the trajectory process parameters
      this->follower_->dtReach() = TF_dt;
      this->follower_->timeMeeting() = TF_timeMeeting;
    }
  
  // initialize the noise parameters
  CLStateT::maxNoiseXY_ = noiseX;
  CLStateT::maxNoiseTH_ = noiseTH;
  CLStateT::maxNoiseS_  = noiseV;
  CLStateT::maxNoisePH_ = noisePH;
  
  CLControl::maxNoiseAcc_ = noiseAcc;
  CLControl::maxNoiseSS_ = noiseSS;

//   CLStateT::maxNoiseXY_ = 500;
//   CLStateT::maxNoiseTH_ = 1.0;
//   CLStateT::maxNoiseS_  = 0;
//   CLStateT::maxNoisePH_ = 0;

//   CLControl::maxNoiseAcc_ = 0;
//   CLControl::maxNoiseSS_ = 0;
}

// execute a step of the following algorithm
int MPC_TeddyIfc::followTraj()
{
  // init the trajectory following process
  this->initTF();
  
  // check if the state of the robot is available
  if(robotStateTF_ == NULL)
    {
      cout << "MPC_TeddyIfc::followTraj : Warning. The robot state is nul" << endl;
      return -1;
    }

  int res;
  if(this->TF_mode == CL_TF_MODE)
    {
      // compute a noised localization from the robot state
      CLStateT noisedS = ((CLStateT*)robotStateTF_)->noisedState(trueLastElapsedT);

      CLStateT sOnTraj;
      ((CLSTraj*)traj_)->getStateAtTime(time,sOnTraj);
      PositionXYTh noisedPos = noisedS.posXYTh();
      PositionXYTh rbPos = robotStateTF_->posXYTh();

//       cout << "time : " << time << endl;
//       cout << "state on traj : "; sOnTraj.display(); cout << endl;
//       cout << "loc : "; noisedS.display(); cout << endl;
//       cout << "robot state : "; robotStateTF_->display(); cout << endl;
      
      TF_X->push_back(sOnTraj.x());
      TF_Y->push_back(sOnTraj.y());
      TF_T->push_back(time);
      TF_locX->push_back(sOnTraj.x() + 3.0*(noisedPos.x() - sOnTraj.x()));
      TF_locY->push_back(sOnTraj.y() + 3.0*(noisedPos.y() - sOnTraj.y()));
      TF_rbX->push_back(rbPos.x());
      TF_rbY->push_back(rbPos.y());

      // init the control that will be computed by the following process
      CLControl ctrl;

      // perform the trajectory following process
      res = ((CLTrajFollow*)follower_)->getControl(noisedS,            // localization of the robot (noised state)
						   noisedS.t() + trueLastElapsedT,  // use the previous TF process duration to evaluate the time at which will be applied the control
						   trueLastElapsedT,   // use the previous TF process duration as the control application duration
						   ctrl);              // control computed by the TF process
      
      // get back the state the trajectory follower try to reach (for displaying)
      if(stateToReachTF_ != NULL) delete stateToReachTF_;
      stateToReachTF_ = new CLStateT;
      ((CLTrajFollow*)follower_)->selectStateToReach(*((CLStateT*)stateToReachTF_));

      // compute a noised control from the control computed by the TF process
      CLControl noisedC = ctrl.noisedCtrl(trueLastElapsedT);
      
      /// compute the state reached by the previous real state and the noised control
	
//       // initialize the init state
//       double* initTS = NULL;
//       CLStateT_1_CL2tabD(*((CLStateT*)robotStateTF_),&initTS,((CarLikeSys*)sys_)->getLength());
      
//       // initialize the system constraints
//       double* cstt = NULL;
//       CLSys_2_CL2TabCstt(*((CarLikeSys*)sys_),&cstt);
      
      // initialize the parameters IG
      double* param = NULL;
      CLctrlt_2_CL2IG(noisedC,trueLastElapsedT,&param);
      
      // compute the state reached
      CLTrajGen trajgen((CarLikeSys*)sys_);
      trajgen.fixFinalXY(true);
      trajgen.fixFinalTH(true);
      trajgen.fixFinalV(false);
      trajgen.fixFinalPH(false);

//       CLTrajGen::setUseVelCst(true);
//       CLTrajGen::setUseAccCst(true);
//       CLTrajGen::setUseStACst(true);
//       CLTrajGen::setUseStSCst(false);

//       trajgen.ComputeFromIG(initTS,cstt,param);
      trajgen.ComputeFromIG(*robotStateTF_,param);
      
      // get back the state reached
      double oldT = robotStateTF_->t();
      double tabSReached[6];
      trajgen.getFinalStateTReached(tabSReached);
      
      // update the local state of the robot
      ((CLStateT*)robotStateTF_)->x() = (tabSReached[0]);
      ((CLStateT*)robotStateTF_)->y() = (tabSReached[1]);
      ((CLStateT*)robotStateTF_)->th() = (tabSReached[2]);
      ((CLStateT*)robotStateTF_)->ph() = (tabSReached[3]);
      ((CLStateT*)robotStateTF_)->s() = (tabSReached[4]);
      ((CLStateT*)robotStateTF_)->t() = (oldT + tabSReached[5]);  
    }
  else if(this->TF_mode == CC_TF_MODE)
    {
      // compute a noised localization from the robot state
      CLStateT noisedS = ((CLStateT*)robotStateTF_)->noisedState(trueLastElapsedT);

      CLStateT sOnTraj;
      ((CLSTraj*)traj_)->getStateAtTime(time,sOnTraj);
      PositionXYTh noisedPos = noisedS.posXYTh();
      PositionXYTh rbPos = robotStateTF_->posXYTh();
      
//       cout << "time : " << time << endl;
//       cout << "trueLastElapsedT : " << trueLastElapsedT << endl;
//       cout << "state on traj : "; sOnTraj.display(); cout << endl;
//       cout << "loc : "; noisedS.display(); cout << endl;
//       cout << "robot state : "; robotStateTF_->display(); cout << endl;
      
      // init the control that will be computed by the following process
      CCControl ctrl;

      // perform the trajectory following process
      res = ((CCTrajFollow*)follower_)->getControl(noisedS,            // localization of the robot (noised state)
						   noisedS.t() + trueLastElapsedT,  // use the previous TF process duration to evaluate the time at which will be applied the control
						   trueLastElapsedT,   // use the previous TF process duration as the control application duration
						   ctrl);              // control computed by the TF process
      
      // get back the state the trajectory follower try to reach (for displaying)
      if(stateToReachTF_ != NULL) delete stateToReachTF_;
      stateToReachTF_ = new CLStateT;
      ((CCTrajFollow*)follower_)->selectStateToReach(noisedS,*((CLStateT*)stateToReachTF_));

      // compute a noised control from the control computed by the TF process
      CCControl noisedC = ctrl.noisedCtrl(trueLastElapsedT);
      CLStateT rbStateCopy(*((CLStateT*)robotStateTF_));
      rbStateCopy.s() = noisedC.v();
      rbStateCopy.ph() = noisedC.ph();
      
//       cout << "CCControl : "; noisedC.display(); cout << endl;

      double dxLoc = noisedPos.x() - sOnTraj.x();
      double dyLoc = noisedPos.y() - sOnTraj.y();
      double dxErrLoc = noisedPos.x() - rbPos.x();
      double dyErrLoc = noisedPos.y() - rbPos.y();
      double dxRb = rbPos.x() - sOnTraj.x();
      double dyRb = rbPos.y() - sOnTraj.y();
      //medDerivLoc_ += (sqrt(dxLoc*dxLoc + dyLoc*dyLoc));
      medDerivLoc_ += (sqrt(dxErrLoc*dxErrLoc + dyErrLoc*dyErrLoc));
      medDerivRb_ += sqrt(dxRb*dxRb + dyRb*dyRb);

      if(II_%40 == 0)
	{
	  cout << "II_ : " << II_ << endl;

	  TF_X->push_back(sOnTraj.x());
	  TF_Y->push_back(sOnTraj.y());
	  TF_T->push_back(time);
	  TF_locX->push_back(noisedPos.x());
	  TF_locY->push_back(noisedPos.y());
	  TF_rbX->push_back(rbPos.x());
	  TF_rbY->push_back(rbPos.y());
	  
	  //TF_DerivLoc->push_back(medDerivLoc_ / 40.0);
	  TF_ErrorLoc->push_back(medDerivLoc_ / 40.0);
	  TF_DerivRb->push_back(medDerivRb_ / 40.0);
	  medDerivLoc_ = 0.0;
	  medDerivRb_ = 0.0;

	  TF_rbCtrl1->push_back(ctrl.v());
	  TF_rbCtrl2->push_back(ctrl.ph());
	  TF_noisedCtrl1->push_back(noisedC.v());
	  TF_noisedCtrl2->push_back(noisedC.ph());
	}
      II_++;

      /// compute the state reached by the previous real state and the noised control
	
      // initialize the init state
      double* initTS = NULL;
      CLStateT_1_CLtabD(rbStateCopy,&initTS,((CarLikeSys*)sys_)->getLength());
      
//       cout << "initTS :(" << initTS[0] << "," << initTS[1] << "," << initTS[2] << "," << 
// 	   initTS[3] << "," << initTS[4] << "," << initTS[5] << ")" << endl;

      // initialize the system constraints
      double* cstt = NULL;
      CLSys_2_CL3TabCstt(*((CarLikeSys*)sys_),&cstt);
      
      // initialize the parameters IG
      double* param = NULL;
      CCctrlt_2_CL3IG(noisedC,trueLastElapsedT,&param,*((CLStateT*)robotStateTF_),((CarLikeSys*)sys_)->getLength());
      
//       cout << "param :(" << param[0] << "," << param[1] << "," << param[2] << "," << param[3] << "," 
// 	   << param[4] << "," << param[5] << "," << param[6] << ")" << endl;

      // compute the state reached
      CLTrajGen trajgen((CarLikeSys*)sys_);
      trajgen.fixFinalXY(true);
      trajgen.fixFinalTH(true);
      trajgen.fixFinalV(false);
      trajgen.fixFinalPH(false);

      //trajgen.setDBG(2);
      trajgen.setTcStatus(TIME_FIXED,0,0);
      trajgen.TrajGen::ComputeFromIG(initTS,param);
      //trajgen.displayProfiles();
      
      // get back the state reached
      double oldT = robotStateTF_->t();
      double tabSReached[6];
      trajgen.getFinalStateTReached(tabSReached);
      
      // update the local state of the robot
      ((CLStateT*)robotStateTF_)->x() = (tabSReached[0]);
      ((CLStateT*)robotStateTF_)->y() = (tabSReached[1]);
      ((CLStateT*)robotStateTF_)->th() = (tabSReached[2]);
      ((CLStateT*)robotStateTF_)->ph() = (tabSReached[3]);
      ((CLStateT*)robotStateTF_)->s() = (tabSReached[4]);
      ((CLStateT*)robotStateTF_)->t() = (oldT + tabSReached[5]);
    }
  else if(this->TF_mode == WC_TF_MODE)
    {
      // compute a noised localization from the robot state
      WCStateT noisedS = ((WCStateT*)robotStateTF_)->noisedState(trueLastElapsedT);
      
      // init the control that will be computed by the following process
      WCControl ctrl;

      // perform the trajectory following process
      res = ((WCTrajFollow*)follower_)->getControl(noisedS,            // localization of the robot (noised state)
						   noisedS.t() + trueLastElapsedT,  // use the previous TF process duration to evaluate the time at which will be applied the control
						   trueLastElapsedT,   // use the previous TF process duration as the control application duration
						   ctrl);              // control computed by the TF process
      
      // get back the state the trajectory follower try to reach (for displaying)
      if(stateToReachTF_ != NULL) delete stateToReachTF_;
      stateToReachTF_ = new WCStateT;
      ((WCTrajFollow*)follower_)->selectStateToReach(noisedS,*((WCStateT*)stateToReachTF_));

      // compute a noised control from the control computed by the TF process
      WCControl noisedC = ctrl.noisedCtrl(trueLastElapsedT);
      WCStateT rbStateCopy(*((WCStateT*)robotStateTF_));
      rbStateCopy.s() = noisedC.v();
      rbStateCopy.om() = noisedC.av();
      
//       cout << "WCControl : "; noisedC.display(); cout << endl;
//       cout << "rbStateCopy : "; rbStateCopy.display(); cout << endl;

      // compute the state reached by the previous real state and the noised control
      
      // initialize the parameters IG
      double* param = NULL;
      WCctrlt_2_WCIG(noisedC,trueLastElapsedT,&param);
      
      // compute the state reached
      WCTrajGen trajgen((WheelChairSys*)sys_);
      //trajgen.setDBG(2);
      trajgen.setTcStatus(TIME_FIXED,0,0);
      trajgen.ComputeFromIG(rbStateCopy,param);
      
      trajgen.displayAllFct();

      // get back the state reached
      double oldT = robotStateTF_->t();
      double tabSReached[6];
      trajgen.getFinalStateTReached(tabSReached);
      
      // update the local state of the robot
      ((WCStateT*)robotStateTF_)->x() = (tabSReached[0]);
      ((WCStateT*)robotStateTF_)->y() = (tabSReached[1]);
      ((WCStateT*)robotStateTF_)->th() = (tabSReached[2]);
      ((WCStateT*)robotStateTF_)->s() = (tabSReached[3]);
      ((WCStateT*)robotStateTF_)->om() = (tabSReached[4]);
      ((WCStateT*)robotStateTF_)->t() = (oldT + tabSReached[5]);

      //cout << "robotStateTF_ : "; robotStateTF_->display(); cout << endl;
    }
  //cout << "res : " << res << endl;
  return res;
}

// display the trajectories computed during the trajectory following process
void MPC_TeddyIfc::displayTFTraj()
{
  cout << "TF_T->size() : " << TF_T->size() << endl;
  cout << "TF_rbCtrl2->size() : " << TF_rbCtrl2->size() << endl;
  cout << "TF_noisedCtrl2->size() : " << TF_noisedCtrl2->size() << endl;
  int N = TF_noisedCtrl2->size();
  cout << "TF_noisedCtrl2 : " << endl;
  for(int i=0; i<N; i++) cout << TF_noisedCtrl2->at(i) << " ";
  cout << endl;
  cout << "TF_rbCtrl2 : " << endl;
  for(int i=0; i<N; i++) cout << TF_rbCtrl2->at(i) << " ";
  cout << endl;

  Gnuplot gTFTraj;
  gTFTraj.reset_plot();
  //gTFTraj.set_style("points").plot_xyz(*TF_locX,*TF_locY,*TF_T,"Localization");
  gTFTraj.set_style("lines").plot_xyz(*TF_X,*TF_Y,*TF_T,"Trajectory to follow");	
  gTFTraj.set_style("points").plot_xyz(*TF_rbX,*TF_rbY,*TF_T,"Real robot's traj");

  Gnuplot gCtrl1;
  gCtrl1.reset_plot();
  gCtrl1.set_style("lines").plot_xy(*TF_T,*TF_rbCtrl1,"Speed control computed");
  gCtrl1.set_style("points").plot_xy(*TF_T,*TF_noisedCtrl1,"Noised speed applied");	

  Gnuplot gCtrl2;
  //gCtrl2.reset_plot();
//   gCtrl2.set_style("lines").plot_xy(*TF_T,*TF_rbCtrl2,"Steering ctrl computed");
//   gCtrl2.set_style("lines").plot_xy(*TF_T,*TF_noisedCtrl2,"Noised steering applied");
  gCtrl2.set_style("lines").plot_xy(*TF_T,*TF_rbCtrl2,"");
  gCtrl2.set_style("lines").plot_xy(*TF_T,*TF_noisedCtrl2,"");

  Gnuplot gDeriv;
  gCtrl2.reset_plot();
  gCtrl2.set_style("lines").plot_xy(*TF_T,*TF_ErrorLoc,"Localization error");
  gCtrl2.set_style("lines").plot_xy(*TF_T,*TF_DerivRb,"Robot derivation");

  wait_for_key();
}

void MPC_TeddyIfc::displayJoinTraj()
{
  CLStateT sinit(150,150,0.25*M_PI,0,0,15);
  CLStateT sgoal;
  ((CLSTraj*)traj_)->getStateAtTime(25,sgoal);

  vector<double> XinitPos;
  XinitPos.push_back(sinit.x());
  vector<double> YinitPos;
  YinitPos.push_back(sinit.y());
  vector<double> TinitPos;
  TinitPos.push_back(sinit.t());

  vector<double> XgoalPos;
  XgoalPos.push_back(sgoal.x());
  vector<double> YgoalPos;
  YgoalPos.push_back(sgoal.y());
  vector<double> TgoalPos;
  TgoalPos.push_back(sgoal.t());

  this->initCarLikeSys();

  CLTrajGen trajgen((CarLikeSys*)sys_);
  trajgen.fixFinalXY(true);
  trajgen.fixFinalTH(true);
  trajgen.fixFinalV(false);
  trajgen.fixFinalPH(false);

  int res = trajgen.ComputeTrajectory(sinit,sgoal,(double**)NULL);

  CLSTraj* joinTraj = trajgen.sampledTraj(0.1);

  vector<double> Xtraj;
  vector<double> Ytraj;
  vector<double> Ttraj;

  int N = traj_->size();
  for(int i=0; i<N; i++)
    {
      PositionXYT pos = traj_->at(i).posXYT();
      Xtraj.push_back(pos.x());
      Ytraj.push_back(pos.y());
      Ttraj.push_back(pos.t());
    }

  vector<double> Xtraj2;
  vector<double> Ytraj2;
  vector<double> Ttraj2;

  int N2 = joinTraj->size();
  for(int i=0; i<N2; i++)
    {
      PositionXYT pos = joinTraj->at(i).posXYT();
      Xtraj2.push_back(pos.x());
      Ytraj2.push_back(pos.y());
      Ttraj2.push_back(pos.t() + sinit.t());
    }

  Gnuplot gTraj;
  gTraj.reset_plot();
  gTraj.set_style("lines").plot_xyz(Xtraj,Ytraj,Ttraj,"Trajectory to follow");
  gTraj.set_style("points").plot_xyz(XinitPos,YinitPos,TinitPos,"Current state-time");
  gTraj.set_style("points").plot_xyz(XgoalPos,YgoalPos,TgoalPos,"State-time to reach (s_r,t_r)");
  gTraj.set_style("lines").plot_xyz(Xtraj2,Ytraj2,Ttraj2,"Join trajectory");

  wait_for_key();
}

// initialize the robot system constraints
void MPC_TeddyIfc::initRobot()
{
  ((CarLikeSys*)sys_)->setVMax(maxSpd);
  ((CarLikeSys*)sys_)->setAMax(maxAcc);
  ((CarLikeSys*)sys_)->setMaxSteering(maxSteer);
  ((CarLikeSys*)sys_)->setMaxSteerSpd(maxSteerS);
  ((CarLikeSys*)sys_)->setLength(robotLength);
}

// initialize the car-like system constraints
void MPC_TeddyIfc::initCarLikeSys()
{
  if(sys_ != NULL) delete sys_;
  sys_ = new CarLikeSys();

  ((CarLikeSys*)sys_)->setVMax(maxSpd);
  ((CarLikeSys*)sys_)->setAMax(maxAcc);
  ((CarLikeSys*)sys_)->setMaxSteering(maxSteer);
  ((CarLikeSys*)sys_)->setMaxSteerSpd(maxSteerS);
  ((CarLikeSys*)sys_)->setLength(robotLength);
}

// initialize the point mass system constraints
void MPC_TeddyIfc::initPointMassSys()
{
  if(sys_ != NULL) delete sys_;
  sys_ = new PointMassSys();

  ((PointMassSys*)sys_)->setRadius(robotLength);
  ((PointMassSys*)sys_)->setVXMax(maxSpd);
  ((PointMassSys*)sys_)->setVYMax(maxSpd);
  ((PointMassSys*)sys_)->setAXMax(maxAcc);
  ((PointMassSys*)sys_)->setAYMax(maxAcc);
}

// initialize the wheelchair system constraints
void MPC_TeddyIfc::initWheelChairSys()
{
  if(sys_ != NULL) delete sys_;
  sys_ = new WheelChairSys();

  ((WheelChairSys*)sys_)->setLength(robotLength);
  ((WheelChairSys*)sys_)->setWidth(robotLength);
//   ((WheelChairSys*)sys_)->setWheelRad();
//   ((WheelChairSys*)sys_)->setWheelBase();
  ((WheelChairSys*)sys_)->setVMax(maxSpd);
  ((WheelChairSys*)sys_)->setAMax(maxAcc);
//   ((WheelChairSys*)sys_)->setAVMax();
//   ((WheelChairSys*)sys_)->setAAMax();
//   ((WheelChairSys*)sys_)->setBrakingDist();

}

// init the tables of internal forces, external forces and nearest points
void MPC_TeddyIfc::initTabForces()
{
  cout << "MPC_TeddyIfc::initTabForces -> traj_->type() :  " << traj_->type() << endl;

  // init the table of state types
  typeState = new int*;
  (*typeState) = new int[this->traj_->getSize()];
  for(int i=0; i < this->traj_->getSize(); i++)
    {
      (*typeState)[i] = 0;
    }
  
//   //init the table of nearest states
//   tableNearState = new PositionXYT**;
//   (*tableNearState) = new PositionXYT* [this->traj_->getSize()];
//   for(int i=0; i < this->traj_->getSize(); i++)
//     {
//       (*tableNearState)[i] = NULL;
//     }
  
//   // init the tables of internal and external forces
//   switch(traj_->type())
//     {
//     case PM_TRAJ : 
//       tableFInt = (StateTime**)(new PMStateT*);
//       *tableFInt = NULL;
//       tableFExt = (StateTime**)(new PMStateT*);
//       *tableFExt = NULL;
//       break;
//     case CL_TRAJ : 
//       tableFInt = (StateTime**)(new CLStateT*);
//       *tableFInt = NULL;
//       tableFExt = (StateTime**)(new CLStateT*);
//       *tableFExt = NULL;
//       break;
//     case DD_TRAJ : 
//       tableFInt = (StateTime**)(new DDStateT*);
//       *tableFInt = NULL;
//       tableFExt = (StateTime**)(new DDStateT*);
//       *tableFExt = NULL;
//       break;
//     case WC_TRAJ : 
//       tableFInt = (StateTime**)(new WCStateT*);
//       *tableFInt = NULL;
//       tableFExt = (StateTime**)(new WCStateT*);
//       *tableFExt = NULL;
//       break;
//     default : 
//       FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::initTabForces","Unknown trajectory type!");
//       break;
//     }
}

// init the viewer after loading of a path
void MPC_TeddyIfc::initViewer_PathLdd()
{
  if(typeMotion == 0) // trajectory tracking
    {
      // get back the path of the trajectory
      path_ = traj_->getPath();
      
      // init the viewer
      PositionXY center;
      double radius; 
      if(env_ != NULL)
	{
	  center = env_->getCenter();
	  radius = env_->getRadius();
	}
      else if(path_ != NULL)
	{
	  center = path_->center();
	  radius = path_->radius();
	}
      else
	{
	  cerr << "MPC_TeddyIfc::initViewer_PathLdd : path and environment are not defined..." << endl;
	  exit(1);
	}
      double depth = 5*radius;
      double dt = 3*(traj_->at(traj_->getSize()-1).t());
      double upx,upy,upz;
      
      Vec v1(3);
      v1.setVal(0,0);
      v1.setVal(1,-dt);
      v1.setVal(2,-depth);
      double v1Norm = v1.norm();
      Vec v2(3);
      v2.setVal(0,0);
      v2.setVal(1,-dt/v1Norm);
      v2.setVal(2,-depth/v1Norm);
      upx = 0;
      upy = -v2[2];
      upz = v2[1];
      
      this->glcanvas->init(center.x(), center.y() + dt, depth, center.x(), center.y(), 0, upx, upy, upz);
            
      // draw the scene
      this->drawScene();
    }
}

// load the system parameters for the car-like system
void MPC_TeddyIfc::loadCarLikeParam()
{
  maxSpd      = 64.0f;
  maxAcc      = 7.0f;
  maxWVel     = 8.0*M_PI;
  maxWAcc     = 1.5*M_PI;

  maxSteer    = 0.25*M_PI;
  maxSteerS   = 0.10*M_PI;
  maxSpdCoef  = 0.6f;
  maxAccCoef  = 0.6f;
  robotRadius = 10.0f;
  robotLength = 20.0f;
  
  genTrajType  = CL_TRAJ;
  trajType     = CL_TRAJ;
  
  reliabilityTime    = 12.0;
  epsilonRT          = 0.05;
  predHorizon        = 3.0;
  
  K_INT           = 1.0;
  K_EXT_SPT       = 0.01;
  K_EXT_TMP       = 0.05;
  K_SPT_TMP       = 3.0;
  K_EXT           = 5.0;
  K_STEXT         = 1.5;

  alphaMin        = 0.1;     // minimal spatial coefficient (for the metric)
  alphaMax        = 1.0;     // maximal spatial coefficient (for the metric)
  betaMin         = 0.0;     // minimal temporal coefficient (for the metric)
  betaMax         = 30.0;    // maximal temporal coefficient (for the metric)
  
  maxSpDist       = 80.0;
  maxTmpDist      = 5.0;
  maxStEffectDist = 80.0;  
  
  useTemporalConsistency = true;
  useConnectivityChecker = false; //true;
  roadmapView = true;
  
  seuilChgt = 0.7;
  maxAngleChgt = 2*M_PI*0.05;
  maxNormChgt = 0.03;
  
  typeMotion = 1;
  
  initXRb = 50;
  initYRb = 50;
  initThRb = 0.0;
  
//   dfltXMap          = 700;
//   dfltYMap          = 700;
  drawMapBoundaries = true;
    
  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;
  
  drawTrajGen = false;

  parallelize = false;
  nbThread = 5;
}

// load the system parameters for the differential drive system
void MPC_TeddyIfc::loadDiffDriveParam()
{
}

// load the system parameters for the intersection scenario
void MPC_TeddyIfc::loadItscParam()
{
  maxSpd      = 64.0f;
  maxAcc      = 7.0f;
  maxWVel     = 8.0*M_PI;
  maxWAcc     = 1.5*M_PI;

  maxSteer    = 0.25*M_PI;
  maxSteerS   = 0.10*M_PI;
  maxSpdCoef  = 0.6f;
  maxAccCoef  = 0.6f;
  robotRadius = 10.0f;
  robotLength = 20.0f;
  
  genTrajType  = CL_TRAJ;
  trajType     = CL_TRAJ;
  
  reliabilityTime    = 12.0;
  epsilonRT          = 0.05;
  predHorizon        = 3.0;
  
  K_INT           = 1.0;
  K_EXT_SPT       = 0.01;
  K_EXT_TMP       = 0.05;
  K_SPT_TMP       = 3.0;
  K_EXT           = 5.0;
  K_STEXT         = 1.5;

  alphaMin        = 0.1;     // minimal spatial coefficient (for the metric)
  alphaMax        = 1.0;     // maximal spatial coefficient (for the metric)
  betaMin         = 0.1;     // minimal temporal coefficient (for the metric)
  betaMax         = 30.0;    // maximal temporal coefficient (for the metric)
  
  maxSpDist       = 50.0;
  maxTmpDist      = 7.0;
  maxStEffectDist = 40.0;  
  
  useTemporalConsistency = true;
  useConnectivityChecker = false; //true;
  roadmapView = false;
  
  seuilChgt = 0.7;
  maxAngleChgt = 0.0; //2*M_PI*0.05;
  maxNormChgt = 0.03;
  
  typeMotion = 0;
  
//   initXRb = 50;
//   initYRb = 50;
//   initThRb = 0.0;
  
//   dfltXMap          = 1370;
//   dfltYMap          = 900;
  drawMapBoundaries = true;
    
  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;
  
  drawTrajGen = false;

  parallelize = false;
  nbThread = 5;
}

// load the system parameters for the motorway scenario
void MPC_TeddyIfc::loadMtwyParam()
{
  maxSpd      = 64.0f;
  maxAcc      = 7.0f;
  maxWVel     = 8.0*M_PI;
  maxWAcc     = 1.5*M_PI;

  maxSteer    = 0.25*M_PI;
  maxSteerS   = 0.10*M_PI;
  maxSpdCoef  = 0.6f;
  maxAccCoef  = 0.6f;
  robotRadius = 10.0f;
  robotLength = 20.0f;
  
  genTrajType  = CL_TRAJ;
  trajType     = CL_TRAJ;
  
  reliabilityTime    = 12.0;
  epsilonRT          = 0.05;
  predHorizon        = 3.0;
  
  K_INT           = 1.0;
  K_EXT_SPT       = 0.01;
  K_EXT_TMP       = 0.05;
  K_SPT_TMP       = 3.0;
  K_EXT           = 5.0;
  K_STEXT         = 1.5;

  alphaMin        = 0.1;     // minimal spatial coefficient (for the metric)
  alphaMax        = 1.0;     // maximal spatial coefficient (for the metric)
  betaMin         = 0.1;     // minimal temporal coefficient (for the metric)
  betaMax         = 30.0;    // maximal temporal coefficient (for the metric)
  
  maxSpDist       = 50.0;
  maxTmpDist      = 7.0;
  maxStEffectDist = 30.0;  
  
  useTemporalConsistency = true;
  useConnectivityChecker = false; //true;
  roadmapView = false;
  
  seuilChgt = 0.7;
  maxAngleChgt = 0.0; //2*M_PI*0.05;
  maxNormChgt = 0.03;
  
  typeMotion = 0;
  
//   initXRb = 50;
//   initYRb = 50;
//   initThRb = 0.0;
  
//   dfltXMap          = 1370;
//   dfltYMap          = 900;
  drawMapBoundaries = true;
    
  useVelCst = true;
  useAccCst = true;
  useAVelCst = true;
  useAAccCst = false;
  
  drawTrajGen = false;

  parallelize = false;
  nbThread = 5;
}

// compute a new random target position
PositionXYTh MPC_TeddyIfc::computeRandTargetPos(StateTime& initSt)
{
  goalPosX = -1;
  goalPosY = -1;

  // get back the dimensions of the environment
  double minX,maxX,minY,maxY;
  if(this->env_ != NULL)
    {
      minX = env_->minX();
      maxX = env_->maxX();
      minY = env_->minY();
      maxY = env_->maxY();
    }
  else
    {
      minX = 0.0;
      maxX = dfltXMap;
      minY = 0.0;
      maxY = dfltYMap;
    }
  
  while(goalPosX == -1 && goalPosY == -1)
    {
      // choose a new goal position randomly
      goalPosX = getIntRandom((int)minX,(int)maxX);
      goalPosY = getIntRandom((int)minY,(int)maxY);

      // check if the goal position isn't obstructed in the environment
      if(env_ != NULL)
	{
	  PositionXY near;
	  double minDist;
	  bool obstruct = env_->staticObstruct(PositionXY(goalPosX,goalPosY),near,minDist);

	  if(obstruct || minDist < 50.0)
	    {
	      goalPosX = -1;
	      goalPosY = -1;
	    }
	}

      // check if the goal position is not too close to the initial position
      double dxIG = goalPosX - initSt.posXYT().x();
      double dyIG = goalPosY - initSt.posXYT().y();
      double sqDistInitGoal = dxIG*dxIG + dyIG*dyIG;
      if(sqDistInitGoal < 100*100)
	{
	  goalPosX = -1;
	  goalPosY = -1;
	}      
    }

  // compute a random heading
  double goalTh = getDoubleRandom(-M_PI,M_PI);

  return PositionXYTh(goalPosX,goalPosY,goalTh);
}

// compute a new straight line trajectory from initCfg
int MPC_TeddyIfc::computeNewRandTrajFrom(StateTime& initSt)
{
  cout << "computeNewRandTrajFrom -> initSt : "; initSt.display(); cout << endl;

  if(this->roadmap_ == NULL)
    {
      cout << "MPC_TeddyIfc::computeNewRandTrajFrom : Not finished yet!" << endl; 
      FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::computeNewRandTrajFrom","Not finished yet!");
      return -1;
    }

  // get a new random goal position
  PositionXYTh posXYTh;
  if(fixRbGoalPos) posXYTh = PositionXYTh(goalXRb,goalYRb,goalThRb);
  else posXYTh = this->computeRandTargetPos(initSt);
  goalPosX = posXYTh.x();
  goalPosY = posXYTh.y();

  cout << "goalPosX : " << goalPosX << "  goalPosY : " << goalPosY << endl;
  
  // reinitialize the roadmap
  this->roadmap_->reinit();

  if(trajType == PM_TRAJ)
    { 
      // compute a path on the roadmap
      PMConfig cfg1(initSt.posXYT().x(), initSt.posXYT().y());
      PMConfig cfg2(goalPosX, goalPosY);

      RoadMapPP rmpp(sys_,env_,roadmap_);
      rmpp.computePath(cfg1,cfg2);
      this->processType_ = PATH_PLNG_PROC;
//       cout << "path : " << endl;
//       rmpp.getPath()->display();
//       cout << "fin path" << endl;
      this->path_ = new PMPath(*(rmpp.getPath()));

      // compute the trajectory until this position
      int res = this->computeTrajectory(NULL, 0, NULL);
      this->processType_ = TRAJ_PLNG_PROC;
            
      return res;
    }
  else if(trajType == CL_TRAJ)
    {
//       CLState s1(((CLState*)(&initSt))->x(),((CLState*)(&initSt))->y(),((CLState*)(&initSt))->th(),0,0);
      CLState s1(initSt.posXYT().x(),initSt.posXYT().y(),((CLStateT*)(&initSt))->th(),0,0);
      CLState s2(posXYTh.x(),posXYTh.y(),posXYTh.th(),0,0);

      CLRoadMapTP rmpp((CarLikeSys*)sys_,env_,roadmap_);
      rmpp.setTimeStep(trajTimeStep);
      rmpp.setMSC(maxSpdCoef);
      rmpp.setMAC(maxAccCoef);
      int res = rmpp.computeTraj(s1,s2);
      this->processType_ = TRAJ_PLNG_PROC;

      if(this->traj_ != NULL) delete this->traj_;
      this->traj_ = new CLSTraj(*(rmpp.getTraj()));
      this->traj_->shift(time);
      curStep = 0;      

      this->drawScene();
      return 0;
    }
  else if(trajType == WC_TRAJ)
    {
      WCState s1(initSt.posXYT().x(),initSt.posXYT().y(),((WCStateT*)(&initSt))->th(),0,0);
      WCState s2(posXYTh.x(),posXYTh.y(),posXYTh.th(),0,0);

//       cout << "s1 : "; s1.display(); cout << endl;
//       cout << "s2 : "; s2.display(); cout << endl;

      WCRoadMapTP rmpp((WheelChairSys*)sys_,env_,roadmap_);
      rmpp.setTimeStep(trajTimeStep);
      rmpp.setMSC(maxSpdCoef);
      rmpp.setMAC(maxAccCoef);
      int res = rmpp.computeTraj(s1,s2);
      this->processType_ = TRAJ_PLNG_PROC;

      if(this->traj_ != NULL) delete this->traj_;
      this->traj_ = new WCSTraj(*(rmpp.getTraj()));
      this->traj_->shift(time);
      curStep = 0;
      
//       this->traj_->display();

      this->drawScene();
      return 0;
    }
}

// compute a new Trajectory for the Car-Like Sys from initCfg
int MPC_TeddyIfc::computeNewCarLikeTraj(StateTime& initCfg)
{
  // load the car-like deformation and trajectory generation predefined parameters
  this->loadCarLikeParam();

  // init the robot configuration
  this->initRobot();

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  // init the random mode
  initRandomDeform();

  /////////////////////////////////////////////////////////////////////////////////
  cout << "MPC_TeddyIfc::computeNewCarLikeTraj : Not finished yet!" << endl; 
  FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::computeNewCarLikeTraj","Not finished yet!");
  return -1;
  /////////////////////////////////////////////////////////////////////////////////
//   AccASTrajGen aastg(sys_,
// 		    env_,
// 		    new CLStateT(initCfg),
// 		    new CLStateT(200.0,200.0,0,0,0,0));
//   int resTP = aastg.computeTraj();
  /////////////////////////////////////////////////////////////////////////////////

//   if(resTP != TRAJ_CPTD)
//     {
//       FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
//       return -2;
//     }
//   else
//     {
//       if(this->traj_ != NULL) delete traj_;
//       this->traj_ = new CLSTraj(*((CLSTraj*)aastg.getTraj()));
//       this->traj_->shift(time);

//       // set the type of the trajectory
//       trajType = CL_TRAJ;

//       // compute the list of initial feddback controls
//       std::list <Vec_DP*>* listFBCtrl = new std::list <Vec_DP*>();
//       for(int i=0; i<aastg.vecCtrl_->size(); i++)
// 	{
//  	  Vec_DP* nvVecParam = new Vec_DP(7); 
// 	  (*nvVecParam)[0] = aastg.vecCtrl_->at(i).getAS();
// 	  (*nvVecParam)[1] = 0.0;
// 	  (*nvVecParam)[2] = 0.0;
// 	  (*nvVecParam)[3] = aastg.vecCtrl_->at(i).getAcc();
// 	  (*nvVecParam)[4] = 0.0;
// 	  (*nvVecParam)[5] = 0.0;
// 	  (*nvVecParam)[6] = aastg.vecCtrl_->at(i).getTime();
// 	  listFBCtrl->push_back(nvVecParam);
// 	}

//       slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
//       curStep = 0;

//       // init the tables of internal forces, external forces and nearest points
//       this->initTabForces();

//       // init the trajectory deformer
//       if(deformer_ != NULL) delete deformer_;
//       deformer_ = new STNHDeformation(this->traj_, this->env_, this->sys_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
//       this->initDeform();

//       // initialize the list of initial feedback controls
//       //((STNHDeformation*)deformer_)->initListCtrl(listFBCtrl);

//       // init the viewer once the path loaded
//       this->initViewer_PathLdd();
//     }  

  return 1;
}

// compute a trajectory for the intersection scenario
int MPC_TeddyIfc::computeIntersectionTraj()
{
  // load the car-like deformation and trajectory generation 
  // predefined parameters for the intersection scenario
  this->loadItscParam();

  // init the robot configuration
  this->initRobot();

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  // init the random mode
  initRandomDeform();

  /////////////////////////////////////////////////////////////////////////////////
  cout << "MPC_TeddyIfc::computeIntersectionTraj : Not finished yet!" << endl; 
  FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::computeNewCarLikeTraj","Not finished yet!");
  return -1;
  /////////////////////////////////////////////////////////////////////////////////

//   AccASTrajGen aastg(sys_,
// 		     env_,
// 		     new StateS6D(28.0,445.0,0.00602402,10.0,0,0),
// 		     new StateS6D(0,0,0,0,0,0));
//   int resTP = aastg.computeTraj2();

//   if(resTP != TRAJ_CPTD)
//     {
//       FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
//       return -2;
//     }
//   else
//     {
//       if(this->traj_ != NULL) delete traj_;
//       this->traj_ = new Trajectory6D(*((Trajectory6D*)aastg.getTraj()));
//       this->traj_->shift(time);

//       // set the type of the trajectory
//       trajType = CL_TRAJ;

//       // compute the list of initial feddback controls
//       std::list <Vec_DP*>* listFBCtrl = new std::list <Vec_DP*>();
//       for(int i=0; i<aastg.vecCtrl_->size(); i++)
// 	{
//  	  Vec_DP* nvVecParam = new Vec_DP(7); 
// 	  (*nvVecParam)[0] = aastg.vecCtrl_->at(i).getAS();
// 	  (*nvVecParam)[1] = 0.0;
// 	  (*nvVecParam)[2] = 0.0;
// 	  (*nvVecParam)[3] = aastg.vecCtrl_->at(i).getAcc();
// 	  (*nvVecParam)[4] = 0.0;
// 	  (*nvVecParam)[5] = 0.0;
// 	  (*nvVecParam)[6] = aastg.vecCtrl_->at(i).getTime();
// 	  listFBCtrl->push_back(nvVecParam);
// 	}

//       slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
//       curStep = 0;

//       // init the tables of internal forces, external forces and nearest points
//       this->initTabForces();

//       // init the trajectory deformer
//       if(deformer_ != NULL) delete deformer_;
//       deformer_ = new STNHDeformation(this->traj_, this->env_, this->sys_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
//       this->initDeform();

//       // dont use the presious parameters for the next trajectory generation
//       ((STNHDeformation*)deformer_)->setUsePrevParam(false);
//       ((STNHDeformation*)deformer_)->setUseResampling(false);

//       // initialize the list of initial feedback controls
//       //((STNHDeformation*)deformer_)->initListCtrl(listFBCtrl);

//       // init the viewer once the path loaded
//       this->initViewer_PathLdd();
//     }  

  return 1;
}

// compute a trajectory for the motorway scenario
int MPC_TeddyIfc::computeMotorwayTraj()
{
  // load the car-like deformation and trajectory generation 
  // predefined parameters for the intersection scenario
  this->loadMtwyParam();

  // init the robot configuration
  this->initRobot();

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  // init the random mode
  initRandomDeform();

  /////////////////////////////////////////////////////////////////////////////////
  cout << "MPC_TeddyIfc::computeMotorwayTraj : Not finished yet!" << endl; 
  FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::computeNewCarLikeTraj","Not finished yet!");
  return -1;
  /////////////////////////////////////////////////////////////////////////////////

//   AccASTrajGen aastg(sys_,
// 		     env_,
// 		     new StateS6D(96.0,161.0,0.47783105,10.0,0,0),
// 		     new StateS6D(0,0,0,0,0,0));
//   int resTP = aastg.computeTraj3();

//   if(resTP != TRAJ_CPTD)
//     {
//       FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
//       return -2;
//     }
//   else
//     {
//       if(this->traj_ != NULL) delete traj_;
//       this->traj_ = new Trajectory6D(*((Trajectory6D*)aastg.getTraj()));
//       this->traj_->shift(time);

//       // set the type of the trajectory
//       trajType = CL_TRAJ;

//       // compute the list of initial feddback controls
//       std::list <Vec_DP*>* listFBCtrl = new std::list <Vec_DP*>();
//       for(int i=0; i<aastg.vecCtrl_->size(); i++)
// 	{
//  	  Vec_DP* nvVecParam = new Vec_DP(7); 
// 	  (*nvVecParam)[0] = aastg.vecCtrl_->at(i).getAS();
// 	  (*nvVecParam)[1] = 0.0;
// 	  (*nvVecParam)[2] = 0.0;
// 	  (*nvVecParam)[3] = aastg.vecCtrl_->at(i).getAcc();
// 	  (*nvVecParam)[4] = 0.0;
// 	  (*nvVecParam)[5] = 0.0;
// 	  (*nvVecParam)[6] = aastg.vecCtrl_->at(i).getTime();
// 	  listFBCtrl->push_back(nvVecParam);
// 	}

//       slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
//       curStep = 0;

//       // init the tables of internal forces, external forces and nearest points
//       this->initTabForces();

//       // init the trajectory deformer
//       if(deformer_ != NULL) delete deformer_;
//       deformer_ = new STNHDeformation(this->traj_, this->env_, this->sys_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
//       this->initDeform();

//       // dont use the presious parameters for the next trajectory generation
//       ((STNHDeformation*)deformer_)->setUsePrevParam(false);
//       ((STNHDeformation*)deformer_)->setUseResampling(false);

//       // initialize the list of initial feedback controls
//       //((STNHDeformation*)deformer_)->initListCtrl(listFBCtrl);

//       // init the viewer once the path loaded
//       this->initViewer_PathLdd();
//     }  

  return 1;
}

// compute a new Trajectory for the DiffDrive Sys from initCfg
int MPC_TeddyIfc::computeNewDiffDriveTraj(StateTime& initCfg)
{
  // load the car-like deformation and trajectory generation predefined parameters
  this->loadDiffDriveParam();

  // init the robot configuration
  ((DiffDriveSys*)sys_)->setWVMax(maxWVel);
  ((DiffDriveSys*)sys_)->setWAMax(maxWAcc);
  ((DiffDriveSys*)sys_)->setLength(robotRadius);

  // reinit the time already past
  //alreadyElapsedTime = 0;
  FPS = 0;
  totalNbTimeout += nbTimeout;
  nbTimeout = 0;

  // init the random mode
  initRandomDeform();

  /////////////////////////////////////////////////////////////////////////////////
  cout << "MPC_TeddyIfc::computeNewDiffDriveTraj : Not finished yet!" << endl; 
  FXMessageBox::error(this,MBOX_OK,"MPC_TeddyIfc::computeNewCarLikeTraj","Not finished yet!");
  return -1;
  /////////////////////////////////////////////////////////////////////////////////

//   WAccTrajGen wacctg(sys_,
// 		     env_,
// // 		     new DDState(initCfg),
// 		     new DDState(50.0,50.0,0,10.0,10.0,0),
// 		     new DDState(200.0,200.0,0,0.0,0.0,0));
//   int resTP = wacctg.computeTraj();

//   if(resTP != TRAJ_CPTD)
//     {
//       FXMessageBox::error(this,MBOX_OK,"Error Computing the Trajectory","Error while computing the trajectory");
//       return -2;
//     }
//   else
//     {
//       if(this->traj_ != NULL) delete traj_;
//       this->traj_ = new DDTrajectory(*((DDTrajectory*)wacctg.getTraj()));
//       this->traj_->shift(time);

//       // set the type of the trajectory
//       trajType = DD_TRAJ;

//       // compute the list of initial feddback controls
//       std::list <Vec_DP*>* listFBCtrl = new std::list <Vec_DP*>();
//       for(int i=0; i<wacctg.vecCtrl_->size(); i++)
// 	{
//  	  Vec_DP* nvVecParam = new Vec_DP(7); 
// 	  (*nvVecParam)[0] = wacctg.vecCtrl_->at(i).getWRAcc();
// 	  (*nvVecParam)[1] = 0.0;
// 	  (*nvVecParam)[2] = 0.0;
// 	  (*nvVecParam)[3] = wacctg.vecCtrl_->at(i).getWLAcc();
// 	  (*nvVecParam)[4] = 0.0;
// 	  (*nvVecParam)[5] = 0.0;
// 	  (*nvVecParam)[6] = wacctg.vecCtrl_->at(i).getTime();
// 	  listFBCtrl->push_back(nvVecParam);
// 	}

//       slider->setRange(0.0,(*(this->traj_))[this->traj_->getSize()-1]->t());
//       curStep = 0;

//       typeState = new int*;
//       (*typeState) = new int[this->traj_->getSize()];
//       for(int i=0; i < this->traj_->getSize(); i++)
// 	{
// 	  (*typeState)[i] = 0;
// 	}

//       //init the table of nearest states
//       tableNearState = new StateR**;
//       (*tableNearState) = new StateR* [this->traj_->getSize()];
//       for(int i=0; i < this->traj_->getSize(); i++)
// 	{
// 	  (*tableNearState)[i] = NULL;
// 	}

//       // init the tables of internal and external forces
//       tableFInt = new StateR*;
//       *tableFInt = NULL;
//       tableFExt = new StateR*;
//       *tableFExt = NULL;

//       // init the trajectory deformer
//       if(deformer_ != NULL) delete deformer_;
      
//       deformer_ = new DDDeformation(this->traj_, this->env_, this->sys_, typeState, (StateR***)tableNearState, (StateR**)tableFInt, (StateR**)tableFExt);
//       ((DDDeformation*)deformer_)->setParallelExec(parallelize);
//       ((DDDeformation*)deformer_)->setNbThread(nbThread);

//       if(optModeDeform == 0) deformer_->setModeDeform(SPATIOTEMP_DEFORM);
//       else if(optModeDeform == 1) deformer_->setModeDeform(SPATONLY_DEFORM);
//       else if(optModeDeform == 2) deformer_->setModeDeform(TEMPONLY_DEFORM);
//       deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &K_SPT_TMP);
//       deformer_->setForcesCoef(&K_INT, &K_EXT, &K_STEXT, &alphaMin, &alphaMax, &betaMin, &betaMax);
// //       deformer_->setMaxEffectDist(&maxEffectDist);
// //       deformer_->setMaxStEffectDist(&maxStEffectDist);
//       deformer_->setMaxEffectDist(&maxSpDist,&maxTmpDist,&maxStEffectDist);
//       deformer_->setReliability(&predHorizon, &reliabilityTime, &epsilonRT);
//       deformer_->setTemporalConst(&useTemporalConsistency);
//       deformer_->setTimeStep(trajTimeStep);
//       //deformer_->setPosStep(trajTimeStep * maxSpd);
//       deformer_->setPosStep(trajPosStep);
      
//       // init the trajectory generation parameters
//       DDTrajGen::setMaxNbStep(maxNbStep);               // set the maximal number of steps
//       DDTrajGen::setMaxResidual(maxResidual);           // set the maximal residual allowed
//       DDTrajGen::setDBG(debugLevel);                    // set the debugging level

//       // display the list of initial feedback controls
//       std::list <Vec_DP*>::iterator itLC;

//       // initialize the list of initial feedback controls
//       //((STNHDeformation*)deformer_)->initListCtrl(listFBCtrl);

//       if(typeMotion == 0) // trajectory tracking
// 	{
// 	  // get back the path of the trajectory
// 	  path_ = traj_->getPath();

// 	  // init the viewer
// 	  ConfigR* center = path_->center();
// 	  double radius = path_->radius();
// 	  double depth = 5*radius;
// 	  double dt = 3*(*traj_)[traj_->getSize()-1]->t();
// 	  double upx,upy,upz;

// 	  Vec v1(3);
// 	  v1.setVal(0,0);
// 	  v1.setVal(1,-dt);
// 	  v1.setVal(2,-depth);
// 	  double v1Norm = v1.norm();
// 	  Vec v2(3);
// 	  v2.setVal(0,0);
// 	  v2.setVal(1,-dt/v1Norm);
// 	  v2.setVal(2,-depth/v1Norm);
// 	  upx = 0;
// 	  upy = -v2[2];
// 	  upz = v2[1];

// 	  this->glcanvas->init(center->x(), center->y() + dt, depth, center->x(), center->y(), 0, upx, upy, upz);
	  
// 	  delete center;

// 	  // draw the scene
// 	  this->drawScene();
// 	}
//     }  

  return 1;
}

void MPC_TeddyIfc::loadMapInViewer()
{
}

void MPC_TeddyIfc::drawMap()
{
  if(this->env_ == NULL)
    {
      // nothing to draw
    }
//   else if(this->env_->getMap() != NULL)
//     {
//       // draw the map of the environment
      
//     }
  else
    {
      if(drawMapBoundaries)
	{
	  // draw the default bouding box of the map
	  glBegin(GL_LINE_LOOP);
	  glVertex3f(0.0, 0.0, time);
	  glVertex3f(dfltXMap, 0.0, time);
	  glVertex3f(dfltXMap, dfltYMap, time);
	  glVertex3f(0.0, dfltYMap, time);
	  glEnd();
	}
    }
}

// verify and correct if needed the configuration of the obstacle
void MPC_TeddyIfc::verifyObsCfg(PMConfig* cfgObs,int i)
{
  cerr << "MPC_TeddyIfc::verifyObsCfg : Not implemented anymore..." << endl;
  exit(1);

//   // test if the new configuration of the obstacle is in or out the environment
//   double limEnvX;
//   double limEnvY;
// //   if(this->env_->getMap() == NULL)
// //     {
//       limEnvX = dfltXMap;
//       limEnvY = dfltYMap;
// //     }
// //   else
// //     {
// //       limEnvX = this->env_->getMap()->getWidth();
// //       limEnvY = this->env_->getMap()->getHeight();
// //     }
//   bool out = false;
//   if(cfgObs->x() < 0)
//     {
//       cfgObs->x() = (0);
//       out = true;
//     }
//   else if(cfgObs->x() > limEnvX)
//     {
//       cfgObs->x() = (limEnvX);
//       out = true;
//     }
//   if(cfgObs->y() < 0)
//     {
//       cfgObs->y() = (0);
//       out = true;
//     }
//   else if(cfgObs->y() > limEnvY)
//     {
//       cfgObs->y() = (limEnvY);
//       out = true;
//     }
//   if(out)
//     {
//       this->env_->getObs(i)->remLastStateStack();
//       Config* lastCfg = this->env_->getObs(i)->lastCfg();

//       if(lastCfg != NULL)
// 	{
// 	  double dx = cfgObs->x() - lastCfg->posXY().x();
// 	  double dy = cfgObs->y() - lastCfg->posXY().y();
// 	  double dt = time - this->env_->getObs(i)->lastTime();
// 	  double dsx = dx / dt;
// 	  double dsy = dy / dt;

// 	  double normSpd = sqrt(dsx*dsx + dsy*dsy);
// 	  if(normSpd == 0.0)
// 	    cout << "Warning : obstacle spd is null :-(" << endl;
// 	  else if(normSpd < 10.0)
// 	    {
// 	      dsx = dsx / normSpd * 10.0;
// 	      dsy = dsy / normSpd * 10.0;
// 	    }
	  
// 	  PMStateT* newST = new PMStateT(cfgObs->posXY().x(),cfgObs->posXY().y(),time,-dsx,-dsy);
// 	  this->env_->getObs(i)->appendStateStack(newST);
// 	  delete lastCfg;
// 	}
//     }
}

// draw the environment components
void MPC_TeddyIfc::drawEnv()
{
//   // draw the circular dynamic obstacles
//   this->drawObstacles();

//   // draw the static obstacles
//   this->drawPolygonObs();

  // draw the map
  this->drawMap();
  
  // draw the generic obstacles
  if(this->env_ != NULL) 
    {
      //this->env_->display();
      this->env_->toGL(dispFuture);
    }
}

// void MPC_TeddyIfc::drawObstacles()
// {
//   // verify that the environment is not null
//   if(this->env_ != NULL)
//     {
//       for(int i=0; i < this->env_->getNbObs(); i++)
// 	{
// 	  const GLfloat redMaterial[]={1.,0.,0.,1.};
// 	  const GLfloat greenMaterial[]={0.,1.,0.,1.};
// 	  const GLfloat greenTsMaterial[]={0.,1.,0.,0.2};
	  
// 	  if(dispFuture)
// 	    {
// 	      // display the entiere future model
// 	      if(this->traj_ != NULL)
// 		{
// 		  int nbSide = 10;

// 		  double radObs = ((ObsDisk2D*)this->env_->getObs(i))->getRadius();
		  
// 		  for(int k=0; k<this->env_->getObs(i)->stackSize(); k++)
// 		    {
// 		      PMConfig* cfgObs = this->env_->getObs(i)->getCfg(time);
		      
// 		      if(cfgObs != NULL)
// 			{
// 			  this->verifyObsCfg(cfgObs,i);
			  
// 			  delete cfgObs;
// 			}

// 		      double initTime;
// 		      double lastTime;
// 		      PMConfig initCfgObs;
// 		      PMConfig goalCfgObs;

// 		      if(k != this->env_->getObs(i)->stackSize()-1)
// 			{
// 			  glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
// 			  glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);

// 			  PMStateT* s1 = this->env_->getObs(i)->getStateStack(k);
// 			  PMStateT* s2 = this->env_->getObs(i)->getStateStack(k+1);
// 			  initTime = s1->t();
// 			  lastTime = s2->t();
// 			  initCfgObs = PMConfig(s1->x(),s1->y());
// 			  goalCfgObs = PMConfig(s2->x(),s2->y());
// 			}
// 		      else
// 			{
// 			  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
// 			  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);

// 			  PMStateT* s1 = this->env_->getObs(i)->getStateStack(k);
// 			  initTime = s1->t();
// 			  lastTime = this->env_->getObs(i)->initTime() + this->env_->getObs(i)->lifeTime();
// 			  PMConfig* cfg2 = this->env_->getObs(i)->getFutureCfg(lastTime);
// 			  initCfgObs = PMConfig(s1->x(),s1->y());
// 			  goalCfgObs = PMConfig(*cfg2);
// 			  delete cfg2;
// 			}
		      
// 		     //  cout << "draw : " << i << " " << k << endl;

// 		      glPushMatrix();
// 		      glTranslated(initCfgObs.x(),initCfgObs.y(),initTime);
// 		      glNormal3f(0,0,-1);
// 		      gluDisk(gluNewQuadric(), 0, radObs, nbSide, 1);
// 		      glPopMatrix();
		      
// 		      glPushMatrix();
// 		      glTranslated(goalCfgObs.x(),goalCfgObs.y(),lastTime);
// 		      glNormal3f(0,0,1);
// 		      gluDisk(gluNewQuadric(), 0, radObs, nbSide, 1);
// 		      glPopMatrix();
		      
// 		      for(int j=0; j<nbSide; j++)
// 			{
// 			  glBegin(GL_QUADS);
// 			  FXVec3d v1(initCfgObs.x() + cos(j*2*M_PI/((double)nbSide)) * radObs,
// 				     initCfgObs.y() + sin(j*2*M_PI/((double)nbSide)) * radObs,
// 				     initTime);
			  
// 			  FXVec3d v2(initCfgObs.x() + cos(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
// 				     initCfgObs.y() + sin(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
// 				     initTime);
			  
// 			  FXVec3d v3(goalCfgObs.x() + cos(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
// 				     goalCfgObs.y() + sin(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
// 				     lastTime);
			  
// 			  FXVec3d v4(goalCfgObs.x() + cos(j*2*M_PI/((double)nbSide)) * radObs,
// 				     goalCfgObs.y() + sin(j*2*M_PI/((double)nbSide)) * radObs,
// 				     lastTime);
			  
// 			  FXVec3d v1v2 = normalize(v2 - v1);
// 			  FXVec3d v1v4 = normalize(v4 - v1);
// 			  FXVec3d n = v1v2^v1v4;
		      
// 			  glNormal3f(n.x,n.y,n.z);
// 			  glVertex3f(v1.x, v1.y, v1.z);
// 			  glNormal3f(n.x,n.y,n.z);
// 			  glVertex3f(v2.x, v2.y, v2.z);
// 			  glNormal3f(n.x,n.y,n.z);
// 			  glVertex3f(v3.x, v3.y, v3.z);
// 			  glNormal3f(n.x,n.y,n.z);
// 			  glVertex3f(v4.x, v4.y, v4.z);
			  
// 			  glEnd();
// 			}
// 		    }
// 		}
// 	    }
// 	  else
// 	    {
// 	      // display obstacles only on the current time
// 	      obs_t typeObs = this->env_->getObs(i)->getType(time);
// 	      PMConfig* cfgObs = this->env_->getObs(i)->getCfg(time);
// 	      double radObs = ((ObsDisk2D*)this->env_->getObs(i))->getRadius();

// 	      if(cfgObs != NULL)
// 		{ 
// 		  if(drawMapBoundaries) this->verifyObsCfg(cfgObs,i);

// 		  if(typeObs == NOT_BORN)
// 		    {
// 		      glMaterialfv(GL_FRONT, GL_AMBIENT, greenTsMaterial);
// 		      glMaterialfv(GL_FRONT, GL_DIFFUSE, greenTsMaterial);

// 		      glPushMatrix();
// 		      glTranslated(cfgObs->x(),cfgObs->y(),time);
// 		      gluSphere(gluNewQuadric(),radObs,8,8);
// 		      glPopMatrix();
// 		    }
// 		  else if(typeObs == ALIVE)
// 		    {
// 		      glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
// 		      glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
		      
// 		      glPushMatrix();
// 		      glTranslated(cfgObs->x(),cfgObs->y(),time);
// 		      //gluSphere(gluNewQuadric(),radObs,8,8);
// 		      gluDisk(gluNewQuadric(), 0, radObs, 20, 1);
// 		      gluCylinder(gluNewQuadric(), radObs, radObs, robotRadius/100, 20, 20);
// 		      glTranslated(0,0,radObs/100);
// 		      gluDisk(gluNewQuadric(), 0, radObs, 20, 1);
// 		      glPopMatrix();
// 		    }
// 		  else // typeObs == DEAD
// 		    {
// 		    }

// 		  delete cfgObs;

// 		  //this->env_->getObs(i)->displayStack();
// 		}

// 	    }
// 	}
//     }
// }

// // draw polygonal obstacles
// void MPC_TeddyIfc::drawPolygonObs()
// {
//   // check if the environment is not null
//   if(this->env_ != NULL)
//     {
//       for(int i=0; i < this->env_->getNbStObs(); i++)
// 	{
// 	  this->env_->getStObs(i)->toGL(time);
// 	}
//     }
// }

void MPC_TeddyIfc::drawPath()
{
  if(this->path_ != NULL)
    {
      this->path_->toGL();
    }
  else
    {
      glBegin(GL_TRIANGLE_STRIP);
      glNormal3f(0.,0.,-1.);
      glVertex3f(-1, -1, -1);
      glVertex3f(-1, 1, -1);
      glVertex3f(1, -1, -1);
      glVertex3f(1, 1, -1);
      glEnd();
    } 
}

void MPC_TeddyIfc::drawTraj()
{
  if(this->traj_ != NULL)
    {
//       cout << "drawTrajGen..." << endl;
      if(drawTrajGen)
	{
	  if(trajType == DD_TRAJ && time != 0)
	    {
	      ((DDTrajDeform*)deformer_)->displayTraj();
	    }
	  if(trajType == CL_TRAJ && time != 0)
	    {
	      ((CLTrajDeform3*)deformer_)->displayTraj();
	    }

	}

//       cout << "traj_->toGL()..." << endl;
      switch(processType_)
	{
	case TRAJ_PLNG_PROC : 
	  this->traj_->toGL();
	  break;

	case TEDDY_PROC : 
	  this->traj_->toGL(time, reliabilityTime, (*typeState));

	  if(deformer_ != NULL)
	    {
	      // display the nearest position-time of the obstacles from each state-time of the trajectory
	      if(dispNearStates) deformer_->dispNearObs();
	      if(dispNearStStates) deformer_->dispNearStObs();
	      
	      // display the internal and external forces
	      if(dispAllForces)
		{
		  deformer_->dispFExt();
		  deformer_->dispFInt();
		  deformer_->dispStFExt();
		}
	    }
	  break;

	case TF_PROC : 
	  this->traj_->toGL(time, reliabilityTime, (*typeState));
	  break;
	}
    }
  else
    {
      this->drawPath();
    }
}

// draw the trajectory (trajectory following mode)
void MPC_TeddyIfc::drawTrajTF()
{
  const GLfloat yellowMaterial[]={1.,1.,0.,1.};
  const GLfloat redMaterial[]={1.,0.,0.,1.};

  if(this->traj_ != NULL)
    {
      this->traj_->toGL();

      // display the state the trajectory follower try to reach
      if(stateToReachTF_ != NULL)
	{
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	  
	  glPushMatrix();
	  glTranslated(stateToReachTF_->posXYT().x(),
		       stateToReachTF_->posXYT().y(),
		       stateToReachTF_->posXYT().t());
	  gluSphere(gluNewQuadric(), 1.5, 5, 5);
	  glPopMatrix();
	}
    }
  else
    {
      this->drawPath();
    }
}

// draw the car-like robotic system
void MPC_TeddyIfc::drawCLRobot(double heading)
{
  const GLfloat blueMaterial[]   = {0.,0.,1.,1.};
  const GLfloat paleblueMaterial[] = {0.1,0.1,0.7,1.};
  const GLfloat redMaterial[]    = {1.,0.,0.,1.};
  const GLfloat greenMaterial[]  = {0.,1.,0.,1.};

	  //draw the frame of the robot
	  glLineWidth(((GLfloat)(3.0)));
	  glBegin(GL_LINES);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	  glVertex3f(0.0,0.0,0.0);
	  glVertex3f(100.0,0.0,0.0);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
	  glVertex3f(0.0,0.0,0.0);
	  glVertex3f(0.0,-100.0,0.0);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
	  glVertex3f(0.0,0.0,0.0);
	  glVertex3f(0.0,0.0,10.0);
	  glEnd();
	  glLineWidth(((GLfloat)(1.0)));

  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);  

  glBegin(GL_QUADS);
  glVertex3f(1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading + 0.5*M_PI),
	     1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading + 0.5*M_PI),
	     0.0);
  glVertex3f(1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading - 0.5*M_PI),
	     1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading - 0.5*M_PI),
	     0.0);
  glVertex3f(-1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading - 0.5*M_PI),
	     -1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading - 0.5*M_PI),
	     0.0);
  glVertex3f(-1.0*robotRadius*cos(heading)+0.5*robotRadius*cos(heading + 0.5*M_PI),
	     -1.0*robotRadius*sin(heading)+0.5*robotRadius*sin(heading + 0.5*M_PI),
	     0.0);
  
  glEnd();
  
  // draw the wheels of the robot
  glMaterialfv(GL_FRONT, GL_AMBIENT, paleblueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, paleblueMaterial);
  
  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading + 0.5*M_PI),
	       1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading + 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading - 0.5*M_PI),
	       1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading - 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(-1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading - 0.5*M_PI),
	       -1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading - 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();
  
  glPushMatrix();
  glTranslated(-1.0*robotRadius*cos(heading)+0.6*robotRadius*cos(heading + 0.5*M_PI),
	       -1.0*robotRadius*sin(heading)+0.6*robotRadius*sin(heading + 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();
  
  // draw the orientation of the robot
  glLineWidth(((GLfloat)(3.0)));
  glBegin(GL_LINES);
  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
  glVertex3f(0.0,0.0,0.0);
  glVertex3f(50.0*cos(heading), 
	     50.0*sin(heading), 
	     0.0);
  glEnd();
  glLineWidth(((GLfloat)(1.0)));
}

// draw the wheelchair robotic system
void MPC_TeddyIfc::drawWCRobot()
{
  const GLfloat redMaterial[]    = {1.,0.,0.,1.};
  const GLfloat blueMaterial[]   = {0.,0.,1.,1.};
  const GLfloat paleblueMaterial[] = {0.1,0.1,0.7,1.};

  // compute the exact state of the differential drive system
  WCState wcState;
  if(time != 0) wcState = ((WCTrajDeform3*)deformer_)->getStateTraj(time).toState();

  // differential drive robot drawing
  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
	  
  glPushMatrix();
//   if(time == 0) glTranslated((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
//   else glTranslated(wcState.x(),wcState.y(),time);

  // draw the robot itself
  //double heading = (*(this->traj_))[(int)curStep]->th();
  double heading;
  if(time == 0) heading = ((WCStateT*)(&((*(this->traj_))[(int)curStep])))->th();
  else heading = wcState.th();

  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
  gluCylinder(gluNewQuadric(), robotRadius, robotRadius, robotRadius*0.01, 20, 20);
  glTranslated(0,0,robotRadius*0.01);
  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);

  // draw the wheels of the robot
  glMaterialfv(GL_FRONT, GL_AMBIENT, paleblueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, paleblueMaterial);

  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading + 0.5*M_PI),
	       1.0*robotRadius*sin(heading + 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();

  glPushMatrix();
  glTranslated(1.0*robotRadius*cos(heading - 0.5*M_PI),
	       1.0*robotRadius*sin(heading - 0.5*M_PI),
	       0.0);
  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
  glPopMatrix();

  // draw the orientation of the robot
  glLineWidth(((GLfloat)(3.0)));
  glBegin(GL_LINES);
  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
  glVertex3f(0.0,0.0,0.0);
  glVertex3f(50.0*cos(heading), 
	     50.0*sin(heading), 
	     0.0);
  glEnd();
  glLineWidth(((GLfloat)(1.0)));

  glPopMatrix();
}

void MPC_TeddyIfc::drawRobot()
{
  if(this->traj_ != NULL)
    {
      const GLfloat redMaterial[]    = {1.,0.,0.,1.};
      const GLfloat greenMaterial[]  = {0.,1.,0.,1.};
      const GLfloat blueMaterial[]   = {0.,0.,1.,1.};
      const GLfloat purpleMaterial[] = {1.,0.,1.,1.};
      const GLfloat paleblueMaterial[] = {0.1,0.1,0.7,1.};

      if(processType_ == TEDDY_PROC)
      	{
	  //draw the frame of the robot
	  glLineWidth(((GLfloat)(3.0)));
	  glBegin(GL_LINES);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	  glVertex3f((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
	  glVertex3f((*(this->traj_))[(int)curStep].posXYT().x() + 100,(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
	  glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
	  glVertex3f((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
	  glVertex3f((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y() + 100,(*(this->traj_))[(int)curStep].posXYT().t());
	  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
	  glVertex3f((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
	  glVertex3f((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t() + 10);
	  glEnd();
	  glLineWidth(((GLfloat)(1.0)));
	}
      
      // draw the robot
      if(trajType == PM_TRAJ)
	{
	  
	  // mass-point robot drawing
	  glMaterialfv(GL_FRONT, GL_AMBIENT, purpleMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, purpleMaterial);
	  
	  glPushMatrix();
	  glTranslated((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
	  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
	  gluCylinder(gluNewQuadric(), robotRadius, robotRadius, robotRadius/100, 20, 20);
	  glTranslated(0,0,robotRadius/100);
	  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
	  glPopMatrix();
	  
	}
      else if(trajType == DD_TRAJ)
	{
	  // compute the exact state of the differential drive system
	  DDState ddState;
	  if(time != 0) ddState = ((DDTrajDeform*)deformer_)->getStateTraj(time).toState();

	  // differential drive robot drawing
	  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);
	  
	  glPushMatrix();
	  if(time == 0) glTranslated((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
	  else glTranslated(ddState.x(),ddState.y(),time);
	  // draw the robot itself
	  //double heading = (*(this->traj_))[(int)curStep]->th();
	  double heading;
	  if(time == 0) heading = ((DDStateT*)(&((*(this->traj_))[(int)curStep])))->th();
	  else heading = ddState.th();

	  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);
	  gluCylinder(gluNewQuadric(), robotRadius, robotRadius, robotRadius/100, 20, 20);
	  glTranslated(0,0,robotRadius/100);
	  gluDisk(gluNewQuadric(), 0, robotRadius, 20, 1);

	  // draw the wheels of the robot
	  glMaterialfv(GL_FRONT, GL_AMBIENT, paleblueMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, paleblueMaterial);

	  glPushMatrix();
	  glTranslated(1.0*robotRadius*cos(heading + 0.5*M_PI),
		       1.0*robotRadius*sin(heading + 0.5*M_PI),
		       0.0);
	  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
	  glPopMatrix();

	  glPushMatrix();
	  glTranslated(1.0*robotRadius*cos(heading - 0.5*M_PI),
		       1.0*robotRadius*sin(heading - 0.5*M_PI),
		       0.0);
	  gluDisk(gluNewQuadric(), 0, 3.0, 20, 1);
	  glPopMatrix();

	  // draw the orientation of the robot
	  glLineWidth(((GLfloat)(3.0)));
	  glBegin(GL_LINES);
	  glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	  glVertex3f(0.0,0.0,0.0);
	  glVertex3f(50.0*cos(heading), 
		     50.0*sin(heading), 
		     0.0);
	  glEnd();
	  glLineWidth(((GLfloat)(1.0)));

	  glPopMatrix();	  
	}
      else if(trajType == WC_TRAJ)
	{
	  if(processType_ == TEDDY_PROC)
	    {
	      // compute the exact state of the differential drive system
	      WCStateT wcState;
	      if(time != 0) wcState = ((WCTrajDeform3*)deformer_)->getStateTraj(time);
	      
	      // car-like robot drawing
	      glPushMatrix();
	      if(time == 0) glTranslated((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
	      else glTranslated(wcState.x(),wcState.y(),time);
	      // draw the robot itself
	      double heading;
	      if(time == 0) heading = ((WCStateT*)(&(((*(this->traj_))[(int)curStep]))))->th();
	      else heading = wcState.th();
	      this->drawWCRobot();
	      
	      glPopMatrix();
	    }
	  else
	    {
	      this->drawWCRobot();
	    }
	}

      else 
	{
	  if(processType_ == TEDDY_PROC)
	    {
	      // compute the exact state of the differential drive system
	      CLStateT clState;
	      if(time != 0) clState = ((CLTrajDeform3*)deformer_)->getStateTraj(time);
	      
// 	      cout << "clState.x() : " << clState.x() << "  clState.y() : " << clState.y() << "  clState.th() : " << clState.th() << "  time : " << time << endl;

	      // car-like robot drawing
	      glPushMatrix();
	      if(time == 0) glTranslated((*(this->traj_))[(int)curStep].posXYT().x(),(*(this->traj_))[(int)curStep].posXYT().y(),(*(this->traj_))[(int)curStep].posXYT().t());
	      else glTranslated(clState.x(),clState.y(),time);
	      // draw the robot itself
	      double heading;
	      if(time == 0) heading = ((CLStateT*)(&(((*(this->traj_))[(int)curStep]))))->th();
	      else heading = clState.th();
	      this->drawCLRobot(heading);
	      
	      glPopMatrix();
	    }
	  else
	    {
	      this->drawCLRobot(0);
	    }
	}
    }
}

// draw the robotic system (trajectory following mode)
void MPC_TeddyIfc::drawRobotTF()
{
  if(robotStateTF_ == NULL)
    {
      cout << "MPC_TeddyIfc::drawRobotTF : Warning. The robotic system state is not defined." << endl;
      return;
    }

  // car-like robot drawing
  glPushMatrix();
  glTranslated(robotStateTF_->posXYT().x(),robotStateTF_->posXYT().y(),robotStateTF_->posXYT().t());
  // draw the robot itself
  double heading = ((CLStateT*)robotStateTF_)->th();
  
  this->drawCLRobot(heading);
  
  glPopMatrix();
}

// draw the roadmap
void MPC_TeddyIfc::drawRoadMap()
{
  if(this->roadmap_ != NULL)
    {
      this->roadmap_->toGL(time);
    }
}

// // draw the trajectories generated for the internal force computation
// void MPC_TeddyIfc::drawGenTraj()
// {
//   if(trajType == PM_TRAJ)
//     {
//       // nothing to do
//     }
//   else if(this->traj_ != NULL && deformer_ != NULL && ((CLTrajDeform*)deformer_)->listCtrl_ != NULL)
//     {
//       std::list <Vec_DP*>::iterator it;
//       int i=1;
      
//       for(it = ((CLTrajDeform*)deformer_)->listCtrl_->begin(); i < this->traj_->getSize()-1 && it != ((CLTrajDeform*)deformer_)->listCtrl_->end(); it++)
// 	{
// 	  if(i != 0)
// 	    {
// 	      CLStateT prev = *((CLStateT*)(&((*(this->traj_))[i-1])));
// 	      CLStateT cur  = *((CLStateT*)(&((*(this->traj_))[i])));
// 	      CLStateT next = *((CLStateT*)(&((*(this->traj_))[i+1])));
	      
// 	      // get bach the i-th control
// 	      Vec_DP* ctrl = *it;
	      
// 	      if(ctrl != NULL)
// 		{
// 		  // compute the state reached by the random control
// 		  CLTrajGen carModel2((CarLikeSys*)sys_);
// 		  carModel2.L_ = ((CarLikeSys*)sys_)->getLength();

// 		  double param[7] = {(*ctrl)[0],(*ctrl)[1],(*ctrl)[2],(*ctrl)[3],(*ctrl)[4],(*ctrl)[5],(*ctrl)[6]};
// 		  carModel2.ComputeFromIG(prev,param);
		  
// 		  vector<vector<double> > vecPoints = carModel2.sampleXYT(prev.t());
		  
// 		  glColor3f(1.0,0.0,0.0);
// 		  glBegin(GL_LINE_STRIP);
// 		  for(int j=0; j<vecPoints.size(); j++)
// 		    {
// 		      glVertex3f(vecPoints[j][0],vecPoints[j][1],vecPoints[j][2]);
// 		    }
// 		  glEnd();
// 		}
// 	    }

// 	  i++;
// 	}
//     }
// }

// draw the trajectories generated for the TF process
void MPC_TeddyIfc::drawGenTrajTF()
{
  // check if the follower is available
  if(this->follower_ != NULL)
    {
      follower_->displayTrajGen();
    }
}

// draw the intermediate trajectory
void MPC_TeddyIfc::drawItdTraj()
{
  const GLfloat purpleMaterial[]={1.,0.,1.,1.};
  const GLfloat yellowMaterial[]={1.,1.,0.,1.};

  if(deformer_ == NULL) return;

  switch(trajType)
    {
    case PM_TRAJ : 
      break;
    case CL_TRAJ : 
      {
	break;

// 	int N = ((CLTrajDeform*)deformer_)->itdTraj_.size();
// 	for(int i=0; i<N; i++)
// 	  {
// 	    if(i%10 == 0)
// 	      {
// 		glMaterialfv(GL_FRONT, GL_AMBIENT, yellowMaterial);
// 		glMaterialfv(GL_FRONT, GL_DIFFUSE, yellowMaterial);
// 	      }
// 	    else
// 	      {
// 		glMaterialfv(GL_FRONT, GL_AMBIENT, purpleMaterial);
// 		glMaterialfv(GL_FRONT, GL_DIFFUSE, purpleMaterial);
// 	      }

// 	    CLStateT cstate = ((CLTrajDeform*)deformer_)->itdTraj_.at(i);
// 	    glPushMatrix();
// 	    glTranslated(cstate.x(),
// 			 cstate.y(),
// 			 cstate.t());
// 	    gluSphere(gluNewQuadric(), 1.0, 5, 5);
// 	    glPopMatrix();
// 	  }
      }
      break;
    case DD_TRAJ : 
      break;
    }

  const GLfloat cyanMaterial[]={0.,1.,1.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, cyanMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, cyanMaterial);
  
  switch(trajType)
    {
    case PM_TRAJ : 
      break;
    case CL_TRAJ : 
      {	
	break;

// 	int N = ((CLTrajDeform*)deformer_)->itdTraj2_.size();
// 	for(int i=0; i<N; i++)
// 	  {
// 	    CLStateT cstate = ((CLTrajDeform*)deformer_)->itdTraj2_.at(i);
// 	    glPushMatrix();
// 	    glTranslated(cstate.x(),
// 			 cstate.y(),
// 			 cstate.t());
// 	    gluSphere(gluNewQuadric(), 1.0, 5, 5);
// 	    glPopMatrix();
// 	  }
      }
      break;
    case DD_TRAJ : 
      break;
    }
}

void MPC_TeddyIfc::updDynamicView()
{
  //if(this->traj_ != NULL && this->path_ != NULL)
  if(this->traj_ != NULL)
    {
      // get back the dimensions of the map
      double dimMapX;
      double dimMapY;
      if(this->env_ == NULL)// || this->env_->getMap() == NULL)
	{
	  dimMapX = dfltXMap;
	  dimMapY = dfltYMap;
	}
//       else
// 	{
// 	  dimMapX = this->env_->getMap()->getWidth();
// 	  dimMapY = this->env_->getMap()->getHeight();
// 	}
      
      // init the viewer
//       PositionXY center(0.5*dimMapX, 0.5*dimMapY);
//       //double radius = path_->radius();
//       double radius = sqrt(0.25*dimMapX*dimMapX * 0.25*dimMapY*dimMapY);

      PositionXY center;
      double radius; 
      if(env_ != NULL)
	{
// 	  cout << "env_->getCenter() : "; env_->getCenter().display(); cout << endl;
// 	  cout << "env_->minX() : " << env_->minX() << "  env_->maxX() : " << env_->maxX() 
// 	       << "  env_->minY() : " << env_->minY() << "  env_->maxY() : " << env_->maxY() << endl;
// 	  cout << "dimMapX : " << dimMapX << "  dimMapY : " << dimMapY << endl;

	  //center = env_->getCenter();
 	  center = PositionXY(0.5*((double)(env_->minX()+env_->maxX())),0.5*((double)(env_->minY()+env_->maxY())));
 	  //center = PositionXY(0.5*dimMapX,0.5*dimMapY);
	  radius = env_->getRadius();
	}
      else if(path_ != NULL)
	{
	  center = path_->center();
	  radius = path_->radius();
	}
      else
	{
	  cerr << "MPC_TeddyIfc::updDynamicView : path and environment are not defined..." << endl;
	  exit(1);
	}
      double depth = 0.022 * radius;
      double dt = 3*(*traj_)[traj_->getSize()-1].posXYT().t();
      double upx,upy,upz;
      
      Vec v1(3);
      v1.setVal(0,0);
      v1.setVal(1,-dt);
      v1.setVal(2,-depth);
      double v1Norm = v1.norm();
      Vec v2(3);
      v2.setVal(0,0);
      v2.setVal(1,-dt/v1Norm);
      v2.setVal(2,-depth/v1Norm);
      upx = 0;
      upy = -v2[2];
      upz = v2[1];

//       cout << "center : (" << center.x() << "," << center.y() << ")  radius : " << radius << "  depth : " << depth << endl;
//       cout << "v1 : "; v1.display(); cout << "  v1.norm : " << v1.norm() << "  v2 : "; v2.display(); cout << endl;
//       cout << "dt : " << dt << "  upx : " << upx << "  upy : " << upy << "  upz : " << upz << endl;
      
      this->glcanvas->init(center.x(), center.y() /*+ dt*/, depth + (time * 10.0), center.x(), center.y(), (time * 10.0), upx, upy, upz);
    }
//   else cout << "rien!!!" << endl;
}

void MPC_TeddyIfc::setLookAt(Environment* env,Path* path,STraj* traj)
{
  if(path != NULL)
    {
      PositionXY center = path->center();
      double radius = path->radius();
      double depth = 5*radius;
      double dt;
      double upx,upy,upz;
      if(traj != NULL)
	{
	  dt = 3*(*traj)[traj->getSize()-1].posXYT().t();
	  
	  Vec v1(3);
	  v1.setVal(0,0);
	  v1.setVal(1,-dt);
	  v1.setVal(2,-depth);
	  double v1Norm = v1.norm();
	  Vec v2(3);
	  v2.setVal(0,0);
	  v2.setVal(1,-dt/v1Norm);
	  v2.setVal(2,-depth/v1Norm);
	  upx = 0;
	  upy = -v2[2];
	  upz = v2[1];
	}
      else
	{
	  dt = 0;
	  upx = 0;
	  upy = 1;
	  upz = 0;
	}
      gluLookAt(center.x(), center.y() + dt, depth, center.x(), center.y(), 0, upx, upy, upz);
    }
  else if(env != NULL)
    {
      cout << "je suis la!!!" << endl;

      PositionXY center = env->getCenter();
      double radius = env->getRadius();
      double depth = 5*radius;
      double dt;
      double upx,upy,upz;

      dt = 100;
	  
      Vec v1(3);
      v1.setVal(0,0);
      v1.setVal(1,-dt);
      v1.setVal(2,-depth);
      double v1Norm = v1.norm();
      Vec v2(3);
      v2.setVal(0,0);
      v2.setVal(1,-dt/v1Norm);
      v2.setVal(2,-depth/v1Norm);
      upx = 0;
      upy = -v2[2];
      upz = v2[1];
      
      gluLookAt(center.x(), center.y() + dt, depth, center.x(), center.y(), 0, upx, upy, upz);      
    }
  else
    {
      gluLookAt(5.,10.,15.,0.,0.,0.,0.,1.,0.);
    }
}

void MPC_TeddyIfc::drawScene()
{
  const GLfloat lightPosition[]={-20.,-100.,100.,1.};
  const GLfloat lightAmbient[]={.3f,.3f,.3f,1.};
  const GLfloat lightDiffuse[]={.9f,.9f,.9f,1.};
  const GLfloat redMaterial[]={1.,0.,0.,1.};
  const GLfloat blueMaterial[]={0.,0.,1.,1.};
  const GLfloat greenMaterial[]={0.,1.,0.,1.};
  const GLfloat otherMaterial[]={1.,0.,1.,1.};

  GLdouble width = glcanvas->getWidth();
  GLdouble height = glcanvas->getHeight();
  GLdouble aspect = height>0 ? width/height : 1.0;

  glcanvas->makeCurrent();

  glViewport(0,0,glcanvas->getWidth(),glcanvas->getHeight());

  glClearColor(1.0,1.0,1.0,1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable(GL_DEPTH_TEST);

  glDisable(GL_DITHER);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(this->path_ != NULL)
    {
      if(!orthoView) gluPerspective(30.,aspect,1.,-1);
      else glOrtho(-0.6*dfltXMap, 0.6*dfltXMap, -0.6*dfltYMap, 0.6*dfltYMap, -1000.0, 10000.0);
    }
  else
    {
      if(!orthoView) gluPerspective(30.,aspect,1.,-1);
      else glOrtho(-0.6*dfltXMap, 0.6*dfltXMap, -0.6*dfltYMap, 0.6*dfltYMap, -1000.0, 10000.0);
    }

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if(dynamicView)
    {
      this->updDynamicView();
    }

  gluLookAt(glcanvas->getEyeX(),glcanvas->getEyeY(),glcanvas->getEyeZ(),
	    glcanvas->getCenterX(),glcanvas->getCenterY(),glcanvas->getCenterZ(),
	    glcanvas->getUpX(),glcanvas->getUpY(),glcanvas->getUpZ());  

  glShadeModel(GL_SMOOTH);
  glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
  glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);

  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);

  glScalef(1.0,1.0,10.0);

//   switch(processType_)
//     {
//     case TEDDY_PROC : cout << "TEDDY_PROC" << endl; break;
//     case TF_PROC : cout << "TF_PROC" << endl; break;
//     case PATH_PLNG_PROC : cout << "PATH_PLNG_PROC" << endl; break;
//     case TRAJ_PLNG_PROC : cout << "TRAJ_PLNG_PROC" << endl; break;
//     default : cout << "ARGH..." << endl; break;
//     }

  if(processType_ == TEDDY_PROC)
    {
      // draw the trajectory
      this->drawTraj();
      // draw the robot
      this->drawRobot();
      // draw the intermediate trajectory
      this->drawItdTraj();
//   // draw the trajectories generated for internal force computation
//   if(drawTrajGen)this->drawGenTraj();
    }
  else if(processType_ == TF_PROC)
    {
      // draw the trajectory
      this->drawTrajTF();
      // draw the robot
      this->drawRobotTF();      
      if(drawTrajGenTF) this->drawGenTrajTF();
    }
  else if(processType_ == PATH_PLNG_PROC)
    {
      // draw the path
      this->drawPath();
      // draw the robot
      //this->drawRobot();
    }
  else if(processType_ == TRAJ_PLNG_PROC)
    {
      // draw the trajectory
      this->drawTraj();
      // draw the robot
      //this->drawRobot();
    }

//   // draw the obstacles
//   this->drawObstacles();
//   // draw the polygonal obstacles
//   this->drawPolygonObs();
//   // draw the map
//   this->drawMap();

  // draw the environment components
  this->drawEnv();

  // draw the roadmap
  if(roadmapView) this->drawRoadMap();

  // draw the forces 
  if(processType_ == TEDDY_PROC)
    {
      for(int i=0; i<lVec.size(); i = i+2)
	{
	  if(i%6 == 0 || i%6 == 1)
	    {
	      glColor3f(1.0f,0.0f,0.0f);
	      glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
	      glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
	    }
	  else if(i%6 == 2 || i%6 == 3)
	    {
	      glColor3f(0.0f,1.0f,0.0f);
	      glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
	      glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
	    }
	  else
	    {
	      glColor3f(1.0f,1.0f,0.0f);
	      glMaterialfv(GL_FRONT, GL_AMBIENT, otherMaterial);
	      glMaterialfv(GL_FRONT, GL_DIFFUSE, otherMaterial);
	    }
	  
	  glBegin(GL_LINES);
	  glVertex3f(lVec[i].x,lVec[i].y,lVec[i].z);
	  glVertex3f(lVec[i+1].x,lVec[i+1].y,lVec[i+1].z);
	  glEnd();
	}
    }

  // Swap if it is double-buffered
  if(glvisual->isDoubleBuffer()){
    glcanvas->swapBuffers();
  }
  
  // Make context non-current
  glcanvas->makeNonCurrent();
}

int MPC_TeddyIfc::elemListExists(const FXString& s)
{
  int i=0;
  bool trouve = false;

  while(!trouve && i<list[0]->getNumItems())
    {
      if(list[0]->getItemText(i).text() == s)
	trouve = true;
      else i++;
    }
  
  if(trouve) return i;
  else return -1;
}

void MPC_TeddyIfc::initRandomDeform()
{
  int seq;

  if(optSeqNum == 0)
    {
      seq = initRandom();
    }
  else
    {
      seq = initRandom(seqNum);
    }
  seqNum = (FXint) seq;

  // init the robotic system
  switch(trajType)
    {
    case PM_TRAJ : 
      initPointMassSys();
      break;
    case CL_TRAJ : 
      initCarLikeSys();
      break;
    case WC_TRAJ : 
      initWheelChairSys();
      break;
    default : 
      cout << "Unknown traj type" << endl;
      break;
    }
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////// Main ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Here we begin
int main(int argc,char *argv[]){

  // Make application
  FXApp application("STTrajectoryDeformation","STTrajectoryDeformation");

  // Open the display
  application.init(argc,argv);

  // Make window
  new MPC_TeddyIfc(&application);

  // Create the application's windows
  application.create();

  // Run the application
  return application.run();
}
